#ifndef __XAM_XASYNC__
#define __XAM_XASYNC__
////////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2007, The Storage Networking Industry Association.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - Neither the name of The Storage Networking Industry Association (SNIA)
//      nor the names of its contributors may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
// ============================================================================
//  Copyright (c) 2007, EMC Corporation.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - The name of EMC Corporation may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
////////////////////////////////////////////////////////////////////////////////
//  XAM SDK Source File Build Version @version.full@
////////////////////////////////////////////////////////////////////////////////

#include "xam_types.h"

#include "XAM_Mutex.h"
#include "XAM_Thread.h"

/** 
 *  Auxiliary object used to manage an asynchronous operation
 *
 *  May be used to spawn a service thread in the XAM library
 *  to execute the requested operation utilizing the synchronous
 *  VIM API method.
 *
 *  NOTE: Currently supports only async XStream operations, which is all the
 *        XAM Library requires.
 */
class XAM_XAsync
{

    private: 
        xam_status mStatus;             // Async object status
        XOPID mXOPID;                   // User provided unique identifier
        XAM_Thread* mThread;            // Pointer to the active thread object
        xasync_callback mCallback;      // Callback function pointer

        XAM_Mutex mAsyncMutex;               // Protects the notifyComplete operation
                                             // from being executed by both halt and complete.
        xam_boolean mNotifyCompleteFinished; // If true, the operation has been halted/completed.
        xam_boolean mHalted;                 // true if halted

        // Results
        xset_handle mXStreamResult;     // XStream result
        xam_int mBytes;                 // Bytes read/written result

        // Parameter cache
        xam_handle_t mHandle;           // Generic handle parameter
        xstream_handle mXStream;        // XStream input parameter
        char* mBuffer;                  // Pointer to a user buffer
        xam_int mBufferLen;             // Length of the user buffer
        xam_string mFieldName;          // Cache for a field name param
        xam_string mXStreamMode;        // XStream open mode parameter

        // Defines the list of supported asynchronous operations
        typedef enum {
            NONE,
            XAM_OPENXSTREAM,
            XSTREAM_READ,
            XSTREAM_WRITE,
            XSTREAM_CLOSE
        } XAM_AsyncMethod;

        XAM_AsyncMethod mMethod;        // The method to service asynchronously

        /**
         * Helper method for notifyComplete
         * @param inStatus The status of the async operation
         */
        void
        notifyCompleteHelper(xam_status inStatus);

    public:
        
        /**
         * Constructor
         *
         * @param inXOPID      Unique identifier specified by the application
         * @param inCallback   The user callback function pointer
         */
        XAM_XAsync(const XOPID inXOPID, 
                   xasync_callback inCallback);

        /**
         * Destructor
         */
        ~XAM_XAsync()
        {
            if(mThread)
                delete mThread;
        }

        /**
         * Determine if this operation has been halted by the caller
         *
         * @return true if halted, false otherwise
         */
        xam_boolean
        isHalted()
        {
            return mHalted;
        }

        /**
         * XAM_XAsync Operations
         *
         *    The following methods map exactly to the public synchronous 
         *    versions of the requested methods. 
         *
         *    An XAsync object may only operate on 1 method at a time.
         *
         *    If the asynchronous request was successfully spawned, true 
         *    is returned. On error (e.g. another asyncronous operation is 
         *    running), false is returned.
         */

        /**
         * Support for XAM_AsyncOpenXStream
         */
        bool
        openXStream(const xam_handle_t inHandle,
                    const xam_string inName,
                    const xam_string inMode);

        /**
         * Support for XStream_AsyncRead
         */
        bool
        read(const xstream_handle inHandle,
             char* ioBuffer,
             const xam_int inBufferLength);

        /**
         * Support for XStream_AsyncWrite
         */
        bool
        write(const xstream_handle inHandle,
              const char* inBuffer,
              const xam_int inByteCount);

        /**
         * Support for XStream_AsyncClose
         */
        bool
        close(const xstream_handle inHandleXStream);

        /**
         * XAM_XAsync Status Methods 
         *
         *    The following methods return status information from the XAsync. 
         */

        /**
         * Retrieve the status result of the asynchronous call
         */
        xam_status 
        getStatus() 
        { 
            return mStatus; 
        }

        /**
         * Determine whether the asynchronous operation has completed or not
         */
        xam_boolean 
        isComplete();
        
        /**
         * Retrieve the XOPID associated with this asynchronous operation
         */
        XOPID 
        getXOPID() 
        { 
            return mXOPID; 
        }
        
        /**
         * Halt this asynchronous operation
         */
        xam_status
        halt();

        /**
         * XAM_XAsync Result Methods 
         *
         *    The following methods return result information from the XAsync. 
         */

        /**
         * Retrieve the number of bytes read
         */
        xam_int 
        getBytesRead();

        /**
         * Retrieve the number of bytes written
         */
        xam_int 
        getBytesWritten();

        /**
         * Retrieve the resulting xstream_handle
         */
        xstream_handle 
        getXStream();

        /**
         *  Run the synchronous version of the method requested 
         *  by the caller. Used by the spawned thread.
         */
        xam_status 
        execSynchronousCall();

        /**
         *  Notifies the user of the complete state by calling any
         *  registered callback method. Used by the spawned thread.
         *
         * @param inStatus The result of the async call
         * @param inObtainLock If true, a lock on the the mAsyncMutex is obtained, 
         *                     false otherwise
         */
        void 
        notifyComplete(xam_status inStatus, xam_boolean inObtainLock = true);

        /**
         * Sets the running state of the active thread (if any)
         *
         * @param inState True if the thread is running, false otherwise.
         */
        void
        setRunning(xam_boolean inState, xam_boolean inObtainLock = true);
};


#endif // __XAM_XASYNC__

