////////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2007, The Storage Networking Industry Association.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - Neither the name of The Storage Networking Industry Association (SNIA)
//      nor the names of its contributors may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
// ============================================================================
//  Copyright (c) 2007, EMC Corporation.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - The name of EMC Corporation may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
////////////////////////////////////////////////////////////////////////////////
//  XAM SDK Source File Build Version @version.full@
////////////////////////////////////////////////////////////////////////////////

#include "XAM_Field.h"
#include "XAM_Exception.h"
#include "XAM_POSIX.h"

const char* 
XAM_Field::getMimeType() 
{ 
    return mMimeType; 
}

const xam_int 
XAM_Field::getSize() 
{ 
    return mDataSizeInBytes;
}

void
XAM_Field::reset()
{
    if(mReadOnly)    
    {
        XAMReadOnlyFieldException e;
        throw e;
    }

    // Free any allocated heap buffer
    if(mData != (char*)0)
    {
        XAM_POSIX::xam_free(mData);
        mData = (char*)0;
    }

    // Delete any temp file used
    if(mTmpFile != (XAM_TempFile*)0)
    {
        delete mTmpFile;
        mTmpFile = (XAM_TempFile*)0;
    }

    mDataSizeInBytes = 0;
}

const xam_int 
XAM_Field::write(const char* inData, 
                 const xam_int inDataOffset, 
                 const xam_int inDataSizeInBytes)
{
    xam_int vBytesWritten = 0;
    xam_int vExpectedTotalDataSize = inDataSizeInBytes;

    // Adjust the total expected data size for any write offset
    if(vExpectedTotalDataSize > (mDataSizeInBytes - inDataOffset))
        vExpectedTotalDataSize = inDataSizeInBytes + inDataOffset;

    // Use a memory buffer 
    // (offload to temp file if the threshold size is exceeded).

    if(mReadOnly)
    {
        XAMReadOnlyFieldException e;
        throw e;
    }

    // Decide whether to use a temp file or memory buffer
    if((mTmpFile == (XAM_TempFile*)0) && 
       (vExpectedTotalDataSize > mTmpFileTriggerSz))
    {
        // Offload data to the temp file
        mTmpFile = new XAM_TempFile();
        if(!mTmpFile)
        {
            XAMOutOfMemoryException e;
            throw e;
        }
        else
        {
            if(mData)
            {
                if(mDataSizeInBytes > 0)
                    mTmpFile->write(mData, 0, mDataSizeInBytes);
                XAM_POSIX::xam_free(mData);
                mData = (char*) 0;
            }
        }     
    }

    // If this field has been offloaded to a temp file, 
    // write the data there.
    if(mTmpFile)
    {
        vBytesWritten = mTmpFile->write(inData, 
                                        inDataOffset, 
                                        inDataSizeInBytes);
    }
    else
    {
        if(vExpectedTotalDataSize > 0)
        {
            // This field is maintained in a memory buffer. Reallocate the buffer 
            // to accomodate the new size.
            mData = (char*)XAM_POSIX::xam_realloc(mData, 
                                                  vExpectedTotalDataSize);
            if(mData == (void*)0)
            {
                XAMOutOfMemoryException e;
                throw e;
            }

            // Copy the data into the heap buffer
            if(inData)
            {
                XAM_POSIX::xam_memcpy(mData + inDataOffset, 
                                      (char*)inData, 
                                      inDataSizeInBytes);

                vBytesWritten = inDataSizeInBytes;
            }
        }
        else
        {
            // No data is expected, so ensure that.
            if(mData != (char*)0)
                XAM_POSIX::xam_free(mData);
            mData = (char*)0;
        }
    }

    // Adjust the actual data size based on bytes written
    if(mDataSizeInBytes < (inDataOffset + vBytesWritten))
        mDataSizeInBytes = (inDataOffset + vBytesWritten);

    return vBytesWritten;
}

const xam_int 
XAM_Field::read(char* const outData, 
                const xam_int inDataOffset, 
                const xam_int inDataSizeInBytes)
{
    xam_int vReadDataSize = inDataSizeInBytes;

    // Cannot read data that does not exist. Adjust for this case.
    if(vReadDataSize > (mDataSizeInBytes - inDataOffset))
        vReadDataSize = mDataSizeInBytes;

    if(outData)
    {
        if(mTmpFile)
        {
            // This field is maintained in a temp file. 
            // Read data from this file.
            vReadDataSize = mTmpFile->read(outData, 
                                           inDataOffset, 
                                           vReadDataSize);       
        }
        else
        {
            // This field is maintained in a memory buffer. 
            // Read data from this buffer.
            XAM_POSIX::xam_memcpy(outData, 
                                  mData + inDataOffset, 
                                  vReadDataSize);
        }
    }

    return vReadDataSize;
}

XAM_Field::XAM_Field(const char* inName, 
                     const char* inMimeType, 
                     const bool inBinding,
                     const bool inReadOnly,
                     const char* inData, 
                     const xam_int inDataSizeInBytes)   

 : mTmpFile(0), 
   mTmpFileTriggerSz(10240), // 10 MB threshold for offloading data to tmp file
   mBinding(inBinding), 
   mDataSizeInBytes(0),
   mIsOpen(false),
   mData((char*)0)
{
    // Maintain a copy of the field name
    XAM_POSIX::xam_strcpy(mName, inName); 

    // Maintain a copy of the mime type
    XAM_POSIX::xam_strcpy(mMimeType, inMimeType);

    // Write the data, then apply read-only status to the field
    if(inData != 0)
    {
        mReadOnly = false;
        if(inDataSizeInBytes > 0)
            write(inData, 0, inDataSizeInBytes);
    }
    mReadOnly = inReadOnly;
}


XAM_Field::~XAM_Field()
{
    reset();
}
