#ifndef __XAM_FIELD_H__
#define __XAM_FIELD_H__
////////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2007, The Storage Networking Industry Association.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - Neither the name of The Storage Networking Industry Association (SNIA)
//      nor the names of its contributors may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
// ============================================================================
//  Copyright (c) 2007, EMC Corporation.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - The name of EMC Corporation may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
////////////////////////////////////////////////////////////////////////////////
//  XAM SDK Source File Build Version @version.full@
////////////////////////////////////////////////////////////////////////////////

#include "xam_types.h"

#include "XAM_Critical.h"
#include "XAM_TempFile.h"

// Defines all field type options
typedef enum 
{
    PROPERTY_FIELD,
    XSTREAM_FIELD
} XAM_FieldType;

/** 
 *  An abstract class representing a XAM field.
 *  All operations on this object must be protected for thread-safety.
 *
 *  Uses a memory buffer to store the field data. If the data grows 
 *  to a size large enough to exceed a configurable limit, the data 
 *  will be offloaded to a temp file. The default trigger size is 
 *  10 MB (10240 KB).
 */
class XAM_Field
{
    /* Only the XAM_FieldContainer class is allowed to 
     * access this object.
     */
    friend class XAM_FieldContainer;

    protected:
        xam_string   mName;        // Unique field name
        xam_string   mMimeType;    // Mime type (ie. "application/vnd.snia.xam.string) 
        xam_boolean  mBinding;     // Binding (true) or Nonbinding (false) content
        xam_boolean  mReadOnly;    // Read only (true) or not (false)
        xam_boolean  mIsOpen;      // The open state

        char* mData;               // Pointer to allocated field data
        xam_int mDataSizeInBytes;  // Number of allocated bytes in mData

        xam_int mTmpFileTriggerSz; // The data size (in KB) which triggers use 
                                   // of a tmp file
        XAM_TempFile* mTmpFile;    // Pointer to a temporary file 
                                   // (used instead of mData if mDataSizeInBytes 
                                   // grows larger than mTmpFileTriggerSz!)

        /**
         * Protected Constructor : only allow derived implementations
         *
         * By default, this object will store and maintain a data field in memory.
         *
         * @param inName             Field name
         * @param inMimeType         Field mime type
         * @param inBinding          Field binding state
         * @param inReadOnly         Field read-only state
         * @param inData             Pointer to field data to copy into the field
         * @param inDataSizeInBytes  Size of the data in the inData buffer
         *
         */
        XAM_Field(const char* inName, 
                  const char* inMimeType, 
                  const bool inBinding,
                  const bool inReadOnly,
                  const char* inData, 
                  const xam_int inDataSizeInBytes);

        /** 
         * isOpen
         *
         * @return true if currently open, false otherwise
         */
        virtual const xam_boolean
        isOpen()
        {
            return mIsOpen;
        }

        /**
         * Sets the non-binding state
         * 
         * @param inVal binding (true) or non-binding (false)
         */
        void 
        setBinding(const xam_boolean inVal) 
        {   
            mBinding = inVal; 
        }

        /**
         * Sets the read only state
         * 
         * @param inVal read only (true) or read/write (false)
         */
        void 
        setReadOnly(const xam_boolean inVal) 
        { 
            mReadOnly = inVal; 
        }

        /**
         * Get the name of the field
         */
        const char* 
        getName() 
        { 
            return mName; 
        }

        /**
         * Get the mime type of this field
         */
        const char* 
        getMimeType();

        /**
         * Get the binding state of this field
         *
         * @return true (binding) or false (non-binding)
         */
        const xam_boolean 
        getBinding() 
        { 
            return mBinding; 
        }

        /**
         * Get the read only state of this field
         *
         * @return true (read only) or false (read/write)
         */
        const xam_boolean 
        getReadOnly() 
        { 
            return mReadOnly; 
        }

        /**
         * Retrieve the size of the data referenced by this field
         */
        virtual const xam_int 
        getSize();

        /**
         * Get the size threshold for offloading field data to tmp file
         *
         * @return The number of bytes (in KB) the field data is allowed 
         *         to grow in memory before being offloaded to a tmp file.
         */
        const xam_int 
        getTmpFileTriggerSize() 
        { 
            return mTmpFileTriggerSz; 
        }

        /**
         * Set the size threshold for offloading field data to tmp file
         *
         * @param inSizeInKb : The number of bytes (in KB) the field data 
         *                     is allowed to grow in memory before being 
         *                     offloaded to a tmp file.
         */
        void 
        setTmpFileTriggerSize(const xam_int inSizeInKb) 
        { 
            mTmpFileTriggerSz = inSizeInKb; 
        }

        /**
         * Returns the XAM_FieldType (ie. PROPERTY_FIELD/XSTREAM_FIELD) 
         */
        virtual const XAM_FieldType 
        getFieldType() = 0;

        /**
         * Deallocate any data in memory and reset data fields
         */
        virtual void 
        reset();

        /**
         * Write data to this field
         *
         * @param inData             Pointer to the buffer of data being 
         *                           written
         * @param inDataOffset       The offset position into the local 
         *                           buffer to start the write to
         * @param inDataSizeInBytes  The number of bytes to be written
         *
         * @return   The number of bytes successfully 
         *           written to the stream (will be <= inDataSizeInBytes)
         */
        const xam_int 
        write(const char* inData, 
              const xam_int inDataOffset, 
              const xam_int inDataSizeInBytes);

        /**
         * Read data from this field
         *
         * @param outData            Pointer to allocated space where the data
         *                           will be written
         * @param inDataOffset       The offset position into the local buffer 
         *                           to start the read from
         * @param inDataSizeInBytes  The number of bytes available in the 
         *                           allocated space
         *
         * @return  The actual number of bytes written to outData 
         *          (will be <= inDataSizeInBytes)
         */
        const xam_int 
        read(char* const outData, 
             const xam_int inDataOffset, 
             const xam_int inDataSizeInBytes); 

    public:

        /**
         * Destructor
         */
        virtual 
        ~XAM_Field();

    private:
        /* Private copy constructor - disallow copies */
        XAM_Field( XAM_Field& ) {}

};

#endif // __XAM_FIELD_H__
