#ifndef __XAM_FIELD_NAME_CACHE_H__
#define __XAM_FIELD_NAME_CACHE_H__
////////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2007, The Storage Networking Industry Association.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - Neither the name of The Storage Networking Industry Association (SNIA)
//      nor the names of its contributors may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
// ============================================================================
//  Copyright (c) 2007, EMC Corporation.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - The name of EMC Corporation may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
////////////////////////////////////////////////////////////////////////////////
//  XAM SDK Source File Build Version @version.full@
////////////////////////////////////////////////////////////////////////////////

#include "xam_types.h"
#include "XAM_FieldContainer.h"

// The threshold at which the cache is reallocated from the heap
// (also the amount of additional space reallocated at each resize event)
static const int CACHE_REALLOC_THRESHOLD = 50;

/**
 * A class for caching a list of field names from a 
 * XAM_FieldContainer for the purpose of thread safe 
 * field name iteration. This class filters all cached 
 * name additions based on a user defined pattern filter.
 */
class XAM_FieldNameCache
{   
    private:
        xam_string* mNames;        // Pointer to the array of cached names
        int mCacheSize;            // The number of allocated slots
        int mCacheSizeAllocCount;  // The number of allocations/reallocations
        int mSize;                 // The number of slots filled

        const char* mPattern;      // A pointer to the required prefix pattern 
        xam_int mPatternLen;       // The length of the required prefix pattern

        const bool 
        patternMatches(const char* inName)
        {
            bool vMatchFound = false;

            // The pattern matches if the "inName" string is found to match the 
            // beginning of the field name.
            if(mPatternLen > 0)
            {
                if(XAM_POSIX::xam_strncmp(inName, mPattern, mPatternLen) == 0)
                    vMatchFound = true;
            }
            else
            {
                // For empty string patterns, everything is a match
                vMatchFound = true;
            }

            return vMatchFound;
        }

    public:    
        
        /**
         * Constructor
         *
         * @param inPattern  A pointer to the pattern filter string 
         *                   (may be empty string for no filtering or non-empty to
         *                    filter all names which do not match the pattern)
         * @param inPatternLen The length (in bytes) of the pattern string 
         */
        XAM_FieldNameCache(const char* inPattern, xam_int inPatternLen)
            : mPattern(inPattern),
              mPatternLen(inPatternLen),
              mCacheSize(CACHE_REALLOC_THRESHOLD), 
              mCacheSizeAllocCount(1), 
              mSize(0)
        {
            // Allocate heap block of mCacheSize
            // (this avoids any extraneous heap fragmenting)
            mNames = (xam_string*) XAM_POSIX::xam_malloc(mCacheSize * 
                                                         sizeof(xam_string));
            if(mNames == (xam_string*)0)
            {
                XAMOutOfMemoryException e;
                throw e;
            }
        }

        /**
         * Destructor
         */
        ~XAM_FieldNameCache()
        {
            // delete the array of names
            if(mNames)
                XAM_POSIX::xam_free(mNames);
        }

        /**
         * Determines whether or not a field name exists in the cache
         *
         * @param inName The field name to look for in the cache
         * @return true if found, false otherwise
         */
        xam_boolean
        exists(const char* inName)
        {
            xam_boolean vFound = false;

            for(int i=0 ; i < getSize() ; i++)
            {
                const char* vName = (const char*) getNameAtIndex(i);
                if(XAM_POSIX::xam_strcmp(inName, vName) == 0)
                {
                    vFound = true;
                    break;
                }
            }

            return vFound;
        }

        /**
         * Add a field name to the cache
         *
         * @param inName The field name to cache
         */
        void
        addField(const char* inName)
        {
            if((inName != (const char*) 0) && patternMatches(inName))
            {
                mSize++;

                // Determine if reallocation of the heap block is needed
                if(mSize > mCacheSize)
                {
                    mCacheSizeAllocCount++;
                    mCacheSize = CACHE_REALLOC_THRESHOLD * mCacheSizeAllocCount;

                    // Reallocate heap block for the new cache size
                    // (this avoids any extraneous heap fragmenting)
                    mNames = (xam_string*) 
                        XAM_POSIX::xam_realloc(mNames,
                                               mCacheSize * sizeof(xam_string));

                    if(mNames == (xam_string*)0)
                    {
                        XAMOutOfMemoryException e;
                        throw e;
                    }
                }

                XAM_POSIX::xam_strcpy(mNames[mSize - 1], inName);
            }
        }

        /**
         * Get the number of names in the cache
         *
         * @return The number of names in the cache
         */
        const int
        getSize()
        {
            return mSize;
        }

        /**
         * Get the next name in the cache
         *
         * @param inIndex The index to retrieve. 
         *                (Must be >= 0 and < return from getSize())
         * @return The next name in the cache or NULL if complete
         */
        const xam_string*
        getNameAtIndex(const int inIndex)
        {
            return &mNames[inIndex];
        }
};


#endif // __XAM_FIELD_NAME_CACHE_H__


