#ifndef __XAM_XSTREAM_H__
#define __XAM_XSTREAM_H__
////////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2007, The Storage Networking Industry Association.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - Neither the name of The Storage Networking Industry Association (SNIA)
//      nor the names of its contributors may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
// ============================================================================
//  Copyright (c) 2007, EMC Corporation.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - The name of EMC Corporation may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
////////////////////////////////////////////////////////////////////////////////
//  XAM SDK Source File Build Version @version.full@
////////////////////////////////////////////////////////////////////////////////

#include "xam_types.h"
#include "XAM_Field.h"
#include "XAM_Mutex.h"

// Defines the start positions available for a seek operation
typedef enum
{
    FROM_START = 0,
    FROM_CURRENT,
    FROM_END
} XStream_Whence;

/** 
 *  Represents an XStream, a XAM_Field whose value is a stream
 *  of bytes. An XStream's MIME type attribute is any defined 
 *  type except for the XAM simple defined types (stypes).
 */
class XAM_XStream : public XAM_Field
{
    /* Only the XAM_FieldContainer class is allowed to 
     * open/close or create this object.
     */
    friend class XAM_FieldContainer;

    /*
     * The XStream instance is allowed to modify this field.
     */
    friend class XAM_XStream_Instance;

    protected:
        XAM_Mutex mWriteLock;            // The lock protecting a write operation
        xam_boolean mWriteLockAcquired;  // True if the write lock was acquired

        xam_boolean mCorrupted;  // If true the xstream is corrupted
        xam_int mOpenInstances;  // The open XStream Instance count

        /**
         * Protected Constructor  
         *
         * This object will store and maintain a data stream in memory.
         *
         * @param inName      Field name
         * @param inMimeType  Field mime type
         * @param inBinding   Field binding state
         * @param inReadOnly  Field read-only state
         *
         */
        XAM_XStream(const char* inName, 
                    const char* inMimeType, 
                    const bool inBinding,
                    const bool inReadOnly);   

        /**
         * Open the stream in the given read/write mode
         *
         * @param inMode  Must be XSTREAM_MODE_READ_ONLY, 
         *                        XSTREAM_MODE_WRITEONLY_TRUNCATE, 
         *                     or XSTREAM_MODE_WRITEONLY_APPEND.
         * @param ioHandle The XStream instance handle is written here
         */
        void 
        open(const char* inMode, xstream_handle* const ioHandle);

        /** 
         * Determine if there are open instances on this XStream field
         *
         * @return true if currently open, false otherwise
         */
        const xam_boolean
        isOpen()
        {
            return (mOpenInstances > 0);
        }

        /*
         * Acquire the write lock
         */
        void
        holdWriteLock();

        /*
         * Release the write lock (if previously grabbed)
         */
        void 
        releaseWriteLock();

        /*
         * Throws corresponding exception if the XStream is corrupt
         */
        void
        assertNotCorrupted();

    public:

        /**
         * Clear the stream of all data
         */
        void 
        reset();

        /**
         * getFieldType
         *
         * @return a string indicating the type of this field 
         *         (i.e. an "XSTREAM" field)
         */
        virtual const XAM_FieldType 
        getFieldType() 
        { 
            return XSTREAM_FIELD; 
        }

        /**
         * Get the name of the XStream field
         *
         * @return The name of the field object associated with this stream
         */
        const char* 
        getName() 
        { 
            return getName(); 
        }

        /**
         * Destructor
         */
        virtual 
        ~XAM_XStream() {};
};

#endif // __XAM_XSTREAM_H__