////////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2007, The Storage Networking Industry Association.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - Neither the name of The Storage Networking Industry Association (SNIA)
//      nor the names of its contributors may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
// ============================================================================
//  Copyright (c) 2007, EMC Corporation.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - The name of EMC Corporation may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
////////////////////////////////////////////////////////////////////////////////
//  XAM SDK Source File Build Version @version.full@
////////////////////////////////////////////////////////////////////////////////

#include "xam_types.h"
#include "xam_strings.h"
#include "xam_errors.h"

#include "XAM_Router.h"
#include "XAM_Status.h"
#include "XAM_Critical.h"
#include "XAM_Utils.h"
#include "XAM_Exception.h"
#include "XAM_Object.h"
#include "XAM_XAsync.h"
#include "XAM_VIM.h"
#include "XAM_POSIX.h"
#include "XAM_Logger.h"

#include "XAsync.h"
#include "XSystem.h"
#include "XStream.h"
#include "XIterator.h"
#include "XSet.h"


// =============================================================================
// methods for evaluating xam_status 
// =============================================================================

xam_boolean
XAM_Router::XAM_GetErrorToken(const xam_handle_t inHandle, 
                              const xam_status inStatusCode, 
                              xam_string* const outToken)
{
    xam_boolean vSuccess = false;

    try
    {
        // Check the required token pointer
        if(outToken == (xam_string*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        if(XAM_Utils::isStandardErrCode(inStatusCode))
        {
            // Standard error tokens can be retreived from 
            // the XAM_Status object
            vSuccess = XAM_Status::getToken(inStatusCode, outToken);
        }
        else
        {
            // Non-standard tokens must be retrieved from the VIM 
            // via the handle provided
            XAM_Handle* vHandle = getHandle(inHandle);   
            vHandle->assertType(XSYSTEM_HANDLE);
            vSuccess = vHandle->getErrorToken(inStatusCode, outToken);
        }
    }
    catch(XAM_Exception &)
    {
        // Retrieve the token for "unknown error"
        XAM_Status::getToken(XAM_UNKNOWN_ERROR, outToken);
    }

    return vSuccess;
}

// =============================================================================
// Methods for the XIterator 
// =============================================================================

xam_status
XAM_Router::XAM_OpenFieldIterator (const xam_handle_t inHandle,  
                                   const xam_string inPattern,
                                   xiterator_handle* const outIterator)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the pattern is in UTF-8
        XAM_Utils::validateUTF8(inPattern);

        // Check the required handle pointer
        if(outIterator == (xiterator_handle*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Create an XIterator and retrieve an xiterator_handle
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->openFieldIterator(inPattern, outIterator);

        // Wrap the retrieved handle with a scoped XAM_Handle type 
        // before returning to caller
        XIterator* vXIter = createXIteratorHandle(*outIterator, vHandle);
        *outIterator = (xiterator_handle) vXIter;
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();      
    }

    return vStatus;
} 

xam_status 
XAM_Router::XIterator_HasNext (const xiterator_handle inHandle, 
                               xam_boolean* const outValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {      
        // Check the required xam_boolean pointer
        if(outValue == (xam_boolean*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Determine if there are more elements in this XIterator
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->hasNext(outValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XIterator_Next (const xiterator_handle inHandle, 
                            xam_string* const outName)      
{
    xam_status vStatus = (xam_status)0;

    try
    {      
        // Check the required name pointer
        if(outName == (xam_string*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Retrieve the next element from the iterator 
        // (empty string indicates completion)
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->next(outName);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 
  
xam_status 
XAM_Router::XIterator_Close (xiterator_handle inHandle)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Close the iterator (assert the type first, since the close 
        // method is generic to all types)
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->assertType(XITERATOR_HANDLE);
        vHandle->close();

        // Deallocate the XAM_Handle wrapper
        deleteHandle(vHandle);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

// =============================================================================
// Methods for managing XAM Fields (properties or XStreams)
// =============================================================================

xam_status 
XAM_Router::XAM_ContainsField (const xam_handle_t inHandle,
                             const xam_string inName,
                             xam_boolean* const outExists)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Check the required boolean pointer
        if(outExists == (xam_boolean*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Retrieve the binding state of this field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->containsField(inName, outExists);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status 
XAM_Router::XAM_SetFieldAsBinding (const xset_handle inHandle,
                                   const xam_string inName)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the pattern is in UTF-8 
        XAM_Utils::validateUTF8(inName);
       
        // Apply the binding state to this field
        XAM_Handle* vHandle = getHandle(inHandle); 

        // Only XSet handles are allowed
        vHandle->assertType(XSET_HANDLE);

        // Apply the binding state
        vHandle->setFieldAsBinding(inName);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_SetFieldAsNonbinding (const xset_handle inHandle,
                                      const xam_string inName)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Apply the non-binding state to this field
        XAM_Handle* vHandle = getHandle(inHandle); 

        // Only XSet handles are allowed
        vHandle->assertType(XSET_HANDLE);

        // Apply the nonbinding state
        vHandle->setFieldAsNonbinding(inName);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_GetFieldType (const xam_handle_t inHandle,
                              const xam_string inName,
                              xam_string* const outType)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Check the required string pointer
        if(outType == (xam_string*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Retrieve the mime type for this field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getFieldType(inName, outType);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_GetFieldLength (const xam_handle_t inHandle,
                                const xam_string inName,
                                xam_int* const outLength)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Check the required length pointer
        if(outLength == (xam_int*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Retrieve the length of this field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getFieldLength(inName, outLength);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_GetFieldBinding (const xam_handle_t inHandle,
                                 const xam_string inName,
                                 xam_boolean* const outBinding)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Check the required boolean pointer
        if(outBinding == (xam_boolean*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Retrieve the binding state of this field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getFieldBinding(inName, outBinding);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_GetFieldReadOnly (const xam_handle_t inHandle,
                                  const xam_string inName,
                                  xam_boolean* const outReadOnly)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Check the required boolean pointer
        if(outReadOnly == (xam_boolean*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Retrieve the read only state of this field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getFieldReadOnly(inName, outReadOnly);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_DeleteField (const xam_handle_t inHandle,
                             const xam_string inName)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Delete this field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->deleteField(inName);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

// =============================================================================
// Methods for managing property fields
// =============================================================================

xam_status 
XAM_Router::XAM_CreateBoolean (const xam_handle_t inHandle,
                               const xam_string inName,
                               const xam_boolean inBinding,
                               const xam_boolean inValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Prevent field creation in the protected namespace
        // or empty string field names.
        if((XAM_POSIX::xam_strncmp(inName, ".", 1) == 0) ||
           (XAM_POSIX::xam_strcmp(inName, "") == 0))
        {
            XAMInvalidFieldNameException e;
            throw e;
        }

        // Create a new boolean field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->createBoolean(inName, inBinding, inValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_CreateInt (const xam_handle_t inHandle,
                           const xam_string inName,
                           const xam_boolean inBinding,
                           const xam_int inValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Prevent field creation in the protected namespace
        // or empty string field names.
        if((XAM_POSIX::xam_strncmp(inName, ".", 1) == 0) ||
           (XAM_POSIX::xam_strcmp(inName, "") == 0))
        {
            XAMInvalidFieldNameException e;
            throw e;
        }

        // Create a new int field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->createInt(inName, inBinding, inValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_CreateDouble (const xam_handle_t inHandle,
                             const xam_string inName,
                             const xam_boolean inBinding,
                             const xam_double inValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Prevent field creation in the protected namespace
        // or empty string field names.
        if((XAM_POSIX::xam_strncmp(inName, ".", 1) == 0) ||
           (XAM_POSIX::xam_strcmp(inName, "") == 0))
        {
            XAMInvalidFieldNameException e;
            throw e;
        }

        // Create a new float field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->createDouble(inName, inBinding, inValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_CreateXUID (const xam_handle_t inHandle,
                            const xam_string inName,
                            const xam_boolean inBinding,
                            const xam_xuid inValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Prevent field creation in the protected namespace
        // or empty string field names.
        if((XAM_POSIX::xam_strncmp(inName, ".", 1) == 0) ||
           (XAM_POSIX::xam_strcmp(inName, "") == 0))
        {
            XAMInvalidFieldNameException e;
            throw e;
        }

        // Create a new xam_xuid field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->createXUID(inName, inBinding, inValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_CreateString (const xam_handle_t inHandle,
                              const xam_string inName,
                              const xam_boolean inBinding,
                              const xam_string inValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Prevent field creation in the protected namespace
        // or empty string field names.
        if((XAM_POSIX::xam_strncmp(inName, ".", 1) == 0) ||
           (XAM_POSIX::xam_strcmp(inName, "") == 0))
        {
            XAMInvalidFieldNameException e;
            throw e;
        }

        // Create a new string field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->createString(inName, inBinding, inValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_CreateDatetime (const xam_handle_t inHandle,
                                const xam_string inName,
                                const xam_boolean inBinding,
                                const xam_datetime inValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Prevent field creation in the protected namespace
        // or empty string field names.
        if((XAM_POSIX::xam_strncmp(inName, ".", 1) == 0) ||
           (XAM_POSIX::xam_strcmp(inName, "") == 0))
        {
            XAMInvalidFieldNameException e;
            throw e;
        }

        // Create a new datetime field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->createDatetime(inName, inBinding, inValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_SetBoolean (const xam_handle_t inHandle,
                            const xam_string inName,
                            const xam_boolean inValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Set the value of this boolean field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->setBoolean(inName, inValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_SetInt (const xam_handle_t inHandle,
                        const xam_string inName,
                        const xam_int inValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Set the value of this int field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->setInt(inName, inValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_SetDouble (const xam_handle_t inHandle,
                          const xam_string inName,
                          const xam_double inValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Set the value of this float field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->setDouble(inName, inValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_SetXUID (const xam_handle_t inHandle,
                         const xam_string inName,
                         const xam_xuid inValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Set the value of this xam_xuid field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->setXUID(inName, inValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_SetString (const xam_handle_t inHandle,
                           const xam_string inName,
                           const xam_string inValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Set the value of this string field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->setString(inName, inValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_SetDatetime (const xam_handle_t inHandle,
                             const xam_string inName,
                             const xam_datetime inValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Set the value of this datetime field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->setDatetime(inName, inValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_GetBoolean (const xam_handle_t inHandle,
                            const xam_string inName,
                            xam_boolean* const outValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Check the required boolean pointer
        if(outValue == (xam_boolean*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the value of this boolean field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getBoolean(inName, outValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_GetInt (const xam_handle_t inHandle,
                        const xam_string inName,
                        xam_int* const outValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Check the required int pointer
        if(outValue == (xam_int*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the value of this int field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getInt(inName, outValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_GetDouble (const xam_handle_t inHandle,
                          const xam_string inName,
                          xam_double* const outValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Check the required float pointer
        if(outValue == (xam_double*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the value of this float field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getDouble(inName, outValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_GetXUID (const xam_handle_t inHandle,
                         const xam_string inName,
                         xam_xuid* const outValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Check the required xam_xuid pointer
        if(outValue == (xam_xuid*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the value of this xam_xuid field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getXUID(inName, outValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_GetString (const xam_handle_t inHandle,
                           const xam_string inName,
                           xam_string* const outValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Check the required string pointer
        if(outValue == (xam_string*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the value of this string field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getString(inName, outValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_GetDatetime (const xam_handle_t inHandle,
                             const xam_string inName,
                             xam_datetime* const outValue)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Check the required datetime pointer
        if(outValue == (xam_datetime*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the value of this datetime field
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getDatetime(inName, outValue);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

// ===========================================================================
// Methods for XStreams and XStream fields
// ===========================================================================

xam_status 
XAM_Router::XAM_CreateXStream (const xam_handle_t inHandle,
                               const xam_string inName,
                               const xam_boolean inBinding,
                               const xam_string inType,
                               xstream_handle* const outXStream)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the name is in UTF-8
        XAM_Utils::validateUTF8(inName);

        // Prevent field creation in the protected namespace
        if(XAM_POSIX::xam_strncmp(inName, ".", 1) == 0)
        {
            XAMInvalidFieldNameException e;
            throw e;
        }

        // Create a new XStream and retrieve an xstream_handle 
        // (opened for XSTREAM_MODE_WRITEONLY_TRUNCATE)
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->createXStream(inName, inBinding, inType, outXStream);

        // Wrap the retrieved handle with a scoped XAM_Handle type 
        // before returning to caller
        XStream* vXStream = createXStreamHandle(*outXStream, vHandle);
        *outXStream = (xstream_handle) vXStream;
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_OpenXStream (const xam_handle_t inHandle,
                             const xam_string inName,
                             const xam_string inMode,
                             xstream_handle* const outXStream)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that both the name and mode are in UTF-8
        XAM_Utils::validateUTF8(inName);
        XAM_Utils::validateUTF8(inMode);

        // Ensure that the mode string passed is valid
        XAM_Utils::validateMode(inMode);

        // Check the required handle pointer
        if(outXStream == (xstream_handle*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Open an existing XStream and retrieve an xstream_handle
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->openXStream(inName, inMode, outXStream);

        // Wrap the retrieved handle with a scoped XAM_Handle type 
        // before returning to caller
        XStream* vXStream = createXStreamHandle(*outXStream, vHandle);
        *outXStream = (xstream_handle) vXStream;
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 


xam_status 
XAM_Router::XStream_Read (const xstream_handle inHandle,
                          char* const ioBuffer,
                          const xam_int inBufferLength,
                          xam_int* const outBytesRead)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check that params are within bounds
        if((outBytesRead == (xam_int*)0) || 
           (ioBuffer == (const char*)0) ||
           (inBufferLength <= 0))
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Request a read of inBufferLength bytes into inBuffer 
        // (return the number of bytes written)
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->read(ioBuffer, inBufferLength, outBytesRead);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XStream_Write (const xstream_handle inHandle,
                           const char* inBuffer,
                           const xam_int inByteCount,
                           xam_int* const outBytesWritten)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check that params are within bounds
        if((outBytesWritten == (xam_int*)0) || 
           (inBuffer == (const char*)0) ||
           (inByteCount <= 0))
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Write inByteCount bytes from inBuffer into the stream 
        // (return number of bytes written)
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->write(inBuffer, inByteCount, outBytesWritten);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XStream_Seek (const xstream_handle inHandle,
                          const xam_int inOffset,
                          const xam_int inWhence)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check for an invalid "whence" parameter
        if((inWhence < 0) || (inWhence > 2))
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Seek the stream position to inOffset from the inWhence position
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->seek(inOffset, inWhence);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XStream_Tell (const xstream_handle inHandle,
                          xam_int* const outPosition)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check for the required int pointer
        if(outPosition == (xam_int*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the current stream position
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->tell(outPosition);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status
XAM_Router::XStream_Abandon(const xstream_handle inHandle)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Abandon the XStream (assert the type first, since the abandon 
        // method is generic to many types)
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->assertType(XSTREAM_HANDLE);
        vHandle->abandon();
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status 
XAM_Router::XStream_Close (xstream_handle inHandle)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
       // Close the XStream (assert the type first, since the close 
        // method is generic to many types)        
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->assertType(XSTREAM_HANDLE);
        vHandle->close();

        // Deallocate the XAM_Handle wrapper
        deleteHandle(vHandle);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

// =============================================================================
// Methods for managing the connection to a XAM storage system
// =============================================================================

xam_status 
XAM_Router::XAMLibrary_Connect (const xam_string inXRI,
                                xsystem_handle* const outHandle)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the XRI is in UTF-8
        XAM_Utils::validateUTF8(inXRI);

        // Check the required handle poiinter
        if(outHandle == (xsystem_handle*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the XAMLibrary object instance
        // (This initializes the XAM object and preloads any VIM libraries
        //  that may have been configured.)
        XAM_Object* vXAMObj = XAM_Object::getInstance();

        // Determine which VIM should be used for the connect
        int vNumRetries = 0;
        xam_boolean vVimNameSpecified = false;
        xam_boolean vVimIsNull = false;
        xam_boolean vVimConnectSuccess = false;
        xam_string vCxnStr = "";  

        // Try all loaded VIMs until either all are exhausted or one succeeds.
        //   - If a VIM name is specified, every attempt will be made to load
        //     the VIM with the given name. If it cannot be located, an error is 
        //     returned
        //   - If no VIM name was specified, continue to discover VIMs until 
        //     one either successfully connects using the given connection 
        //     string or until no additional VIMs can be found.
        XAM_VIM* vVim = (XAM_VIM*)0;
        while(!vVimConnectSuccess)
        {    
            try
            {
                // Select a VIM for this operation (may throw on VIM loading 
                // error, but returns NULL when no VIM could be selected)
                vVim = selectVim(inXRI, 
                                 vCxnStr, 
                                 vVimNameSpecified, 
                                 vNumRetries);

                // No VIM library could be located
                if(vVim == (XAM_VIM*)0)
                {
                    vVimIsNull = true;

                    XAMVimNotFoundException e;
                    throw e;
                }

                // Perform the connect and retrieve an xsystem_handle 
                // (The XAM Object serves as the XSystem factory)
                vXAMObj->connect(vVim, vCxnStr, outHandle);

                // Break the retry loop on successful connect
                vVimConnectSuccess = true;           
            }
            catch(XAMInvalidParamException &e)
            {
                // Always return the invalid param exception
                throw e;
            }
            catch(XAMInvalidXRIException &e)
            {
                // Always return the invalid XRI exception
                throw e;
            }
            catch(XAM_Exception &e)
            {
                // Propagate the error to the caller if no VIM could 
                // be selected, or if the specified VIM could not be located.
                if(vVimNameSpecified || vVimIsNull)
                    throw e;
                else
                {
                    // A retryable error was thrown, so log the failure, and
                    // try selecting another VIM for this operation as we
                    // reenter the loop
                    XAM_Logger::Log(LOGLEVEL_DEBUG, component_xam_debug, 
                                    "Failed to connect to VIM[%i]: %s (%ld)",
                                    vNumRetries - 1, 
                                    vVim ? vVim->getName() : "<unknown>", 
                                    e.getStatusCode());
                }
            }
        }

        // Wrap the retrieved handle with a scoped XAM_Handle type 
        // before returning to caller
        XSystem* vXSystem = createXSystemHandle(*outHandle, vVim);
        *outHandle = (xsystem_handle) vXSystem;
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSystem_Authenticate (const xsystem_handle inHandle,
                                  const char* inBuffer,
                                  const xam_int inByteCount,
                                  xstream_handle* const outXStream)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required handle pointer
        if(outXStream == (xstream_handle*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Authenticate to the XSystem
        XAM_Handle* vHandle = getHandle(inHandle); 
        try
        {
            // This handle is currently invalid. Init to zero.
            *outXStream = 0;

            // Authenticate (throws an exception on VIM API error)
            vHandle->authenticate(inBuffer, inByteCount, outXStream);  

            // Authentication succeeded, check for an output stream
            if(*outXStream != 0)
            {
                // Wrap the retrieved handle with a scoped XAM_Handle type 
                // before returning to caller
                XStream* vXStream = createXStreamHandle(*outXStream, vHandle);
                *outXStream = (xstream_handle) vXStream;
            }
        }
        catch(XAM_Exception &e)
        {
            // Authentication failed, check for an output stream
            if(*outXStream != 0)
            {
                // Wrap the retrieved handle with a scoped XAM_Handle type 
                // before returning to caller
                XStream* vXStream = createXStreamHandle(*outXStream, vHandle);
                *outXStream = (xstream_handle) vXStream;
            }

            throw e;
        }
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSystem_Close (const xsystem_handle inHandle)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Close the XStream (assert the type first, since the close 
        // method is generic to many types)
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->assertType(XSYSTEM_HANDLE);
        vHandle->close();

        // Deallocate the XAM_Handle wrapper
        deleteHandle(vHandle);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSystem_Abandon (const xsystem_handle inHandle)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Abandon the XSystem (assert the type first, since the abandon 
        // method is generic to many types)
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->assertType(XSYSTEM_HANDLE);
        vHandle->abandon();
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

// =============================================================================
// Methods for XSet management
// =============================================================================

xam_status 
XAM_Router::XSystem_DeleteXSet (const xsystem_handle inHandle,
                                const xam_xuid inXUID)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Delete an XSet from this XSystem
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->deleteXSet(inXUID);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSystem_HoldXSet (const xsystem_handle inHandle,
                              const xam_xuid inXUID,
                              const xam_string inHoldID)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the ID is in UTF-8
        XAM_Utils::validateUTF8(inHoldID);

        // Hold this XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->holdXSet(inXUID, inHoldID);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSystem_ReleaseXSet (const xsystem_handle inHandle,
                                 const xam_xuid inXUID,
                                 const xam_string inHoldID)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the ID is in UTF-8
        XAM_Utils::validateUTF8(inHoldID);

        // Release the XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->releaseXSet(inXUID, inHoldID);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSystem_AccessXSet (const xsystem_handle inHandle,
                                const xam_xuid inXUID,
                                const xam_int inMode,
                                xam_boolean* const outIsAccessible)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the require boolean pointer
        if(outIsAccessible == (xam_boolean*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Determine the accessibility of this XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->accessXSet(inXUID, inMode, outIsAccessible);  
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSystem_IsXSetRetained (const xsystem_handle inHandle,
                                    const xam_xuid inXUID,
                                    xam_boolean* const outIsRetained)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the require boolean pointer
        if(outIsRetained == (xam_boolean*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Determine if this XSet is under retention
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->isXSetRetained(inXUID, outIsRetained);  
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status
XAM_Router::XSystem_GetXSetAccessTime (const xsystem_handle inHandle,
                                       const xam_xuid inXUID,
                                       xam_datetime* const outAccessTime)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required datetime pointer
        if(outAccessTime == (xam_datetime*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the time of the last open/commit for this XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getXSetAccessTime(inXUID, outAccessTime);  
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

// =============================================================================
// Methods for creating an XSet
// =============================================================================

xam_status 
XAM_Router::XSystem_CreateXSet (const xsystem_handle inHandle,
                                const xam_string inMode,
                                xset_handle* const outXSet)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required handle pointer
        if(outXSet == (xset_handle*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Create the XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->createXSet(inMode, outXSet);

        // Wrap the retrieved handle with a scoped XAM_Handle type 
        // before returning to caller
        XSet* vXSet = createXSetHandle(*outXSet, vHandle);
        *outXSet = (xset_handle) vXSet;
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

// ===========================================================================
// Methods for XSet policy management
// ===========================================================================

xam_status
XAM_Router::XSet_ApplyAccessPolicy (const xset_handle inHandle,
                                    const xam_boolean inBinding,
                                    const xam_string inPolicy)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the policy string is in UTF-8
        XAM_Utils::validateUTF8(inPolicy);

        // Apply the management policy
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->applyAccessPolicy(inBinding, inPolicy);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status
XAM_Router::XSet_ResetAccessFields (const xset_handle inHandle)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Reset management policies for this XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->resetAccessFields();
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status 
XAM_Router::XSet_ApplyManagementPolicy (const xset_handle inHandle,
                                        const xam_boolean inBinding,
                                        const xam_string inPolicy)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the policy string is in UTF-8
        XAM_Utils::validateUTF8(inPolicy);

        // Apply the management policy
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->applyManagementPolicy(inBinding, inPolicy);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status
XAM_Router::XSet_ResetManagementFields (const xset_handle inHandle)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Reset management policies for this XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->resetManagementFields();
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status
XAM_Router::XSet_CreateRetention (const xset_handle inHandle,
                                  const xam_boolean inBinding,
                                  const xam_string inRetentionID)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Create retention identifier on this XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->createRetention(inBinding, inRetentionID);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status
XAM_Router::XSet_SetRetentionEnabledFlag (const xset_handle inHandle,
                                          const xam_string inRetentionID,
                                          const xam_boolean inBinding,
                                          const xam_boolean inEnabled)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Enable this retention for this XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->setRetentionEnabledFlag(inRetentionID, inBinding, inEnabled);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status
XAM_Router::XSet_ApplyRetentionEnabledPolicy (const xset_handle inHandle,
                                              const xam_string inRetentionID,
                                              const xam_boolean inBinding,
                                              const xam_string inPolicy)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Apply this retention enabled policy for this XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->applyRetentionEnabledPolicy(inRetentionID, inBinding, inPolicy);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status
XAM_Router::XSet_SetRetentionDuration (const xset_handle inHandle,
                                       const xam_string inRetentionID,
                                       const xam_boolean inBinding,
                                       const xam_int inDuration)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Set this retention duration for this XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->setRetentionDuration(inRetentionID, inBinding, inDuration);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status
XAM_Router::XSet_ApplyRetentionDurationPolicy (const xset_handle inHandle,
                                               const xam_string inRetentionID,
                                               const xam_boolean inBinding,
                                               const xam_string inPolicy)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Apply the retention duration policy for this XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->applyRetentionDurationPolicy(inRetentionID, inBinding, inPolicy);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status
XAM_Router::XSet_SetRetentionStarttime (const xset_handle inHandle,
                                        const xam_string inRetentionID,
                                        const xam_boolean inBinding)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Reset this retention start time for this XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->setRetentionStarttime(inRetentionID, inBinding);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status
XAM_Router::XSet_SetBaseRetention (const xset_handle inHandle,
                                   const xam_boolean inBinding,
                                   const xam_int inDuration)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Set this base retention/duration for this XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->setBaseRetention(inBinding, inDuration);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status
XAM_Router::XSet_ApplyBaseRetentionPolicy (const xset_handle inHandle,
                                           const xam_boolean inBinding,
                                           const xam_string inPolicy)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Set this base retention policy for this XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->applyBaseRetentionPolicy(inBinding, inPolicy);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status 
XAM_Router::XSet_ApplyAutoDeletePolicy (const xset_handle inHandle,
                                        const xam_boolean inBinding,
                                        const xam_string inPolicy)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the policy string is in UTF-8
        XAM_Utils::validateUTF8(inPolicy);

        // Apply the auto delete policy
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->applyAutoDeletePolicy(inBinding, inPolicy);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSet_SetAutoDelete (const xset_handle inHandle,
                                const xam_boolean inBinding,
                                const xam_boolean inAutoDelete)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Set the auto delete state
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->setAutoDelete(inBinding, inAutoDelete);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSet_ApplyShredPolicy (const xset_handle inHandle,
                                   const xam_boolean inBinding,
                                   const xam_string inPolicy)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the policy string is in UTF-8
        XAM_Utils::validateUTF8(inPolicy);

        // Apply the shredding policy
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->applyShredPolicy(inBinding, inPolicy);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status 
XAM_Router::XSet_SetShred (const xset_handle inHandle,
                           const xam_boolean inBinding,
                           const xam_boolean inShred)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Set the shredding state
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->setShred(inBinding, inShred);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status 
XAM_Router::XSet_ApplyStoragePolicy (const xset_handle inHandle,
                                     const xam_boolean inBinding,
                                     const xam_string inPolicy)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Ensure that the policy string is in UTF-8
        XAM_Utils::validateUTF8(inPolicy);

        // Apply the storage policy
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->applyStoragePolicy(inBinding, inPolicy);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSet_GetActualRetentionDuration (const xset_handle inHandle,
                                             const xam_string inRetentionID,
                                             xam_int* const outDuration)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required boolean pointer
        if(outDuration == (xam_int*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the actual retention duration state
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getActualRetentionDuration(inRetentionID, outDuration);  
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status 
XAM_Router::XSet_GetActualRetentionEnabled (const xset_handle inHandle,
                                            const xam_string inRetentionID,
                                            xam_boolean* const outEnabled)
{
 xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required boolean pointer
        if(outEnabled == (xam_boolean*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the actual retention enabled state
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getActualRetentionEnabled(inRetentionID, outEnabled);  
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status 
XAM_Router::XSet_GetActualAutoDelete (const xset_handle inHandle,
                                      xam_boolean* const outEnabled)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required boolean pointer
        if(outEnabled == (xam_boolean*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the actual auto delete state
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getActualAutoDelete(outEnabled);  
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSet_GetActualShred (const xset_handle inHandle,
                                 xam_boolean* const outEnabled)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required boolean pointer
        if(outEnabled == (xam_boolean*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the actual auto delete state
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getActualShred(outEnabled);  
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

// ===========================================================================
// Methods for editing an XSet
// ===========================================================================

xam_status 
XAM_Router::XSystem_OpenXSet (const xsystem_handle inHandle,
                              const xam_xuid inXUID,
                              const xam_string inMode,
                              xset_handle* const outXSet)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required handle pointer
        if(outXSet == (xset_handle*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Open the XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->openXSet(inXUID, inMode, outXSet);

        // Wrap the retrieved handle with a scoped XAM_Handle type before returning to caller
        XSet* vXSet = createXSetHandle(*outXSet, vHandle);
        *outXSet = (xset_handle) vXSet;
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSystem_CopyXSet (const xsystem_handle inHandle,
                              const xam_xuid inXUID,
                              const xam_string inMode,
                              xset_handle* const outXSet)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required handle pointer
        if(outXSet == (xset_handle*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Open the XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->copyXSet(inXUID, inMode, outXSet);

        // Wrap the retrieved handle with a scoped XAM_Handle type before returning to caller
        XSet* vXSet = createXSetHandle(*outXSet, vHandle);
        *outXSet = (xset_handle) vXSet;
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSet_Commit (const xset_handle inHandle,
                         xam_xuid* const outXUID)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required xam_xuid pointer
        if(outXUID == (xam_xuid*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Commit this XSet
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->commit(outXUID);   
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSet_Abandon (const xset_handle inHandle)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Abandon the XSet (assert the type first, since the abandon 
        // method is generic to many types)        
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->assertType(XSET_HANDLE);
        vHandle->abandon();
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSet_Close (const xset_handle inHandle)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Close the XSet (assert the type first, since the close 
        // method is generic to many types)                
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->assertType(XSET_HANDLE);
        vHandle->close();

        // Deallocate the XAM_Handle wrapper
        deleteHandle(vHandle);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

// ===========================================================================
// Methods for XSet migration
// ===========================================================================

xam_status 
XAM_Router::XSet_OpenExportXStream (const xset_handle inHandle,
                                    xstream_handle* const outXStream)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required handle pointer
        if(outXStream == (xstream_handle*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Open an "export" stream from which canonical XSet contents 
        // may be read
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->openExportXStream(outXStream);
    
        // Wrap the retrieved handle with a scoped XAM_Handle type 
        // before returning to caller        
        XStream* vXStream = createXStreamHandle(*outXStream, vHandle);
        *outXStream = (xstream_handle) vXStream;
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSet_OpenImportXStream (const xset_handle inHandle,
                                    xstream_handle* const outXStream)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required handle pointer
        if(outXStream == (xstream_handle*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Open an "import" stream to which canonical XSet contents 
        // may be written
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->openImportXStream(outXStream);

        // Wrap the retrieved handle with a scoped XAM_Handle type 
        // before returning to caller     
        XStream* vXStream = createXStreamHandle(*outXStream, vHandle);
        *outXStream = (xstream_handle) vXStream;
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

// ===========================================================================
// Methods for job control
// ===========================================================================

xam_status 
XAM_Router::XSet_SubmitJob (const xset_handle inHandle)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Submit the job
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->submitJob();
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XSet_HaltJob (const xset_handle inHandle)      
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Halt the job
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->haltJob();
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

// ======================================================================
//  method prototypes for async i/o
// ======================================================================

xam_status
XAM_Router::XSystem_AsyncOpenXSet (const xsystem_handle inHandle,
                                   const xam_xuid inXUID,
                                   const xam_string inMode,
                                   const XOPID inXOPID,
                                   xasync_callback inCallback,
                                   xasync_handle* const outAsyncHandle)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required handle pointer
        if((outAsyncHandle == (xasync_handle*)0) || !inHandle)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Create an XAsync operation manager handle
        // (We do this first, because the operation below may complete
        // very quickly, and we need the callback to catch notifications).
        XAM_Handle* vHandle = getHandle(inHandle); 
        XAsync* vXAsync = createXAsyncHandle(inCallback, vHandle);

        // Open an XSet asynchronously (returns an xasync_handle)
        vHandle->asyncOpenXSet(inXUID, 
                               inMode, 
                               inXOPID, 
                               inCallback,
                               outAsyncHandle);      

        // Wrap the retrieved handle with a scoped XAM_Handle type 
        // before returning to caller         
        vXAsync->setHandle(outAsyncHandle);
        *outAsyncHandle = (xasync_handle) vXAsync;  
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 


xam_status
XAM_Router::XSystem_AsyncCopyXSet (const xsystem_handle inHandle,
                                   const xam_xuid inXUID,
                                   const xam_string inMode,
                                   const XOPID inXOPID,
                                   xasync_callback inCallback,
                                   xasync_handle* const outAsyncHandle)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required handle pointer
        if((outAsyncHandle == (xasync_handle*)0) || !inHandle)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Create an XAsync operation manager handle
        // (We do this first, because the operation below may complete
        // very quickly, and we need the callback to catch notifications).
        XAM_Handle* vHandle = getHandle(inHandle); 
        XAsync* vXAsync = createXAsyncHandle(inCallback, vHandle);

        // Open an XSet asynchronously (returns an xasync_handle)
        vHandle->asyncCopyXSet(inXUID, 
                               inMode, 
                               inXOPID,
                               inCallback,
                               outAsyncHandle);      

        // Wrap the retrieved handle with a scoped XAM_Handle type 
        // before returning to caller         
        vXAsync->setHandle(outAsyncHandle);
        *outAsyncHandle = (xasync_handle) vXAsync;  
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status 
XAM_Router::XAM_AsyncOpenXStream (const xam_handle_t inHandle,
                                  const xam_string inName,
                                  const xam_string inMode,
                                  const XOPID inXOPID,
                                  xasync_callback inCallback,
                                  xasync_handle* const outAsyncHandle)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required handle pointer
        if(outAsyncHandle == (xasync_handle*)0)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Create an XAsync operation manager handle
        // (We do this first, because the operation below may complete
        // very quickly, and we need the callback to catch notifications).
        XAM_Handle* vHandle = getHandle(inHandle); 
        XAsync* vXAsync = createXAsyncHandle(inCallback, vHandle);

        // Open an XSet asynchronously (returns an xasync_handle)
        vHandle->asyncOpenXStream(inName, 
                                  inMode, 
                                  inXOPID, 
                                  inCallback,
                                  outAsyncHandle);      

        // Wrap the retrieved handle with a scoped XAM_Handle type 
        // before returning to caller    
        vXAsync->setHandle(outAsyncHandle);
        *outAsyncHandle = (xasync_handle) vXAsync;  
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
}

xam_status
XAM_Router::XStream_AsyncRead (const xstream_handle inHandle,
                               char* const ioBuffer,
                               const xam_int inBufferLength,
                               const XOPID inXOPID,
                               xasync_callback inCallback,
                               xasync_handle* const outAsyncHandle)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required handle pointer
        if((outAsyncHandle == (xasync_handle*)0) || !inHandle)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Create an XAsync operation manager handle
        // (We do this first, because the operation below may complete
        // very quickly, and we need the callback to catch notifications).
        XAM_Handle* vHandle = getHandle(inHandle); 
        XAsync* vXAsync = createXAsyncHandle(inCallback, vHandle);

        // Read from the stream asynchronously 
        // (returns an xasync_handle)
        vHandle->asyncRead(ioBuffer, 
                           inBufferLength, 
                           inXOPID, 
                           inCallback,
                           outAsyncHandle);  

        // Wrap the retrieved handle with a scoped XAM_Handle type 
        // before returning to caller 
        vXAsync->setHandle(outAsyncHandle);
        *outAsyncHandle = (xasync_handle) vXAsync;  
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status
XAM_Router::XStream_AsyncWrite (const xstream_handle inHandle,
                                const char* inBuffer,
                                const xam_int inByteCount,
                                const XOPID inXOPID,
                                xasync_callback inCallback,
                                xasync_handle* const outAsyncHandle)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required handle pointer
        if((outAsyncHandle == (xasync_handle*)0) || !inHandle)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Create an XAsync operation manager handle
        // (We do this first, because the operation below may complete
        // very quickly, and we need the callback to catch notifications).
        XAM_Handle* vHandle = getHandle(inHandle); 
        XAsync* vXAsync = createXAsyncHandle(inCallback, vHandle);

        // Write to the stream asynchronously 
        // (returns an xasync_handle)
        vHandle->asyncWrite(inBuffer, 
                            inByteCount, 
                            inXOPID, 
                            inCallback,
                            outAsyncHandle); 

        // Wrap the retrieved handle with a scoped XAM_Handle type 
        // before returning to caller 
        vXAsync->setHandle(outAsyncHandle);
        *outAsyncHandle = (xasync_handle) vXAsync;  
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status
XAM_Router::XStream_AsyncClose (const xstream_handle inHandle,
                                const XOPID inXOPID,
                                xasync_callback inCallback,
                                xasync_handle* const outAsyncHandle)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required handle pointer
        if((outAsyncHandle == (xasync_handle*)0) || 
            !inHandle)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Create an XAsync operation manager handle
        // (We do this first, because the operation below may complete
        // very quickly, and we need the callback to catch notifications).
        XAM_Handle* vHandle = getHandle(inHandle); 
        XAsync* vXAsync = createXAsyncHandle(inCallback, vHandle);

        // Close the stream asynchronously 
        // (returns an xasync_handle)
        vHandle->asyncClose(inXOPID,
                            inCallback,
                            outAsyncHandle);  

        // Wrap the retrieved handle with a scoped XAM_Handle type 
        // before returning to caller 
        vXAsync->setHandle(outAsyncHandle);
        *outAsyncHandle = (xasync_handle) vXAsync;  
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status
XAM_Router::XSet_AsyncCommit (const xset_handle inHandle,
                              const XOPID inXOPID,
                              xasync_callback inCallback,
                              xasync_handle* const outAsyncHandle)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required handle pointer
        if((outAsyncHandle == (xasync_handle*)0) || !inHandle)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Create an XAsync operation manager handle
        // (We do this first, because the operation below may complete
        // very quickly, and we need the callback to catch notifications).
        XAM_Handle* vHandle = getHandle(inHandle); 
        XAsync* vXAsync = createXAsyncHandle(inCallback, vHandle);

        // Commit the XSet asynchronously 
        // (returns an xasync_handle)
        vHandle->asyncCommit(inXOPID, 
                             inCallback, 
                             outAsyncHandle);        

        // Wrap the retrieved handle with a scoped XAM_Handle type 
        // before returning to caller 
        vXAsync->setHandle(outAsyncHandle);
        *outAsyncHandle = (xasync_handle) vXAsync;  
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 


// ===========================================================================
// method prototypes for managing asynchronous operations
// ===========================================================================

xam_status
XAM_Router::XAsync_Halt (const xasync_handle inHandle)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Halt the asynchronous operation
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->halt();
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status
XAM_Router::XAsync_Close (const xasync_handle inHandle)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Close the XAsync (assert the type first, since the close 
        // method is generic to many types)
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->assertType(XASYNC_HANDLE);
        vHandle->close();

        // Deallocate the XAM_Handle wrapper
        deleteHandle(vHandle);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status
XAM_Router::XAsync_IsComplete (const xasync_handle inHandle,
                               xam_boolean* const outIsComplete)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required boolean pointer
        if((outIsComplete == (xam_boolean*)0) || !inHandle)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Determine whether or not the asynchronous operation has completed
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->isComplete(outIsComplete);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status
XAM_Router::XAsync_GetStatus (const xasync_handle inHandle,
                              xam_status* const outStatus)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required xam_status pointer
        if((outStatus == (xam_status*)0) || !inHandle)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the status result of the operation (only valid on complete)
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getStatus(outStatus);
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status
XAM_Router::XAsync_GetXOPID (const xasync_handle inHandle,
                             XOPID* const outXOPID)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required XOPID pointer
        if((outXOPID == (XOPID*)0) || !inHandle)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the XOPID associated with this operation
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getXOPID(outXOPID);        
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status
XAM_Router::XAsync_GetXSet (const xasync_handle inHandle,
                            xset_handle* const outXSet)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required handle pointer
        if((outXSet == (xset_handle*)0) || !inHandle)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the XSet associated with this operation
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getXSet(outXSet);     

        // Get the parent of the Async to use as the parent of the XSet
        XAM_Handle* vParent = vHandle->getParent();

        // Wrap the retrieved handle with a scoped XAM_Handle type 
        // before returning to caller
        XSet* vXSet = createXSetHandle(*outXSet, vParent);
        *outXSet = (xset_handle) vXSet;
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status
XAM_Router::XAsync_GetXStream (const xasync_handle inHandle,
                               xstream_handle* const outXStream)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required handle pointer
        if((outXStream == (xstream_handle*)0) || !inHandle)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the XStream associated with this operation
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getXStream(outXStream);   

        // Get the parent of the Async to use as the parent of the XStream
        XAM_Handle* vParent = vHandle->getParent();

        // Wrap the retrieved handle with a scoped XAM_Handle type 
        // before returning to caller
        XStream* vXStream = createXStreamHandle(*outXStream, vParent);
        *outXStream = (xstream_handle) vXStream;
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status
XAM_Router::XAsync_GetXUID (const xasync_handle inHandle,
                            xam_xuid* const outXUID)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required xam_xuid pointer
        if((outXUID == (xam_xuid*)0) || !inHandle)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the xam_xuid result (only valid on complete)
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getXUID(outXUID);             
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status
XAM_Router::XAsync_GetBytesRead (const xasync_handle inHandle,
                                   xam_int* const outBytesRead)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required int pointer
        if((outBytesRead == (xam_int*)0) || !inHandle)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the number of bytes read (only valid on complete)
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getBytesRead(outBytesRead);          
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 

xam_status
XAM_Router::XAsync_GetBytesWritten (const xasync_handle inHandle,
                                      xam_int* const outBytesWritten)
{
    xam_status vStatus = (xam_status)0;

    try
    {
        // Check the required int pointer
        if((outBytesWritten == (xam_int*)0) || !inHandle)
        {
            XAMInvalidParamException e;
            throw e;
        }

        // Get the number of bytes written (only valid on complete)
        XAM_Handle* vHandle = getHandle(inHandle); 
        vHandle->getBytesWritten(outBytesWritten);       
    }
    catch(XAM_Exception &e)
    {
        vStatus = e.getStatusCode();
    }

    return vStatus;
} 



