#ifndef __RemoteVIM_H__
#define __RemoteVIM_H__
////////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2009, The Storage Networking Industry Association.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - Neither the name of The Storage Networking Industry Association (SNIA)
//      nor the names of its contributors may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
////////////////////////////////////////////////////////////////////////////////
//  XAM SDK Source File Build Version @version.full@
////////////////////////////////////////////////////////////////////////////////
#include "SingleRemoteVIM.h"

/**
 * A class for forming a linkage (an association) between an XSystem and a
 * VIM that does not reside in the same process space.
 *
 * NOTE: The search algorithm to find the xsystem_handle is SLOW at scale, 
 * and should be improved
 */
class RemoteVIM
{
private:
	xam_int mBlockSize;

	// direct xsystem to vim map
	xsystem_handle* mHandle;
	xsystem_handle* mRemoteHandle;
	std::string** mAddress;
	XferBuffer** mCache;
	xam_boolean* mActive; // can replace all above with a class ptr
	xam_int mNumActive;
	xam_int mNumAllocated;

	// indirect handle to xsystem map
	xam_handle_t* mIndirectHandle;
	xasync_callback* mCallback;
	xsystem_handle* mReferencedHandle;
	xam_int* mBuffersize;
	char** mBuffer;
	xam_boolean* mIndirectActive;
	xam_int mNumIndirectActive;
	xam_int mNumIndirectAllocated;

	// VIM list
	SingleRemoteVIM** mSingleVIM;
	xam_boolean* mIsVIMReserved;
	xam_int mNumVIMActive;
	xam_int mNumVIMAllocated;

	// thread info for polling loop
	XAM_ThreadInfo mInfo;

	////////////////////////////////////////////////////////////////////////////
	// internal utilities
	////////////////////////////////////////////////////////////////////////////
	void ExpandDirectBuffers(void);                         // lock to use
	void ExpandIndirectBuffers(void);                       // lock to use
	void ExpandVIMBuffers(void);                            // lock to use
	xam_int FindDirectIndex(const xsystem_handle inHandle); // lock to use
	xam_int FindIndirectIndex(const xam_handle_t inHandle); // lock to use
	xam_int FindAvailableVIMIndex(void);                    // lock to use

public:
	////////////////////////////////////////////////////////////////////////////
	// manage the linkage/association between VIM and xsystem_handle
	////////////////////////////////////////////////////////////////////////////
	/**
	 * Allocate resources to hold bindings to the xsystem
	 * @param inLocalHandle the handle of the xsystem to associate with
	 */
	void RegisterXSystem(const xsystem_handle inLocalHandle);

	/**
	 * Sets the XSystem handle associated with this VIM instance
	 * @param inLocalHandle the xsystem handle assigned by the local VIM 
	 * @param inRemoteHandle the xsystem handle assigned by the remote VIM 
	 */
	void BindToRemoteXSystem(const xsystem_handle inLocalHandle,
	                         const xsystem_handle inRemoteHandle);

	/**
	 * Associate a handle with an xsystem_handle
	 * @param inHandle the handle to associate
	 * @param inXSystemHandle the handle of the xsystem to associate with
	 */
	void BindToHandle(const xam_handle_t inHandle, const xsystem_handle inXSystemHandle);

	/**
	 * Associate a handle with an xsystem_handle
	 * @param inHandle the handle to associate
	 * @param inCallback the callback to be invoked
	 * @param inXSystemHandle the handle of the xsystem to associate with
	 */
	void BindToHandle(const xam_handle_t inHandle, 
	                  const xasync_callback inCallback,
	                  const xsystem_handle inXSystemHandle);

	/**
	 * Associate the xsystem_handle with a single remote VIM
	 * @param inHandle the handle of the xsystem to associate
	 * @param inAddress an STL string containing the address of a remote VIM 
	 */
	void BindToAddress(const xsystem_handle inHandle, const std::string inAddress);

	void BindToBuffer (const xam_handle_t inHandle, 
                       xam_int inBuffersize,
                       char* inBuffer);

	/**
	 * Get the single remote VIM associated with the xsystem_handle
	 * @param inHandle the handle of the local xsystem
	 * @return the handle of the remote VIM 
	 */
	xsystem_handle GetRemoteHandle(const xsystem_handle inHandle);

	/**
	 * Get the xsystem_handle associated with a non-xsystem handle
	 * @param inHandle the associated non-xsystem handle
	 * @return the handle of the referenced xsystem
	 */
	xsystem_handle GetRegisteredXSystem(const xam_handle_t inHandle);

	/**
	 * Get the field cache for writing to a VIM; returns null if 
	 * there is no cache
	 * @param inHandle the xsystem handle
	 * @return a pointer to the cache if one exists
	 */
	XferBuffer* GetCache(const xsystem_handle inHandle);

	void GetBuffer (const xam_handle_t inHandle, 
                    xam_int* outBuffersize,
                    char** outBuffer);

	/**
	 * Delete the cache if one exists
	 * @param inHandle the xsystem handle
	 */
	void DeleteCache(const xsystem_handle inHandle);

	/**
	 * Dissolve the binding between the xsystem and the single remote VIM,
	 * also releasing any resources used in the linkage.
	 * @param inHandle the handle of the xsystem to disassociate
	 */
	void ReleaseDirectBinding(const xsystem_handle inHandle);

	/**
	 * Dissolve the binding between the handle and the xsystem.
	 * @param inHandle the handle to disassociate
	 */
	void ReleaseIndirectBinding(const xam_handle_t inHandle);

	////////////////////////////////////////////////////////////////////////////
	// interact with a single remote VIM
	////////////////////////////////////////////////////////////////////////////
	/**
	 * Get a reserved reference to a managed VIM
	 * @return reference to SingleRemoteVIM
	 */
	SingleRemoteVIM& GetVIM(void);

	/**
	 * Get a reserved reference to a managed VIM
	 * @param inXSystemHandle the xsystem_handle  
	 * @param inCommand the c string containing the command
	 * @param inOPHandle the handle associated with the command
	 * @return reference to SingleRemoteVIM
	 */
	SingleRemoteVIM& GetVIM(const xsystem_handle inXSystemHandle, 
	                        const char* const inCommand,
							const xam_handle_t inOpHandle = 0);

	/**
	 * Release the reserved reference to the managed VIM
	 * @param reference to the managed SingleRemoteVIM
	 */
	void ReleaseVIM(SingleRemoteVIM& inVIM);

	/**
	 * Modifies the URL by appending the command string
	 * @param inHandle the handle of the associated xsystem
	 * @param inCommand the string containing the name of the command
	 */
    void SetCommand(const xsystem_handle inHandle, const char* const inCommand);

	/**
	 * Resets any argument list
	 * @param inHandle the handle of the associated xsystem
	 * @param inUseXSystem flag to set the first name value pair to the xsystem handle
	 */
	void ResetArgs(const xsystem_handle inHandle, const xam_boolean inUseXSystem);

	/**
	 * Appends the specified name/value pair to the argument list
	 * @param inName the string containing the name of the argument
	 * @param inValue the string containing the value of the argument
	 */
    void AppendArg(const xsystem_handle inHandle,
	               const char* const inName,
	               const char* const inValue);

	/**
	 * Appends the specified name/value pair to the argument list
	 * @param inName the string containing the name of the argument
	 * @param inValue the xam_int containing the value of the argument
	 */
    void AppendArg(const xsystem_handle inHandle,
	               const char* const inName,
	               const xam_int inValue);

	/**
	 * Appends the specified name/value pair to the argument list
	 * @param inName the string containing the name of the argument
	 * @param inValue the xam_boolean containing the value of the argument
	 */
    void AppendArg(const xsystem_handle inHandle,
	               const char* const inName,
	               const xam_boolean inValue);

	/**
	 * Appends the specified name/value pair to the argument list
	 * @param inName the string containing the name of the argument
	 * @param inValue the xam_double containing the value of the argument
	 */
    void AppendArg(const xsystem_handle inHandle,
	               const char* const inName,
	               const xam_double inValue);

	/**
	 * Appends the specified name/value pair to the argument list
	 * @param inName the string containing the name of the argument
	 * @param inValue the xam_xuid containing the value of the argument
	 */
    void AppendXUID(const xsystem_handle inHandle,
	                const char* const inName,
	                const xam_xuid inValue);

	////////////////////////////////////////////////////////////////////////////
	// command execution
	////////////////////////////////////////////////////////////////////////////
	/**
	 * Issues the queued up command as a GET with the current args
	 * @param ioHeaderStr the header of the response
	 * @param ioDataBuffer the body of the response
	 */
	void GET(const xsystem_handle inHandle, 
	         std::string* ioHeaderStr, 
	         XferBuffer* ioDataBuffer);

	/**
	 * Issues the queued up command as a POST with the current args
	 * @param inBuffer binary data to attach to the POST
	 * @param inByteCount number of bytes in the buffer
	 * @param ioHeaderStr the header of the response
	 * @param ioDataBuffer the body of the response
	 */
	void POST(const xsystem_handle inHandle, 
	          const char* inBuffer,
	          const xam_int inByteCount,
	          std::string* ioHeaderStr, 
	          XferBuffer* ioDataBuffer);

	xam_status GetAsyncData (const xasync_handle inHandle);

	/**
	 * Checks status of ongoing async ops and invokes callbacks as needed
	 */
	xam_boolean POLL (void);

	/**
	 * Marks the polling loop as complete
	 */
	void HaltPOLL (void);

	////////////////////////////////////////////////////////////////////////////
	// constructors
	////////////////////////////////////////////////////////////////////////////
private:
    RemoteVIM(const RemoteVIM&); // Copy constructor not available
    RemoteVIM& operator = (const RemoteVIM&); // Assignment op not available

public:
	/**
	 * Create a class to bind XSystem handles to Remote VIMs
	 */
	RemoteVIM(void);

	////////////////////////////////////////////////////////////////////////////
	// destructor
	////////////////////////////////////////////////////////////////////////////
    virtual ~RemoteVIM(void);

};
#endif  // #ifndef __RemoteVIM_H__
