////////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2009, The Storage Networking Industry Association.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - Neither the name of The Storage Networking Industry Association (SNIA)
//      nor the names of its contributors may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
////////////////////////////////////////////////////////////////////////////////
//  XAM SDK Source File Build Version @version.full@
////////////////////////////////////////////////////////////////////////////////

#include "SingleRemoteVIM.h"
#include "XferBuffer.h"

// the segments of a double (not normalized)
static xam_int sMANTISSA = 0x000FFFFFFFFFFFFFLL;
static xam_int sEXPONENT = 0x7FF0000000000000LL;
static xam_int sSIGN     = 0x8000000000000000LL;
static xam_int sEXPONENT_MASK = 0x7ff;
static xam_int sEXPONENT_SHIFT = 52;

xam_int SingleRemoteVIM::sID = 0;

////////////////////////////////////////////////////////////////////////////////
// callback routines - required by CURL
////////////////////////////////////////////////////////////////////////////////
/*
 * The string callback will be called once for each header and only
 * complete header lines are passed on to the callback.
 */
static size_t 
data_to_string_cb (void *inPtr, size_t inSz, size_t inNMemb, void *ioStream)
{
	// copy the data into a string
    size_t vBytecount = inNMemb*inSz;
	if ( ioStream == (void*)0 || inPtr == (void*)0 )
		vBytecount = 0;
	else
	{
		std::string* vStr = (std::string*)ioStream;
		try { vStr->assign((const char*)inPtr, vBytecount); }
		catch (...) // suppress the error
		{
			vBytecount = 0;
		}
	}
	return vBytecount;
}

/*
 * Capture data into the provided buffer.
 */
static size_t
data_to_buffer_cb(void *inPtr, size_t inSz, size_t inNMemb, void *ioStream)
{
	// copy the data into a string
    xam_int vBytecount = inNMemb*inSz;
    size_t vBytesHandled = 0;
	if ( ioStream == (void*)0 || inPtr == (void*)0 )
		vBytecount = 0;
	else
	{
		XferBuffer* vBuf = (XferBuffer*)ioStream;
		try { vBuf->AddBytes((char*)inPtr, vBytecount); }
		catch (...) // suppress the error
		{
			vBytecount = 0;
		}
	}
	return (size_t)vBytecount;
}

static void
xam_int_to_cstring(const xam_int inVal, char ioCStr[21])
{
	xam_int vVal = inVal;
	xam_int vIndex = 19;
	while ( vIndex >= 0 )
	{
		ioCStr[vIndex--] = 48 + vVal%10;
		vVal /= 10;
	}
	ioCStr[20] = 0;
}

static void
xam_double_to_cstring(const xam_double inValue, char ioCStr[23])
{
	// use the unnormalized values
	xam_boolean vSign = ((xam_int)inValue) & sSIGN ? true : false;
	xam_int vMantissa = ((xam_int)inValue) & sMANTISSA;
	xam_int vExponent = (((xam_int)inValue) >> sEXPONENT_SHIFT) & sEXPONENT_MASK;

	// make a cstring of the form +|-eeee+mmmmmmmmmmmmmmmm
	ioCStr[0] = vSign ? 45 : 43;
	xam_int vIndex = 4;
	while ( vIndex > 0 )
	{
		ioCStr[vIndex--] = 48 + vExponent%10;
		vExponent /= 10;
	}
	ioCStr[5] = 43;
	vIndex = 21;
	while ( vIndex > 5 )
	{
		ioCStr[vIndex--] = 48 + vMantissa%10;
		vMantissa /= 10;
	}
	ioCStr[22] = (char)0;
}

////////////////////////////////////////////////////////////////////////////////
// SingleRemoteVIM methods
////////////////////////////////////////////////////////////////////////////////

bool 
SingleRemoteVIM::Initialize ()
{
	bool vSuccess = false;
	if ( mCURLSession == NULL )
		mCURLSession = curl_easy_init();
	else
		curl_easy_reset(mCURLSession);
	if ( mCURLSession != NULL )
	{
		curl_easy_setopt(mCURLSession, CURLOPT_NOPROGRESS, 1);
		vSuccess = true;
	}
	return vSuccess ? false : true;
}

bool 
SingleRemoteVIM::IsInitialized (void) const
{ 
	return mCURLSession != NULL ? true : false; 
}

void 
SingleRemoteVIM::AppendArg (const char* const inName, const char* const inValue)
{
	std::string vArgStr(inName);
	vArgStr.append(": ");
	vArgStr.append(inValue);
    mCURLHeaders = curl_slist_append(mCURLHeaders, vArgStr.c_str());
    if ( mCURLHeaders == NULL )
    {
        XAMUnknownException e;
        throw e;
    }
}

void 
SingleRemoteVIM::AppendArg (const char* const inName, const xam_int inValue)
{
	char vCStr[21];
	xam_int_to_cstring((xam_int)inValue, vCStr);
	AppendArg(inName, vCStr);
}
	
void 
SingleRemoteVIM::AppendArg(const char* const inName, const xam_status inValue)
{
	AppendArg(inName, (xam_int)inValue);
}

void 
SingleRemoteVIM::AppendArg (const char* const inName, const xam_boolean inValue)
{
	if ( inValue )
		AppendArg(inName, "true");
	else
		AppendArg(inName, "false");
}

void 
SingleRemoteVIM::AppendArg (const char* const inName, const xam_double inValue)
{
	char vCStr[23];
	xam_double_to_cstring(inValue, vCStr);
	AppendArg(inName, vCStr);
}

void 
SingleRemoteVIM::AppendXUID(const char* const inName, const xam_xuid inValue)
{
	char vCopy[81];
	XAM_POSIX::xam_strncpy(vCopy, inValue, 80);
	vCopy[80] = (char)0;
	AppendArg(inName, vCopy);
}

void 
SingleRemoteVIM::ResetArgs(const xam_boolean inUseXSystem)
{
	// free any existing name/value pairs
	if ( mCURLHeaders != NULL )
	{
		curl_slist_free_all(mCURLHeaders);
		mCURLHeaders = NULL;
	}

	// create a new list with the name/value pair as the first entry
	if ( inUseXSystem )
		AppendArg("handle", mXSystem);
}

void
SingleRemoteVIM::SetCommand (const std::string inAddress, const char* const inCommand)
{
	// reset all the options to the default
	Initialize();

	// format the command
	mURL = "http://" + inAddress + "/";
	mCommand = mURL + inCommand;
}

void 
SingleRemoteVIM::SetXSystemHandle(const xsystem_handle inHandle)
{
	mXSystem = inHandle;
}

void
SingleRemoteVIM::GET(std::string* ioHeaderString, XferBuffer* ioDataBuffer)
{
	// Set the formatted command into the CURL que...
	CURLcode vStat = curl_easy_setopt(mCURLSession, CURLOPT_URL, mCommand.c_str()); 
	if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }

    // ...then add the arguments...
	vStat = curl_easy_setopt(mCURLSession, CURLOPT_HTTPHEADER, mCURLHeaders);
	if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }

    // ...make it a GET, and not a POST...
    vStat = curl_easy_setopt(mCURLSession, CURLOPT_HTTPGET, 1);
	if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }

    // ...set the max size of each chunk of data received...
    vStat = curl_easy_setopt(mCURLSession, CURLOPT_BUFFERSIZE, CURL_MAX_WRITE_SIZE);
	if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }

    // ...set the callback function to be invoked when the response arrives...
    vStat = curl_easy_setopt(mCURLSession, CURLOPT_WRITEFUNCTION, data_to_buffer_cb);
	if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }
    vStat = curl_easy_setopt(mCURLSession, CURLOPT_WRITEDATA, ioDataBuffer);
    if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }

    // ...set the callback function to capture the header(s) of the HTTP GET response...
    vStat = curl_easy_setopt(mCURLSession, CURLOPT_HEADERFUNCTION, data_to_string_cb);
	if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }
    vStat = curl_easy_setopt(mCURLSession, CURLOPT_WRITEHEADER, ioHeaderString);
	if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }

    // Execute the queued HTTP GET and asyncronously read the response
    vStat = curl_easy_perform(mCURLSession);
	if ( vStat != CURLE_OK )
	{
		XAMUnknownException e;
        throw e;
	}
}

void
SingleRemoteVIM::POST(const char* inBuffer,
                             const xam_int inByteCount,
                             std::string* ioHeaderString, 
                             XferBuffer* ioDataBuffer)
{
	// Set the formatted command into the CURL que...
	curl_easy_setopt(mCURLSession, CURLOPT_VERBOSE, TRUE);
	CURLcode vStat = curl_easy_setopt(mCURLSession, CURLOPT_URL, mCommand.c_str()); 
	if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }

	// ...then add the arguments...
	char* vContentType = "Content-Type: application/octet-stream";
	mCURLHeaders = curl_slist_append(mCURLHeaders, vContentType);
    if ( mCURLHeaders == NULL )
    {
        XAMUnknownException e;
        throw e;
    }
	vStat = curl_easy_setopt(mCURLSession, CURLOPT_HTTPHEADER, mCURLHeaders);
	if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }

    // ...make it a POST, and not a GET...
    vStat = curl_easy_setopt(mCURLSession, CURLOPT_HTTPPOST, 1);
	if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }
	vStat = curl_easy_setopt(mCURLSession, CURLOPT_POSTFIELDS, inBuffer);
	if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }
	vStat = curl_easy_setopt(mCURLSession, CURLOPT_POSTFIELDSIZE, inByteCount);
	if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }

    // ...set the max size of each chunk of data received...
    vStat = curl_easy_setopt(mCURLSession, CURLOPT_BUFFERSIZE, CURL_MAX_WRITE_SIZE);
	if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }

    // ...set the callback function to be invoked when the response arrives...
    vStat = curl_easy_setopt(mCURLSession, CURLOPT_WRITEFUNCTION, data_to_buffer_cb);
	if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }
    vStat = curl_easy_setopt(mCURLSession, CURLOPT_WRITEDATA, ioDataBuffer);
    if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }

    // ...set the callback function to capture the header(s) of the HTTP GET response...
    vStat = curl_easy_setopt(mCURLSession, CURLOPT_HEADERFUNCTION, data_to_string_cb);
	if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }
    vStat = curl_easy_setopt(mCURLSession, CURLOPT_WRITEHEADER, ioHeaderString);
	if ( vStat != CURLE_OK )
    {
        XAMUnknownException e;
        throw e;
    }

    // Execute the queued HTTP POST and asyncronously read the response
    vStat = curl_easy_perform(mCURLSession);
	if ( vStat != CURLE_OK )
	{
		XAMUnknownException e;
        throw e;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Public utility routines
////////////////////////////////////////////////////////////////////////////////
void 
SingleRemoteVIM::CacheBoolean(const xam_string inName,
                              const xam_boolean inValue, 
                              const xam_boolean inBinding,
                              XferBuffer& outCache)
{
	// create the string
	std::string vCacheStr(inName);
	vCacheStr.append(".value=");
	vCacheStr.append(inValue?"true":"false");
	vCacheStr.append(" ");
	vCacheStr.append(inName);
	vCacheStr.append(".binding=");
	vCacheStr.append(inBinding?"true":"false");
	vCacheStr.append(" ");
	vCacheStr.append(inName);
	vCacheStr.append(".type=xam_boolean");

	// add the bytes to the cache
	outCache.AddBytes(vCacheStr.c_str(), vCacheStr.length());
}

void 
SingleRemoteVIM::CacheInt(const xam_string inName,
	                      const xam_int inValue, 
	                      const xam_boolean inBinding,
	                      XferBuffer& outCache)
{
	// encode the xam_int to a c string
	char vValue[21];
	xam_int_to_cstring(inValue, vValue);

	// create the string
	std::string vCacheStr(inName);
	vCacheStr.append(".value=");
	vCacheStr.append(vValue);
	vCacheStr.append(" ");
	vCacheStr.append(inName);
	vCacheStr.append(".binding=");
	vCacheStr.append(inBinding?"true":"false");
	vCacheStr.append(" ");
	vCacheStr.append(inName);
	vCacheStr.append(".type=xam_int");

	// add the bytes to the cache
	outCache.AddBytes(vCacheStr.c_str(), vCacheStr.length());
}

void 
SingleRemoteVIM::CacheDouble(const xam_string inName,
	                         const xam_double inValue, 
	                         const xam_boolean inBinding,
	                         XferBuffer& outCache)
{
	// encode the xam_double to a c string
	char vValue[23];
	xam_double_to_cstring(inValue, vValue);

	// create the string
	std::string vCacheStr(inName);
	vCacheStr.append(".value=");
	vCacheStr.append(vValue);
	vCacheStr.append(" ");
	vCacheStr.append(inName);
	vCacheStr.append(".binding=");
	vCacheStr.append(inBinding?"true":"false");
	vCacheStr.append(" ");
	vCacheStr.append(inName);
	vCacheStr.append(".type=xam_double");

	// add the bytes to the cache
	outCache.AddBytes(vCacheStr.c_str(), vCacheStr.length());
}

void 
SingleRemoteVIM::CacheXUID(const xam_string inName,
	                       const xam_xuid inValue, 
	                       const xam_boolean inBinding,
	                       XferBuffer& outCache)
{
	// encode the xam_xuid to a c string
	char vValue[81];
	XAM_POSIX::xam_strncpy(vValue, inValue, 80);
	vValue[80] = (char)0;

	// create the string
	std::string vCacheStr(inName);
	vCacheStr.append(".value=");
	vCacheStr.append(vValue);
	vCacheStr.append(" ");
	vCacheStr.append(inName);
	vCacheStr.append(".binding=");
	vCacheStr.append(inBinding?"true":"false");
	vCacheStr.append(" ");
	vCacheStr.append(inName);
	vCacheStr.append(".type=xam_xuid");

	// add the bytes to the cache
	outCache.AddBytes(vCacheStr.c_str(), vCacheStr.length());
}

void
SingleRemoteVIM:: CacheString(const xam_string inName,
	                          const xam_string inValue, 
	                          const xam_boolean inBinding,
	                          XferBuffer& outCache)
{
	// create the string
	std::string vCacheStr(inName);
	vCacheStr.append(".value=");
	vCacheStr.append(inValue);
	vCacheStr.append(" ");
	vCacheStr.append(inName);
	vCacheStr.append(".binding=");
	vCacheStr.append(inBinding?"true":"false");
	vCacheStr.append(" ");
	vCacheStr.append(inName);
	vCacheStr.append(".type=xam_string");

	// add the bytes to the cache
	outCache.AddBytes(vCacheStr.c_str(), vCacheStr.length());
}

void 
SingleRemoteVIM::CacheDatetime(const xam_string inName,
	                           const xam_datetime inValue, 
	                           const xam_boolean inBinding,
	                           XferBuffer& outCache)
{
	// create the string
	std::string vCacheStr(inName);
	vCacheStr.append(".value=");
	vCacheStr.append(inValue);
	vCacheStr.append(" ");
	vCacheStr.append(inName);
	vCacheStr.append(".binding=");
	vCacheStr.append(inBinding?"true":"false");
	vCacheStr.append(" ");
	vCacheStr.append(inName);
	vCacheStr.append(".type=xam_datetime");

	// add the bytes to the cache
	outCache.AddBytes(vCacheStr.c_str(), vCacheStr.length());
}

////////////////////////////////////////////////////////////////////////////////
// Private utility routines
////////////////////////////////////////////////////////////////////////////////
void 
SingleRemoteVIM::CopyMembers (const SingleRemoteVIM& src)
{
    if ( mCURLSession != NULL )
    {
        curl_easy_cleanup(mCURLSession);
		mCURLSession = NULL;
	}
    mXSystem = src.mXSystem;
    mURL = src.mURL;
	mCommand = src.mURL;
	mCURLHeaders = NULL;
}

////////////////////////////////////////////////////////////////////////////////
// constructors
////////////////////////////////////////////////////////////////////////////////
SingleRemoteVIM::SingleRemoteVIM (void): 
	mCURLSession(NULL),
	mXSystem(0),
	mCURLHeaders(NULL),
	mURL(""),
	mCommand("")
{
	mID = sID++;
//    mURL = "http://" + inAddress + "/";
//	mCommand = mURL;
}

SingleRemoteVIM::SingleRemoteVIM (const SingleRemoteVIM& inVIM)
{
    CopyMembers(inVIM);
}

SingleRemoteVIM& 
SingleRemoteVIM::operator = (const SingleRemoteVIM& inVIM)
{
    if ( this != &inVIM )
        CopyMembers(inVIM);
    return *this;
}

////////////////////////////////////////////////////////////////////////////////
// destructor
////////////////////////////////////////////////////////////////////////////////
SingleRemoteVIM::~SingleRemoteVIM (void)
{
    if ( mCURLSession != NULL )
    {
        curl_easy_cleanup(mCURLSession);
        mCURLSession = NULL;
    }
	if ( mCURLHeaders != NULL )
	{
		curl_slist_free_all(mCURLHeaders);
		mCURLHeaders = NULL;
	}
}
