#ifndef __SingleRemoteVIM_H__
#define __SingleRemoteVIM_H__
////////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2009, The Storage Networking Industry Association.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - Neither the name of The Storage Networking Industry Association (SNIA)
//      nor the names of its contributors may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
////////////////////////////////////////////////////////////////////////////////
//  XAM SDK Source File Build Version @version.full@
////////////////////////////////////////////////////////////////////////////////
#include <string>
#include "vim.h"
#include "curl/curl.h"
#include "XferBuffer.h"
#include "XAM_Exception.h"

/**
 * The exception thrown by the local VIM
 */
class XAMUnknownException : public XAM_Exception
{
 public:
    XAMUnknownException() : XAM_Exception(XAM_UNKNOWN_ERROR) {}
};

/**
 * A class representing the linkage between an XSystem and a VIM that does
 * not reside in the same process space. The scope of this class is a single
 * transaction.
 *
 * NOTE: This class can be reused for multiple serial transactions.
 */
class SingleRemoteVIM
{
 private:
    xsystem_handle mXSystem;  // handle of remote XSystem (was mVimHandle)

    std::string mURL;         // URL of the remote VIM HTTP server

	std::string mCommand;     // modified URL with command string appended
	curl_slist* mCURLHeaders; // arguments to command as name/value pairs
    CURL* mCURLSession;       // pointer to the handle of a CURL session

	static xam_int sID;       // the static counter for generating IDs
	xam_int mID;              // the ID of the instance

	friend class RemoteVIM;

 public:
	/**
	 * Opens a CURL session and stores the handle.
	 * @returns false for success, true if there was some sort of error.
	 */
    bool Initialize(void);

    /**
	 * Returns the state of the CURL session.
	 * @returns true if the CURL session was successfully opened
	 */
    bool IsInitialized(void) const;

	/**
	 * Add a field to a cache to pass in when the connect
	 * method is invoked.
	 * @param inName the name of the field
	 * @param inValue the value of the field
	 * @param inBinding true if the field should be binding
	 * @param outCache reference to XferBuffer containing the cache
	 */
	static void CacheBoolean(const xam_string inName,
	                         const xam_boolean inValue, 
	                         const xam_boolean inBinding,
	                         XferBuffer& outCache);

	/**
	 * Add a field to a cache to pass in when the connect
	 * method is invoked.
	 * @param inName the name of the field
	 * @param inValue the value of the field
	 * @param inBinding true if the field should be binding
	 * @param outCache reference to XferBuffer containing the cache
	 */
	static void CacheInt(const xam_string inName,
	                     const xam_int inValue, 
	                     const xam_boolean inBinding,
	                     XferBuffer& outCache);

	/**
	 * Add a field to a cache to pass in when the connect
	 * method is invoked.
	 * @param inName the name of the field
	 * @param inValue the value of the field
	 * @param inBinding true if the field should be binding
	 * @param outCache reference to XferBuffer containing the cache
	 */
	static void CacheDouble(const xam_string inName,
	                        const xam_double inValue, 
	                        const xam_boolean inBinding,
	                        XferBuffer& outCache);

	/**
	 * Add a field to a cache to pass in when the connect
	 * method is invoked.
	 * @param inName the name of the field
	 * @param inValue the value of the field
	 * @param inBinding true if the field should be binding
	 * @param outCache reference to XferBuffer containing the cache
	 */
	static void CacheXUID(const xam_string inName,
	                      const xam_xuid inValue, 
	                      const xam_boolean inBinding,
	                      XferBuffer& outCache);

	/**
	 * Add a field to a cache to pass in when the connect
	 * method is invoked.
	 * @param inName the name of the field
	 * @param inValue the value of the field
	 * @param inBinding true if the field should be binding
	 * @param outCache reference to XferBuffer containing the cache
	 */
	static void CacheString(const xam_string inName,
	                        const xam_string inValue, 
	                        const xam_boolean inBinding,
	                        XferBuffer& outCache);

	/**
	 * Add a field to a cache to pass in when the connect
	 * method is invoked.
	 * @param inName the name of the field
	 * @param inValue the value of the field
	 * @param inBinding true if the field should be binding
	 * @param outCache reference to XferBuffer containing the cache
	 */
	static void CacheDatetime(const xam_string inName,
	                          const xam_datetime inValue, 
	                          const xam_boolean inBinding,
	                          XferBuffer& outCache);

	/**
	 * Sets the XSystem handle associated with this VIM instance
	 * @param inHandle the xsystem handle assigned by the remote VIM 
	 */
    void SetXSystemHandle(const xsystem_handle inHandle);

	/**
	 * Modifies the URL by appending the command string
	 * @throws XAMUnknownException if there is a CURL error
	 * @param inAddress the std::string containing the remote HTTP address
	 * @param inCommand the c string containing the name of the command
	 */
	void SetCommand(const std::string inAddress, const char* const inCommand);

	/**
	 * Resets any argument list
	 * @throws XAMUnknownException if there is a CURL error
	 * @param inUseXSystem flag to set the first name value pair to the xsystem handle
	 */
	void ResetArgs(const xam_boolean inUseXSystem);

	/**
	 * Appends the specified name/value pair to the argument list
	 * @throws XAMUnknownException if there is a CURL error
	 * @param inName the string containing the name of the argument
	 * @param inValue the string containing the value of the argument
	 */
    void AppendArg(const char* const inName, const char* const inValue);

    /**
	 * Appends the specified name/value pair to the argument list
	 * @param inName the string containing the name of the argument
	 * @param inValue the xam_int containing the value of the argument
	 */
	void AppendArg(const char* const inName, const xam_int inValue);

    /**
	 * Appends the specified name/value pair to the argument list
	 * @param inName the string containing the name of the argument
	 * @param inValue the xam_boolean containing the value of the argument
	 */
	void AppendArg(const char* const inName, const xam_boolean inValue);

    /**
	 * Appends the specified name/value pair to the argument list
	 * @param inName the string containing the name of the argument
	 * @param inValue the xam_double containing the value of the argument
	 */
	void AppendArg(const char* const inName, const xam_double inValue);

    /**
	 * Appends the specified name/value pair to the argument list
	 * @param inName the string containing the name of the argument
	 * @param inValue the xam_status containing the value of the argument
	 */
	void AppendArg(const char* const inName, const xam_status inValue);

	/**
	 * Appends the specified name/value pair to the argument list
	 * @param inName the string containing the name of the argument
	 * @param inValue the xam_xuid containing the value of the argument
	 */
	void AppendXUID(const char* const inName, const xam_xuid inValue);

	/**
	 * Issues the queued up command with the current args
	 * @param ioHeaderStr the header of the response
	 * @param ioDataBuffer the body of the response
	 */
	void GET(std::string* ioHeaderString, XferBuffer* ioDataBuffer);

	/**
	 * Issues the queued up command as a POST with the current args
	 * @param inBuffer binary data to attach to the POST
	 * @param inByteCount number of bytes in the buffer
	 * @param ioHeaderStr the header of the response
	 * @param ioDataBuffer the body of the response
	 */
	void POST(const char* inBuffer,
	          const xam_int inByteCount,
	          std::string* ioHeaderStr, 
	          XferBuffer* ioDataBuffer);

 private:
	////////////////////////////////////////////////////////////////////////////
	// Private utility routines
	////////////////////////////////////////////////////////////////////////////
	/* Copy member variables from the referenced RemoteVIM.
	 * @param source The RemoteVIM to copy the data from.
	 */
	void CopyMembers(const SingleRemoteVIM& source);

    /*
	 * Tells Curl to perform the http operation.
	 */
    void CurlPerform(void) const;

	////////////////////////////////////////////////////////////////////////////
	// constructors (all private)
	////////////////////////////////////////////////////////////////////////////
	/**
	  * Default constructor
	  */
    SingleRemoteVIM(void);

	/**
	 * Overwrite a SingleRemoteVim by assigning an existing SingleRemoteVIM
	 * @param inSingleRemoteVIM a reference to the SingleRemoteVIM to copy from
	 * @returns the number of bytes
	 */
    SingleRemoteVIM& operator = (const SingleRemoteVIM& inSingleRemoteVIM);

	/**
	 * Create a SingleRemoteVim copying from a reference 
	 * @param inSingleRemoteVIM a reference to the SingleRemoteVIM to copy from
	 */
	SingleRemoteVIM(const SingleRemoteVIM& inSingleRemoteVIM);

	////////////////////////////////////////////////////////////////////////////
	// destructor (private)
	////////////////////////////////////////////////////////////////////////////
    virtual ~SingleRemoteVIM(void);
};

#endif // __SingleRemoteVIM_H__