#ifndef __XAM_LOGBASE__
#define __XAM_LOGBASE__
////////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2007, The Storage Networking Industry Association.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - Neither the name of The Storage Networking Industry Association (SNIA)
//      nor the names of its contributors may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
// ============================================================================
//  Copyright (c) 2007, EMC Corporation.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - The name of EMC Corporation may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
////////////////////////////////////////////////////////////////////////////////
//  XAM SDK Source File Build Version @version.full@
////////////////////////////////////////////////////////////////////////////////

#include "XAM_LogTypes.h"
#include "XAM_POSIX.h"
#include "XAM_LogState.h"

/* This is the string which will appear in log rollover files 
 * (e.g. my_log_ovr1.log)
 */
static const char*  XAM_LOG_ROLLOVER_STRING = "_ovr";

/**
 * Pure abstract class that must be specialized for a specific log formatting.
 *
 * (e.g. tab delimited format, etc). It defines the interface available on 
 * all formatter types.
 *
 *  XAM_LogFormatter derived classes are typically used as follows:
 *
 *   - instantiate a new logger (refer to the constructor documentation 
 *     of the various XAM_LogFormatter implementations for details on arguments.)
 *   - Call the log or logList methods in the base class.
 *   - The base implementation calls the derived logEvent() methods.
 *   - The logEvent() and other derived methods use the services in the base 
 *     class for writing to the log file.
 *
 *  Note that this same interface may be used for formatters which do
 *  not write to a file directly. A formatter may be created which directs
 *  logging output to a third party logging framework.
 */
class XAM_LogFormatter
{
    protected:
             
        XAM_LogState mLogState;       // Active log state configuration

        FILE* mLogFile;               // FILE handle to the actual log file 
        xam_boolean mKeepExisting;    // If true, keep a backup of the file 
                                      // before overwriting it           
        xam_int mBytesWritten;        // Keep track of the number of bytes 
                                      // written

       /**
         *  Return the verbosity string corresponding with the argument
         *  @param inVerbosity Verbosity to obtain
         *  @return pointer to a string corresponding with the verbosity 
         *          argument
         */
        const char* 
        getVerbosityString(const loglevel_t inVerbosity);

        /**
         * Returns a component string associated with a component   
         *
         * @param inComponent A component mask containing a single component
         * @return A string corresponding to the inComponent parameter
         */         
        const char*
        getComponentMaskString(const component_t inComponent); 

    public:

        /**
         * Default Constructor
         */
        XAM_LogFormatter();

        /**
         * Destructor
         */
        virtual 
        ~XAM_LogFormatter();

       /**
        * Adds infomation to the log using logEvent
        * @param inLevel     Verbosity level
        * @param inComponent the kind of event to log
        * @param inFormat    the event message format string
        * @param ...         parameter list
        */
        void 
        log (const loglevel_t inLevel,
             const component_t inComponent,
             const char *inFormat,
             ...);

       /**
        *  Adds information to the log using the derived formatter classes
        *  Derived classes MUST define this method, and should make use of the
        *  logToFile service methods to write information to the file.
        *
        *  @param inLevel     Verbosity level
        *  @param inComponent the kind of event to log
        *  @param inFormat    the event message format string
        *  @param ap          parameter list
        */
        virtual void 
        logEvent (const loglevel_t inLevel,
                  const component_t inComponent,
                  const char *inFormat,
                  va_list ap) = 0;

        /**
         * This hook method will be called by the createLogFile after file 
         * creation. It may be defined by derived formatters in order to write
         * header data into the file.
         */
        virtual void 
        postOpenEvent () {}

        /**
         * This hook method will be called whenever files are closed using the
         * rollover services. It may be defined by derived formatters in order 
         * to write footer data into the file (ie. &lt;/xml&gt;).
         */
        virtual void 
        preCloseEvent () {}

        /**
         * Retrieve the length of the initial string automatically generated 
         * for each log event. It's used in determining whether or not we 
         * should rollover at the beginning of a log line.
         */
        virtual const xam_int 
        getPrependLength() 
        { 
            return(0); 
        }

        /**
         * Update this object to reflect a new logging state
         */
        virtual void 
        update(XAM_LogState* inState);

       /**
        *  Handle a log rollover event
        *  @param inFormat           the event message format string
        *  @param ap                 parameter list
        *  @throws XAM_FileSystemException on error
        */
        void 
        handleRolloverEvent(const char* inFormat, va_list ap);

       /**
        *  This method is called for each successful file write with an xam_int 
        *  parameter indicating the number of bytes written as a result of this
        *  call.
        */
        void 
        bytesWritten(const xam_int inNumWritten);

       /**
        *  Get the current time
        *  @param outTimestamp Variable filled with the current time
        */
        void 
        getTimestamp(struct timeval& outTimestamp);

       /**
        * Create and return a pointer to the log file
        * @param inKeepExisting If set, change the filename if one already 
        *                       exists
        * @return A pointer to the created file; NULL on error.
        * @throws XAM_FileSystemException on error
        */
        FILE*
        createLogFile(const xam_boolean inKeepExisting);

        /**
         * Service provided to derived classes in order to support rollover.
         * Rolls all existing log files over to the next highest number 
         * ("max rollover" times allowed)
         */
        void 
        rolloverLogFiles();

        /**
         *  Service provided to derived classes in order to support rollover.
         *  Determines if the configured rollover threshold has been reached.
         *  @return true if threshold reached, false otherwise
         *  @throws XAM_OutOfMemoryException
         */
        const xam_boolean 
        rolloverThresholdReached(const char* inFormat, 
                                 va_list ap, 
                                 const xam_int inBytesWritten);

        /**
         * Service provided to derived classes.
         * Logs a given format string to the provided file handle.
         */
        const int 
        logToFile(FILE* inFile, const char* inFormat, ...);

        /**
         * Service provided to derived classes.
         * Logs a given format string to the provided file handle.
         * This version may be used to explictly handle an existing va_list.
         */
        const int 
        logListToFile(FILE* inFile, const char* inFormat, va_list ap);

        /**
         * Returns a pointer to the full path to the active log file
         */ 
        const char* 
        getLogPath() 
        {
            return mLogState.getLogPath();
        }

    private:
        /**
         * Allocates enough memory to fit the resulting string
         *
         * @param inFormat
         * @param ap
         * @param outLen  The length of the allocated buffer
         * @return The string in an allocated buffer. Caller must deallocate 
         *          with XAM__free()!
         */
        char* 
        allocateBuffer(const char* inFormat, va_list ap, xam_int &outLen);

        /**
         * Private Copy Constructor (disallow copy)
         */
        XAM_LogFormatter(XAM_LogFormatter&) {}
};


#endif // __XAM_LogFormatter__

