#ifndef __XAM_LOGSTATE__
#define __XAM_LOGSTATE__
////////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2007, The Storage Networking Industry Association.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - Neither the name of The Storage Networking Industry Association (SNIA)
//      nor the names of its contributors may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
// ============================================================================
//  Copyright (c) 2007, EMC Corporation.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - The name of EMC Corporation may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
////////////////////////////////////////////////////////////////////////////////
//  XAM SDK Source File Build Version @version.full@
////////////////////////////////////////////////////////////////////////////////

#include "XAM_LogTypes.h"
#include "XAM_Properties.h"

/**
 * A class that encapsulates all log state information and provides 
 * serialization/de-serialization (save/open) to and from a 
 * properties file.
 */
class XAM_LogState 
{
    private:
         XAM_Properties* mProps;          // Properties for serialization/
                                          // deserialization

         xam_int mLogLevel;               // Log level 
                                          // (e.g. debug, errors, warnings)
         xam_int mVerbosityLevel;         // Verbosity level (component level)
         char* mLogFileName;              // Specified log file name
         xam_int mLogFormat;              // The log format to use 
                                          // (e.g. XML, TAB)
         xam_int mMaxLogRollovers;        // Max number of rollover files kept
         xam_int mCompBitmask;            // Mask of components to actively log
         xam_boolean mLogAppend;          // Append to existing file (true), 
                                          // or overwrite (false)
         xam_int mMaxLogSizeInKb;         // Max size in KB allowed before 
                                          // a rollover event occurs
         xam_string mCompFilter;          // A component string filter
         xam_string mMsgFilter;           // A msg string filter


         /*
          * Deallocates all resources used by this object.
          * Should be called anytime there's a state change that
          * reloads all options and at destruct time.
          */
         void 
         cleanup();

        /*
         * Reads the env vars associated with all log state options, and 
         * updates the outLogOptions with any values found.
         */
        void 
        loadEnvOptions();

        /*
         *  Loads the default options into the provided XAM_LogState object.
         *  Will read the corresponding env vars for initialization, or will
         *  set a default value if the env var has not been set.
         *
         */
        void 
        loadDefaultOptions();

        /**
         * Get a component mask for the given debug level
         *
         * @param inDebugLevel The debug level
         * @return A component bit mask based on the debug log level param
         */
        xam_int
        getComponentList(xam_int inDebugLevel);

    public:
         /**
          * Loads all default parameters
          *
          * @throws XAM_FileSystemException if the default config file
          *  is found in the working directory, but can't be read.
          */
         void 
         loadDefaults();

         /**
          * Loads all parameters from the given file.
          * Deserializes/reads an XAM_LogState properties file.
          *
          * @throws XAM_ParameterException if the inFilePath is null.
          * @throws XAM_FileSystemException if the inFilePath can't be read
          */
         void 
         loadFile(const char* inFilePath);

         /**
          * Serializes/creates an XAM_LogState properties file
          * based on the current state of this object.
          */
         void 
         writeToFile(const char* inFilePath);

         /**
          * Set the log file path
          *
          * @param inFilePath The full path to the requested log file target
          */
         void 
         setPath(const char* inFilePath);

         /**
          * Get the log file path
          *
          * @return Pointer to the current log path
          */
         const char* 
         getLogPath();

         /**
          * Set the append mode
          *
          * @param inAppend true for append, false for overwrite
          */
         void 
         setAppend(const xam_boolean inAppend);

         /** 
          * Get the append mode 
          *
          * @return true for append, false for overwrite
          */
         const xam_boolean 
         getAppend();

         /**
          * Set the log format
          *
          * @param inFormat The integer corresponding to the requested 
          *                 formatter class (e.g. TAB=1, XML=2)
          */
         void 
         setFormat(const xam_int inFormat);

         /**
          * Get the log format
          *
          * @return The index into the log formatter array
          */
         const xam_int 
         getFormat();

         /**
          * Set the log level
          */
         void 
         setLevel(const xam_int inLevel);

         /**
          * Get the log level
          *
          * @return The current log level
          */
         const xam_int 
         getLevel();

         /**
          * Set the debug log level
          */
         void 
         setVerbosityLevel(const xam_int inLevel);

         /**
          * Get the debug verbosity log level
          *
          * @return The current verbosity log level
          */
         const xam_int 
         getVerbosityLevel();

         /**
          * Get the component mask for the current debug level
          *
          * @return The component mask associated with the debug level
          */
         const xam_int 
         getComponentMask();

         /**
          * Set the maximum allowed log file size in KB
          */
         void 
         setMaxSizeInKb(const xam_int inMaxKbSize);

         /**
          * Get the maximum allowed log file size in KB
          *
          * @return The max size a log file is a llowed to grow (in KB)
          */
         const xam_int 
         getMaxSizeInKb();

         /**
          * Set the maximum number of log rollover files to retain
          */
         void 
         setMaxRollovers(const xam_int inMaxRollovers);

         /**
          * Get the maximum number of log rollover files to retain
          *
          * @return The mamimum number of log file rollovers to retain.
          */
         const xam_int 
         getMaxRollovers();

         /**
          * Set the message filter string
          */
         void 
         setMsgFilter(const char* inMsg);

         /**
          * Get the message filter string
          *
          * @return The current message filter string
          */
         const char* 
         getMsgFilter();

         /**
          * Set the component filter string
          */
         void 
         setCompFilter(const char* inMsg);

         /**
          * Get the component filter string
          *
          * @return The current component filter string
          */
         const char* 
         getCompFilter();

         /**
          * Sets the current log state data structure
          */
         void 
         setLogState(XAM_LogState* inLogOptions);

         /**
          * Determines whether or not logging should be enabled for this state
          */
         xam_boolean 
         isEnabled();

         /**
          * Default constructor 
          *
          *    Uses all defaults (initialized from env vars)
          */
         XAM_LogState();

         /**
          * Constructor which loads an existing log state configuration file
          *
          *  @param inFilePath Full path to a config properties file on disk
          */
         XAM_LogState(char* inFilePath); 

         /**
          * Destructor
          */
         ~XAM_LogState();
};


#endif // __XAM_LOGSTATE__


