////////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2007, The Storage Networking Industry Association.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - Neither the name of The Storage Networking Industry Association (SNIA)
//      nor the names of its contributors may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
// ============================================================================
//  Copyright (c) 2007, EMC Corporation.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - The name of EMC Corporation may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
////////////////////////////////////////////////////////////////////////////////
//  XAM SDK Source File Build Version @version.full@
////////////////////////////////////////////////////////////////////////////////

#include "XAM_Logger.h"
#include "XAM_Exception.h"
#include "XAM_Utils.h"
#include "XAM_Router.h"

// The current log state
LoggingState XAM_Logger::mLogState = UNSET;

// This lock covers mLogFactory
XAM_Mutex XAM_Logger::mLoggerLock;

// The logger factory
XAM_LogFactory XAM_Logger::mLogFactory;

char* XAM_Logger::mMsgBuf = (char*)0; // Log message buffer
xam_int XAM_Logger::mMsgBufLen = 256; // Length of log message buffer

// The verbosity level of the active logger
xam_int XAM_Logger::mActiveVerbosityLevel = (xam_int) LOGLEVEL_DEBUG_PLUS;

// The component filter of the active logger (default to all enabled)
xam_int XAM_Logger::mActiveComponentFilter = (xam_int) component_all;

//-----------------------------------------------------------------------------

xam_boolean
XAM_Logger::isEnabled ()
{
    return mLogState ? true : false;
}

xam_boolean
XAM_Logger::isEnabled(const loglevel_t inLevel, const component_t inComponent)
{
    xam_boolean vRetVal = false;

    if ((inLevel <= mActiveVerbosityLevel) && (inComponent & mActiveComponentFilter))
        vRetVal = isEnabled();

    return vRetVal;
}

xam_boolean
XAM_Logger::logList (const loglevel_t inLevel,
                     const component_t inComponent,
                     const char *inFormat,
                     va_list ap)
{
    xam_boolean vRetVal = true;

    switch(mLogState)
    {
    case UNSET:
        {
            initLogging();

            // If we just turned logging on, just fall through to 
            // the ON case. Break otherwise.
            if(mLogState != ON)
                break;
        }
    case ON:
        {
            // Only log this event if it falls within the configured constraints
            // (performance consideration)
            if ((inLevel <= mActiveVerbosityLevel) && (inComponent & mActiveComponentFilter))
            {
                vRetVal = logToFormatter(inLevel, inComponent, inFormat, ap);
            }
        }
    }

    return vRetVal;
}

xam_boolean 
XAM_Logger::logToFormatter(const loglevel_t inLevel, 
                           const component_t inComponent, 
                           const char* inFormat, 
                           va_list ap)
{
    xam_boolean vRetVal = true;

    // This lock covers XAM_LogFactory, XAM_LogFormatter, 
    // and our mMsgBuf buffer
    XAM_CriticalSection vCritical(&mLoggerLock);

    XAM_LogFormatter* vLog = mLogFactory.getInstance();
    if (vLog)
    {
        xam_int vResult = 0 ;

        // Allocate the log message buffer as needed
        if (mMsgBuf == (char*)0)
        {
            mMsgBuf = 
                (char*) XAM_POSIX::xam_malloc ((size_t) mMsgBufLen) ;

            if ( mMsgBuf == (char*)0 )
            { 
                XAMOutOfMemoryException e;
                throw e;
            }
        }

        vResult = XAM_POSIX::xam_vsnprintf (mMsgBuf, 
            (size_t) mMsgBufLen-1,
            inFormat, 
            ap) ;

        if ((vResult < 0) || (vResult >= (mMsgBufLen - 1)))
        {
            XAM_POSIX::xam_free (mMsgBuf) ;
            mMsgBuf = (char*)0 ;
            mMsgBufLen += 256 ;
            vRetVal = false;
        }

        if(vRetVal)
        {
    #ifdef __MVS__
            XAM_Utils::convert_ASCII_to_EBCDIC (mMsgBuf, mMsgBuf, strlen(mMsgBuf));
    #endif             
            try 
            {       
                // Log the message using the configured formatter object
                vLog->log(inLevel, inComponent, mMsgBuf);
            }
            catch(XAM_Exception&)
            {
                // An error occured, disable logging and rethrow to API 
                // level. Need to disable logging in case the API call 
                // logs exceptions.
                mLogState = OFF;
                throw;
            }
        }

        // Update the active logging parameters
        mActiveVerbosityLevel = mLogFactory.getLevel();
        mActiveComponentFilter = mLogFactory.getComponentMask();

    } // if(vLog)

    return vRetVal;
}

void 
XAM_Logger::initLogging(xam_boolean inObtainLock)
{
    if(inObtainLock)
    {
        XAM_CriticalSection vCritical(&mLoggerLock); 
        XAM_LogFormatter* vLog = mLogFactory.getInstance();    
        initLoggingHelper(vLog);
    }
    else
    {
        XAM_LogFormatter* vLog = mLogFactory.getInstance();    
        initLoggingHelper(vLog);
    }
}

void 
XAM_Logger::initLoggingHelper(XAM_LogFormatter* inLog)
{

    // Logging has not been previously set up before.
    // Create a log factory instance if logging is enabled.
    LoggingState vPrevLogState = mLogState;

    // Update the active logging parameters
    mActiveVerbosityLevel = mLogFactory.getLevel();
    mActiveComponentFilter = mLogFactory.getComponentMask();

    if (inLog == (XAM_LogFormatter *)0)
        mLogState = OFF;
    else
    {
        mLogState = ON;
        if(vPrevLogState != mLogState)
            logSettings(inLog);
    }
}

void 
XAM_Logger::Log (const loglevel_t inLevel,
                 const component_t inComponent,
                 const char *inFormat,
                 ...)
{
    va_list ap;
    va_start(ap, inFormat);
    int i = 0;
    // if buffer was enlarged more than 5 times - there is something wrong 
    // with the log, just drop out
    while (i < 5 && (! logList (inLevel, inComponent, inFormat, ap)))
    {
        va_end(ap);
        va_start(ap, inFormat);
        i++;
    };
    va_end(ap);
}

void
XAM_Logger::logSettings(XAM_LogFormatter* inLogger)
{    
    const char* vLogPath = mLogFactory.getLogPath();

    inLogger->log(LOGLEVEL_DEBUG, 
                  component_logger, 
                  "  LOG_CONFIG[ %s = \"%i\" ]", 
                  XAM_LOG_LEVEL, 
                  mLogFactory.getLevel());

    inLogger->log(LOGLEVEL_DEBUG, 
                  component_logger, 
                  "  LOG_CONFIG[ %s = \"%i\" ]", 
                  XAM_LOG_VERBOSITY, 
                  mLogFactory.getVerbosityLevel());

    inLogger->log(LOGLEVEL_DEBUG, 
                  component_logger,  
                  "  LOG_CONFIG[ %s = \"%s\" ]", 
                  XAM_LOG_PATH, 
                  vLogPath ? vLogPath : "(null)");

    inLogger->log(LOGLEVEL_DEBUG, 
                  component_logger, 
                  "  LOG_CONFIG[ %s = \"%i\" ]", 
                 XAM_LOG_FORMAT, 
                 mLogFactory.getFormat());

    inLogger->log(LOGLEVEL_DEBUG, 
                  component_logger, 
                  "  LOG_CONFIG[ %s = \"%s\" ]", 
                  XAM_LOG_APPEND, 
                  mLogFactory.getAppend() ? "true" : "false");

    inLogger->log(LOGLEVEL_DEBUG, 
                  component_logger, 
                  "  LOG_CONFIG[ %s = \"%i\" ]", 
                  XAM_LOG_MAX_SIZE, 
                  mLogFactory.getMaxSizeInKb());

    inLogger->log(LOGLEVEL_DEBUG, 
                  component_logger, 
                  "  LOG_CONFIG[ %s = \"%i\" ]", 
                  XAM_LOG_MAX_ROLLOVERS, 
                  mLogFactory.getMaxRollovers()); 

    inLogger->log(LOGLEVEL_DEBUG, 
                  component_logger, 
                  "  LOG_CONFIG[ %s = \"%s\" ]", 
                  XAM_LOG_MSG_FILTER, 
                  mLogFactory.getMsgFilter()); 

    inLogger->log(LOGLEVEL_DEBUG, 
                  component_logger, 
                  "  LOG_CONFIG[ %s = \"%s\" ]", 
                  XAM_LOG_COMP_FILTER, 
                  mLogFactory.getCompFilter()); 
}

void
XAM_Logger::Start(XAM_LogState *inLogState) 
{
    XAM_LogState* vLogState = inLogState;
    if(vLogState == (XAM_LogState*)0) {
        // Start logging using all defaults (inited based on env vars)
        vLogState = new XAM_LogState();
    }

    // This lock covers XAM_LogFactory
    XAM_CriticalSection vCritical(&mLoggerLock);

    // Build a logger matching the current config  
    mLogFactory.applyOptions(vLogState);
    mLogState = ON;     
}

void
XAM_Logger::Log(xam_int inLevel, const char* inMessage)
{
    Log((loglevel_t)inLevel, component_app, inMessage);
}

xam_status
XAM_Logger::Stop()
{
    xam_status vStat = 0;

    // Covers the log factory for this state change
    XAM_CriticalSection vCritical(&mLoggerLock);

    // Close the log
    mLogState = OFF;
    mLogFactory.closeLog();   

    return vStat;
}

void
XAM_Logger::logStatus(xam_handle_t inHandle, xam_status inStatus)
{
    if(inStatus != 0)
    {
        xam_string vStatStr;
  
        try 
        {
            // XSystem handles can provide VIM specific tokens
            XAM_Router::XAM_GetErrorToken(inHandle, inStatus, &vStatStr);
        }
        catch(XAMInvalidHandleException &)
        {
            // All other handles must use XAM_LIBRARY_HANDLE tokens
            XAM_Router::XAM_GetErrorToken(XAM_LIBRARY_HANDLE, inStatus, &vStatStr);
        }

        if(XAM_POSIX::xam_strlen(vStatStr) > 0)
        {
            XAM_Logger::Log(LOGLEVEL_ERROR, 
                            component_api_status, "ERROR: %s (%i)", vStatStr, (int)inStatus);
        }
    }
}

void 
XAM_Logger::setState(LoggingState inLogState, xam_int inLevel, xam_int inComponentMask)
{
    mLogState = inLogState;

    // Update the active logging parameters
    mActiveVerbosityLevel = inLevel;
    mActiveComponentFilter = inComponentMask;
}

LoggingState 
XAM_Logger::getState()
{
    return mLogState;
}

