#ifndef __XAM_ARRAY_H__
#define __XAM_ARRAY_H__
////////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2007, The Storage Networking Industry Association.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - Neither the name of The Storage Networking Industry Association (SNIA)
//      nor the names of its contributors may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
// ============================================================================
//  Copyright (c) 2007, EMC Corporation.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - The name of EMC Corporation may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
////////////////////////////////////////////////////////////////////////////////
//  XAM SDK Source File Build Version @version.full@
////////////////////////////////////////////////////////////////////////////////

#include "xam_types.h"

#include "XAM_POSIX.h"
#include "XAM_Exception.h"

// Array element destruction mode
enum XADestructMode {
       kXA_USE_FREE = 0,    // Dealloc array elements with "xam_free"
       kXA_USE_DELETE = 1   // Dealloc array elements with "delete"
};

/**
 *  Template class for a dynamic array
 */
template <class AType> 
class XAM_Array
{
    private:
        int mStartSize;     // The initial size of the array 
        int mGrowSize;      // The size to grow at each reallocation
        int mElemCnt;       // The current element count
        int mArrSize;       // The current size of the array
        AType **mArr;       // The internal pointer array

        xam_boolean mDeepDestruct;  // Destruct all elements (defaults to true)
        int mElementDestructMode;   // The destruction mode (e.g. free/delete)

        /*
         * Allocate an array of a given size
         *
         * @param inSize The size of the array to allocate
         */
        AType**
        arrAlloc(const int inSize)
        {
            AType **vArr = NULL;
            try
            {
                // don't check this ptr as it isn't always a leak
                vArr = new AType*[inSize];
            }
            catch(...)
            {
                XAMOutOfMemoryException e;
                throw e;
            }
            return vArr;
        }

        /* disallow copy and asignment b/c the contained object may not 
         * be able to be copied or assigned (deep copy)
         */

        XAM_Array(const XAM_Array&) { }
        XAM_Array& operator= (const XAM_Array& inObj) { return *this; } 

    public:        

        /** 
            Default constructor

            @param pDeepDestruct If true the memory relating to the elements 
                                 contained in this object is released 
                                 (default is true).

            @param pElementDestructMode If kXA_USE_DELETE, during a deep 
                                        destruct this object will release 
                                        the elements memory using the 
                                        delete keyword (default).

                                        If kXA_USE_FREE, during a deep 
                                        destuct this object will release 
                                        the elements memory using 
                                        xam_free.
                                              
         */
        XAM_Array(xam_boolean pDeepDestruct = true, 
                  XADestructMode pElementDestructMode = kXA_USE_DELETE)
            : mStartSize(10), 
              mGrowSize(10), 
              mElemCnt(0),
              mArrSize(mStartSize), 
              mArr(arrAlloc(mStartSize)), 
              mDeepDestruct(pDeepDestruct),
              mElementDestructMode(pElementDestructMode) { }
        
        /** 
            Constructor for specifying memory allocation increments

            @param pStartSize The init size of this array

            @param pGrowSize How many more element locations will be allocated
                             when the array grows.

            @param pDeepDestruct If true the memory relating to the elements 
                                 contained in this object is released.

            @param pElementDestructMode If kXA_USE_DELETE, during a deep 
                                        destruct this object will release
                                        the elements memory using the 
                                        delete keyword.

                                        If kXA_USE_FREE, during a deep 
                                        destuct this object will release 
                                        the elements memory using 
                                        xam_free.
                                              
         */
        XAM_Array(int pStartSize, 
                  int pGrowSize, 
                  xam_boolean pDeepDestruct,
                  XADestructMode pElementDestructMode = kXA_USE_DELETE)
            : mStartSize(pStartSize), 
              mGrowSize(pGrowSize), 
              mElemCnt(0),
              mArrSize(mStartSize), 
              mArr(arrAlloc(mStartSize)),
              mDeepDestruct(pDeepDestruct),
              mElementDestructMode(pElementDestructMode) { }


        /** 
           Destructor
         */
        ~XAM_Array()
        {
            clear();
            delete[] mArr;
        }

        /** 
           Add an element to the end of the array

           @param inObj The pointer to add
           @return the index in the array where added
         */
        xam_int
        push_back(AType* inObj)
        {
            if (mElemCnt == mArrSize)
            {
                // grow the array
                AType **mTempArr;
                try
                {
                    mArrSize += mGrowSize;
                    mTempArr = mArr;
                    mArr = new AType *[mArrSize];
                    XAM_POSIX::xam_memcpy(mArr, 
                                          mTempArr, 
                                          mElemCnt * sizeof(AType *));
                    delete [] mTempArr;
                }
                catch(...)
                {
                    // restore the old situation
                    mArr = mTempArr;
                    mArrSize -= mGrowSize;

                    XAMOutOfMemoryException e;
                    throw e;
                }
            }
            mArr[mElemCnt++] = inObj;
            return (mElemCnt - 1);
        }

        /** 
           Insert an element to a specific array index

           @param inObj The pointer to add
           @param inIndex The index where to insert
           @return the actual index where added
         */
        xam_int
        insert(AType *inObj, xam_int inIndex)
        {
            xam_int vRetVal = 0;

            if (inIndex < 0)
                inIndex = 0;
    
            if (inIndex >= mElemCnt)
                vRetVal = push_back(inObj);
            else
            {
                if (mElemCnt == mArrSize)
                {
                    // grow the array
                    AType **mTempArr;
                    try
                    {
                        mArrSize += mGrowSize;
                        mTempArr = mArr;
                        mArr = new AType *[mArrSize];
                        XAM_POSIX::xam_memcpy(mArr, 
                                              mTempArr, 
                                              mElemCnt * sizeof(AType *));
                        delete [] mTempArr;
                    }
                    catch(...)
                    {
                        // restore the old situation
                        mArr = mTempArr;
                        mArrSize -= mGrowSize;

                        XAMOutOfMemoryException e(); 
                        throw e;
                    }
                }
                // Move the existing elements back one spot
                xam_int i;
                for (i = mElemCnt-1; i >= inIndex; i--)
                {
                    mArr[i+1] = mArr[i];
                }
                // Insert the new element
                vRetVal = i+1;
                mArr[vRetVal] = inObj;
                mElemCnt++;
            }
            return vRetVal;
        }


        AType*
        operator[](int inIndex) const
        {
          return mArr[inIndex];
        }

        /** 
           Locate an AType pointer

           @param inObj The pointer to find
           @return the actual index where found or -1 if not found
         */
        int
        find(const AType *inObj) const
        {
          for(int i = 0 ; i < mElemCnt ; ++i)
            if (mArr[i] == inObj)    
              return i ;
          return -1 ;
        }

        /** 
           Locate a value by dereferencing a stored AType pointer
           and determining if the value is stored at an index

           @param inObj The pointer to dereference
           @return the actual index where found or -1 if not found
         */
        int
        findByValue(const AType* inObj) const
        {
          for(int i = 0 ; i < mElemCnt ; ++i)
            if (*(mArr[i]) == *inObj)    
              return i ;
          return -1 ;
        }

        /** 
           Remove an AType pointer

           @param inIndex The index to remove
           @return the remove AType pointer
         */
        AType*
        remove(const int inIndex)
        {
            AType *vObj = mArr[inIndex];
            int i;
            for (i = inIndex; i < (mElemCnt - 1); ++i)
                mArr[i] = mArr[i + 1];
            mArr[i] = 0;
            --mElemCnt;
            return vObj;
        }

        /** 
           Remove an AType pointer

           @param inObj The AType pointer to remove
           @return true on success, false otherwise
         */
        xam_boolean
        remove(const AType *inObj)
        {
            int vIndex = find(inObj) ;
            if (vIndex >= 0)
               remove(vIndex);
            return vIndex >= 0;
        }

        /**
          Get the size of the array

          @return The size of the array
          */
        int
        size() const
        {
            return mElemCnt;
        }

        /**
         Clear the array
         */
        void
        clear()
        {
            for (; mElemCnt != 0;)
            {
                if (mDeepDestruct && mArr[mElemCnt-1])
                {
                    if (mElementDestructMode == kXA_USE_DELETE)   
                        delete mArr[mElemCnt-1];
                    else
                        XAM_POSIX::xam_free((void*) mArr[mElemCnt-1]);
                }
                mArr[--mElemCnt] = 0;
            }
        }

        /**
         Swap the contents of this array with another array

         @param inOther The array to swap element with 
         */
        void
        swap(XAM_Array &inOther)
        {
            xam_int vTempElemCnt = inOther.mElemCnt;
            xam_int vTempArrSize = inOther.mArrSize;
            AType **vTempArr = inOther.mArr;
            inOther.mElemCnt = this->mElemCnt;
            inOther.mArrSize = this->mArrSize;
            inOther.mArr = this->mArr;
            this->mElemCnt = vTempElemCnt;
            this->mArrSize = vTempArrSize;
            this->mArr = vTempArr;
        }
};

#endif // __XAM_ARRAY_H__


