#ifndef __XAM_PROPERTIES__
#define __XAM_PROPERTIES__
/******************************************************************************
 *
 * XAM_Properties.h
 *                                             
 * XAM SDK Header File Build Version @version.full@
 *
 *  Copyright (c) 2007, EMC Corporation.
 *
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *    - Redistributions of source code must retain the above copyright notice, 
 *      this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright 
 *      notice, this list of conditions and the following disclaimer in the 
 *      documentation and/or other materials provided with the distribution.
 *    - The name of EMC Corporation may not be used to endorse or promote 
 *      products derived from this software without specific prior written 
 *      permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

#include "XAM_AssociativeArray.h"
#include "XAM_Critical.h"

/** 
 *   Manages a hash table of name/value pairs that may
 *   serialized/deserialized to or from a file on disk.
 */
class XAM_Properties
{
    private:
        XAM_AssociativeArray mHashTable;   // The table of name/value pairs
        XAM_Mutex  mMutex;                 // Mutex covering this object
        char* mName;                       // Name of this properties file

    public:
        /**
         * Sets or updates the given name with the given string value. 
         * This method is thread safe.
         *
         * @param inName - Name of the property to modify
         * @param inValue - Value to associate with the name
         * 
         * @return true on success, false on failure.
         */
        xam_boolean 
        set(const char* inName, const char* inValue);

        /**
         * Sets or updates the given name with the given int value. 
         * This method is thread safe.
         *
         * @param inName - Name of the property to modify
         * @param inValue - Value to associate with the name
         * 
         * @return true on success, false on failure.
         */
        xam_boolean 
        set(const char* inName, xam_int inValue);

        /**
         * Remove the given property name from the map
         * This method is thread safe.
         *
         * @param inName - Name of the property to remove
         * @return true on success, false on failure.
         */
        xam_boolean 
        remove(char* inName);

        /**
         * Clears out the properties map.
         * This method is thread safe.
         */
        void 
        clear();

        /**
         * Returns the value in the map given the property name.
         * This method is thread safe.
         *
         * @param inName - Name of the parameter
         * @param outName The value is written here
         * @return true if successfully written to outName, false otherwise
         */
        xam_boolean 
        getValue(const char* inName, xam_string* const outName);

        /**
         * Returns the value in the map given the property name converted 
         * to a xam_int.
         *
         * This method is thread safe.
         *
         * @param inName - Name of the parameter
         * @return xam_int value converted from property string or -1 on error
         */
        xam_int 
        getIntValue(const char* inName);

        /**
         * Returns the number of properties in the map.
         */
        int 
        getSize();
        
        /**
         * Returns the name string found at the given index in the array.
         *
         * @param inIndex The index into the array. Must be >= 0 and < getSize()
         */
        const char*
        getNameAtIndex(int inIndex)
        {
            return mHashTable.getName(inIndex);
        }

        /**
         * Returns the value string found at the given index in the array.
         *
         * @param inIndex The index into the array. Must be >= 0 and < getSize()
         */
        const char*
        getValueAtIndex(int inIndex)
        {
            return mHashTable.getValue(inIndex);
        }

        /** 
         * Serializes the properties to the given file.
         * This method is thread safe.
         *
         * @param inFilePath - Full path to the properties file on disk.
         * @return true on success, false otherwise.
         * 
         * @throws XAM_FileSystemException if inFilePath cannot be written to.
         */
        xam_boolean 
        writeToFile(const char* inFilePath);

        /** 
         * Deserializes the given properties file to this properties object.
         * The properties object is cleared before reading the file.
         * This method is thread safe.
         *
         * @param inFilePath - Full path to the properties file on disk.
         * @return true on success, false otherwise.
         *
         * @throws XAM_FileSystemException if inFilePath cannot be read.
         */
        xam_boolean 
        readFromFile(const char* inFilePath);

        /**
         * Allocates a new string buffer and writes all property/value pairs to
         * it, delimited by new lines and with name/values separated by '='.
         * This method is NOT thread safe and requires the lock.
         *
         * @throws XAMOutOfMemoryException If the memory cannot be allocated
         */
        char* 
        toString();

        /** lock the XAM_Properties mutex for thread safety */
        void 
        lock () 
        { 
            mMutex.lock(); 
        }

        /** unlock the XAM_Properties mutex */
        void 
        unlock () 
        { 
            mMutex.unlock(); 
        }

        /**
         * Sub-class for locking a XAM_Properties instance
         * 
         * This class is useful for creating a locked scope on the stack, 
         * as it will lock this object on construction and will cleanly
         * unlock the object on destruction.
         */
        class XAM_PropertiesLocker 
        {      
            private:
                XAM_Properties *mProperties;
            public:
                void unlock () { 
                    if (mProperties) 
                        mProperties->unlock(); 
                    mProperties = NULL; 
                }

                XAM_PropertiesLocker (const XAM_Properties *inProps) {
                    mProperties = (XAM_Properties*) inProps; 
                    mProperties->lock(); 
                }

                ~XAM_PropertiesLocker () { unlock (); }      
        };

        XAM_Properties(char* inName);
        ~XAM_Properties();

    private:

        /*
         * Parses the name/value fields from the properties file 
         * into the local associative array
         *
         * @param vString A heap buffer containing the data read
         *                from the properties file
         * @throws XAM_Exception on error (vString is deallocated)
         */
        void parsePropertiesFile(char* vString);
};

#endif // __XAM_PROPERTIES__


