#ifndef __XAM_VIM_MANAGER_H__
#define __XAM_VIM_MANAGER_H__
////////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2007, The Storage Networking Industry Association.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - Neither the name of The Storage Networking Industry Association (SNIA)
//      nor the names of its contributors may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
// ============================================================================
//  Copyright (c) 2007, EMC Corporation.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - The name of EMC Corporation may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
////////////////////////////////////////////////////////////////////////////////
//  XAM SDK Source File Build Version @version.full@
////////////////////////////////////////////////////////////////////////////////

#include "XAM_Array.h"
#include "XAM_Critical.h"
#include "XAM_VIM.h"
#include "XAM_FieldContainer.h"
#include "XAM_AssociativeArray.h"

/** 
 *  A manager class for XAM_VIM objects
 *
 *  Given an XRI connection string, use the "selectVim" method to
 *  perform the VIM selection process. This process may involve
 *  loading a new VIM library or selecting a library which has
 *  already been loaded.
 */
class XAM_VIM_Manager
{
    private:

        static XAM_Array<XAM_VIM> mVimArray;  // active VIM objects
        static XAM_Mutex mVimArrayMutex;      // covers the VIM array
        static const char* kProtocol;         // Protocol string
        static const char* kVimToken;         // VIM name token

        // Namespaces for VIM aliases
        static const char* sVimAliasNamespace;      
        static const char* sVimStackAliasNamespace; 

        // Character sequence between VIM libraries (or alias) in a VIM stack
        static const char* sVimStackDelimiter;

        // Namespace for preloaded VIMs
        static const char* sVimPreloadNamespace;

        // Namespace for definind lookup order
        static const char* sVimLookupOrderNamespace;

    protected:

        /** 
         * Selects a VIM to use based on the XRI connection string.
         *
         *  If the requested VIM library has already been loaded, the 
         *  corresponding XAM_VIM object will be returned.
         *
         *  If the VIM name was specified (and has not yet been loaded), 
         *  the VIM library will be loaded into a new XAM_VIM object 
         *  and returned.
         *
         *  If the VIM name was not specified, the VIM discovery
         *  sequence will be initiated (see discoverVIM()). 
         *
         *  The caller may select a different VIM library calling this method 
         *  again with the same (already incremented) retry count parameter. 
         *
         *  If an error occurs during VIM selection, a XAM_Exception is thrown.
         *  In these cases, this call may be made again (with the same retry 
         *  count parameter) in order to select a different VIM.
         *
         *  When no additional VIMs can be selected, a definitive NULL 
         *  pointer is returned.        
         *
         * @param inXRI  An "XSystem Resource Identifier" connection string
         *
         * @param outVimCxnStr  The VIM connection string is written here. 
         *
         * @param outSpecifiedVimName If a VIM name was specified, true is
         *                            written here. Otherwise, false is 
         *                            written.
         *
         * @param inRetryCount The number of discovered VIMs to skip 
         *                     (defaults to 0). May be incremented for 
         *                     subsequent calls until a returned VIM works or 
         *                     a definitive NULL is returned.  
         *
         *                     If this value is non-zero, the XRI will not be 
         *                     parsed, and the buffers provided will be assumed
         *                     to contain the parsed information from a 
         *                     previous call.
         *
         * @return A pointer to a fully initialized XAM_VIM object or NULL if 
         *         no VIM could be selected and retry is not possible.
         * @throws XAM_Exception on error (subsequent retry calls are possible)
         * @throws XAMInvalidParamException on parameter error
         *
         */
        static XAM_VIM* 
        selectVim(const char* inXRI, 
                  xam_string &outVimCxnStr,
                  xam_boolean &outSpecifiedVimName,
                  int& ioRetryCount);
    private:
        /* 
         * Parses a valid "XRI" connection string
         *
         * @param inXRI      An "XSystem Resource Identifier" connection string
         *                   (e.g. "snia-xam://MyLibrary!MyConnectionString")
         *
         * @param outVimName   A pointer to a xam_string which the VIM name
         *                     will be written to (if specified). This string 
         *                     is cleared if the library name is not specified.
         *                      
         * @param outVimCxnStr   The VIM connection string parsed from the XRI
         *                       is written here. The caller provides the  
         *                       allocated xam_string.
         *
         * @throws XAM_Exception on error
         *
         */
        static void 
        parseXRI(const char* inXRI, 
                 xam_string &outVimName, 
                 xam_string &outVimCxnStr);

        /* 
         * Searches for a VIM library in a given directory     
         *
         * @param inLibName     The lib name to search for 
         *                      (or specify NULL for the first found) 
         * @param inSearchPath  The directory in which the VIM lib will be searched
         *
         * @param outFoundPath  A pointer to a xam_string (allocated by the caller)
         *                      where the full path to the located VIM library is 
         *                      written (if found).
         *
         * @param inRetryCount  If non-zero, requests that the Nth VIM library
         *                      located is returned to the caller.
         *
         * @return true if the VIM library was found, false otherwise
         *
         * @throws XAM_Exception on error
         *
         */
        static bool 
        findVimLibInPath(const char* inLibName, 
                         const char* inSearchPath, 
                         xam_string* const outFoundPath,
                         int inRetryCount=0);

        /* 
         * Searches for any available VIM to use for a XAM operation
         *
         *   If a VIM library has been previously loaded, the most
         *   recently loaded VIM will be used.
         *
         *   If no VIM has been loaded, the directory pointed to by
         *   the XAM_VIM_PATH env var will be searched for VIM libraries
         *   (only libraries with "vim" in the lib name will be searched).
         *
         *   If no VIM has been found, a XAMVimNotFoundException is thrown.
         *
         * @param inRetryCount The number of times this call was retried.
         *                     If non-zero, the inRetryCount determines 
         *                     which discovered VIM should be returned.
         *                     The retry count must start with 0 and
         *                     should be increased monotonically for each
         *                     call to this method until we get a 
         *                     XAMVimNotFoundException indicating that 
         *                     no other VIM libraries could be discovered.
         * @return A pointer to a fully initialized XAM_VIM object or NULL
         *         on any error
         * @throws XAMVimNotFoundException
         *
         */
        static XAM_VIM* 
        discoverVIM(int inRetryCount=0);

        /* 
         * Loads a VIM object corresponding to the specified VIM library.
         *
         *   First checks for an existing VIM table entry (if one exists). 
         *   If an entry for this library does not exists, the method will 
         *   attempt to build a new VIM object for this library, add it to 
         *   the array, and return a pointer to the VIM object.
         *
         * @param inLibPath  Path to a given VIM library
         * @param outAlreadyLoaded If this library was already loaded, true
         *                         is written here. Otherwise, false is written
         *
         * @return A pointer to a fully initialized XAM_VIM object or NULL
         *         on any failure
         */
        static XAM_VIM* 
        loadVIM(const char* inLibPath, xam_boolean* const outAlreadyLoaded);

        /*
         * Selects an existing VIM from the loaded VIM array 
         *
         * @param inLibPath  The name of the library requested OR
         *                   The full path to the library requested OR
         *                   NULL which is a request for the VIM at the 
         *                   given index.
         * @param inRetryCount Only valid if inLibPath is NULL. Allows the
         *                     caller to select the next index in the 
         *                     existing VIM array.
         * @param inAdjustLibPath If true, the lib name will be adjusted to
         *                        include the expected platform specific
         *                        prefix, suffix, and/or architecture strings.
         */
        static XAM_VIM*
        selectExistingVim(const char* inLibPath, 
                          int inRetryCount = 0, 
                          xam_boolean inAdjustLibPath = false);

        /*
         * Obtains the count of lookuporder configuration fields
         *
         * @return The number of times to call into selectVimFromLookupOrderConfig
         *         to check all fields. Note that this may be 1 more than the 
         *         total count of the fields in the case that the configuration
         *         field count starts at 1 instead of 0.
         */
        static int
        getLookupOrderConfigCount();

        /*
         * Selects a VIM from the "lookuporder" configuration fields:
         *    .xam.config.vim.lookuporder.<n>=<aliasvalue>
         *
         *   Where "n" is a positive integer value that is unique to the 
         *   entire lookuporder namespace and indicates the ordinal
         *   position of the VIM to utilize for the lookup.
         *
         *   And "aliasvalue" is any predefined VIM alias from the config.
         *
         * @param ioIndex The index in the lookuporder fields to check for 
         *                (must be >= 0). Is incremented by this call.
         * @param ioKeepLooking true is written here if the next index exists,
         *                      false if the next index does not exist.
         */
        static XAM_VIM*
        selectVimFromLookupOrderConfig(int& ioIndex, 
                                       xam_boolean& ioKeepLooking);

        /**
         * Builds a VIM alias map from predefined fields on the field container.
         *
         *    VIM aliases are defined under the namespace:
         *      .xam.config.vim.alias.{aliasvalue}={vimpath}
         *
         *    Where "aliasvalue" is the actual alias and "vimpath" is the 
         *    full path to the VIM library on the filesystem.
         *
         *    VIM stack aliases (aliases for chains of VIMs) are defined as:
         *      .xam.config.vim.alias.stack.{aliasvalue}={aliasvalue}:{aliasvalue}:...
         *
         * @param inContainer A field container to search for the alias fields
         * @param outVimAliasMap A caller provided hashmap reference
         * @param inIncludeVimAliases If true, normal VIM aliases are added to the map
         * @param inIncludeVimStackAliases If true, VIM stack aliases are added to the map
         */
        static void
        buildVimAliasMap(XAM_FieldContainer* inContainer,
                         XAM_AssociativeArray& outVimAliasMap,
                         xam_boolean inIncludeVimAliases = true,
                         xam_boolean inIncludeVimStackAliases = true);

        /*
         * Determine if there's an alias defined for the given name
         *
         * @param inContainer A field container to search for the alias fields
         * @param inAliasName The name to find in the alias list
         * @param outAliasValue The alias value is written here
         * @param inAliasMap If NULL, a map is generated. Otherwise this list is used.
         * @param inCheckAliasForAlias True to check aliases for an alias
         * @return false if no alias exists, or true if outAliasValue contains the alias.
         *         Also returns false if aliases are infinitely recursive.
         */
        static xam_boolean
        findAliasValue(XAM_FieldContainer* inContainer,
                       const char* inAliasName,
                       xam_string& outAliasValue,
                       XAM_AssociativeArray* inAliasMap = 0,
                       xam_boolean inCheckAliasForAlias = true);

        /*
         * Appends any configured name/value pairs for a given VIM alias 
         * to the connection string.         
         *
         *    VIM aliases paris are defined under the namespace:
         *      .xam.config.vim.alias.{aliasvalue}.param.{name}={value}
         *
         *    Where "aliasvalue" is the actual alias and name/value is the 
         *    property pair to append to the XRI using "name=value" syntax
         *    (separated by "&" for multiple pairs).
         *
         *    If a xam_string is not large enough to hold a name/value pairs, 
         *    then the name/value pair will not be appended to the connection 
         *    string.
         *
         * @param inContainer A container to search for the alias param fields
         * @param inSpecifiedVimName The VIM name to use for checking param fields
         * @param ioCxnStr The connection string to append name/value pairs to
         */
        static void
        appendConfiguredNameValuePairs(XAM_FieldContainer* inContainer,
                                       xam_string inSpecifiedVimName,
                                       xam_string& ioCxnStr);

    public:

        /**
         * Preloads a list of VIM Libraries based on predefined fields 
         * in the specified field container.
         *
         *    All xam_string fields in the following namespace are checked:
         *      .xam.config.vim.preload.{unique string}={aliasvalue} 
         *
         *    Where "aliasvalue" is a VIM alias defined in the namespace:
         *      .xam.config.vim.alias.{aliasvalue}={path to vim library}
         *
         *    Where "unique string" is any string which would result in
         *    a unique field name under the preload namespace.                 
         *
         * @param inContainer The field container to check for predefined fields
         */
        static void
        preloadVims(XAM_FieldContainer* inContainer);
};

#endif // __XAM_VIM_MANAGER_H__

