/******************************************************************************
 *
 * Copyright  2007, The Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *    - Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    - Neither the name of The Storage Networking Industry Association (SNIA)
 *      nor the names of its contributors may be used to endorse or promote
 *      products derived from this software without specific prior written
 *      permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * ============================================================================
 *  Copyright  2007, EMC Corporation.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *    - Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    - The name of EMC Corporation may not be used to endorse or promote
 *      products derived from this software without specific prior written
 *      permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/
/*****************************************************************************
 * XAMLibraryObj Version @version.full@
 *****************************************************************************/

package org.snia.xam;

import java.util.Calendar;
import java.util.Properties;
import java.util.Enumeration;
import org.snia.xam.natives.XAM_Native;
import org.snia.xam.toolkit.XAMXUID;
import org.snia.xam.XAMException;
import org.snia.xam.util.ISO8601Date; 

public class XAMLibraryObj implements XAMLibrary {
    
    private static XAMLibrary sSingleton      = null;
    private long              mHandle         = 1;
    static private String     mLOAD_EXCEPTION = "Error loading XAM Library or its dependencies";
    static String             sFORMAT         = "yyyy-MM-dd'T'HH:mm:ss.SSSZ";
    
    static String             sIntPrefix      = "xam.int.";
    static String             sDoublePrefix   = "xam.double.";
    static String             sXUIDPrefix     = "xam.xuid.";
    static String             sDatetimePrefix = "xam.datetime.";
    static String             sBooleanPrefix  = "xam.boolean.";  
    
    // /////////////////////////////////////////////////////////////////////////
    // XAM methods
    // /////////////////////////////////////////////////////////////////////////
    /**
     * Connect to the specified XAM archive and return an XSystem object to
     * allow the application to use the XAM API. XRI values are in the form VIM
     * names, and parameters are optional. Use of optional parameters is
     * specific to any XAM Storage System and its configuration. Note: A
     * connected XSystem instance is not fully usable until authenticated. 
     * SNIA Public Java API Reference
     * 
     * @param inXRI
     *            The XAM Storage System Resource specification
     * @return The XSystem object representing the connection
     * @throws InvalidXRIException -
     *             The XRI is null or malformed.
     * @throws ConnectException -
     *             A connection could not be made, or a VIM could not be found.
     * @throws XAMException -
     *             Another error exists with the underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    
    public XSystem connect(String inXRI) throws XAMException {
        
        // check for null XRI
        if (inXRI == null)
            throw new InvalidXRIException(
                    XAMErrors.sInvalidXRIExceptionMessage, XAMErrors.sInvalidXRIException);
        
        XSystem vXSystem = null;
        vXSystem = new XSystemImpl(inXRI, this);
        return vXSystem;
    }
    
    // /////////////////////////////////////////////////////////////////////////
    // field container property creation methods
    // /////////////////////////////////////////////////////////////////////////
    /**
     * Create a XAM Boolean valued property. XAM Boolean is <i>boolean</i> in
     * Java.
     * 
     * @param inFieldName
     *            The name of this property
     * @param inBinding
     *            Indicates the BINDING of the field, TRUE will create a bound
     *            field, FALSE will create an unbound field.
     * @param inValue
     *            The boolean value for this property
     * @throws InvalidFieldName -
     *             fieldName is null or malformed. FieldExistsException - The
     *             specified field already exists. MaximumFieldException - Too
     *             many fields in the XSet XSetException - The current mode is
     *             invalid, the XSet has open import/export streams, or the XSet
     *             is in a corrupt or abandoned state. XSystemException - The
     *             XSystem is in a corrupt or abandoned state. XAMException -
     *             Another error exists with the underlying XAM Storage
     *             System.XAMException
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public void createProperty(String inFieldName, boolean inBinding,
            boolean inValue) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        try {
            XAM_Native.XAM_CreateBoolean(mHandle, inFieldName, inBinding,
                    inValue);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
    }
    
    /**
     * Create a XAM Int valued property. XAM Int is <i>long</i> in Java.
     * 
     * @param inFieldName
     *            The name of this property
     * @param inBinding
     *            Indicates the BINDING of the field, TRUE will create a bound
     *            field, FALSE will create an unbound field.
     * @param inValue
     *            The boolean value for this property
     * @throws 
     *             InvalidFieldNameException - fieldName is null or malformed.
     *             FieldExistsException - The specified field already exists.
     *             MaximumFieldException - Too many fields in the XSet
     *             XSetException - The current mode is invalid, the XSet has
     *             open import/export streams, XSet is in a corrupt or abandoned
     *             state. XSystemException - The XSystem is in a corrupt or
     *             abandoned state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public void createProperty(String inFieldName, boolean inBinding,
            long inValue) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        try {
            XAM_Native.XAM_CreateInt(mHandle, inFieldName, inBinding, inValue);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
    }
    
    /**
     * Create a XAM Float valued property. XAM Float is <i>double</i> in Java
     * 
     * @param inFieldName
     *            The name of this property
     * @param inBinding
     *            Indicates the BINDING of the field, TRUE will create a bound
     *            field, FALSE will create an unbound field.
     * @param inValue
     *            The double value for this property
     * @throws 
     *             InvalidFieldNameException - fieldName is null or malformed.
     *             FieldExistsException - The specified field already exists.
     *             MaximumFieldException - Too many fields in the XSet
     *             XSetException - The current mode is invalid, the XSet has
     *             open import/export streams, XSet is in a corrupt or abandoned
     *             state. XSystemException - The XSystem is in a corrupt or
     *             abandoned state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public void createProperty(String inFieldName, boolean inBinding,
            double inValue) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        try {
            XAM_Native.XAM_CreateDouble(mHandle, inFieldName, inBinding,
                    inValue);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
    }
    
    /**
     * Create a XUID valued property.
     * 
     * @param inFieldName
     *            The name of this property
     * @param inBinding
     *            Indicates the BINDING of the field, TRUE will create a bound
     *            field, FALSE will create an unbound field.
     * @param inValue
     *            The XUID value of the property
     * @throws 
     *             InvalidFieldNameException - fieldName is null or malformed.
     *             FieldExistsException - The specified field already exists.
     *             MaximumFieldException - Too many fields in the XSet
     *             XSetException - The current mode is invalid, the XSet has
     *             open import/export streams, XSet is in a corrupt or abandoned
     *             state. XSystemException - The XSystem is in a corrupt or
     *             abandoned state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public void createProperty(String inFieldName, boolean inBinding,
            XUID inValue) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        // check for null XUID
        if (inValue == null)
            throw new InvalidArgumentException(
                    XAMErrors.sInvalidArgumentExceptionMessage, XAMErrors.sInvalidArgumentException);
        
        // get the bytes of the XUID
        byte[] vBytes = null;
        vBytes = inValue.toBytes();
        
        // create a XUID field
        try {
            XAM_Native.XAM_CreateXUID(mHandle, inFieldName, inBinding, vBytes);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
    }
    
    /**
     * Create a XAM String valued property. Note that Java strings may be much
     * longer than XAM strings. If the Java string is longer than the XAM MAX
     * String, an exception will be thrown.
     * 
     * @param inFieldName
     *            The name of this property
     * @param inBinding
     *            Indicates the BINDING of the field, TRUE will create a bound
     *            field, FALSE will create an unbound field.
     * @param inValue
     *            The String value of the property
     * @throws 
     *             InvalidFieldNameException - fieldName is null or malformed.
     *             FieldExistsException - The specified field already exists.
     *             MaximumFieldException - Too many fields in the XSet
     *             XSetException - The current mode is invalid, the XSet has
     *             open import/export streams, XSet is in a corrupt or abandoned
     *             state. XSystemException - The XSystem is in a corrupt or
     *             abandoned state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public void createProperty(String inFieldName, boolean inBinding,
            String inValue) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        // check for invalid argument.
        if ((inValue == null) || (inValue.length() > XAM_MAX_STRING))
            throw new InvalidArgumentException(
                    XAMErrors.sInvalidArgumentExceptionMessage, XAMErrors.sInvalidArgumentException);
        
        try {
            XAM_Native.XAM_CreateString(mHandle, inFieldName, inBinding,
                    inValue);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
    }
    
    /**
     * Create a XAM Date valued property. XAM Date is <i>java.util.Calendar</i>
     * in Java.
     * 
     * @param inFieldName
     *            The name of this property
     * @param inBinding
     *            Indicates the BINDING of the field, TRUE will create a bound
     *            field, FALSE will create an unbound field.
     * @param inValue
     *            The date and time value of the property
     * @throws 
     *             InvalidFieldNameException - fieldName is null or malformed.
     *             FieldExistsException - The specified field already exists.
     *             MaximumFieldException - Too many fields in the XSet
     *             XSetException - The current mode is invalid, the XSet has
     *             open import/export streams, XSet is in a corrupt or abandoned
     *             state. XSystemException - The XSystem is in a corrupt or
     *             abandoned state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public void createProperty(String inFieldName, boolean inBinding,
            Calendar inValue) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        // Check for NULL 
        assertOnInvalidDate(inValue);
        
        // Convert calendar object to ISO8601 formatted string 
        String vDateString = null;
        try {
        	vDateString = ISO8601Date.toISO8601String(inValue);	
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        
        // create the field
        try {
            XAM_Native.XAM_CreateDatetime(mHandle, inFieldName, inBinding,
                    vDateString);
        } catch (Exception vExcption){ 
            XAMErrors.throwConvertedException(mHandle, vExcption);
        }
    }
    
    // /////////////////////////////////////////////////////////////////////////
    // field container property mutator methods
    // /////////////////////////////////////////////////////////////////////////
    /**
     * Set the named field to the specified boolean value
     * 
     * @param inFieldName
     *            The name of the Property field to set.
     * @param inValue
     *            The boolean value to set
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. InvalidFieldTypeException -
     *             The field is not a xam_boolean stype. XSetException - The
     *             current mode is invalid, the XSet has open import/export
     *             streams, or the XSet is in a corrupt or abandoned state.
     *             XSystemException - The XSystem is in a corrupt or abandoned
     *             state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public void setProperty(String inFieldName, boolean inValue)
            throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        try {
            XAM_Native.XAM_SetBoolean(mHandle, inFieldName, inValue);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
    }
    
    /**
     * Set the named field to the specified Date/Time
     * 
     * @param inFieldName
     *            The name of the Property field to set.
     * @param inValue
     *            The Date/Time value to set
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. InvalidFieldTypeException -
     *             The field is not a xam_datetime stype. XSetException - The
     *             current mode is invalid, the XSet has open import/export
     *             streams, or the XSet is in a corrupt or abandoned state.
     *             XSystemException - The XSystem is in a corrupt or abandoned
     *             state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public void setProperty(String inFieldName, Calendar inValue)
            throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        // Check for invalid Calendar object 
        assertOnInvalidDate(inValue);
        
        // Convert calendar object to ISO8601 formated string 
        String vDateString = null;
        try {
        	vDateString = ISO8601Date.toISO8601String(inValue);	
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        
        // set the value on the field
        try {
            XAM_Native.XAM_SetDatetime(mHandle, inFieldName, vDateString);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
    }
    
    /**
     * Set the named field to the specified XAM Float
     * 
     * @param inFieldName
     *            The name of the Property field to set.
     * @param inValue
     *            The XAM Float value to set
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. InvalidFieldTypeException -
     *             The field is not a xam_float stype. XSetException - The
     *             current mode is invalid, the XSet has open import/export
     *             streams, or the XSet is in a corrupt or abandoned state.
     *             XSystemException - The XSystem is in a corrupt or abandoned
     *             state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public void setProperty(String inFieldName, double inValue)
            throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        try {
            XAM_Native.XAM_SetDouble(mHandle, inFieldName, inValue);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
    }
    
    /**
     * Set the named field to the specified XAM Int
     * 
     * @param inFieldName
     *            The name of the Property field to set.
     * @param inValue
     *            The XAM Int value to set
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. InvalidFieldTypeException -
     *             The field is not a xam_int stype. XSetException - The current
     *             mode is invalid, the XSet has open import/export streams, or
     *             the XSet is in a corrupt or abandoned state. XSystemException -
     *             The XSystem is in a corrupt or abandoned state. XAMException -
     *             Another error exists with the underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public void setProperty(String inFieldName, long inValue)
            throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        try {
            XAM_Native.XAM_SetInt(mHandle, inFieldName, inValue);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
    }
    
    /**
     * Set the named field to the specified String
     * 
     * @param inFieldName
     *            The name of the Property field to set.
     * @param inValue
     *            The String value to set. String length must not exceed the
     *            maximum XAM String limit.
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. InvalidFieldTypeException -
     *             The field is not a xam_string stype. XSetException - The
     *             current mode is invalid, the XSet has open import/export
     *             streams, or the XSet is in a corrupt or abandoned state.
     *             XSystemException - The XSystem is in a corrupt or abandoned
     *             state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public void setProperty(String inFieldName, String inValue)
            throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        // check for invalid argument
        if (inValue == null)
            throw new InvalidArgumentException(
                    XAMErrors.sInvalidArgumentExceptionMessage, XAMErrors.sInvalidArgumentException);
        // check for InvalidFieldTypeException
        if (inValue.length() > XAM_MAX_STRING)
            throw new InvalidFieldTypeException(
                    XAMErrors.sInvalidFieldTypeExceptionMessage, XAMErrors.sInvalidFieldTypeException);
        
        try {
            XAM_Native.XAM_SetString(mHandle, inFieldName, inValue);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
    }
    
    /**
     * Set the named field to the specified XUID
     * 
     * @param inFieldName
     *            The name of the Property field to set.
     * @param inValue
     *            The XUID value to set.
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. InvalidFieldTypeException -
     *             The field is not a xam_xuid stype. XSetException - The
     *             current mode is invalid, the XSet has open import/export
     *             streams, or the XSet is in a corrupt or abandoned state.
     *             XSystemException - The XSystem is in a corrupt or abandoned
     *             state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public void setProperty(String inFieldName, XUID inValue)
            throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        // check for invalid argument
        if (inValue == null)
            throw new InvalidArgumentException(
                    XAMErrors.sInvalidArgumentExceptionMessage, XAMErrors.sInvalidArgumentException);
        
        try {
            XAM_Native.XAM_SetXUID(mHandle, inFieldName, inValue.toBytes());
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
    }
    
    public boolean getBoolean(String inFieldName) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        boolean vBoolean = false;
        try {
            vBoolean = XAM_Native.XAM_GetBoolean(mHandle, inFieldName);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vBoolean;
    }
    
    public Calendar getDateTime(String inFieldName) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        // get the datetime from the field
        String vDateTime = null;
        try {
            vDateTime = XAM_Native.XAM_GetDatetime(mHandle, inFieldName);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        
        // create the calendar using the formatted date
        Calendar vCalendar = null;
        try {
        	vCalendar = ISO8601Date.toCalendar(vDateTime);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        
        return (vCalendar);
    }
    
    /**
     * Get the XAM Float value of the named field.
     * 
     * @param inFieldName
     *            The name of the property to get the value
     * @return The Java double value of the named field.
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. InvalidFieldTypeException -
     *             The field is not a xam_boolean stype. XSetException - The
     *             current mode is invalid, the XSet has open import/export
     *             streams, or the XSet is in a corrupt or abandoned state.
     *             XSystemException - The XSystem is in a corrupt or abandoned
     *             state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public double getDouble(String inFieldName) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        double vDbl = 0.0;
        try {
            vDbl = XAM_Native.XAM_GetDouble(mHandle, inFieldName);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vDbl;
    }
    
    /**
     * Get the XAM Int value of the named field.
     * 
     * @param inFieldName
     *            The name of the property to get the value
     * @return The Java long value of the named field.
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. InvalidFieldTypeException -
     *             The field is not a xam_boolean stype. XSetException - The
     *             current mode is invalid, the XSet has open import/export
     *             streams, or the XSet is in a corrupt or abandoned state.
     *             XSystemException - The XSystem is in a corrupt or abandoned
     *             state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public long getLong(String inFieldName) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        long vLong = 0;
        try {
            vLong = XAM_Native.XAM_GetInt(mHandle, inFieldName);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vLong;
    }
    
    /**
     * Get the XAM String value of the named field.
     * 
     * @param inFieldName
     *            The name of the property to get the value
     * @return The Java String value of the named field.
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. InvalidFieldTypeException -
     *             The field is not a xam_boolean stype. XSetException - The
     *             current mode is invalid, the XSet has open import/export
     *             streams, or the XSet is in a corrupt or abandoned state.
     *             XSystemException - The XSystem is in a corrupt or abandoned
     *             state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public String getString(String inFieldName) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        String vString = null;
        try {
            vString = XAM_Native.XAM_GetString(mHandle, inFieldName);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vString;
    }
    
    /**
     * Get the XUID value of the named field.
     * 
     * @param inFieldName
     *            The name of the property to get the value
     * @return The XUID value of the named field.
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. InvalidFieldTypeException -
     *             The field is not a xam_boolean stype. XSetException - The
     *             current mode is invalid, the XSet has open import/export
     *             streams, or the XSet is in a corrupt or abandoned state.
     *             XSystemException - The XSystem is in a corrupt or abandoned
     *             state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public XUID getXUID(String inFieldName) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        XUID vXUID = null;
        byte[] vXUIDBytes = new byte[XAMXUID.sXUIDLENGTH];
        try {
            vXUIDBytes = XAM_Native.XAM_GetXUID(mHandle, inFieldName);
            vXUID = new XAMXUID(vXUIDBytes);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vXUID;
    }
    
    // /////////////////////////////////////////////////////////////////////////
    // field container XStream methods (creation/accessor/mutator)
    // /////////////////////////////////////////////////////////////////////////
    /**
     * Create a new XStream valued field in the XSet. The new stream will be
     * opened in WRITE-Truncate mode, with a zero length value.
     * 
     * @param inFieldName
     *            The name of this property
     * @param inBinding
     *            Indicates the BINDING of the field, TRUE will create a bound
     *            field, FALSE will create an unbound field.
     * @param inMimeType
     *            The mime type of the new XStream.
     * @throws InvalidArgumentException -
     *             mimeType is null or malformed. InvalidFieldNameException -
     *             fieldName is null or malformed. FieldExistsException - The
     *             specified field already exists. MaximumFieldException - Too
     *             many fields in the XSet. XSetException - The current mode is
     *             invalid, the XSet has open import/export streams, or the XSet
     *             is in a corrupt or abandoned state. XSystemException - The
     *             XSystem is in a corrupt or abandoned state. XAMException -
     *             Another error exists with the underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public XStream createXStream(String inFieldName, boolean inBinding,
            String inMimeType) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        // check for InvalidArgumentException
        if ((inMimeType == null) || (inMimeType.length() > XAM_MAX_STRING))
            throw new InvalidArgumentException(
                    XAMErrors.sInvalidArgumentExceptionMessage, XAMErrors.sInvalidArgumentException);
        
        long vXStreamHandle = 0;
        try {
            vXStreamHandle = XAM_Native.XAM_CreateXStream(mHandle, inFieldName,
                    inBinding, inMimeType);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        
        // return the stream
        XStreamImpl vXStream = new XStreamImpl(vXStreamHandle, this, null);
        return vXStream;
    }
    
    /**
     * Open the named XStream for read or write. If the XStream has not been
     * opened, the data manipulation methods will throw an exception.
     * 
     * @param inFieldName
     * @param inMode
     * @return The XStream containing the data
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. FieldDoesNotExistException -
     *             The specified field does not exist. XSetException - The
     *             current mode is invalid, the XSet has open import/export
     *             streams, or the XSet is in a corrupt or abandoned state.
     *             XSystemException - The XSystem is in a corrupt or abandoned
     *             state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public XStream openXStream(String inFieldName, String inMode)
            throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        // check for InvalidArgumentException
        if ((inMode == null) || (inMode.length() > XAM_MAX_STRING))
            throw new InvalidArgumentException(
                    XAMErrors.sInvalidArgumentExceptionMessage, XAMErrors.sInvalidArgumentException);
        
        long vXStreamHandle = 0;
        try {
            vXStreamHandle = XAM_Native.XAM_OpenXStream(mHandle, inFieldName,
                    inMode);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        
        // return the stream
        XStreamImpl vXStream = new XStreamImpl(vXStreamHandle, this, null);
        return vXStream;
    }
    
    // /////////////////////////////////////////////////////////////////////////
    // field container general field methods (XStreams or properties)
    // /////////////////////////////////////////////////////////////////////////
    
    /**
     * Return the MIME type of the specified field.
     * 
     * @param inFieldName
     *            Name of the field from which to get the type
     * @return The string value of the field's type
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. FieldDoesNotExistException -
     *             The specified field does not exist. XSetException - The
     *             current mode is invalid, the XSet has open import/export
     *             streams, or the XSet is in a corrupt or abandoned state.
     *             XSystemException - The XSystem is in a corrupt or abandoned
     *             state. XAMException - Another error exists with the
     *             underlying XAM Storage System.XAMException
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public String getFieldType(String inFieldName) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        String vType = null;
        try {
            vType = XAM_Native.XAM_GetFieldType(mHandle, inFieldName);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vType;
    }
    
    /**
     * Return the length of the field, in bytes.
     * 
     * @param inFieldName
     *            Name of the field from which to get the length
     * @return The number of bytes for the length of the field
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. FieldDoesNotExistException -
     *             The specified field does not exist. XSetException - The
     *             current mode is invalid, the XSet has open import/export
     *             streams, or the XSet is in a corrupt or abandoned state.
     *             XSystemException - The XSystem is in a corrupt or abandoned
     *             state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public long getFieldLength(String inFieldName) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        long vLength = 0;
        try {
            vLength = XAM_Native.XAM_GetFieldLength(mHandle, inFieldName);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vLength;
    }
    
    /**
     * Return TRUE if the field is bound, FALSE otherwise.
     * 
     * @param inFieldName
     *            Name of the field from which to get the binding info
     * @return The boolean value of the binding mode of the field.
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. FieldDoesNotExistException -
     *             The specified field does not exist. XSetException - The
     *             current mode is invalid, the XSet has open import/export
     *             streams, or the XSet is in a corrupt or abandoned state.
     *             XSystemException - The XSystem is in a corrupt or abandoned
     *             state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public boolean getFieldBinding(String inFieldName) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        boolean vBinding = false;
        try {
            vBinding = XAM_Native.XAM_GetFieldBinding(mHandle, inFieldName);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vBinding;
    }

    /**
     * Set the Field Binding attribute to FALSE. This method will only work on
     * Fields contained in an XSet.
     * 
     * @param inFieldName
     *            Name of the field to set the binding value
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. FieldDoesNotExistException -
     *             The specified field does not exist. XSetException - The
     *             current mode is invalid, the XSet has open import/export
     *             streams, XSet is in a corrupt or abandoned state.
     *             XSystemException - The XSystem is in a corrupt or abandoned
     *             state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public void setFieldAsNonbinding(String inFieldName) 
        throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        try {
            XAM_Native.XAM_SetFieldAsNonbinding(mHandle, inFieldName);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
    }

    // /////////////////////////////////////////////////////////////////////////
    // field container general field methods (XStreams or properties)
    // /////////////////////////////////////////////////////////////////////////
    /**
     * Set the Field Binding attribute to TRUE. This method will only work on
     * Fields contained in an XSet.
     * 
     * @param inFieldName
     *            Name of the field to set the binding value
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. FieldDoesNotExistException -
     *             The specified field does not exist. XSetException - The
     *             current mode is invalid, the XSet has open import/export
     *             streams, XSet is in a corrupt or abandoned state.
     *             XSystemException - The XSystem is in a corrupt or abandoned
     *             state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public void setFieldAsBinding(String inFieldName) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);

        try {
            XAM_Native.XAM_SetFieldAsBinding(mHandle, inFieldName);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
    }
    
    /**
     * Return TRUE if the Read-Only attribute of the field is TRUE, return FALSE
     * otherwise.
     * 
     * @param inFieldName
     *            Name of the field from which to get the read-only info
     * @return The boolean value of the read-only mode of the property
     * @throws InvalidFieldNameException -
     *             fieldName is null or malformed. FieldDoesNotExistException -
     *             The specified field does not exist. XSetException - The
     *             current mode is invalid, the XSet has open import/export
     *             streams, or the XSet is in a corrupt or abandoned state.
     *             XSystemException - The XSystem is in a corrupt or abandoned
     *             state. XAMException - Another error exists with the
     *             underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public boolean getFieldReadOnly(String inFieldName) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        boolean vReadOnly = false;
        try {
            vReadOnly = XAM_Native.XAM_GetFieldReadOnly(mHandle, inFieldName);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vReadOnly;
    }
    
    /**
     * Removes the named field from the FieldContainer. This method applies only
     * to XSets.
     * 
     * @param inFieldName
     *            The name of the Field to remove
     * @throws 
     *             InvalidFieldNameException - fieldName is null or malformed.
     *             FieldDoesNotExistException - The specified field does not
     *             exist. XSetException - The current mode is invalid, the XSet
     *             has open import/export streams, or the XSet is in a corrupt
     *             or abandoned state. XSystemException - The XSystem is in a
     *             corrupt or abandoned state. XAMException - Another error
     *             exists with the underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public void deleteField(String inFieldName) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        try {
            XAM_Native.XAM_DeleteField(mHandle, inFieldName);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
    }
    
    // /////////////////////////////////////////////////////////////////////////
    // field container XIterator methods
    // /////////////////////////////////////////////////////////////////////////
    
    /**
     * Iterate over a set of fields belonging to the XAMLibrary object. These
     * iterators yield String objects, the name of the field selected. The
     * filter pattern is a simple regular expression limited to a prefix of
     * character literals. To find all fields begining with "com.example" make
     * the call openFieldIterator( "com.example" );
     * 
     * @param inPrefix
     *            A prefix string describing the fields to be returned by the
     *            iterator. The expression is simple, a series of byte values
     *            which are treated as a "prefix" to match. For example
     *            "com.example" will match all fields starting with
     *            "com.example".
     * @throws InvalidFieldNameException -
     *             prefix is null or malformed.
     * @throws XSetException -
     *             The current mode is invalid, the XSet has open import/export
     *             streams, or the XSet is in a corrupt or abandoned state.
     * @throws XSystemException -
     *             The XSystem is in a corrupt or abandoned state.
     * @throws XAMException -
     *             Another error exists with the underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     * @since 1.0
     */
    public XIterator openFieldIterator(String inPrefix) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inPrefix);
        
        // create a native iterator handle
        long vHandle = 0;
        try {
            vHandle = XAM_Native.XAM_OpenFieldIterator(mHandle, inPrefix);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        
        // create a field iterator that wraps the native iterator handle
        FieldIterator vIterator = new FieldIterator(vHandle, this, null);
        return vIterator;
    }
    
    /**
     * Returns a boolean value of TRUE indicating if the named field is
     * contained within the referenced FieldContainer object (XAM Library,
     * XSystem, or XSet); returns FALSE in all other cases/
     * 
     * @param fieldName -
     *            The name of the field to test.
     * @throws InvalidFieldName -
     *             fieldName is null or malformed.
     * @throws XSetException -
     *             The XSet is in a corrupt or abandoned state.
     * @throws XSystemException -
     *             The XSystem is in a corrupt or abandoned state.
     * @throws XAMException -
     *             Another error exists with the underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public boolean containsField(String inFieldName) throws XAMException {
        // check for invalid field name
        assertOnInvalidFieldName(inFieldName);
        
        boolean vFieldExists = false;
        try {
            vFieldExists = XAM_Native.XAM_ContainsField(mHandle, inFieldName);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vFieldExists;
    }
    
    // == Asynchronous Operations ==
    /**
     * This method will begin the asynchronous opening of XStream in either
     * readonly, writeonly or appendonly mode, based on the mode argument. The
     * specified callback will be invoked as part of the asynchronous opening.
     * To monitor the status of this operation, the application can poll the
     * Async instance that is generated by this method. A handle to an Async
     * instance is also passed to any provided callback method when that
     * callback method is invoked.
     * 
     * @param name
     *            The name of the stream to be opened.
     * @param mode
     *            The mode to open the stream, one of XStream.MODE_READ_ONLY,
     *            XStream.MODE_WRITE_TRUNCATE, XStream.MODE_WRITE_APPEND.
     * @param xopid
     *            The application supplied XOPID of this XAsync operation.
     * @param listener
     *            listener An optional parameter, may be null, of the
     *            XASyncListener object to be notified when the operation has
     *            completed.
     * @return XAsync object representing this operation
     * @throws XAMException
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe, the XAsyncListener must be
     *             coded in a thread safe manner
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public XAsync asyncOpenXStream(String inName, String inMode, long inXOPID,
            XAsyncListener inListener) throws XAMException {
        
        // check for invalid field name
        assertOnInvalidFieldName(inName);
        
        XAsync vRetVal = null;
        try {
            vRetVal = XAM_Native.XAM_AsyncOpenXStream(mHandle, inName, inMode,
                    inXOPID, inListener);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        // set the XAMLibrary and XSystem Handles
        ((XAsyncImpl) vRetVal).setXAMLibrary(this);
        return vRetVal;
    }
    
    // /////////////////////////////////////////////////////////////////////////
    // singleton pattern support methods
    // /////////////////////////////////////////////////////////////////////////
    /**
     * This synchronized method provides access to the XAMLibrary singleton
     * 
     * @return XAMLibrary singleton
     * @throws CloneNotSupportedException
     *             in all cases
     * @since 1.0
     */
    public static synchronized XAMLibrary getLibrary() throws XAMException {
        if (sSingleton == null)
            sSingleton = new XAMLibraryObj();
        return sSingleton;
    }
    
    /**
     * This synchronized method provides access to the XAMLibrary singleton
     * and loads the specified Properties
     * 
     * @param inProps The configuration Properties to load into the returned instance
     * @return XAMLibrary singleton
     * @throws CloneNotSupportedException
     *             in all cases
     * @since 1.0
     */
    public static synchronized XAMLibrary getLibrary(Properties inProps) throws XAMException {
        if (sSingleton == null)
            sSingleton = new XAMLibraryObj();
        ((XAMLibraryObj)sSingleton).loadConfig(inProps);        
        return sSingleton;
    }
        
    /**
     * This method prevents inadvertent copying of the singleton
     * 
     * @throws CloneNotSupportedException
     *             in all cases
     * @since 1.0
     */
    public Object clone() throws CloneNotSupportedException {
        throw new CloneNotSupportedException();
    }
    
    /**
     * @return the XAMLibrary Handle
     */
    long getHandle() {
        return mHandle;
    }
    
    /**
     * Checks to see whether the input Field Name is a valid xam_string.
     * 
     * @param inFieldName
     *            field name
     * @throws InvalidFieldNameException
     */
    private void assertOnInvalidFieldName(String inFieldName)
            throws InvalidFieldNameException {
        
        // check for invalid field name
        if ((inFieldName == null) || (inFieldName.length() > XAM_MAX_STRING))
            throw new InvalidFieldNameException(
                    XAMErrors.sInvalidFieldNameExceptionMessage, XAMErrors.sInvalidFieldNameException);
    }
    
    /**
     * Checks to see whether the input Calendar is a valid object.
     * 
     * @param vCal
     *            Calendar object
     * @throws sInvalidArgumentException
     */
    private void assertOnInvalidDate(Calendar vCal)
            throws InvalidArgumentException {
        
        // check for invalid field name
        if (vCal == null)
            throw new InvalidArgumentException(
                    XAMErrors.sInvalidArgumentExceptionMessage, XAMErrors.sInvalidArgumentException);
    }

    /**
     * Load configuration Properties into the XAMLibrary instance
     * @param inProps 
     * 			The Properties to load onto the XAMLibrary instance
     */
    private void loadConfig(Properties inProps) {
    	if(inProps != null) {
    	   // Create (or set) each property found in the config file
		   Enumeration e = inProps.keys();
		   while( e.hasMoreElements() )
		   {
			   String key = (String) e.nextElement();
			   String value = inProps.getProperty(key);
		
			   try {
				   // Attempt to create a Property field with this name/value
				   if(key.startsWith(sIntPrefix, 0))
					   createProperty( key, false, Long.parseLong(value));
				   else if(key.startsWith(sDoublePrefix, 0))
					   createProperty( key, false, Double.parseDouble(value));
				   else if(key.startsWith(sBooleanPrefix, 0))
					   createProperty( key, false, (Boolean.valueOf(value)).booleanValue());
				   else if(key.startsWith(sDatetimePrefix, 0))
					   createProperty( key, false, getDateTime(value));	
				   else if(key.startsWith(sXUIDPrefix, 0))
					   createProperty( key, false, new XAMXUID(value));					   
				   else
				   	   createProperty( key, false, value );
			   } catch(XAMException ignore) {	
				   
				   // Attempt to set the Property field values on "create" error.				   
				   try {
					   if(key.startsWith(sIntPrefix, 0))
						   setProperty( key, Long.parseLong(value));
					   else if(key.startsWith(sDoublePrefix, 0))
						   setProperty( key, Double.parseDouble(value));
					   else if(key.startsWith(sBooleanPrefix, 0))
						   setProperty( key, (Boolean.valueOf(value)).booleanValue());
					   else if(key.startsWith(sDatetimePrefix, 0))
						   setProperty( key, getDateTime(value));	
					   else if(key.startsWith(sXUIDPrefix, 0))
						   setProperty( key, new XAMXUID(value));						   
					   else
					   	   setProperty( key, value );  
				   } catch(XAMException ignore2) {
				 	  // Ignore exceptions in setting field values
					  // while reading configuration Properties
				   }
			   }
		   }
    	}
    }
    
    // /////////////////////////////////////////////////////////////////////////
    // constructor methods
    // /////////////////////////////////////////////////////////////////////////
    
    /**
     * Private constructor (singleton pattern)
     */
    private XAMLibraryObj() throws XAMException {
        // make sure the shared library is loaded
        try {
            XAM_Native.loadXAMLibrary();
        } catch (Throwable e) // ReMap 'throwable' to a catchable exception
        {
            throw new XAMException(XAMLibraryObj.mLOAD_EXCEPTION);
        }
    }
}
