/******************************************************************************
 *
 * Copyright  2007, The Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *    - Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    - Neither the name of The Storage Networking Industry Association (SNIA)
 *      nor the names of its contributors may be used to endorse or promote
 *      products derived from this software without specific prior written
 *      permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * ============================================================================
 *  Copyright  2007, EMC Corporation.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *    - Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    - The name of EMC Corporation may not be used to endorse or promote
 *      products derived from this software without specific prior written
 *      permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/
/*****************************************************************************
 * XAsyncImpl Version @version.full@
 *****************************************************************************/

package org.snia.xam;

import org.snia.xam.natives.XAM_Native;
import org.snia.xam.toolkit.XAMXUID;

/**
 * The XAsynch object represents an asynchronous XAM operations. The XAsynch
 * object contains the state information regarding the asynchronous operation.
 * Asynchronous operations are in one of two states: pending and completed. When
 * the operation is first initiated, it is in the pending state. Because the
 * operation has not completed, it is only possible to query whether the
 * operation has completed, retrieve the XOPID that was specified when the
 * operation was initiated, and to halt the operation.
 */

public class XAsyncImpl implements XAsync {
    
    protected long     mHandle     = 0;
    private XAMLibraryObj mXAMLibrary = null;
    private XSystem    mXSystem    = null;
    
    /**
     * This method stops the execution of the operation associated with the
     * XAsync instance. This method may be used at any time.
     * 
     * @throws XAMException
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public void halt() throws XAMException {
        
        // Check to see whether XStream is closed?
        assertClosed();
        
        try {
            XAM_Native.XAsync_Halt(mHandle);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
    }
    
    /**
     * This method retrieves the completed state of the operation associated
     * with the XAsync instance. It may be used at any time.
     * 
     * @throws XAMException
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public boolean isComplete() throws XAMException {
        
        // Check to see whether XStream is closed?
        assertClosed();
        
        boolean vRetVal = false;
        try {
            vRetVal = XAM_Native.XAsync_IsComplete(mHandle);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vRetVal;
    }
    
    /**
     * This method retrieves the XOPID of the operation associated with the
     * XAsync instance. This method may be used at any time.
     * 
     * @throws XAMException
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public long getXOPID() throws XAMException {
        
        // Check to see whether XStream is closed?
        assertClosed();
        
        long vRetVal = 0;
        try {
            vRetVal = XAM_Native.XAsync_GetXOPID(mHandle);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vRetVal;
    }
    
    /**
     * This method retrieves the XAM Status of the operation associated with the
     * XAsync instance. This method may be used after the operation has
     * transitioned to the completed state. If no error has occurred, the method
     * returns. If an error occurred, the exception generated will be thrown.
     * 
     * @throws XAMException
     *             Any exception possibly generated by the asynchronous
     *             operation.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public void getStatus() throws XAMException {
        
        // Check to see whether XStream is closed?
        assertClosed();
        
        long vRetVal = 0;
        try {
            vRetVal = XAM_Native.XAsync_GetStatus(mHandle);
        } catch (Exception vException) {
            // ignore the Exception as we throw the exception anyway
            // if the vRetVal is != 0.
        }
        if (vRetVal != 0) {
            XAMErrors.throwConvertedException(mHandle, (int) vRetVal);
        }
    }
    
    /**
     * This method retrieves the XSet reference associated with this XAsync
     * instance. This is only valid if the operation was an "asyncOpenXSet" or
     * "asyncCopyXSet".
     * 
     * @return The XSet reference from the XAsync.
     * @throws XAMException
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public XSet getXSet() throws XAMException {
        
        // Check to see whether XStream is closed?
        assertClosed();
        
        XSet vRetVal = null;
        long vXSetHandle = 0;
        try {
            vXSetHandle = XAM_Native.XAsync_GetXSet(mHandle);
            vRetVal = new XSetImpl(vXSetHandle, mXSystem);
            
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vRetVal;
    }
    
    /**
     * This method retrieves the XStream reference associated with this XAsync
     * instance. This is only valid if the operation was an "asyncOpenXStream".
     * This method may be used at any time.
     * 
     * @return The XStream reference from the XAsync.
     * @throws XAMException
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public XStream getXStream() throws XAMException {
        
        // Check to see whether XStream is closed?
        assertClosed();
        
        XStream vRetVal = null;
        long vXStreamHandle = 0;
        try {
            vXStreamHandle = XAM_Native.XAsync_GetXStream(mHandle);
            vRetVal = new XStreamImpl(vXStreamHandle, mXAMLibrary, mXSystem);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vRetVal;
    }
    
    /**
     * This method retrieves the XUID reference (if any) of the XSet associated
     * with this XAsync instance. This method may be used at any time.
     * 
     * @throws XAMException
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public XUID getXUID() throws XAMException {
        
        // Check to see whether XStream is closed?
        assertClosed();
        
        XUID vRetVal = null;
        byte[] vXUIDBytes = null;
        try {
            vXUIDBytes = XAM_Native.XAsync_GetXUID(mHandle);
            vRetVal = new XAMXUID(vXUIDBytes);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vRetVal;
    }
    
    /**
     * This method retrieves the number of bytes read by the operation
     * associated with this XAsync instance. Not all operations read bytes, and
     * for those operations it will always be set to zero. This method may be
     * used at any time.
     * 
     * @throws XAMException
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public long getBytesRead() throws XAMException {
        
        // Check to see whether XStream is closed?
        assertClosed();
        
        long vRetVal = 0;
        try {
            vRetVal = XAM_Native.XAsync_GetBytesRead(mHandle);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vRetVal;
    }
    
    /**
     * This method retrieves the number of bytes written by the operation
     * associated with this XAsync instance. Not all operations write bytes, and
     * for those operations it will always be set to zero. This method may be
     * used at any time.
     * 
     * @throws XAMException
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public long getBytesWritten() throws XAMException {
        
        // Check to see whether XStream is closed?
        assertClosed();
        
        long vRetVal = 0;
        try {
            vRetVal = XAM_Native.XAsync_GetBytesWritten(mHandle);
        } catch (Exception vException) {
            XAMErrors.throwConvertedException(mHandle, vException);
        }
        return vRetVal;
    }
    
    /**
     * This method releases the resources of the operation associated with the
     * XAsync instance and resources of the XAsync operation itself. This method
     * may be used after the operation has transitioned to the completed state.
     * 
     * @throws XAMException
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public void close() throws XAMException {
        // Check to see whether XStream is closed?
        assertClosed();
        
        synchronized(this){
            try {
                XAM_Native.XAsync_Close(mHandle);
                mHandle = 0;
            } catch (Exception vException) {
                XAMErrors.throwConvertedException(mHandle, vException);
            }
        }
    }
    
    /**
     * @return the XAsyncHandle
     */
    long getHandle() {
        return mHandle;
    }
    
    protected void finalize() {
        if (mHandle != 0) {
            try {
                close();
            } catch (Exception ignore) {
            }
            mHandle = 0;
        }
    }
    
    private void assertClosed() throws XAMException {
        synchronized(this){
            if (this.mHandle == 0)
                throw new XAMException("xam/XAsync already closed.");
        }
    }
    
    /**
     * Sets the XAMLibrary Object on this Async object
     * @param inXAMLibrary  inXAMLibrary Object
     */
    void setXAMLibrary(XAMLibraryObj inXAMLibrary) {
        mXAMLibrary = inXAMLibrary;
    }
    
    /**
     * Sets the XSystem Object on this Async object
     * 
     * @param XSystem inXSystem Object
     */
    void setXSystem(XSystem inXSystem) {
        mXSystem = inXSystem;
    }
    
    // /////////////////////////////////////////////////////////////////////////
    // constructor methods
    // /////////////////////////////////////////////////////////////////////////
    /**
     * Default constructor
     */
    XAsyncImpl(long inHandle) {
        mHandle = inHandle;
    }
}
