/******************************************************************************
 *
 * Copyright  2007, The Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *    - Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    - Neither the name of The Storage Networking Industry Association (SNIA)
 *      nor the names of its contributors may be used to endorse or promote
 *      products derived from this software without specific prior written
 *      permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * ============================================================================
 *  Copyright  2007, EMC Corporation.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *    - Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    - The name of EMC Corporation may not be used to endorse or promote
 *      products derived from this software without specific prior written
 *      permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/
/*****************************************************************************
 * CRC16 Version @version.full@
 *****************************************************************************/

package org.snia.xam.toolkit.util;

public class CRC16 {
    
    private final short  sWidth      = 16;
    private final int    sPoly       = 0x8005;
    private final int    sInit       = 0x0000;
    private final int    sXorOut     = 0x0000;
    private final int    sCheck      = 0xBB3D;
    static int           sCrcTable[] = null;
    private static CRC16 sInstance   = null;
    
    // private final String sName = "CRC-16";
    // private final boolean sRefIn = true;
    // private final boolean sRefOut = true;
    
    public static CRC16 getInstancec() {
        if (sInstance == null)
            sInstance = new CRC16();
        return sInstance;
    }
    
    /**
     * Reorder the bits of a binary sequence, by reflecting them about the
     * middle position.
     * 
     * @param data
     *            the original data
     * @param nBits
     *            number of bits
     * @return The reflection of the original data.
     */
    private static int reflect(int data, int nBits) {
        int vReflection = 0x00000000;
        short vBit;
        
        /*
         * Reflect the user data.
         */
        for (vBit = 0; vBit < nBits; ++vBit) {
            /*
             * If the LSB bit is set, set the reflection of it.
             */
            if ((data & 0x01) > 0) {
                vReflection |= (1 << ((nBits - 1) - vBit));
            }
            data = (data >> 1);
        }
        return vReflection;
    }
    
    /**
     * This method populates the CRC lookup table. To get higher performance
     * this method should only be called once and the CRC lookup table should be
     * cached.
     */
    
    private CRC16() {
        sCrcTable = new int[256];
        int vRemainder;
        int vDividend;
        char vBit;
        
        /*
         * Compute the remainder of each possible dividend.
         */
        for (vDividend = 0; vDividend < 256; ++vDividend) {
            /*
             * Start with the dividend followed by zeros.
             */
            vRemainder = vDividend << (sWidth - 8);
            
            /*
             * Perform modulo-2 division, a bit at a time.
             */
            for (vBit = 8; vBit > 0; --vBit) {
                /*
                 * Try to divide the current data bit.
                 */
                int vTopBit = (1 << (sWidth - 1));
                if ((vRemainder & vTopBit) > 0) {
                    vRemainder = (vRemainder << 1) ^ sPoly;
                } else {
                    vRemainder = (vRemainder << 1);
                }
            }
            
            vRemainder = vRemainder & 0xFFFF;
            /*
             * Store the result into the table.
             */
            sCrcTable[vDividend] = (short) vRemainder;
        }
    }
    
    /**
     * This method calculates the CRC-16 of nBytes from the provided byte[]. We
     * use the following parameters. Name : "CRC-16" Width : 16 Poly : 0x8005
     * Init : 0x0000 RefIn : True RefOut : True XorOut : 0x0000 Check : 0xBB3D
     * 
     * @param message
     *            byte[] consisting of the actual message that we need to run
     *            CRC-16
     * @param nBytes
     *            Number of bytes from the byte[] that we need to run CRC-16
     * @return CRC checksum
     */
    public int calcCRC(byte message[], int nBytes) {
        int vRemainder = sInit;
        short data = 0;
        int vByte = 0;
        int vReflection = 0;
        
        /*
         * Divide the message by the polynomial, a byte at a time.
         */
        for (vByte = 0; vByte < nBytes; ++vByte) {
            vReflection = reflect(message[vByte], 8) & 0xFFFF;
            vRemainder &= 0xFFFF;
            data = (short) ((vReflection) ^ ((vRemainder) >> 8));
            vRemainder = ((sCrcTable[data]) ^ ((vRemainder) << 8));
        }
        
        /*
         * The final remainder is the CRC.
         */
        return reflect(vRemainder & 0xFFFF, sWidth) ^ sXorOut;
    }
}