/* Copyright (c) 2008, The Storage Networking Industry Association.

   Redistribution and use in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

   * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

   * Neither the name of The Storage Networking Industry Association (SNIA) nor
   the names of its contributors may be used to endorse or promote products
   derived from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
   THE POSSIBILITY OF SUCH DAMAGE.
*/

/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */

/* * ============================================================================
   Copyright  2007, EMC Corporation.
 
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
     - The name of EMC Corporation may not be used to endorse or promote
       products derived from this software without specific prior written
       permission.
 
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.util;

import java.text.ParseException;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.TimeZone;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A class for converting Calendar objects into ISO8601 formatted
 * date time strings. The ISO8601 date time format looks like:
 * <pre>
 * [+|-]YYYYMMDDThhmmss.SSS[Z|TZD]
 * +|-  = Era designator, where - is BC
 * YYYY = The year as four digits, ex: 2007
 * MM   = The month of the year (01-12)
 * DD   = The day of the month (01-31)
 * T    = The literal character 'T'
 * hh   = Hour of day (0-23)
 * mm   = Minute of day (0-59)
 * ss   = Second of day (0-59)
 * SSS  = Millisecond of day (000-999)
 * Z    = Zero offset from GMT
 * TZD  = Time offset from GMT [+|-]hh:mm
 *        +|- = offset direction, - is before GMT
 *        hh  = hour offset
 *        mm  = minute offset
 * </pre>
 */
public class ISO8601Date
{
    protected static final String COLON = ":";
    protected static final String DECPT = ".";
    protected static final String MINUS = "-";
    protected static final String PLUS  = "+";
    protected static final String TIME  = "T";
    protected static final String ZULU  = "Z";

    protected static final int MIN_IN_MS  = 60 * 1000;
    protected static final int HOUR_IN_MS = 60 * MIN_IN_MS;

    protected static final String DATE =
        "([\\+\\-])?(\\d{4})\\-(\\d{2})\\-(\\d{2})T(\\d{2})\\:(\\d{2})\\:(\\d{2})(?:\\.(\\d{1,3}+))?(Z|[\\+\\-]\\d{2}\\:\\d{2})";

    protected static final Pattern  s_date = Pattern.compile(DATE);
    protected static final TimeZone s_utc  = TimeZone.getTimeZone("GMT");

    // Pattern group location
    private static final int ERA     =  1;
    private static final int YEAR    =  2;
    private static final int MONTH   =  3;
    private static final int DAY     =  4;
    private static final int HOUR    =  5;
    private static final int MIN     =  6;
    private static final int SEC     =  7;
    private static final int MS      =  8;
    private static final int TZ      =  9;

    /**
     * Converts a java.util.Calendar object to an ISO8601 formatted
     * date time string.
     *
     * @param cal The date time as a Calendar object.
     * @return The ISO8601 formatted date time string.
     */
    public static String toISO8601String(Calendar cal)
    {
        StringBuffer buf = new StringBuffer(30);
        if (cal.get(Calendar.ERA) == GregorianCalendar.BC)
            buf.append(MINUS);
        pad(buf, Integer.toString(cal.get(Calendar.YEAR)), 4, '0');
        buf.append(MINUS);
        pad(buf, Integer.toString(cal.get(Calendar.MONTH)+1), 2, '0');
        buf.append(MINUS);
        pad(buf, Integer.toString(cal.get(Calendar.DAY_OF_MONTH)), 2, '0');
        buf.append(TIME);
        pad(buf, Integer.toString(cal.get(Calendar.HOUR_OF_DAY)), 2, '0');
        buf.append(COLON);
        pad(buf, Integer.toString(cal.get(Calendar.MINUTE)), 2, '0');
        buf.append(COLON);
        pad(buf, Integer.toString(cal.get(Calendar.SECOND)), 2, '0');
        buf.append(DECPT);
        pad(buf, Integer.toString(cal.get(Calendar.MILLISECOND)), 3, '0');
        TimeZone tz = cal.getTimeZone();
        int offset = tz.getOffset(cal.get(Calendar.ERA),
            	cal.get(Calendar.YEAR),
            	cal.get(Calendar.MONTH),
            	cal.get(Calendar.DAY_OF_MONTH),
            	cal.get(Calendar.DAY_OF_WEEK),
            	cal.get(Calendar.MILLISECOND));
        if (offset != 0)
        {
            if (offset < 0)
            {
                buf.append(MINUS);
                offset = Math.abs(offset);
            }
            else
                buf.append(PLUS);
            int hours = offset / HOUR_IN_MS;
            int mins = (offset % HOUR_IN_MS) / MIN_IN_MS;
            pad(buf, Integer.toString(hours), 2, '0');
            buf.append(COLON);
            pad(buf, Integer.toString(mins), 2, '0');
        }
        else
            buf.append(ZULU);
        return buf.toString();
    }

    /**
     * Parses the string argument into a Calendar object. The
     * argument must be an ISO8601 date time string as specified
     * in the class description
     *
     * @param date An ISO8601 date formatted String
     * @return The date as a java.util.Calendar object
     * @throws ParseException
     * 		If the string does not contain a property formatted ISO-8601 date
     */
    public static Calendar toCalendar(String date)
        throws ParseException
    {
    	if (date == null)
    		throw new IllegalArgumentException("The date cannot be null");

        Matcher m = s_date.matcher(date);
        if (!m.matches())
            throw new ParseException("Invalid date format", 0);

        int year = Integer.parseInt(m.group(YEAR));
        if (year == 0)
            throw new ParseException("Invalid year", 0);

        GregorianCalendar cal = new GregorianCalendar(s_utc);
        String grp = m.group(ERA);
        if (MINUS.equals(grp))
            cal.set(Calendar.ERA, GregorianCalendar.BC);
        else
            cal.set(Calendar.ERA, GregorianCalendar.AD);
        cal.set(Calendar.YEAR, year);
        cal.set(Calendar.MONTH, Integer.parseInt(m.group(MONTH))-1);
        cal.set(Calendar.DAY_OF_MONTH, Integer.parseInt(m.group(DAY)));
        cal.set(Calendar.HOUR_OF_DAY, Integer.parseInt(m.group(HOUR)));
        cal.set(Calendar.MINUTE, Integer.parseInt(m.group(MIN)));
        cal.set(Calendar.SECOND, Integer.parseInt(m.group(SEC)));
        grp = m.group(MS);
        if (grp != null)
        	cal.set(Calendar.MILLISECOND, getMillisecond(grp));
        else
        	cal.set(Calendar.MILLISECOND, 0);
        String tz = m.group(TZ);
        if (!tz.equals(ZULU))
            cal.setTimeZone(TimeZone.getTimeZone("GMT"+tz));
        return cal;
    }

    protected static void pad(StringBuffer buf, String val, int count, char pad)
    {
        for (int i = val.length(); i < count; i++)
            buf.append(pad);
        buf.append(val);
    }

    protected static int getMillisecond(String val)
    {
    	int ms = Integer.parseInt(val);
    	int len = val.length();
    	if (len == 1)
    		ms *= 100;
    	else if (len == 2)
    		ms *= 10;
    	return ms;
    }
}
