/*
 * Copyright (c) 2009, Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
package org.snia.xam.vim.http.client;

import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.Calendar;
import java.util.HashMap;
import java.util.logging.Logger;

import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.FieldContainer;
import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.FieldExistsException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidFieldNameException;
import org.snia.xam.InvalidFieldTypeException;
import org.snia.xam.InvalidOperationException;
import org.snia.xam.MaximumFieldException;
import org.snia.xam.ObjectInUseException;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XIterator;
import org.snia.xam.XSetAbandonException;
import org.snia.xam.XSetCorruptException;
import org.snia.xam.XStream;
import org.snia.xam.XSystemAbandonException;
import org.snia.xam.XSystemCorruptException;
import org.snia.xam.XUID;
import org.snia.xam.base.DefaultXUID;
import org.snia.xam.vim.isolation.server.http.handlers.ProtocolConstants;

abstract class HTTPFieldContainer
   implements FieldContainer
{
   String m_host         = null;
   String m_handle_label = null;
   String m_handle       = null;
   
   // HTTP Commands from the subclass
   String m_create_property_command         = null;
   String m_set_property_command            = null;
   String m_get_property_command            = null;
   String m_delete_field_command            = null;
   String m_get_field_length_command        = null;
   String m_get_field_type_command          = null;
   String m_get_field_readonly_command      = null;
   String m_get_field_binding_command       = null;
   String m_set_field_as_binding_command    = null;
   String m_set_field_as_nonbinding_command = null;
   String m_create_xstream_command          = null;
   String m_open_xstream_command            = null;
   String m_open_field_iterator_command     = null;
   String m_xiterator_type_property         = null;
   String m_contains_field_command          = null;
   String m_async_open_xstream_command      = null;
   
   Logger m_logger = null;
   
   public XAsync asyncOpenXStream(String name,
                                  String mode,
                                  long xopid,
                                  XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      InvalidFieldNameException, FieldDoesNotExistException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_async_open_xstream_command,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(name, ProtocolConstants.UTF8),
            ProtocolConstants.MODE_PROPERTY, URLEncoder.encode(mode, ProtocolConstants.UTF8),
            ProtocolConstants.ASYNC_XOPID, Long.toString(xopid),
            m_handle_label, m_handle);
         String handle = responses.get(ProtocolConstants.XASYNC_HANDLE);
         responses.clear();
         return new HTTPXAsync( m_host, handle, listener, m_logger );
      } catch( UnsupportedEncodingException use ) {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, use );
      }
   }

   public boolean containsField(String fieldName)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_contains_field_command,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            m_handle_label, m_handle);
         boolean contains = Boolean.parseBoolean(responses.get(ProtocolConstants.ATTRIBUTE_VALUE));
         responses.clear();
         return contains;
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public void createProperty(String fieldName, boolean binding, boolean value)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldExistsException, InvalidOperationException, MaximumFieldException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_create_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_BOOLEAN,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            ProtocolConstants.ATTRIBUTE_VALUE, Boolean.toString(value),
            ProtocolConstants.ATTRIBUTE_BINDING, Boolean.toString(binding),
            m_handle_label, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public void createProperty(String fieldName, boolean binding, long value)
      throws AuthenticationExpiredException, FieldExistsException,
      InvalidOperationException, MaximumFieldException, ObjectInUseException,
      XSetAbandonException, XSetCorruptException, XSystemAbandonException,
      XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_create_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_INTEGER,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            ProtocolConstants.ATTRIBUTE_VALUE, Long.toString(value),
            ProtocolConstants.ATTRIBUTE_BINDING, Boolean.toString(binding),
            m_handle_label, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public void createProperty(String fieldName, boolean binding, double value)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldExistsException, InvalidOperationException, MaximumFieldException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_create_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_DOUBLE,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            ProtocolConstants.ATTRIBUTE_VALUE, Double.toString(value),
            ProtocolConstants.ATTRIBUTE_BINDING, Boolean.toString(binding),
            m_handle_label, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public void createProperty(String fieldName, boolean binding, XUID value)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldExistsException, InvalidArgumentException,
      InvalidOperationException, MaximumFieldException, ObjectInUseException,
      XSetAbandonException, XSetCorruptException, XSystemAbandonException,
      XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_create_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_XUID,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            ProtocolConstants.ATTRIBUTE_VALUE, URLEncoder.encode(value.toString(), ProtocolConstants.UTF8),
            ProtocolConstants.ATTRIBUTE_BINDING, Boolean.toString(binding),
            m_handle_label, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public void createProperty(String fieldName, boolean binding, String value)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldExistsException, InvalidArgumentException,
      InvalidOperationException, MaximumFieldException, ObjectInUseException,
      XSetAbandonException, XSetCorruptException, XSystemAbandonException,
      XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_create_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_STRING,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            ProtocolConstants.ATTRIBUTE_VALUE, URLEncoder.encode(value, ProtocolConstants.UTF8),
            ProtocolConstants.ATTRIBUTE_BINDING, Boolean.toString(binding),
            m_handle_label, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public void createProperty(String fieldName, boolean binding, Calendar value)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldExistsException, InvalidArgumentException,
      InvalidOperationException, MaximumFieldException, ObjectInUseException,
      XSetAbandonException, XSetCorruptException, XSystemAbandonException,
      XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_create_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_DATETIME,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            ProtocolConstants.ATTRIBUTE_VALUE, Long.toString(value.getTimeInMillis()),
            ProtocolConstants.ATTRIBUTE_BINDING, Boolean.toString(binding),
            m_handle_label, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public XStream createXStream(String fieldName,
                                boolean binding,
                                String mimeType)
      throws AuthenticationExpiredException, InvalidArgumentException,
      InvalidFieldNameException, InvalidOperationException,
      FieldExistsException, MaximumFieldException, ObjectInUseException,
      XSetAbandonException, XSetCorruptException, XSystemAbandonException,
      XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_create_xstream_command,
            ProtocolConstants.TYPE_PROPERTY, URLEncoder.encode(mimeType, ProtocolConstants.UTF8),
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            ProtocolConstants.ATTRIBUTE_BINDING, Boolean.toString(binding),
            m_handle_label, m_handle);
         String stream_handle = responses.get(ProtocolConstants.XSTREAM_HANDLE);
         responses.clear();
         return new HTTPXStream( m_host, stream_handle, m_logger );
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public void deleteField(String fieldName)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldDoesNotExistException, ObjectInUseException, XSetAbandonException,
      XSetCorruptException, XSystemAbandonException, XSystemCorruptException,
      XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_delete_field_command,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            m_handle_label, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public boolean getBoolean(String fieldName)
      throws AuthenticationExpiredException, FieldDoesNotExistException,
      InvalidFieldNameException, InvalidFieldTypeException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSetCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_get_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_BOOLEAN,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            m_handle_label, m_handle);
         return Boolean.parseBoolean(responses.get(ProtocolConstants.ATTRIBUTE_VALUE));
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public Calendar getDateTime(String fieldName)
      throws AuthenticationExpiredException, FieldDoesNotExistException,
      InvalidFieldNameException, InvalidFieldTypeException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSetCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_get_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_DATETIME,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            m_handle_label, m_handle);
         long millis = Long.valueOf(responses.get(ProtocolConstants.ATTRIBUTE_VALUE));
         Calendar c = Calendar.getInstance();
         c.setTimeInMillis(millis);
         return c;
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }
   
   public double getDouble(String fieldName)
      throws AuthenticationExpiredException, FieldDoesNotExistException,
      InvalidFieldNameException, InvalidFieldTypeException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSetCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_get_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_DOUBLE,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            m_handle_label, m_handle);
         return Double.parseDouble(responses.get(ProtocolConstants.ATTRIBUTE_VALUE));
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }
   
   public boolean getFieldBinding(String fieldName)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldDoesNotExistException, ObjectInUseException, XSetAbandonException,
      XSetCorruptException, XSystemAbandonException, XSystemCorruptException,
      XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_get_field_binding_command,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            m_handle_label, m_handle);
         return Boolean.parseBoolean(responses.get(ProtocolConstants.ATTRIBUTE_VALUE));
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public long getFieldLength(String fieldName)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldDoesNotExistException, ObjectInUseException, XSetAbandonException,
      XSetCorruptException, XSystemAbandonException, XSystemCorruptException,
      XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_get_field_length_command,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            m_handle_label, m_handle);
         return Long.parseLong(responses.get(ProtocolConstants.ATTRIBUTE_VALUE));
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public boolean getFieldReadOnly(String fieldName)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldDoesNotExistException, ObjectInUseException, XSetAbandonException,
      XSetCorruptException, XSystemAbandonException, XSystemCorruptException,
      XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_get_field_readonly_command,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            m_handle_label, m_handle);
         return Boolean.parseBoolean(responses.get(ProtocolConstants.ATTRIBUTE_VALUE));
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public String getFieldType(String fieldName)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldDoesNotExistException, ObjectInUseException, XSetAbandonException,
      XSetCorruptException, XSystemAbandonException, XSystemCorruptException,
      XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_get_field_type_command,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            m_handle_label, m_handle);
         return responses.get(ProtocolConstants.ATTRIBUTE_VALUE);
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public long getLong(String fieldName)
      throws AuthenticationExpiredException, FieldDoesNotExistException,
             InvalidFieldNameException, InvalidFieldTypeException,
             ObjectInUseException, XSetAbandonException, XSetCorruptException,
             XSystemAbandonException, XSetCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_get_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_INTEGER,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            m_handle_label, m_handle);
         return Long.parseLong(responses.get(ProtocolConstants.ATTRIBUTE_VALUE));
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }
   
   public String getString(String fieldName)
      throws AuthenticationExpiredException, FieldDoesNotExistException,
      InvalidFieldNameException, InvalidFieldTypeException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSetCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_get_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE,  ProtocolConstants.PTYPE_STRING,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            m_handle_label, m_handle);
         String value = responses.get(ProtocolConstants.ATTRIBUTE_VALUE);
         return URLDecoder.decode( value, ProtocolConstants.UTF8 );
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }
   
   public XUID getXUID(String fieldName)
      throws AuthenticationExpiredException, FieldDoesNotExistException,
      InvalidFieldNameException, InvalidFieldTypeException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSetCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_get_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_XUID,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            m_handle_label, m_handle);
         String value = responses.get(ProtocolConstants.ATTRIBUTE_VALUE);
         return DefaultXUID.valueOf(URLDecoder.decode( value, ProtocolConstants.UTF8 ));
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }
   
   public XIterator openFieldIterator(String prefix)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_open_field_iterator_command,
            ProtocolConstants.CLASS_TYPE_PROPERTY, m_xiterator_type_property,
            ProtocolConstants.PATTERN_PROPERTY, URLEncoder.encode(prefix, ProtocolConstants.UTF8),
            m_handle_label, m_handle);
         return new HTTPXIterator(m_host, responses.get(ProtocolConstants.XITERATOR_HANDLE));
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public XStream openXStream(String fieldName, String mode)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldDoesNotExistException, ObjectInUseException, XSetAbandonException,
      XSetCorruptException, XSystemAbandonException, XSystemCorruptException,
      XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_open_xstream_command,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            ProtocolConstants.MODE_PROPERTY, URLEncoder.encode(mode, ProtocolConstants.UTF8), 
            m_handle_label, m_handle);
         return new HTTPXStream( m_host, responses.get(ProtocolConstants.XSTREAM_HANDLE), m_logger);
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public void setFieldAsBinding(String fieldName)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      InvalidOperationException, FieldDoesNotExistException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_set_field_as_binding_command,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            m_handle_label, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public void setFieldAsNonbinding(String fieldName)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldDoesNotExistException, ObjectInUseException, XSetAbandonException,
      XSetCorruptException, XSystemAbandonException, XSystemCorruptException,
      XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_set_field_as_nonbinding_command,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            m_handle_label, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public void setProperty(String fieldName, boolean value)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      InvalidFieldTypeException, FieldDoesNotExistException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_set_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_BOOLEAN,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            ProtocolConstants.ATTRIBUTE_VALUE, Boolean.toString(value),
            m_handle_label, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public void setProperty(String fieldName, Calendar value)
      throws AuthenticationExpiredException, FieldDoesNotExistException,
      InvalidArgumentException, InvalidFieldNameException,
      InvalidFieldTypeException, ObjectInUseException, XSetAbandonException,
      XSetCorruptException, XSystemAbandonException, XSystemCorruptException,
      XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_set_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_DATETIME,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            ProtocolConstants.ATTRIBUTE_VALUE, Long.toString(value.getTimeInMillis()),
            m_handle_label, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public void setProperty(String fieldName, double value)
      throws AuthenticationExpiredException, FieldDoesNotExistException,
      InvalidFieldNameException, InvalidFieldTypeException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_set_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_DOUBLE,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            ProtocolConstants.ATTRIBUTE_VALUE, Double.toString(value),
            m_handle_label, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public void setProperty(String fieldName, long value)
      throws AuthenticationExpiredException, FieldDoesNotExistException,
      InvalidFieldNameException, InvalidFieldTypeException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_set_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_INTEGER,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            ProtocolConstants.ATTRIBUTE_VALUE, Long.toString(value),
            m_handle_label, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public void setProperty(String fieldName, String value)
      throws AuthenticationExpiredException, FieldDoesNotExistException,
      InvalidArgumentException, InvalidFieldNameException,
      InvalidFieldTypeException, ObjectInUseException, XSetAbandonException,
      XSetCorruptException, XSystemAbandonException, XSystemCorruptException,
      XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_set_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_STRING,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            ProtocolConstants.ATTRIBUTE_VALUE, value,
            m_handle_label, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public void setProperty(String fieldName, XUID value)
      throws AuthenticationExpiredException, FieldDoesNotExistException,
      InvalidArgumentException, InvalidFieldNameException,
      InvalidFieldTypeException, ObjectInUseException, XSetAbandonException,
      XSetCorruptException, XSystemAbandonException, XSystemCorruptException,
      XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            m_set_property_command,
            ProtocolConstants.ATTRIBUTE_TYPE, ProtocolConstants.PTYPE_XUID,
            ProtocolConstants.ATTRIBUTE_NAME, URLEncoder.encode(fieldName, ProtocolConstants.UTF8),
            ProtocolConstants.ATTRIBUTE_VALUE, value.toString(),
            m_handle_label, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }
}
