/*
 * Copyright (c) 2009, Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
package org.snia.xam.vim.http.client;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.HashMap;

import org.snia.xam.AsyncHaltedException;
import org.snia.xam.AsyncPendingException;
import org.snia.xam.AuthenticationException;
import org.snia.xam.AuthorizationException;
import org.snia.xam.ConnectException;
import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.FieldExistsException;
import org.snia.xam.FieldInUseException;
import org.snia.xam.FieldReadOnlyException;
import org.snia.xam.HoldIdException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidFieldNameException;
import org.snia.xam.InvalidFieldTypeException;
import org.snia.xam.InvalidOperationException;
import org.snia.xam.InvalidXRIException;
import org.snia.xam.InvalidXSetModeException;
import org.snia.xam.InvalidXStreamModeException;
import org.snia.xam.InvalidXUIDException;
import org.snia.xam.JobCommandException;
import org.snia.xam.JobPermissionsException;
import org.snia.xam.JobResourcesException;
import org.snia.xam.JobRunningException;
import org.snia.xam.JobUnsupportedException;
import org.snia.xam.MaximumFieldException;
import org.snia.xam.ObjectInUseException;
import org.snia.xam.PolicyMismatchException;
import org.snia.xam.PolicyNameException;
import org.snia.xam.RetentionValueException;
import org.snia.xam.VIMLoadException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XSetAbandonException;
import org.snia.xam.XSetCorruptException;
import org.snia.xam.XSetInaccessibleException;
import org.snia.xam.XSetUnderHoldException;
import org.snia.xam.XSetUnderRetentionException;
import org.snia.xam.XStreamAbandonException;
import org.snia.xam.XStreamCorruptException;
import org.snia.xam.XSystemAbandonException;
import org.snia.xam.XSystemCorruptException;

import org.snia.xam.vim.isolation.server.http.handlers.ProtocolConstants;

public class HTTPRequest
{
   static final boolean debug = false;
   
   /**
    * Send a simple request with one argument and value.
    * @param host The host with the remote HTTP VIM Server
    * @param method The method to execute
    * @param arg1 First argument name
    * @param value1 First value
    * @return The response map
    * @throws XAMException
    */
   static HashMap<String, String> request(String host,
                                          String method,
                                          String arg1,
                                          String value1)
      throws XAMException
   {
      HashMap<String, String> response = new HashMap<String, String>();
      String composedURL = "http://" + host + "/" + method ;

      if(debug)
         System.out.println( composedURL );
      try
      {
         URL remote = new URL(composedURL);
         URLConnection conn = remote.openConnection();
         conn.setRequestProperty(arg1, value1);
         InputStream in = conn.getInputStream();
         BufferedReader rdr = new BufferedReader(new InputStreamReader(in));
         processResponse(rdr, response);
         String responseStr = response.get(ProtocolConstants.STATUS_CODE);
         int errno = Integer.valueOf(responseStr);
         String errMsg = response.get(ProtocolConstants.STATUS_MSG);
         mapError( method, host, remote, errno, errMsg );
         return response;
      }
      catch (XAMException xe)
      {
         throw xe; // Don't need to wrap it again.
      }
      catch (Throwable t)
      {
         throw new XAMException("Request failed " + composedURL, t);
      }
   }

   /**
    * Send a request with two arguments and values.
    * @param host The host with the remote HTTP VIM Server
    * @param method The method to execute
    * @param arg1 First argument name
    * @param value1 First value
    * @param arg2 Second argument name
    * @param value2 Second value
    * @return The response map
    * @throws XAMException
    */
   static HashMap<String, String> request(String host,
                                          String method,
                                          String arg1,
                                          String value1,
                                          String arg2,
                                          String value2 )
      throws XAMException
   {
      HashMap<String, String> response = new HashMap<String, String>();
      String composedURL = "http://" + host + "/" + method ;

      if(debug)
         System.out.println( composedURL );
      try
      {
         URL remote = new URL(composedURL);
         URLConnection conn = remote.openConnection();
         conn.addRequestProperty(arg1, value1);
         conn.addRequestProperty(arg2, value2);
         InputStream in = conn.getInputStream();
         BufferedReader rdr = new BufferedReader(new InputStreamReader(in));
         processResponse(rdr, response);
         int errno = Integer.valueOf(response.get(ProtocolConstants.STATUS_CODE));
         String errMsg = response.get(ProtocolConstants.STATUS_MSG);
         mapError( method, host, remote, errno, errMsg );
         return response;
      }
      catch (XAMException xe)
      {
         throw xe; // Don't need to wrap it again.
      }
      catch (Throwable t)
      {
         throw new XAMException("Request failed " + composedURL, t);
      }
   }

   /**
    * Send a request with three arguments and values.
    * @param host The host with the remote HTTP VIM Server
    * @param method The method to execute
    * @param arg1 First argument name
    * @param value1 First value
    * @param arg2 Second argument name
    * @param value2 Second value
    * @param arg3 Third argument name
    * @param value3 Third value
    * @return The response map
    * @throws XAMException
    */
   static HashMap<String, String> request(String host,
                                          String method,
                                          String arg1,
                                          String value1,
                                          String arg2,
                                          String value2,
                                          String arg3,
                                          String value3)
      throws XAMException
   {
      HashMap<String, String> response = new HashMap<String, String>();
      String composedURL = "http://" + host + "/" + method ;

      if(debug)
         System.out.println( composedURL );
      try
      {
         URL remote = new URL(composedURL);
         URLConnection conn = remote.openConnection();
         conn.addRequestProperty(arg1, value1);
         conn.addRequestProperty(arg2, value2);
         conn.addRequestProperty(arg3, value3);
         InputStream in = conn.getInputStream();
         BufferedReader rdr = new BufferedReader(new InputStreamReader(in));
         processResponse(rdr, response);
         int errno = Integer.valueOf(response.get(ProtocolConstants.STATUS_CODE));
         String errMsg = response.get(ProtocolConstants.STATUS_MSG);
         mapError( method, host, remote, errno, errMsg );
         return response;
      }
      catch (XAMException xe)
      {
         throw xe; // Don't need to wrap it again.
      }
      catch (Throwable t)
      {
         throw new XAMException("Request failed " + composedURL, t);
      }
   }

   /**
    * Send a request with four arguments and values.
    * @param host The host with the remote HTTP VIM Server
    * @param method The method to execute
    * @param arg1 First argument name
    * @param value1 First value
    * @param arg2 Second argument name
    * @param value2 Second value
    * @param arg3 Third argument name
    * @param value3 Third value
    * @param arg4 Fourth argument name
    * @param value4 Fourth value
    * @return The response map
    * @throws XAMException
    */
   static HashMap<String, String> request(String host,
                                          String method,
                                          String arg1,
                                          String value1,
                                          String arg2,
                                          String value2,
                                          String arg3,
                                          String value3,
                                          String arg4,
                                          String value4 )
      throws XAMException
   {
      HashMap<String, String> response = new HashMap<String, String>();
      String composedURL = "http://" + host + "/" + method;
      
      if(debug)
         System.out.println( composedURL );
      try
      {
         URL remote = new URL(composedURL);
         URLConnection conn = remote.openConnection();
         conn.addRequestProperty(arg1, value1);
         conn.addRequestProperty(arg2, value2);
         conn.addRequestProperty(arg3, value3);
         conn.addRequestProperty(arg4, value4);
         InputStream in = conn.getInputStream();
         BufferedReader rdr = new BufferedReader(new InputStreamReader(in));
         processResponse(rdr, response);
         int errno = Integer.valueOf(response.get(ProtocolConstants.STATUS_CODE));
         String errMsg = response.get(ProtocolConstants.STATUS_MSG);
         mapError( method, host, remote, errno, errMsg );
         return response;
      }
      catch (XAMException xe)
      {
         throw xe; // Don't need to wrap it again.
      }
      catch (Throwable t)
      {
         throw new XAMException("Request failed " + composedURL, t);
      }
   }

   /**
    * Send a request with five arguments and values.
    * @param host The host with the remote HTTP VIM Server
    * @param method The method to execute
    * @param arg1 First argument name
    * @param value1 First value
    * @param arg2 Second argument name
    * @param value2 Second value
    * @param arg3 Third argument name
    * @param value3 Third value
    * @param arg4 Fourth argument name
    * @param valuye4 Fourth value
    * @param arg5 Fifth argument name
    * @param value5 Fifth value
    * @return The response map
    * @throws XAMException
    */
   static HashMap<String, String> request(String host,
                                          String method,
                                          String arg1,
                                          String value1,
                                          String arg2,
                                          String value2,
                                          String arg3,
                                          String value3,
                                          String arg4,
                                          String value4,
                                          String arg5,
                                          String value5)
      throws XAMException
   {
      HashMap<String, String> response = new HashMap<String, String>();
      String composedURL = "http://" + host + "/" + method;

      if(debug)
         System.out.println( composedURL );
      try
      {
         URL remote = new URL(composedURL);
         URLConnection conn = remote.openConnection();
         conn.addRequestProperty(arg1, value1);
         conn.addRequestProperty(arg2, value2);
         conn.addRequestProperty(arg3, value3);
         conn.addRequestProperty(arg4, value4);
         conn.addRequestProperty(arg5, value5);
         InputStream in = conn.getInputStream();
         BufferedReader rdr = new BufferedReader(new InputStreamReader(in));
         processResponse(rdr, response);
         int errno = Integer.valueOf(response.get(ProtocolConstants.STATUS_CODE));
         String errMsg = response.get(ProtocolConstants.STATUS_MSG);
         mapError( method,  host, remote, errno, errMsg );
         return response;
      }
      catch (XAMException xe)
      {
         throw xe; // Don't need to wrap it again.
      }
      catch (Throwable t)
      {
         throw new XAMException("Request failed " + composedURL, t);
      }
   }

   /**
    * Send a request with two arguments and values.
    * @param host The host with the remote HTTP VIM Server
    * @param method The method to execute
    * @param arg1 First argument name
    * @param value1 First value
    * @param arg2 Second argument name
    * @param value2 Second value
    * @return The response map
    * @throws XAMException
    */
   static HashMap<String, String> requestWrite(String host,
                                          String method,
                                          String arg1,
                                          String value1,
                                          int    offset,
                                          int    length,
                                          byte   buffer[] )
      throws XAMException
   {
      HashMap<String, String> response = new HashMap<String, String>();
      String composedURL = "http://" + host + "/" + method ;
      if(debug)
         System.out.println( composedURL );
      
      try
      {
         URL remote = new URL(composedURL);
         
         // Send buffer
         URLConnection conn = remote.openConnection();
         conn.setDoOutput(true);
         conn.setDoInput(true);
         conn.setAllowUserInteraction(false);
         
         conn.setRequestProperty("Content-Type",  XAMLibrary.DEFAULT_MIME_TYPE );
         conn.setRequestProperty("Content-Length", Integer.toString(buffer.length) );
         conn.setRequestProperty(arg1, value1 );
         OutputStream os = conn.getOutputStream();
         os.write(buffer, offset, length );
         os.flush();

         // Read Response, forces the POST to complete.
         InputStream in = conn.getInputStream();
         BufferedReader rdr = new BufferedReader(new InputStreamReader(in));
         processResponse(rdr, response);
         in.close();

         os.close();
         int errno = Integer.valueOf(response.get(ProtocolConstants.STATUS_CODE));
         String errMsg = response.get(ProtocolConstants.STATUS_MSG);
         mapError( method, host, remote, errno, errMsg );
         return response;
      }
      catch (XAMException xe)
      {
         throw xe; // Don't need to wrap it again.
      }
      catch (Throwable t)
      {
         throw new XAMException("Request failed " + composedURL, t);
      }
   }

   /**
    * Send a request with two arguments and values.
    * @param host The host with the remote HTTP VIM Server
    * @param method The method to execute
    * @param arg1 First argument name
    * @param value1 First value
    * @param arg2 Second argument name
    * @param value2 Second value
    * @return The response map
    * @throws XAMException
    */
   static HashMap<String, String> requestWrite(String host,
                                          String method,
                                          String arg1,
                                          String value1,
                                          String arg2,
                                          String value2,
                                          int    offset,
                                          int    length,
                                          byte   buffer[] )
      throws XAMException
   {
      HashMap<String, String> response = new HashMap<String, String>();
      String composedURL = "http://" + host + "/" + method ;
      if(debug)
         System.out.println( composedURL );
      
      try
      {
         URL remote = new URL(composedURL);
         
         // Send buffer
         URLConnection conn = remote.openConnection();
         conn.setDoOutput(true);
         conn.setDoInput(true);
         conn.setAllowUserInteraction(false);
         
         conn.setRequestProperty("Content-Type",  XAMLibrary.DEFAULT_MIME_TYPE );
         conn.setRequestProperty("Content-Length", Integer.toString(buffer.length) );
         conn.setRequestProperty( arg1, value1 );
         conn.setRequestProperty( arg2, value2 );
         OutputStream os = conn.getOutputStream();
         os.write(buffer, offset, length );
         os.flush();

         // Read Response, forces the POST to complete.
         InputStream in = conn.getInputStream();
         BufferedReader rdr = new BufferedReader(new InputStreamReader(in));
         processResponse(rdr, response);
         in.close();

         os.close();
         int errno = Integer.valueOf(response.get(ProtocolConstants.STATUS_CODE));
         String errMsg = response.get(ProtocolConstants.STATUS_MSG);
         mapError( method, host, remote, errno, errMsg );
         return response;
      }
      catch (XAMException xe)
      {
         throw xe; // Don't need to wrap it again.
      }
      catch (Throwable t)
      {
         throw new XAMException("Request failed " + composedURL, t);
      }
   }

   /**
    * Send a request with three arguments and values.
    * @param host The host with the remote HTTP VIM Server
    * @param method The method to execute
    * @param arg1 First argument name
    * @param value1 First value
    * @param arg2 Second argument name
    * @param value2 Second value
    * @param arg3 Third argument name
    * @param value3 Third value
    * @return The response map
    * @throws XAMException
    */
   static int requestNoStatus(String host,
                              String method,
                              String arg1,
                              String value1,
                              String arg2,
                              String value2,
                              byte   buffer[],
                              int    offset,
                              int    bufferLength )
   throws XAMException
   {
      String composedURL = "http://" + host + "/" + method + "?" + arg1 + "="
         + value1 + "&" + arg2 + "=" + value2 ;
      try
      {
         URL remote = new URL(composedURL);
         InputStream strm = remote.openStream();
         byte b ;
         int  count = 0;
         int  index = offset; 
         while( (b = (byte) strm.read()) > 0 )
         {
            buffer[index++] = b;
            count++;
         }
         return count;
      }
      catch (Throwable t)
      {
         throw new XAMException("Request failed " + composedURL, t);
      }
   }

   /**
    * Process all response codes for this request.
    * 
    * @param rdr The reader with the responses
    * @param response The response map.
    * @throws IOException
    */
   private static void processResponse(BufferedReader rdr,
                                       HashMap<String, String> response)
      throws XAMException, IOException
   {
      boolean bt = false;
      String respLine = null;
      while ((respLine = rdr.readLine()) != null)
      {
         if( bt ) System.out.println( respLine );
         
         int equalIdx = respLine.indexOf("=");
         if( equalIdx > 0 )
         {
            String respCode = respLine.substring(0, equalIdx);
            String valueStr = respLine.substring(equalIdx + 1);
            response.put(respCode, valueStr);
         }
         else 
            throw new XAMException( "Malformed method response: " + respLine );
      }
   }

   /**
    * Given an error code from the remote server, turn it into the appropriate
    * XAM Error.
    * 
    * @param errno Error status returned by remote server
    * @throws XAMException
    */
   static void mapError( String method, String host, URL url, int errno, String errMsg )
      throws XAMException
   {
      String msg = errMsg;
      if( msg == null )
         msg = "Method " + method + " failed on host " + host + ": " + url;
      
      switch (errno)
      {
         case XAMException.SUCCESS :
            return;
            
         case XAMException.sXAM_INVALID_HANDLE :
            throw new XAMException( "Invalid Handle " + url, errno );
            
         case XAMException.sXAM_OUT_OF_MEMORY:
            if( errMsg == null )
               msg = "Remote VIM reports out of memory";
            throw new XAMException( msg, errno );
            
         case XAMException.sInvalidArgumentException:
            if( errMsg == null )
              msg = "Invalid argument to method";
            throw new XAMException( msg, errno );
            
         case XAMException.sXAM_NOT_SUPPORTED :
            if( errMsg == null )
               msg = "Operation not supported";
            throw new XAMException( msg, errno );
            
         case XAMException.sXAM_FILESYSTEM_ERROR:
            if( errMsg == null )
               msg = "File system exception";
            throw new XAMException( msg, errno );
            
         case XAMException.sXAMException:
            throw new XAMException(msg + " " + errno, errno);
            
         case XAMException.sFieldDoesNotExistException :
            throw new FieldDoesNotExistException(msg, errno);

         case XAMException.sInvalidFieldNameException:
            throw new InvalidFieldNameException(msg, errno);

         case XAMException.sXAM_PARAM_NOT_UTF8:
            throw new InvalidArgumentException(msg,errno);
            
         case XAMException.sInvalidFieldTypeException:
            throw new InvalidFieldTypeException(msg,errno);
            
         case XAMException.sInvalidXStreamModeException:
            throw new InvalidXStreamModeException(msg,errno);
            
         case XAMException.sInvalidXSetModeException:
            throw new InvalidXSetModeException(msg,errno);

         case XAMException.sInvalidXRIException:
            throw new InvalidXRIException(msg,errno);
            
         case XAMException.sVIMLoadException:
         case XAMException.sXAM_VIM_SYMBOL_NOT_FOUND:
            throw new VIMLoadException(msg,errno);
            
         case XAMException.sFieldReadOnlyException:
            throw new FieldReadOnlyException(msg,errno);
            
         case XAMException.sFieldExistsException:
            throw new FieldExistsException(msg,errno);
            
         case XAMException.sFieldInUseException:
            throw new FieldInUseException(msg,errno);
            
         case XAMException.sMaximumFieldException:
            throw new MaximumFieldException(msg,errno);
            
         case XAMException.sXSystemAbandonException:
            throw new XSystemAbandonException(msg,errno);
            
         case XAMException.sXSetAbandonException:
            throw new XSetAbandonException(msg,errno);
            
         case XAMException.sXStreamAbandonException:
            throw new XStreamAbandonException(msg,errno);
            
         case XAMException.sXSystemCorruptException:
            throw new XSystemCorruptException(msg,errno);
            
         case XAMException.sXSetCorruptException:
            throw new XSetCorruptException(msg,errno);
            
         case XAMException.sXStreamCorruptException:
            throw new XStreamCorruptException(msg,errno);
            
         case XAMException.sConnectException:
            throw new ConnectException(msg,errno);
            
         case XAMException.sXAM_AUTH_DATA_NEEDED:
         case XAMException.sAuthorizationException :
            throw new AuthorizationException(msg,errno);
            
         case XAMException.sAuthenticationException:
            throw new AuthenticationException(msg,errno);
            
         case XAMException.sInvalidXUIDException:
            throw new InvalidXUIDException(msg,errno);
            
         case XAMException.sXSetDoesNotExistException:
            throw new XSetInaccessibleException(msg,errno);
            
         case XAMException.sInvalidOperationException:
            throw new InvalidOperationException(msg,errno);
            
         case XAMException.sObjectInUseException:
            throw new ObjectInUseException(msg,errno);
            
         case XAMException.sXAM_NOT_A_JOB:
         case XAMException.sJobCommandException:
         case XAMException.sXAM_JOB_INVALID_CMD_SYNTAX:
            throw new JobCommandException(msg,errno);
            
         case XAMException.sJobUnsupportedException:
            throw new JobUnsupportedException(msg,errno);
            
         case XAMException.sJobPermissionsException:
            throw new JobPermissionsException(msg,errno);
            
         case XAMException.sJobResourcesException:
            throw new JobResourcesException(msg,errno);
            
         case XAMException.sJobRunningException:
            throw new JobRunningException(msg,errno);
            
         case XAMException.sXSetUnderRetentionException:
            throw new XSetUnderRetentionException(msg,errno);
            
         case XAMException.sXSetUnderHoldException:
            throw new XSetUnderHoldException(msg,errno);
            
         case XAMException.sHoldIdException:
            throw new HoldIdException(msg,errno);
            
         case XAMException.sRetentionValueException:
            throw new RetentionValueException(msg,errno);
            
         case XAMException.sPolicyNameException:
            throw new PolicyNameException(msg,errno);
            
         case XAMException.sPolicyMismatchException:
            throw new PolicyMismatchException(msg,errno);
            
         case XAMException.sAsyncPendingException :
            throw new AsyncPendingException(msg,errno);
            
         case XAMException.sXAM_XASYNC_HALTED:
            throw new AsyncHaltedException( "Asynchronous operation failure " + errno, errno );
         default :
            throw new XAMException("Unmapped error " + errno, errno);
      }
   }
}
