/*
 * Copyright (c) 2009, Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
package org.snia.xam.vim.http.client;

import java.net.URL;
import java.util.HashMap;
import java.util.logging.Logger;

import org.snia.xam.AsyncHaltedException;
import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XSet;
import org.snia.xam.XStream;
import org.snia.xam.XUID;
import org.snia.xam.base.DefaultXUID;
import org.snia.xam.vim.isolation.server.http.handlers.ProtocolConstants;
import org.snia.xam.vim.isolation.server.util.XAsyncCache;

/**
 * Implements the XAsync client model for the XAM HTTP VIM. Because of the HTTP protocol
 * implementation, this class contains a singleton instance of a thread to manage outstanding
 * requests and manage call backs when complete. 
 */
public class HTTPXAsync
   implements XAsync
{
   private String m_host     = null;
   private String m_handle   = null;
   XAsyncListener m_listener = null;
   private boolean m_is_complete = false;
   
   // Used if the request was a read request.
   private byte   m_buffer[] = null;
   
   Logger m_logger = null;
   
   /**
    * Create a new HTTP XAsync object to handle these async operations.
    * @param host The host:port to be used for the operations.
    * @param handle The XAsync handle for this operation
    */
   HTTPXAsync( String host, String handle, XAsyncListener listener, Logger logger )
      throws XAMException
   {
      final long NULL_XSYSTEM_HANDLE = -1;
      
      m_host     = host;
      m_handle   = handle;
      m_listener = listener;
      m_logger   = logger;
      
      XAsyncCache xac = XAsyncCache.getInstance( logger );
      try
      {
         xac.addXAsyncWithHandle( NULL_XSYSTEM_HANDLE, Long.valueOf(m_handle), this );
      }
      catch (NumberFormatException e)
      {
         throw new XAMException( "HTTPXAsync creation error with bad handle: " + m_handle, XAMException.sXAM_INVALID_HANDLE );
      }
   }
   
   public String toString( ) { return "HTTPXAsync " + m_handle + " " + m_buffer ; }
   
  
   /**
    * Set the buffer, as a result of an asyncRead operation.
    */
   void setBuffer( byte buffer[] )
   {
      m_buffer = buffer;
   }
   
   /**
    * Return the byte buffer to be used in case the operation was an asyncRead.
    */
   byte[] getBuffer() 
   {
      return m_buffer;
   }
   
   /**
    * Called from the callback to set completion. 
    */
   void setIsComplete() { m_is_complete = true; } 

   public void close()
      throws AuthenticationExpiredException, XAMException
   {
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.XASYNC_CLOSE_HANDLER,
         ProtocolConstants.XASYNC_HANDLE, m_handle );
      responses.clear();
      XAsyncCache xac = XAsyncCache.getInstance(m_logger);
      xac.removeXAsync(Long.parseLong(m_handle));
      m_host   = null;
      m_handle = null;
   }

   public long getBytesRead()
      throws AuthenticationExpiredException, AsyncHaltedException, XAMException
   {
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.XASYNC_GET_BYTES_READ_HANDLER,
         ProtocolConstants.XASYNC_HANDLE, m_handle );
      String bytesStr = responses.get(ProtocolConstants.ASYNC_BYTES_READ);
      responses.clear();
      return Long.parseLong(bytesStr);
   }

   public long getBytesWritten()
      throws AuthenticationExpiredException, AsyncHaltedException, XAMException
   {
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.XASYNC_GET_BYTES_WRITTEN_HANDLER,
         ProtocolConstants.XASYNC_HANDLE, m_handle );
      String bytesStr = responses.get(ProtocolConstants.ASYNC_BYTES_WRITTEN);
      responses.clear();
      return Long.parseLong(bytesStr);
   }

   public void getStatus()
      throws AuthenticationExpiredException, AsyncHaltedException, XAMException
   {
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.XASYNC_GET_STATUS,
         ProtocolConstants.XASYNC_HANDLE, m_handle );
      String errno = responses.get(ProtocolConstants.ASYNC_STATUS);
      String errMsg = responses.get(ProtocolConstants.ASYNC_STATUS_MSG );
      HTTPRequest.mapError( "HTTPXAsync", m_host, (URL) null, Integer.parseInt(errno), errMsg );
      responses.clear();
   }

   public long getXOPID()
      throws AuthenticationExpiredException, AsyncHaltedException, XAMException
   {
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.XASYNC_GET_XOPID,
         ProtocolConstants.XASYNC_HANDLE, m_handle );
      String xopid = responses.get(ProtocolConstants.ASYNC_XOPID);
      responses.clear();
      return Long.parseLong(xopid);
   }

   public XSet getXSet()
      throws AuthenticationExpiredException, AsyncHaltedException, XAMException
   {
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.XASYNC_GET_XSET,
         ProtocolConstants.XASYNC_HANDLE, m_handle );
      String xsethandle = responses.get(ProtocolConstants.ASYNC_XSET);
      responses.clear();
      if( xsethandle.equals("-1"))
         return null;
      else
         return new HTTPXSet( m_host, xsethandle );
   }

   public XStream getXStream()
      throws AuthenticationExpiredException, AsyncHaltedException, XAMException
   {
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.XASYNC_GET_XSTREAM,
         ProtocolConstants.XASYNC_HANDLE, m_handle );
      String xstreamhandle = responses.get(ProtocolConstants.ASYNC_XSTREAM);
      responses.clear();
      if( xstreamhandle.equals(-1))
         return null;
      else
         return new HTTPXStream( m_host, xstreamhandle, m_logger );
   }

   public XUID getXUID()
      throws AuthenticationExpiredException, AsyncHaltedException, XAMException
   {
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.XASYNC_GET_XUID,
         ProtocolConstants.XASYNC_HANDLE, m_handle );
      String xuid = responses.get(ProtocolConstants.ASYNC_XUID);
      responses.clear();
      return DefaultXUID.valueOf(xuid);
   }

   public void halt()
      throws AuthenticationExpiredException, XAMException
   {
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.XASYNC_HALT,
         ProtocolConstants.XASYNC_HANDLE, m_handle );
      responses.clear();
      m_is_complete = true;
   }

   public boolean isComplete()
      throws AuthenticationExpiredException, XAMException
   {
      return m_is_complete;
//         HashMap<String, String> responses = HTTPRequest.request(m_host,
//            ProtocolConstants.XASYNC_IS_COMPLETE,
//            ProtocolConstants.XASYNC_HANDLE, m_handle );
//         boolean isComplete = Boolean.parseBoolean(responses.get(ProtocolConstants.ASYNC_COMPLETE));
//         responses.clear();
//         return isComplete;
   }
}
