/*
 * Copyright (c) 2009, Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
package org.snia.xam.vim.http.client;

import java.util.HashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.snia.xam.XAMException;
import org.snia.xam.vim.isolation.server.http.handlers.ProtocolConstants;
import org.snia.xam.vim.isolation.server.util.XAsyncCache;

/**
 * The HTTPXAsyncCallbackManager monitors all active HTTPXAsync operations. Periodically the callback
 * manager will call he POLL operation to determine which HTTPXAsync operations have completed. When 
 * an Async operation is complete, the HTTPXAsync object is marked as being complete. 
 * 
 * If an HTTPXAsync object has completed, two things are done. First, if the HTTPXasync object was 
 * the result of an XStream Read, the data will be retrieved using the XAsyncGetData method.
 * 
 * Second, the application supplied callback will be called, if it has been set for the Async object.
 * 
 * This class will do lots of logging in the event of errors. The logger used is supplied by the creator.
 */
public class HTTPXAsyncCallbackManager
   implements Runnable
{
   private String  m_host           = null;
   private String  m_xsystem_handle = null;
   private long    m_period         = 500;
   private boolean m_running        = true;
   
   protected Logger m_logger = null;

   HTTPXAsyncCallbackManager( String host, String xsystemHandle, long period, Logger logger )
   {
      m_host           = host;
      m_xsystem_handle = xsystemHandle;
      m_period         = period;
      m_logger         = logger;

      Thread worker = new Thread( this, "XAsyncCallbackManager." + xsystemHandle );
      worker.start();
   }
   
   /**
    * The core of the manager. Periodically run the poll method, and issue callbacks.
    */
   public void run()
   {
      while( m_running )
      {
         try
         {
            pollCallback();
            synchronized(this) {
               wait(m_period);
            }
         }
         catch( Exception e )
         {
            m_logger.log( Level.SEVERE, "Error processing XAsync callbacks", e );
         }
      }
      m_logger.fine( "Exiting the XAsync Manager Thread Loop" );
   }
   
   /**
    * Causes the asynchronous manager to stop running.
    */
   public void stop() 
   {
      m_running = false;
      synchronized(this) {
         notifyAll();
         m_logger.fine( "Notified XAsync Manager Thread to stop" );
      }
   }
   
   /**
    * Issue the HTTP POLL method to the server, obtaining a list of completed XAsync operations.
    * @return The HTTPXAsync operations which are completed.
    * @throws XAMException
    */
   private void pollCallback()
      throws XAMException
   {
      XAsyncCache xac = XAsyncCache.getInstance(m_logger);
      
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.ASYNC_CALLBACK_POLL,
         ProtocolConstants.QUALIFIED_XSYSTEM_HANDLE, m_xsystem_handle );
      
      for( String response : responses.keySet() )
      {
         if(response.startsWith( ProtocolConstants.ASYNC_COMPLETE ))
         {
            String handle = (String) responses.get(response);
            m_logger.finest( "XAsync callback, recovered handle: " + response + "/" + handle );
            HTTPXAsync async = null;
            try 
            { 
               async = (HTTPXAsync) xac.getXAsync(Long.parseLong(handle));
               byte buffer[] = async.getBuffer();
               // Get data for 
               
               if( buffer != null )
                  getData( async, handle, m_host, buffer );
               
               async.setIsComplete();
               
               if( (async != null) && (async.m_listener != null) )
                  async.m_listener.XAsyncCallback(async);
            }
            catch( XAMException xe )
            {
               // Invalid handles may show up because the server returned a handle of an async
               // object not being tracked. HTTPXAsync objects are not tracked if they have no
               // listener attached. The sever has no way of knowing, and returns all completed
               // handles. This is a little workaround to avoid false negatives in the log.
               if( xe.getStatusCode() != XAMException.sXAM_INVALID_HANDLE )
                  m_logger.log( Level.SEVERE, "ERROR calling async callback ", xe );
            }
            catch( Throwable t ) 
            {
               m_logger.log( Level.SEVERE, "ERROR calling async callback ", t );
            }
         }
      }
      responses.clear();
      return;
   }

   /**
    * Send a request to read the data.
    * @param handle The XAsync handle
    * @param m_host The host:port string for the request.
    */
   private void getData( HTTPXAsync async, String handle, String m_host, byte buffer[] )
      throws XAMException
   {
      try {
         /* Length return is ignored */
         HTTPRequest.requestNoStatus(m_host,
            ProtocolConstants.ASYNC_CALLBACK_GETDATA,
            ProtocolConstants.XASYNC_HANDLE, handle,
            ProtocolConstants.XSTREAM_NUMBYTES, Long.toString(buffer.length),
            buffer, 0, buffer.length );
      } catch( Exception e ) { e.printStackTrace(); } 
   }
}
