/*
 * Copyright (c) 2009, Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
package org.snia.xam.vim.http.client;

import java.util.HashMap;
import java.util.logging.Logger;

import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XSetUnderHoldException;
import org.snia.xam.XStream;
import org.snia.xam.XStreamAbandonException;
import org.snia.xam.XStreamCorruptException;
import org.snia.xam.XStreamException;
import org.snia.xam.vim.isolation.server.http.handlers.ProtocolConstants;

public class HTTPXStream
   implements XStream
{
   String LATER = "Please implement this function later";
   private String m_host;
   private String m_handle;
   private Logger m_logger = null;
   
   HTTPXStream( String host, String handle, Logger logger )
   {
      m_host   = host;
      m_handle = handle;
      m_logger = logger;
   }
   public void abandon()
      throws AuthenticationExpiredException, XAMException
   {
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.XSTREAM_ABANDON_HANDLER,
         ProtocolConstants.XSTREAM_HANDLE, m_handle );
      responses.clear();
   }

   public XAsync asyncClose(long xopid, XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamCorruptException, XStreamException, XAMException
   {
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.XSTREAM_ASYNC_CLOSE_HANDLER,
         ProtocolConstants.XSTREAM_HANDLE, m_handle,
         ProtocolConstants.ASYNC_XOPID, Long.toString(xopid) );
      String asyncHandle = responses.get(ProtocolConstants.XASYNC_HANDLE);
      responses.clear();
      return new HTTPXAsync( m_host, asyncHandle, listener, m_logger );
   }

   public XAsync asyncRead(byte[] buffer,
                           long bufferLength,
                           long xopid,
                           XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XAMException
   {
      if( bufferLength > Integer.MAX_VALUE )
         throw new InvalidArgumentException( "bufferLength must be <= to Integer.MAX_VALUE" );
      
      if( bufferLength > buffer.length )
         throw new InvalidArgumentException( "bufferLength is longer than available buffer space" );
      
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.XSTREAM_ASYNC_READ_HANDLER,
         ProtocolConstants.XSTREAM_HANDLE, m_handle,
         ProtocolConstants.XSTREAM_NUMBYTES, Long.toString(bufferLength),
         ProtocolConstants.ASYNC_XOPID, Long.toString(xopid) );
      String asyncHandle = responses.get(ProtocolConstants.XASYNC_HANDLE);
      responses.clear();
      HTTPXAsync async = new HTTPXAsync( m_host, asyncHandle, listener, m_logger );
      async.setBuffer( buffer );
      return async;
   }

   public XAsync asyncWrite(byte[] buffer,
                            long bufferLength,
                            long xopid,
                            XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XAMException
   {
      if( bufferLength > buffer.length )
         throw new InvalidArgumentException( "bufferLength is beyond the end of the actual buffer" );
      
     if( bufferLength < 0 )
        throw new InvalidArgumentException( "bufferLength may not be negative" );
      
      HashMap<String, String> responses = HTTPRequest.requestWrite(m_host,
         ProtocolConstants.XSTREAM_ASYNC_WRITE_HANDLER,
         ProtocolConstants.XSTREAM_HANDLE, m_handle,
         ProtocolConstants.ASYNC_XOPID, Long.toString(xopid),
         (int) 0, (int) bufferLength, buffer );
      String asyncHandle = responses.get(ProtocolConstants.XASYNC_HANDLE);
      responses.clear();
      HTTPXAsync async = new HTTPXAsync( m_host, asyncHandle, listener, m_logger );
      return async;
   }

   public void close()
      throws AuthenticationExpiredException, XSetUnderHoldException,
      XStreamCorruptException, XStreamException, XAMException
   {
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.XSTREAM_CLOSE_HANDLER,
         ProtocolConstants.XSTREAM_HANDLE, m_handle );
      responses.clear();
   }

   public long read(byte[] buffer, long offset, long bytesToRead)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XAMException
   {
      if( (bytesToRead + offset) > buffer.length )
         throw new InvalidArgumentException( "bytesToRead + offset is beyond the end of the buffer" );
      
     if( (bytesToRead < 0) || (offset < 0) )
        throw new InvalidArgumentException( "bytesToRead and offset may not be negative" );
      
      int length = HTTPRequest.requestNoStatus(m_host,
         ProtocolConstants.XSTREAM_READ_HANDLER,
         ProtocolConstants.XSTREAM_HANDLE, m_handle,
         ProtocolConstants.XSTREAM_NUMBYTES, Long.toString(bytesToRead), 
         buffer, (int) offset, (int) bytesToRead );
      return length;      
   }

   public long read(byte[] buffer, long bytesToRead)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XAMException
   {
      if( bytesToRead > buffer.length )
         throw new InvalidArgumentException( "bytesToRead argument must be <= to buffer length" );
      
      if( bytesToRead < 0 )
         throw new InvalidArgumentException( "bytesToRead may not be negative" );
       
      int length = HTTPRequest.requestNoStatus(m_host,
         ProtocolConstants.XSTREAM_READ_HANDLER,
         ProtocolConstants.XSTREAM_HANDLE, m_handle,
         ProtocolConstants.XSTREAM_NUMBYTES, Long.toString(bytesToRead), 
         buffer, 0, (int) bytesToRead );
      return length;      
   }

   public long read(byte[] buffer)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XAMException
   {
      int length = HTTPRequest.requestNoStatus(m_host,
         ProtocolConstants.XSTREAM_READ_HANDLER,
         ProtocolConstants.XSTREAM_HANDLE, m_handle,
         ProtocolConstants.XSTREAM_NUMBYTES, Integer.toString(buffer.length), 
         buffer, 0, buffer.length );
      return length;
   }

   public void seek(long offset, long whence)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XAMException
   {
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.XSTREAM_SEEK_HANDLER,
         ProtocolConstants.XSTREAM_HANDLE, m_handle,
         ProtocolConstants.XSTREAM_OFFSET, Long.toString(offset),
         ProtocolConstants.XSTREAM_WHENCE, Long.toString(whence) );
      responses.clear();
   }

   public long tell()
      throws AuthenticationExpiredException, XStreamAbandonException,
      XStreamCorruptException, XAMException
   {
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.XSTREAM_TELL_HANDLER,
         ProtocolConstants.XSTREAM_HANDLE, m_handle );
      long position = Long.valueOf(responses.get(ProtocolConstants.XSTREAM_POSITION));
      responses.clear();
      return position;
   }

   public long write(byte[] buffer)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XSetUnderHoldException, XAMException
   {
      HashMap<String, String> responses = HTTPRequest.requestWrite(m_host,
         ProtocolConstants.XSTREAM_WRITE_HANDLER,
         ProtocolConstants.XSTREAM_HANDLE, m_handle,
         0, buffer.length, buffer );
      long numbytes = Long.parseLong(responses.get(ProtocolConstants.XSTREAM_NUMBYTES));
      responses.clear();
      return numbytes;
   }

   public long write(byte[] buffer, long bytesToWrite)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XSetUnderHoldException, XAMException
   {
      if( bytesToWrite > buffer.length )
         throw new XStreamException( "bytesToWrite is larger than the buffer" );
      HashMap<String, String> responses = HTTPRequest.requestWrite(m_host,
         ProtocolConstants.XSTREAM_WRITE_HANDLER,
         ProtocolConstants.XSTREAM_HANDLE, m_handle,
         0, (int) bytesToWrite, buffer );
      long numbytes = Long.parseLong(responses.get(ProtocolConstants.XSTREAM_NUMBYTES));
      responses.clear();
      return numbytes;
   }

   public long write(byte[] buffer, long offset, long bytesToWrite)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XSetUnderHoldException, XAMException
   {
      if( (bytesToWrite + offset) > buffer.length )
         throw new InvalidArgumentException( "bytesToRead + offset is beyond the end of the buffer" );
      
     if( (bytesToWrite < 0) || (offset < 0) )
        throw new InvalidArgumentException( "bytesToRead and offset may not be negative" );
      
      HashMap<String, String> responses = HTTPRequest.requestWrite(m_host,
         ProtocolConstants.XSTREAM_WRITE_HANDLER,
         ProtocolConstants.XSTREAM_HANDLE, m_handle,
         (int) offset, (int) bytesToWrite, buffer );
      long numbytes = Long.parseLong(responses.get(ProtocolConstants.XSTREAM_NUMBYTES));
      responses.clear();
      return numbytes;
   }
}
