/*
 * Copyright (c) 2009, Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
package org.snia.xam.vim.http.client;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.Calendar;
import java.util.HashMap;
import java.util.logging.Logger;

import org.snia.xam.AuthenticationException;
import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.AuthorizationException;
import org.snia.xam.ConnectException;
import org.snia.xam.FieldContainer;
import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.FieldExistsException;
import org.snia.xam.HoldIdException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidFieldNameException;
import org.snia.xam.InvalidOperationException;
import org.snia.xam.InvalidXSetModeException;
import org.snia.xam.InvalidXUIDException;
import org.snia.xam.MaximumFieldException;
import org.snia.xam.ObjectInUseException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XIterator;
import org.snia.xam.XSet;
import org.snia.xam.XSetAbandonException;
import org.snia.xam.XSetCorruptException;
import org.snia.xam.XSetInaccessibleException;
import org.snia.xam.XSetUnderHoldException;
import org.snia.xam.XSetUnderRetentionException;
import org.snia.xam.XStream;
import org.snia.xam.XSystem;
import org.snia.xam.XSystemAbandonException;
import org.snia.xam.XSystemCorruptException;
import org.snia.xam.XSystemException;
import org.snia.xam.XUID;

import org.snia.xam.base.DefaultXUID;
import org.snia.xam.base.XAMImplementation;
import org.snia.xam.base.XRI;
import org.snia.xam.util.LogManager;
import org.snia.xam.vim.isolation.server.http.handlers.ProtocolConstants;
import org.snia.xam.vim.isolation.server.util.Constants;

/**
 * The HTTPXSystem class implements the client half of the HTTP Protocol VIM for Java. The operations of
 * this VIM are not optimized for throughput or scalability, but show one example of a stackable and remote
 * VIM construction. 
 * 
 * NOTE: Because of the way that the XAMLibrary initialization protocol works, this object keeps track of 
 * all copied fields from the XAMLibrary (most likely logging fields). This is because the HTTP protocol
 * does not exactly match the VIM protocol (there is no createVIM call). Upon connect() this object copies
 * the cached fields to the remote VIM. Because of the way this is done, and in full knowledge of the 
 * XAMLibrary implementation, the deleteField() operation will not pass through during the initialization 
 * phase. Instead this class understands that the field .xsystem.initializing is to be deleted prior to 
 * commencing the connect method.
 * 
 * A better way to implement this is would be to implement a field operation journal, then replay the 
 * the journal just prior to connect.
 *
 */
public class HTTPXSystem extends HTTPFieldContainer
   implements org.snia.xam.XSystem
{
   static final String JOURNAL_MSG     = "General field manipulation not supported during intialization, HTTPXSystem needs a field journaling mechanism for the initialization phase.";
   static final String AUTH_XFER_ERROR = "Unable to transfer the authenticate response to the ByteArrayOutputStream";
   
   private boolean m_initializing = true;
   
   public final String CLIENT_LOG_PATH = "org.snia.xam.http.log.path";

   /** Used to hold existing fields until connect is called. */
   FieldContainer m_temp_field_container = new TempFieldContainer();
   
   final long CALLBACK_PERIOD = 500; /* ms */

   /** Callback Manager to issue callbacks when XAsync operations are finished. */
   HTTPXAsyncCallbackManager m_callback_mgr = null;
   
   /**
    * Set up the HTTP XSystem, client side.
    */
   public HTTPXSystem()
   {
      m_handle_label                   = ProtocolConstants.XSYSTEM_HANDLE;
      m_create_property_command         = ProtocolConstants.XSYSTEM_CREATEPROPERTY_HANDLER_NAME;
      m_set_property_command            = ProtocolConstants.XSYSTEM_SETPROPERTY_HANDLER_NAME;
      m_get_property_command            = ProtocolConstants.XSYSTEM_GETPROPERTY_HANDLER_NAME;
      m_delete_field_command            = ProtocolConstants.XSYSTEM_DELETEFIELD_HANDLER_NAME;
      m_get_field_length_command        = ProtocolConstants.XSYSTEM_GET_FIELDLENGTH_HANDLER_NAME;
      m_get_field_type_command          = ProtocolConstants.XSYSTEM_GETPROPERTYTYPE_HANDLER_NAME;
      m_get_field_readonly_command      = ProtocolConstants.XSYSTEM_GET_FIELD_READONLY_HANDLER;
      m_get_field_binding_command       = ProtocolConstants.XSYSTEM_GET_FIELD_BINDING_HANDLER;
      m_set_field_as_binding_command    = ProtocolConstants.XSYSTEM_SET_FIELD_AS_BINDING_HANDLER;
      m_set_field_as_nonbinding_command = ProtocolConstants.XSYSTEM_SET_FIELD_AS_NONBINDING_HANDLER;
      m_create_xstream_command          = ProtocolConstants.XSYSTEM_CREATE_XSTREAM_HANDLER;
      m_open_xstream_command            = ProtocolConstants.XSYSTEM_OPEN_XSTREAM_HANDLER;
      m_open_field_iterator_command     = ProtocolConstants.XAM_CREATE_FIELD_ITERATOR_HANDLER;
      m_xiterator_type_property         = ProtocolConstants.XITERATOR_CLASS_XSYSTEM;
      m_contains_field_command          = ProtocolConstants.XSYSTEM_CONTAINS_FIELD_HANDLER;
      m_async_open_xstream_command      = ProtocolConstants.XSYSTEM_ASYNC_OPENXSTREAM_HANDLE;
   }

   public void abandon()
      throws AuthenticationExpiredException, XAMException
   {
      HashMap<String, String> responses = HTTPRequest.request(m_host,
         ProtocolConstants.XSYSTEM_ABANDON_HANDLER_NAME,
         ProtocolConstants.XSYSTEM_HANDLE, m_handle);
      responses.clear();
   }

   public boolean accessXSet(XUID xsetid, long mode)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      try 
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            ProtocolConstants.XSYSTEM_ACCESSXSET_HANDLER_NAME,
            ProtocolConstants.PTYPE_XUID, URLEncoder.encode(xsetid.toString(), ProtocolConstants.UTF8),
            ProtocolConstants.MODE_PROPERTY, Long.toString(mode),
            ProtocolConstants.XSYSTEM_HANDLE, m_handle);
         boolean returnValue = Boolean.parseBoolean(responses.get(ProtocolConstants.XSYSTEM_ACCESSIBLE_PROPERTY));
         responses.clear();
         return returnValue;
      }
      catch( UnsupportedEncodingException uee )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR );
      }
   }

   public XAsync asyncCopyXSet(XUID inXuid,
                               String mode,
                               long xopid,
                               XAsyncListener listener)
      throws AuthenticationExpiredException, AuthorizationException,
      InvalidArgumentException, XSetInaccessibleException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      try 
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            ProtocolConstants.XSYSTEM_ASYNC_COPYXSET_HANDLER_NAME,
            ProtocolConstants.XSYSTEM_HANDLE, m_handle,
            ProtocolConstants.PTYPE_XUID, URLEncoder.encode(inXuid.toString(), ProtocolConstants.UTF8),
            ProtocolConstants.MODE_PROPERTY, URLEncoder.encode(mode, ProtocolConstants.UTF8),
            ProtocolConstants.ASYNC_XOPID, Long.toString(xopid) );
         String async_handle = responses.get(ProtocolConstants.XASYNC_HANDLE);
         responses.clear();
         return new HTTPXAsync( m_host, async_handle, listener, m_logger );
      }
      catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR );
      }
   }

   public XAsync asyncOpenXSet(XUID inXuid,
                               String mode,
                               long xopid,
                               XAsyncListener listener)
      throws AuthorizationException, AuthenticationExpiredException,
      InvalidArgumentException, XSetInaccessibleException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      try 
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            ProtocolConstants.XSYSTEM_ASYNC_OPENXSET_HANDLER,
            ProtocolConstants.XSYSTEM_HANDLE, m_handle,
            ProtocolConstants.PTYPE_XUID, URLEncoder.encode(inXuid.toString(), ProtocolConstants.UTF8),
            ProtocolConstants.MODE_PROPERTY, URLEncoder.encode(mode, ProtocolConstants.UTF8),
            ProtocolConstants.ASYNC_XOPID, Long.toString(xopid) );
         String async_handle = responses.get(ProtocolConstants.XASYNC_HANDLE);
         responses.clear();
         return new HTTPXAsync( m_host, async_handle, listener, m_logger );
      }
      catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR );
      }
   }

   public byte[] authenticate(byte[] buffer)
      throws AuthenticationException, InvalidArgumentException, XAMException
   {
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      
      /* Note return code is ignored in this method, which is why it's been deprecated. */
      authenticate(buffer, baos );
      return baos.toByteArray();
   }

   public int authenticate(byte[] buffer, ByteArrayOutputStream responseOutputStream )
      throws AuthenticationException, InvalidArgumentException, XAMException
   {
      HashMap<String, String> responses = HTTPRequest.requestWrite(m_host,
         ProtocolConstants.XSYSTEM_AUTHENTICATE_HANDLER,
         ProtocolConstants.XSYSTEM_HANDLE, m_handle,
         0, buffer.length, buffer );
      String responseHandle = responses.get(ProtocolConstants.XSTREAM_HANDLE);
      HTTPXStream responseStream = new HTTPXStream( m_host, responseHandle, m_logger );
      
      // Get length of stream
      responseStream.seek(0, XStream.SEEK_END );
      long length = responseStream.tell();
      responseStream.seek(0, XStream.SEEK_SET);
      
      // Allocate enough bytes, then return.
      byte responseBytes[] = new byte[(int) length];
      responseStream.read(responseBytes);
      responseStream.close();
      
      try { responseOutputStream.write(responseBytes); }
      catch( IOException ioe ) { throw new AuthenticationException( AUTH_XFER_ERROR, ioe ); }
      
      String authCode = responses.get(ProtocolConstants.AUTHENTICATE_STATUS);
      responses.clear();
      return Integer.parseInt(authCode);
   }

   public void close()
      throws AuthenticationExpiredException, ObjectInUseException, XAMException
   {
      try 
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            ProtocolConstants.XSYSTEM_CLOSE_HANDLER_NAME,
            ProtocolConstants.XSYSTEM_HANDLE, m_handle);
         responses.clear();
      } 
      finally 
      {
         m_callback_mgr.stop();
      }
   }

   public void connect(String xri)
      throws XAMException
   {
      try
      {
         XRI incomingXRI = XRI.valueOf(xri);
         String host = incomingXRI.getParameterValue(ProtocolConstants.XRI_HOST_ARG);
         String port = incomingXRI.getParameterValue(ProtocolConstants.XRI_PORT_ARG);
         
         if( (host == null) || (host == null) )
            throw new ConnectException( "XRI did not contain the required arguments: " +
                                        ProtocolConstants.XRI_HOST_ARG + " or " +
                                        ProtocolConstants.XRI_PORT_ARG );
         m_host = host + ":" + port;
         
         byte buffer[] = createInitializeBuffer(m_temp_field_container);
         HashMap<String, String> responses = HTTPRequest.requestWrite(m_host,
            ProtocolConstants.XSYSTEM_CONNECT_HANDLER_NAME,
            "xri", URLEncoder.encode(xri, ProtocolConstants.UTF8),
            0, buffer.length, buffer );
         m_handle = responses.get(m_handle_label);
         m_initializing = false;
         m_logger = updateLogger(m_temp_field_container);
         m_callback_mgr = new HTTPXAsyncCallbackManager( m_host, m_handle, CALLBACK_PERIOD, m_logger );
         m_temp_field_container = null;
      }
      catch( XAMException xe )
      { 
         throw xe; // So it doesn't get wrapped again.
      }
      catch( Exception e )
      { 
         throw new ConnectException( "Failed connection ", e );
      }
   }

   /**
    * Used to turn the temporary field container into a byte buffer to be sent on the 
    * XSystem.connect().
    * @param fc The FieldContainer containing the properties to be sent.
    * @return The byte buffer of the field container
    */
   private byte[] createInitializeBuffer(FieldContainer fc )
      throws XAMException
   {
      final String EQ       = "=";
      final String SPACE    = " ";
      final String VALUE    = ".value";
      final String BINDING  = ".binding";
      final String TYPE     = ".type";
      
      XIterator xi = fc.openFieldIterator("");
      StringBuffer strBuf = new StringBuffer();
      try { 
         while( xi.hasNext() )
         {
            String property = xi.next();
            String mimeType = fc.getFieldType(property);
            String encProperty = URLEncoder.encode( property, ProtocolConstants.UTF8 );
            String protocolType = null;
            String encValue     = null;
         
            if( mimeType.equals(XAMLibrary.STYPE_BOOLEAN_MIME_TYPE )) {
               protocolType = ProtocolConstants.XAM_BOOLEAN;
               encValue = Boolean.toString(fc.getBoolean(property));
            } else if( mimeType.equals(XAMLibrary.STYPE_DATETIME_MIME_TYPE)) {
               protocolType = ProtocolConstants.XAM_DATETIME;
               encValue = Long.toString(fc.getDateTime(property).getTimeInMillis());
            } else if( mimeType.equals(XAMLibrary.STYPE_DOUBLE_MIME_TYPE)) {
               protocolType = ProtocolConstants.XAM_DOUBLE;
               encValue = Double.toString(fc.getDouble(property));
            } else if( mimeType.equals(XAMLibrary.STYPE_INT_MIME_TYPE)) {
               protocolType = ProtocolConstants.XAM_INTEGER;
               encValue = Long.toString(fc.getLong(property));
            } else if( mimeType.equals(XAMLibrary.STYPE_STRING_MIME_TYPE )) {
               protocolType = ProtocolConstants.XAM_STRING;
               encValue = URLEncoder.encode(fc.getString(property), ProtocolConstants.UTF8);
            } else if( mimeType.equals(XAMLibrary.STYPE_XUID_MIME_TYPE)) {
               protocolType = ProtocolConstants.XAM_XUID ;
               encValue = URLEncoder.encode(fc.getXUID(property).toString(), ProtocolConstants.UTF8);         
            } else
               throw new ConnectException( "Unrecognized STYPE in initialization phase of connect: " + mimeType );

            // Value
            strBuf.append(SPACE);
            strBuf.append( encProperty );
            strBuf.append( VALUE );
            strBuf.append(EQ);
            strBuf.append( encValue );
            strBuf.append(SPACE);
         
            // type
            strBuf.append(property);
            strBuf.append(TYPE);
            strBuf.append(EQ);
            strBuf.append(protocolType);
            strBuf.append(SPACE);
         
            // binding
            strBuf.append(property);
            strBuf.append(BINDING);
            strBuf.append(EQ);
            strBuf.append(fc.getFieldBinding(property));
            strBuf.append(SPACE);
         }
      
         return strBuf.toString().getBytes(ProtocolConstants.UTF8); 
      }
      catch( UnsupportedEncodingException uee ) {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR );
      }
   }

   /** When the logging parameters are set up, make sure the logger is initialized and created 
    * according the to the XAM SDK logging rules. 
    * @param fc The FieldContainer holding properties which control logging.
    * @return The logger object.
    * @throws XAMException
    */
   private Logger updateLogger(FieldContainer fc )
      throws XAMException
   {
      // Read the log configuration from the property files.
      String path    = Constants.LOGGER_NAME;
      long   lsize   = LogManager.DEFAULT_LOG_SIZE;
      long   lcount  = LogManager.DEFAULT_LOG_COUNT;
      long   level   = XAMLibrary.XAM_LOG_FATAL;
      boolean append = false;
  
      try { path = fc.getString(XAMLibrary.XAM_LOG_PATH); }
      catch( FieldDoesNotExistException fdne ) { /* go with default */ }
      
      try { path = fc.getString(CLIENT_LOG_PATH); }
      catch( FieldDoesNotExistException fdne ) { /* go with default */ }
      
      try { level = fc.getLong(XAMLibrary.XAM_LOG_LEVEL); }
      catch( FieldDoesNotExistException fdne ) { /* go with default */ }
   
      try { append = fc.getBoolean( XAMImplementation.XAM_LOG_APPEND ); }
      catch( FieldDoesNotExistException fdne ) { /* go with default */ }

      try 
      { 
         lsize = fc.getLong( XAMLibrary.XAM_LOG_MAX_SIZE ) * 1000; // The property is in kB 
         if( (lsize > Integer.MAX_VALUE) || (lsize < 0) )
            throw new InvalidArgumentException( "Log file size is too large, or less than zero" );
      }
      catch( FieldDoesNotExistException x ) { }
   
      try 
      { 
         lcount = fc.getLong( XAMLibrary.XAM_LOG_MAX_ROLLOVERS ); 
         if( (lcount > Integer.MAX_VALUE) || (lcount < 0) )
            throw new InvalidArgumentException( "Log count is too large, or less than zero" );
      }
      catch(FieldDoesNotExistException x) { }
   
      Logger logger = LogManager.GetLogger(path, (int) lsize, (int) lcount, append);
      logger.setLevel(LogManager.mapLogLevelToJavaLevel(level));
      return logger;
   }

   public XSet copyXSet(XUID xsetid, String copyMode)
      throws AuthenticationExpiredException, AuthorizationException,
      InvalidArgumentException, InvalidXSetModeException,
      XSetInaccessibleException, XSystemAbandonException,
      XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            ProtocolConstants.XSYSTEM_COPY_XSET_HANDLER_NAME,
            ProtocolConstants.XUID_PROPERTY, URLEncoder.encode(xsetid.toString(), ProtocolConstants.UTF8),
            ProtocolConstants.MODE_PROPERTY, copyMode,
            ProtocolConstants.XSET_HANDLE, m_handle);
         String xset_handle = responses.get(m_handle_label);
         responses.clear();
         return new HTTPXSet(m_host, xset_handle);
      }
      catch( UnsupportedEncodingException uee )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, uee );
      }
   }

   public XSet createXSet(String createMode)
      throws AuthenticationExpiredException, AuthorizationException,
      InvalidArgumentException, XSystemAbandonException,
      XSystemCorruptException, XAMException
   {
      String xset_handle = null;
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            ProtocolConstants.XSYSTEM_CREATEXSET_HANDLER_NAME,
            ProtocolConstants.XSYSTEM_HANDLE, m_handle,
            ProtocolConstants.MODE_PROPERTY, URLEncoder.encode(createMode, ProtocolConstants.UTF8) );
         xset_handle = responses.get(m_handle_label);
         responses.clear();
      }
      catch( UnsupportedEncodingException uee )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, uee );
      }
      return new HTTPXSet(m_host, xset_handle);
   }

   public XUID createXUID(String inBase64XUID)
      throws AuthenticationExpiredException, InvalidXUIDException, XAMException
   {
      return DefaultXUID.valueOf(inBase64XUID);
   }

   public XUID createXUID(byte[] inBinaryXUID)
      throws AuthenticationExpiredException, InvalidXUIDException, XAMException
   {
      return DefaultXUID.valueOf(inBinaryXUID);
   }

   public void deleteXSet(XUID xsetid)
      throws AuthorizationException, AuthenticationExpiredException,
      XSetUnderRetentionException, XSetUnderHoldException,
      InvalidArgumentException, XSetInaccessibleException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            ProtocolConstants.XSYSTEM_DELETEXSET_HANDLER_NAME,
            ProtocolConstants.XUID_PROPERTY, URLEncoder.encode(xsetid.toString(), ProtocolConstants.UTF8),
            ProtocolConstants.XSYSTEM_HANDLE, m_handle);
         responses.clear();
      }
      catch( UnsupportedEncodingException uee )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, uee );
      }
   }

   public Calendar getXSetAccessTime(XUID xsetid)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XSetInaccessibleException, XSystemAbandonException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            ProtocolConstants.XSYSTEM_GET_XSET_ACCESS_TIME,
            ProtocolConstants.PTYPE_XUID, URLEncoder.encode(xsetid.toString(), ProtocolConstants.UTF8 ),
            ProtocolConstants.XSYSTEM_HANDLE, m_handle);
         long millis = Long.valueOf(responses.get(ProtocolConstants.ARGUMENT_ACCESSTIME));
         Calendar c = Calendar.getInstance();
         c.setTimeInMillis(millis);
         responses.clear();
         return c;
      }
      catch( UnsupportedEncodingException uee )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, uee );
      }
   }

   public void holdXSet(XUID xsetid, String holdID)
      throws AuthenticationExpiredException, AuthorizationException,
      InvalidFieldNameException, HoldIdException, InvalidArgumentException,
      XSetInaccessibleException, XSystemAbandonException,
      XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            ProtocolConstants.XSYSTEM_HOLD_XSET,
            ProtocolConstants.XUID_PROPERTY, URLEncoder.encode(xsetid.toString(), ProtocolConstants.UTF8),
            ProtocolConstants.HOLD_ID, URLEncoder.encode(holdID, ProtocolConstants.UTF8),
            ProtocolConstants.XSYSTEM_HANDLE, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public boolean isXSetRetained(XUID xsetid)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XSetInaccessibleException, XSystemAbandonException,
      XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            ProtocolConstants.XSYSTEM_IS_XSET_RETAINED,
            ProtocolConstants.XUID_PROPERTY, URLEncoder.encode(xsetid.toString(), ProtocolConstants.UTF8 ),
            ProtocolConstants.XSYSTEM_HANDLE, m_handle);
         boolean retained = Boolean.valueOf(responses.get(ProtocolConstants.ARGUMENT_RETAINED));
         responses.clear();
         return retained;
      }
      catch( UnsupportedEncodingException uee )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, uee );
      }
   }

   public XSet openXSet(XUID xsetid, String openMode)
      throws AuthenticationExpiredException, AuthorizationException,
      InvalidArgumentException, InvalidXSetModeException,
      XSetInaccessibleException, XSystemAbandonException,
      XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            ProtocolConstants.XSYSTEM_OPENXSET_HANDLER_NAME,
            ProtocolConstants.XSYSTEM_HANDLE, m_handle,
            ProtocolConstants.XUID_PROPERTY, URLEncoder.encode(xsetid.toString(), ProtocolConstants.UTF8 ),
            ProtocolConstants.MODE_PROPERTY, URLEncoder.encode(openMode, ProtocolConstants.UTF8) );
         String xset_handle = responses.get(m_handle_label);
         responses.clear();
         return new HTTPXSet(m_host, xset_handle);
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   public void releaseXSet(XUID xsetid, String holdID)
      throws AuthenticationExpiredException, AuthorizationException,
      HoldIdException, InvalidArgumentException, XSetInaccessibleException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      try
      {
         HashMap<String, String> responses = HTTPRequest.request(m_host,
            ProtocolConstants.XSYSTEM_RELEASE_XSET,
            ProtocolConstants.XUID_PROPERTY, URLEncoder.encode(xsetid.toString(), ProtocolConstants.UTF8 ),
            ProtocolConstants.HOLD_ID, URLEncoder.encode(holdID, ProtocolConstants.UTF8),
            ProtocolConstants.XSYSTEM_HANDLE, m_handle);
         responses.clear();
      } catch( UnsupportedEncodingException e )
      {
         throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR, e );
      }
   }

   // Field Container Methods.
   public void createProperty(String fieldName, boolean binding, boolean value)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldExistsException, InvalidOperationException, MaximumFieldException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      if( m_initializing )
      {
         // Hold the properties temporarily until connect time.
         m_temp_field_container.createProperty(fieldName, binding, value );
      }
      else
         super.createProperty( fieldName, binding, value );
   }

   public void createProperty(String fieldName, boolean binding, long value)
      throws AuthenticationExpiredException, FieldExistsException,
      InvalidOperationException, MaximumFieldException, ObjectInUseException,
      XSetAbandonException, XSetCorruptException, XSystemAbandonException,
      XSystemCorruptException, XAMException
   {
      if( m_initializing )
      {
         // Hold the properties temporarily until connect time.
         m_temp_field_container.createProperty(fieldName, binding, value );
      }
      else
         super.createProperty( fieldName, binding, value );
   }

   public void createProperty(String fieldName, boolean binding, double value)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldExistsException, InvalidOperationException, MaximumFieldException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      if( m_initializing )
      {
         // Hold the properties temporarily until connect time.
         m_temp_field_container.createProperty(fieldName, binding, value );
      }
      else
         super.createProperty( fieldName, binding, value );
   }

   public void createProperty(String fieldName, boolean binding, XUID value)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldExistsException, InvalidArgumentException,
      InvalidOperationException, MaximumFieldException, ObjectInUseException,
      XSetAbandonException, XSetCorruptException, XSystemAbandonException,
      XSystemCorruptException, XAMException
   {
      if( m_initializing )
      {
         // Hold the properties temporarily until connect time.
         m_temp_field_container.createProperty(fieldName, binding, value );
      }
      else
         super.createProperty( fieldName, binding, value );
   }

   public void createProperty(String fieldName, boolean binding, String value)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldExistsException, InvalidArgumentException,
      InvalidOperationException, MaximumFieldException, ObjectInUseException,
      XSystemCorruptException, XAMException
   {
      if( m_initializing )
      {
         // Hold the properties temporarily until connect time.
         m_temp_field_container.createProperty(fieldName, binding, value );
      }
      else
         super.createProperty( fieldName, binding, value );
   }

   public void createProperty(String fieldName, boolean binding, Calendar value)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldExistsException, InvalidArgumentException,
      InvalidOperationException, MaximumFieldException, ObjectInUseException,
      XSetAbandonException, XSetCorruptException, XSystemAbandonException,
      XSystemCorruptException, XAMException
   {
      if( m_initializing )
      {
         // Hold the properties temporarily until connect time.
         m_temp_field_container.createProperty(fieldName, binding, value );
      }
      else
         super.createProperty( fieldName, binding, value );
   }
   
   public void deleteField(String fieldName)
      throws AuthenticationExpiredException, InvalidFieldNameException,
             FieldDoesNotExistException, ObjectInUseException, XSetAbandonException,
             XSetCorruptException, XSystemAbandonException, XSystemCorruptException,
             XAMException
   {
      if( m_initializing )
      {
         // No delete supported during initializing phase.
         if( ! fieldName.equals(XSystem.XAM_XSYSTEM_INITIALIZING))
            throw new UnsupportedOperationException( JOURNAL_MSG );
         return;
      }
      super.deleteField(fieldName);
   }

}
