/*
 * 
 * Copyright (c) 2007-2009, Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* Copyright (c) 2007, Sun Microsystems, Inc. */

package org.snia.xam.vim.isolation.server;

import java.util.Enumeration;
import java.util.Properties;
import java.util.logging.Logger;
import java.util.logging.Level;

import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.base.XAMImplementation;
import org.snia.xam.util.LogManager;

import org.snia.xam.vim.isolation.server.handlers.StorageHandler;
import org.snia.xam.vim.isolation.server.http.HttpStorageServer;
import org.snia.xam.vim.isolation.server.ifc.StorageServerException;
import org.snia.xam.vim.isolation.server.util.Constants;
import org.snia.xam.vim.isolation.server.util.HandlerConfig;
import org.snia.xam.vim.isolation.server.util.ServerConfig;
import org.snia.xam.vim.isolation.server.util.ServerConfigKeys;
import org.snia.xam.vim.isolation.server.util.VIMCache;

/**
  * <b>Introduction</b>
  *
  * The set of interfaces/classes described in this package allow to build a 
  * StorageServer and its associated StorageHandler.
  * 
  * <b>Basic Concepts and Key Classes</b>
  * 
  * The StorageServer interface specifies the set of methods to start, stop and 
  * configure the server. After the start method has been called, the server will 
  * start listening for incoming request using the port number that was specified 
  * by the setPort method.
  * 
  * When a new request comes in, either there is a thread available from the pool 
  * to handle the request, or the request is blocked until a thread becomes available. 
  * The number of threads can be specified using the setMaxThreads and setMinThreads 
  * methods. Once a thread is allocated to handle the request, it looks through all 
  * the StorageHandler that have been registered and chooses which one is appropriate 
  * by matching the name of the handler, as returned by the getName method of the 
  * StorageHandler object, to specific information coming from the request. This part 
  * is left implementation dependent and is not specified.
  * 
  * If no registered StorageHandler match the incoming request an exception is returned. 
  * Otherwise the handler is called and it should handle the request by first calling 
  * the createContext method. Then, depending if this is a WriteHandler or a ReadHandler 
  * it should call the methods read or write as long as there are data to be read/written. 
  * In the case of a WriteHandler it can also invoke the flush method to make sure 
  * the state is saved on persistent storage. Finally, the handler should call 
  * deleteContext to free the resource. 
  */
public class IsolationVimServer 
{
    
   private static final int     DEFAULT_LOG_SIZE = 10; // kB
   private static final int     DEFAULT_LOG_COUNT = 1;

   /** Creates a new instance of IsolationVimServer */
    public IsolationVimServer() {
    }
    
    /**
     * Creates a default set of properties for the isolation vim server
     * @return A populated property bundle with all possible keys and
     * default values for those keys.  The keys will match those used
     * in a valid properties file.
     */
    public static Properties getDefaultStartupProperties() {
        Properties p = new Properties();
        p.setProperty(ServerConfigKeys.SERVER_HOST_KEY,Constants.SERVER_HOST);
        p.setProperty(ServerConfigKeys.SERVER_PORT_KEY,Constants.SERVER_PORT);
        p.setProperty(ServerConfigKeys.SERVER_VERSION_KEY,Constants.SERVER_VERSION);
        return p;
    }
    
    /**
     * Creates a default set of properties for the handlers to load
     * into the Isolation VIM
     * @return A populated property bundle with all possible keys and
     * default values for those keys.  The keys will match those used
     * in a valid properties file.
     */
    public static Properties getDefaultHandlerProperties() {
        Properties p = new Properties();
        
        for(int i=0 ; i<Constants.AVAILABLE_HANDLERS.length ; i++) {
            p.setProperty(Constants.AVAILABLE_HANDLERS[i][0], Constants.AVAILABLE_HANDLERS[i][1]);
        }
        
        return p;
    }    
    
    /**
     * Initialized (but does not start) an instance of the
     * storage server.  If storage server startup properties
     * are available, they will be used.  If they are not available,
     * the properties in the Constants interface will be used.
     * @see org.snia.xam.vim.isolation.server.util.Constants
     */
    public static HttpStorageServer getStorageServer()
    throws StorageServerException,XAMException
    {
        
        ServerConfig sc = null;
        HttpStorageServer hss = null;
        
        // attempt to load the properties from a file in the current directory
        try {
            sc = new ServerConfig();
        } catch (Exception e ) {
           Logger logger = LogManager.GetLogger( Constants.LOGGER_NAME );
           logger.log( Level.SEVERE, "Error while processing configuration", e );
        }
        
        // if the properties were not loaded from a file, use defaults
        if(sc==null) {
            sc = new ServerConfig(getDefaultStartupProperties());
        }
        
        updateLogger(sc);
        
        VIMCache xlc = VIMCache.getInstance();
        xlc.setConfiguration(sc);

        hss = new HttpStorageServer(
                sc.getString(ServerConfigKeys.SERVER_VERSION_KEY),
                sc.getString(ServerConfigKeys.SERVER_HOST_KEY),
                (int) sc.getLong(ServerConfigKeys.SERVER_PORT_KEY) );      
        return hss;
    }
    
    /**
     * Initialize the log file, based on configuration file parameters.
     * @param config The configuration object.
     * @throws XAMException
     */
    static void updateLogger(ServerConfig config)
    throws XAMException
    {
       // Read the log configuration from the property files.
       String path    = Constants.LOGGER_NAME;
       long   lsize   = DEFAULT_LOG_SIZE;
       long   lcount  = DEFAULT_LOG_COUNT;
       long   level   = XAMLibrary.XAM_LOG_FATAL;
       boolean append = false;
   
       try { path = config.getString(XAMLibrary.XAM_LOG_PATH); }
       catch( FieldDoesNotExistException fdne ) { /* go with default */ }
       
       try { level = config.getLong(XAMLibrary.XAM_LOG_LEVEL); }
       catch( FieldDoesNotExistException fdne ) { /* go with default */ }
    
       try { append = config.getBoolean( XAMImplementation.XAM_LOG_APPEND ); }
       catch( FieldDoesNotExistException fdne ) { /* go with default */ }

       try 
       { 
          lsize = config.getLong( XAMLibrary.XAM_LOG_MAX_SIZE ) * 1000; // The property is in kB 
          if( (lsize > Integer.MAX_VALUE) || (lsize < 0) )
             throw new InvalidArgumentException( "Log file size is too large, or less than zero" );
       }
       catch( FieldDoesNotExistException x ) { }
    
       try 
       { 
          lcount = config.getLong( XAMLibrary.XAM_LOG_MAX_ROLLOVERS ); 
          if( (lcount > Integer.MAX_VALUE) || (lcount < 0) )
             throw new InvalidArgumentException( "Log count is too large, or less than zero" );
       }
       catch(FieldDoesNotExistException x) { }
    
       Logger logger = LogManager.GetLogger(path, (int) lsize, (int) lcount, append);
       logger.setLevel(LogManager.mapLogLevelToJavaLevel(level));
    }

    /**
     * Loads the handlers into the Isolation VIM.  The handlers should be
     * available in the handler properties file.  If the file is not available,
     * a default list of handlers will be used from the Constants interface.
     * @see org.snia.xam.vim.isolation.server.util.Constants
     */
    public static void initializeHandlers(HttpStorageServer hss) {
        HandlerConfig hc = null;
        
        // attempt to load the properties from a file in the current directory
        try {
            hc = new HandlerConfig();
        } catch (Exception e) {
            Logger logger = null;
            try { logger = LogManager.GetLogger( Constants.LOGGER_NAME ); }
            catch( Exception e2 ) {
               logger = Logger.getLogger( Constants.LOGGER_NAME );
            }
            logger.log( Level.SEVERE, "Error while reading the handler configuration", e );
        }
        
        // if the properties were not loaded from a file, use defaults
        if(hc==null) {
            hc = new HandlerConfig(getDefaultHandlerProperties());
        }
        
        // load the handlers
        Properties p = hc.getRuntimeHandlers();
        
        Enumeration e = p.keys();
        while(e.hasMoreElements()) {
            String key = (String)e.nextElement();
            String value = p.getProperty(key);
            StorageHandler sh = null;
            try {
                sh = (StorageHandler)Class.forName(value).newInstance();
                hss.register(sh, Constants.DEFAULT_VERSION);
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }

    }
    
    public static void main(String[] args) {
        try {
            HttpStorageServer ss = getStorageServer();
            initializeHandlers(ss);
            ss.start();
        } catch(StorageServerException sse) {
            sse.printStackTrace();
        } catch(Exception e) {
            e.printStackTrace();
        }
    }
}
