/*
 *
 * Copyright (c) 2007, Sun Microsystems, Inc.
 * Copyright (c) 2007-2009, Storage Networking Industry Association
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *  
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *  
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *  
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


package org.snia.xam.vim.isolation.server.handlers;

import java.util.Map;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;

import org.snia.xam.vim.isolation.server.ifc.StorageServerException;
import org.snia.xam.vim.isolation.server.contextes.Context;
import org.snia.xam.vim.isolation.server.contextes.ContextKey;

public interface StorageHandler {

    /**
     * Returns the name identifying this handler.
     * 
     * @return          name of the specific handler.
     */
    public String getName();


    /**
     * Returns the size of the <code>ByteBufferList</code>
     * that should be used to write data (store) or read 
     * data (retrieve).
     *
     * @see <code>write</code> of <code>WriteHandler</code>
     * @see <code>read</code> of <code>ReadHandler</code>
     */
    public int getOptimalRequiredSize();


    /**
     * Creates the context for this handler from the information 
     * passed in the properties and the input stream.
     *
     * @param inProps    name/value (IN)
     * @param input      input stream of formatted data 
     * @param outProps   name/value (OUT)
     * @param output     output stream of formatted data 
     * @param ctx        context to be filled
     * @return           key identifiying the created context
     * @throws IOException when there is an error reading/writing to
     *         to the <code>input</code> or <code>output</code> stream
     * @throws StorageServerException (TBD)
     */

    public ContextKey createContext(Map<String,String> inProps,
      InputStream input,
      Map<String,String> outProps,
      OutputStream output,
      Context ctx) throws IOException, StorageServerException;

    /**
     * Recreates the context for this handler from the information 
     * passed in the properties and the input stream. This is used
     * to resume from an operation that has been aborted.
     * <p>
     * In order to resume a write operation, the flush API of 
     * <code>WriteHandler</code> must be used to continuously
     * save the data and context on persistent storage.
     *
     * @param inProps name/value pair of specific information
     * @param input      input stream of formatted data
     * @param outProps name/value pair of specific information
     * @param output     output stream of formatted data  
     * @param ctxKey     key identifying the context
     * @param ctx        context to be filled
     * @throws IOException when there is an error reading/writing to
     *         to the <code>input</code> or <code>output</code> stream
     * @throws StorageServerException (TBD)
     */
    public void reopenContext(Map<String,String> inProps,
      InputStream input,
      Map<String,String> outProps,
      OutputStream output,
      ContextKey ctxKey,
      Context ctx) throws StorageServerException;

    /**
     * Delete the context corresponding to the <code>key</code>.
     *
     * @param ctx        context de be deleted
     * @throws StorageServerException when the context is invalid
     */
    public void closeContext(Context ctx, OutputStream output)
        throws StorageServerException;
}

