/*
 * Copyright (c) 2007-2009, Storage Networking Industry Association
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *  
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *  
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *  
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* Copyright (c) 2007, Sun Microsystems, Inc. */

package org.snia.xam.vim.isolation.server.http;

import java.util.Iterator;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.logging.Logger;
import java.util.logging.Level;

import org.mortbay.http.HttpOutputStream;
import org.mortbay.http.HttpInputStream;
import org.mortbay.http.HttpRequest;
import org.mortbay.http.HttpResponse;
import org.mortbay.http.handler.AbstractHttpHandler;

import org.snia.xam.util.LogManager;
import org.snia.xam.vim.isolation.server.ifc.StorageServerException;
import org.snia.xam.vim.isolation.server.util.Constants;
import org.snia.xam.vim.isolation.server.handlers.StorageHandler;
import org.snia.xam.vim.isolation.server.contextes.Context;
import org.snia.xam.vim.isolation.server.contextes.ContextKey;

public abstract class HttpGenericHandler 
    extends AbstractHttpHandler {

    protected Logger logger = null;

    protected StorageHandler handler;

    public HttpGenericHandler(StorageHandler hdl, String contextPath)
        throws StorageServerException 
    {
       try { 
          logger = LogManager.GetLogger( Constants.LOGGER_NAME );
       } catch( Exception e ) { 
          logger = Logger.getLogger( Constants.LOGGER_NAME );
       }
       
        if (hdl == null) {
            throw new StorageServerException("need to specify a valid handler");
        }
        handler = hdl;
        if (logger.isLoggable(Level.INFO)) {
            logger.info("Created handler "  + handler.getName() +
            ", contextPath = " + contextPath);
        }        
    }

    public void handle(final String pathInContext,
      String pathParams,
      HttpRequest request,
      HttpResponse response)
        throws IOException {

        StringBuffer bufLog = null;
        if (logger.isLoggable(Level.FINE)) {
            bufLog = new StringBuffer();
            bufLog.append("Request " + pathParams + ", HTTP parameters : ");
        }

        Map<String,String> inProps = new HashMap<String,String>();
        Iterator it = request.getParameterNames().iterator();
        while (it.hasNext()) {
            String name = (String) it.next();
            String value = request.getParameter(name);
            inProps.put(name, value);
        }
        
        Enumeration names = request.getFieldNames();
        while (names.hasMoreElements()) {
            String name = (String) names.nextElement();
            String value = request.getField(name);
            inProps.put(name, value);
        }

        if (logger.isLoggable(Level.FINE) &&
          bufLog != null) {
            it = inProps.keySet().iterator();
            while (it.hasNext()) {
                String key = (String) it.next();
                String value = (String) inProps.get(key);
                bufLog.append(key + " = ");
                bufLog.append((value == null) ? "null" : value);
            }
            logger.fine(bufLog.toString());
        }
        
        HttpInputStream  inputStr = (HttpInputStream)response.getInputStream();
        HttpOutputStream outputStr = (HttpOutputStream)response.getOutputStream();

        Map<String,String> outProps = new HashMap<String,String>();
        ContextKey ctxKey = null;
        Context ctx = new Context();
        try {
            //
            // Jetty uses a BufferedOutputStream until we set the chunking
            // true. Therefore we can pass safely the HttpOutputStream to
            // createContext and process the HTTP headers returned by the
            // outProps after we return from createContext.
            //
            ctxKey = createContext(inProps, inputStr, outProps, outputStr, ctx);
            setHttpHeaders(response, outProps);
            outputStr.flush();
            outputStr.setChunking();
            handle(ctx, outputStr, response);
            outputStr.flush();

            closeContext(ctx, outputStr);

        } catch (Exception e) {
            logger.log(Level.SEVERE, 
		       "Failed to handle request ", e);
            request.setHandled(true);
            return;
        } finally {

        }
    }
    
    protected abstract void handle(Context ctx,
      OutputStream out, HttpResponse response)
        throws IOException, StorageServerException;

    protected int getOptimalRequiredSize() {
        return handler.getOptimalRequiredSize();
    }

    private void setHttpHeaders(HttpResponse response, Map<String,String> outProps) {
        Iterator<String> it = outProps.keySet().iterator();
        while (it.hasNext()) {
            String name = it.next();
            String value = outProps.get(name);
            try {
                response.addField(name, value);
            } catch (IllegalStateException ils) {
                logger.severe("Failed to set HTTP header " + 
                  name + " = " + value);
            }
         }
    }

    protected ContextKey createContext(Map<String,String> inProps,
      InputStream input,
      Map<String,String> outProps,
      OutputStream output,
      Context ctx) throws IOException, StorageServerException {
        return handler.createContext(inProps, input, outProps, output, ctx);
    }

    protected void reopenContext(Map<String,String> inProps,
      InputStream input,
      Map<String,String> outProps,
      OutputStream output,
      ContextKey key,
      Context ctx) throws StorageServerException {
        handler.reopenContext(inProps, input, outProps, output, key, ctx);
    }

    protected void closeContext(Context ctx, OutputStream out)
        throws StorageServerException {
        handler.closeContext(ctx, out);
    }
}
