/*
 *
 * Copyright (c) 2007, Sun Microsystems, Inc.
 * Copyright (c) 2007-2009, Storage Networking Industry Association
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *  
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *  
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *  
 *      * Neither the name of Storage Networking Industry Association, Inc.
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

package org.snia.xam.vim.isolation.server.http;

import java.util.Map;
import java.util.HashMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.net.UnknownHostException;
import java.net.ServerSocket;
import java.io.IOException;

import org.mortbay.http.HttpServer;
import org.mortbay.http.HttpListener;
import org.mortbay.http.SocketListener;
import org.mortbay.http.HttpContext;
import org.mortbay.util.InetAddrPort;
import org.mortbay.util.MultiException;

import org.snia.xam.util.LogManager;
import org.snia.xam.vim.isolation.server.ifc.StorageServer;
import org.snia.xam.vim.isolation.server.ifc.StorageServerException;
import org.snia.xam.vim.isolation.server.util.Constants;
import org.snia.xam.vim.isolation.server.handlers.StorageHandler;
import org.snia.xam.vim.isolation.server.handlers.ReadHandler;
import org.snia.xam.vim.isolation.server.handlers.WriteHandler;

/**
 * The HttpStorageServer implements the generic StorageServer interface using
 * HTTP to receive messages and an additional "model" on top of the standard
 * HTTP header and body to provide contents for a particular message.
 */
public class HttpStorageServer implements StorageServer {

    
    private static final int DEFAULT_MIN_THREAD = 1;
    private static final int DEFAULT_MAX_THREAD = 50;
    private static final int MAX_IDLE_TIME = 1000 * 30; // 30 sec

    private Map defContextes;
    private Map vContextes;     // the set of handlers for various wire operations
    private int minThreads;
    private int maxThreads;
    private int httpPort;
    private String addressIP;
    private String defVersion;
    private HttpServer httpServer;

    private Logger logger = null;

    /**
     * Sets up the HttpStorageServer, including the listener on the proper
     * ports, but does not start the storage server
     * @param defaultVersion the version of this storage server using an x.y notation
     * @param address the IP address to listen on, must be a valid a.b.c.d address
     * @param port the address port to listen on, must not conflict with another port
     * @exception StorageServerException thrown if version number not formatted properly,
     * or if Http Server could not be created
     */
    public HttpStorageServer(String defaultVersion, String address, int port) 
        throws StorageServerException {
        
        if (defaultVersion == null ||
          (!defaultVersion.matches("\\d+\\.\\d+"))) {
            throw new StorageServerException("invalid default version " +
                " should be of the form major.minor");
        }

        minThreads = DEFAULT_MIN_THREAD;
        maxThreads = DEFAULT_MAX_THREAD;
        httpPort = port;
        addressIP = address;
        defVersion = defaultVersion;
        defContextes = new HashMap();
        vContextes = new HashMap();
        httpServer = new HttpServer();
        initLogging(true);
    }

    
    private void initLogging(boolean defaultLogging) {
       try { 
          logger = LogManager.GetLogger( Constants.LOGGER_NAME );
       } catch( Exception e ) { 
          logger = Logger.getLogger( Constants.LOGGER_NAME );
       }
    }


    public void register(StorageHandler handler, String version)
        throws StorageServerException {
        
        if (logger.isLoggable(Level.INFO)) {
            logger.info("HttpStorageServer registers handler " +
              handler.getName() + ", version = " + version);
        }

        version = checkArguments(handler, version);
        synchronized(vContextes) {
            Map contextes = (Map) vContextes.get(version);
            if (contextes == null) {
                contextes = new HashMap();
                vContextes.put(version, contextes);
            }
            HttpContext httpContext =
              (HttpContext) contextes.get(handler.getName());
            if (httpContext != null) {
                throw new StorageServerException("handler " +
                  handler.getName() + ", version " +
                    version + " already registered");
            }
            //
            // Registers default version under "/" 
            //
            httpContext = createHttpContext(handler, version);
            httpServer.addContext(httpContext);
            defContextes.put(handler.getName(), httpContext);
        }
    }

    public void unregister(StorageHandler handler, String version)
        throws StorageServerException {

        version = checkArguments(handler, version);
        synchronized(vContextes) {
            Map contextes = (Map) vContextes.get(version);
            if (contextes != null) {
                stopContext(contextes, handler.getName());
            }
            if (version.equals(defVersion)) {
                stopContext(defContextes, handler.getName());                
            }
        }
        throw new StorageServerException("handler " +
          handler.getName() + ", version " +
          version + " was not registered");
    }

    public void setMinThreads(int nbThreads)
        throws StorageServerException {
        minThreads = nbThreads;
        setThreads(nbThreads, true);
    }

    public void setMaxThreads(int nbThreads)
        throws StorageServerException {
        maxThreads = nbThreads;
        setThreads(nbThreads, false);
    }

    public void start()
        throws StorageServerException {

        createHttpServer();
        if (logger.isLoggable(Level.INFO)) {
            logger.info("Will start HttpStorgaeServer threads [" + 
              minThreads + " - " + maxThreads);
        }
        try {
            httpServer.start();
        } catch(MultiException e) {
            throw new StorageServerException("failed to start " +
              "HttpStorageServer", e);
        }
    }

    public void stop(boolean force)
        throws InterruptedException {

        if (logger.isLoggable(Level.INFO)) {
            logger.info("Will stop HttpStorageServer ");
        }

        SocketListener listener = getListener();
        if (listener != null) {
            ServerSocket socket = listener.getServerSocket();
            if (socket != null) {
                try {
                    socket.close();
                } catch (IOException ioe) {
                    logger.warning("Failed to close the socket");
                }
            }
        }
        httpServer.stop(force);
        httpServer.join();
        httpServer.destroy();
    }

    public boolean isStarted() {
        if (httpServer != null) {
            return httpServer.isStarted();
        } else {
            return false;
        }
    }

    private HttpContext createHttpContext( StorageHandler handler,
                                           String version ) throws StorageServerException {

        StringBuffer contextPath = new StringBuffer();
        contextPath.append(handler.getName());

        HttpGenericHandler httpHandler = null;
        HttpContext res = new HttpContext(httpServer, contextPath.toString());
        if (handler instanceof ReadHandler) {
            httpHandler = new HttpReadGlue(handler,  contextPath.toString());
        } else if (handler instanceof WriteHandler) {
            httpHandler = new HttpWriteGlue(handler, contextPath.toString());
        } else {
            throw new StorageServerException("invalid handler");
        }
        res.addHandler(httpHandler);
        res.setRedirectNullPath(false);
        res.setStatsOn(true);
        if (isStarted()) {
            try {
                res.start();
            } catch (Exception e) {
                throw new StorageServerException("can't start handler for "+
                  " context " + handler.getName() + 
                  ", version " + version);
            }
        }
        return res;
    }

    private String checkArguments(StorageHandler handler, String version)
        throws StorageServerException {

        if (handler == null) {
            throw new StorageServerException("handler is null");
        }

        if (version == null) {
            version =  defVersion;
        } else if (!version.matches("\\d+\\.\\d+")) {
            throw new StorageServerException("version should be of the form " +
                "major.minor");
        }
        return version;
    }


    private void createHttpServer()
        throws StorageServerException {

        try {
            httpServer.setStatsOn(true);
            httpServer.setRequestsPerGC(16);
            
            InetAddrPort inet = null;
            if (addressIP != null) {
                inet = new InetAddrPort(addressIP, httpPort);
            } else {
                inet = new InetAddrPort(httpPort);
            }
            SocketListener listener = new SocketListener(inet);
            listener.setMaxIdleTimeMs(MAX_IDLE_TIME);
            listener.setMinThreads(minThreads);
            listener.setMaxThreads(maxThreads);
            listener.setPoolName("HttpStorageServer");
        
            httpServer.addListener(listener);
        } catch (UnknownHostException uhe) {
            throw new StorageServerException("failed to create Jetty server",
                uhe);
        }

    }

    private SocketListener getListener() {

        HttpListener [] listeners = null;
        if (httpServer != null) {
            listeners = httpServer.getListeners();
            if (listeners == null || listeners.length == 0) {
                return null;
            }
        }
        if (! (listeners[0] instanceof SocketListener)) {
            return null;
        }
        return (SocketListener) listeners[0];
    }

    private void setThreads(int nbThreads, boolean min) 
        throws StorageServerException {

        if (isStarted()) {
            SocketListener listener = getListener();
            if (listener == null) {
                throw new StorageServerException("can't set the number " +
                  "of threads");
            }
            if (min) {
                listener.setMinThreads(minThreads);
            } else {
                listener.setMaxThreads(maxThreads);
            }
        }
    }

    private void stopContext(Map contextTable, String contextPath) {

        HttpContext httpContext =
          (HttpContext) contextTable.remove(contextPath);
        if (httpContext != null) {
            try {
                httpContext.stop();
            } catch (InterruptedException ignore) {
            }
            httpServer.removeContext(httpContext);
            return;
        }
    }

}

