/*
 * Copyright (c) 2009, Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
package org.snia.xam.vim.isolation.server.http.handlers;

import java.io.InputStream;
import java.io.OutputStream;
import java.net.URLDecoder;
import java.nio.ByteBuffer;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.snia.xam.InvalidArgumentException;
import org.snia.xam.XAMException;
import org.snia.xam.util.LogManager;
import org.snia.xam.vim.isolation.server.contextes.Context;
import org.snia.xam.vim.isolation.server.contextes.ContextKey;
import org.snia.xam.vim.isolation.server.handlers.ReadHandler;
import org.snia.xam.vim.isolation.server.ifc.StorageServerException;
import org.snia.xam.vim.isolation.server.resources.ByteBufferList;
import org.snia.xam.vim.isolation.server.resources.ByteBufferPool;
import org.snia.xam.vim.isolation.server.util.Constants;

/**
 * Implements boilerplate methods for operation handlers in the HTTP wire
 * operations. This eliminates the need for individual handlers to re-implement
 * common functionality.
 */
public abstract class AbstractReadHandler
   implements ReadHandler
{
   protected int m_optimal_size = 1024; // A guess unless the implementing class
                                        // overrides it.

   protected String m_name = null;
   
   protected static final Long    DEFAULT_HANDLE = new Long(-1);
   
   /* These are for getAndRegister calls in the handlers. Allows having typed null, without casting in line. */
   protected static final String  NULL_STRING  = null;
   protected static final Boolean NULL_BOOLEAN = null;
   protected static final Long    NULL_LONG    = null;

   protected Logger logger = null;

   protected void setupLogger()
   {
      if( logger == null )
      {
         try {
         logger = LogManager.GetLogger(Constants.LOGGER_NAME);
         } catch( Exception e ) {
            Logger.getLogger(Constants.LOGGER_NAME);
         }
      }
   }
   
   final public String getName()
   {
      return m_name;
   }

   final public int getOptimalRequiredSize()
   {
      return m_optimal_size;
   }

   final public void reopenContext(Map<String,String> inProps,
                                   InputStream input,
                                   Map<String,String> outProps,
                                   OutputStream output,
                                   ContextKey ctxKey,
                                   Context ctx)
      throws StorageServerException
   {
   }

   public void closeContext(Context ctx, OutputStream output)
      throws StorageServerException
   {
      ctx.releaseResources();
      ctx.dispose();
   }

   /**
    * Given the property map, log for debug purposes.
    * @param inProps
    */
   final protected void logProperties(Map<String,String> inProps)
   {
      if (logger.isLoggable(Level.FINEST))
      {
         StringBuffer propString = new StringBuffer();
         Set<String> keys = inProps.keySet();
         Iterator<String> i = keys.iterator();
         while (i.hasNext())
         {
            String key = i.next();
            String value = inProps.get(key);
            propString.append(key);
            propString.append("=");
            propString.append(value);
            propString.append("\n");
         }
         logger.finest(propString.toString());
      }
   }

   /**
    * Get the named Boolean property and register it as a transient object.
    * @param inProps Property map containing properties from URL
    * @param ctx Context to be used for this method call
    * @param label The name of the property to get
    * @param defaultValue A default value to be used if no property is found. May be null.
    * @return
    */
   protected Boolean getAndRegister(Map<String,String> inProps,
                                   Context ctx,
                                   String  label,
                                   Boolean defaultValue )
  {
      Boolean value = null;
      String longAsString = inProps.get(label);
      if(longAsString!=null) {
          try {
              value = new Boolean(longAsString);
          } catch(NumberFormatException nfe) {
              value = defaultValue;
          }
      }
      else
         value = defaultValue;

      if( value != null )
         ctx.registerTransientObject(label,value);
      return value;
  }

   /**
    * Get the named Long property and register it as a transient object.
    * @param inProps Property map containing properties from URL
    * @param ctx Context to be used for this method call
    * @param label The name of the property to get
    * @param defaultValue A default value to be used if no property is found. May be null.
    * @return
    */
   protected Long getAndRegister(Map<String,String> inProps,
                                 Context ctx,
                                 String  label,
                                 Long    defaultValue )
  {
      Long value = null;
      String longAsString = inProps.get(label);
      if(longAsString!=null) {
          try {
              value = new Long(longAsString);
          } catch(NumberFormatException nfe) {
              value = defaultValue;
          }
      }
      else
         value = defaultValue;

      if( value != null )
         ctx.registerTransientObject(label,value);
      return value;
  }

   /**
    * Get the named String property and register it as a transient object.
    * @param inProps Property map containing properties from URL
    * @param ctx Context to be used for this method call
    * @param label The name of the property to get
    * @param defaultValue A default value to be used if no property is found. May be null.
    * @return
    */
   protected String getAndRegister(Map<String,String> inProps,
                                   Context ctx,
                                   String  label,
                                   String  defaultValue )
  {
      String value = inProps.get(label);
      if( value == null )
         value = defaultValue;
      
      if( value != null )
         ctx.registerTransientObject(label,value);
      return value;
  }

   /**
    * Get the named Boolean transient object from the context.
    * @param ctx
    * @param label
    * @return
    */
   protected Boolean getBoolean(Context ctx, String label)
      throws InvalidArgumentException
   {
      Boolean value = (Boolean) ctx.getTransientObject(label);
      if( value == null )
         throw new InvalidArgumentException( "Missing argument to method: " + label );
      return value;
   }

   /**
    * Get the named Long transient object from the context.
    * @param ctx
    * @param label
    * @return
    */
   protected Long getLong(Context ctx, String label)
      throws InvalidArgumentException
   {
      Long value = (Long) ctx.getTransientObject(label);
      if( value == null )
         throw new InvalidArgumentException( "Missing argument to method: " + label );
      return value;
   }

   /**
    * Get the named String transient object from the context.
    * Because Strings (and XUID base64) values are from the application space, they have
    * been encoded to be URL safe.
    * @param ctx
    * @param label
    * @return
    */
   protected String getString(Context ctx, String label)
      throws InvalidArgumentException, XAMException
   {
      String value = (String) ctx.getTransientObject(label);
      if( value == null )
         throw new InvalidArgumentException( "Missing argument to method: " + label );
      else {
         try { value = URLDecoder.decode( value, ProtocolConstants.UTF8 ); }
         catch( Exception e ) {
            throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR + " " + label );
         }
      }

      return value;
   }

   /**
    * Get the named Long transient object from the context.
    * @param ctx
    * @param label
    * @return
    */
   protected Long getHandle(Context ctx, String label) throws XAMException
   {
      Long value = (Long) ctx.getTransientObject(label);
      if( value == null )
         throw new InvalidArgumentException( "Invalid handle", XAMException.sXAM_INVALID_HANDLE );
      return value;
   }

   /**
    * Get the named Boolean transient object from the context.
    * @param ctx
    * @param label
    * @return
    */
   protected boolean exists(Context ctx, String label)
   {
      Object value = ctx.getTransientObject(label);
      return (value != null);
   }

   /** Utility to place a single string value in the response buffer */
   protected void respond( StringBuffer response, String label, String value )
   {
      if( response.length() > 0 )
         response.append( "\n" );
      response.append( label );
      response.append( "=" );
      response.append( value );
   }

   /** Utility to place a single char arry value in the response buffer */
   protected void respond( StringBuffer response, String label, char value[] )
   {
      if( response.length() > 0 )
         response.append( "\n" );
      response.append( label );
      response.append( "=" );
      response.append( value );
   }

   /** Utility to place a single long value in the response buffer */
   protected void respond( StringBuffer response, String label, long value )
   {
      if( response.length() > 0 )
         response.append( "\n" );
      response.append( label );
      response.append( "=" );
      response.append( Long.toString(value) );
   }
   
   /** Utility to place a single boolean value in the response buffer */
   protected void respond( StringBuffer response, String label, boolean value )
   {
      if( response.length() > 0 )
         response.append( "\n" );
      response.append( label );
      response.append( "=" );
      response.append( Boolean.toString(value) );
   }
   
   /**
    * Take the string buffer reponse and pack it for return. All handlers call this at the end of 
    * their read method.
    * @param ctx Handler context
    * @param bufferList Handler response buffer
    * @param response Reponse composed by the read handler
    * @return
    */
   int packResponse( Context ctx, ByteBufferList bufferList, StringBuffer response )
   {
      String rtrnCode = response.toString();
      if( logger.isLoggable(Level.FINEST))
         logger.log(Level.FINEST,"rtrnCode="+rtrnCode);
      
      ByteBufferPool pool = ByteBufferPool.getInstance();
      ByteBuffer bb = pool.checkOutBuffer(rtrnCode.getBytes().length);
      bb.clear();
      bb.put(rtrnCode.getBytes());
      bb.flip();
      bufferList.appendBuffer(bb);
      pool.checkInBuffer(bb);
      
      // Put a flag into the context as we will be called again
      ctx.registerTransientObject(ProtocolConstants.COMPLETION_FLAG, Boolean.TRUE );
      return rtrnCode.getBytes().length;
   }
}
