package org.snia.xam.vim.isolation.server.http.handlers;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URLDecoder;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.snia.xam.InvalidArgumentException;
import org.snia.xam.XAMException;
import org.snia.xam.util.LogManager;
import org.snia.xam.vim.isolation.server.contextes.Context;
import org.snia.xam.vim.isolation.server.contextes.ContextKey;
import org.snia.xam.vim.isolation.server.handlers.WriteHandler;
import org.snia.xam.vim.isolation.server.ifc.StorageServerException;
import org.snia.xam.vim.isolation.server.resources.ByteBufferList;
import org.snia.xam.vim.isolation.server.util.Constants;

public abstract class AbstractWriteHandler
   implements WriteHandler
{
   static final Long DEFAULT_HANDLE = new Long(-1);
   
   protected String m_name         = null;
   protected int    m_optimal_size = 1024;
   protected Logger logger         = null;
   
   public abstract void write(Context ctx, ByteBufferList bufferList)
   throws IOException, StorageServerException;

   public abstract void closeContext(Context ctx, OutputStream output)
   throws StorageServerException;

   public abstract ContextKey createContext(Map<String, String> inProps,
                                            InputStream input,
                                            Map<String, String> outProps,
                                            OutputStream output,
                                            Context ctx)
   throws StorageServerException;

   public void flush(Context ctx)
      throws IOException, StorageServerException
   {
   }

   public String getName()
   {
      return m_name;
   }

   public int getOptimalRequiredSize()
   {
      return m_optimal_size;
   }

   public void reopenContext(Map<String, String> inProps,
                             InputStream input,
                             Map<String, String> outProps,
                             OutputStream output,
                             ContextKey ctxKey,
                             Context ctx)
      throws StorageServerException
   {
   }

   /**
    * Get the named Long property and register it as a transient object.
    * @param inProps Property map containing properties from URL
    * @param ctx Context to be used for this method call
    * @param label The name of the property to get
    * @param defaultValue A default value to be used if no property is found. May be null.
    * @return
    */
   protected Long getAndRegister(Map<String,String> inProps,
                                 Context ctx,
                                 String  label,
                                 Long    defaultValue )
   {
      Long value = null;
      String longAsString = inProps.get(label);
      if(longAsString!=null) {
          try {
              value = new Long(longAsString);
          } catch(NumberFormatException nfe) {
              value = defaultValue;
          }
      }
      else
         value = defaultValue;

      if( value != null )
         ctx.registerTransientObject(label,value);
      return value;
  }

   /**
    * Get the named String property and register it as a transient object.
    * @param inProps Property map containing properties from URL
    * @param ctx Context to be used for this method call
    * @param label The name of the property to get
    * @param defaultValue A default value to be used if no property is found. May be null.
    * @return
    */
   protected String getAndRegister(Map<String,String> inProps,
                                   Context ctx,
                                   String  label,
                                   String  defaultValue )
  {
      String value = inProps.get(label);
      if( value == null )
         value = defaultValue;
      
      if( value != null )
         ctx.registerTransientObject(label,value);
      return value;
  }

   /**
    * Get the named Boolean transient object from the context.
    * @param ctx
    * @param label
    * @return
    */
   protected Boolean getBoolean(Context ctx, String label, Boolean defaultValue )
   {
      Boolean value = (Boolean) ctx.getTransientObject(label);
      if( value == null )
         return defaultValue;
      return value;
   }
   
   protected Boolean getBoolean( Context ctx, String label )
      throws InvalidArgumentException
   {
      Boolean v = getBoolean( ctx, label, null );
      if( v == null )
         throw new InvalidArgumentException( "Missing argument " + label + " to method " );
      return v;
   }



   protected Long getLong(Context ctx, String label, Long defaultValue )
   {
      Long value = (Long) ctx.getTransientObject(label);
      if( value == null )
         return defaultValue;
      return value;
   }

   protected Long getLong( Context ctx, String label )
   throws InvalidArgumentException
   {
      Long l = getLong( ctx, label, null );
      if( l == null )
         throw new InvalidArgumentException( "Missing argument " + label + " to method " );
      return l;
   }

   /**
    * Get the named String transient object from the context.
    * Because Strings (and XUID base64) values are from the application space, they have
    * been encoded to be URL safe.
    * @param ctx
    * @param label
    * @return
    */
   protected String getString(Context ctx, String label, String defaultValue )
      throws XAMException
   {
      String value = (String) ctx.getTransientObject(label);
      if( value == null )
         return defaultValue;
      else {
         try { value = URLDecoder.decode( value, ProtocolConstants.UTF8 ); }
         catch( Exception e ) {
            throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR + " " + label );
         }
      }

      return value;
   }

   protected String getString( Context ctx, String label )
   throws XAMException
   {
      String v = getString( ctx, label, null );
      if( v == null )
         throw new InvalidArgumentException( "Missing argument " + label + " to method " );
      return v;
   }
   
   /**
    * Get the named Long transient object from the context.
    * @param ctx
    * @param label
    * @return
    */
   protected Long getHandle(Context ctx, String label) throws XAMException
   {
      Long value = (Long) ctx.getTransientObject(label);
      if( value == null )
         throw new InvalidArgumentException( "Invalid handle", XAMException.sXAM_INVALID_HANDLE );
      return value;
   }

   /** Utility to place a single string value in the response buffer */
   protected void respond( StringBuffer response, String label, String value )
   {
      if( response.length() > 0 )
         response.append( "\n" );
      response.append( label );
      response.append( "=" );
      response.append( value );
   }

   /** Utility to place a single char arry value in the response buffer */
   protected void respond( StringBuffer response, String label, char value[] )
   {
      if( response.length() > 0 )
         response.append( "\n" );
      response.append( label );
      response.append( "=" );
      response.append( value );
   }

   /** Utility to place a single long value in the response buffer */
   protected void respond( StringBuffer response, String label, long value )
   {
      if( response.length() > 0 )
         response.append( "\n" );
      response.append( label );
      response.append( "=" );
      response.append( Long.toString(value) );
   }
   
   /** Utility to place a single boolean value in the response buffer */
   protected void respond( StringBuffer response, String label, boolean value )
   {
      if( response.length() > 0 )
         response.append( "\n" );
      response.append( label );
      response.append( "=" );
      response.append( Boolean.toString(value) );
   }

   /**
    * Grab the Vim Server logger
    */
   final protected void setupLogger()
   {
      try { 
         logger = LogManager.GetLogger(Constants.LOGGER_NAME);
      } catch( Exception e ) {
         logger = Logger.getLogger(Constants.LOGGER_NAME);
      }
   }
   /**
    * Given the property map, log for debug purposes.
    * @param inProps
    */
   final protected void logProperties(Map inProps)
   {
      if (logger.isLoggable(Level.FINER))
      {
         StringBuffer propString = new StringBuffer();
         Set keys = inProps.keySet();
         Iterator i = keys.iterator();
         while (i.hasNext())
         {
            String key = (String) i.next();
            String value = (String) inProps.get(key);
            propString.append(key);
            propString.append("=");
            propString.append(value);
            propString.append("\n");
         }
         logger.info(propString.toString());
      }
   }
}
