/*
 *
 * Copyright (c) 2007-2009, Storage Networking Industry Association
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of the Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* Copyright (c) 2007, Sun Microsystems, Inc. */

package org.snia.xam.vim.isolation.server.http.handlers;

import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.util.Map;
import java.util.logging.Level;

import org.snia.xam.AsyncPendingException;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.vim.isolation.server.contextes.ConnectContextKey;
import org.snia.xam.vim.isolation.server.contextes.Context;
import org.snia.xam.vim.isolation.server.contextes.ContextKey;
import org.snia.xam.vim.isolation.server.ifc.StorageServerException;
import org.snia.xam.vim.isolation.server.resources.ByteBufferList;
import org.snia.xam.vim.isolation.server.resources.ByteBufferPool;
import org.snia.xam.vim.isolation.server.util.XAsyncCache;
import org.snia.xam.vim.isolation.server.util.XAsyncCacheEntry;

/**
 * Implements the handler for the HTTP wire "XAsyncGetData" operation.
 */
public class XAsyncGetData extends AbstractReadHandler {
    
    public static final int CHUNK_SIZE = 131072;
    protected ContextKey contextKey = null;
    
    public XAsyncGetData() throws StorageServerException {
        m_optimal_size = CHUNK_SIZE;
        m_name = ProtocolConstants.ASYNC_CALLBACK_GETDATA;
        setupLogger();
    }
    
    public ContextKey createContext(Map<String,String> inProps, InputStream input, Map<String,String> outProps, OutputStream output, Context ctx) {
        contextKey = new ConnectContextKey();
        logProperties( inProps );
        
        getAndRegister( inProps, ctx, ProtocolConstants.XASYNC_HANDLE, DEFAULT_HANDLE );
        
        return contextKey;
    }
    
    /**
     * Read will work in a chunk size and use the transport library's chunking
     * support to return chunk by chunk.  This may have to be called many, many
     * times based on the chunk size.
     */
    public long read(Context ctx, ByteBufferList bufferList) {
        if(exists( ctx, ProtocolConstants.COMPLETION_FLAG) ) return 0;
        logger.entering(this.getClass().getName(), "read");
        
        long status = XAMException.SUCCESS;
        String statusMsg = null;
        int bytesRead = -1;
        byte buffer[] = null;
        try { 
           Long handleAsLong = getHandle( ctx, ProtocolConstants.XASYNC_HANDLE);

           long handle = handleAsLong.longValue();

           XAsyncCache xac = XAsyncCache.getInstance(logger);
           XAsyncCacheEntry asyncCacheEntry = xac.getCacheEntry(handle);
           XAsync async = asyncCacheEntry.getXAsync();
           if( ! async.isComplete() )
              throw new AsyncPendingException( "Async is not yet complete" );
           
           if( logger.isLoggable( Level.INFO ))
              logger.info( "XASync[" + handle + "].GetData()" );

           buffer = asyncCacheEntry.getBuffer();           
           bytesRead = buffer.length;

        } catch (XAMException ex) {
           status = ex.getStatusCode();
           statusMsg = ex.getMessage();
        } catch(Exception e) {
           status = XAMException.sXAMException;
           statusMsg = ProtocolConstants.UNEXPECTED_EXCEPTION + e.getMessage();
           logger.log( Level.SEVERE, ProtocolConstants.UNEXPECTED_EXCEPTION, e );
        }
        
        // assemble the response
        StringBuffer responseBody = new StringBuffer();
        respond( responseBody, ProtocolConstants.STATUS_CODE, status );
        if( statusMsg != null )
           respond( responseBody, ProtocolConstants.STATUS_MSG, statusMsg );

        // get the pool
        ByteBufferPool pool = ByteBufferPool.getInstance();
        
        ctx.registerTransientObject(ProtocolConstants.COMPLETION_FLAG, Boolean.TRUE );
        ctx.registerTransientObject(ProtocolConstants.XSTREAM_NUMBYTES, new Integer(bytesRead));
        // if the read returned more than 0 bytes, write the bytes to the pool
        
        if(bytesRead>=0) {
            ByteBuffer bb = pool.checkOutBuffer(bytesRead);
            bb.clear();
            bb.put(buffer, 0, bytesRead);
            bb.flip();
            bufferList.appendBuffer(bb);
            pool.checkInBuffer(bb);
        }   
        logger.exiting(this.getClass().getName(), "read");
        return bytesRead;
    }

}
