/*
 *
 * Copyright (c) 2007-2009, Storage Networking Industry Association
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of the Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* Copyright (c) 2007, Sun Microsystems, Inc. */

package org.snia.xam.vim.isolation.server.http.handlers;

import java.io.InputStream;
import java.io.OutputStream;
import java.net.URLEncoder;
import java.util.Calendar;
import java.util.Map;
import java.util.logging.Level;
import org.snia.xam.XAMException;
import org.snia.xam.XSet;
import org.snia.xam.XUID;
import org.snia.xam.vim.isolation.server.contextes.ConnectContextKey;
import org.snia.xam.vim.isolation.server.contextes.Context;
import org.snia.xam.vim.isolation.server.contextes.ContextKey;
import org.snia.xam.vim.isolation.server.ifc.StorageServerException;
import org.snia.xam.vim.isolation.server.resources.ByteBufferList;
import org.snia.xam.vim.isolation.server.util.XSetCache;

/**
 * Implements the handler for the HTTP wire "XSetGetProperty" operation.
 */
public class XSetGetProperty extends AbstractReadHandler {

    protected ContextKey contextKey = null;
    
    public XSetGetProperty() throws StorageServerException {
       m_optimal_size = 65000;
       m_name = ProtocolConstants.XSET_GETPROPERTY_HANDLER_NAME;
       setupLogger();
    }

    public ContextKey createContext(Map<String,String> inProps, InputStream input, Map<String,String> outProps, OutputStream output, Context ctx) {
        contextKey = new ConnectContextKey();
        logProperties( inProps );

        getAndRegister( inProps, ctx, ProtocolConstants.XSET_HANDLE, DEFAULT_HANDLE );
        getAndRegister( inProps, ctx, ProtocolConstants.ATTRIBUTE_TYPE, NULL_STRING );
        getAndRegister( inProps, ctx, ProtocolConstants.ATTRIBUTE_NAME, NULL_STRING );

        return contextKey;
    }

    /** Perform the XSetGetProperty logic */
    public long read(Context ctx, ByteBufferList bufferList) {
        if(exists( ctx, ProtocolConstants.COMPLETION_FLAG) ) return 0;
        logger.entering(this.getClass().getName(), "read");
        
        long status      = XAMException.SUCCESS;
        String statusMsg = null;
        String value     = null;
        try { 
           Long xsystemHandleAsLong = getHandle( ctx, ProtocolConstants.XSET_HANDLE);
           String ptype = getString( ctx, ProtocolConstants.ATTRIBUTE_TYPE);
           String name = getString( ctx, ProtocolConstants.ATTRIBUTE_NAME);

           // retrieve the xset
           long xsetHandle = xsystemHandleAsLong.longValue();
           XSetCache xsc    = XSetCache.getInstance();
           XSet xs          = xsc.getXSet(xsetHandle);
        
           if( logger.isLoggable( Level.INFO ))
              logger.info( "XSet[" + xsetHandle + "].getProperty(" + name + ")" );
           value = getProperty(xs,ptype,name);
        } catch (XAMException ex) {
           status = ex.getStatusCode();
           statusMsg = ex.getMessage();
        } catch(Exception e) {
           status = XAMException.sXAMException;
           statusMsg = ProtocolConstants.UNEXPECTED_EXCEPTION + e.getMessage();
           logger.log( Level.SEVERE, ProtocolConstants.UNEXPECTED_EXCEPTION, e );
        }

        // assemble the response
        StringBuffer responseBody = new StringBuffer();
        respond( responseBody, ProtocolConstants.STATUS_CODE, status );
        if( statusMsg != null ) 
           respond( responseBody, ProtocolConstants.STATUS_MSG, statusMsg );
        if( status == XAMException.SUCCESS) 
           respond( responseBody, ProtocolConstants.ATTRIBUTE_VALUE, value );

        logger.exiting(this.getClass().getName(), "read");
        return packResponse( ctx, bufferList, responseBody );
    }

    /**
     * Calls the appropriate property creation method on the target XSet
     * @param xs xset that the property will be created on
     * @param ptype the property type that is being created
     * @param name the name of the new property
     * @return String the status of the operation on the XSet
     */
    protected String getProperty(XSet xs, String ptype, String name) throws XAMException {
        String value = null;
        if(ProtocolConstants.PTYPE_BOOLEAN.equals(ptype)) {
            boolean valueAsType = xs.getBoolean(name);
            value = new Boolean(valueAsType).toString();
        } else if(ProtocolConstants.PTYPE_INTEGER.equals(ptype)) {
            long valueAsType = xs.getLong(name);
            value = new Long(valueAsType).toString();
        } else if(ProtocolConstants.PTYPE_DOUBLE.equals(ptype)) {
            double valueAsType = xs.getDouble(name);
            value = new Double(valueAsType).toString();
        } else if(ProtocolConstants.PTYPE_XUID.equals(ptype)) {
            XUID valueAsType = xs.getXUID(name);
            value = valueAsType.toString();
            try { value = URLEncoder.encode( value, ProtocolConstants.UTF8 ); }
            catch( Exception e ) {
               throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR + " " + name );
               }
        } else if(ProtocolConstants.PTYPE_STRING.equals(ptype)) {
            value = xs.getString(name);
            try { value = URLEncoder.encode( value, ProtocolConstants.UTF8 ); }
            catch( Exception e ) {
               throw new XAMException( ProtocolConstants.TRANSPORT_ENCODE_ERROR + " " + name );
               }
        } else if(ProtocolConstants.PTYPE_DATETIME.equals(ptype)) {
            Calendar valueAsType = xs.getDateTime(name);
            long valueAsLong = valueAsType.getTimeInMillis();
            value = new Long(valueAsLong).toString();
        }

        return value;
    }
}
