/*
 *
 * Copyright (c) 2009, Storage Networking Industry Association
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of the Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
package org.snia.xam.vim.isolation.server.http.handlers;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.util.Map;
import java.util.logging.Level;

import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.XStream;
import org.snia.xam.vim.isolation.server.contextes.ConnectContextKey;
import org.snia.xam.vim.isolation.server.contextes.Context;
import org.snia.xam.vim.isolation.server.contextes.ContextKey;
import org.snia.xam.vim.isolation.server.ifc.StorageServerException;
import org.snia.xam.vim.isolation.server.resources.ByteBufferList;
import org.snia.xam.vim.isolation.server.resources.ByteBufferPool;
import org.snia.xam.vim.isolation.server.util.XAsyncCache;
import org.snia.xam.vim.isolation.server.util.XStreamCache;

/**
 * Implements the handler for the HTTP wire "XStreamAsyncWrite" operation.
 */
public class XStreamAsyncWrite extends AbstractWriteHandler {
    
    public static final int CHUNK_SIZE = 4096;
    protected ContextKey contextKey = null;
    
    public XStreamAsyncWrite() throws StorageServerException {
       setupLogger();
       m_name = ProtocolConstants.XSTREAM_ASYNC_WRITE_HANDLER;
       m_optimal_size = CHUNK_SIZE;
    }
    
    public ContextKey createContext(Map inProps, InputStream input, Map outProps, OutputStream output, Context ctx) 
       throws StorageServerException {
        contextKey = new ConnectContextKey();
        logProperties( inProps );
        getAndRegister( inProps, ctx, ProtocolConstants.XSTREAM_HANDLE, DEFAULT_HANDLE );
        getAndRegister( inProps, ctx, ProtocolConstants.ASYNC_XOPID, (Long) null );
        return contextKey;
    }
    
    public void closeContext(Context ctx, OutputStream output) throws StorageServerException {
        try {
           StringBuffer response = new StringBuffer();  
           Long status =  getLong( ctx, ProtocolConstants.STATUS_CODE, null );
           if( status != null ) 
              respond( response, ProtocolConstants.STATUS_CODE, status );
           
         Long asyncHandle = getLong( ctx, ProtocolConstants.XASYNC_HANDLE, null );
         if( asyncHandle != null ) 
            respond( response, ProtocolConstants.XASYNC_HANDLE, asyncHandle );
         
         String statusMsg = getString( ctx, ProtocolConstants.STATUS_MSG, null );
         if( statusMsg != null ) 
            respond( response, ProtocolConstants.STATUS_MSG, statusMsg );
          
         output.write(response.toString().getBytes());
         output.flush();
        } catch (Exception e) {
           logger.log( Level.SEVERE, "Fatal error completing XStream.asyncWrite method", e );
            e.printStackTrace();
        }
        ctx.releaseResources();
        
        ctx.dispose();
    }
    
    public void flush(Context ctx) throws IOException, StorageServerException {
    }
    
    /** Perform the XStreamAsyncWrite logic */
    public void write(Context ctx, ByteBufferList bufferList) throws IOException, StorageServerException {
        logger.entering(this.getClass().getName(), "write");
        
        long   status      = XAMException.SUCCESS;
        String statusMsg   = null;
        long   asyncHandle = -1;
        XAsync async       = null;
        try {
           Long handleAsLong = getHandle( ctx, ProtocolConstants.XSTREAM_HANDLE);
           Long xopid        = getLong( ctx, ProtocolConstants.ASYNC_XOPID );
           long handle = handleAsLong.longValue();

           // retrieve a handle to the proper XStream
           XStreamCache xsc  = XStreamCache.getInstance();
           XStream xs        = xsc.getXStream(handle);
        
           // retrieve the buffers from the list that have to be written
           ByteBuffer[] bbs = bufferList.getBuffers();
           // get the pool
           ByteBufferPool pool = ByteBufferPool.getInstance();
           
           // go through each of the byte buffers and write them
           for(int i=0; i<bbs.length ; i++) {
              int remaining = bbs[i].remaining();
              byte[] bytes = new byte[remaining];
              for(int j=0;j<remaining ;j++) {
                 bytes[j] = bbs[i].get();
              }
              if( logger.isLoggable( Level.INFO ))
                 logger.info( "XStream[" + handle + "].asyncWrite(buffer[]," + bytes.length +"," + xopid + ")" );
               async = xs.asyncWrite(bytes, bytes.length, xopid, null );
               XAsyncCache asc = XAsyncCache.getInstance(logger);
               asyncHandle = asc.addXAsync(xsc.getXSystemHandle(handle), async);
               break;
           }
           pool.checkInBufferList(bufferList);
        } catch (XAMException ex) {
           status = ex.getStatusCode();
           statusMsg = ex.getMessage();
        } catch( Exception e ) {
           status = XAMException.sXAM_INVALID_HANDLE;
           statusMsg = ProtocolConstants.UNEXPECTED_EXCEPTION;
           logger.log( Level.SEVERE, ProtocolConstants.UNEXPECTED_EXCEPTION, e );
        }
        
        if( statusMsg != null )
           ctx.registerTransientObject(ProtocolConstants.STATUS_MSG, statusMsg );
        ctx.registerTransientObject(ProtocolConstants.XASYNC_HANDLE, new Long(asyncHandle));
        ctx.registerTransientObject(ProtocolConstants.STATUS_CODE,new Long(status));
        ctx.registerTransientObject(ProtocolConstants.COMPLETION_FLAG, new Boolean("true"));

        logger.exiting(this.getClass().getName(), "write");
        return;
    }
}
