/*
 * Copyright (c) 2009, Storage Networking Industry Association
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
package org.snia.xam.vim.isolation.server.http.handlers;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.Map;
import java.util.logging.Level;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.XSystem;
import org.snia.xam.base.DefaultXUID;
import org.snia.xam.vim.isolation.server.contextes.ConnectContextKey;
import org.snia.xam.vim.isolation.server.contextes.Context;
import org.snia.xam.vim.isolation.server.contextes.ContextKey;
import org.snia.xam.vim.isolation.server.ifc.StorageServerException;
import org.snia.xam.vim.isolation.server.resources.ByteBufferList;
import org.snia.xam.vim.isolation.server.util.XAsyncCache;
import org.snia.xam.vim.isolation.server.util.XSystemCache;

/**
 * Implements the handler for the HTTP wire "XSystemAsyncOpenXSet" operation.
 */
public class XSystemAsyncOpenXSet extends AbstractReadHandler {

    protected ContextKey contextKey = null;
    
    /** Creates a new instance of XSystem.createXSet */
    public XSystemAsyncOpenXSet() throws StorageServerException {
       m_name = ProtocolConstants.XSYSTEM_ASYNC_OPENXSET_HANDLER;
       setupLogger();
    }
    public ContextKey createContext(Map<String,String> inProps, InputStream input, Map<String,String> outProps, OutputStream output, Context ctx) {
        contextKey = new ConnectContextKey();
        logProperties(inProps);

        getAndRegister( inProps, ctx, ProtocolConstants.XSET_HANDLE, DEFAULT_HANDLE );
        getAndRegister( inProps, ctx, ProtocolConstants.MODE_PROPERTY, NULL_STRING );
        getAndRegister( inProps, ctx, ProtocolConstants.XUID_PROPERTY, NULL_STRING );
        getAndRegister( inProps, ctx, ProtocolConstants.ASYNC_XOPID, NULL_LONG );

        return contextKey;
    }

    /** Perform the XSystemAsyncOpenXSet logic */
    public long read(Context ctx, ByteBufferList bufferList) {
        if (exists( ctx, ProtocolConstants.COMPLETION_FLAG) ) return 0; 
        logger.entering(this.getClass().getName(), "read");
        
        long   status       = XAMException.SUCCESS;
        String statusMsg    = null;
        long   xasyncHandle = -1;
        long   xopid        = -1;
        try {
           Long xsystemHandleAsLong = getHandle( ctx, ProtocolConstants.XSYSTEM_HANDLE);
           String xuid = getString( ctx, ProtocolConstants.XUID_PROPERTY);
           String mode = getString( ctx, ProtocolConstants.MODE_PROPERTY);
           Long   xopidLong = getLong( ctx, ProtocolConstants.ASYNC_XOPID );
           xopid = xopidLong.longValue();
           
           // retrieve the xsystem
           long xsystemHandle = xsystemHandleAsLong.longValue();
           XSystemCache xsc = XSystemCache.getInstance();
           XSystem xs = xsc.getXSystem(xsystemHandle);

           // open the xset
           if( logger.isLoggable( Level.INFO ))
              logger.info( "XSystem[" + xsystemHandle + "].asyncOpenXSet(" + xuid + "," + mode + "," + xopid + ")" );
           XAsync async = xs.asyncOpenXSet(DefaultXUID.valueOf(xuid), mode, xopid, null );
           
           // an exception would occur if the xset was not found
           //  continue on and add the xset to the cache
           XAsyncCache asyncCache = XAsyncCache.getInstance(logger);
           xasyncHandle = asyncCache.addXAsync(xsystemHandle, async);
        } catch (XAMException ex) {
           status = ex.getStatusCode();
           statusMsg = ex.getMessage();
        } catch(Exception e) {
           status = XAMException.sXAMException;
           statusMsg = ProtocolConstants.UNEXPECTED_EXCEPTION + e.getMessage();
           logger.log( Level.SEVERE, ProtocolConstants.UNEXPECTED_EXCEPTION, e );
        }

        // assemble the response
        // create a string buffer to process body information, there will be many appends
        StringBuffer responseBody = new StringBuffer();
        respond( responseBody, ProtocolConstants.STATUS_CODE, status );
        if( statusMsg != null )
           respond( responseBody, ProtocolConstants.STATUS_MSG, statusMsg );
        
        if ( status == XAMException.SUCCESS )
           respond( responseBody, ProtocolConstants.XASYNC_HANDLE, xasyncHandle );

        logger.exiting(this.getClass().getName(), "read");
        return packResponse( ctx, bufferList, responseBody );
    }
}
