/*
 *
 * Copyright (c) 2007-2009, Storage Networking Industry Association
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of the Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* Copyright (c) 2007, Sun Microsystems, Inc. */

package org.snia.xam.vim.isolation.server.http.handlers;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.util.Map;
import java.util.logging.Level;

import org.snia.xam.XAMException;
import org.snia.xam.XSystem;
import org.snia.xam.vim.isolation.server.contextes.ConnectContextKey;
import org.snia.xam.vim.isolation.server.contextes.Context;
import org.snia.xam.vim.isolation.server.contextes.ContextKey;
import org.snia.xam.vim.isolation.server.handlers.WriteHandler;
import org.snia.xam.vim.isolation.server.ifc.StorageServerException;
import org.snia.xam.vim.isolation.server.resources.ByteBufferList;
import org.snia.xam.vim.isolation.server.resources.ByteBufferPool;
import org.snia.xam.vim.isolation.server.util.AuthenticateByteArrayXStream;
import org.snia.xam.vim.isolation.server.util.XStreamCache;
import org.snia.xam.vim.isolation.server.util.XSystemCache;

/**
 * Implements the handler for the HTTP wire "XSystemAuthenticate" operation.
 */
public class XSystemAuthenticate extends AbstractWriteHandler implements WriteHandler {
    
    public static final int CHUNK_SIZE = 4096;
    protected ContextKey contextKey = null;
    
    /** Creates a new instance of Connect */
    public XSystemAuthenticate() throws StorageServerException {
       setupLogger();
       m_optimal_size = CHUNK_SIZE;
       m_name = ProtocolConstants.XSYSTEM_AUTHENTICATE_HANDLER;
    }
    
    public ContextKey createContext(Map inProps, InputStream input, Map outProps, OutputStream output, Context ctx) throws StorageServerException {
        contextKey = new ConnectContextKey();
        logProperties( inProps );

        getAndRegister( inProps, ctx, ProtocolConstants.XSYSTEM_HANDLE, DEFAULT_HANDLE );
        return contextKey;
    }
    
    public void closeContext(Context ctx, OutputStream output) throws StorageServerException {
        try {
           StringBuffer response = new StringBuffer();  
           Long status =  getLong( ctx, ProtocolConstants.STATUS_CODE);
           if( status == null ) {
              respond( response, ProtocolConstants.STATUS_CODE, XAMException.sAuthenticationException );
              respond( response, ProtocolConstants.STATUS_MSG, 
                        "Write method for XSystem.authenticate not called, faulty HTTP POST?" );
            
           } else 
              respond( response, ProtocolConstants.STATUS_CODE, status );
           
         String statusMsg = getString( ctx, ProtocolConstants.STATUS_MSG, null );
         if( statusMsg != null ) 
            respond( response, ProtocolConstants.STATUS_MSG, statusMsg );
         
         Long responseHandle = getLong( ctx, ProtocolConstants.XSTREAM_HANDLE, null);
         if( responseHandle != null ) 
            respond( response, ProtocolConstants.XSTREAM_HANDLE, responseHandle );
          
         Long authStatus = getLong( ctx, ProtocolConstants.AUTHENTICATE_STATUS, null );
         if( authStatus != null )
            respond( response, ProtocolConstants.AUTHENTICATE_STATUS, authStatus );
          
         output.write(response.toString().getBytes());
         output.flush();
        } catch (Exception e) {
           logger.log( Level.SEVERE, "XSystem.authentication completion fauilure", e );
            e.printStackTrace();
        }
        
        ctx.releaseResources();
        ctx.dispose();
    }
    
    public void write(Context ctx, ByteBufferList bufferList) throws IOException, StorageServerException {
        logger.entering(this.getClass().getName(), "write");

        long status           = XAMException.SUCCESS;
        String statusMsg      = null;
        long responseHandle   = -1;
        long authenticateCode = -1;

        try {        
           Long handleAsLong = getHandle( ctx, ProtocolConstants.XSYSTEM_HANDLE);
           long handle = handleAsLong.longValue();
        
           // first retrieve the xsystem
           XSystemCache xsc = XSystemCache.getInstance();
           XSystem xs = xsc.getXSystem(handle);

           // retrieve the buffers from the list that have to be written
           ByteBuffer[] bbs = bufferList.getBuffers();
           // get the pool
           ByteBufferPool pool = ByteBufferPool.getInstance();
           
           // go through each of the byte buffers and write them
           for(int i=0; i<bbs.length ; i++) {
              int remaining = bbs[i].remaining();
              byte[] bytes = new byte[remaining];
              for(int j=0;j<remaining ;j++) {
                 bytes[j] = bbs[i].get();
              }
              
              ByteArrayOutputStream baos = new ByteArrayOutputStream();
              if( logger.isLoggable( Level.INFO ))
                 logger.info( "XSystem[" + handle + "].authenticate()" );
              authenticateCode = xs.authenticate(bytes, baos);
              AuthenticateByteArrayXStream responseXStream = new AuthenticateByteArrayXStream(baos.toByteArray());
              
              XStreamCache streamCache = XStreamCache.getInstance();
              responseHandle = streamCache.addXStream( handle, responseXStream );
           }
           pool.checkInBufferList(bufferList);
        } catch (XAMException ex) {
           status = ex.getStatusCode();
           statusMsg = ex.getMessage();
        } catch( Exception e ) {
           status = XAMException.sXAM_INVALID_HANDLE;
           statusMsg = ProtocolConstants.UNEXPECTED_EXCEPTION;
           logger.log( Level.SEVERE, ProtocolConstants.UNEXPECTED_EXCEPTION, e );
        }
        
        if( statusMsg != null )
           ctx.registerTransientObject(ProtocolConstants.STATUS_MSG, statusMsg );
        
        ctx.registerTransientObject(ProtocolConstants.STATUS_CODE,new Long(status) );
        
        if( status == XAMException.SUCCESS ) {
           ctx.registerTransientObject(ProtocolConstants.XSTREAM_HANDLE, responseHandle );
           ctx.registerTransientObject(ProtocolConstants.AUTHENTICATE_STATUS, Long.valueOf(authenticateCode) );
        }
        
        ctx.registerTransientObject(ProtocolConstants.COMPLETION_FLAG, Boolean.TRUE );
        logger.exiting(this.getClass().getName(), "write");
        return;
    }
}
