/*
 *
 * Copyright (c) 2007-2009, Storage Networking Industry Association
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of the Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* Copyright (c) 2007, Sun Microsystems, Inc. */

package org.snia.xam.vim.isolation.server.http.handlers;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.nio.ByteBuffer;
import java.util.Calendar;
import java.util.Enumeration;
import java.util.GregorianCalendar;
import java.util.Hashtable;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.logging.Level;

import org.snia.xam.ConnectException;
import org.snia.xam.FieldExistsException;
import org.snia.xam.VIMLoadException;
import org.snia.xam.XAMException;
import org.snia.xam.XSystem;
import org.snia.xam.base.DefaultXUID;
import org.snia.xam.vim.VIM;
import org.snia.xam.vim.isolation.server.contextes.ConnectContextKey;
import org.snia.xam.vim.isolation.server.contextes.Context;
import org.snia.xam.vim.isolation.server.contextes.ContextKey;
import org.snia.xam.vim.isolation.server.ifc.StorageServerException;
import org.snia.xam.vim.isolation.server.resources.ByteBufferList;
import org.snia.xam.vim.isolation.server.resources.ByteBufferPool;
import org.snia.xam.vim.isolation.server.util.VIMCache;
import org.snia.xam.vim.isolation.server.util.XRIParser;
import org.snia.xam.vim.isolation.server.util.XSystemCache;

/**
 * Implements the handler for the HTTP wire "XSystemAuthenticate" operation.
 */
public class XSystemConnect extends AbstractWriteHandler {
    
    public static final int CHUNK_SIZE = 4096;
    protected ContextKey contextKey = null;
    
    /** Creates a new instance of Connect */
    public XSystemConnect() throws StorageServerException {
       setupLogger();
       m_name = ProtocolConstants.XSYSTEM_CONNECT_HANDLER_NAME;
       m_optimal_size = CHUNK_SIZE;
    }
    
    public ContextKey createContext(Map inProps, InputStream input, Map outProps, OutputStream output, Context ctx) 
       throws StorageServerException {
        contextKey = new ConnectContextKey();
        logProperties( inProps );
        getAndRegister( inProps, ctx, ProtocolConstants.CONNECT_XRI, (String) null );
        return contextKey;
    }
    
    public void closeContext(Context ctx, OutputStream output) throws StorageServerException {
        try {
           StringBuffer response = new StringBuffer();  
           Long status =  getLong( ctx, ProtocolConstants.STATUS_CODE, null );
           if( status == null ) {
              respond( response, ProtocolConstants.STATUS_CODE, XAMException.sConnectException );
              respond( response, ProtocolConstants.STATUS_MSG, 
                 "Write method for XSystem.connect not called, faulty HTTP POST?" );
           } else 
              respond( response, ProtocolConstants.STATUS_CODE, status );
           
         String statusMsg = getString( ctx, ProtocolConstants.STATUS_MSG, null);
         if( statusMsg != null ) 
            respond( response, ProtocolConstants.STATUS_MSG, statusMsg );
         
         Long responseHandle = getLong( ctx, ProtocolConstants.XSYSTEM_HANDLE, null );
         if( responseHandle != null ) 
            respond( response, ProtocolConstants.XSYSTEM_HANDLE, responseHandle );

         output.write(response.toString().getBytes());
         output.flush();
        } catch (Exception e) {
           logger.log( Level.SEVERE, "Fatal error completing the XSystem.connect method", e );
            e.printStackTrace();
        }
        ctx.releaseResources();
        
        ctx.dispose();
    }
    
    public void write(Context ctx, ByteBufferList bufferList) throws IOException, StorageServerException {
        logger.entering(this.getClass().getName(), "write");
        
        long status       = XAMException.SUCCESS;
        String statusMsg  = null;
        long handle       = -1;
        try {        
           String xri   = getString( ctx, ProtocolConstants.CONNECT_XRI);
           xri = URLDecoder.decode( xri, ProtocolConstants.UTF8 );
           
           VIMCache xlc = VIMCache.getInstance();
           logger.log(Level.INFO, "xri="+xri);
        
           VIM theVim = null;

           xri = xlc.getNextXri(xri);
           XRIParser xp = new XRIParser(xri);
           String vimName = xp.getVimName();
           
           logger.info( "Loading VIM for name '" + vimName + "'" );
           theVim = xlc.getVIM(vimName);
 
           XSystem xs = null;

           if (theVim == null)
              throw new VIMLoadException("Can not find a VIM for " + vimName);

           logger.log(Level.INFO, "creating XSystem="+xri);
           xs = theVim.createXSystem();
               
           if( logger.isLoggable( Level.INFO ))
              logger.info( "XSystem[" + handle + "].connect()" );
               
           // now connect
           xs.connect(xri);
           XSystemCache xsc = XSystemCache.getInstance();

           handle = xsc.addXSystem(xri, xs);

           // Process the incoming buffer and write new fields to the XSystem
           ByteBuffer[] bbs = bufferList.getBuffers();
           ByteBufferPool pool = ByteBufferPool.getInstance();
           
           // go through each of the byte buffers and write them
           ByteArrayOutputStream baos = new ByteArrayOutputStream();
           for(int i=0; i<bbs.length ; i++) {
              int remaining = bbs[i].remaining();
              for( int j = 0; j < remaining; j++ )
                 baos.write(bbs[i].get());
           }
           
           // Process values
           pool.checkInBufferList(bufferList);
           processConnectBuffer( xs, baos.toByteArray() );
           xs.deleteField(XSystem.XAM_XSYSTEM_INITIALIZING);
        } catch (XAMException ex) {
           status = ex.getStatusCode();
           statusMsg = ex.getMessage();
        } catch( Exception e ) {
           status = XAMException.sXAM_INVALID_HANDLE;
           statusMsg = ProtocolConstants.UNEXPECTED_EXCEPTION;
           logger.log( Level.SEVERE, ProtocolConstants.UNEXPECTED_EXCEPTION, e );
        }
        
        if( statusMsg != null )
           ctx.registerTransientObject(ProtocolConstants.STATUS_MSG, statusMsg );
        ctx.registerTransientObject(ProtocolConstants.STATUS_CODE, new Long(status));
        if( status == XAMException.SUCCESS ) {
           ctx.registerTransientObject(ProtocolConstants.XSYSTEM_HANDLE, handle );
        }
        ctx.registerTransientObject(ProtocolConstants.COMPLETION_FLAG, Boolean.TRUE );
        logger.exiting(this.getClass().getName(), "write");
        return;
    }
    
    class InitializationRecord {
       String  name     = null;
       String  type     = null;
       String  value    = null;
       boolean binding  = false;
    }
    
    /** 
     * Process the connect buffer, setting the properties appropriately.
     * @param xs
     * @param byteArray
     */
    private void processConnectBuffer(XSystem xs, byte[] byteArray)
       throws XAMException
    {
       
       Hashtable<String,InitializationRecord> props = new Hashtable<String,InitializationRecord>();
       String connectProps = new String(byteArray);
       StringTokenizer nameValuePair = new StringTokenizer( connectProps, " " );
       while( nameValuePair.hasMoreElements() )
       {
          String pair = nameValuePair.nextToken();
          parsePair( props, pair );
       }
       Enumeration<String> keys = props.keys();
       while( keys.hasMoreElements() )
       {
          InitializationRecord ir = props.get(keys.nextElement());
          if( ir.type.equals(ProtocolConstants.XAM_BOOLEAN))
             try { xs.createProperty( ir.name, ir.binding, Boolean.parseBoolean(ir.value)); }
             catch( FieldExistsException fee ) { xs.setProperty( ir.name, Boolean.parseBoolean(ir.value)); }
             
          else if( ir.type.equals(ProtocolConstants.XAM_DATETIME )) {
             Calendar value = GregorianCalendar.getInstance();
             value.setTimeInMillis(new Long(ir.value).longValue());
             try { xs.createProperty( ir.name, ir.binding, value); }
             catch( FieldExistsException fee ) { xs.setProperty( ir.name, value ); }
          }
          
          else if( ir.type.equals(ProtocolConstants.XAM_DOUBLE))
             try { xs.createProperty( ir.name, ir.binding, Double.parseDouble(ir.value)); }
             catch( FieldExistsException fee ) { xs.setProperty( ir.name, Double.parseDouble(ir.value)); }
             
          else if( ir.type.equals(ProtocolConstants.XAM_INTEGER))
             try { xs.createProperty( ir.name, ir.binding, Long.parseLong(ir.value)); }
             catch( FieldExistsException fee ) { xs.setProperty( ir.name, Long.parseLong(ir.value)); }
          
          else if( ir.type.equals(ProtocolConstants.XAM_STRING ))
             try { xs.createProperty( ir.name, ir.binding, ir.value ); }
             catch( FieldExistsException fee ) { xs.setProperty( ir.name, ir.value ); }
             
          else if( ir.type.equals(ProtocolConstants.XAM_XUID))
             try { xs.createProperty( ir.name, ir.binding, DefaultXUID.valueOf(ir.value) ); }
             catch( FieldExistsException fee ) { xs.setProperty( ir.name, DefaultXUID.valueOf(ir.value) ); }
             
          else 
             throw new ConnectException( "Unknown property type in initialization protocol: " +
                ir.type + " for property " + ir.name );
       }
    }

    /**
     * Given a property initialization pair, such as ".xam.identity.value=foo", parse it and add it to 
     * the property table, for later storage in the XSystem.
     * @param props The hashtable containing the InitializationRecords
     * @param pair The pair to be parsed.
     */
   private void parsePair(Hashtable<String, InitializationRecord> props,
                          String pair)
      throws XAMException
   {
      final String VALUE    = ".value";
      final String BINDING  = ".binding";
      final String TYPE     = ".type";
      
      int eqIdx = pair.indexOf('=');
      if( eqIdx < 0 )
         return;
      String name  = pair.substring(0, eqIdx );
      String value = pair.substring( eqIdx + 1 );
      try 
      {
         name = URLDecoder.decode( name, ProtocolConstants.UTF8 );
         value = URLDecoder.decode( value, ProtocolConstants.UTF8 );
      }
      catch( UnsupportedEncodingException uee )
      {
         logger.log( Level.SEVERE, "Unsupported encoding in XSystem.connect protocol", uee );
         return;
      }
      int lastDot = name.lastIndexOf('.');
      String propertyName = name.substring(0,lastDot);
      InitializationRecord ir = props.get(propertyName);
      if( ir == null ) {
         ir = new InitializationRecord();
         ir.name = propertyName;
         props.put(propertyName, ir );
      }
      if( name.endsWith(VALUE))
         ir.value = value;
      else if( name.endsWith(BINDING))
         ir.binding = Boolean.parseBoolean(value);
      else if( name.endsWith(TYPE))
         ir.type = value;
      else
         throw new ConnectException( "Unexpected initialization property in connect protocol: " + name );
   }

}
