/*
 *
 * Copyright (c) 2007-2009, Storage Networking Industry Association
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of the Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* Copyright (c) 2007, Sun Microsystems, Inc. */

package org.snia.xam.vim.isolation.server.http.handlers;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Map;
import java.util.logging.Level;
import org.snia.xam.XAMException;
import org.snia.xam.XSystem;
import org.snia.xam.XUID;
import org.snia.xam.base.DefaultXUID;
import org.snia.xam.vim.isolation.server.contextes.ConnectContextKey;
import org.snia.xam.vim.isolation.server.contextes.Context;
import org.snia.xam.vim.isolation.server.contextes.ContextKey;
import org.snia.xam.vim.isolation.server.ifc.StorageServerException;
import org.snia.xam.vim.isolation.server.resources.ByteBufferList;
import org.snia.xam.vim.isolation.server.util.XSystemCache;

/**
 * Implements the handler for the HTTP wire "XSystemSetProperty" operation.
 */
public class XSystemSetProperty extends AbstractReadHandler {

    protected ContextKey contextKey = null;
    
    public XSystemSetProperty() throws StorageServerException {
       m_name = ProtocolConstants.XSYSTEM_SETPROPERTY_HANDLER_NAME;
       setupLogger();
    }

    public ContextKey createContext(Map<String,String> inProps, InputStream input, Map<String,String> outProps, OutputStream output, Context ctx) {
        contextKey = new ConnectContextKey();
        logProperties(inProps);

        getAndRegister( inProps, ctx, ProtocolConstants.XSYSTEM_HANDLE, DEFAULT_HANDLE );
        getAndRegister( inProps, ctx, ProtocolConstants.ATTRIBUTE_TYPE, NULL_STRING );
        getAndRegister( inProps, ctx, ProtocolConstants.ATTRIBUTE_NAME, NULL_STRING );
        getAndRegister( inProps, ctx, ProtocolConstants.ATTRIBUTE_VALUE, NULL_STRING );
        
        return contextKey;
    }

    /**
     * Perform the XSystemSetProperty logic.
     */
    public long read(Context ctx, ByteBufferList bufferList)  {
        if(exists( ctx, ProtocolConstants.COMPLETION_FLAG) ) return 0;
        logger.entering(this.getClass().getName(), "read");
        
        long   status    = XAMException.SUCCESS;
        String statusMsg = null;
        try {
           Long xsystemHandleAsLong = getHandle( ctx, ProtocolConstants.XSYSTEM_HANDLE);
           String ptype = getString( ctx, ProtocolConstants.ATTRIBUTE_TYPE);
           String name  = getString( ctx, ProtocolConstants.ATTRIBUTE_NAME);
           String value = getString( ctx, ProtocolConstants.ATTRIBUTE_VALUE);

           // retrieve the xsystem
           long xsystemHandle = xsystemHandleAsLong.longValue();
           XSystemCache xsc = XSystemCache.getInstance();
           XSystem xs = xsc.getXSystem(xsystemHandle);

           if( logger.isLoggable( Level.INFO ))
              logger.info( "XSystem[" + xsystemHandle + "].setProperty(" + name + ", '" + value + "')" );
           setProperty(xs,ptype,name,value);
        } catch (XAMException ex) {
           status = ex.getStatusCode();
           statusMsg = ex.getMessage();
        } catch(Exception e) {
           status = XAMException.sXAMException;
           statusMsg = ProtocolConstants.UNEXPECTED_EXCEPTION + e.getMessage();
           logger.log( Level.SEVERE, ProtocolConstants.UNEXPECTED_EXCEPTION, e );
        }

        // assemble the response
        // create a string buffer to process body information, there will be many appends
        StringBuffer responseBody = new StringBuffer();
        respond( responseBody, ProtocolConstants.STATUS_CODE, status );
        if( statusMsg != null )
           respond( responseBody, ProtocolConstants.STATUS_MSG, statusMsg );

        logger.exiting(this.getClass().getName(), "read");
        return packResponse( ctx, bufferList, responseBody );
    }

    /**
     * Calls the appropriate property set method on the target XSet
     * @param xs xsystem that the property will be created on
     * @param ptype the property type that is being created
     * @param name the name of the new property
     * @param valueAsString the value the property should be set to
     * @return int the status of the operation on the XSet
     */
    protected void setProperty( XSystem xs, 
                                String  ptype, 
                                String  name,
                                String  valueAsString) 
       throws Exception
    {
       if(ProtocolConstants.PTYPE_BOOLEAN.equals(ptype)) {
          boolean value = new Boolean(valueAsString).booleanValue();
          xs.setProperty(name, value);
       } else if(ProtocolConstants.PTYPE_INTEGER.equals(ptype)) {
          long value = new Long(valueAsString).longValue();
          xs.setProperty(name, value);
       } else if(ProtocolConstants.PTYPE_DOUBLE.equals(ptype)) {
          double value = new Double(valueAsString).doubleValue();
          xs.setProperty(name, value);
       } else if(ProtocolConstants.PTYPE_XUID.equals(ptype)) {
          XUID value = DefaultXUID.valueOf(valueAsString);
          xs.setProperty(name, value);
       } else if(ProtocolConstants.PTYPE_STRING.equals(ptype)) {
          xs.setProperty(name, valueAsString);
       } else if(ProtocolConstants.PTYPE_DATETIME.equals(ptype)) {
          Calendar value = GregorianCalendar.getInstance();
          value.setTimeInMillis(new Long(valueAsString).longValue());
          xs.setProperty(name, value);
       }
    }
}
