/* Copyright (c) 2008-2009, The Storage Networking Industry Association.

 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */

package org.snia.xam.vim.isolation.server.util;

import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidOperationException;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XSetUnderHoldException;
import org.snia.xam.XStream;
import org.snia.xam.XStreamAbandonException;
import org.snia.xam.XStreamCorruptException;
import org.snia.xam.XStreamException;

/**
 * A class to wrap a byte[] in a read only {@link XStream} interface. This is
 * used to wrap the SASL authentication response in an XStream.
 */
public class AuthenticateByteArrayXStream
   implements XStream
{
   private static final String READ_ONLY_ERROR = "Authentication response XStream is READ ONLY";
   private static final String UNSUPPORTED = "HTTP VIM does not support asynchronous operations on authenticate XStream response";
   
   private byte[] m_bytes;
   private int    m_ptr;
   private boolean m_is_closed    = false;
   private boolean m_is_abandoned = false;

   /**
    * Constructs a new ByteArrayXStream with the specified data. This creates a
    * read only {@link XStream}.
    * 
    * @param bytes The XStream data
    */
   public AuthenticateByteArrayXStream(byte[] bytes)
   {
      m_bytes = bytes;
      m_ptr = 0;
   }

   public void abandon()
      throws XAMException
   {
      checkState();
      m_is_abandoned = true;
   }

   public void close()
      throws XAMException
   {
      checkState();
      m_ptr = -1;
      m_is_closed = true;
   }

   public XAsync asyncClose(long xopid, XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamCorruptException, XStreamException, XAMException
   {
      throw new InvalidOperationException( UNSUPPORTED );
   }

   public XAsync asyncRead(byte[] buffer,
                           long bufferLength,
                           long xopid,
                           XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XAMException
   {
      throw new InvalidOperationException( UNSUPPORTED );
   }

   public XAsync asyncWrite(byte[] buffer,
                            long bufferLength,
                            long xopid,
                            XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XAMException
   {
      checkState();
      throw new XStreamException( READ_ONLY_ERROR );
   }

   public long read(byte[] buffer, long offset, long bytesToRead)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XAMException
   {
      checkState();
      
      if( bytesToRead > Integer.MAX_VALUE )
         throw new InvalidArgumentException( "bytesToRead argument is too large, must be <= " + Integer.MAX_VALUE );
      
      if( offset > Integer.MAX_VALUE )
         throw new InvalidArgumentException( "offset argument is too large, must be <= " + Integer.MAX_VALUE );
      
      int endPoint = (int) (offset+bytesToRead);
      if( (endPoint > Integer.MAX_VALUE) || (endPoint > buffer.length) )
         throw new InvalidArgumentException( "offset+bytesToRead is too large, or exceded buffer length" );
      
      if (m_ptr == m_bytes.length) return -1;
      int total = Math.min(m_bytes.length - m_ptr, (int) bytesToRead);
      System.arraycopy(m_bytes, m_ptr, buffer, (int) offset, total);
      m_ptr += total;
      return total;
   }

   public long read(byte[] buffer, long bytesToRead)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XAMException
   {
      return read( buffer, 0, bytesToRead );
   }

   public long read(byte[] buffer)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XAMException
   {
      return read( buffer, 0, buffer.length );
   }

   public void seek(long offset, long whence)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XAMException
   {
      checkState();
      long ptr = m_ptr;
      switch ( (int) whence)
      {
         case XStream.SEEK_SET :
            ptr = offset;
            break;
         case XStream.SEEK_END :
            ptr = m_bytes.length + offset;
            break;
         case XStream.SEEK_CUR :
            ptr += offset;
            break;
      }
      ptr = Math.max(0, Math.min(ptr, (long) m_bytes.length));
      m_ptr = (int) ptr;
   }

   public long tell()
   throws XAMException
   {
      checkState();
      return m_ptr;
   }

   /**
    * XStream write method, not supported for the ByteArrayXStream class.
    */
   public long write(byte[] buffer)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XSetUnderHoldException, XAMException
   {
      checkState();
      throw new XStreamException( READ_ONLY_ERROR );
   }

   /**
    * XStream write method, not supported for the ByteArrayXStream class.
    */
   public long write(byte[] buffer, long bytesToWrite)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XSetUnderHoldException, XAMException
   {
      checkState();
      throw new XStreamException( READ_ONLY_ERROR );
   }

   /**
    * XStream write method, not supported for the ByteArrayXStream class.
    */
   public long write(byte[] buffer, long offset, long bytesToWrite)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XSetUnderHoldException, XAMException
   {
      checkState();
      throw new XStreamException( READ_ONLY_ERROR );
   }
   
   private void checkState() throws XAMException
   {
      if( m_is_closed )
         throw new XStreamException( "Authentication response XStream is closed" );
      
      if( m_is_abandoned )
         throw new XStreamAbandonException( "Authentication response XStream is abandoned" );
   }
}
