/*
 * Copyright (c) 2007-2009, Storage Networking Industry Association
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* Copyright (c) 2007, Sun Microsystems, Inc. */

package org.snia.xam.vim.isolation.server.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Properties;

/**
 *
 */
public class HandlerConfig {
    
   private Properties config;
   
    /**
     * Creates a new instance of HandlerConfig with
     * a startup properties file from the "current" directory.
     *
     * @see Constants#LOAD_HANDLER_FILE
     */
    public HandlerConfig() throws IOException {
        this(new File(".", Constants.LOAD_HANDLER_FILE));
    }
    
   /*
    * Create a new handler configuration from an identified file.
    * @param f a property file that contains the handler parameters for the server
    * @exception IOException thrown if there are any problems opening or reading
    * the property file (such as FileNotFound, etc...).
    */
    public HandlerConfig(File f) throws IOException {
        config = new Properties();
        FileInputStream fis = null;
        try {
            fis = new FileInputStream(f);
            config.load(fis);
            fis.close();
        } finally {
            if (fis != null) fis.close();
        }
    }
    
    /**
     * Create a new handler configuration from a set of properties
     * @param p is a set of properties to use for handlers, use keys as if they were
     * read directly from a properties file
     */
    public HandlerConfig(Properties p) {
        config = p;
    }
    
    public void setProperty(String name, String value) {
        if (value == null)
            config.remove(name);
        else
            config.setProperty(name, value);
    }
    
    /**
     * Returns a particular property if it is found in the
     * configuration file.  If it was not, the defaultValue will
     * be returned.
     * @param name key value for the property
     * @param defaultValue value to be returned if the property is not found
     * @return value associated with the key
     */
    public String getProperty(String name, String defaultValue) {
        return config.getProperty(name, defaultValue);
    }
    
    /**
     * Returns a particular property if it is found in the
     * configuration file.
     * @param name key value for the property
     * @return value associated with the key
     */
    public String getProperty(String name) {
        return config.getProperty(name);
    }
    
    /**
     * Returns a map of all handlers that the property file
     * designates as handlers that should be loaded for use
     * in the isolation vim.  This set is a subset (not proper) 
     * of all of the handlers available within the file.
     * @return a property set containing key value pairs of 
     * handlers for loading (Context, Handler Class)
     */
    public Properties getRuntimeHandlers() {
        return config;
    }
}
