/*
 * Copyright (c) 2009, Storage Networking Industry Association
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
package org.snia.xam.vim.isolation.server.util;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.logging.Logger;

import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.FieldContainer;
import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.FieldExistsException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidFieldNameException;
import org.snia.xam.InvalidOperationException;
import org.snia.xam.MaximumFieldException;
import org.snia.xam.ObjectInUseException;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XSetAbandonException;
import org.snia.xam.XSetCorruptException;
import org.snia.xam.XStream;
import org.snia.xam.XSystemAbandonException;
import org.snia.xam.XSystemCorruptException;
import org.snia.xam.base.AbstractFieldContainer;
import org.snia.xam.base.XAMConfiguration;
import org.snia.xam.util.LogManager;

/**
 * 
 */
public class ServerConfig 
   extends AbstractFieldContainer
   implements FieldContainer
{
   static final String NON_COMPLIANT = "ServerConfig is not a complaint FieldContainer";
   
   /**
     * Creates a new instance of StartupConfig with 
     * a startup properties file from the "current" directory.
     * 
     * @see Constants#STARTUP_PROPERTY_FILE
     */
   public ServerConfig() throws IOException,XAMException
   {
       this(new File(".", Constants.STARTUP_PROPERTY_FILE));
   }
   
   /* 
    * Create a new startup config from a file.
    * @param f a property file that contains the startup parameters for the VIM
    * @exception IOException thrown if there are any problems opening or reading
    * the property file (such as FileNotFound, etc...).
    */
   public ServerConfig(File f) throws IOException,XAMException
   {
      this.setProtectNamespace(false);
      Logger logger = LogManager.GetLogger( Constants.LOGGER_NAME );
      XAMConfiguration xc = new XAMConfiguration(logger);
      xc.load(f.getAbsolutePath(), this);
   }
   
   /**
    * Create a new startup configuration from a set of properties
    * @param p is a set of properties to use for startup, use keys as if they were
    * read directly from a properties file
    */
   public ServerConfig(Properties p) throws XAMException
   {
      this.setProtectNamespace(false);
      Logger logger = LogManager.GetLogger( Constants.LOGGER_NAME );
      XAMConfiguration xc = new XAMConfiguration(logger);
      xc.load(p, this);
   }

   /**
    * Returns an ordered list of stacked VIMs that are 
    * "underneath" the aliased VIM.  These VIMs should be aliases
    * themselves.
    * @param alias top level alias value
    */
   public String[] getStackedVims(String alias)
      throws XAMException
   {
       String fullValue = getString(Constants.ALIAS_NAMESPACE + "." + alias);
   
       if(fullValue==null) {
           return null;
       }

       StringTokenizer tokenizer = new StringTokenizer(fullValue, "|");
       // skip the first element
       String element = (String)tokenizer.nextElement();
       ArrayList<String> al = new ArrayList<String>();
       while(tokenizer.hasMoreElements()) {
           element = (String) tokenizer.nextElement();
           al.add(element);
       }
       return al.toArray(new String[al.size()]);
   }

   /**
    * Returns the value of an alias from the properties.  If the alias
    * contains an additional stack of VIMs, it is not returned here, only
    * the initial value is returned.
    * @return value of alias
    * @see #getStackedVims
    */
   public String getAliasValue(String alias) 
      throws XAMException
   {
       String fullValue = getString(Constants.ALIAS_NAMESPACE + "." + alias);
   
       if(fullValue==null) {
           return null;
       }

       StringTokenizer tokenizer = new StringTokenizer(fullValue, "|");
       return (String)tokenizer.nextElement();
   }
   
   // Following methods added because the abstract field container requires them.
   public XAsync asyncOpenXStream(String name,
                                  String mode,
                                  long xopid,
                                  XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      InvalidFieldNameException, FieldDoesNotExistException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      throw new InvalidOperationException(NON_COMPLIANT);
   }

   public XStream createXStream(String fieldName,
                                boolean binding,
                                String mimeType)
      throws AuthenticationExpiredException, InvalidArgumentException,
      InvalidFieldNameException, InvalidOperationException,
      FieldExistsException, MaximumFieldException, ObjectInUseException,
      XSetAbandonException, XSetCorruptException, XSystemAbandonException,
      XSystemCorruptException, XAMException
   {
      throw new InvalidOperationException(NON_COMPLIANT);
   }

   public XStream openXStream(String fieldName, String mode)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldDoesNotExistException, ObjectInUseException, XSetAbandonException,
      XSetCorruptException, XSystemAbandonException, XSystemCorruptException,
      XAMException
   {
      throw new InvalidOperationException(NON_COMPLIANT);
   }
}
