/*
 *
 * Copyright (c) 2007, Sun Microsystems, Inc.
 * Copyright (c) 2007-2009, Storage Networking Industry Association
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

package org.snia.xam.vim.isolation.server.util;

import java.util.HashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidXRIException;
import org.snia.xam.XAMException;
import org.snia.xam.XIterator;
import org.snia.xam.util.LogManager;
import org.snia.xam.vim.VIM;

/**
 * This class caches VIM instances for later use. 
 */
public class VIMCache {
    
   private final String XRI_LIST_PREFIX = ".org.snia.xam.vim.xri.list";
   
   public static VIMCache            cache          = null;
   private       HashMap<String,VIM> m_vimInstances = new HashMap<String,VIM>();
   private       ServerConfig        config         = null;
   
   protected Logger logger = null;
   
    /**
     * The constructor for XamLibraryCache is private, use
     * getInstance to retrieve it (this is a singleton class for
     * the virtual machine).
     */
    private VIMCache() {
       try {
          logger = LogManager.GetLogger(Constants.LOGGER_NAME);
       } catch( Exception e ) {
          logger = Logger.getLogger( Constants.LOGGER_NAME );
       }
    }
    
    /**
     * The singleton constructor for the VIMCache.  Only one of the caches can
     * exist per virtual machine.
     * @return the single instance of the VIMCache
     */
    public static VIMCache getInstance() {
        if(cache==null) {
            cache = new VIMCache();
        }
        return cache;
    }
    
    /**
     * Sets the configuration file into the VIMCache.
     * Also creates a map of legal XRI values, so it can reject unknown XRIs.
     */
    public void setConfiguration(ServerConfig sc) {
        config = sc;
    }
    
    /**
     * Adds a VIM to the cache.  VIMs are cached based on VIM name. There should
     * only need to be a single instance of a VIM per class name.
     * @param vimName The class name of the VIM, cannot be null
     * @param theVim The VIM associated with the class name.
     * @return boolean indicating whether the class name already had an associated VIM
     */
    public synchronized boolean addVim( String vimName, VIM theVim ) 
    {
       if( theVim==null ) 
          throw new NullPointerException("Parameter theVim cannot be null");
       
        if( vimName==null ) 
           throw new NullPointerException("Parameter String (vimName) cannot be null");
        
        boolean alreadyExists = m_vimInstances.containsKey(vimName);
        if(!alreadyExists) {
            m_vimInstances.put(vimName,theVim);
            if( logger.isLoggable(Level.FINEST))
               logger.log(Level.FINEST, "cache total items: "+m_vimInstances.size());
        }
        return alreadyExists;
    }
    
    /**
     * Removes an class name and VIM from the cache.
     * @param vimName The class name to be removed
     * @return boolean indicating whether the class anme was already in the cache
     */
    public synchronized boolean removeVim(String vimName) 
    {
       if(vimName==null) 
          throw new NullPointerException("Parameter String (xri) cannot be null");
       
       boolean alreadyExists = m_vimInstances.containsKey(vimName);
       if(alreadyExists) {
          m_vimInstances.remove(vimName);
       }
       return alreadyExists;
    }
    
    /**
     * Retrieves the VIM instance associated with the vimName if one exists 
     * in the cache
     * @param vimName The name that identifies the particular associated VIM to retrieve, it cannot be null
     * @return the VIM instance instance associated with the vimName, null if the vimName is not a part of the cache
     */
    public synchronized VIM getVIM( String vimName ) 
       throws XAMException, InvalidXRIException 
    {
        if(vimName==null) 
           throw new InvalidArgumentException("Parameter String (vimName) cannot be null");
        
        // parse the XRI into its constituent parts
        
        VIM theVim = m_vimInstances.get(vimName);
        
        if( theVim == null )
        {
           try
           {
              Class xamClass = Class.forName(vimName);
              theVim = (VIM)xamClass.newInstance();
           } catch( Exception e )
           {
              throw new InvalidXRIException( "Failed to create instance of VIM : " + vimName, e );
           }
           addVim(vimName,theVim);
        }
        
        return theVim;
    }
    
    /**
     * Retrieves the XRI after this one is "unstacked"
     */
    public synchronized String getNextXri(String xri) 
    throws InvalidXRIException, XAMException
    {
        if(xri==null) throw new InvalidXRIException("Parameter String (xri) cannot be null");
        
        // parse the XRI into its constituent parts
        XRIParser xp = new XRIParser(xri);
        
        // retrieve the vim name
        String vimName = xp.getVimName();
        if(vimName==null) {
            throw new InvalidXRIException( "No VIM name present in XRI: " + xri );
        } else {
            // since this is the HTTP VIM, we have to pop the stack to the next
            //  target via the configuration file, this logic should be moved UP
            //  to the top half of the VIM later
            String vims[] = config.getStackedVims(vimName);
            if( (vims == null) || (vims.length == 0) )
               throw new InvalidXRIException( "No VIM found for VIM name '" + vimName + "'" );
            String nextInStackAlias = vims[0];
            vimName = config.getAliasValue(nextInStackAlias);
        }
        
        // if the vimName is not set, must return
        if(vimName==null) {
            throw new InvalidXRIException("vimName not set on XRI and no default vimname configured");
        }
        
        xp.setVimName(vimName);
        return xp.toString();
    }
}
