/*
 *
 * Copyright (c) 2007, Sun Microsystems, Inc.
 * Copyright (c) 2007, Storage Networking Industry Association
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
package org.snia.xam.vim.isolation.server.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Properties;
import java.util.*;

/**
 *
 */
public class VimConfig {
    /**
     * A cache of XRIs to the class used to contact them.  This is only
     * populated on the first XRI to class map, a "lazy" population used
     * only if the surrounding framework uses this direction rather than
     * direct property manipulation.
     */
    protected HashMap<String,String> xriCache = null;
    
    /**
     * Creates a new instance of VimCOnfig with
     * a vim properties file from the "current" directory.
     *
     * @see Constants#STARTUP_PROPERTY_FILE
     */
    public VimConfig() throws IOException {
        this(new File(".", Constants.LOAD_VIM_FILE));
    }
    
   /*
    * Create a new vim config from a file.
    * @param f a property file that contains the properties for vims to load
    * @exception IOException thrown if there are any problems opening or reading
    * the property file (such as FileNotFound, etc...).
    */
    public VimConfig(File f) throws IOException {
        config = new Properties();
        FileInputStream fis = null;
        try {
            fis = new FileInputStream(f);
            config.load(fis);
            fis.close();
        } finally {
            if (fis != null) fis.close();
        }
    }
    
    /**
     * Create a new vim configuration from a set of properties
     * @param p is a set of properties to use for which vims to load, use keys as if they were
     * read directly from a properties file
     */
    public VimConfig(Properties p) {
        config = p;
    }
    
    public void setProperty(String name, String value) {
        if (value == null)
            config.remove(name);
        else
            config.setProperty(name, value);
    }
    
    /**
     * Returns a particular property if it is found in the
     * configuration file.  If it was not, the defaultValue will
     * be returned.
     * @param name key value for the property
     * @param defaultValue value to be returned if the property is not found
     * @return value associated with the key
     */
    public String getProperty(String name, String defaultValue) {
        return config.getProperty(name, defaultValue);
    }
    
    /**
     * Returns a particular property if it is found in the
     * configuration file.
     * @param name key value for the property
     * @return value associated with the key
     */
    public String getProperty(String name) {
        return config.getProperty(name);
    }
    
    private void initValues() {
        config = new Properties();
    }
    
    public Properties getProperties() {
        return config;
    }
    
    private Properties config;
    
    public Properties getConfig() {
        return config;
    }
    
    public void setConfig(Properties val) {
        this.config = val;
    }
    
    /**
     * Returns the value of an alias from the properties.  If the alias
     * contains an additional stack of VIMs, it is not returned here, only
     * the initial value is returned.
     * @return value of alias
     * @see #getStackedVims
     */
    public String getAliasValue(String alias) {
        String fullValue = getProperty(Constants.ALIAS_NAMESPACE + "." + alias);
    
        if(fullValue==null) {
            return null;
        }

        StringTokenizer tokenizer = new StringTokenizer(fullValue, "|");
        return (String)tokenizer.nextElement();
    }
    
    /**
     * Returns an ordered list of stacked VIMs that are 
     * "underneath" the aliased VIM.  These VIMs should be aliases
     * themselves.
     * @param alias top level alias value
     */
    public String[] getStackedVims(String alias) {
        String fullValue = getProperty(Constants.ALIAS_NAMESPACE + "." + alias);
    
        if(fullValue==null) {
            return null;
        }

        StringTokenizer tokenizer = new StringTokenizer(fullValue, "|");
        // skip the first element
        String element = (String)tokenizer.nextElement();
        ArrayList al = new ArrayList();
        while(tokenizer.hasMoreElements()) {
            element = (String)tokenizer.nextElement();
            al.add(element);
        }
        return (String[])al.toArray(new String[al.size()]);
    }
    
}
