/*
 * Copyright (c) 2009, Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* Copyright (c) 2007, Sun Microsystems, Inc.                                       */

package org.snia.xam.vim.isolation.server.util;

import java.util.HashMap;
import java.util.List;
import java.util.Set;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.snia.xam.InvalidArgumentException;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;

/**
 * A cache of XAsync operations that are in use.  
 */
public class XAsyncCache {
    
    HashMap<String,XAsyncCacheEntry> hm = new HashMap<String,XAsyncCacheEntry>();
    public static XAsyncCache cache = null;
    
    // Note that the logger must be passed in, as this class is used in the client.
    protected Logger logger         = null;
    
    /**
     * The ctor for XSetCache is private, use
     * getInstance to retrieve it (this is a singleton class for
     * the virtual machine).
     */
    private XAsyncCache(Logger logger) {
       this.logger = logger;
    }
    
    /**
     * The singleton ctor for the XSetCache.  Only one of the caches can
     * exist per virtual machine.
     * @return the single instance of the XSystemCache
     */
    public static XAsyncCache getInstance(Logger logger) {
        if(cache==null) {
            cache = new XAsyncCache(logger);
        }
        return cache;
    }
    
    /**
     * Adds an XAsync to the cache of sets.  This method returns
     * a handle to the object instance for use in future operations.
     * @param xsystemhandle The handle of the XSystem
     * @param async The XAsync operation to be added
     *
     * @return long a handle to the XSet for future use
     */
    public synchronized long addXAsync(long xsystemHandle, XAsync async ) 
       throws XAMException
    {
        if(async==null) throw new InvalidArgumentException("Parameter XAsync cannot be null");

        // this isn't very sophisticated but since we are synchronized it will probably do the job
        // perhaps get a real UUID generator later
        long handle = System.currentTimeMillis();
        
        addXAsyncWithHandle( xsystemHandle, handle, async );
        
        return handle;
    }
    
    /**
     * Adds an XAsync to the cache of sets.  The handle is expected to be generated externally to this method.
     * Typically this will be used by client VIMs tracking XAsync resources.
     * @param xsystemhandle The handle of the XSystem
     * @param async The XAsync operation to be added
     */
    public synchronized void addXAsyncWithHandle(long xsystemHandle, long asyncHandle, XAsync async ) 
       throws XAMException
    {
        if(async==null) throw new InvalidArgumentException("Parameter XAsync cannot be null");

        XAsyncCacheEntry xace = new XAsyncCacheEntry(xsystemHandle, async, asyncHandle );
        hm.put(new Long(asyncHandle).toString(), xace);
        if( logger.isLoggable(Level.FINEST))
           logger.log(Level.FINEST, "Added " + asyncHandle + ": cache total items: "+hm.size());
        return;
    }
    
    /**
     * Removes associated XAsync from the cache.
     * @param handle is the identifier for the system to be removed
     * @return boolean indicating whether the item was already in the cache
     */
    public synchronized boolean removeXAsync(long handle) {
        String handleAsString = new Long(handle).toString();
        
        boolean alreadyExists = hm.containsKey(handleAsString);
        if(alreadyExists) {
            hm.remove(handleAsString);
        }
        return alreadyExists;
    }

    /**
     * Retrieves the XAsync associated with the handle if one exists in the 
     * cache
     * @param handle is the identifier for the system to be removed
     * @return the XAsync object instance associated with the handle, null if
     * the handle is not a part of the cache
     */
    public synchronized XAsync getXAsync(long handle) 
       throws XAMException
    {
        XAsync xa = null;
        String handleAsString = new Long(handle).toString();
        XAsyncCacheEntry xace = hm.get(handleAsString);
        if(xace!=null) {
            xa = xace.getXAsync();
        }
        
        if( xa == null )
           throw new XAMException( "Handle " + handle + " does not refer to an active XASync",
                                   XAMException.sXAM_INVALID_HANDLE );
        return xa;
    }
    
    /**
     * Retrieves the XAsync associated with the handle if one exists in the 
     * cache
     * @param handle is the identifier for the system to be removed
     * @return the XAsync object instance associated with the handle, null if
     * the handle is not a part of the cache
     */
    public synchronized long getXSystemHandle(long handle) 
       throws XAMException
    {
        String handleAsString = new Long(handle).toString();
        XAsyncCacheEntry xace = hm.get(handleAsString);
        if( xace == null )
           throw new XAMException( "Handle " + handle + " does not refer to an active XASync",
                                   XAMException.sXAM_INVALID_HANDLE );
        else
           return xace.getXSystemHandle();
    }
    
    /**
     * Retrieves AND removes the XSystem in one operation, this prevents
     * the system from being retrieved before it is shut down gracefully.
     * @param handle is the identifier for the system to be removed
     * @return XSystem identified by the handle, null if it didn't exist
     */
    public synchronized XAsync getAndRemoveXSet(long handle) 
       throws XAMException
    {
        XAsync xa = getXAsync(handle);
        if(xa!=null) {
            removeXAsync(handle);
        }
        
        return xa;
    }

    /**
     * Return a list of all XAsyncCache objects matching a specific XSystem Handle. Because this is linear,
     * this structure will not scale to large numbers of active XAsync operations. 
     * @param xsystemHandle
     * @return The matching list.
     */
   public synchronized List<XAsyncCacheEntry> getXSystemList(Long xsystemHandle)
   {
      Vector<XAsyncCacheEntry> asyncList = new Vector<XAsyncCacheEntry>();
      
      Set<String> keys = hm.keySet();
      for( String key : keys ) 
      {
         XAsyncCacheEntry xace = hm.get(key);
         if( xace.getXSystemHandle() == xsystemHandle )
            asyncList.add(xace);
      }
      return asyncList;
   }

   public synchronized XAsyncCacheEntry getCacheEntry(long asyncHandle)
      throws XAMException
   {
      String handleAsString = new Long(asyncHandle).toString();
      XAsyncCacheEntry xace = hm.get(handleAsString);
      
      if( xace == null )
         throw new XAMException( "Handle " + asyncHandle + " does not refer to an active XASync",
                                 XAMException.sXAM_INVALID_HANDLE );
      return xace;
   }
}
