/*
 *
 * Copyright (c) 2007, Sun Microsystems, Inc.
 * Copyright (c) 2007-2009, Storage Networking Industry Association
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

package org.snia.xam.vim.isolation.server.util;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import org.snia.xam.InvalidXRIException;

/**
 */
public class XRIParser {
    String xri = null;
    String vimName = null;
    String xSystemName = null;
    Map<String,String> parameters = null;
    
    /** Creates a new instance of XRIParser */
    public XRIParser(String xri) throws InvalidXRIException {
        this.xri = xri;
        boolean valid = xri.startsWith("snia-xam://");
        if(!valid) {
            throw new InvalidXRIException("Invalid XRI, does not begin with snia-xam:// "+xri);
        }
        
        // strip off snia-xam://
        String xriBody = xri.substring("snia-xam://".length());
        
        // if there are no characters left, there is no xsystemname and the XRI is invalid
        if(xriBody.length()<=0) {
            throw new InvalidXRIException("Invalid XRI, does not contain xsystemname: "+xri);
        }
        
        // divide into vimname+(xsystemname+params)
        int endOfVimName = xriBody.indexOf('!');
        String xSystemAndParams = xriBody;
        if(endOfVimName!=-1) {
            vimName = xriBody.substring(0,endOfVimName);
            xSystemAndParams = xriBody.substring(endOfVimName+1);
        }
        
        // divide xsystemname+params
        int startOfParameters = xSystemAndParams.indexOf('?');
        String params = null;
        if(startOfParameters==-1) {
            xSystemName = xSystemAndParams;
        } else {
            xSystemName = xSystemAndParams.substring(0,startOfParameters);
            params = xSystemAndParams.substring(startOfParameters+1);
        }
        
        // divide params into Map
        if(params!=null) {
            parameters = new HashMap<String,String>();
            StringTokenizer tokenizer = new StringTokenizer(params,"&");
            // go through the elements
            while(tokenizer.hasMoreElements()) {
                String element = (String)tokenizer.nextElement();
                int equalIndex = element.indexOf("=");
                String name = element.substring(0,equalIndex);
                String value = element.substring(equalIndex+1);
                parameters.put(name,value);
            }
        }
        
    }
    
    /**
     * Retrieves the VIM Name
     */
    public String getVimName() {
        return vimName;
    }
    
    /**
     * Retrieves the XSystemName from the XRI
     */
    public String getXSystemName() {
        return xSystemName;
    }
    
    /**
     * Retrieves XRI Parameters
     */
    public Map<String,String> getParameters() {
        return parameters;
    }
    
    public String getParameter( String name )
    {
       return parameters.get(name);
    }
    
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("snia-xam://");
        if(vimName!=null) {
            sb.append(vimName);
            sb.append("!");
        }
        
        sb.append(xSystemName);
        if(parameters!=null) {
            sb.append("?");
            Set keys = parameters.keySet();
            Iterator i = keys.iterator();
            while(i.hasNext()) {
                String key = (String)i.next();
                String value = (String)parameters.get(key);
                sb.append(key);
                sb.append("=");
                sb.append(value);
                if(i.hasNext()) sb.append("&");
            }
        }
        
        return sb.toString();
    }
    
    public static void main(String args[]) {
        try {
            XRIParser xp = new XRIParser("snia-xam://vimSun!123.122.1.5?name=value&name2=value");
            XRIParser xp2 = new XRIParser("snia-xam://123.122.1.5");
            XRIParser xp3 = new XRIParser("snia-xam://vimSun!localhost");
        } catch (InvalidXRIException ex) {
            ex.printStackTrace();
        }
    }

    /**
     * Setter for property XSystemName.
     * @param xSystemName New value of property XSystemName.
     */
    public void setXSystemName(java.lang.String xSystemName) {
        this.xSystemName = xSystemName;
    }

    /**
     * Setter for property parameters.
     * @param parameters New value of property parameters.
     */
    public void setParameters(Map<String,String> parameters) {
        this.parameters = parameters;
    }

    /**
     * Setter for property vimName.
     * @param vimName New value of property vimName.
     */
    public void setVimName(java.lang.String vimName) {
        this.vimName = vimName;
    }
}
