/*
 * Copyright (c) 2007-2009, Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* Copyright (c) 2007, Sun Microsystems, Inc.                                       */

package org.snia.xam.vim.isolation.server.util;

import java.util.HashMap;
import java.util.Set;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.snia.xam.InvalidArgumentException;
import org.snia.xam.XAMException;
import org.snia.xam.XStream;
import org.snia.xam.util.LogManager;

/**
 * A cache of systems that are in use.  This cache will auto clean the
 * connections if they are not used for an agreed upon amount of time.
 * @author Owner
 */
public class XStreamCache {

    HashMap<String,XStreamCacheEntry> hm = new HashMap<String,XStreamCacheEntry>();
    public static XStreamCache cache = null;
    protected Logger logger = null;
    
    /**
     * The ctor for XSystemCache is private, use
     * getInstance to retrieve it (this is a singleton class for
     * the virtual machine).
     */
    private XStreamCache() {
       try { 
          logger = LogManager.GetLogger( Constants.LOGGER_NAME );
       } catch( Exception e ) { 
          logger = Logger.getLogger( Constants.LOGGER_NAME );
       }
    }

    /**
     * The singleton ctor for the XSystemCache.  Only one of the caches can
     * exist per virtual machine.
     * @return the single instance of the XSystemCache
     */
    public static XStreamCache getInstance() {
        if(cache==null) {
            cache = new XStreamCache();
        }
        return cache;
    }

    /**
     * Adds an XSystem to the cache of connected systems.  This method returns
     * a handle to the object instance for use in future operations.
     * @param xs The XSystem associated with the XRI, cannot be null
     * the next operation on it, -1 indicates the default lease time should
     * be applied
     *
     * @return long a handle to the XSystem for future use
     */
    public synchronized long addXStream(long xsystemhandle, XStream xs) 
       throws XAMException
    {
        if(xs==null) throw new InvalidArgumentException("Parameter XStream cannot be null");

        // this isn't very sophisticated but since we are synchronized it will probably do the job
        // perhaps get a real UUID generator later
        long handle = System.currentTimeMillis();

        XStreamCacheEntry xsce = new XStreamCacheEntry(xsystemhandle, xs, handle);
        hm.put(new Long(handle).toString(), xsce);
        if( logger.isLoggable(Level.FINEST))
           logger.log(Level.FINEST, "cache total items: "+hm.size());
        return handle;
    }

    /**
     * Removes an xri and associated XSystem from the cache.
     * @param handle is the identifier for the system to be removed
     * @return boolean indicating whether the xri was already in the cache
     */
    public synchronized boolean removeXStream(long handle) {
        String handleAsString = new Long(handle).toString();

        boolean alreadyExists = hm.containsKey(handleAsString);
        if(alreadyExists) {
            hm.remove(handleAsString);
        }
        return alreadyExists;
    }

    /**
     * Retrieves the XSystem associated with the XRI if one exists in the
     * cache
     * @param handle is the identifier for the system to be removed
     * @return the XSystem object instance associated with the handle, null if
     * the handle is not a part of the cache
     */
    public synchronized XStream getXStream(long handle) 
       throws XAMException
    {
        XStream xs = null;
        String handleAsString = new Long(handle).toString();
        XStreamCacheEntry xsce = (XStreamCacheEntry)hm.get(handleAsString);
        if(xsce!=null) {
            xs = xsce.getXStream();
        }
        
        if( xs == null )
           throw new XAMException( "Handle " + handle + " does not refer to an active XStream",
                                   XAMException.sXAM_INVALID_HANDLE );
        return xs;
    }

    /**
     * Retrieves the XSystemHandle associated with xstream if one exists in the
     * cache
     * @param handle is the identifier for the stream to be retrieved
     * @return the XSystem object instance associated with the handle, null if
     * the handle is not a part of the cache
     */
    public synchronized long getXSystemHandle(long xstreamHandle) 
       throws XAMException
    {
        String handleAsString = new Long(xstreamHandle).toString();
        XStreamCacheEntry xsce = (XStreamCacheEntry)hm.get(handleAsString);
        if( xsce == null )
           throw new XAMException( "Handle " + xstreamHandle + " does not refer to an active XStream",
                                   XAMException.sXAM_INVALID_HANDLE );
        else
           return xsce.getXSystemHandle();
    }

    /**
     * Retrieves AND removes the XSystem in one operation, this prevents
     * the system from being retrieved before it is shut down gracefully.
     * @param handle is the identifier for the system to be removed
     * @return XSystem identified by the handle, null if it didn't exist
     */
    public synchronized XStream getAndRemoveXStream(long handle) 
       throws XAMException
    {
        XStream xs = getXStream(handle);
        if(xs!=null) {
            removeXStream(handle);
        }

        return xs;
    }
    
    /**
     * Used when the XStream is known and need to retrieve a handle. Sadly the map is not setup 
     * to efficiently do this, so it becomes a linear search. 
     * @param xs
     * @return The found handle, or -1 if the XSet is not in the map.
     */
    public long getHandle( XStream xs )
    {
       Set<String> keys = hm.keySet();
       for( String key : keys )
       {
          XStreamCacheEntry xsce = hm.get(key);
          if( xsce.getXStream() == xs )
             return xsce.getHandle();
       }
       
       return -1;
    }

    /** When the XSystem is closed or abandoned, remove items referenced by this XSystem handle.
     * @param handle The handle of the XSystem which is no longer available.
     */
   public synchronized void xsystemRemoved(long handle)
   {
      Set<String> keys = hm.keySet();
      Vector<String> removeList = new Vector<String>();
      for( String key : keys )
      {
         XStreamCacheEntry xsce = hm.get(key);
         if( xsce.getXSystemHandle() == handle )
            removeList.add(key);
      }
      
      for( String xsHandle : removeList )
      {
         hm.remove(xsHandle);
      }
   }
}
