/*
 * Copyright (c) 2007-2009, Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *      * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *      * Neither the name of Storage Networking Industry Association
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* Copyright (c) 2007, Sun Microsystems, Inc.                                       */

package org.snia.xam.vim.isolation.server.util;

import java.util.HashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.snia.xam.InvalidArgumentException;
import org.snia.xam.XAMException;
import org.snia.xam.XSystem;
import org.snia.xam.util.LogManager;

/**
 * A cache of systems that are in use.  This cache will auto clean the
 * connections if they are not used for an agreed upon amount of time.
 * @author Owner
 */
public class XSystemCache {
    
    HashMap<String,XSystemCacheEntry> hm = new HashMap<String,XSystemCacheEntry>();
    public static XSystemCache cache = null;
    protected Logger logger = null;
    
    /**
     * The ctor for XSystemCache is private, use
     * getInstance to retrieve it (this is a singleton class for
     * the virtual machine).
     */
    private XSystemCache() {
       try { 
          logger = LogManager.GetLogger( Constants.LOGGER_NAME );
       } catch( Exception e ) { 
          logger = Logger.getLogger( Constants.LOGGER_NAME );
       }
    }
    
    /**
     * The singleton ctor for the XSystemCache.  Only one of the caches can
     * exist per virtual machine.
     * @return the single instance of the XSystemCache
     */
    public static XSystemCache getInstance() {
        if(cache==null) {
            cache = new XSystemCache();
        }
        return cache;
    }
    
    /**
     * Adds an XSystem to the cache of connected systems.  This method returns
     * a handle to the object instance for use in future operations.
     * @param xri The XRI that identifies the particular XSystem, cannot be null
     * @param xs The XSystem associated with the XRI, cannot be null
     * @param lease How long the XSystem should be maintained in the cache before
     * the next operation on it, -1 indicates the default lease time should
     * be applied
     *
     * @return long a handle to the XSystem for future use
     */
    public synchronized long addXSystem(String xri, XSystem xs ) 
       throws XAMException
    {
        if(xs==null) throw new InvalidArgumentException("Parameter XSystem cannot be null");
        if(xri==null) throw new InvalidArgumentException("Parameter String (xri) cannot be null");
        
        // this isn't very sophisticated but since we are synchronized it will probably do the job
        // perhaps get a real UUID generator later
        long handle = System.currentTimeMillis();
        
        XSystemCacheEntry xsce = new XSystemCacheEntry(xri, xs, handle);
        hm.put(new Long(handle).toString(), xsce);
        if( logger.isLoggable(Level.FINEST))
           logger.log(Level.FINEST, "cache total items: "+hm.size());
        
        return handle;
    }
    
    /**
     * Removes an xri and associated XSystem from the cache.
     * @param handle is the identifier for the system to be removed
     * @return boolean indicating whether the xri was already in the cache
     */
    public synchronized boolean removeXSystem(long handle) {
        String handleAsString = new Long(handle).toString();
        
        boolean alreadyExists = hm.containsKey(handleAsString);
        if(alreadyExists) {
            hm.remove(handleAsString);
        }
        return alreadyExists;
    }
    
    /**
     * Retrieves the XSystem associated with the XRI if one exists in the
     * cache
     * @param handle is the identifier for the system to be removed
     * @return the XSystem object instance associated with the handle, null if
     * the handle is not a part of the cache
     */
    public synchronized XSystem getXSystem(long handle) 
       throws XAMException
    {
        XSystem xs = null;
        String handleAsString = new Long(handle).toString();
        XSystemCacheEntry xsce = (XSystemCacheEntry)hm.get(handleAsString);
        if(xsce!=null) {
            xs = xsce.getXsystem();
        }
        if( xs == null )
        {
           String msg = "Handle " + handle + " does not refer to an active XSystem";
           logger.severe(msg);
           throw new XAMException( msg, XAMException.sXAM_INVALID_HANDLE );
        }
        return xs;
    }
    
    /**
     * Retrieves AND removes the XSystem in one operation, this prevents
     * the system from being retrieved before it is shut down gracefully.
     * @param handle is the identifier for the system to be removed
     * @return XSystem identified by the handle, null if it didn't exist
     */
    public synchronized XSystem getAndRemoveXSystem(long handle) 
       throws XAMException
    {
        XSystem xs = getXSystem(handle);
        if(xs!=null) {
            removeXSystem(handle);
        }
        
        return xs;
    }
}
