/* Copyright (c) 2009, The Storage Networking Industry Association.

   Redistribution and use in source and binary forms, with or without modification, 
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.
   
   * Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.
   
   * Neither the name of The Storage Networking Industry Association (SNIA) nor 
   the names of its contributors may be used to endorse or promote products 
   derived from this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
   THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.snia.xam.vim.http;

import org.snia.xam.InvalidOperationException;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XSet;
import org.snia.xam.XStream;
import org.snia.xam.XSystem;
import org.snia.xam.XUID;
import org.snia.xam.util.SASLUtils;

import junit.framework.Test;
import junit.framework.TestResult;
import junit.framework.TestSuite;

public class HTTPXAsyncTest
   extends XAMTest
{
   final String PROP_PREFIX   = "com.example.plist.";
   final String PROP_BOOL     = PROP_PREFIX + "bool";
   final String PROP_INT      = PROP_PREFIX + "int";
   final String PROP_DOUBLE   = PROP_PREFIX + "double";
   final String PROP_DATETIME = PROP_PREFIX + "datetime";
   final String PROP_STRING   = PROP_PREFIX + "string";
   final String PROP_XUID     = PROP_PREFIX + "xuid";
   
   class asyncListener implements XAsyncListener 
   {
      boolean   m_complete  = false;
      Exception m_exception = null;
      boolean   m_fail      = false;
      
      public void XAsyncCallback(XAsync asyncState)
      {
         try {
            m_complete = asyncState.isComplete();
            synchronized(this) {
//               System.out.println( "Notifying waiting threads");
               notifyAll();
            }
         } catch( Exception e ) { m_exception = e; }
      }
   }
   
   public HTTPXAsyncTest( String name )
   { 
      super(name);
   }
   
   public static int runTest(String[] args, Test test)
   {
       TestResult res = junit.textui.TestRunner.run(test);
       return Math.max(res.failureCount(), res.errorCount());
   }
   
  public static Test suite()
  {
     TestSuite suite = new TestSuite();
     suite.addTest(new HTTPXAsyncTest("testCreateCopyXSet"));
     suite.addTest(new HTTPXAsyncTest("testXSystemOpenStream"));
     suite.addTest(new HTTPXAsyncTest("testHalt"));
     suite.addTest(new HTTPXAsyncTest("testAsyncStream"));
     suite.addTest(new HTTPXAsyncTest("testListeners"));
     suite.addTest(new HTTPXAsyncTest("testGetStatus"));
     return suite;
  }

   public void testCreateCopyXSet() throws Exception
   {
      XSystem xsystem = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsystem, "testuser", "testpasswd" );
      XSet xs = xsystem.createXSet(XSet.MODE_UNRESTRICTED);
      
      xs.createProperty(PROP_BOOL, false, false);
      assertFalse( "Expected bool prop to be false", xs.getBoolean( PROP_BOOL ));
      xs.setProperty( PROP_BOOL, true );
      assertTrue( "Expected bool prop to be true", xs.getBoolean( PROP_BOOL ));
      XUID xuid = xs.commit();
      xs.close();
      xs = null;
      
      XAsync xa = xsystem.asyncOpenXSet(xuid, XSet.MODE_READ_ONLY, 4242, null );
      waitForXA(xa);
      
      xs = xa.getXSet();
      assertTrue( "Expected bool prop to be true after async open", xs.getBoolean( PROP_BOOL ));
      
      long testId = xa.getXOPID();
      assertEquals( "Wrong xopid from XAsync object", 4242, testId );
      
      XUID xuid2 = xa.getXUID();
      assertEquals( "Expected the retrieved XUID to match the Xset", xuid, xuid2 );
      
      // Should just work.
      xa.getStatus();
      
      xa.close();
      
      // Test a copy
      xa = xsystem.asyncCopyXSet(xuid, XSet.MODE_UNRESTRICTED, 12345, null );
      waitForXA(xa);
      XSet xsCopy = xa.getXSet();
      XUID xuidCopy = xsCopy.commit();
      assertFalse( "Expected xuids to different after a copy", xuid.equals(xuidCopy) );
      assertTrue( "Expected the boolean property to be present in the copied XSet", xsCopy.getBoolean(PROP_BOOL));
      xa.close();
      
      xa = xsCopy.asyncCommit(789, null);
      waitForXA(xa);
      xa.getStatus();
      XUID asyncXuid = xa.getXUID();
      assertEquals( "Expected same XUID to come back from asynchronous XSet commit", xuidCopy, asyncXuid );
      
      xa.close();
      xs.abandon();
      xs.close();
      xsCopy.close();
      xsystem.close();
   }
   
   public void testXSystemOpenStream() throws Exception
   {
      XSystem xsystem = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsystem, "testuser", "testpasswd" );
      
      try { 
         XAsync xa = xsystem.asyncOpenXStream("test.stream", XStream.MODE_READ_ONLY, 1234, null);
         waitForXA(xa);
         xa.close();
      }
      catch( InvalidOperationException e ) { /* ignore for now */ }
      
      xsystem.abandon();
      xsystem.close();
   }
   
   public void testHalt() throws Exception
   {
      XSystem xsystem = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsystem, "testuser", "testpasswd" );
      XSet xs = xsystem.createXSet(XSet.MODE_UNRESTRICTED);
      
      xs.createProperty(PROP_BOOL, false, false);
      assertFalse( "Expected bool prop to be false", xs.getBoolean( PROP_BOOL ));
      xs.setProperty( PROP_BOOL, true );
      assertTrue( "Expected bool prop to be true", xs.getBoolean( PROP_BOOL ));
      XUID xuid = xs.commit();
      xs.close();
      xs = null;
      
      XAsync xa = xsystem.asyncOpenXSet(xuid, XSet.MODE_READ_ONLY, 1234, null );
      
      xa.halt();
      
      assertTrue( "Expected XAsync to be complete after halt", xa.isComplete() );
      
      // Should just work.
      xa.getStatus();
      
      xs = xa.getXSet();
      if( xs != null )
         xs.close();
      
      xa.close();
      xsystem.abandon();
      xsystem.close();
   }
   
   public void testAsyncStream() throws Exception
   {
      XSystem xsystem = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsystem, "testuser", "testpasswd" );
      XSet xs = xsystem.createXSet(XSet.MODE_UNRESTRICTED);
      
      xs.createProperty(PROP_BOOL, false, false);
      assertFalse( "Expected bool prop to be false", xs.getBoolean( PROP_BOOL ));
      xs.setProperty( PROP_BOOL, true );
      assertTrue( "Expected bool prop to be true", xs.getBoolean( PROP_BOOL ));
      XUID xuid = xs.commit();
      
      XStream testStream = xs.createXStream("test.stream", false, "text/plain" );
       
      String testString = "certum est quia impossible est";
      byte chars[] = testString.getBytes();
      XAsync xa = testStream.asyncWrite( chars, chars.length, 1234, null );
      waitForXA(xa);
      assertEquals( "Expected the correct number of bytes to be written", chars.length, xa.getBytesWritten() );
      xa.close();
      
      xa = testStream.asyncClose(1234, null);
      waitForXA(xa);
      xa.close();
      
      xa = xs.asyncOpenXStream( "test.stream", XStream.MODE_READ_ONLY, 1234, null );
      
      waitForXA(xa);
      
      testStream = xa.getXStream();
      byte buffer[] = new byte[1024];
      xa = testStream.asyncRead( buffer, buffer.length, 123, null );
      waitForXA(xa);
      int bytesRead = (int) xa.getBytesRead();
      assertEquals( "Expected the correct number of bytes to be read", chars.length, bytesRead );
      String testAsyncString = new String( buffer, 0, bytesRead );
      assertEquals( "Expected asynchronous read to match the original string", testString, testAsyncString );
      
      xa.close();
      testStream.close();
      
      xs.close();
      xs = null;

      xsystem.close();
   }
   
   
   public void testListeners() throws Exception
   {
//      System.out.println( "\n===Asynchronous test===" );
      XSystem xsystem = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsystem, "testuser", "testpasswd" );
      XSet xs = xsystem.createXSet(XSet.MODE_UNRESTRICTED);
      
      xs.createProperty(PROP_BOOL, false, false);
      assertFalse( "Expected bool prop to be false", xs.getBoolean( PROP_BOOL ));
      xs.setProperty( PROP_BOOL, true );
      assertTrue( "Expected bool prop to be true", xs.getBoolean( PROP_BOOL ));
      XUID xuid = xs.commit();
      xs.close();
      xs = null;
      
      asyncListener al = new asyncListener();
      XAsync xa = xsystem.asyncOpenXSet(xuid, XSet.MODE_READ_ONLY, 1234, al ); 
      synchronized(al) {
//         System.out.println( "Waiting for the listener" );
         al.wait(20000);
      }
      xs = xa.getXSet();
      xs.close();

      if( al.m_exception != null )
         throw al.m_exception;
      assertTrue( "Expected listener to have returned isComplete == true", al.m_complete );
            
      xsystem.close();
   }
   
   public void testGetStatus() throws Exception
   {
      XSystem xsystem = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsystem, "testuser", "testpasswd" );
      XSet xs = xsystem.createXSet(XSet.MODE_UNRESTRICTED);
      
      xs.createProperty(PROP_BOOL, false, false);
      assertFalse( "Expected bool prop to be false", xs.getBoolean( PROP_BOOL ));
      xs.setProperty( PROP_BOOL, true );
      assertTrue( "Expected bool prop to be true", xs.getBoolean( PROP_BOOL ));
      XUID xuid = xs.commit();
      xs.close();
      xs = null;
      
      asyncListener al = new asyncListener();
      al.m_fail = true;
      XAsync xa = xsystem.asyncOpenXSet(xuid, XSet.MODE_READ_ONLY, 1234, al ); 
      synchronized(al) {
         al.wait(20000);
      }

      try 
      { 
         xa.getStatus();
      }
      catch( Throwable t )
      {
         /* expected */
      }
      assertTrue( "Expected listener to have returned isComplete == true", al.m_complete );
       
      xa.close();
      xsystem.abandon();
      xsystem.close();
   }

   private void waitForXA( XAsync xa ) throws XAMException
   {
      while( ! xa.isComplete() )
      {
         try { Thread.sleep(100); }
         catch( InterruptedException ie ) { /* ignore */ }
      }
   }
   /**
    * Stand alone tests.
    */
   public static void main(String[] args)
   {
      System.exit(runTest(args, suite()));
   }
}
