/* Copyright (c) 2009, The Storage Networking Industry Association.

   Redistribution and use in source and binary forms, with or without modification, 
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.
   
   * Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.
   
   * Neither the name of The Storage Networking Industry Association (SNIA) nor 
   the names of its contributors may be used to endorse or promote products 
   derived from this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
   THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.snia.xam.vim.http;

import java.util.Calendar;

import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XIterator;
import org.snia.xam.XSet;
import org.snia.xam.XStream;
import org.snia.xam.XSystem;
import org.snia.xam.XUID;
import org.snia.xam.base.DefaultXUID;
import org.snia.xam.util.SASLUtils;

import junit.framework.Test;
import junit.framework.TestResult;
import junit.framework.TestSuite;

public class HTTPXSetTest
   extends XAMTest
{
   final String PROP_PREFIX   = "com.example.plist.";
   final String PROP_BOOL     = PROP_PREFIX + "bool";
   final String PROP_INT      = PROP_PREFIX + "int";
   final String PROP_DOUBLE   = PROP_PREFIX + "double";
   final String PROP_DATETIME = PROP_PREFIX + "datetime";
   final String PROP_STRING   = PROP_PREFIX + "string";
   final String PROP_XUID     = PROP_PREFIX + "xuid";
   
   public HTTPXSetTest( String name )
   { 
      super(name);
   }
   
   public static int runTest(String[] args, Test test)
   {
       TestResult res = junit.textui.TestRunner.run(test);
       return Math.max(res.failureCount(), res.errorCount());
   }
   
  public static Test suite()
  {
     TestSuite suite = new TestSuite();
     suite.addTest(new HTTPXSetTest("testPropertyCreateSetGet"));
     suite.addTest(new HTTPXSetTest("testAttributeFunctions"));
     suite.addTest(new HTTPXSetTest("testCreateXStream"));
     suite.addTest(new HTTPXSetTest("testJobMethods"));
     suite.addTest(new HTTPXSetTest("testIterator"));
     suite.addTest(new HTTPXSetTest("testAccessPolicy"));
     suite.addTest(new HTTPXSetTest("testBaseRetention"));
     suite.addTest(new HTTPXSetTest("testManagement"));
     suite.addTest(new HTTPXSetTest("testRetention"));
     suite.addTest(new HTTPXSetTest("testDisposition"));
     suite.addTest(new HTTPXSetTest("testStoragePolicy"));
     suite.addTest(new HTTPXSetTest("testImportExport"));
     return suite;
  }

   public void testPropertyCreateSetGet() throws Exception
   {
      XSystem xsystem = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsystem, "testuser", "testpasswd" );
      XSet xs = xsystem.createXSet(XSet.MODE_UNRESTRICTED);
      
      xs.createProperty(PROP_BOOL, false, false);
      assertFalse( "Expected bool prop to be false", xs.getBoolean( PROP_BOOL ));
      xs.setProperty( PROP_BOOL, true );
      assertTrue( "Expected bool prop to be true", xs.getBoolean( PROP_BOOL ));
      
      xs.createProperty( PROP_INT, false, 42 );
      assertEquals( PROP_INT, 42, xs.getLong(PROP_INT ));
      xs.setProperty( PROP_INT, 43 );
      assertEquals( PROP_INT, 43, xs.getLong(PROP_INT ));
      
      xs.createProperty( PROP_DOUBLE, false, -42.0 );
      assertEquals( PROP_DOUBLE, -42.0, xs.getDouble(PROP_DOUBLE ));
      xs.setProperty( PROP_DOUBLE, -43.0 );
      assertEquals( PROP_DOUBLE, -43.0, xs.getDouble(PROP_DOUBLE ));
      
      xs.createProperty( PROP_STRING, false, "foo" );
      assertEquals( PROP_STRING, "foo", xs.getString(PROP_STRING ));
      xs.setProperty( PROP_STRING, "bar" );
      assertEquals( PROP_STRING, "bar", xs.getString(PROP_STRING));
      
      Calendar c = Calendar.getInstance();
      xs.createProperty( PROP_DATETIME, false, c );
      assertEquals( PROP_DATETIME, c, xs.getDateTime(PROP_DATETIME));
      try { Thread.sleep( 321 ); }
      catch( InterruptedException ie ) { ; } 
      Calendar c2 = Calendar.getInstance();
      xs.setProperty( PROP_DATETIME, c2 );
      assertEquals( PROP_DATETIME, c2, xs.getDateTime(PROP_DATETIME));
      
      XUID xuid = DefaultXUID.valueOf("AAA6AwA0ymgwMDUwNTZDMDAwMDgtMjAwOC0wNy0xOFQxMV81OF8zMl80ODMtMDhfMDAtYQ==" );
      xs.createProperty( PROP_XUID, false, xuid );
      assertEquals( PROP_XUID, xuid, xs.getXUID(PROP_XUID));
      XUID xuid2 = DefaultXUID.valueOf("AAAEcwA9419GVk1LQ05PNlRPQlE0ZTVJUTc1UzNNSlQ2Q0pHNDEzRjE1VkcySDFLNjFHMFFUTVRLOFNWNw==");
      xs.setProperty( PROP_XUID, xuid2 );
      assertEquals( PROP_XUID, xuid2, xs.getXUID(PROP_XUID));
      
      assertEquals( "Wrong return from XSetContainsField", true, xs.containsField(PROP_STRING));
      
      xs.deleteField( PROP_BOOL );
      xs.deleteField( PROP_INT );
      xs.deleteField( PROP_DOUBLE );
      xs.deleteField( PROP_STRING );
      xs.deleteField( PROP_DATETIME );
      xs.deleteField( PROP_XUID );
      xs.abandon();
      xs.close();
      xsystem.close();
   }
   
   public void testAttributeFunctions() throws Exception
   {
      XSystem xsystem = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsystem, "testuser", "testpasswd" );
      XSet xs = xsystem.createXSet(XSet.MODE_UNRESTRICTED);
      xs.createProperty( PROP_BOOL, false, false );
      xs.createProperty( PROP_INT,  false, 42 );
      xs.createProperty( PROP_STRING, false, "12345" );
      
      assertEquals( "Wrong attr length on boolean prop", 1, xs.getFieldLength(PROP_BOOL));
      assertEquals( "Wrong attr length on int prop", 8, xs.getFieldLength(PROP_INT));
      assertEquals( "Wrong attr length on string prop", 5, xs.getFieldLength(PROP_STRING));
      
      assertEquals( "Wrong binding on boolean prop", false, xs.getFieldBinding(PROP_BOOL));
      
      assertEquals( "Wrong readonly on boolean prop", false, xs.getFieldReadOnly(PROP_BOOL));
      assertEquals( "Wrong readonly on identiy prop", false, xs.getFieldReadOnly(PROP_BOOL));
      
      assertEquals( "Wrong type on bool prop", XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, xs.getFieldType(PROP_BOOL));
      
      final String COMPLICATED_STRING = "a complicated string with control chars\nstuff\rthings=tidbits:to confuse";
      final String COMP_PROP = PROP_PREFIX + "complicate_string";
      xs.createProperty( COMP_PROP, false, COMPLICATED_STRING );
      assertEquals( "Wrong value for complicated string", COMPLICATED_STRING, xs.getString(COMP_PROP));
      
      xs.setFieldAsBinding(PROP_BOOL);
      XUID xuid2 = xs.commit();
      assertEquals( "Wrong binding after setFieldAsBinding", true, xs.getFieldBinding(PROP_BOOL));
      
      xs.setFieldAsNonbinding(PROP_BOOL);
      XUID xuid3 = xs.commit();
      assertEquals( "Wrong binding after setFieldAsNonbinding", false, xs.getFieldBinding( PROP_BOOL ));
      
      assertFalse( "Expected different xuid values from each binding change commmit", xuid2.equals(xuid3) );
      xs.close();
      xsystem.close();
   }
   
   /** 
    * Tests the XSystem create XStream method, which will fail with the ReferenceVIM.
    * @throws Exception
    */
   public void testCreateXStream() throws Exception
   {
      XSystem system = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( system, "testuser", "testpasswd" );
      XSet xs = system.createXSet(XSet.MODE_UNRESTRICTED);
      XStream strm = xs.createXStream("test", false, XAMLibrary.DEFAULT_MIME_TYPE );
      strm.abandon();
      strm.close();
      xs.close();
      system.close();
   }
   
   public void testJobMethods() throws Exception
   {
      String testProp = "testJobMethods." + System.currentTimeMillis();
      String qstring = "select \".xset.xuid\" where exists(\"" + testProp + "\")";
      
      XSystem system = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( system, "testuser", "testpasswd" );
      
      try {
      
         // Create a single test XSet
         XSet txs = system.createXSet(XSet.MODE_UNRESTRICTED);
         txs.createProperty( testProp, false, 1234);
         XUID testXuid = txs.commit();
      
         // Set up the query
         XSet xs = system.createXSet(XSet.MODE_UNRESTRICTED);
         xs.createProperty( XSet.XAM_JOB_COMMAND, false, XSet.XAM_JOB_QUERY );
         XStream qstream = xs.createXStream( XSet.XAM_JOB_QUERY_COMMAND, false, "test/plain;charset=utf-8" );
         qstream.write( qstring.getBytes("utf-8"));
         qstream.close();     
         xs.submitJob();
      
         // wait until job has started running
         int numTries = 0;
         String jobStatusString = xs.getString(XSet.XAM_JOB_STATUS);
         while (numTries < 250 && !jobIsDoneExecution(jobStatusString) ) {
            Thread.sleep(100);
            numTries++;
            jobStatusString = xs.getString(XSet.XAM_JOB_STATUS);
         }
      
         assertEquals( "Wrong number of results returned from query",
            1, xs.getLong(XSet.XAM_JOB_QUERY_RESULTS_COUNT) );
      
         // Resubmit, then halt it right away.
         xs.submitJob();
         xs.haltJob();
         jobStatusString = xs.getString(XSet.XAM_JOB_STATUS);
         assertEquals( "Wrong state on job after it was halted", XSet.XAM_JOB_STATUS_HALTED, jobStatusString );
      }
      finally {
         system.abandon();
         system.close();
      }
   }
   
   /**
    * Determines if the job has finished running. 
    * @return true if the job has finished running; false otherwise
    */
   private boolean jobIsDoneExecution(String status) {
       boolean jobDone = true;
       if (status.equals(XSet.XAM_JOB_STATUS_SHUTTING_DOWN) ||
          status.equals(XSet.XAM_JOB_STATUS_NEW) ||
          status.equals(XSet.XAM_JOB_STATUS_RUNNING) ||
          status.equals(XSet.XAM_JOB_STATUS_STARTING ) ) {
          jobDone = false;
       }
       return jobDone;
   }
   
   /**
    * Tests the path through to the XIterator methods for XSet.
    * @throws Exception
    */
   public void testIterator() throws Exception
   {
      XSystem xsys = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsys, "testuser", "testpasswd" );
      XSet xs = xsys.createXSet(XSet.MODE_UNRESTRICTED);
      XUID xuid = xs.commit();
      
      XIterator xi = xs.openFieldIterator("");
      int count = 0;
      while( xi.hasNext() )
      {
         String property = xi.next();
//         System.out.println( property );
         count ++;
      }
      xi.close();
      assertTrue( "Expected to read at least one field name from XSet XIterator ", count > 0 );
      
      xs.close();
      xsys.close();
   }
   
   /**
    * Tests the path through to the AccessPolicy method
    * @throws Exception
    */
   public void testAccessPolicy() throws Exception
   {
      XSystem xsys = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsys, "testuser", "testpasswd" );
      XSet xs = xsys.createXSet(XSet.MODE_UNRESTRICTED);
      
      xs.applyAccessPolicy( false, "testuser" );
      assertEquals( "Expected access policy to be set to testuser", "testuser", xs.getString( XSet.XAM_ACCESS_POLICY ));
      assertEquals( "Expected access policy to be unbound", false, xs.getFieldBinding( XSet.XAM_ACCESS_POLICY ));
      xs.applyAccessPolicy( true, "testuser" );
      assertEquals( "Expected access policy to be bound", true, xs.getFieldBinding( XSet.XAM_ACCESS_POLICY ));
      
      xs.resetAccessFields();
      try 
      {
         String access = xs.getString( XSet.XAM_ACCESS_POLICY );
         fail( "Should have failed because .xset.access.policy was removed" );
      } catch( FieldDoesNotExistException fdne )
      { /* ignored */ }
      
      xs.close();
      xsys.close();
   }
   
   /**
    * Tests the path through to the AccessPolicy method
    * @throws Exception
    */
   public void testBaseRetention() throws Exception
   {
      XSystem xsys = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsys, "testuser", "testpasswd" );
      XSet xs = xsys.createXSet(XSet.MODE_UNRESTRICTED);
      
      XIterator xi = xsys.openFieldIterator( XSystem.XAM_RETENTION_DURATION_POLICY_LIST );
      String policyPropertyName = xi.next();
      String policy = xsys.getString(policyPropertyName);
      xs.applyBaseRetentionPolicy( false, policy );
      
      assertEquals( "Wrong value for the base retention policy", policy, xs.getString(XSet.XAM_RETENTION_BASE_DURATION_POLICY));
      
      long oldDuration = xs.getActualRetentionDuration("base");
      xs.setBaseRetention(false, oldDuration+1);
      assertEquals( "Wrong value for base retention after setBaseRetention() method", oldDuration+1,
                    xs.getActualRetentionDuration("base"));
      xs.close();
      xsys.close();
   }
   
   /**
    * Tests the path through to the AccessPolicy method
    * @throws Exception
    */
   public void testManagement() throws Exception
   {
      XSystem xsys = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsys, "testuser", "testpasswd" );
      XSet xs = xsys.createXSet(XSet.MODE_UNRESTRICTED);

      String policy = xsys.getString( XSystem.XAM_MANAGEMENT_POLICY_DEFAULT );
      xs.applyManagementPolicy( false, policy );
      
      assertEquals( "Wrong value for the management policy", policy, xs.getString(XSet.XAM_MANAGEMENT_POLICY));
      
      xs.close();
      xsys.close();
   }
   
   /**
    * Tests the path through to the generic retention methods
    * @throws Exception
    */
   public void testRetention() throws Exception
   {
      final String RET_ID = "org.snia.xam.unit.test.testRetention";
      
      XSystem xsys = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsys, "testuser", "testpasswd" );
      XSet xs = xsys.createXSet(XSet.MODE_UNRESTRICTED);
      
      xs.createRetention( false, RET_ID );
      String test = xs.getString( XSet.XAM_RETENTION_LIST + RET_ID );
      assertEquals( "Wrong value for the retention list", RET_ID, test );
      
      XIterator xi = xsys.openFieldIterator( XSystem.XAM_RETENTION_DURATION_POLICY_LIST );
      String policyPropertyName = xi.next();
      String policy = xsys.getString(policyPropertyName);
      xi.close();
      xs.applyRetentionDurationPolicy( RET_ID, false, policy );
      assertEquals( "Wrong retention duration policy value", policy, 
                    xs.getString(".xset.retention." + RET_ID + ".duration.policy" ));
      long oldRetention = xs.getActualRetentionDuration( RET_ID );
      xs.setRetentionDuration( RET_ID, false, oldRetention+1 );
      assertEquals( "Wrong retention duration value", oldRetention+1, xs.getActualRetentionDuration(RET_ID));
      
      xi = xsys.openFieldIterator( XSystem.XAM_RETENTION_ENABLED_POLICY_LIST );
      policyPropertyName = xi.next();
      policy = xsys.getString(policyPropertyName);
      xi.close();
      xs.applyRetentionEnabledPolicy( RET_ID, false, policy );
      assertEquals( "Wrong retention enabled policy value", policy,
                    xs.getString(".xset.retention." + RET_ID + ".enabled.policy" ));
      boolean oldEnabled = xs.getActualRetentionEnabled(RET_ID);
      xs.setRetentionEnabledFlag(RET_ID, false, !oldEnabled);
      assertEquals( "Wrong retention enabled value", !oldEnabled, xs.getActualRetentionEnabled(RET_ID));
      
      xs.setRetentionEnabledFlag( RET_ID, false, true );
      xs.setRetentionStarttime( RET_ID, false );
      Calendar now = Calendar.getInstance();
      Calendar starttime = xs.getDateTime(".xset.retention." + RET_ID + ".starttime" );
      // Be a little permissive here, because of network lags, give it a couple of milliseconds
      // to return, and the value "now" will be a little later than the start time.
      long delta = now.getTimeInMillis() - starttime.getTimeInMillis();
      // The time check is quite liberal here to allow for very slow
      // machines running clients or servers (e.g. VMWare instances).
      assertEquals( "Starttime is not correct", true, (delta < 100) );
      
      // The following code might be useful for debugging.
      //
//      String now_str   = ISO8601Date.toISO8601String(now);
//      String start_str = ISO8601Date.toISO8601String(starttime);
//      System.out.println( now_str + " / " + start_str );
//      assertEquals( "Wrong starttime for retention event", now_str, start_str );
      
      policyPropertyName = ".xset.retention." + RET_ID + ".enabled.policy";
      xs.resetManagementFields();
      try
      {
         policy = xs.getString(policyPropertyName);
         System.out.println( policyPropertyName + " erroneously returned : " + policy );
         fail( "Expected an expection at this point, property should not exist: " + policyPropertyName );
      }
      catch( FieldDoesNotExistException fdne )
      {
         /* ignore */
      }

      xs.close();
      xsys.close();
   }
   
   /**
    * Tests the path through to the generic retention methods
    * @throws Exception
    */
   public void testDisposition() throws Exception
   {
      XSystem xsys = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsys, "testuser", "testpasswd" );
      XSet xs = xsys.createXSet(XSet.MODE_UNRESTRICTED);
      
      XIterator xi = xsys.openFieldIterator(XSystem.XAM_DELETION_AUTODELETE_POLICY_LIST);
      String policyPropertyName = xi.next();
      String policy = xsys.getString(policyPropertyName);
      xi.close();
      xs.applyAutoDeletePolicy(false, policy);
      assertEquals("Wrong autodelete policy on XSet", policy, xs.getString(XSet.XAM_DELETION_AUTODELETE_POLICY));
      boolean firstAutoDelete = xs.getActualAutoDelete();
      xs.setAutoDelete(false, !firstAutoDelete);
      assertEquals("Wrong explicit autodelete setting", !firstAutoDelete, xs.getActualAutoDelete());

      xi = xsys.openFieldIterator(XSystem.XAM_DELETION_SHRED_POLICY_LIST);
      policyPropertyName = xi.next();
      policy = xsys.getString(policyPropertyName);
      xi.close();
      xs.applyShredPolicy(false, policy);
      assertEquals("Wrong shred policy on XSet", policy, xs.getString(XSet.XAM_DELETION_SHRED_POLICY));
      boolean firstShred = xs.getActualShred();
      xs.setShred(false, !firstShred);
      assertEquals("Wrong explicit shred setting", !firstShred, xs.getActualShred());

      xs.close();
      xsys.close();
   }
   
   /**
    * Tests the path through to the generic retention methods
    * @throws Exception
    */
   public void testStoragePolicy() throws Exception
   {
      XSystem xsys = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsys, "testuser", "testpasswd" );
      XSet xs = xsys.createXSet(XSet.MODE_UNRESTRICTED);
      
      String policy = "tape";
      xs.applyStoragePolicy(false, policy);
      assertEquals("Wrong autodelete policy on XSet", policy, xs.getString(XSet.XAM_STORAGE_POLICY));

      xs.close();
      xsys.close();
   }
   
   public void testImportExport() throws Exception
   {
      final String PROP_NAME = "property.name";
      final String VALUE     = "Weather forcast for tonight and tomorrow is dark, followed by periods of light";
      XSystem xsys = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsys, "testuser", "testpasswd" );
      XSet xs = xsys.createXSet(XSet.MODE_UNRESTRICTED);
      xs.createProperty( PROP_NAME, false, VALUE );
      XUID xuid = xs.commit();

      XSet dest = xsys.createXSet(XSet.MODE_UNRESTRICTED);
      XStream expStream = xs.openExportXStream();
      XStream dstStream = dest.openImportXStream();
      
      byte buffer[] = new byte[1024];
      long count = 0;
      
      while( (count = expStream.read(buffer)) > 0 )
      {
         dstStream.write(buffer, count);
      }
      expStream.close();
      dstStream.close();
      
      String test = dest.getString(PROP_NAME);
      assertEquals( "Expected import process to work correctly", VALUE, test );
      
      dest.close();
      xs.close();
      xsys.close();
   }
   
   /**
    * Stand alone tests.
    */
   public static void main(String[] args)
   {
      System.exit(runTest(args, suite()));
   }
}
