/* Copyright (c) 2009, The Storage Networking Industry Association.

   Redistribution and use in source and binary forms, with or without modification, 
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.
   
   * Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.
   
   * Neither the name of The Storage Networking Industry Association (SNIA) nor 
   the names of its contributors may be used to endorse or promote products 
   derived from this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
   THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.snia.xam.vim.http;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.util.Calendar;

import org.snia.xam.InvalidOperationException;
import org.snia.xam.InvalidXRIException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XIterator;
import org.snia.xam.XSet;
import org.snia.xam.XStream;
import org.snia.xam.XSystem;
import org.snia.xam.XUID;
import org.snia.xam.base.DefaultXUID;
import org.snia.xam.util.SASLUtils;

import junit.framework.Test;
import junit.framework.TestResult;
import junit.framework.TestSuite;

public class HTTPXSystemTest
   extends XAMTest
{
   final String PROP_PREFIX   = "com.example.plist.";
   final String PROP_BOOL     = PROP_PREFIX + "bool";
   final String PROP_INT      = PROP_PREFIX + "int";
   final String PROP_DOUBLE   = PROP_PREFIX + "double";
   final String PROP_DATETIME = PROP_PREFIX + "datetime";
   final String PROP_STRING   = PROP_PREFIX + "string";
   final String PROP_XUID     = PROP_PREFIX + "xuid";
   
   public HTTPXSystemTest( String name )
   { 
      super(name);
   }
   
   public static int runTest(String[] args, Test test)
   {
       TestResult res = junit.textui.TestRunner.run(test);
       return Math.max(res.failureCount(), res.errorCount());
   }
   
  public static Test suite()
  {
     TestSuite suite = new TestSuite();
     suite.addTest(new HTTPXSystemTest("testConnect"));
     suite.addTest(new HTTPXSystemTest("testDefaultSASL"));
     suite.addTest(new HTTPXSystemTest("testPropertyCreateSetGet"));
     suite.addTest(new HTTPXSystemTest("testXSetCreation"));
     suite.addTest(new HTTPXSystemTest("testXSetManipulation"));
     suite.addTest(new HTTPXSystemTest("testAttributeFunctions"));
     suite.addTest(new HTTPXSystemTest("testCreateXStream"));
     suite.addTest(new HTTPXSystemTest("testIterator"));
     suite.addTest(new HTTPXSystemTest("testAccessXSet"));
     suite.addTest(new HTTPXSystemTest("testAccessTime"));
     suite.addTest(new HTTPXSystemTest("testHoldRelease"));
     suite.addTest(new HTTPXSystemTest("testIsXSetRetained"));
     suite.addTest(new HTTPXSystemTest("testXStream"));
     return suite;
  }

  /**
    * Test the ability to connect to a remove via the HTTP server.
    */
   public void testConnect() throws Exception
   {
      final String PROP = "com.example.test";
      
      XSystem system = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( system, "testuser", "testpasswd" );
      System.out.println( system.getString(XSystem.XAM_XSYSTEM_IDENTITY));
      system.createProperty( PROP, false, true );
      assertTrue( "Expected property to be true", system.getBoolean( PROP ));
      system.abandon();
      system.close();
   }
   
   /**
    * Test to see if reference VIM will work with default SASL PLAIN.
    */
   public void testDefaultSASL() throws Exception {
       XSystem xsys = s_xam.connect(s_xri);
       String creds = "\000testuser\000testpasswd";
       ByteArrayOutputStream baos = new ByteArrayOutputStream();
       
       xsys.authenticate(creds.getBytes(), baos );
       
       String result = new String(baos.toByteArray());
       assertEquals( "Expected DONE as reult from authtenticate", "DONE", result );
       xsys.close();
       xsys = null;
       baos.close();
       baos = null;
       
       // See if authentication can be done in a single buffer, with mechanism
       XSystem xsys2 = s_xam.connect(s_xri);
       String creds2 = "PLAIN\000testuser\000testpasswd";
       ByteArrayOutputStream baos2 = new ByteArrayOutputStream();
       
       xsys2.authenticate(creds2.getBytes(), baos2 );
       
       String result2 = new String(baos2.toByteArray());
       assertEquals( "Expected DONE as reult from authtenticate", "DONE", result2 );
       baos2.close();
       xsys2.abandon();
       xsys2.close();
   }

   public void testPropertyCreateSetGet() throws Exception
   {
      XSystem system = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( system, "testuser", "testpasswd" );
     
      system.createProperty(PROP_BOOL, false, false);
      assertFalse( "Expected bool prop to be false", system.getBoolean( PROP_BOOL ));
      system.setProperty( PROP_BOOL, true );
      assertTrue( "Expected bool prop to be true", system.getBoolean( PROP_BOOL ));
      
      system.createProperty( PROP_INT, false, 42 );
      assertEquals( PROP_INT, 42, system.getLong(PROP_INT ));
      system.setProperty( PROP_INT, 43 );
      assertEquals( PROP_INT, 43, system.getLong(PROP_INT ));
      
      system.createProperty( PROP_DOUBLE, false, -42.0 );
      assertEquals( PROP_DOUBLE, -42.0, system.getDouble(PROP_DOUBLE ));
      system.setProperty( PROP_DOUBLE, -43.0 );
      assertEquals( PROP_DOUBLE, -43.0, system.getDouble(PROP_DOUBLE ));
      
      system.createProperty( PROP_STRING, false, "foo" );
      assertEquals( PROP_STRING, "foo", system.getString(PROP_STRING ));
      system.setProperty( PROP_STRING, "bar" );
      assertEquals( PROP_STRING, "bar", system.getString(PROP_STRING));
      
      Calendar c = Calendar.getInstance();
      system.createProperty( PROP_DATETIME, false, c );
      assertEquals( PROP_DATETIME, c, system.getDateTime(PROP_DATETIME));
      try { Thread.sleep( 321 ); }
      catch( InterruptedException ie ) { ; } 
      Calendar c2 = Calendar.getInstance();
      system.setProperty( PROP_DATETIME, c2 );
      assertEquals( PROP_DATETIME, c2, system.getDateTime(PROP_DATETIME));
      
      XUID xuid = DefaultXUID.valueOf("AAA6AwA0ymgwMDUwNTZDMDAwMDgtMjAwOC0wNy0xOFQxMV81OF8zMl80ODMtMDhfMDAtYQ==" );
      system.createProperty( PROP_XUID, false, xuid );
      assertEquals( PROP_XUID, xuid, system.getXUID(PROP_XUID));
      XUID xuid2 = DefaultXUID.valueOf("AAAEcwA9419GVk1LQ05PNlRPQlE0ZTVJUTc1UzNNSlQ2Q0pHNDEzRjE1VkcySDFLNjFHMFFUTVRLOFNWNw==");
      system.setProperty( PROP_XUID, xuid2 );
      assertEquals( PROP_XUID, xuid2, system.getXUID(PROP_XUID));
      
      assertEquals( "XSystemContains field fails", true, system.containsField(PROP_STRING));
      
      system.deleteField( PROP_BOOL );
      system.deleteField( PROP_INT );
      system.deleteField( PROP_DOUBLE );
      system.deleteField( PROP_STRING );
      system.deleteField( PROP_DATETIME );
      system.deleteField( PROP_XUID );
      system.abandon();
      system.close();
   }
   
   /** Test XSet creation, copy, delete. */
   public void testXSetCreation() throws Exception
   {
      final String TEST_STR = "Raven, a raven! I see a raven! Run, Hide, Scream!";
      
      XSystem system = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( system, "testuser", "testpasswd" );
      
      XSet xs = system.createXSet(XSet.MODE_UNRESTRICTED);
      xs.createProperty( PROP_STRING, false, TEST_STR );
      XUID xuid = xs.commit();
      xs.close();
      
      XSet xs2 = system.openXSet(xuid, XSet.MODE_UNRESTRICTED);
      assertEquals( PROP_STRING, TEST_STR, xs2.getString(PROP_STRING));
      xs2.abandon();
      xs2.close();
      
      XSet xs3 = system.copyXSet(xuid, XSet.MODE_UNRESTRICTED );
      assertEquals( PROP_STRING, TEST_STR, xs3.getString(PROP_STRING));
      XUID xuid3 = xs3.commit();
      assertFalse( "Need new XUID value", xuid.equals(xuid3) );
      xs3.close();
      
      system.abandon();
      system.close();
   }
   
   public void testXSetManipulation() throws Exception
   {
      XSystem system = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( system, "testuser", "testpasswd" );
      
      XSet xs = system.createXSet(XSet.MODE_UNRESTRICTED);
      xs.createProperty( PROP_STRING, false, "for testXSetManipulation test" );
      XUID xuid = xs.commit();
      xs.close();
      
      system.deleteXSet(xuid);
      try {
         XSet failSet = system.openXSet( xuid, XSet.MODE_READ_ONLY );
         fail( "XSet should not have opened, it was to be deleted." );
      } catch( org.snia.xam.XSetInaccessibleException xie ) {
         /* Expected */
      }
      system.abandon();
      system.close();
   }
   
   public void testAttributeFunctions() throws Exception
   {
      XSystem system = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( system, "testuser", "testpasswd" );
      system.createProperty( PROP_BOOL, false, false );
      system.createProperty( PROP_INT,  false, 42 );
      system.createProperty( PROP_STRING, false, "12345" );
      
      assertEquals( "Wrong attr length on boolean prop", 1, system.getFieldLength(PROP_BOOL));
      assertEquals( "Wrong attr length on int prop", 8, system.getFieldLength(PROP_INT));
      assertEquals( "Wrong attr length on string prop", 5, system.getFieldLength(PROP_STRING));
      
      assertEquals( "Wrong binding on boolean prop", false, system.getFieldBinding(PROP_BOOL));
      
      assertEquals( "Wrong readonly on boolean prop", false, system.getFieldReadOnly(PROP_BOOL));
      assertEquals( "Wrong readonly on identiy prop", true, system.getFieldReadOnly(XSystem.XAM_XSYSTEM_IDENTITY));
      
      assertEquals( "Wrong type on bool prop", XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, system.getFieldType(PROP_BOOL));
      
      try 
      {
         system.setFieldAsBinding(PROP_BOOL);
         fail( "Expected XSystem.setFieldAsBinding to fail with exception" );
      }
      catch( InvalidOperationException ioe )
      {
         /* ignored */
      }
      
      try 
      {
         system.setFieldAsNonbinding(PROP_BOOL);
         fail( "Expected XSystem.setFieldAsBinding to fail with exception" );
      }
      catch( InvalidOperationException ioe )
      {
         /* ignored */
      }
      system.abandon();
      system.close();
   }
   
   /** 
    * Tests the XSystem create XStream method, which will fail with the ReferenceVIM.
    * @throws Exception
    */
   public void testCreateXStream() throws Exception
   {
      XSystem system = s_xam.connect(s_xri);
      try
      {
         XStream strm = system.createXStream("test", false, XAMLibrary.DEFAULT_MIME_TYPE );
         fail( "Should have thrown an exception from XSystem" );
      }
      catch( Exception /*InvalidOperationException*/ ioe )
      {
         // Acceptable for this test.
      }
      system.abandon();
      system.close();
   }
   
   /**
    * Tests the path through to the XIterator methods for XSystem.
    * @throws Exception
    */
   public void testIterator() throws Exception
   {
      XSystem xsys = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( xsys, "testuser", "testpasswd" );
      
      XIterator xi = xsys.openFieldIterator("");
      int count = 0;
      while( xi.hasNext() )
      {
         String property = xi.next();
//         System.out.println( property );
         count ++;
      }
      xi.close();
      assertTrue( "Expected to read at least one field name from XSystem XIterator ", count > 0 );
      xsys.abandon();
      xsys.close();
   }
   
   /** Test XSet access method */
   public void testAccessXSet() throws Exception
   {
      XSystem system = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( system, "testuser", "testpasswd" );
      
      XSet xs = system.createXSet(XSet.MODE_UNRESTRICTED);
      XUID xuid = xs.commit();
      xs.close();
      
      boolean result = system.accessXSet( xuid, XSystem.ACCESS_READ_OK );
      assertTrue( "Expected access READ_OK to return true", result );
      system.abandon();
      system.close();
   }

   /** Test XSet access method */
   public void testAccessTime() throws Exception
   {
      final long DELAY = 100;
      
      XSystem system = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( system, "testuser", "testpasswd" );
      
      XSet xs = system.createXSet(XSet.MODE_UNRESTRICTED);
      XUID xuid = xs.commit();
      
      Calendar time1 = system.getXSetAccessTime(xuid);
      
      try { Thread.sleep(DELAY); } catch( InterruptedException ie ) { ; } 
      xs.commit();
      
      Calendar time2 = system.getXSetAccessTime(xuid);
      
      assertTrue( "Expected second access time to be greater than first access time", 
                   time2.getTimeInMillis() > time1.getTimeInMillis() );
      
      long deltaTime = time2.getTimeInMillis() - time1.getTimeInMillis();
      assertTrue( "Expected time difference between two access time to be at least " + DELAY + " ms ",
                  deltaTime > DELAY );
      system.abandon();
      system.close();
   }

   /** Test Hold/Release XSet method */
   public void testHoldRelease() throws Exception
   {
      final long DELAY = 100;
      
      XSystem system = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( system, "testuser", "testpasswd" );
      
      XSet xs = system.createXSet(XSet.MODE_UNRESTRICTED);
      XUID xuid = xs.commit();
      system.holdXSet( xuid, "UnitTest" );
      system.releaseXSet( xuid, "UnitTest" );
      
      xs.close();
      system.deleteXSet(xuid); // Hygenic Unit Testing
      system.abandon();
      system.close();
   }

   /** Test Hold/Release XSet method */
   public void testIsXSetRetained() throws Exception
   {
      XSystem system = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( system, "testuser", "testpasswd" );
      
      XSet xs = system.createXSet(XSet.MODE_UNRESTRICTED);
      XUID xuid = xs.commit();
      xs.close();
      assertTrue( "Expected XSystem.isXSetRetained to return false", ! system.isXSetRetained(xuid));
      system.deleteXSet(xuid); // Hygenic Unit Testing
      system.abandon();
      system.close();
   }
   
   /**
    * Test the openXStream method
    */
   public void testXStream() throws Exception
   {
      XSystem system = s_xam.connect(s_xri);
      SASLUtils.authenticatePlain( system, "testuser", "testpasswd" );
      
      try {
         try { 
            XStream test = system.openXStream( "does not exist", XStream.MODE_READ_ONLY );
            fail( "Expected open XStream to fail" );
         } catch( InvalidOperationException ioe ) {
            /* ignore */
         }
      
         try { 
            XStream test = system.createXStream( "should fail", false, XAMLibrary.DEFAULT_MIME_TYPE );
            fail( "Expected create XStream to fail" );
         } catch( InvalidOperationException ioe ) {
            /* ignore */
         }
      }
      finally {
         system.abandon();
         system.close();
      }
   }

   /**
    * Stand alone tests.
    */
   public static void main(String[] args)
   {
      System.exit(runTest(args, suite()));
   }

}
