/******************************************************************************
 *
 * Copyright (c) 2007-2008, The Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *    - Redistributions of source code must retain the above copyright notice, 
 *      this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright 
 *      notice, this list of conditions and the following disclaimer in the 
 *      documentation and/or other materials provided with the distribution.
 *    - Neither the name of The Storage Networking Industry Association (SNIA)
 *      nor the names of its contributors may be used to endorse or promote 
 *      products derived from this software without specific prior written 
 *      permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/
package org.snia.xam;

import java.util.Calendar;

/**
 * Details methods common to all first class XAM Objects which have Property
 * Field associated with them.
 */
public interface FieldContainer
{
    // == XIterator ==
    /**
     * Iterate over a set of fields belonging to the XamSystem object. These
     * iterators yield String objects, the name of the field selected. The
     * filter pattern is a simple regular expression limited to a prefix of
     * character literals. To find all fields beginning with "com.example" make
     * the call openFieldIterator( "com.example" );
     * 
     * @param prefix
     *            A prefix string describing the fields to be returned by the
     *            iterator. The expression is simple, a series of byte values
     *            which are treated as a "prefix" to match. For example
     *            "com.example" will match all fields starting with
     *            "com.example".
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidFieldNameException Prefix is null or malformed.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public XIterator openFieldIterator(String prefix) 
       throws AuthenticationExpiredException, InvalidFieldNameException,
              ObjectInUseException, XSetAbandonException, 
              XSetCorruptException, XSystemAbandonException,
              XSystemCorruptException, XAMException;
    
    // == Field Interogation ==
    /**
     * Tests to see if the specified field exists in this FieldContainer.
     * 
     * @param fieldName The name of the field to test for existence.
     * @return TRUE if the field exists in this FieldContainer, FALSE otherwise.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public boolean containsField( String fieldName )
       throws AuthenticationExpiredException, InvalidFieldNameException,
              ObjectInUseException, XSetAbandonException, 
              XSetCorruptException, XSystemAbandonException,
              XSystemCorruptException, XAMException;
    
    // == Property creation ==
    /**
     * Create a XAM Boolean valued property. XAM Boolean is <i>boolean</i> in
     * Java.
     * 
     * @param fieldName
     *            The name of this property
     * @param binding
     *            True if this property is to be BOUND.
     * @param value
     *            The boolean value for this property
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidOperationException An attempt to create a bound field on a XAMLibrary or XSystem.
     * @throws FieldExistsException The specified field already exists.
     * @throws MaximumFieldException Too many fields in the FieldContainer.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void createProperty(String fieldName, boolean binding, boolean value)
       throws AuthenticationExpiredException, InvalidFieldNameException,
              FieldExistsException,
              InvalidOperationException, MaximumFieldException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSystemCorruptException,
              XAMException;

    /**
     * Create a XAM Int valued property. XAM Int is <i>long</i> in Java.
     * 
     * @param fieldName
     *            The name of this property
     * @param binding
     *            True if this property is to be BOUND.
     * @param value
     *            The boolean value for this property
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidOperationException An attempt to create a bound field on a XAMLibrary or XSystem.
     * @throws FieldExistsException The specified field already exists.
     * @throws MaximumFieldException Too many fields in the FieldContainer.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void createProperty(String fieldName, boolean binding, long value)
       throws AuthenticationExpiredException,
              FieldExistsException, 
              InvalidOperationException, MaximumFieldException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSystemCorruptException,
             XAMException;

    /**
     * Create a XAM Float valued property. XAM Float is <i>double</i> in Java
     * 
     * @param fieldName
     *            The name of this property
     * @param binding
     *            True if this property is to be BOUND.
     * @param value
     *            The double value for this property
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidOperationException An attempt to create a bound field on a XAMLibrary or XSystem.
     * @throws FieldExistsException The specified field already exists.
     * @throws MaximumFieldException Too many fields in the FieldContainer.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void createProperty(String fieldName, boolean binding, double value)
       throws AuthenticationExpiredException, InvalidFieldNameException,
              FieldExistsException,
              InvalidOperationException, MaximumFieldException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSystemCorruptException,
              XAMException;

    /**
     * Create a XUID valued property.
     * 
     * @param fieldName
     *            The name of the property
     * @param binding
     *            True if the property is to be BOUND
     * @param value
     *            The XUID value of the property
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException value is null or malformed.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidOperationException An attempt to create a bound field on a XAMLibrary or XSystem.
     * @throws FieldExistsException The specified field already exists.
     * @throws MaximumFieldException Too many fields in the FieldContainer.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void createProperty(String fieldName, boolean binding, XUID value)
       throws AuthenticationExpiredException, InvalidFieldNameException,
              FieldExistsException, InvalidArgumentException,
              InvalidOperationException, MaximumFieldException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSystemCorruptException,
              XAMException;

    /**
     * Create a XAM String valued property. Note that Java strings may be much
     * longer than XAM strings. If the Java string is longer than the XAM MAX
     * String, an exception will be thrown.
     * 
     * @param fieldName
     *            The name of the property
     * @param binding
     *            True if the property is to be BOUND.
     * @param value
     *            The String value of the property
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException value is null or malformed.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidOperationException An attempt to create a bound field on a XAMLibrary or XSystem.
     * @throws FieldExistsException The specified field already exists.
     * @throws MaximumFieldException Too many fields in the FieldContainer.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void createProperty(String fieldName, boolean binding, String value)
       throws AuthenticationExpiredException, InvalidFieldNameException,
              FieldExistsException, InvalidArgumentException,
              InvalidOperationException, MaximumFieldException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSystemCorruptException,
              XAMException;

    /**
     * Create a XAM Date valued property. XAM Date is <i>java.util.Calendar</i>
     * in Java.
     * 
     * @param fieldName
     *            The name of the property.
     * @param binding
     *            True if the property is is to be BOUND.
     * @param value
     *            The date and time value of the property
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException value is null or malformed.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidOperationException An attempt to create a bound field on a XAMLibrary or XSystem.
     * @throws FieldExistsException The specified field already exists.
     * @throws MaximumFieldException Too many fields in the FieldContainer.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void createProperty(String fieldName, boolean binding, Calendar value)
       throws AuthenticationExpiredException, InvalidFieldNameException,
              FieldExistsException, InvalidArgumentException,
              InvalidOperationException, MaximumFieldException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSystemCorruptException,
              XAMException;

    // == Property value mutators (setters) ==
    /**
     * Set the named field to the specified xam_boolean (Java boolean)
     * 
     * @param fieldName
     *            The name of the field to set.
     * @param value
     *            The boolean value to set
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidFieldTypeException The field is not a xam_boolean stype.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void setProperty(String fieldName, boolean value) 
       throws AuthenticationExpiredException, InvalidFieldNameException,
              InvalidFieldTypeException, FieldDoesNotExistException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSystemCorruptException,
              XAMException;
    
    /**
     * Set the named field to the specified xam_datetime (Java Calendar)
     * 
     * @param fieldName
     *            The name of the field to set.
     * @param value
     *            The Date/Time value to set
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException value is null or malformed.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidFieldTypeException The field is not a xam_datetime stype.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void setProperty(String fieldName, Calendar value)
       throws AuthenticationExpiredException, FieldDoesNotExistException, InvalidArgumentException, 
              InvalidFieldNameException, InvalidFieldTypeException,
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XSystemAbandonException,
              XSystemCorruptException, XAMException;
    
    /**
     * Set the named field to the specified xam_float (Java double)
     * 
     * @param fieldName
     *            The name of the field to set
     * @param value
     *            The XAM Float value to set
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidFieldTypeException The field is not a xam_double stype.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void setProperty(String fieldName, double value) 
       throws AuthenticationExpiredException, FieldDoesNotExistException, InvalidFieldNameException,
              InvalidFieldTypeException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSystemCorruptException,
              XAMException;
    
    /**
     * Set the named field to the specified xam_int (Java long)
     * 
     * @param fieldName
     *            The name of the field to set
     * @param value
     *            The XAM Int value to set
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidFieldTypeException The field is not a xam_int stype.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void setProperty(String fieldName, long value) 
       throws AuthenticationExpiredException, FieldDoesNotExistException, InvalidFieldNameException,
              InvalidFieldTypeException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSystemCorruptException,
              XAMException;
    
    /**
     * Set the named field to the specified xam_string (Java String)
     * 
     * @param fieldName
     *            The name of the field to set
     * @param value
     *            The String value to set. String length must not exceed the
     *            maximum XAM String limit.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws InvalidArgumentException value is null or malformed.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidFieldTypeException The field is not a xam_string stype.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void setProperty(String fieldName, String value) 
       throws AuthenticationExpiredException, FieldDoesNotExistException, InvalidArgumentException, 
              InvalidFieldNameException, InvalidFieldTypeException,
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XSystemAbandonException,
              XSystemCorruptException, XAMException;
    
    /**
     * Set the named field to the specified xam_xuit (XUID)
     * 
     * @param fieldName
     *            The name of the field to set
     * @param value
     *            The XUID value to set.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws InvalidArgumentException value is null or malformed.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidFieldTypeException The field is not a xam_xuid stype.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void setProperty(String fieldName, XUID value) 
       throws AuthenticationExpiredException, FieldDoesNotExistException, InvalidArgumentException, 
              InvalidFieldNameException, InvalidFieldTypeException,
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XSystemAbandonException,
              XSystemCorruptException, XAMException;
    
    // == Property value accessors (getters) ==
    /**
     * Get the boolean value of the named field.
     * 
     * @param fieldName
     *            The name of the property to get the value
     * @return The boolean value of the named field.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidFieldTypeException The field is not a xam_boolean stype.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public boolean getBoolean(String fieldName) 
       throws AuthenticationExpiredException, FieldDoesNotExistException, InvalidFieldNameException, 
              InvalidFieldTypeException, ObjectInUseException, 
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSetCorruptException, 
              XAMException;
    
    /**
     * Get the DateTime value of the named field.
     * 
     * @param fieldName
     *            The name of the property to get the value
     * @return The Calendar instance of the named field.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidFieldTypeException The field is not a xam_datetime stype.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public Calendar getDateTime(String fieldName) 
       throws AuthenticationExpiredException, FieldDoesNotExistException, InvalidFieldNameException, 
              InvalidFieldTypeException, ObjectInUseException, 
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSetCorruptException, 
              XAMException;
    
    /**
     * Get the XAM Float value of the named field.
     * 
     * @param fieldName
     *            The name of the property to get the value
     * @return The Java double value of the named field.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidFieldTypeException The field is not a xam_double stype.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public double getDouble(String fieldName) 
       throws AuthenticationExpiredException, FieldDoesNotExistException, InvalidFieldNameException, 
              InvalidFieldTypeException, ObjectInUseException, 
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSetCorruptException, 
              XAMException;
    
    /**
     * Get the XAM Int value of the named field.
     * 
     * @param fieldName
     *            The name of the property to get the value
     * @return The Java long value of the named field.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidFieldTypeException The field is not a xam_long stype.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public long getLong(String fieldName) 
       throws AuthenticationExpiredException, FieldDoesNotExistException, InvalidFieldNameException, 
              InvalidFieldTypeException, ObjectInUseException, 
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSetCorruptException, 
              XAMException;
    
    /**
     * Get the XAM String value of the named field.
     * 
     * @param fieldName
     *            The name of the property to get the value
     * @return The Java String value of the named field.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidFieldTypeException The field is not a xam_string stype.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public String getString(String fieldName) 
       throws AuthenticationExpiredException, FieldDoesNotExistException, InvalidFieldNameException, 
              InvalidFieldTypeException, ObjectInUseException, 
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSetCorruptException, 
              XAMException;
    
    /**
     * Get the XUID value of the named field.
     * 
     * @param fieldName
     *            The name of the property to get the value
     * @return The XUID value of the named field.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidFieldTypeException The field is not a xam_xuid stype.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public XUID getXUID(String fieldName)
       throws AuthenticationExpiredException, FieldDoesNotExistException, InvalidFieldNameException, 
              InvalidFieldTypeException, ObjectInUseException, 
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSetCorruptException, 
             XAMException;

    // == XStream Creation ==
    /**
     * Create a new XStream valued field in the XSet. The new stream will be
     * opened in WRITE-Truncate mode, with a zero length value.
     * 
     * @param fieldName
     *            The name of the new XStream.
     * @param binding
     *            True if this field is to be BOUND.
     * @param mimeType
     *            The mime type of the new XStream.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException mimeType is null or malformed.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidOperationException An attempt to create a bound field on a XAMLibrary or XSystem.
     * @throws FieldExistsException The specified field already exists.
     * @throws MaximumFieldException Too many fields in the FieldContainer.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public XStream createXStream(String fieldName, boolean binding, String mimeType)
       throws AuthenticationExpiredException, InvalidArgumentException,
              InvalidFieldNameException, InvalidOperationException,
              FieldExistsException, MaximumFieldException,
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XSystemAbandonException,
              XSystemCorruptException, XAMException;
    
    // == XStream Mutator/Accessor (setter/getter) ==
    /**
     * Open the named XStream for read or write. If the XStream has not been
     * opened, the data manipulation methods will throw an exception.
     * 
     * @param fieldName
     * @param mode
     * @return The XStream containing the data
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public XStream openXStream(String fieldName, String mode)
       throws AuthenticationExpiredException, InvalidFieldNameException,
              FieldDoesNotExistException, ObjectInUseException, 
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSystemCorruptException,
              XAMException;
    
    // == Field Operations (XStreams or Properties) ==
    /**
     * Return the MIME type of the specified field.
     * 
     * @param fieldName
     *            Name of the field from which to get the type
     * @return The string value of the field's type
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public String getFieldType(String fieldName) 
       throws AuthenticationExpiredException, InvalidFieldNameException,
              FieldDoesNotExistException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSystemCorruptException,
              XAMException;
    
    /**
     * Return the length of the field, in bytes.
     * 
     * @param fieldName
     *            Name of the field from which to get the length
     * @return The number of bytes for the length of the field
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public long getFieldLength(String fieldName) 
       throws AuthenticationExpiredException, InvalidFieldNameException,
              FieldDoesNotExistException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSystemCorruptException,
              XAMException;

    /**
     * Return TRUE if the Field is Bound, FALSE otherwise.
     * 
     * @param fieldName
     *            Name of the field from which to get the binding info
     * @return The boolean value of the binding mode of the field.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public boolean getFieldBinding(String fieldName) 
       throws AuthenticationExpiredException, InvalidFieldNameException,
              FieldDoesNotExistException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSystemCorruptException,
              XAMException;

    /**
     * Return TRUE if the Read-Only attributeof the Field is TRUE, return FALSE
     * otherwise.
     * 
     * @param fieldName
     *            Name of the field from which to get the read-only info
     * @return The boolean value of the read-only mode of the property
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public boolean getFieldReadOnly(String fieldName) 
       throws AuthenticationExpiredException, InvalidFieldNameException,
              FieldDoesNotExistException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSystemCorruptException,
              XAMException;

    /**
     * Removes the named field from the FieldContainer. This method applies only
     * to XSets.
     * 
     * @param fieldName
     *            The name of the Field to remove
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void deleteField(String fieldName) 
       throws AuthenticationExpiredException, InvalidFieldNameException,
              FieldDoesNotExistException, ObjectInUseException, 
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSystemCorruptException,
              XAMException; 
    
    /**
     * Set the Field Binding attribute to TRUE. This method will only work on
     * Fields contained in an XSet.
     * 
     * @param fieldName
     *            Name of the field to set the binding value
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws InvalidOperationException An attempt to create a bound field on a XAMLibrary or XSystem.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void setFieldAsBinding(String fieldName) 
       throws AuthenticationExpiredException, InvalidFieldNameException,
              InvalidOperationException, FieldDoesNotExistException,
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XSystemAbandonException, 
              XSystemCorruptException, XAMException;

    /**
     * Set the Field Binding attribute to FALSE. This method will only work on
     * Fields contained in an XSet.
     * 
     * @param fieldName
     *            Name of the field to set the binding value
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void setFieldAsNonbinding(String fieldName) 
       throws AuthenticationExpiredException, InvalidFieldNameException,
              FieldDoesNotExistException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException,
              XSystemAbandonException, XSystemCorruptException,
              XAMException;
    
    // == Asynchronous Operations ==
    /**
     * This method will begin the asynchronous opening of XStream in either 
     * readonly, writeonly or appendonly mode, based on the mode argument. 
     * The specified callback will be invoked as part of the asynchronous
     * opening. To monitor the status of this operation, the application 
     * can poll the Async instance that is generated by this method. 
     * A handle to an Async instance is also passed to any provided callback method
     * when that callback method is invoked.
     * @param name The name of the stream to be opened.
     * @param mode The mode to open the stream, one of XStream.MODE_READ_ONLY, XStream.MODE_WRITE_TRUNCATE, XStream.MODE_WRITE_APPEND.
     * @param xopid The application supplied XOPID of this XAsync operation.
     * @param listener listener An optional parameter, may be null, of the XASyncListener object to be notified when the operation has completed.
     * @return XAsync object representing this operation
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException mode is null, xopid is already in use..
     * @throws InvalidFieldNameException fieldName is null or malformed.
     * @throws FieldDoesNotExistException The specified field does not exist.
     * @throws ObjectInUseException The XSet has open import/export streams.
     * @throws XSetAbandonException The XSet is in the abandoned state and may only be closed.
     * @throws XSetCorruptException The XSet is in the corrupt state and may only be abandoned.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe, the XAsyncListener must be coded in a thread safe manner
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public XAsync asyncOpenXStream( String         name,
                                    String         mode,
                                    long           xopid,
                                    XAsyncListener listener )
       throws AuthenticationExpiredException, InvalidArgumentException,
              InvalidFieldNameException, FieldDoesNotExistException,
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XSystemAbandonException,
              XSystemCorruptException, XAMException;
}


