/******************************************************************************
 *
 * Copyright (c) 2007-2008, The Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *    - Redistributions of source code must retain the above copyright notice, 
 *      this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright 
 *      notice, this list of conditions and the following disclaimer in the 
 *      documentation and/or other materials provided with the distribution.
 *    - Neither the name of The Storage Networking Industry Association (SNIA)
 *      nor the names of its contributors may be used to endorse or promote 
 *      products derived from this software without specific prior written 
 *      permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

package org.snia.xam;

/**
 * An XamSystem represents the Xam library running locally. An application
 * wanting to use the XAM library must create an instance of the XAM library.
 * This library object will manage and create instances of VIMs in order to
 * connect to XAM Storage Systems. 
 */
public interface XAMLibrary extends FieldContainer
{
    /** The MIME Type of XAM_BOOLEAN STypes. */
    public final static String STYPE_BOOLEAN_MIME_TYPE = "application/vnd.snia.xam.boolean";

    /** The MIME Type of XAM_INT STypes. */
    public final static String STYPE_INT_MIME_TYPE = "application/vnd.snia.xam.int";

    /** The MIME Type of XAM_FLOAT STypes. */
    public final static String STYPE_DOUBLE_MIME_TYPE = "application/vnd.snia.xam.double";

    /** The MIME Type of XAM_STRING STypes. */
    public final static String STYPE_STRING_MIME_TYPE = "application/vnd.snia.xam.string";

    /** The MIME Type of XAM_DATETIME STypes. */
    public final static String STYPE_DATETIME_MIME_TYPE = "application/vnd.snia.xam.datetime";

    /** The MIME Type of XAM_XUID STypes. */
    public final static String STYPE_XUID_MIME_TYPE = "application/vnd.snia.xam.xuid";
    
    /** Plain text MIME type */
    public final static String TEXT_PLAIN_MIME_TYPE = "text/plain";
        
    /** MIME type of the query result stream */
    public final static String XUID_LIST_MIME_TYPE = "application/vnd.snia.query.xuid_list";

    /** XAM XRI strings must begin with this protocol designator */
    public final static String XAM_PROTOCOL = "snia-xam://";

    /** This character separates the VIM name from the XAM Storage System Name */
    public final static String XAM_VIM_LIBRARY_TOKEN = "!";

    /**
     * Applications may use this MIME type for data not defined by any other
     * appropriate MIME type
     */
    public static final String DEFAULT_MIME_TYPE = "application/octet-stream";

    // == XAM Library Properties ==
    /**
     * This XAM_STRING property indicates the origin of the XAM Library. It is
     * intended for debugging and applications should not code specific behavior
     * with respect to this value.
     */
    public final static String XAM_IDENTITY = ".xam.identity";
    
    /**
     * This XAM_STRING property is used to indicate the current level of normal
     * library logging. The higher the value, the more detail is logged.
     * Applications may set this value to control the log.
     */
    public final static String XAM_LOG_LEVEL = ".xam.log.level";

    /**
     * This XAM_STRING property is used to indicate the current level of debug
     * library logging. The higher the value, the more detail is logged.
     * Applications may set this value to control the log.
     */
    public final static String XAM_LOG_VERBOSITY = ".xam.log.verbosity";
    
    /**
     * This XAM_STRING property is used to indicate where the XAM Library's 
     * log file is to be found. The log file is a locally named file running within
     * the same scope as the XAM Library (i.e. not on a different computer)
     */
    public final static String XAM_LOG_PATH = ".xam.log.path";

    /**
     * This XAM_INT property indicates how many logfiles the XAMLibrary instance will 
     * maintain.
     */
    public static final String XAM_LOG_MAX_ROLLOVERS = ".xam.log.max.rollovers";
    
    /**
     * This XAM_INT property indicates how approximate maximum size a log file is allowed to 
     * grow before the XAMLibrary instance will close it and open a new one. The size
     * of the logfile is measured in kB.
     */
    public static final String XAM_LOG_MAX_SIZE      = ".xam.log.max.size";
    
    /**
     * This XAM_BOOLEAN property indicates if the XAMLibrary instance shall begin by 
     * appending to the existing logfile, instead of opening a new log file.
     */
    public static final String XAM_LOG_APPEND        = ".xam.log.append";      

    /**
     * This XAM_STRING property is used to indicate which version of the XAM
     * API is implemented.
     */
    public final static String XAM_API_LEVEL = ".xam.apiLevel";

    /**
     * This is a prefix for the properties, in a XAM Library, listing the names
     * of the available VIMs.
     */
    public final static String XAM_VIM_LIST = ".xam.vim.list";

    // Useful constants.
    /** The maximum number of characters allowed in a XAM_STRING field. */
    public final static int XAM_MAX_STRING = 512;
    
    public final static long XAM_LOG_ALL   = 5;
    public final static long XAM_LOG_INFO  = 4;
    public final static long XAM_LOG_WARN  = 3;
    public final static long XAM_LOG_ERROR = 2;
    public final static long XAM_LOG_FATAL = 1;
    public final static long XAM_LOG_NONE  = 0;

    /**
     * Connect to the specified XAM archive and return am XSystem object to
     * allow the application to use the XAM API. XRI values are in the form
     * "snia-xam://[vimname!]xsystemname[?param=value[{&param=value}]]" VIM
     * names, and parameters are optional. Use of optional parameters is
     * specific to any XAM Storage System and its configuration. NOTE: A
     * connected XSystem instance is not fully usable until authenticated.
     * 
     * @param XRI
     *            The XAM Storage System Resource specification
     * @return The XSystem object representing the connection
     * @throws InvalidXRIException The XRI is null or malformed.
     * @throws ConnectException A connection could not be made.
     * @throws VIMLoadException A VIM could not be found to handle this connection.
     * @throws XAMException Another error exists with the underlying XAM Storage System.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public XSystem connect(String XRI) 
       throws InvalidXRIException, ConnectException, VIMLoadException, XAMException;
}

