/******************************************************************************
 *
 * Copyright (c) 2007-2008, The Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *    - Redistributions of source code must retain the above copyright notice, 
 *      this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright 
 *      notice, this list of conditions and the following disclaimer in the 
 *      documentation and/or other materials provided with the distribution.
 *    - Neither the name of The Storage Networking Industry Association (SNIA)
 *      nor the names of its contributors may be used to endorse or promote 
 *      products derived from this software without specific prior written 
 *      permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

package org.snia.xam;

/** The XAsynch object represents an asynchronous XAM operations. The XAsynch object
 * contains the state information regarding the asynchronous operation.
 * Asynchronous operations are in one of two states: pending and completed. 
 * When the operation is first initiated, it is in the pending state. 
 * Because the operation has not completed, it is only possible to query
 * whether the operation has completed, retrieve the XOPID that was specified 
 * when the operation was initiated, and to halt the operation. 
 */
public interface XAsync
{
    /**
     * This method stops the execution of the operation associated with the XAsync 
     * instance. This method may be used at any time. If this method causes the XAsync instance
     * to halt before it is complete, any get method invoked on that XAsync will throw an AsyncHaltedException.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public void halt() 
        throws AuthenticationExpiredException, XAMException;
    
    /**
     * This method retrieves the completed state of the operation associated with the 
     * XAsync instance. It may be used at any time.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public boolean isComplete() 
        throws AuthenticationExpiredException, XAMException;
    /**
     * This method retrieves the XOPID of the operation associated with the 
     * XAsync instance. This method may be used at any time.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws AsyncHaltedException This operation was programmatically halted and results are not available.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public long getXOPID() 
        throws AuthenticationExpiredException, AsyncHaltedException, XAMException;
    
    /**
     * This method retrieves the XAM Status of the operation associated with the XAsync instance. 
     * This method may be used after the operation has transitioned to the completed state. If no
     * error has occurred, the method returns. If an error occurred, the exception generated will be 
     * thrown. 
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws AsyncHaltedException This operation was programmatically halted and results are not available.
     * @throws XAMException Any exception possibly generated by the asynchronous operation.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public void getStatus() 
        throws AuthenticationExpiredException, AsyncHaltedException, XAMException;
    
    /**
     * This method retrieves the XSet reference associated with this XAsync instance.
     * This is only valid if the operation was an "asyncOpenXSet" or "asyncCopyXSet".
     * @return The XSet reference from the XAsync.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws AsyncHaltedException This operation was programmatically halted and results are not available.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
   public XSet getXSet() 
        throws AuthenticationExpiredException, AsyncHaltedException, XAMException;
    
   /**
    * This method retrieves the XStream reference associated with this XAsync instance.
    * This is only valid if the operation was an "asyncOpenXStream".
    * This method may be used at any time.
    * @return The XStream reference from the XAsync.
    * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
    * @throws AsyncHaltedException This operation was programmatically halted and results are not available.
    * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
    *             <DT><b>Thread Safety:</b>
    *             <DD>This method is thread safe
    *             <DT><b>Blocking:</b>
    *             <DD>This method will return immediately
    */
  public XStream getXStream() 
       throws AuthenticationExpiredException, AsyncHaltedException, XAMException;
   
   /**
    * This method retrieves the XUID reference (if any) of the XSet associated with this
    * XAsync instance. This method may be used at any time.
    * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
    * @throws AsyncHaltedException This operation was programmatically halted and results are not available.
    * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
    *             <DT><b>Thread Safety:</b>
    *             <DD>This method is thread safe
    *             <DT><b>Blocking:</b>
    *             <DD>This method will return immediately
    */
    public XUID getXUID() 
        throws AuthenticationExpiredException, AsyncHaltedException, XAMException;
    
    /**
     * This method retrieves the number of bytes read by the operation associated
     * with this XAsync instance. Not all operations read bytes, and for those operations
     * it will always be set to zero. This method may be used at any time.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws AsyncHaltedException This operation was programmatically halted and results are not available.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public long getBytesRead() throws AuthenticationExpiredException, AsyncHaltedException, XAMException;
    
    /**
     * This method retrieves the number of bytes written by the operation associated
     * with this XAsync instance. Not all operations write bytes, and for those 
     * operations it will always be set to zero. This method may be used at any time.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws AsyncHaltedException This operation was programmatically halted and results are not available.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public long getBytesWritten() throws AuthenticationExpiredException, AsyncHaltedException, XAMException;
    
    /**
     * This method releases the resources of the operation associated with the 
     * XAsync instance and resources of the XAsync operation itself. 
     * This method may be used after the operation has transitioned to the completed state.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public void close() 
        throws AuthenticationExpiredException, XAMException;
}
