/******************************************************************************
 *
 * Copyright (c) 2007-2008, The Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *    - Redistributions of source code must retain the above copyright notice, 
 *      this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright 
 *      notice, this list of conditions and the following disclaimer in the 
 *      documentation and/or other materials provided with the distribution.
 *    - Neither the name of The Storage Networking Industry Association (SNIA)
 *      nor the names of its contributors may be used to endorse or promote 
 *      products derived from this software without specific prior written 
 *      permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

package org.snia.xam;

/**
 * XStreams are the interface to the content portions of a XAM storage system.
 * XStream implementations are provided by a storage vendor in the the vendor's
 * packages. XStream instances perform the storage vendor specific functionality
 * to interact with that storage vendor's XAM Storage System.
 */
public interface XStream
{
    /** Specify an offset seek() from the beginning of the XStream. */
    public static final int SEEK_SET = 0;

    /** Specify an offset seek() from the current cursor of the XStream. */
    public static final int SEEK_CUR = 1;

    /** Specify an offset seek() from the end of the XStream. */
    public static final int SEEK_END = 2;

    /** Open mode read only. Used in openXStream(). */
    public final static String MODE_READ_ONLY = "readonly";

    /** Open mode write only, truncate contents. Used in openXStream(). */
    public final static String MODE_WRITE_TRUNCATE = "writeonly";

    /** Open mode write only, append to contents. Used in openXStream(). */
    public final static String MODE_WRITE_APPEND = "appendonly";
    
    /** End of File value returned from a read() method when the end of XStream has been reached. */
    public final static long EOF = -1;

    /**
     * Returns the position of the position indicator in the XStream
     * 
     * @return The cursor position
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws XStreamAbandonException The XStream is in the abandoned state and may only be closed.
     * @throws XStreamCorruptException The XSstream is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public long tell() 
       throws AuthenticationExpiredException, XStreamAbandonException,
              XStreamCorruptException, XAMException;
    
    /**
     * Seeks to a specific location within the XStream
     * 
     * @param offset
     *            The number bytes of offset from the cursor position (whence)
     * @param whence
     *            One of SEEK_CUR, SEEK_SET, SEEK_END
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException the value of <i>whence</i> is invalid.
     * @throws XStreamAbandonException The XStream is in the abandoned state and may only be closed.
     * @throws XStreamCorruptException The XSstream is in the corrupt state and may only be abandoned.
     * @throws XStreamException - Attempt to seek outside boundaries of the stream.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void seek( long offset, long whence ) 
       throws AuthenticationExpiredException, InvalidArgumentException,
              XStreamAbandonException, XStreamCorruptException,
              XStreamException, XAMException;
    
    /** Transfer byte values to the XStream. This method will not block if a 
     * partial write occurred. Multiple calls to XStream.write may be required to
     * completely transfer the buffer.
     * 
     * @param buffer 
     *          The array of bytes to transfer.
     * @return The number of bytes transfered.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException buffer is null, or size mismatch between arguments.
     * @throws XSetUnderHoldException The XSet is being held and may not be changed.
     * @throws XStreamAbandonException The XStream is in the abandoned state and may only be closed.
     * @throws XStreamCorruptException The XSstream is in the corrupt state and may only be abandoned.
     * @throws XStreamException - Maximum stream size has been reached.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method does not block until all data is written. It
     *             returns the number of bytes written. Subsequent calls to this
     *             method may be required to write the buffer.
     */
    public long write( byte buffer[] ) 
       throws AuthenticationExpiredException, InvalidArgumentException, 
              XStreamAbandonException, XStreamCorruptException,
              XStreamException, XSetUnderHoldException, XAMException;
    
    /** Transfer byte values to the XStream. This method will not block if a 
     * partial write occurred. Multiple calls to XStream.write may be required to
     * completely transfer the buffer.
     * 
     * @param buffer 
     *          The array of bytes to transfer.
     * @param bytesToWrite
     *          The number of bytes to transfer, from the beginning of the buffer.
     * @return The number of bytes transfered.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException buffer is null, or size mismatch between arguments.
     * @throws XSetUnderHoldException The XSet is being held and may not be changed.
     * @throws XStreamAbandonException The XStream is in the abandoned state and may only be closed.
     * @throws XStreamCorruptException The XSstream is in the corrupt state and may only be abandoned.
     * @throws XStreamException - Maximum stream size has been reached.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method does not block until all data is written. It
     *             returns the number of bytes written. Subsequent calls to this
     *             method may be required to write the buffer.
     */
    public long write( byte buffer[], long bytesToWrite ) 
       throws AuthenticationExpiredException, InvalidArgumentException,
              XStreamAbandonException, XStreamCorruptException, 
              XStreamException, XSetUnderHoldException, XAMException;
    
    /**
     * Transfer byte values to the XStream. This method will not block if a
     * partial write occurred. Multiple calls to this method may be required to
     * completely transfer the buffer.
     * 
     * @param buffer
     *            The array of bytes to transfer
     * @param offset
     *            The byte offset in the buffer from which to start writing
     * @param bytesToWrite
     *            The number of bytes transfered.
     * @return The number of bytes transfered to the stream.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException buffer is null, or size mismatch between arguments.
     * @throws XSetUnderHoldException The XSet is being held and may not be changed.
     * @throws XStreamAbandonException The XStream is in the abandoned state and may only be closed.
     * @throws XStreamCorruptException The XSstream is in the corrupt state and may only be abandoned.
     * @throws XStreamException - Maximum stream size has been reached.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method does not block until all data is written. It
     *             returns the number of bytes written. Subsequent calls to this
     *             method may be required to write the buffer.
     */
    public long write( byte buffer[], long offset, long bytesToWrite )
       throws AuthenticationExpiredException, InvalidArgumentException, 
              XStreamAbandonException, XStreamCorruptException, 
              XStreamException, XSetUnderHoldException, XAMException;    

    /**
     * Transfer byte values from the XStream to buffer. If there is not enough
     * data left in the stream to fill the buffer, that data is copied to the
     * buffer, leaving the remaining bytes unchanged. This method does not block
     * if there are insufficient bytes to fill the buffer. Subsequent calls may
     * be required to read the remaining data in the XStream.
     * 
     * @param buffer
     *            An array of byte values to read
     * @param offset
     *            The byte offset in the buffer to start placing bytes read from
     *            the stream
     * @param bytesToRead
     *            The maximum number of bytes to read. The value will be less that or equal to bytesToRead. When there is no more data to be read, a value of -1 will be returned.
     * @return The actual number of bytes read
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException buffer is null, or size mismatch between arguments.
     * @throws XStreamAbandonException The XStream is in the abandoned state and may only be closed.
     * @throws XStreamCorruptException The XSstream is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method does not block until all data is read. It
     *             returns the number of bytes read. Subsequent calls to this
     *             method may be required to read the remaining data.
     */
    public long read(byte buffer[], long offset, long bytesToRead)
       throws AuthenticationExpiredException, InvalidArgumentException,
              XStreamAbandonException, XStreamCorruptException, 
              XAMException;
    
    /**
     * Transfer byte values from the XStream to buffer. If there is not enough
     * data left in the stream to fill the buffer, that data is copied to the
     * buffer, leaving the remaining bytes unchanged. This method does not block
     * if there are insufficient bytes to fill the buffer. Subsequent calls may
     * be required to read the remaining data in the XStream.
     * 
     * @param buffer
     *            An array of byte values to read
     * @param bytesToRead
     *            The maximum number of bytes to read. The value will be less that or equal to bytesToRead. When there is no more data to be read, a value of -1 will be returned.
     * @return The actual number of bytes read
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException buffer is null, or size mismatch between arguments.
     * @throws XStreamAbandonException The XStream is in the abandoned state and may only be closed.
     * @throws XStreamCorruptException The XSstream is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method does not block until all data is read. It
     *             retusnt the number of bytes read. Subsequent calls to this
     *             method may be required to read the remaining data.
     */
    public long read(byte buffer[], long bytesToRead)
       throws AuthenticationExpiredException, InvalidArgumentException,
              XStreamAbandonException, XStreamCorruptException,
              XAMException;
    
    /**
     * Transfer byte values from the XStream to buffer. If there is not enough
     * data left in the stream to fill the buffer, that data is copied to the
     * buffer, leaving the remaining bytes unchanged. This method does not block
     * if there are insufficient bytes to fill the buffer. Subsequent calls may
     * be required to read the remaining data in the XStream.
     * 
     * @param buffer
     *            An array of byte values to read
     * @return The maximum number of bytes to read. The value will be less that or equal to the buffer length. When there is no more data to be read, a value of -1 will be returned.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException buffer is null, or size mismatch between arguments.
     * @throws XStreamAbandonException The XStream is in the abandoned state and may only be closed.
     * @throws XStreamCorruptException The XSstream is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method does not block until all data is read. It
     *             returns the number of bytes read. Subsequent calls to this
     *             method may be required to read the remaining data.
     */
    public long read(byte buffer[] )
       throws AuthenticationExpiredException, InvalidArgumentException,
              XStreamAbandonException, XStreamCorruptException,
              XAMException;
    
   /**
     * Closes the XStream, making this XStream instance unavailable for further
     * use. The XStream needs to be reopened to access this the stream data.
     * 
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws XSetUnderHoldException The XSet is being held and may not be changed.
     * @throws XStreamAbandonException The XStream is in the abandoned state and may only be closed.
     * @throws XStreamCorruptException The XSstream is in the corrupt state and may only be abandoned.
     * @throws XStreamException The stream was not open.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void close() 
       throws AuthenticationExpiredException, XSetUnderHoldException,
              XStreamCorruptException, XStreamException, XAMException;
    
    /**
     * Abandon (lose) all pending work on this XStream allowing it to be closed.
     * This method is to be used when the XStream is in the corrupt state and
     * the application needs to perform error recovery and clean up. <b>WARNING:
     * Calling abandon() may result in data loss if used inappropriately.</b>
     * 
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws XStreamAbandonException The XStream is in the abandoned state and may only be closed.
     * @throws XStreamCorruptException The XSstream is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void abandon() throws AuthenticationExpiredException, XAMException;
    
    // == Asynchronous Operations ==
    /**
     * This method will begin the asynchronous transfer of data from the storage 
     * system into the target buffer, up to the number of bytes requested. 
     * The specified callback will be invoked as part of the asynchronous
     * transfer. To monitor the status of this operation, the application 
     * can poll the Async instance that is generated by this method. 
     * A handle to an Async instance is also passed to any provided callback method
     * when that callback method is invoked.
     * @param buffer The byte buffer into which the date is read.
     * @param bufferLength The number of bytes to read. This must be less than or equal to the buffer length.
     * @param xopid The application supplied XOPID of this XAsync operation.
     * @param listener listener An optional parameter, may be null, of the XASyncListener object to be notified when the operation has completed.
     * @return XAsync object representing this operation
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException buffer is null, or xopid is already in use.
     * @throws XStreamAbandonException The XStream is in the abandoned state and may only be closed.
     * @throws XStreamCorruptException The XSstream is in the corrupt state and may only be abandoned.
     * @throws XStreamException The stream is open in the wrong mode (write only).
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe, the XAsyncListener must be coded in a thread safe manner
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public XAsync asyncRead( byte           buffer[],
                             long           bufferLength,
                             long           xopid,
                             XAsyncListener listener )
       throws AuthenticationExpiredException, InvalidArgumentException,
              XStreamAbandonException, XStreamCorruptException, 
              XStreamException, XAMException;   
    
    /**
     * This method will begin the asynchronous transfer of data from the source
     * buffer to the XAM Storage System, up to the number of bytes requested. 
     * The specified callback will be invoked as part of the asynchronous transfer. 
     * To monitor the status of this operation, the application can poll the Async instance
     * that is generated by this method. A handle to an Async instance is also passed 
     * to any provided callback method when that callback method is invoked.<br>
     * Note: This method may fail with an error if the maximum number of bytes supported 
     * in an XStream is reached. To determine the actual maximum number of bytes allowed 
     * in an XStream, an application should evaluate the '.xsystem.limits.maxSizeOfXStream' 
     * field on the XSystem instance. For more information on this topic, please consult [XAM-ARCH].
     * @param buffer A byte array containing the data to be written to the stream.
     * @param bufferLength The number of bytes to write from the buffer. This value must be less than or equal to the actual buffer length.
     * @param xopid The application supplied XOPID of this XAsync operation.
     * @param listener listener An optional parameter, may be null, of the XASyncListener object to be notified when the operation has completed.
     * @return XAsync object representing this operation
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException buffer is null, or xopid is already in use.
     * @throws XStreamAbandonException The XStream is in the abandoned state and may only be closed.
     * @throws XStreamCorruptException The XSstream is in the corrupt state and may only be abandoned.
     * @throws XStreamException The stream is open in the wrong mode (read only).
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe, the XAsyncListener must be coded in a thread safe manner
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public XAsync asyncWrite( byte           buffer[],
                              long           bufferLength,
                              long           xopid,
                              XAsyncListener listener )
       throws AuthenticationExpiredException, InvalidArgumentException,
              XStreamAbandonException, XStreamCorruptException, 
              XStreamException, XAMException;   
    
    /**
     * This method will begin the asynchronous closing of a previously opened XStream. 
     * Any resources that were allocated can be released at this point. The specified 
     * callback will be invoked as part of the asynchronous close. To monitor the status 
     * of this operation, the application can poll the Async instance that is generated 
     * by this method. A handle to an Async instance is also passed to any provided callback
     * method when that callback method is invoked.
     * @param xopid The application supplied XOPID of this XAsync operation.
     * @param listener listener An optional parameter, may be null, of the XASyncListener object to be notified when the operation has completed.
     * @return XAsync object representing this operation
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException <i>xopid</i> is already in use.
     * @throws XStreamAbandonException The XStream is in the abandoned state and may only be closed.
     * @throws XStreamCorruptException The XSstream is in the corrupt state and may only be abandoned.
     * @throws XStreamException The Xstream is not open.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe, the XAsyncListener must be coded in a thread safe manner
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public XAsync asyncClose( long           xopid,
                              XAsyncListener listener )
       throws AuthenticationExpiredException, InvalidArgumentException,
              XStreamCorruptException, XStreamException, XAMException;   
}
