/******************************************************************************
 *
 * Copyright (c) 2007-2008, The Storage Networking Industry Association
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *    - Redistributions of source code must retain the above copyright notice, 
 *      this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright 
 *      notice, this list of conditions and the following disclaimer in the 
 *      documentation and/or other materials provided with the distribution.
 *    - Neither the name of The Storage Networking Industry Association (SNIA)
 *      nor the names of its contributors may be used to endorse or promote 
 *      products derived from this software without specific prior written 
 *      permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

package org.snia.xam;

import java.util.Calendar;
import java.io.ByteArrayOutputStream;

/**
 * Represents a connection to a XAM Storage System. An XSystem instance has been
 * implemented in the storage vendor's packages. Storage vendor specific XSystem
 * instances will perform their specified interactions with XAM Storage Systems
 * as well as acting as factory sources for XSet and XStream objects (via
 * create/open XSet or XStream).
 */
public interface XSystem extends FieldContainer
{
    // The following bit fields are defined for accessXSet(). These values must 
    // be consistent with all XAM API language bindings because these values will
    // be passed into VIMS.
    
    /** Read access is allowed, used in AccessXSet() */
    public final static long ACCESS_READ_OK             = 0x80000000L;

    /** Write access is allowed, used in AccessXSet() */
    public final static long ACCESS_WRITE_APPLICATION_OK = 0x40000000L;

    /** Write access to system fields is allowed, used in AccessXSet() */
    public final static long ACCESS_WRITE_SYSTEM_OK     = 0x20000000L;
    
    /** Creating XSet is allowed, not used for AccessXSet(), defined as auth granule */
    public final static long ACCESS_CREATE_OK           = 0x10000000;
    
    /** Delete access is allowed, used in AccessXSet() */
    public final static long ACCESS_DELETE_OK           = 0x08000000L;
    
    /** Hold/Release operation on an XSet is allowed, used in AccessXSet() */
    public final static long ACCESS_HOLD_OK             = 0x04000000L;
    
    /** Event based retention operation on an XSet is OK, used in AccessXSet() */
    public final static long ACCESS_RETENTION_EVENT_OK  = 0x02000000L;
    
    /** Submit of a job is allowed, used in AccessXSet() */
    public final static long ACCESS_JOB_OK              = 0x01000000L;
    
    /** Commit of a job is allowed, used in AccessXSet() */
    public final static long ACCESS_JOB_COMMIT_OK       = 0x00800000L;

    /** Returned from the authenticate call if authentication is complete 
     * @since 1.0.1 
     */
    public final int XAM_SASL_COMPLETE   = 0;

    /** Returned from the authenticate call if more information is required 
     * @since 1.0.1
     */
    public final int XAM_SASL_IN_PROCESS = 1;

    // == XSystem Properties ==
    /**
     * This XAM String property has the vendor identity of the connected XSystem
     * instance
     */
    public final String XAM_XSYSTEM_IDENTITY = ".xsystem.identity";
    
    /**
     * Indicates if the XSystem is in the intialization state. Initially set to TRUE, 
     * then when the XAM Library is done with the initialization protocol, the XAM Library 
     * will set this false. At that point, the XSystem will delete this field.
     */
    public final String XAM_XSYSTEM_INITIALIZING = ".xsystem.initializing";

    /**
     * This XAM Date property contains the current time of the connected XSystem
     * instance. This value is advisory for application.
     */
    public final String XAM_XSYSTEM_TIME = ".xsystem.time";

    /**
     * This XAM Int property (Java long) indicates the maximum number of Fields
     * that may be created in an XSet
     */
    public final String XAM_XSYSTEM_LIMITS_MAX_FIELDS = ".xsystem.limits.maxFieldsPerXSet";

    /**
     * This XAM Int Property (Java long) indicates the maximum size of an
     * XStream
     */
    public final String XAM_XSYSTEM_LIMITS_MAX_STREAM_SIZE = ".xsystem.limits.maxSizeOfXStream";
    
    /**
     * This Property name prefix may be used to get the list of SASL supported
     * authentication mechanisms supported by the connected XSystem.
     */
    public final String XAM_XSYSTEM_AUTH_SASL_LIST = ".xsystem.auth.SASLmechanism.list.";

    /**
     * This XAM String property has the name of default SASL mechanism for the
     * connected XSystem.
     */
    public final String XAM_XSYSTEM_AUTH_SASL_DEFAULT = ".xsystem.auth.SASLmechanism.default";

    /** Prefix of the auth granule list, to be used in an XIterator pattern */
    public final String XAM_XSYSTEM_AUTH_GRANULE_LIST = ".xsystem.auth.granule.list.";

    public final String XAM_XSYSTEM_AUTH_IDENTITY_AUTHENTICATION = ".xsystem.auth.identity.authentication";

    public final String XAM_XSYSTEM_AUTH_IDENTITY_AUTHORIZATION = ".xsystem.auth.identity.authorization";

    public final String XAM_XSYSTEM_AUTH_EXPIRATION = ".xsystem.auth.expiration";
    
    /** This boolean field indicates if the system supports XSet access policy */
    public final String XAM_XSYSTEM_ACCESS = ".xsystem.access";
    
    /**
     * This field name prefix is used to create XIterators interrogating the XSystem
     * for allowable Access Policy names.
     */
    public final String XAM_XSYSTEM_ACCESS_POLICY_LIST = ".xsystem.access.policy.list";

    public final String XAM_XSYSTEM_JOB_COMMIT_SUPPORTED = ".xsystem.job.commit.supported";

    /** Prefix of the job list, to be used in an XIterator pattern */
    public final String XAM_XSYSTEM_JOB_LIST = ".xsystem.job.list.";

    public final String XAM_XSYSTEM_JOB_LIST_QUERY = ".xsystem.job.list.xam.job.query";

    public final String XAM_XSYSTEM_JOB_QUERY_CONTINUANCE_SUPPORTED = ".xsystem.job.xam.job.query.continuance.supported";

    public final String XAM_XSYSTEM_JOB_QUERY_LEVEL1_SUPPORTED = ".xsystem.job.xam.job.query.level1.supported";

    public final String XAM_XSYSTEM_JOB_QUERY_LEVEL2_SUPPORTED = ".xsystem.job.xam.job.query.level2.supported";

    /**
     * Prefix of the retention policy list, to be used in an XIterator
     * pattern
     */
    public final String XAM_RETENTION_DURATION_POLICY_LIST = ".xsystem.retention.duration.policy.list.";

    /**
     * Prefix of the event retention enabled policy list, to be used in an
     * XIterator pattern
     */
    public final String XAM_RETENTION_ENABLED_POLICY_LIST = ".xsystem.retention.enabled.policy.list.";

    public final String XAM_DELETION_AUTODELETE = ".xsystem.deletion.autodelete";
    
    /** Prefix of the autodelete policy list, to be used in an XIterator pattern */
    public final String XAM_DELETION_AUTODELETE_POLICY_LIST = ".xsystem.deletion.autodelete.policy.list.";

    public final String XAM_DELETION_SHRED = ".xsystem.deletion.shred";

    /** Prefix of the shred policy list, to be used in an XIterator pattern */
    public final String XAM_DELETION_SHRED_POLICY_LIST = ".xsystem.deletion.shred.policy.list.";

    /** Prefix of the storage policy list, to be used in an XIterator pattern */
    public final String XAM_STORAGE_POLICY_LIST = ".xsystem.storage.policy.list.";

    /** Prefix of the management policy list, to be used in an XIterator pattern */
    public final String XAM_MANAGEMENT_POLICY_LIST = ".xsystem.management.policy.list.";

    public final String XAM_MANAGEMENT_POLICY_DEFAULT = ".xsystem.management.policy.default";

    // == XAM Storage System Management
    
    /**
     * The connect method is called by the VIM after the XSystem has been created and 
     * initialized. This connect method operates identically to the specified method 
     * for XAMLibrary.connect() but shall not be  not called by the application. 
     * This method shall be called only once by the XAM Library, and shall throw an exception
     * if called more than once. 
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     */
    
    public void connect( String xri ) throws XAMException;
    
    /**
     * This method will allow an application to authenticate an XSystem
     * instance. It provides a generic interface to exchange data as part of the
     * authentication process. The application must check the XSystem
     * properties, prefixed constant value XAM_XSYSTEM_SASL_LIST, to determine
     * which patterns of authentication are available for use. After a pattern
     * is selected, the appropriate sequence of data exchanges should be made
     * (using this call) in order to authenticate. A failed authentication will
     * make the XSystem Instance unusable. APplication cannot repeat failed
     * authentications using the same XSystem instance.
     * 
     * @param buffer
     *            Data being passed to the authentication system. The entire buffer is
     *            used by the authentication process.
     * @return Output from the authentication method.
     * @throws AuthenticationException The authentication process encountered an error.
     * @throws InvalidArgumentException The buffer argument is null.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     *
     * @deprecated
     */
    public byte[] authenticate(byte[] buffer) 
       throws AuthenticationException, InvalidArgumentException, XAMException;

    /**
     * This method will allow an application to authenticate an XSystem
     * instance. It provides a generic interface to exchange data as part of the
     * authentication process. The application must check the XSystem
     * properties, prefixed constant value XAM_XSYSTEM_SASL_LIST, to determine
     * which patterns of authentication are available for use. After a pattern
     * is selected, the appropriate sequence of data exchanges should be made
     * (using this call) in order to authenticate. A failed authentication will
     * make the XSystem Instance unusable. APplication cannot repeat failed
     * authentications using the same XSystem instance.
     * 
     * @param buffer
     *            Data being passed to the authentication system. The entire buffer is
     *            used by the authentication process.
     * @param response 
     *            The response from the XSystem for this call to authenticate.
     *            Previous contents of the ByteArrayOutputStream are ignored.
     *
     * @return A status code indicating if more information is required for 
     *         the authentication process. A value, either XAM_SASL_COMPLETE or 
     *         XAM_SASL_IN_PROCESS shall be returned.
     *
     * @throws AuthenticationException The authentication process encountered an error.
     * @throws InvalidArgumentException The buffer argument is null.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     *
     * @since 1.0.1
     */
    public int authenticate( byte[] buffer, ByteArrayOutputStream response ) 
       throws AuthenticationException, InvalidArgumentException, XAMException;

    /**
     * Closes this XSystem instance, preventing any further changes. Any
     * resources associated with this XSystem instance will be released. This
     * method blocks until complete.
     * 
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws ObjectInUseException There are open XStreams on the XSystem preventing the close method from completing.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void close() 
       throws AuthenticationExpiredException, ObjectInUseException, XAMException;

    /**
     * Abandons all changes to this XSsystem. No unsaved data is written, and
     * all open XSets are abandoned, allowing close to be called. <b>WARNING:
     * Inappropriate use of this call can and will result in data loss.
     * Applications are encouraged to track open XSets and close them properly
     * instead of using this method.</b>
     * 
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void abandon() throws AuthenticationExpiredException, XAMException;

    /**
     * Create a XUID from the base64 string.
     * @param inBase64XUID - The base 64 string representing the XUID
     * @throws AuthenticationExpiredException authentication process encountered an error.
     * @throws InvalidXUIDException The base64 string was invalid, the decoded XUID failed the CRC check, or the XUID value was malformed.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public XUID createXUID( String inBase64XUID ) 
       throws AuthenticationExpiredException, InvalidXUIDException, XAMException;

    /**
     * Create a XUID from a byte buffer.
     * @param inBinaryXUID - The binary buffer representing the XUID
     * @throws AuthenticationExpiredException authentication process encountered an error.
     * @throws InvalidXUIDException The XUID value was malformed or failed a CRC check.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public XUID createXUID( byte inBinaryXUID[] ) 
       throws AuthenticationExpiredException, InvalidXUIDException, XAMException;

    // == XSet management ==
    /**
     * Delete the specified XSet without opening it.
     * 
     * @param xsetid
     *            XUID of the XSet to be deleted.
     * @throws AuthorizationException The delete operation is not allowed.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws XSetUnderRetentionException The XSet may not be deleted because of retention.
     * @throws XSetUnderHoldException The XSet may not be deleted because it is under hold.
     * @throws InvalidArgumentException The XUID is null.
     * @throws XSetInaccessibleException XSet does not exist or is not accessible; it cannot be deleted.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void deleteXSet(XUID xsetid)     
       throws AuthorizationException, AuthenticationExpiredException,
              XSetUnderRetentionException, XSetUnderHoldException,
              InvalidArgumentException, XSetInaccessibleException,
              XSystemAbandonException, XSystemCorruptException,
              XAMException;

    /**
     * This method evaluates all retention criteria that exists on the XSet, specified as a XUID.
     * A non-fatal exception shall be thrown if the specified XUID is improperly formatted, 
     * or does not exist in the XSystem, or if the application is not authorized to access the XSet.
     * 
     * @param xsetid The XUID of the XSet to evaluate.
     * @return TRUE if the XSet is under retention, FALSE otherwise.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException The XUID is null.
     * @throws XSetInaccessibleException XSet does not exist or cannot be accessed.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public boolean isXSetRetained( XUID xsetid ) 
       throws AuthenticationExpiredException, InvalidArgumentException, 
              XSetInaccessibleException, XSystemAbandonException, 
              XSystemCorruptException, XAMException;
    
    /**
     * Puts the specified XSet on hold. This prevents applications from making
     * changes to the XSet. An XSet may participate in in multiple holds. The
     * HoldId is specified by the application.
     * 
     * @param xsetid
     *            The XSet to hold.
     * @param holdID
     *            The ID of the HOLD operation.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws AuthorizationException The application is not authorized to place a hold.
     * @throws InvalidArgumentException xsetid or holdid arguments are null.
     * @throws InvalidFieldNameException The field name for the hold is too long or malformed.
     * @throws HoldIdException The hold id is already in use
     * @throws XSetInaccessibleException XSet does not exist or is not accessible.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void holdXSet(XUID xsetid, String holdID) 
       throws AuthenticationExpiredException, AuthorizationException,
              InvalidFieldNameException, HoldIdException, 
              InvalidArgumentException, XSetInaccessibleException, 
              XSystemAbandonException, XSystemCorruptException, 
              XAMException;
    /**
     * Releases XSet from a specific hold, making it subject to normal retention
     * processing.
     * 
     * @param xsetid
     *            The XSet to be released.
     * @param holdID
     *            The ID of the corresponding HOLD operation.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws AuthorizationException The application is not authorized to place a hold.
     * @throws HoldIdException The hold id is not in use on this XSet.
     * @throws InvalidArgumentException xsetid or holdid arguments are null.
     * @throws XSetInaccessibleException XSet does not exist or is not accessible.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public void releaseXSet(XUID xsetid, String holdID) 
       throws AuthenticationExpiredException, AuthorizationException,
              HoldIdException, InvalidArgumentException, XSetInaccessibleException, 
              XSystemAbandonException, XSystemCorruptException, 
              XAMException;
    /**
     * Tests if a specified XSet is accessible according to the modes specified.
     * 
     * @param xsetid
     *            The XSet to test.
     * @param mode
     *            Bitset of modes (ACCESS_READ_OK, ACCESS_WRITE_APPLICATION_OK, ACCESS_WRITE_SYSTEM_OK,
     *                             ACCESS_DELETE_OK, ACCESS_HOLD_OK, ACCESS_RETENTION_EVENT_OK,
     *                             ACCESS_JOB_OK, ACCESS_JOB_COMMIT_OK )
     * @return TRUE if the application can access the XSet, FALSE otherwise.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException xsetid or holdid arguments are null.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public boolean accessXSet(XUID xsetid, long mode) 
       throws AuthenticationExpiredException, InvalidArgumentException, 
              XSystemAbandonException, XSystemCorruptException,
              XAMException;
    
    /**
     * Retrieves the XSet's Last Access Time, most recent of either time of last
     * open or commit. This method will not alter the access time of the XSet.
     * 
     * @param xsetid
     *            The XUID of the XSet
     * @return The last access time of the XSet.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException xsetid or holdid arguments are null.
     * @throws XSetInaccessibleException XSet does not exist or is not accessible.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public Calendar getXSetAccessTime(XUID xsetid) 
       throws AuthenticationExpiredException, InvalidArgumentException, 
              XSetInaccessibleException, XSystemAbandonException, 
              XAMException;

    /**
     * Create a new, unsaved XSet, in this XSystem. The open XSet will initially
     * be fully writeable, but will be in the specified mode after the first
     * commit.
     * 
     * @param createMode
     *            The mode the XSet should be in after the first commit. One of
     *            XSET_MODE_RESTRICTED or XSET_MODE_UNRESTRICTED.
     * @return The new XSet object.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws AuthorizationException The application is not authorized to create an XSet.
     * @throws InvalidArgumentException createMode is either null or an invalid mode value.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             Any error during creation of the XSet.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public XSet createXSet(String createMode) 
       throws AuthenticationExpiredException, AuthorizationException, 
              InvalidArgumentException, XSystemAbandonException,
              XSystemCorruptException, XAMException;
    
    /**
     * Open the specified XSet for access by the application.
     * 
     * @param xsetid
     *            The XUID of the XSet to open
     * @param openMode
     *            One of XSET_MODE_READ_ONLY, XSET_MODE_RESTRICTED,
     *            XSET_MODE_UNRESTRICTED
     * 
     * @return The XSet object to be opened.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application
     * @throws AuthorizationException The application is not authorized to create an XSet.
     * @throws InvalidArgumentException xsetid or openMode are null, or the mode is invalid.
     * @throws InvalidXSetModeException The specified mode is not recognized by the XSystem.
     * @throws XSetInaccessibleException XSet does not exist or is not accessible. must reauthenticate.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public XSet openXSet(XUID xsetid, String openMode) 
       throws AuthenticationExpiredException, AuthorizationException, 
              InvalidArgumentException, InvalidXSetModeException, XSetInaccessibleException, 
              XSystemAbandonException, XSystemCorruptException,
              XAMException;
    
    /**
     * Copy the specified XSet to a new XSet. The new XSet will be "open" in the
     * mode specified.
     * 
     * @param xsetid
     *            The source XSet to copy.
     * @param copyMode
     *            One of XSET_MODE_RESTRICTED or XSET_MODE_UNRESTRICED
     * @return The newly copied XSet.
     * @throws AuthorizationException The application is not authorized to perform the copy.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException xsetid or copyMode are null, or copyMode is invalid.
     * @throws InvalidXSetModeException The specified mode is not recognized by the XSystem.
     * @throws XSetInaccessibleException The specified XSet does not exist or is not accessible.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe
     *             <DT><b>Blocking:</b>
     *             <DD>This method blocks until completion
     */
    public XSet copyXSet(XUID xsetid, String copyMode) 
       throws AuthenticationExpiredException, AuthorizationException, 
              InvalidArgumentException, InvalidXSetModeException, XSetInaccessibleException, 
              XSystemAbandonException, XSystemCorruptException,
              XAMException;
    
    // == Asynchronous Operations ==
    /**
     * This method will begin the asynchronous opening of an XSet in the 
     * XSystem, ultimately returning a handle to an XSet instance associated 
     * with the XSystem. The specified callback will be invoked as part of the
     * asynchronous opening. To monitor the status of this operation, the 
     * application can poll the Async instance that is generated by this 
     * method. A handle to an Async instance is also passed to any provided 
     * callback method when that callback method is invoked.
     * @param inXuid The XUID of the XSet to be opened.
     * @param mode The mode the XSet should be opened in. One of XSet.MODE_READ_ONLY, XSet.MODE_RESTRICTED, XSet.MODE_UNRESTRICTED
     * @param xopid The application XOPID of this XAsync operation
     * @param listener An optional parameter, may be null, of the XASyncListener object to be notified when the operation has completed.
     * @return XAsync object representing this operation.
     * @throws AuthorizationException The application is not authorized to open XSets.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException inXuid or mode are null, or the Xopid may already be in use.
     * @throws XSetInaccessibleException The specified XSet does not exist or is not accessible.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe, the XAsyncListener must be coded in a thread safe manner
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public XAsync asyncOpenXSet( XUID           inXuid,
                                 String         mode,
                                 long           xopid,
                                 XAsyncListener listener )
       throws AuthorizationException, AuthenticationExpiredException, 
              InvalidArgumentException, XSetInaccessibleException, 
              XSystemAbandonException, XSystemCorruptException, 
              XAMException;
    
    /**
     * This method will begin the asynchronous copying of an XSet in the XSystem, 
     * ultimately returning a handle to an XSet instance associated with the XSystem. 
     * The specified callback will be invoked as part of the asynchronous copying. 
     * To monitor the status of this operation, the application can poll the Async instance 
     * that is generated by this method. A handle to an Async instance is also passed to any 
     * provided callback method when that callback method is invoked.
     * @param inXuid The XUID of the XSet to copy.
     * @param mode The mode in which the XSet will be opened in. One of XSet.MODE_UNRESTRICTED or XSet.MODE_RESTRICTED.
     * @param xopid The application supplied XOPID of this XAsync operation.
     * @param listener listener An optional parameter, may be null, of the XASyncListener object  to be notified when the operation has completed.
     * @return XAsync object representing this operation
     * @throws AuthorizationException The application is not authorized to perform this operation.
     * @throws AuthenticationExpiredException The latest authentication has expired and the application must reauthenticate.
     * @throws InvalidArgumentException inXuid or mode are null, or the Xopid id may already be in use.
     * @throws XSetInaccessibleException XSet does not exist or is not accessible.
     * @throws XSystemAbandonException The XSystem is in the abandoned state and may only be closed.
     * @throws XSystemCorruptException The XSystem is in the corrupt state and may only be abandoned.
     * @throws XAMException An error exists with the underlying XAM Storage System or VIM.
     *             <DT><b>Thread Safety:</b>
     *             <DD>This method is thread safe, the XAsyncListener must be coded in a thread safe manner
     *             <DT><b>Blocking:</b>
     *             <DD>This method will return immediately
     */
    public XAsync asyncCopyXSet( XUID           inXuid,
                                 String         mode,
                                 long           xopid,
                                 XAsyncListener listener )
       throws AuthenticationExpiredException, AuthorizationException, 
              InvalidArgumentException, XSetInaccessibleException, 
              XSystemAbandonException, XSystemCorruptException,
              XAMException;
}
