/*
 * Copyright (c) 2009, Sun Microsystems, Inc.
 * Copyright (c) 2009, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package examples;

import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.util.Properties;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XSystem;
import org.snia.xam.base.XAMImplementation;
import org.snia.xam.util.SASLUtils;
import org.snia.xam.vim.reference.ReferenceAuthenticationStatus;
import org.snia.xam.vim.reference.utils.ReferenceSaslUtils;


/**
 * Simple client example program that loads and connects to the Reference VIM.
 * Note that for this example the Reference VIM is started and loaded into the
 * current process' JVM. In a production VIM implementation the VIM is likely
 * to be running in a separate process as a server. 
 * So in this example application calls are made to the Java XAM Library to 
 * load and register the Reference VIM.
 */
public class ExampleBase {
	/** 
         * System property that specifies the test property file name.
         * This property file contains the definitions for the properties
         * listed below.
         */
	public static final String TEST_PROP_FILE = "xam.test.props";
	
	/** 
         * Test property that specifies the connection XRI (XSystem Resource Identifier).
         * This property specifies the VIM to connect to and any other argument
         * values accepted by the VIM on the connect method.
         * 
         * For example in the Reference VIM's xam.test.props file the
         * xam.test.xri property is set to 
         *        xam.test.xri=snia-xam://SNIA_Reference_VIM!localhost
         * This is takes the default repository location as specified by
         * java.io.tmpdir.
         * When connecting to the reference VIM it is possible to specify a
         * different location for the backing store by using the "dir"
         * parameter. For example:
         * Directory location of /home/mytest/xam_storage (Unix)
         * xam.test.xri=snia-xam://SNIA_Reference_VIM!localhost?dir=/home/mytest/xam_storage
         * 
         * Directory location of C:\mytest\xam_storage (Windows)
         * xam.text.xri=snia-xam://SNIA_Reference_VIM!localhost?dir=C:\mytest\xam_storage
         */
	public static final String XRI_PROP  = "xam.test.xri";
	
	/** 
         * Test property that specifies the properties file containing the
         * definition for '.xam.config.vim.alias.<VIM-Name> and similar 
         * properties. The last part of the property name has to match the name 
         * of the VIM. The property value must specify the VIM class name. 
         * For example:
         * .xam.config.vim.alias.SNIA_Reference_VIM=org.snia.xam.vim.reference.ReferenceVIM
         */
	public static final String CONFIG_PROP = "xam.test.vims";
	
	/** Test property that specifies the username */
	public static final String USER_PROP = "xam.test.username";
	
	/** Test property that specifies the username */
	public static final String PASS_PROP = "xam.test.password";
	
	protected static final String DEFAULT_USER = "test";
	protected static final String DEFAULT_PASS = "test";
        
	protected static String     s_pass;
	protected static String     s_user;
	protected static XAMLibrary s_xam;
	protected static String     s_xri;

    /**
     * Because the VIM is being run out of this JVM must use the XAM_Library 
     * to load the VIM
     * @throws java.lang.Exception
     */
    public static void initLibrary() throws Exception {
        if (s_xam == null) {
            System.out.println("\nInitializing VIM");
            // Get the test properties
            Properties props = new Properties();
            String testPropFile = System.getProperty(TEST_PROP_FILE);
            if (testPropFile == null)
                testPropFile = TEST_PROP_FILE;

            System.out.println("Loading test properties from file: " +
                    testPropFile);
            props.load(new FileInputStream(testPropFile));

            s_xri = props.getProperty(XRI_PROP);

            s_user = props.getProperty(USER_PROP, DEFAULT_USER);
            s_pass = props.getProperty(PASS_PROP, DEFAULT_PASS);

            // XAM Library loads the VIM name and associated 
            // Java implementation class
            System.out.println("Loading the VIM using the Java " +
                    "XAM Library. ");
            System.out.println("VIM Configuration contained in file: " +
                    props.getProperty(CONFIG_PROP));
            s_xam = new XAMImplementation(props.getProperty(CONFIG_PROP));
        }
    }

    
    /**
     * Client application connect to the VIM.
     * @param xri XSystem Resource Identifier (VIM name) and connection args
     * @return XSystem instance
     * @throws java.lang.Exception
     */
    public static XSystem connectToVIM(String xri) throws Exception {
        XSystem xsystem = null;
        // The xri string contains the name of the VIM to connect to
        // e.g. xam.test.xri=snia-xam://SNIA_Reference_VIM!localhost
        System.out.println("Connection arguments: " +
                xri);
        xsystem = s_xam.connect(xri);
        return xsystem;
    }

    /**
     * Send our user credentials to the reference VIM.
     * 
     * @param system
     * @throws org.snia.xam.XAMException
     */
    public static void authenticate(XSystem system) throws XAMException {
        String defMech = system.getString(XSystem.XAM_XSYSTEM_AUTH_SASL_DEFAULT);
        ByteArrayOutputStream response = new ByteArrayOutputStream(200);
        byte[] inputData = null;
        int retValue = 0;
        if (defMech.equals(ReferenceSaslUtils.SASL_MECHANISM_ANONYMOUS)) {
            retValue = system.authenticate(inputData, response);
        } else if (defMech.equals(SASLUtils.SASL_PLAIN)) {
            byte[] creds = ReferenceSaslUtils.encodeSASLPlain(null,
                    ReferenceAuthenticationStatus.TEST_USERNAME,
                    ReferenceAuthenticationStatus.TEST_PASSWORD);
            retValue = system.authenticate(creds, response);
         } else {
            throw new XAMException("Unknown default SASL mechanism " + defMech);
        }
        
        // Validate return status value
        if (retValue != XSystem.XAM_SASL_COMPLETE) {
            throw new XAMException("Failed to authenticate.");
        }
    }
}
