/*
 *  Copyright (c) 2009, Sun Microsystems, Inc.
 *  Copyright (c) 2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.FieldExistsException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidFieldNameException;
import org.snia.xam.InvalidOperationException;
import org.snia.xam.InvalidXSetModeException;
import org.snia.xam.MaximumFieldException;
import org.snia.xam.ObjectInUseException;
import org.snia.xam.XAMException;
import org.snia.xam.XSet;
import org.snia.xam.XSetAbandonException;
import org.snia.xam.XSetCorruptException;
import org.snia.xam.XSystem;
import org.snia.xam.XSystemAbandonException;
import org.snia.xam.XSystemCorruptException;

/**
 * Manage Access Policies for an XSystem instance. To work, this object needs a list of the
 * users so it can create all the access policies. 
 * 
 * This model creates a unique policy name for each user. The policy itself restricts all 
 * access to an XSet to only the owner of the policy (the authenticated user). Two additional 
 * policies are defined: SniaAccessReadAll and SniaAccessReadWriteAll. 
 */
public class AccessPolicyManager
{
   /* Access policies. One per user, each user's XSets are isolated from the others, 
    * except for SniaAccessReadAll, SniaAccessReadWriteAll.
    */
   public final static String STR_ACCESS_POLICY_READ_ALL       = "org.snia.refvim.access.read.all";            
   public final static String STR_ACCESS_POLICY_READ_WRITE_ALL = "org.snia.refvim.access.read.write.all"; 
   
   /* All users may read. */
   private final static AccessPolicy ACCESS_POLICY_READ_ALL = 
      new AccessPolicy( AccessPolicy.ALL, true, false );
   
   /* All users may read/write */
   private final static AccessPolicy ACCESS_POLICY_READ_WRITE_ALL = 
      new AccessPolicy( AccessPolicy.ALL, true, true );
   
   private final Vector<String> m_readPolicies = new Vector<String>();
   
   private ReferenceXSystem m_xsystem_instance;
   private Hashtable<String,AccessPolicy> m_access_policies;
   
   public AccessPolicyManager( ReferenceXSystem xsystemInstance, ReferenceAuthenticationStatus authMgr ) 
      throws InvalidFieldNameException, FieldExistsException, MaximumFieldException, XSetAbandonException, 
      XSetCorruptException, XSystemAbandonException, XSystemCorruptException, 
      AuthenticationExpiredException, InvalidArgumentException, InvalidOperationException, 
      ObjectInUseException, XAMException
   {
      m_access_policies = new Hashtable<String,AccessPolicy>();
      m_xsystem_instance = xsystemInstance;
      
      loadAccessTable( authMgr );
      loadPoliciesInXSystem();
   }

   /**
    * Create policy objects for each username + the two extra policies.
    * @param authMgr
    */
   private void loadAccessTable( ReferenceAuthenticationStatus authMgr )
   {
      Enumeration<String> users = authMgr.getUsernameEnumeration();
      
      while( users.hasMoreElements() )
      {
         String username = users.nextElement();
         m_access_policies.put( username, new AccessPolicy(username, true, true) );
      }
      
      m_access_policies.put( STR_ACCESS_POLICY_READ_ALL, ACCESS_POLICY_READ_ALL );
      m_access_policies.put( STR_ACCESS_POLICY_READ_WRITE_ALL, ACCESS_POLICY_READ_WRITE_ALL );    
   }

   /**
    * Publish all the access policy names for each policy in the XSystem instance.
    * @throws InvalidFieldNameException
    * @throws FieldExistsException
    * @throws MaximumFieldException
    * @throws XSetAbandonException
    * @throws XSetCorruptException
    * @throws XSystemAbandonException
    * @throws XSystemCorruptException
    * @throws AuthenticationExpiredException
    * @throws InvalidArgumentException
    * @throws InvalidOperationException
    * @throws ObjectInUseException
    * @throws XAMException
    */
   private void loadPoliciesInXSystem() 
      throws InvalidFieldNameException, FieldExistsException, MaximumFieldException, 
             XSetAbandonException, XSetCorruptException, XSystemAbandonException, 
             XSystemCorruptException, AuthenticationExpiredException, InvalidArgumentException, 
             InvalidOperationException, ObjectInUseException, XAMException
   {
      Enumeration<String> keys = m_access_policies.keys();
      String prefix = XSystem.XAM_XSYSTEM_ACCESS_POLICY_LIST+".";
      
      while( keys.hasMoreElements() )
      {
         String userName = keys.nextElement();
         m_xsystem_instance.createXSystemProperty( prefix+userName, true /*readonly*/, userName );
      }
      
      /* Load the readable access policies. Query will want to use this list. */
      m_readPolicies.add(m_xsystem_instance.getAuthenticatedUsername());
      m_readPolicies.add(STR_ACCESS_POLICY_READ_ALL);
      m_readPolicies.add(STR_ACCESS_POLICY_READ_WRITE_ALL);
      m_readPolicies.add(""); // Same as read_write_all
   }
   
   /** 
    * Return the list of policies which allow read access to an XSet.
    * @return The policy list.
    */
   public Vector<String> getReadPolicies() { return m_readPolicies; }
   
   /**
    * Checks an XSet policy to determine if it is compatible for read access.
    * @param xsetAccessPolicy
    * @param authenticatedUser
    * @return True if read access is allowed, false otherwise.
    */
   public boolean allowReadAccess( String xsetAccessPolicy, String authenticatedUser )
   {
      if ((xsetAccessPolicy == null)||xsetAccessPolicy.equals(""))
         return true;
      
      AccessPolicy policy = m_access_policies.get(xsetAccessPolicy);
      if( policy == null )
         return false;
      else
         return policy.canRead(authenticatedUser);
   }

   /**
    * Checks an XSet policy to determine if it is compatible for write access.
    * @param xsetAccessPolicy
    * @param authenticatedUser
    * @return True if write access is allowed, false otherwise.
    */
   public boolean allowWriteAccess( String xsetAccessPolicy, String authenticatedUser )
   {
      if (xsetAccessPolicy == null)
         return true;
      
      AccessPolicy policy = m_access_policies.get(xsetAccessPolicy);
      return policy.canWrite(authenticatedUser);
   }
   
   /**
    * Check compatibility for XSet modes, the XSet access policy, and the authenticated user.
    * @param xsetMode
    * @param xsetPolicy
    * @param user
    * @return True if the XSet mode is compatible with the XSet access policy.
    * @throws InvalidXSetModeException
    */
   public boolean allowAccess( String xsetMode, String xsetPolicy, String user ) throws InvalidXSetModeException
   {
      if( xsetMode.equals( XSet.MODE_READ_ONLY ))
         return allowReadAccess( xsetPolicy, user );
      else if( xsetMode.equals( XSet.MODE_RESTRICTED ))
         return allowWriteAccess( xsetPolicy, user );
      else if( xsetMode.equals( XSet.MODE_UNRESTRICTED ))
         return allowWriteAccess( xsetPolicy, user );
      else
         throw new InvalidXSetModeException();
   }
}
