package org.snia.xam.vim.reference;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.Term;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.PhraseQuery;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TopDocs;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;
import org.snia.xam.QueryException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XIterator;
import org.snia.xam.XSet;
import org.snia.xam.XSetException;
import org.snia.xam.XUID;
import org.snia.xam.vim.reference.query.ParsedFactorNode;


public class ContentIndexManager
{
   // Maximum number of results from the Lucene query. 
   static final int MAX_RESULTS = 10000000;
   
   private ReferenceXSystem m_xsystem;
   private Logger           m_logger;
   private String           m_index_path;
   
   // Lucene state variables
   private Directory        m_fsDirectory;

   private static Hashtable<String,ContentIndexManager> s_instance = new Hashtable<String,ContentIndexManager>();
   private static Boolean   s_lock     = Boolean.valueOf(false);
   
   /**
    * Create the index manager.
    * @param xsys The reference XSystem creating this initially
    * @param path Where the database is stored.
    * @param create If true, cause Lucene to erase and create the index directory
    * @throws XAMException
    * @throws IOException
    */
   private ContentIndexManager( ReferenceXSystem xsys, String path, boolean create ) 
      throws XAMException, IOException
   {
      m_xsystem    = xsys;
      m_logger     = xsys.getLogger();
      
      File dbPath = new File(path);
      String fileSeparator = System.getProperty("file.separator");
      m_index_path    = dbPath.getCanonicalPath() + fileSeparator + "contentIndex" ;
      
      File contentIndexPath = new File( m_index_path );
      if( ! contentIndexPath.exists() )
         contentIndexPath.mkdir();
      
      m_fsDirectory = FSDirectory.getDirectory( contentIndexPath, create );
   }

   /**
    * Because the store directory may be shared over several ReferenceXSystems
    * there may only be a single instance of the ContentIndexManager.
    * @param xsys A referring ReferenceXSystem
    * @param dbPath The path to the database directory
    * @param create If true causes the directory to be erased and created.
    * @return The ContentIndexManager for this store location.
    * @throws XAMException
    */
   public static ContentIndexManager GetInstance( ReferenceXSystem xsys, 
                                                  String           dbPath,
                                                  boolean          create ) 
   {
      synchronized(s_lock) {
         String storeLocation = xsys.getPersistenceMgr().getRepositoryLocation();
         ContentIndexManager instance = s_instance.get(storeLocation);
         if( instance == null )
         {
            try { 
               instance = new ContentIndexManager( xsys, dbPath, create ); 
               s_instance.put( storeLocation, instance );
            }
            catch(Exception e ) { instance = null; } 
         }
         
         return instance;
      }
   }

   /**
    * Indexes all text/plain streams for later query.
    * Adds each stream to a Lucene document, and adds the XUID as a stored field. Each InputStream
    * must be closed AFTER to the document has been added to the index, so these are saved in 
    * a simple vector until post indexing time.
    * @param xset The XSet containing any streams.
    * @throws XAMException
    */
   synchronized public void storeXSet( ReferenceXSet xset ) 
      throws XAMException
   {
      Vector<Reader> readers = new Vector<Reader>();
      XIterator iter = xset.openFieldIterator("");
      
      IndexWriter writer = null;
      Document doc = new Document();
      StandardAnalyzer analyzer = new StandardAnalyzer();
      
      // This is a work around to determine if an index exists. Modern versions of 
      // Lucene will create if needed, but 1.9 doesn't have that capability.
      boolean createIndex = false;
      try
      {
         IndexReader testReader = IndexReader.open(m_fsDirectory);
         testReader.close();
      }
      catch(Exception ex)
      {
         createIndex = true;
      }
      
      try 
      {
         writer = new IndexWriter( m_fsDirectory, analyzer, createIndex );
         while( iter.hasNext() )
         {
            String fieldName = (String) iter.next();
         
            if( xset.getFieldType(fieldName).startsWith( XAMLibrary.TEXT_PLAIN_MIME_TYPE) )
            {
               readers.add(storeStream( xset, fieldName, doc ));
            }
         }
         
         // The XUID is a special field
         if( readers.size() > 0 )
         {
            String xuidString = xset.getXUID(XSet.XAM_XUID).toString();
            doc.add( new Field(XSet.XAM_XUID, xuidString, Field.Store.YES, Field.Index.TOKENIZED ));
         
            writer.addDocument(doc);
            writer.optimize();
         }
         writer.close();
         
         Enumeration<Reader> r = readers.elements();
         while( r.hasMoreElements() )
            r.nextElement().close(); // Closes isXStream and the underlying XStream
      }
      catch (Exception e)
      {
         m_logger.log( Level.SEVERE, "Content index service unspecified exception", e );
         throw new XSetException( "Failed to store XStream content", e );
      }
   }

   /**
    * When handed the parent XSet and the stream, index the content. This implementation
    * requires that the stream is of MIME type, "text/plain". Note that this method 
    * requires the use of the Java Toolkit wrapper class XStreamInputStream.
    * @param xset The XSet containing the stream
    * @param name The stream name
    * @param doc The Lucene document to store the fields (streams)
    * @return The InputStreamReader used, which must be closed AFTER the content is added to the Lucene Document.
    */
   private Reader storeStream( ReferenceXSet xset, String name, Document doc )
      throws IOException, XAMException
   {
      ReferenceStreamField stream = (ReferenceStreamField) xset.getFieldNoChecking(name);
      File f = new File(stream.getAbsoluteLocation());
      m_logger.finest( "Content Index: Adding " + name + " " + f);
      Reader rdr = new FileReader(f);
      doc.add(new Field(name, rdr ));
      return rdr;
   }

   /**
    * Indexes all text/plain streams for later query.
    * Adds each stream to a Lucene document, and adds the XUID as a stored field. Each InputStream
    * must be closed AFTER to the document has been added to the index, so these are saved in 
    * a simple vector until post indexing time.
    * 
    * This method is synchronized to prevent any other IndexWriters from being open at
    * the same time this method is running. Lucene does not allow an IndexReader to 
    * delete a document while one ore more IndexWriters are open. A more robust and 
    * scalable implementation would provide a more sophisticated synchronization technique.
    * 
    * @param xset The XSet containing any streams.
    * @throws XAMException
    */
   synchronized void deleteXSet( XUID xuid ) 
      throws XAMException
   {
      try
      {
         StandardAnalyzer analyzer = new StandardAnalyzer();
         IndexReader reader = IndexReader.open( m_fsDirectory );
         reader.deleteDocuments( new Term( XSet.XAM_XUID, xuid.toString() ) );        
      }
      catch( FileNotFoundException fne )
      {
         // This index has not been created. The assumption is that there is 
         // no XSet data to delete. 
      }
      catch( Exception ex )
      {
         throw new XSetException("Error deleting an XSet from the content index", ex );
      }
   }

   /** 
    * Search the content index to get a list of documents satisfying the 
    * level 2 sub expression.
    * @param pfn The ParsedFactorNode which is the level 2 sub expression
    * @return The list of documents
    */
   public String[] search(ParsedFactorNode pfn)
      throws QueryException
   {
      try 
      {
         Vector<String> results = new Vector<String>();
         
         // Search 
         IndexSearcher isearcher = new IndexSearcher(m_fsDirectory);
         
         // Parameters
         Vector v = (Vector) pfn.getExpression().getLiteral();
         String fieldName = pfn.getProperty();
         
         Query query = null;
         StandardAnalyzer analyzer = new StandardAnalyzer();
         QueryParser parser = new QueryParser( fieldName, analyzer);
         switch ( pfn.getOperator() )
         {
            case ParsedFactorNode.CONTAINS:
               // Parse a simple query that searches for "text":
               String containsText = (String) v.get(0);
               query = parser.parse(containsText);
               m_logger.finest( "Content Contains: " + fieldName + " cont " + containsText );
               break;
               
            case ParsedFactorNode.BEFORE:
               PhraseQuery pquery_b = new PhraseQuery();
               String text1_b = (String) v.get(0);
               String text2_b = (String) v.get(1);
               pquery_b.add(new Term(fieldName, text1_b) );
               pquery_b.add(new Term(fieldName, text2_b) );
               pquery_b.setSlop(Integer.MAX_VALUE);
               query = pquery_b;
               m_logger.finest( "Content Before: " + fieldName + " before " + 
                  text1_b + " " + text2_b );
               break;
               
            case ParsedFactorNode.AFTER:
               PhraseQuery pquery_a = new PhraseQuery();
               String text1_a = (String) v.get(0);
               String text2_a = (String) v.get(1);
               pquery_a.add(new Term(fieldName, text2_a) );
               pquery_a.add(new Term(fieldName, text1_a) );
               pquery_a.setSlop(Integer.MAX_VALUE);
               query = pquery_a;
               m_logger.finest( "Content After: " + fieldName + " after " + 
                  text2_a + " " + text1_a );
               break;
               
            case ParsedFactorNode.WITHIN:
               String text1_w  = (String) v.get(0);
               String text2_w  = (String) v.get(1);
               Integer distance = (Integer) v.get(2);
               StringBuffer query_string = new StringBuffer();
               query_string.append("\"");
               query_string.append(text1_w);
               query_string.append(" ");
               query_string.append(text2_w);
               query_string.append("\" ~");
               query_string.append(distance);
               query_string.append(" OR \"");
               query_string.append(text2_w);
               query_string.append(" ");
               query_string.append(text1_w);
               query_string.append("\" ~");
               query_string.append(distance);
               query = parser.parse( query_string.toString() );
               m_logger.finest( "Content Within: " + fieldName + " within " + 
                  text1_w + " " + text2_w );
               break;
               
            default:
               throw new QueryException( "Unknown level 2 operator: " + pfn.getOperator() );
         }
         
         TopDocs docs = isearcher.search(query, null, MAX_RESULTS );

         // Iterate through the results:
         ScoreDoc[] docList = docs.scoreDocs;
         
         for (int i = 0; i < docList.length; i++) {
           Document hitDoc = isearcher.doc(docList[i].doc);
           results.add(hitDoc.get(XSet.XAM_XUID) );
           m_logger.finest( "Content Found: " + hitDoc.get(XSet.XAM_XUID) );
         }
         isearcher.close();
         
         // Allocate 10 items, hoping the results will fit, otherwise a new 
         // array is to be allocated but the toArray call.
         String cache[] = new String[10];
         return results.toArray(cache);
      }
      catch( QueryException qe )
      {
         throw qe; // prevents the qe from getting wrapped in the other catch.
      }
      catch( Exception e )
      {
         throw new QueryException( "Error during table population for level2", e );
      }
   }
}
