/*
 *  Copyright (c) 2009, Sun Microsystems, Inc.
 *  Copyright (c) 2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.File;
import java.io.FilenameFilter;
import java.io.StringBufferInputStream;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.Calendar;
import java.util.Hashtable;
import java.util.Vector;
import java.util.logging.Logger;
import java.util.logging.Level;

import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.QueryException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XIterator;
import org.snia.xam.XSet;
import org.snia.xam.XSystemException;
import org.snia.xam.XUID;
import org.snia.xam.base.Field;
import org.snia.xam.base.Property;
import org.snia.xam.util.ISO8601Date;
import org.snia.xam.vim.reference.query.ParsedFactorNode;
import org.snia.xam.vim.reference.query.ParsedQuery;
import org.snia.xam.vim.reference.query.XamQlParser;

/**
 * Database Manager of the XAM DB for properties. Initially this implementation will
 * store the properties of XSets. Other housekeeping tables may be added at some
 * point in the future.
 * 
 * There exists an instance of the DBManager for each storage location, since each location
 * will have its own database. The DBManager instances will live for the life of the 
 * JVM, and under the current implementation, don't close. This prevents multiple 
 * ReferenceVIM instances, in separate JVMs from accessing the same storage location. 
 * 
 * It should be possible to use multiple client JVMs accessing the same storage location
 * if the client JVMs are isolated from the single ReferenceVIM instance using the HTTP VIM.
 */
public class DBManager
{
   static final String SQLDRIVER  = "org.apache.derby.jdbc.EmbeddedDriver";
   static final String CONNECTURL_PREFIX = "jdbc:derby:";
   static final String CONNECTURL_CREATE = ";create=true";
   
   // This version is inserted into the version table. The DB should
   // rebuilt whenever the version of the code and DB don't match.
   // 1 == Initial version
   // 2 == WIP implementing level 1
   // 3 == Completed Level 1 query
   // 4 == Add content index for Level 2 query
   private static final int    SCHEMA_VERSION = 4;
   private static final String VERSION_TABLE  = "versions";
   private static final String VERSION_COLUMN = "schemaversion";
   
   /** The vertical table of field values. */
   static final String VALUETABLE = "xfieldvalues";
   
   static final String DB_NAME = "ReferenceVimDB";
   
   // Column names in the XFieldValues table
   static final String PROPNAME = "propname";
   static final String PTYPE    = "type";
   static final String BINDING  = "binding";
   static final String READONLY = "readonly";
   static final String LENGTH   = "length";
   static final String SVALUE   = "svalue";
   static final String NUMVALUE = "numvalue";
   static final String BVALUE   = "bvalue";
   static final String XVALUE   = "xuid";
   
   // SQL Operators used in generating SQL strings.
   static final String SQL_OP_EQ = "=";
   static final String SQL_OP_NE = "<>";
   static final String SQL_OP_GT = ">";
   static final String SQL_OP_GE = ">=";
   static final String SQL_OP_LT = "<";
   static final String SQL_OP_LE = "<=";
   static final String SQL_OP_LIKE = " like ";

   // Types used to construct the tables. 
   static final String SQL_TYPE_STRING  = "varchar(600)"; // For XAM_STRING, XAM_DATETIME, XAM_XUID, Type
   static final String SQL_TYPE_DOUBLE  = "double";       // For XAM_DOUBLE, XAM_INTEGER
   static final String SQL_TYPE_INTEGER = "bigint";       // For length
   static final String SQL_TYPE_BOOLEAN = "smallint";     // For XAM_BOOLEAN, Binding, Readonly
   
   // Used in the db tables
   static final int  INT_TRUE  = 1;
   static final int  INT_FALSE = 0;
   
   // Fake property, synthesized and applied to XSets (just in XFieldValues) so 
   // query can restrict against policies. Not visible to programs cause it's in the 
   // DB, not in the XSets. 
   static final String VND_ACCESS_POLICY = ".vnd.org.snia.xam.reference.access_policy";
   
   private Connection m_conn;
   private PreparedStatement m_store_value_prep_stmt;
   private PreparedStatement m_delete_xset_prep_stmt;
   
   private static Hashtable<String,DBManager> s_instance = new Hashtable<String,DBManager>();
   private static Boolean   s_lock     = Boolean.valueOf(false);
   
   private String              m_connect_url;
   private Logger              m_logger = null;
   private String              m_repoPosition = null;
   private String              m_db_path = null;
   
   /** An adjunct to the database manager, indexes content and adds support
    * for Level 2 query.
    */
   private ContentIndexManager m_content_index_manager = null;

   public DBManager( ReferenceXSystem xsys ) throws ClassNotFoundException, SQLException, XAMException
   {
      m_logger = xsys.getLogger();
      
      // Construct the DBManager for this storage location, but it is important 
      // to never keep a reference to the XSystem. The XSystem is only used when
      // loading XSets from the persistent store. 
      m_repoPosition  = xsys.getPersistenceMgr().getRepositoryLocation();
      m_db_path = m_repoPosition + DB_NAME;
      m_connect_url = CONNECTURL_PREFIX + m_db_path ;
      setupDB(xsys);
   }
   
   public static DBManager GetInstance( ReferenceXSystem xsys ) throws ClassNotFoundException, SQLException, XAMException
   {
      synchronized(s_lock) {
         String storeLocation = xsys.getPersistenceMgr().getRepositoryLocation();
         DBManager instance = s_instance.get(storeLocation);
         if( instance == null )
         {
            instance = new DBManager( xsys );
            s_instance.put( storeLocation, instance );
         }
         
         return instance;
      }
   }
   
   /** Indicates if Level 2 query is supported. */
   public boolean supportsLevel2() { return m_content_index_manager != null; }
      
   /**
    * Print out the sub tables, for debugging.
    * 
    * @throws SQLException
    */
   @SuppressWarnings("unused")
   private void dumpSubTables( String table) throws SQLException
   {
      String query = "select * from " + table;
      CallableStatement stmt = m_conn.prepareCall(query);
      ResultSet results = stmt.executeQuery();
      
      int xcol    = results.findColumn(XVALUE);
      
      // Print the header
      System.out.println( "=== " + table + " table ===" );
      ResultSetMetaData md = results.getMetaData();
      int column_count = md.getColumnCount();
      for( int c = 0; c < column_count; c++ )
      {
         System.out.print( md.getColumnName(c+1));
         System.out.print( " " );
      }
      System.out.println("");
      
      // Print the table contents.
      while( results.next() )
      {
         String prop_name = "Not impl";
         String xuid      = results.getString(xcol);
         System.out.println( prop_name + "/" + xuid );
      }
      results.close();
      stmt.close();
   }

   /** Given the parts of a general field, insert them into the values table.
    * @param field_name The name of the field
    * @param type The MIME type of the field
    * @param svalue The string value for the property (strings, datetime, xuid)
    * @param numvalue The numeric value for numeric poperties (double, int)
    * @param bvalue The boolean value for the property
    * @param binding The boolean value of the binding attribute for the field
    * @param readonly The boolean value of the readonly attribute for the field
    * @param length The value of the length attribute for the field, as measured in bytes
    * @param xuid The XUID of the XSet to which this field belongs
    * @throws SQLException 
    */
    public void storeField( String  field_name,
                    String  type,
                    String  svalue,
                    double  numvalue,
                    int     bvalue,
                    boolean binding,
                    boolean readonly,
                    long    length,
                    XUID    xuid) throws SQLException
   {
      try
      {
         m_store_value_prep_stmt.setString(1, field_name);
         m_store_value_prep_stmt.setString(2, svalue); 
         m_store_value_prep_stmt.setDouble(3, numvalue);
         m_store_value_prep_stmt.setInt(4, bvalue );
         m_store_value_prep_stmt.setInt(5, (binding?INT_TRUE:INT_FALSE));
         m_store_value_prep_stmt.setInt(6, (readonly?INT_TRUE:INT_FALSE));
         m_store_value_prep_stmt.setLong(7, length);
         m_store_value_prep_stmt.setString(8, type );
         m_store_value_prep_stmt.setString(9, xuid.toString());
         m_store_value_prep_stmt.execute();
      }
      catch (SQLException e)
      {
         m_logger.log( Level.SEVERE, "Can't insert generic field '" + field_name + "': " + e.toString());
         throw e;
      }
   }

   /**
    * Check the schema in the data base to make sure it is consistent with what the code needs.
    * @return True if the schema is compatible with this version of the DBManager.
    * @throws SQLException 
    */
   private boolean checkSchema() throws SQLException
   {
      boolean isCompatible = false;

      // Implementation note: It seems as though the best way to do this would be 
      // to grab the DatabaseMetaData and get a list of tables. Seems as though this
      // isn't working with Derby, so run a simple query against the values table
      // and if it craps out, represent the schema as being incompatible.
      try {
         m_conn = DriverManager.getConnection(m_connect_url);
         
         String testSQL = "select " + VERSION_COLUMN + " from " + VERSION_TABLE;
         CallableStatement stmt = m_conn.prepareCall(testSQL);
         ResultSet rs = stmt.executeQuery();
         if( rs.next() )
         {
            int version = rs.getInt(1);
            if( version ==  SCHEMA_VERSION)
               isCompatible = true;
         }
         rs.close();
      }
      catch( SQLException se ) 
      { 
         if(m_conn != null)
            m_conn.close();
         isCompatible = false;
      }
      
      return isCompatible;
   }

   /**
    * Create the basic tables needed by the data base.
    * 
    * @throws ClassNotFoundException
    * @throws SQLException
    * @throws XAMException 
    */
   private void setupDB( ReferenceXSystem xsys ) throws ClassNotFoundException, SQLException, XAMException
   {
      String storePath = xsys.getRepositoryLocation();
      if( m_logger != null )
         m_logger.log( Level.INFO, "Connecting to the DB: " + storePath );
      
      // Create the DB
      Class.forName(SQLDRIVER); 
      
      if( checkSchema() )
      {
         createPreparedStatements();
         try { 
            m_content_index_manager = ContentIndexManager.GetInstance(xsys, m_db_path, false );
         } catch( Throwable t)
         {
            m_content_index_manager = null;
            m_logger.log( Level.WARNING, "Level 2 is not supported: ", t );
         }
      }
      else
      {
         m_logger.log( Level.INFO, "Incompatible schema, rebuilding DB: " + storePath );
         
         m_conn = DriverManager.getConnection(m_connect_url+ CONNECTURL_CREATE);
         
         // Drop any old tables
         try {
            CallableStatement drop = m_conn.prepareCall( "drop table " + VALUETABLE );
            drop.execute();
            drop.close();
         } catch( SQLException sqe ) { /* sqe.printStackTrace(); */ }
         
         try {
            CallableStatement drop = m_conn.prepareCall( "drop table " + VERSION_TABLE );
            drop.execute();
            drop.close();
         }
         catch( SQLException sqe ) { /* sqe.printStackTrace(); */ }
         
         // Create the tables
         String createSql = "create table " + VALUETABLE + 
         " ( propname "  + SQL_TYPE_STRING +
         "," + SVALUE    + " " + SQL_TYPE_STRING +
         "," + NUMVALUE  + " " + SQL_TYPE_DOUBLE + 
         "," + BVALUE    + " " + SQL_TYPE_BOOLEAN +
         "," + BINDING   + " " + SQL_TYPE_BOOLEAN +
         "," + READONLY  + " " + SQL_TYPE_BOOLEAN + 
         "," + LENGTH    + " " + SQL_TYPE_INTEGER +
         "," + PTYPE     + " " + SQL_TYPE_STRING + 
         "," + XVALUE    + " " + SQL_TYPE_STRING + ")";
         CallableStatement stmt = m_conn.prepareCall( createSql );
         stmt.execute();
         stmt.close();
         
         CallableStatement stmt2 = m_conn.prepareCall( "create table " + VERSION_TABLE + 
            "( " + VERSION_COLUMN + " smallint )" );
         stmt2.execute();
         stmt2.close();
         
         stmt2 = m_conn.prepareCall( "insert into " + VERSION_TABLE + 
            "(" + VERSION_COLUMN + ") values(" + SCHEMA_VERSION + ")" );
         stmt2.execute();
         stmt2.close();
         
         createPreparedStatements();
         
         try { 
            m_content_index_manager = ContentIndexManager.GetInstance(xsys, m_db_path, true);
         } catch( Throwable t)
         {
            m_content_index_manager = null;
            if( m_logger != null )
               m_logger.log( Level.WARNING, "Level 2 is not supported: ", t );
         }

         reloadExistingXSets(xsys);
      } 
   }

   /**
    * Create the prepared statements used in maintaining the DB
    * @throws SQLException
    */
   private void createPreparedStatements() throws SQLException
   {
      // * Create Prepared statements for inserting values
      String sql = "insert into " + VALUETABLE + " values(?,?,?,?,?,?,?,?,?)";
      m_store_value_prep_stmt = m_conn.prepareStatement(sql);

      sql = "delete from " + VALUETABLE + " where xuid=?";
      m_delete_xset_prep_stmt = m_conn.prepareStatement(sql);
   }

   /**
    * Find all the existing XSets in the repository and store them in the database.
    * @throws XAMException 
    * @throws XAMException 
    */
   private void reloadExistingXSets( ReferenceXSystem xsys ) throws XAMException 
   {
      long start = System.currentTimeMillis();
      
      String reposDir = m_repoPosition;
      File dir = new File(reposDir);
      FilenameFilter hexFilter = new FilenameFilter() {
         public boolean accept(File dir, String name) {
            return (name.startsWith(ReferenceVIMPersistenceMgr.XSET_FILE_PREFIX) &&
               name.endsWith(ReferenceVIMPersistenceMgr.XSET_FILE_SUFFIX));
         }
      };

      XUID xuid = null;
      String[] files = dir.list(hexFilter); 
         
      for (int idx = 0; idx < files.length; idx++) {
         int startIdx = files[idx].indexOf('_');
         int endIdx = files[idx].indexOf('.');
         String xuidString = files[idx].substring(++startIdx, endIdx);
         try { 
            xuid = SniaXUID.createFromSafeString(xuidString);
            if( (idx % 100) == 0)
               m_logger.info( "Reloaded " + idx + " XSets" );
            storeXSet( xsys, xuid );
         }
         catch( Exception e ) {
            m_logger.log( Level.SEVERE, "Can not store XUID: '" + xuid + "' " + e.toString() );
         }
      }
      long end = System.currentTimeMillis();
      double delta = (end - start) / 1000.0;
      double rate = files.length / delta;
      m_logger.log( Level.INFO, "Loaded " + files.length + " in " + delta + " seconds " + rate + " files/sec." );
   }

   /**
    * Given a reference XSystem and XUID, store the entire XSet field attributes into the database.
    * @param xsys The ReferenceXSystem, to give context to the XSet
    * @param xuid The XUID of the XSet to store.
    * @throws XAMException 
    */ 
   public void storeXSet( ReferenceXSystem xsys, XUID xuid ) throws XAMException
   {
      ReferenceXSet xset = new ReferenceXSet( xsys, xuid );
      try { storeXSet( xset ); }
      finally {
         xset.close();
      }
   }
   
   /**
    * Delete the entire XSet from the DB. Used by things like delete, autodelete, and storeXSet.
    * @param xsys
    * @param xuid
    * @throws XAMException
    */
   public void deleteXSet( ReferenceXSystem xsys, XUID xuid ) throws XAMException
   {
      try
      {
         m_delete_xset_prep_stmt.setString(1, xuid.toString());
         m_delete_xset_prep_stmt.execute();
      }
      catch( SQLException sqe )
      {
         String msg = "Can't delete XSet from DB : " + xuid + " " + sqe.toString();
         m_logger.log( Level.SEVERE, msg, sqe );
         throw new XSystemException( msg, sqe);
      }
      if( m_content_index_manager != null )
         m_content_index_manager.deleteXSet(xuid);
   }
   
   /** Given an XSet, store the entire XSet field attributes into the database.
    * @param xset The XSet to store in the database.
    * @throws XAMException
    */
   public void storeXSet( ReferenceXSet xset ) throws XAMException
   {
      XIterator iter     = xset.openFieldIterator("");
      XUID      xuid     = xset.getXUID();
      String    propName = null;
      try
      {
         deleteXSet( xset.getXSystem(), xuid );
         while( iter.hasNext() )
         {
            propName = (String) iter.next();
            
            Field  f        = xset.getField(propName);
            String type     = f.getType();
            String svalue   = null;
            double numvalue = 0.0;
            int    bvalue   = 0;
            
            if( type.equals(XAMLibrary.STYPE_BOOLEAN_MIME_TYPE)) {
               Property p = (Property) f;
               bvalue = ((Boolean) p.getValue()).booleanValue() ? INT_TRUE : INT_FALSE;
            }
            else if( type.equals(XAMLibrary.STYPE_DATETIME_MIME_TYPE)) {
               Property p = (Property) f;
               svalue = ISO8601Date.toISO8601String( ((Calendar) p.getValue()) );
            }
            else if( type.equals(XAMLibrary.STYPE_DOUBLE_MIME_TYPE) ) {
               Property p = (Property) f;
               numvalue = ((Double) p.getValue()).doubleValue();
            }
            else if( type.equals(XAMLibrary.STYPE_INT_MIME_TYPE)) {
               Property p = (Property) f;
               numvalue = ((Long) p.getValue()).longValue();
            }
            else if( type.equals(XAMLibrary.STYPE_STRING_MIME_TYPE)) {
               Property p = (Property) f;
               svalue = ((String) p.getValue());
            }
            else if( type.equals(XAMLibrary.STYPE_XUID_MIME_TYPE)) {
               Property p = (Property) f;
               svalue = ((XUID) p.getValue()).toString();
            }

            storeField(propName,
               type,
               svalue, numvalue, bvalue,
               f.getBinding(),
               f.getReadOnly(), 
               f.getLength(),
               xuid );            
         }
         
         // Special property for all XSets, to enable query access restrictions.
         Property accessPolicy = null;
         try { accessPolicy = (Property) xset.getField( XSet.XAM_ACCESS_POLICY ); } 
         catch( FieldDoesNotExistException fdne ) { /* Ignore it */ }

         String accessPolicyName = ((accessPolicy == null)? "" : accessPolicy.getValue().toString());
         storeField( VND_ACCESS_POLICY,
            XAMLibrary.STYPE_STRING_MIME_TYPE,
            accessPolicyName,
            0.0, 0, 
            false, 
            true,
            accessPolicyName.getBytes( "utf-8" ).length,
            xuid );
         
         // Store index content for Level 2
         if( m_content_index_manager != null )
            m_content_index_manager.storeXSet( xset );
      }
      catch( XAMException xe )
      {
         throw xe; // Prevent another XAMException from being wrapped.
      }
      catch( Exception sqe )
      {
         String msg = "Can't persist " + xuid.toString() + "." + propName;
         m_logger.log(Level.SEVERE, msg, sqe);
         throw new XSystemException( msg, sqe );
      }
      finally
      {
         // Don't close the XSet, since it isn't "owned" by the DBManager.
         iter.close();
      }
   }

   /**
    * This method is the heart of the query execution. It translates the XAM QL to SQL, and runs the query, returning the results.
    * This is also the location where optimization opportunities are detected and optimal methods are executed.
    * @param id The query id, used to generate a table, if required.
    * @param parsed_query The query parse tree.
    * @return ResultSet from the SQL.
    * @throws XAMException 
    * @throws SQLException 
    */
   public ResultSet executeXamQuery(Job            qJob,
                                    String         id, 
                                    ParsedQuery    parsed_query,
                                    Vector<String> readPolicies ) throws SQLException, XAMException
   {
      ResultSet rs = null;
      
      // Do the core work, in this case run the SQL
      if( (parsed_query.getWhereExpression() == null) &&
          (readPolicies == null ))
         rs = DBOptimizedSqlTranslator.selectAllXUIDs(m_conn);    // Optimized select * variant.
      
      else if( (parsed_query.getWherePropertyList().size() == 1)  &&
               (parsed_query.getExistsPropertyList().size() == 0) &&
               (parsed_query.getWhereExpression().isFactor() )    &&
               (readPolicies == null ))
         rs = DBOptimizedSqlTranslator.selectSimpleWhere( m_conn, parsed_query ); // Optimization
      
      else if( (parsed_query.getWherePropertyList().size() == 0)  &&
               (parsed_query.getExistsPropertyList().size() == 1) &&
               (parsed_query.getWhereExpression().isFactor() )    &&
               (readPolicies == null ))
         rs = DBOptimizedSqlTranslator.selectSimpleExists( m_conn, parsed_query );
      
      else
      {
         DBSqlTranslator translator = new DBSqlTranslator( m_conn, 
                                                           m_content_index_manager, 
                                                           m_logger, 
                                                           readPolicies );
         rs = translator.selectXamQuery( qJob, id, parsed_query, readPolicies );
      }

      return rs;
   }
   
   /**
    * Parse and execute a query.
    * @param id An id for the query, may be used to create a table.
    * @param query The query string iteself.
    * @return SQL Result set of the query
    * @throws XAMException
    */
   public ResultSet executeXamQuery( String id, String query, Vector<String> readPolicies ) throws XAMException
   {
      try 
      {
         XamQlParser xql = new XamQlParser(new StringBufferInputStream( query ));

         ParsedQuery q = xql.Start();
         return executeXamQuery( null, id, q, readPolicies );
      }
      catch( Exception sqe )
      {
         String msg = "Query failed: ";
         m_logger.log( Level.SEVERE, msg, sqe );
         throw new XAMException( msg, sqe );
      }
   }

   /**
    * Given a XAM QL operator, return the equivalent SQL operator
    * @param operator The XAM QL operator
    * @return The SQL Operator (string)
    * @throws QueryException
    */
   static String mapXamOpToSqlOp(int operator) throws QueryException
   {
      switch( operator ) {
         case ParsedFactorNode.EQUALS:
            return SQL_OP_EQ;
            
         case ParsedFactorNode.NOT_EQUALS:
            return SQL_OP_NE;
            
         case ParsedFactorNode.GREATER_THAN:
            return SQL_OP_GT;
            
         case ParsedFactorNode.GREATER_THAN_EQ:
            return SQL_OP_GE;
            
         case ParsedFactorNode.LESS_THAN:
            return SQL_OP_LT;
            
         case ParsedFactorNode.LESS_THAN_EQ:
            return SQL_OP_LE;
            
         case ParsedFactorNode.LIKE:
            return SQL_OP_LIKE;
            
         case ParsedFactorNode.CONTAINS:
         case ParsedFactorNode.BEFORE:
         case ParsedFactorNode.AFTER:
         case ParsedFactorNode.WITHIN:
            return SQL_OP_EQ;
            
         default:
            throw new QueryException( "Unknown operator in SQL generator : " + operator );
      }
   }
}
