/*
 *  Copyright (c) 2009, Sun Microsystems, Inc.
 *  Copyright (c) 2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

 package org.snia.xam.vim.reference;

import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Calendar;

import org.snia.xam.QueryException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XSet;
import org.snia.xam.util.ISO8601Date;
import org.snia.xam.vim.reference.query.AnnotatedLiteral;
import org.snia.xam.vim.reference.query.ParsedFactorNode;
import org.snia.xam.vim.reference.query.ParsedQuery;
import org.snia.xam.vim.reference.query.Token;
import org.snia.xam.vim.reference.query.TypeCode;
import org.snia.xam.vim.reference.query.WhereNode;

public class DBOptimizedSqlTranslator
{
   /**
    * Select XUIDs for a single property name. This is an optimized call that does not
    * need to construct the secondary query table, as it can operate off of the 
    * field value table directly.
    * 
    * The trickiest part of this call is to construct a reasonable where clause. 
    * For instance to check for a boolean value
    * equal to true, the call would look like:
    *     selectSingleProperty( "propname", mimeType, "ivalue=1")
    * Searching for a string value would look like:
    *     selectSingleProperty( "propname", mimeType, "svalue='foo'")
    *     
    * Because this exposes the construction of the tables, this method should not 
    * be available as a public part of the DBManager interface.
    * 
    * All boolean values are mapped to 0 for FALSE and 1 for TRUE and are stored in 
    * the integer column. 
    * 
    * All strings, dates, and xuids are stored in the string column (svalue).
    * 
    * All doubles are stored in the double column (dvalue).
    * 
    * Binding values are stored in the binding column (binding).
    * 
    * Readonly values are stored in the readonly column (readonly).
    * 
    * Length values are stored in the length column (length).
    * 
    * Type values are stored in the type column (type).
    * 
    * @param propName The unadorned simple property name
    * @param mimeType The mime type of the property
    * @param whereClause A naked simple where clause to be appended to the query
    * @return True if the schema is compatible with this version of the DBManager.
    * @throws SQLException 
    */
   static private ResultSet selectSingleProperty( Connection conn,
                                                 String     propName, 
                                                 String     mimeType, 
                                                 String     whereClause ) 
      throws SQLException, XAMException
   {
      StringBuffer singleSQL = new StringBuffer("select xuid from ");
      singleSQL.append( DBManager.VALUETABLE );
      singleSQL.append( " where propname='");
      singleSQL.append( propName );
      singleSQL.append( "' and type ='");
      singleSQL.append( mimeType );
      singleSQL.append( "' and " );
      singleSQL.append( whereClause );

      CallableStatement stmt = null;
      stmt = conn.prepareCall(singleSQL.toString());
      ResultSet rs = stmt.executeQuery();
      return rs;
   }

   /**
    * Simple method to get all XUIDs in the store. This is likely to be called by
    * developers trying out query with the world's simplest query 'select ".xset.xuid" '
    * This method is provided as an optimization for the Query Job to use.
    * @throws SQLException Pass it back to the calling job and let the job decide on the proper error.
    */
   static public ResultSet selectAllXUIDs( Connection conn ) throws SQLException, XAMException
   {
      ResultSet rs = selectSingleProperty( 
         conn, 
         XSet.XAM_XUID, 
         XAMLibrary.STYPE_XUID_MIME_TYPE, 
         "propname='.xset.xuid'" );
      return rs;
   }
   
   /**
    * An optimized method dealing with a single factor where node from the query. Typically
    * this is a simple query of the form: select ".xset.xuid" where propname=42
    */
   static public ResultSet selectSimpleWhere( Connection conn, ParsedQuery pq ) 
      throws SQLException, QueryException, XAMException
   {
      WhereNode where = pq.getWhereExpression();
      if( ! where.isFactor() )
         throw new QueryException( "Can't handle parenthesized expressions right now" );
      else {
         ParsedFactorNode pfn = (ParsedFactorNode) where;
         int    valueOf  = pfn.getValueRef();
         String      sql = null;
         
         if( valueOf != ParsedFactorNode.VALUE_OF ) // Attribute function
            sql = generateAttributeFunction( pfn );
         else
            sql = generatePropertyBinaryComparison( pfn );
         
         
         CallableStatement stmt = null;
         stmt = conn.prepareCall(sql);
         ResultSet rs = stmt.executeQuery();
                    
         return rs;
      }
   }

   /**
    * An optimized method dealing with a single factor where node from the query. Typically
    * this is a simple query of the form: select ".xset.xuid" where exists("propname")
    */
   static public ResultSet selectSimpleExists( Connection conn, ParsedQuery pq ) 
      throws SQLException, QueryException, XAMException
   {
      WhereNode where = pq.getWhereExpression();
      if( ! where.isFactor() )
         throw new QueryException( "Can't handle parenthesized expressions right now" );
      else {
         ParsedFactorNode pfn = (ParsedFactorNode) where;
         int    valueOf  = pfn.getValueRef();
         StringBuffer sql = null;
         
         if( valueOf != ParsedFactorNode.EXISTS) // Attribute function
            throw new QueryException( "Optimizer failure, bad value operation: " + pfn.getOperator() );
         else
         {
            sql = new StringBuffer( "select xuid from " );
            sql.append(DBManager.VALUETABLE);
            sql.append(" where ");
            sql.append(DBManager.PROPNAME);
            sql.append("='");
            sql.append(pfn.getProperty() );
            sql.append("'");
         }
         
         CallableStatement stmt = null;
         stmt = conn.prepareCall(sql.toString());
         ResultSet rs = stmt.executeQuery();
                    
         return rs;
      }
   }

   /**
    * Generate SQL for the attribute functions: length, readonly, binding, typeof
    * @param pfn
    * @return SQL String
    * @throws QueryException
    */
   static private String generateAttributeFunction(ParsedFactorNode pfn) throws QueryException
   {
      String propName    = pfn.getProperty();
      String sqlOperator = null;
      int    valueOf     = pfn.getValueRef();
      Long   length      = null;
      Boolean bool       = null;
      String  type       = null;

      AnnotatedLiteral literal     = pfn.getExpression();
      
      // If binding() or readonly() are used in bare form, implied '== true'
      if( pfn.getOperator() != 0 )
         sqlOperator = DBManager.mapXamOpToSqlOp( pfn.getOperator() );

      StringBuffer singleSQL = new StringBuffer("select xuid from ");
      singleSQL.append( DBManager.VALUETABLE );
      singleSQL.append( " where propname = '");
      singleSQL.append( propName );
      singleSQL.append( "' and " );
      
      // LENGTHOF
      if( valueOf == ParsedFactorNode.VALUE_LENGTH )
      {
         try { length = (Long) literal.getLiteral(); }
         catch( ClassCastException cce ) {
            throw new QueryException( "Length function requires a XAM_INT comparison literal" );
         }
   
         singleSQL.append( DBManager.LENGTH );
         singleSQL.append( sqlOperator );
         singleSQL.append( length.toString() );
      } 
      
      // BINDING READONLY
      else if( (valueOf == ParsedFactorNode.VALUE_BINDING) || 
               (valueOf == ParsedFactorNode.VALUE_READONLY ))
      {
         if( literal != null )
         {  
            try {
               Token boolToken = (Token) literal.getLiteral();
               bool = Boolean.valueOf(boolToken.image);
            }
            catch( ClassCastException cce ) {
               throw new QueryException( "Attribute function requires a XAM_BOOELAN comparison literal" );
            }
         }
         else
         {
            // Not specified, implied true
            bool = Boolean.TRUE;
            sqlOperator = " = ";
         }
   
         if( valueOf == ParsedFactorNode.VALUE_BINDING )
            singleSQL.append( DBManager.BINDING );
         else
            singleSQL.append( DBManager.READONLY );
         singleSQL.append( sqlOperator );
         singleSQL.append( (bool.booleanValue()?DBManager.INT_TRUE:DBManager.INT_FALSE) );
      }
      
      // TYPEOF
      else if( valueOf == ParsedFactorNode.VALUE_TYPEOF )
      {
         try {type = (String) literal.getLiteral(); }
         catch( ClassCastException cce ) {
            throw new QueryException( "Typeof function requires a XAM_STRING comparison literal" );
         }
   
         singleSQL.append( DBManager.PTYPE );
         singleSQL.append( sqlOperator );
         singleSQL.append( " '" );
         singleSQL.append( type );
         singleSQL.append( "'" );
      }
      else
         throw new QueryException( "Unknown attribute function : " + valueOf );
      
      return singleSQL.toString();
   }

   /**
    * Generate SQL for simple property literal comparisons, such as "com.example.prop" = 42
    * @param pfn
    * @return SQL String
    * @throws QueryException
    */
   static private String generatePropertyBinaryComparison(ParsedFactorNode pfn) throws QueryException
   {
      String sql = null;
      String sqlOperator = DBManager.mapXamOpToSqlOp( pfn.getOperator() );
      String propName    = pfn.getProperty();
      AnnotatedLiteral literal  = (AnnotatedLiteral) pfn.getExpression();
     
      switch( literal.getType() ) {
         case TYPE_STRING:
            sql = generateSingleProperty(  propName, 
               XAMLibrary.STYPE_STRING_MIME_TYPE,
               DBManager.SVALUE + sqlOperator + "'" + literal + "'" );
            break;
            
         case TYPE_DATETIME:
            Calendar c = (Calendar) literal.getLiteral();
            sql = generateSingleProperty( propName,
               XAMLibrary.STYPE_DATETIME_MIME_TYPE,
               DBManager.SVALUE + sqlOperator + "'" + ISO8601Date.toISO8601String(c) + "'" );
            break;

         case TYPE_XUID:
            sql = generateSingleProperty( propName,
               XAMLibrary.STYPE_XUID_MIME_TYPE,
               DBManager.SVALUE + sqlOperator + "'" + literal + "'" );
            break;
      
         case TYPE_INTEGER:
            sql = generateSingleNumericProperty( propName, sqlOperator, literal );
            break;
      
         case TYPE_DOUBLE:
            sql = generateSingleNumericProperty( propName, sqlOperator, literal );
            break;
      
         case TYPE_BOOLEAN:
            boolean b = ( (Boolean) (literal.getLiteral())).booleanValue();
            sql = generateSingleProperty( propName,
               XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
               DBManager.BVALUE + sqlOperator + (b?DBManager.INT_TRUE:DBManager.INT_FALSE) );
            break; 
            
         default:
            throw new QueryException( "Unknown literal type, unable to process " + literal + " " + pfn );
      }
      
      return sql;
   }

   static private String generateSingleProperty( String propName, String mimeType, String whereClause ) 
   {
      StringBuffer singleSQL = new StringBuffer("select xuid from ");
      singleSQL.append( DBManager.VALUETABLE );
      singleSQL.append( " where propname='");
      singleSQL.append( propName );
      singleSQL.append( "' and " );
      singleSQL.append( DBManager.PTYPE );
      singleSQL.append( "='");
      singleSQL.append( mimeType );
      singleSQL.append( "' and " );
      singleSQL.append( whereClause );

      return singleSQL.toString();
   } 
   
   /**
    * Select XUIDs for a single property name comparison with a numeric. This is a 
    * special construct because the XAM Architecture specification requires "promotion" of 
    * integer literal comparisons to double when a double property or literal is involved.
    * 
    * This method queries directly to the values table.
    *  
    * @param propName The unadorned simple property name
    * @param operator The comparison operator, from the parsed query
    * @param value The literal value in the operation
    * @return True if the schema is compatible with this version of the DBManager.
    * @throws SQLException 
    */
   static private String generateSingleNumericProperty(
                                                 String           propName, 
                                                 String           operator,
                                                 AnnotatedLiteral value )
   {
      final String SQL_TRUE  = "(bvalue=0 or bvalue<>1)";
      final String SQL_FALSE = "(bvalue=0 and bvalue=1)";
      
      StringBuffer singleSQL = new StringBuffer("select xuid from ");
      singleSQL.append( DBManager.VALUETABLE );
      singleSQL.append( " where propname='");
      singleSQL.append( propName );
      singleSQL.append( "' and (" );

      if( ! value.getType().equals( TypeCode.TYPE_DOUBLE ) )
      {
         // Regular old number.
         singleSQL.append( DBManager.NUMVALUE );
         singleSQL.append( " " );
         singleSQL.append( operator );
         singleSQL.append( " " );
         singleSQL.append( String.valueOf(value));
      }
      else
      {
         Double literal = (Double) value.getLiteral();
         if( literal.equals( Double.NaN ))
            singleSQL.append(SQL_FALSE);  // NaN never compares with anything.
         
         else if( operator.equals(DBManager.SQL_OP_EQ) && 
                  (literal.equals(Double.NEGATIVE_INFINITY) || literal.equals( Double.POSITIVE_INFINITY )) )
            singleSQL.append(SQL_FALSE);  // Never equal infinity
         
         else if( (operator.equals(DBManager.SQL_OP_LT) || operator.equals(DBManager.SQL_OP_LE) ) &&
                  literal.equals(Double.POSITIVE_INFINITY) )
            singleSQL.append(SQL_TRUE);

         else if( (operator.equals(DBManager.SQL_OP_LT) || operator.equals(DBManager.SQL_OP_LE) ) &&
                  literal.equals(Double.NEGATIVE_INFINITY) )
            singleSQL.append(SQL_FALSE);
         
         else if( (operator.equals(DBManager.SQL_OP_GT) || operator.equals(DBManager.SQL_OP_GE)) &&
                  literal.equals(Double.POSITIVE_INFINITY) )
            singleSQL.append(SQL_FALSE);
         
         else if( (operator.equals(DBManager.SQL_OP_GT) || operator.equals(DBManager.SQL_OP_GE)) &&
                  literal.equals(Double.NEGATIVE_INFINITY) )
            singleSQL.append(SQL_TRUE);
         
         else
         {
            // Regular old number.
            singleSQL.append( DBManager.NUMVALUE );
            singleSQL.append( " " );
            singleSQL.append( operator );
            singleSQL.append( " " );
            singleSQL.append( String.valueOf(value));
         }
      }
      singleSQL.append( ") and (type='" );
      singleSQL.append( TypeCode.TYPE_DOUBLE );
      singleSQL.append( "' or type='" );
      singleSQL.append( TypeCode.TYPE_INTEGER );
      singleSQL.append( "')" );
      return singleSQL.toString();
   }
}
