/*
 * Copyright (c) 2008-2009, Sun Microsystems, Inc.
 * Copyright (c) 2008-2009, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Calendar;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;
import java.util.Vector;
import java.util.logging.Logger;

import org.snia.xam.QueryException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XSet;
import org.snia.xam.util.ISO8601Date;
import org.snia.xam.vim.reference.query.AnnotatedLiteral;
import org.snia.xam.vim.reference.query.FactorNode;
import org.snia.xam.vim.reference.query.ParsedFactorNode;
import org.snia.xam.vim.reference.query.ParsedQuery;
import org.snia.xam.vim.reference.query.ParsedTermNode;
import org.snia.xam.vim.reference.query.TermNode;
import org.snia.xam.vim.reference.query.TypeCode;
import org.snia.xam.vim.reference.query.WhereNode;

public class DBSqlTranslator
{
   /* Use temporary tables for generalized XAM query? Set to true to use temp tables, false otherwise. 
    * Using non-temporary tables will leave create permanent tables in the database. Typically this 
    * is only used for debugging.
    */
   private static final boolean USE_TEMP_TABLES = true;
   
   private Connection     m_conn   = null;
   private Logger         m_logger = null;
   private Vector<String> m_readPolicies = null;
   private ContentIndexManager m_contentMgr = null;

   public DBSqlTranslator( Connection conn, ContentIndexManager contentMgr, Logger log, Vector<String> readPolicies )
   {
      m_conn         = conn;
      m_contentMgr   = contentMgr;
      m_logger       = log;
      m_readPolicies = readPolicies;
   }
   
   /**
    * Select query based on generalized XAM QL. This involves creating a temporary table
    * used to support this single query.
    * @param name Name of the job, used to name the table.
    * @param parsed_query Parsed query tree
    * @return The ResultSet of the values.
    * @throws SQLException
    * @throws QueryException
    */
   public ResultSet selectXamQuery( Job            qJob,
                                    String         name, 
                                    ParsedQuery    parsed_query,
                                    Vector<String> accessList ) 
      throws SQLException, QueryException
   {
      m_logger.fine( "-------- Begin Query Job --------");
      String sanitizedTableName = name.replace( ':', '_' );
      m_logger.fine( "Building table: " + sanitizedTableName );

      if( accessList != null )
         addAccessRestrictions( accessList, parsed_query );
      
      Hashtable<String,String> propertyNameMap = createQueryTable( sanitizedTableName, parsed_query );
      
      // Copy values from XFieldValues to temporary query table
      populateQueryTable( qJob, sanitizedTableName, parsed_query, propertyNameMap );
      
      if( parsed_query.getLevel() == 2 )
         populateLevel2Query( qJob, sanitizedTableName, parsed_query, propertyNameMap );
      
      // Generate the SQL
      ResultSet rs = executeGeneralizedXamQuery( sanitizedTableName, parsed_query,
                                                 propertyNameMap );
      return rs;
   }

   /**
    * Add sub-clauses to the where node so that access policies are honored. This uses the vendor dynamic field
    * .vnd.org.snia.xam.reference.access_policy, which is added to the DB (not the XSet itself).
    * 
    * In order to add the restrictions, the property references need to be added in the same manner 
    * as the XAM Query parser would.
    * @param accessList
    * @param parsed_query
    */
   private void addAccessRestrictions(Vector<String> accessList,
                                      ParsedQuery    parsed_query)
   {
      ParsedFactorNode pfn = null;
      WhereNode        accessRestrictor = null;
      
      Enumeration<String> e = accessList.elements();
      
      while( e.hasMoreElements() )
      {
         String policy = e.nextElement();
         pfn = new ParsedFactorNode();
         pfn.setProperty(DBManager.VND_ACCESS_POLICY);
         pfn.setExpression(new AnnotatedLiteral(policy, TypeCode.TYPE_STRING));
         pfn.setOperator(ParsedFactorNode.EQUALS);
         
         if( accessRestrictor == null )
            accessRestrictor = pfn;
         else
         {
            ParsedTermNode ptn = new ParsedTermNode();
            ptn.setLeftNode( pfn );
            ptn.setRightNode( accessRestrictor );
            ptn.setOperator(ParsedTermNode.OPERATOR_OR);
            accessRestrictor = ptn;
         }
      }
      
      if( accessRestrictor != null )
      {
         ParsedTermNode andNode = new ParsedTermNode();
         andNode.setLeftNode(accessRestrictor);
         andNode.setRightNode(parsed_query.getWhereExpression());
         andNode.setOperator(ParsedTermNode.OPERATOR_AND);
         parsed_query.setWhereExpression(andNode);
         
         TypeCode tc = TypeCode.TYPE_STRING;
         parsed_query.getWherePropertyList().put( DBManager.VND_ACCESS_POLICY+tc, tc );
      }
   }

   /**
    * Given a completely generalized query, create the SQL on the temporary table and execute it.
    * @param sanitizedTableName A cleaned up table name, created in the previous phase.
    * @param m_parsed_query The parsed query.
    * @param propertyList The list of required properties in the where clause
    * @param propertyNameMap The list of property aliases, these are the column names.
    * @return The result set from the sql query/
    * @throws QueryException
    * @throws SQLException
    */
   private ResultSet executeGeneralizedXamQuery(String      sanitizedTableName,
                                                ParsedQuery parsed_query,
                                                Hashtable<String, String> propertyNameMap) throws QueryException, SQLException
   {
      StringBuffer sql = new StringBuffer( "select xuid from " );
      if( USE_TEMP_TABLES )
         sql.append( "SESSION." );
      sql.append(sanitizedTableName);
      sql.append( " where " );
      
      WhereNode node = (WhereNode) parsed_query.getWhereExpression(); 
      generateXamSQL( sql, node, propertyNameMap );
      m_logger.fine( "Translated SQL: " + sql );
      
      CallableStatement stmt = null;
      stmt = m_conn.prepareCall(sql.toString());
      ResultSet rs = stmt.executeQuery();
                 
      return rs;
   }

   /**
    * Generate the sql based on the parsed query.
    * @param sql The SQL buffer
    * @param node The Where expression
    * @param propertyAliases The list of column names, keyed by property name
    * @throws QueryException
    */
   static private void generateXamSQL( StringBuffer sql, 
                                WhereNode node,
                                Hashtable<String,String> propertyAliases ) throws QueryException
   {
      if( node.isTerm() )
      {
         TermNode tn = (TermNode) node;
         
         if( tn.isNotExpression() )
            sql.append( " not " );
         sql.append( "(" );
         generateXamSQL( sql, tn.getLeftNode(), propertyAliases );
         sql.append( ")" );
         
         // If the node was not(expr) there will be no right node, only left.
         if( tn.getRightNode() == null )
            return;
         
         switch( tn.getOperator() ) {
            case TermNode.OPERATOR_AND:
               sql.append(" and ");
               break;
            case TermNode.OPERATOR_OR:
               sql.append(" or ");
               break;
            default: throw new QueryException( "Uknown Term Node operator: "  + tn );          
         }
         sql.append( "(" );
         generateXamSQL( sql, tn.getRightNode(), propertyAliases );
         sql.append( ")" );
      }
      else
      {
         FactorNode fn = (FactorNode) node;
         generateXamSQLFactor( sql, fn, propertyAliases );
      }
   }

   /** 
    * Create the sql for the detailed portions of the query, where the properties and values are
    * compared. 
    * @param sql
    * @param fn
    * @throws QueryException
    */
   private static void generateXamSQLFactor( StringBuffer sql, 
                                      FactorNode fn, 
                                      Hashtable<String,String> propertyAliases) throws QueryException
   {
      AnnotatedLiteral literal = fn.getExpression();
      
      int valueRef = fn.getValueRef();
      if( valueRef == FactorNode.VALUE_OF )
      {
         TypeCode tc = literal.getType();
         String columnName = propertyAliases.get(fn.getProperty()+tc);
         String operator   = DBManager.mapXamOpToSqlOp(fn.getOperator());
         
         switch( tc ) {
            case TYPE_STRING:
            case TYPE_XUID:
               sql.append( columnName );
               sql.append( operator );
               sql.append( "'" );
               sql.append( escapeString(literal.toString()) );
               sql.append( "'" );
               break;
            
            case TYPE_DATETIME:
               sql.append( columnName );
               sql.append( operator );
               sql.append( "'" );
               sql.append( ISO8601Date.toISO8601String( (Calendar) literal.getLiteral() ) );
               sql.append( "'" );
               break;
               
            case TYPE_BOOLEAN:
               sql.append( columnName );
               sql.append( operator );
               Boolean b = (Boolean) literal.getLiteral();
               sql.append( (b.booleanValue()?DBManager.INT_TRUE:DBManager.INT_FALSE));
               break;
            
            case TYPE_DOUBLE:
               generateNumericComparison( sql, columnName, operator, literal );
               break;
               
            case TYPE_CONTAINS:
            case TYPE_BEFORE:
            case TYPE_AFTER:
            case TYPE_WITHIN:
               columnName = propertyAliases.get( fn.toString() );
               sql.append( columnName );
               sql.append( operator );
               sql.append( DBManager.INT_TRUE );
               break;
            
            default:
               sql.append( columnName );
               sql.append( operator );
               sql.append( literal );
               break;
         }
      }
      else 
      {
         // If the literal is null, it's most likely something like
         // "where binding(x)" or "where readonly(x)"
         
         int operator = fn.getOperator();
         if( literal == null )
         {
            literal  = AnnotatedLiteral.TRUE;
            operator = FactorNode.EQUALS;
         }
         switch( valueRef ) 
         {
            case FactorNode.VALUE_LENGTH:
               sql.append( propertyAliases.get(fn.getProperty()+TypeCode.TYPE_LENGTH) );
               sql.append( DBManager.mapXamOpToSqlOp(operator));
               sql.append(literal);
               break;
            case FactorNode.VALUE_READONLY:
               sql.append( propertyAliases.get(fn.getProperty()+TypeCode.TYPE_READONLY) );
               sql.append( DBManager.mapXamOpToSqlOp(operator));
               Boolean b_ro = (Boolean) literal.getLiteral();
               sql.append( (b_ro.booleanValue()?DBManager.INT_TRUE:DBManager.INT_FALSE));
               break;
            case FactorNode.VALUE_BINDING:
               sql.append( propertyAliases.get(fn.getProperty()+TypeCode.TYPE_BINDING) );
               sql.append( DBManager.mapXamOpToSqlOp(operator));
               Boolean b_bnd = (Boolean) literal.getLiteral();
               sql.append( (b_bnd.booleanValue()?DBManager.INT_TRUE:DBManager.INT_FALSE));
               break;     
            case FactorNode.VALUE_TYPEOF:
               sql.append( propertyAliases.get(fn.getProperty()+TypeCode.TYPE_TYPEOF) );
               sql.append( DBManager.mapXamOpToSqlOp(operator));
               sql.append( "'" );
               sql.append( literal );
               sql.append( "'" );
               break;
               
            case FactorNode.EXISTS:
               sql.append( propertyAliases.get(fn.getProperty()+TypeCode.TYPE_EXISTS) );
               sql.append( DBManager.mapXamOpToSqlOp(operator));
               Boolean b_exists = (Boolean) literal.getLiteral();
               sql.append( (b_exists.booleanValue()?DBManager.INT_TRUE:DBManager.INT_FALSE));
               break;
               
            default:
               throw new QueryException( "Unknown value reference function: " + valueRef );
         }
      }  
   }

   /** This method generates the SQL for a numeric compare, potentially qualified if the programmer
    * included Inf, -Inf, or Nan
    * @param sql
    * @param columnName
    * @param operator
    * @param literal
    */
   private static void generateNumericComparison(StringBuffer sql,
                                                 String columnName,
                                                 String operator,
                                                 AnnotatedLiteral annotatedLiteral)
   {
      Double literal = (Double) annotatedLiteral.getLiteral();
      
      final String SQL_TRUE  = "(" + columnName + "=0 or " + columnName + "<>1)";
      final String SQL_FALSE = "(" + columnName + "=0 and " + columnName + "=1)";
      
      if( literal.equals( Double.NaN ))
         sql.append(SQL_FALSE);  // NaN never compares with anything.
      
      else if( operator.equals(DBManager.SQL_OP_EQ) && 
               (literal.equals(Double.NEGATIVE_INFINITY) || literal.equals( Double.POSITIVE_INFINITY )) )
         sql.append(SQL_FALSE);  // Never equal infinity
      
      else if( (operator.equals(DBManager.SQL_OP_LT) || operator.equals(DBManager.SQL_OP_LE) ) &&
               literal.equals(Double.POSITIVE_INFINITY) )
         sql.append(SQL_TRUE);

      else if( (operator.equals(DBManager.SQL_OP_LT) || operator.equals(DBManager.SQL_OP_LE) ) &&
               literal.equals(Double.NEGATIVE_INFINITY) )
         sql.append(SQL_FALSE);
      
      else if( (operator.equals(DBManager.SQL_OP_GT) || operator.equals(DBManager.SQL_OP_GE)) &&
               literal.equals(Double.POSITIVE_INFINITY) )
         sql.append(SQL_FALSE);
      
      else if( (operator.equals(DBManager.SQL_OP_GT) || operator.equals(DBManager.SQL_OP_GE)) &&
               literal.equals(Double.NEGATIVE_INFINITY) )
         sql.append(SQL_TRUE);
      
      else
      {
         // Regular old number.
         sql.append( columnName );
         sql.append( operator );
         sql.append( literal );
      }
      
   }

   /**
    * Selects the values from the field values, and inserts them into the query table.
    * @param sanitizedTableName The name for the temporary table. This method is, 
    * unfortunately, rather long and complicated. 
    * @param propertyList The list of properties from the where clause
    * @param propertyNameMap Aliases of the values, with column names.
    * @throws SQLException
    * @throws QueryException
    */
   private void populateQueryTable(
                                   Job        qJob,
                                   String     sanitizedTableName,
                                   ParsedQuery query,
                                   Hashtable<String, String> propertyNameMap
                           ) throws SQLException, QueryException
   {    
      Properties propertyList = query.getWherePropertyList();
      Properties existsList   = query.getExistsPropertyList();

      // This table tracks the values to be retrieved for each property.
      Hashtable<String,boolean[]> valueList = new Hashtable<String,boolean[]>();
      
      StringBuffer selectSQL = new StringBuffer();
      int propertyCount = selectValuesForGeneralizedQuery( sanitizedTableName,
                                   propertyList,
                                   existsList,
                                   propertyNameMap,
                                   valueList,
                                   selectSQL );
      
      // Step 3, populate the final table.
      m_logger.fine( "Selecting properties for a query" + selectSQL.toString() );
      try
      {
         CallableStatement stmt = m_conn.prepareCall(selectSQL.toString());
         ResultSet results = stmt.executeQuery();
         
         int pcol   = results.findColumn( DBManager.PROPNAME );
         int scol   = results.findColumn( DBManager.SVALUE );
         int numcol = results.findColumn( DBManager.NUMVALUE );
         int bcol   = results.findColumn( DBManager.BVALUE );
         int xcol   = results.findColumn( DBManager.XVALUE );
         int rcol   = results.findColumn( DBManager.READONLY );
         int bindcol   = results.findColumn( DBManager.BINDING );
         int lengthcol = results.findColumn( DBManager.LENGTH );
         int typecol   = results.findColumn( DBManager.PTYPE );

         int values_this_xuid   = 0;
         int columns_this_xuid  = 0;
         String current_xuid = null;
         StringBuffer insert = null;
         StringBuffer values = null;
         
         // Process results, inserting rows, as needed.
         while (results.next() && ((qJob != null) && ! qJob.isCancelRequested()) )
         {
            String propname = results.getString(pcol);
            String xuid     = results.getString(xcol);
           
          if( (values_this_xuid == propertyCount) &&
                ! xuid.equals(current_xuid) && 
                (columns_this_xuid > 0) )
          {
             // The old rows processed are sufficient, insert it into the new table.
             insertIntoQueryTable( m_conn, insert, values, current_xuid );
             values_this_xuid = 0;
             current_xuid  = null;
          }

            String value   = null;
            boolean isString = false;
            
            // Starting a new XSet.
            if (current_xuid == null ||
               ! current_xuid.equals(xuid))
            {
               // New XUID, start of new insert
               current_xuid = xuid;
               values_this_xuid = 0;
               columns_this_xuid = 0;
               insert = new StringBuffer( "insert into " );
               if( USE_TEMP_TABLES )
                  insert.append( "SESSION." );
               insert.append(sanitizedTableName);
               insert.append( " (" );
               values = new StringBuffer(" values(" );
            } 

            boolean valueColumns[] = valueList.get(propname);
            
            if( existsList.containsKey(propname) )
            {
               // Special Exists function.
               if( columns_this_xuid++ > 0 )
               {
                  insert.append(",");
                  values.append(",");
               }
               insert.append( propertyNameMap.get(propname+TypeCode.TYPE_EXISTS) );
               values.append( DBManager.INT_TRUE );
            }
            
            if( valueColumns != null )
            for( int type = TypeCode.FIRST.ordinal(); type <= TypeCode.LAST.ordinal(); type++ )
            {
               if( ! valueColumns[type] )
                  continue;
               
               values_this_xuid++;
               
               TypeCode tc = TypeCode.LIST[type];
               String column_name = propertyNameMap.get(propname+tc);
               isString = false;
               switch( tc ) {
                  case TYPE_STRING:
                     value   = escapeString(results.getString(scol));
                     isString = true;
                     break;
                  case TYPE_DATETIME:
                     value   = results.getString(scol);
                     isString = true;
                     break;
                  case TYPE_XUID:
                     value   = results.getString(scol);
                     isString = true;
                     break;
                  case TYPE_INTEGER:
                     value   = String.valueOf(results.getDouble(numcol));
                     break;
                  case TYPE_DOUBLE:
                     value   = String.valueOf(results.getDouble(numcol));
                     break;
                  case TYPE_BOOLEAN:
                     value   = String.valueOf(results.getInt(bcol));
                     break;
                  case TYPE_LENGTH:
                     value   = String.valueOf(results.getInt( lengthcol ));
                     break;
                  case TYPE_READONLY:
                     value   = String.valueOf(results.getInt( rcol ));
                     break;
                  case TYPE_BINDING:
                     value   = String.valueOf(results.getInt( bindcol ));
                     break;  
                  case TYPE_TYPEOF:
                     value   = results.getString(typecol);
                     isString = true;
                     break;
                  default:
                     value = "???";
                  break;
               }

               // Add values to insert statement
               if( columns_this_xuid++ > 0 )
               {
                  insert.append(",");
                  values.append(",");
               }
               insert.append( column_name );
    
               if( isString )
                  values.append( "'" );
               values.append( value );
               if( isString )
                  values.append( "'" );
            }
         }
         
         // There may be one more XSet in the insert/values buffer, flush them
         if( columns_this_xuid > 0 )
            insertIntoQueryTable( m_conn, insert, values, current_xuid );
      }
      catch( SQLException e2 )
      {
         m_logger.severe( "Can't build query table: " + e2 );
         throw e2;
      }
   }

   /**
    * Escapes quotes to keep SQL happy later when performing inserts.
    * Ideally the code should be rewritten to use parameterized insert statements
    * rather than string concatenting sql statements. This may also cause problems 
    * with a RDMS other than Derby (e.g. MySQL).
    * @param string The string to be escaped
    * @return
    */
   private static String escapeString(String str)
   {
      if( str != null )
         return str.replace("\'", "\'\'");
      else
         return null;
   }

   /**
    * Runs a series of Lucene queries to get the level 2 information and adds it to 
    * the temporary table.
    * @param job The query job that is running
    * @param sanitizedTableName Name of the temporary table to populate
    * @param parsed_query The output of the query parser
    * @param propertyNameMap Property aliases
    * @throws SQLException 
    */
   private void populateLevel2Query(Job job,
                                    String sanitizedTableName,
                                    ParsedQuery parsed_query,
                                    Hashtable<String, String> propertyNameMap)
      throws QueryException, SQLException
   {
      m_logger.fine( "Integrating level 2 sub-expressions " );
      Properties level2Fields = parsed_query.getLevel2Fields();
      Enumeration field = level2Fields.elements();
      while( field.hasMoreElements() )
      {
         ParsedFactorNode l2Pfn = (ParsedFactorNode) field.nextElement();
         
         String documents[] = m_contentMgr.search( l2Pfn );
         for( int d = 0; d < documents.length; d++ )
         {
            if( documents[d] == null )
               continue;
            StringBuffer insert = new StringBuffer( "update " );
            if( USE_TEMP_TABLES )
               insert.append( "SESSION." );
            insert.append(sanitizedTableName);
            insert.append( " set " );
            insert.append( propertyNameMap.get( l2Pfn.toString() ) );
            insert.append( "=" );
            insert.append( DBManager.INT_TRUE );
            insert.append( " where xuid='" );
            insert.append( documents[d] );
            insert.append( "'" );
            
            CallableStatement stmt = null;
            try {
               stmt = m_conn.prepareCall(insert.toString());
               stmt.execute();
            } 
            finally {
               if( stmt != null )
                  stmt.close();
            }
         }
      }
   }

   /** Grabs the values from the values table which are required to satisfy the query. 
    * @param sanitizedTableName Cleaned up table name to insert into
    * @param propertyList The where properties
    * @param existsList The exists properties
    * @param propertyNameMap The property aliases
    * @param valueList A <b>side effect</b> set hash which has the values for each property.
    * @return The sql string to select values.
    * @throws QueryException
    */
   static private int selectValuesForGeneralizedQuery(String sanitizedTableName,
                                                      Properties propertyList,
                                                      Properties existsList,
                                                      Hashtable<String, String> propertyNameMap,
                                                      Hashtable<String,boolean[]> valueList,
                                                      StringBuffer sql ) throws QueryException
   {
      // Step 2 build the select statement from the values table
      final String COMMA = ",";
      
      sql.append(  "select " );
      sql.append( DBManager.PROPNAME );
      sql.append( COMMA );
      sql.append( DBManager.SVALUE );
      sql.append( COMMA );
      sql.append( DBManager.NUMVALUE );
      sql.append( COMMA );
      sql.append( DBManager.BVALUE );
      sql.append( COMMA );
      sql.append( DBManager.BINDING );
      sql.append( COMMA );
      sql.append( DBManager.READONLY );
      sql.append( COMMA );
      sql.append( DBManager.LENGTH );
      sql.append( COMMA );
      sql.append( DBManager.PTYPE );
      sql.append( COMMA );
      sql.append( DBManager.XVALUE );
      sql.append( " from " );
      sql.append( DBManager.VALUETABLE );
      sql.append( " where (" );
      
      int propertyCount = 0;
      Enumeration props = propertyList.keys();
      while( props.hasMoreElements() )
      {
         String propertyName = (String) props.nextElement();
         TypeCode type = (TypeCode) propertyList.get(propertyName);
         if( propertyCount++ > 0 )
            sql.append( " or " );
         sql.append( "(" );
         sql.append( DBManager.PROPNAME );
         sql.append( "='" ); 
         
         String cleanPropName = propertyName.substring( 0, propertyName.indexOf(type.toString()));
         if( type.isAttribute() )
         {
            // Clean up the property name
            sql.append( cleanPropName );
            sql.append( "'" );
         }
         else
         {
            sql.append( cleanPropName );
            sql.append( "' and " );
            if( type.equals( TypeCode.TYPE_INTEGER) || type.equals(TypeCode.TYPE_DOUBLE ))
            {
               sql.append("(");
               sql.append( DBManager.PTYPE );
               sql.append( "='" );
               sql.append( TypeCode.TYPE_INTEGER );
               sql.append( "' or " );
               sql.append( DBManager.PTYPE );
               sql.append( "='" );
               sql.append( TypeCode.TYPE_DOUBLE );
               sql.append("')");
            }
            else
            {
               sql.append( DBManager.PTYPE );
               sql.append( "='" );
               sql.append( type.toString() );
               sql.append( "'" );
            }
         }
         sql.append( ")");
         boolean values[] = valueList.get(cleanPropName);
         if( values == null )
         {
            values = new boolean[TypeCode.TYPE_OTHER.ordinal()];
            valueList.put( cleanPropName, values );
         }
         if( ! values[type.ordinal()] )
            values[type.ordinal()] = true;
      }
      
      // EXISTS properties
      props = existsList.keys();
      int tmpCount = propertyCount; // Don't increment propertyCount for exists clauses.
      while( props.hasMoreElements() )
      {
         String propertyName = (String) props.nextElement();
         TypeCode tc = (TypeCode) existsList.get(propertyName);
         if( ! tc.equals( TypeCode.TYPE_EXISTS ))
            throw new QueryException( "Parser malfunction! Exists function should not have entered " + propertyName + " exits list." );
         if( tmpCount++ > 0 )
            sql.append( " or " );
         sql.append( "(" );
         sql.append( DBManager.PROPNAME ); 
         sql.append( "='" ); 
         sql.append(propertyName);
         sql.append( "')" );
      }
      sql.append( ") order by " );
      sql.append( DBManager.XVALUE );
      
      return propertyCount;
   }

   /** This is a small worker method used by populateQueryTable to insert rows into 
    * the query table. 
    * @param sql The StringBuffer containing the current insert statement
    * @param xuid The XUID these values are associated with
    * @throws SQLException 
    */
   private static void insertIntoQueryTable( Connection conn,
                                             StringBuffer sql, 
                                             StringBuffer values, 
                                             String xuid ) throws SQLException
   {
      // The old rows processed are sufficient, insert it into the new table.
      sql.append( ", xuid" );
      sql.append( ")" );
      sql.append( values );
      sql.append( ",'" );
      sql.append( xuid );
      sql.append( "')" );

      CallableStatement stmt = null;
      try {
         stmt = conn.prepareCall(sql.toString());
         stmt.execute();
      } 
      finally {
         if( stmt != null )
            stmt.close();
      }
   }
   
   /**
    * Create a temporary query table, return a hash table with the mapped property names.
    * @return The hashtable of the property aliases to be used for the generated table.
    * @throws SQLException 
    * @throws QueryException 
    */
   private Hashtable<String,String> createQueryTable( String table_name, 
                                                      ParsedQuery query )
          throws SQLException, QueryException
   {
      Properties props      = query.getWherePropertyList();
      Properties existProps = query.getExistsPropertyList();
      Properties level2Fields = query.getLevel2Fields();
      
     // Step 1, 
      Enumeration propertyNames = props.keys();
      Hashtable<String,String>   propertyAliases = new Hashtable<String,String>();

      // Create the temporary table and create the aliases at the same time.
      // Alias property names must be create since properties are UTF-8 and
      // databases have limitations on column names. There could also be a query like:
      //      select ".xset.xuid" where "property" = 42 or "property" = true or "property" = 'string' or "property" = date('2009-03-01T00:00.00-08.00')
      // Which is a legal query and a legal data situation.
 
      StringBuffer sql = null;
      if( USE_TEMP_TABLES )
         sql = new StringBuffer("declare global temporary table ");
      else
         sql = new StringBuffer("create table "); // For debugging
      
      if( props.size() == 0 )
      {
         // Either the where clause is empty, or an expression of non-possesion rule
         // properties (e.g. exists sub-clauses). This is an implied "scan everything".
         // In this case, force a column for .xset.xuid, even though it won't be used in 
         // the expression.
         props.put( XSet.XAM_XUID, TypeCode.TYPE_XUID );
      }
      
      try
      {
         Enumeration propertyList = props.keys();
         int x = 0;
         sql.append(table_name);
         sql.append("(");
         
         // Do the normal value expressions
         while( propertyList.hasMoreElements() )
         {
            String propertyName = (String) propertyList.nextElement();
            TypeCode tc = (TypeCode) props.get(propertyName);
            
            String columnName = null;
            String propertyAlias = propertyAliases.get(propertyName);
            if( propertyAlias == null )
            {
               propertyAlias = "p" + x;
               propertyAliases.put( propertyName, propertyAlias );
            }
            
            columnName = propertyAlias;
            
            if ((x++) > 0) sql.append(",");
            
            sql.append(columnName);

            String sqlType = mapLiteralTypeToSqlType(tc);
            sql.append( " " );
            sql.append( sqlType );
         }
         
         // Do the exist properties
         propertyList = existProps.keys();
         while( propertyList.hasMoreElements() )
         {
            String propertyName = (String) propertyList.nextElement();
            String encodedPropertyName = propertyName + TypeCode.TYPE_EXISTS;
            String propertyAlias = propertyAliases.get(encodedPropertyName);
            if( propertyAlias == null )
            {
               propertyAlias = "p" + x;
               propertyAliases.put( encodedPropertyName, propertyAlias );
            }
            if( (x++) > 0 ) sql.append(",");
            
            sql.append( propertyAlias );
            sql.append( " smallint default 0" );
         }
         
         // Do the level 2 expressions.
         if( query.getLevel() == 2 )
         {
            Enumeration l2 = level2Fields.keys();
            while( l2.hasMoreElements() )
            {
               String expressionKey = (String) l2.nextElement();
               ParsedFactorNode pfn = (ParsedFactorNode) level2Fields.get(expressionKey);
               TypeCode tc = pfn.getExpression().getType();
               String encodedFieldName = expressionKey;
               String fieldAlias = propertyAliases.get(encodedFieldName);
               if( fieldAlias == null )
               {
                  fieldAlias = "p" + x;
                  propertyAliases.put( encodedFieldName, fieldAlias );
               }
               if( (x++ > 0) ) sql.append(",");
            
               sql.append( fieldAlias );
               sql.append( " smallint default 0" );
            }
         }
         
         // Finish off the table
         sql.append(",xuid varchar(200)");
         sql.append(")" );
         if( USE_TEMP_TABLES )
            sql.append( " on commit preserve rows not logged");
         m_logger.fine( "Building temporary table: " + sql );
         CallableStatement stmt = m_conn.prepareCall(sql.toString());
         stmt.execute();
         stmt.close();
      }
      catch (SQLException e)
      {
         m_logger.severe( "Can not create query table: " + sql.toString() );
         throw e;
      }
      return propertyAliases;
   }

   /**
    * Given a literal type, return the correct SQL type.
    * @param literal
    * @return
    * @throws QueryException
    */
   private static String mapLiteralTypeToSqlType(TypeCode literalType) throws QueryException
   {
      switch( literalType ) {
         case TYPE_STRING:
            return DBManager.SQL_TYPE_STRING;
      
         case TYPE_DATETIME:
            return DBManager.SQL_TYPE_STRING;
      
         case TYPE_XUID:
            return DBManager.SQL_TYPE_STRING;
      
         case TYPE_INTEGER:
            return DBManager.SQL_TYPE_INTEGER;
      
         case TYPE_DOUBLE:
            return DBManager.SQL_TYPE_DOUBLE;
      
         case TYPE_BOOLEAN:
            return DBManager.SQL_TYPE_BOOLEAN;
            
         case TYPE_READONLY:
            return DBManager.SQL_TYPE_BOOLEAN;
            
         case TYPE_BINDING:
            return DBManager.SQL_TYPE_BOOLEAN;
            
         case TYPE_TYPEOF:
            return DBManager.SQL_TYPE_STRING;
            
         case TYPE_LENGTH:
            return DBManager.SQL_TYPE_INTEGER;
      
         case TYPE_CONTAINS:
            return DBManager.SQL_TYPE_BOOLEAN;
         
         case TYPE_BEFORE:
            return DBManager.SQL_TYPE_BOOLEAN;
         
         case TYPE_AFTER:
            return DBManager.SQL_TYPE_BOOLEAN;
         
         case TYPE_WITHIN:
            return DBManager.SQL_TYPE_BOOLEAN;
         
         default:
            throw new QueryException( "Unknown type, can not generate query table: " + literalType);
      }
   }

   /**
    * Given the XAM Mime type, return the TypeCode
    * @param mime
    * @return
    */
   TypeCode mapFromMIME( String mime )
   {
      if( mime.equals(XAMLibrary.STYPE_BOOLEAN_MIME_TYPE) )
         return TypeCode.TYPE_BOOLEAN;
      
      else if( mime.equals(XAMLibrary.STYPE_STRING_MIME_TYPE))
         return TypeCode.TYPE_STRING;
      
      else if( mime.equals(XAMLibrary.STYPE_DATETIME_MIME_TYPE))
         return TypeCode.TYPE_DATETIME;
      
      else if( mime.equals(XAMLibrary.STYPE_INT_MIME_TYPE))
         return TypeCode.TYPE_INTEGER;
      
      else if( mime.equals(XAMLibrary.STYPE_DOUBLE_MIME_TYPE))
         return TypeCode.TYPE_DOUBLE;
      
      else if( mime.equals(XAMLibrary.STYPE_XUID_MIME_TYPE))
         return TypeCode.TYPE_XUID;
      
      else return TypeCode.TYPE_OTHER;
   }
 }
