/*
 *  Copyright (c) 2008, Sun Microsystems, Inc.
 *  Copyright (c) 2008, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.URL;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;
import org.xml.sax.helpers.DefaultHandler;

/**
 * 
 */
public class DefaultXMLWriter extends DefaultHandler
{

    private static final char QUOTE = '\'';
    private static final String INDENT = "  ";
    private static final int INDENT_LENGTH = INDENT.length();
    /** 
     * An upper bound on the number of characters that can be printed to the output writer 
     * per one line. A new line is started every time when the number of characters in a line
     * (excluding indent spaces) after @link #startElement or @link #addAttribute exceeds MAX.
     */
    private static final int MAX = 75;
    private static final String DEFAULT_ENCODING = "UTF-8";
    // variables ................................................................
    //
    private OutputStreamWriter writer;
    // document locator
    private Locator locator;
    //
    private String encoding = null;
    // true if the currently written XML element has at least one sub-element
    private boolean hasContent = true; // inited by true due to condition in @link #startElement
    // true if the currently written XML element has some characters in its content

    private boolean hasCharacters = false;
    // indentation spaces to be currently used
    private String indentSpaces = "";
    // number of charecters written on current line so far (excluding indent spaces)
    private int charsCount = 0;

    // init .....................................................................
    public DefaultXMLWriter()
    {
    }

    public DefaultXMLWriter(OutputStreamWriter writer, String encoding)
    {
        this.writer = writer;
        this.encoding = encoding;
    }

    public void init() 
    throws SAXException
    {
        hasContent = true;
        hasCharacters = true;
        indentSpaces = "";
        charsCount = 0;

        if (writer == null) {
            try {
                OutputStream ostr = new URL(locator.getSystemId())
                    .openConnection().getOutputStream();
                writer = new OutputStreamWriter(ostr);
            }
            catch (IOException e) {
                throw new SAXException(e);
            }
        }
        write("");
        writeln();
    }

    // methods ..................................................................
    public Writer getWriter()
    {
        return writer;
    }

    /**
     * Writes a string to the output writer.
     *
     * @param text a string to be written in the output writer
     */
    private void write(String text) 
    throws SAXException
    {
        try {
            writer.write(text);
        }
        catch (IOException e) {
            throw new SAXException(e);
        }
    }

    /**
     * Writes end of line to the output writer.
     */
    private void writeln() 
    throws SAXException
    {
        try {
            writer.write("\r\n");
        }
        catch (IOException e) {
            throw new SAXException(e);
        }
        charsCount = 0;
    }

    /**
     * Locates occurences of special XML charecters (like '<', '&', etc.) 
     * in a string and replaces them by sequences of the form &X...X;
     *
     * In addition, if replaceWhitechars parameter is true, '\n', '\r' and '\t' 
     * are replaced by &#NN; sequences. 
     */
    private String replaceSpecialChars(String s, boolean replaceWhitechars)
    {
        int length = s.length();
        char[] chars = new char[6 * length];
        int n = 0;
        for (int x = 0; x < length; x++) {
            char c = s.charAt(x);
            switch (c) {
                case '&':
                    chars[n] = '&';
                    n++;
                    chars[n] = 'a';
                    n++;
                    chars[n] = 'm';
                    n++;
                    chars[n] = 'p';
                    n++;
                    chars[n] = ';';
                    n++;
                    break;
                case '\'':
                    chars[n] = '&';
                    n++;
                    chars[n] = 'a';
                    n++;
                    chars[n] = 'p';
                    n++;
                    chars[n] = 'o';
                    n++;
                    chars[n] = 's';
                    n++;
                    chars[n] = ';';
                    n++;
                    break;
                case '\"':
                    chars[n] = '&';
                    n++;
                    chars[n] = 'q';
                    n++;
                    chars[n] = 'u';
                    n++;
                    chars[n] = 'o';
                    n++;
                    chars[n] = 't';
                    n++;
                    chars[n] = ';';
                    n++;
                    break;
                case '<':
                    chars[n] = '&';
                    n++;
                    chars[n] = 'l';
                    n++;
                    chars[n] = 't';
                    n++;
                    chars[n] = ';';
                    n++;
                    break;
                case '>':
                    chars[n] = '&';
                    n++;
                    chars[n] = 'g';
                    n++;
                    chars[n] = 't';
                    n++;
                    chars[n] = ';';
                    n++;
                    break;
                default:
                    if (replaceWhitechars) {
                        switch (c) {
                            case '\n':
                                chars[n] = '&';
                                n++;
                                chars[n] = '#';
                                n++;
                                chars[n] = '1';
                                n++;
                                chars[n] = '0';
                                n++;
                                chars[n] = ';';
                                n++;
                                break;
                            case '\r':
                                chars[n] = '&';
                                n++;
                                chars[n] = '#';
                                n++;
                                chars[n] = '1';
                                n++;
                                chars[n] = '3';
                                n++;
                                chars[n] = ';';
                                n++;
                                break;
                            case '\t':
                                chars[n] = '&';
                                n++;
                                chars[n] = '#';
                                n++;
                                chars[n] = '9';
                                n++;
                                chars[n] = ';';
                                n++;
                                break;
                            default:
                                chars[n] = c;
                                n++;
                        }
                    } else {
                        chars[n] = c;
                        n++;
                    }
            } // switch

        } // for

        return new String(chars, 0, n);
    }

    /**
     * Writes start of a XML element to the output writer.
     *
     * @param name name of the XML element to be written
     */
    private void startElement(String name, Attributes attrs) 
    throws SAXException
    {
        if (!hasContent && !hasCharacters) {
            write(">");
            writeln();
        }
        hasContent = false;
        hasCharacters = false;
        write(indentSpaces);
        write("<" + name);
        charsCount += name.length() + 1;
        indentSpaces = indentSpaces + INDENT;
        for (int x = 0; x < attrs.getLength(); x++) {
            addAttribute(attrs.getQName(x), attrs.getValue(x));
        }
    }

    /**
     * Writes end of an XML element to the output writer.
     *
     * @param name name of the XML element to be written
     */
    private void endElement(String name) throws SAXException
    {
        indentSpaces = indentSpaces.substring(0, 
                indentSpaces.length() - INDENT_LENGTH);
        if (hasContent) {
            write(indentSpaces);
            write("");
        } else if (hasCharacters) {
            write("");
        } else {
            write("/>");
        }
        writeln();
        hasContent = true;
    }

    /**
     * Writes an attribute of the currenly written XML elemnt to the 
     * output writer.
     *
     * @param name attribute name
     * @param value value of the attribute
     */
    private void addAttribute(String name, String value) 
    throws SAXException
    {
        value = replaceSpecialChars(value, true);
        // [PENDING] ??? can be special characters in name too ???
        if (charsCount > MAX) {
            writeln();
            write(indentSpaces);
        } else {
            write(" ");
            charsCount++;
        }
        write(name + "=" + QUOTE + value + QUOTE);
        charsCount += name.length() + value.length() + 3;
    }

    /**
     * Writes characters into body of the currenly written XML elemnt.
     * Before the string is written, @link #replaceSpecialChars is called
     * on it to replace special XML characters.
     *
     * @param text string to be written
     */
    private void characters(String text) throws SAXException
    {
        text = replaceSpecialChars(text, false);
        if (!hasContent) {
            write(">");
        }
        write(text);
        hasCharacters = true;
    }

    // org.xml.sax.Contentwriter interface implementation ......................
    public void startDocument() 
    throws org.xml.sax.SAXException
    {
        init();
    }

    public void endDocument() 
    throws org.xml.sax.SAXException
    {
        try {
            writer.flush();
            writer.close();
        }
        catch (IOException e) {
            throw new SAXException(e);
        }
    }

    public void startElement(String uri, String sName, String qName, 
            Attributes attributes) 
    throws org.xml.sax.SAXException
    {
        startElement(qName, attributes);
    }

    public void endElement(String uri, String sName, String qName) 
    throws org.xml.sax.SAXException
    {
        endElement(qName);
    }

    public void characters(char[] buf, int offset, int len) 
    throws org.xml.sax.SAXException
    {
        characters(new String(buf, offset, len));
    }

    public void setDocumentLocator(Locator locator)
    {
        this.locator = locator;
    }
    
    public static void main(String[] args) throws Exception {
        OutputStreamWriter streamWriter = new OutputStreamWriter(System.out);
        DefaultXMLWriter writer = new DefaultXMLWriter(streamWriter, "UTF-8");
        writer.startDocument();
        
        AttributesImpl attrs = new AttributesImpl();
        attrs.addAttribute("123", "1234", "readOnly", "boolean", "true");
        attrs.addAttribute("123", "1234", "value", "string", "test");
        writer.startElement("", "", "properties", attrs);
        attrs = new AttributesImpl();
        attrs.addAttribute("123", "1234", "readOnly", "boolean", "true");
        attrs.addAttribute("123", "1234", "value", "string", "test");
        writer.startElement("", "", "data", attrs);
        writer.endElement("", "", "data");
        writer.endElement("","", "properties");;
        writer.endDocument();
    }
}
