/*
 *  Copyright (c) 2009, Sun Microsystems, Inc.
 *  Copyright (c) 2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Hashtable;
import java.util.logging.Logger;
import java.util.logging.Level;

import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.XAMException;
import org.snia.xam.XSet;
import org.snia.xam.XUID;

/**
 * Performs auto deletion functions, asynchronous from API, and has "shred" functionality.
 * NOTE: The shred functionality contained here is to provide functionality that is 
 * exercised from the API and is not intended to provide a secure deletion function.
 */
public class DispositionManager
   implements Runnable
{
   private long m_daemon_period = 5 * 60 * 1000; /* 5 minutes */
   
   private static Hashtable<String,DispositionManager> s_instance = 
         new Hashtable<String,DispositionManager>();
   private static Boolean   s_lock     = Boolean.valueOf(false);
 
   private ReferenceXSystem m_xsystem_instance;
   private Logger           m_logger;
   private Thread           m_daemon;
   private boolean          m_cancel_execution = false;
   
   public DispositionManager( ReferenceXSystem xsystemInstance ) 
      throws XAMException
   {
      m_xsystem_instance = xsystemInstance;
      m_logger = xsystemInstance.getLogger();
      
      // Get the period, if it's configured via the properties
      try { 
         long tmp = m_xsystem_instance.getLong(ReferenceXSystem.PROPERTY_AUTODELETE_PERIOD) * 1000;
         m_daemon_period = tmp;
      }
      catch( FieldDoesNotExistException fdne ) { /* Ignore it, use built in */ }
      catch (XAMException e)
      {
         m_logger.log( Level.WARNING, "Unexpected error while getting autodelete daemon period " + e.getMessage() );
      }
      
      if( m_daemon_period < 0 )
         m_logger.log( Level.INFO, "Disabling the autodelete daemon" );
      else 
      {
         m_daemon = new Thread( this, "SNIA Reference VIM Disposition Manager: " + 
            m_xsystem_instance.getPersistenceMgr().getRepositoryLocation() );
         m_daemon.setDaemon(true);
         m_daemon.start();
      }
   }

   /**
    * Get the unique instance of the autodelete daemon for a specific store directory.
    * @param xsys
    * @return The disposition manager instance for the store instance directory. 
    * @throws ClassNotFoundException
    * @throws SQLException
    * @throws XAMException
    */
   public static DispositionManager GetInstance( ReferenceXSystem xsys ) throws ClassNotFoundException, SQLException, XAMException
   {
      synchronized(s_lock) {
         String storeLocation = xsys.getPersistenceMgr().getRepositoryLocation();
         DispositionManager instance = s_instance.get(storeLocation);
         if( instance == null )
         {
            instance = new DispositionManager( xsys );
            s_instance.put( storeLocation, instance );
         }
         
         return instance;
      }
   }
   
   /** The heart of the Auto delete engine. This method is called by the Timer task and
    *  will search for auto delete enabled XSets. Each of the XSets is examined and if no
    *  longer under retention, they are deleted. 
    *  
    *  Suggestion : Use a better algorithm, opening each XSet in turn and evaluating the retention
    *  periods is rather expensive. Could compute an expiration time on setting 
    *  the retention, and set a vendor field with expiration date.
    */
   public void run()
   {
     while( ! m_cancel_execution )
      {
        if( m_logger != null )
           m_logger.log( Level.FINER, "Running autodelete" );
        
        try { Thread.sleep(m_daemon_period); }
        catch( InterruptedException ie ) { /* ignore */ }
        autodeleteSweep();
      }
   }
   
   /**
    * Called once per run cycle. to process all the items in the store.
    * which need to be deleted.
    *
    * This does two separate queries, once for explicit settings, and a second one for policy settings.
    * Doing these queries separately allows the DBM to optimize the queries to direct table query instead
    * of the more complicated process where a temporary table is created. This is exposing some behavior from 
    * the DBM, but it's worth the execution optimization for this process.
    */
   private void autodeleteSweep()
   {
      ResultSet results = null;
      try {
         DBManager dbm = m_xsystem_instance.getDBManager();
         
         ReferenceVIMPersistenceMgr pMgr = m_xsystem_instance.getPersistenceMgr();
         
         // Note: It has been suggested that running the query with a sub clause of ".xset.hold" = false will
         // speed up the query, the problem here is that the hold boolean is not present in all cases. Adding 
         // that property will ignore most XSets which are auto delete and not being held. Best to evaluate the 
         // hold at delete time.
         
         String tempId = "DispositionManager." + String.valueOf( System.currentTimeMillis() );
         results = dbm.executeXamQuery( tempId, 
            "select \"" + XSet.XAM_XUID + "\" where \"" + 
            XSet.XAM_DELETION_AUTODELETE + "\"",
            null /* Read Access policies */ ); 
         processResults( results, dbm, pMgr );

         results = dbm.executeXamQuery( tempId, 
            "select \"" + XSet.XAM_XUID + "\" where \"" + 
            XSet.XAM_DELETION_AUTODELETE_POLICY + "\" <> ''",
            null /* Read Access policies */ ); 
         processResults( results, dbm, pMgr );
      }
      catch( Exception e )
      {
         m_logger.severe( "Can't successfully run autodelete: " + e.toString() );
      }
      finally
      {
         try {
            if( results != null )
            {
               results.getStatement().close();
               results.close();
            }
         }
         catch( SQLException sqe )
         {
            m_logger.log( Level.SEVERE, "Error closing results set, shouldn't happen here. " + 
                  sqe.toString()  );
         }
      }
   }

   /** 
    * Walk the each XSet in the results and delete them. Double check to make sure autodelete is truly called for.
    * @param results The query results
    * @param dbm The Database manager, so the XSet can be removed from its tables.
    * @param pMgr The persistence manager so the XSet can be removed from the store.
    * @throws SQLException
    */
   private void processResults(ResultSet results,
                               DBManager dbm,
                               ReferenceVIMPersistenceMgr pMgr ) throws SQLException
   {
      while( results.next() && ! m_cancel_execution )
      {
         String base64Xuid = results.getString(1);
         try 
         {
            XUID xsetid = m_xsystem_instance.createXUID(base64Xuid);
            
            ReferenceXSet xset = new ReferenceXSet(m_xsystem_instance, xsetid );
            if( ! xset.getActualAutoDelete() )
            {
               m_logger.log( Level.SEVERE, 
                  m_daemon.getName() + ": Unexpected XSet in autodelete, doesn't have autodelete set " + xsetid );
            }

            // check to see if the XSet is retained
            if( m_xsystem_instance.isXSetRetained(xsetid) ||
                m_xsystem_instance.isXSetUnderHold(xsetid) ||
                JobMgr.isJobRunningForXSet(xset)) 
               xset.close();
            else  
            { 
               boolean shred = false;
               shred = xset.getActualShred();
               xset.close();
               dbm.deleteXSet( m_xsystem_instance, xsetid );
               pMgr.deleteXSet(xsetid, shred);
               m_logger.log( Level.FINEST, "Autodelete " + xsetid.toString() );
            }
         }
        catch( XAMException xe )
        {
           // Don't print stack trace, don't complain too much (at least be at warning level)
           // so the log file doesn't get to full of petty complaints.
           m_logger.log( Level.WARNING, 
              m_daemon.getName() + ": Unexpected XAM Error in autodelete " + xe.getMessage() );
        }
      }
   }

   /**
    * Cancel the execution of the auto delete daemon.
    */
   public boolean cancel()
   {
      m_cancel_execution = true;
      return true;
   }
}
