/*
 *  Copyright (c) 2009, Sun Microsystems, Inc.
 *  Copyright (c) 2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.util.Hashtable;

/**
 * These objects represent different disposition policies which can be applied to an
 * XSet, indicating how to dispose the XSet when it is deleted. These policies are
 * rather "canned" in that there is no external policy configuration mechanism 
 * supported.
 */
public class DispositionPolicy
{
   public static final String DISPOSITION_AUTODELETE           = "org.snia.refvim.disposition.autodelete";
   public static final String DISPOSITION_SHRED                = "org.snia.refvim.disposition.shred";
   public static final String DISPOSITION_AUTODELETE_AND_SHRED = "org.snia.refvim.disposition.autodelete.and.shred";

   static  Hashtable<String,DispositionPolicy> s_policies;
   static  Boolean s_lock = new Boolean(false);
   
   boolean m_autodelete = false;
   boolean m_shred      = false;
   
   /**
    * This constructor will only be available for the static class constructor
    * "createPolicies".
    * 
    * @param autodelete
    * @param shred
    */
   private DispositionPolicy( boolean autodelete, boolean shred )
   {
      m_autodelete = autodelete;
      m_shred      = shred;
   }
   
   public static DispositionPolicy GetPolicy( String name )
   {
      synchronized( s_lock )
      {
         if( s_policies == null )
         {
            s_policies = new Hashtable<String,DispositionPolicy>();
            s_policies.put( DISPOSITION_AUTODELETE, 
               new DispositionPolicy( true, false ));
            s_policies.put( DISPOSITION_SHRED, 
               new DispositionPolicy( false, true ));
            s_policies.put( DISPOSITION_AUTODELETE_AND_SHRED, 
               new DispositionPolicy( true, true ));
         }
      }
      return s_policies.get( name );
   }
   
   /**
    * Get the autodelete status of the policy
    * @return True of the XSet should deleted, False otherwise.
    */
   public boolean getAutodelete() { return m_autodelete; }
   
   /**
    * Get the shred status of the policy
    * @return True of the XSet should shredded, False otherwise.
    */
   public boolean getShred() { return m_shred; }
   
   /**
    * Test a given policy to make sure it is autodelete enabled.
    * @param policy
    * @return True if the policy is set for auto delete, false otherwise.
    */
   public static boolean isAutodelete( String policy )
   {
      boolean autodel = false;
      DispositionPolicy p = GetPolicy(policy);
      if( p != null )
         autodel = p.getAutodelete();
      return autodel;
   }

   /**
    * Test a given policy to make sure it is shred enabled.
    * @param policy
    * @return True if the policy is set for shred, false otherwise.
    */
   public static boolean isShred( String policy )
   {
      boolean shred = false;
      DispositionPolicy p = GetPolicy(policy);
      if( p != null )
         shred = p.getAutodelete();
      return shred;
   }
}
