/*
 * Copyright (c) 2008, Sun Microsystems, Inc.
 * Copyright (c) 2008, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.TimeZone;
import java.util.logging.Logger;

import org.snia.xam.ObjectInUseException;
import org.snia.xam.XAMException;
import org.snia.xam.XSet;

/**
 * Job manager class is responsible for submitting, keeping track of and 
 * canceling jobs.
 * 
 */
public class JobMgr {
    
    /**
     * Map to keep track of jobs created and submitted.  The map's key is the 
     * job identifier and its value is a reference to its Job object.  The map
     * is used by the JobMgr to access a Job when a call is made to cancel
     * the job, determine if the job is running, and determine the total number
     * of jobs running within the ReferenceXSystem.
     */
    private static HashMap<String, ReferenceJob> jobMap = 
            new HashMap<String, ReferenceJob>();
    /**
     * Map to keep track of the relationship between XSet instances and their
     * respective job identifiers.  This map is useful when the XSet instance
     * has submitted the job, but has not been committed. Since the XSet will
     * not have a XUID until it's committed, being able to pass in the XSet
     * instance and find its associated job identifier is important.
     */
    private static HashMap<ReferenceXSet, String> xsetInstanceMap = 
            new HashMap<ReferenceXSet, String>();
    /**
     * Map to keep track of the relationship between the xset identifier, a xuid,
     * and the XSet instance responsible for submitting the job.  This map is 
     * useful when the XSet instance has been committed since all XSets at the
     * time of commit have an xuid.
     */
    private static HashMap<String, ReferenceXSet> xuidMap = 
            new HashMap<String, ReferenceXSet>();
    
    Logger m_logger = null;
    
    /*
     * Job counter is incremented for each submitted job.  When it is 
     * combined with the datetime of submission, it provides a unique jobID. 
     */
    private static int jobCounter = 0;
    
    /**
     * Flag for debugging purposes only
     */
    private static boolean debug = false;

    /**
     * Calls the cancel method on the job associated with the specified
     * XSet instance.  If wait is set to true, then the method will wait for
     * the job to be successfully cancelled until the timeout duration has
     * passed.  
     *
     * @param xset The XSet instance whose associated job needs to be cancelled.
     * @param wait flag indicating whether or not the method should wait for
     *             the job to be cancelled prior to returning.
     *
     * @throws RRMException   If job cancellation fails.
     */
    public static void cancelJobForXSet(ReferenceXSet xset, boolean wait)
                                                        throws XAMException {
        ReferenceJob job = getJobRunningForXSet(xset);
        if (job != null) {
            JobMgr.cancel(job.getID(), wait);
        }
     }
    
    /**
     * Checks to see if the job associated with the XSet instance is running.
     *
     * @param xset The XSet instance whose associated job needs to be checked
     *             to determine if it is still running.
     * 
     * @return ReferenceJob running job associated with the XSet.  If the job
     *                      is not running, then null is returned.
     *
     * @throws RRMException   
     */
    public static  ReferenceJob getJobRunningForXSet(ReferenceXSet xset) 
            throws XAMException {
        boolean isRunning = false;
        ReferenceJob job = null;
        if (xset != null) {
            /* The xsetInstanceMap contains XSet instances and their respective 
             job ID's at the time the jobs were submitted -- the XSet instances
             may have been committed and contain an xuid or they may NOT be 
             committed yet and the xuid is null.*/
            synchronized(JobMgr.class) {
                if (xsetInstanceMap.containsKey(xset)) {
                    // check to see if the job is still active
                    job = jobMap.get(xsetInstanceMap.get(xset));
                    if (job != null && !job.getJobStatus().isDone()) {
                        isRunning = true;
                    }
                } else {
                    // Make sure have valid XUID before check
                    if (xset.getXUIDString().length() != 0) {
                        /* The xset may be a different instance than the set 
                         that submitted the job -- thus, need to iterate 
                         through all ReferenceXSets in the jobMap and see 
                         if a job is running at all.*/
                        Iterator<ReferenceJob> jobs = jobMap.values().iterator();
                        while (jobs.hasNext()) {
                            job = jobs.next();
                            ReferenceXSet refXSet = job.getXSet();
                            if (refXSet.sameXUID(xset) && 
                                            !job.getJobStatus().isDone()) {
                                isRunning = true;
                                break;
                            }
                        }
                    }
                }
            }
            
        }
        return isRunning ? job : null;
    }
    
    /**
     * Checks to see if the job associated with the XSet instance is running.
     *
     * @param xset The XSet instance whose associated job needs to be checked
     *             to determine if it is still running.
     * 
     * @return boolean flag indicating if the job associated with the XSet
     *                 is still running (true) or not (false)
     *
     * @throws RRMException   
     */
    public static boolean isJobRunningForXSet(ReferenceXSet xset)
                                                        throws XAMException {
        boolean isRunning = false;
        ReferenceJob job = JobMgr.getJobRunningForXSet(xset);
        if (job != null) {
            isRunning = true;
        }
        return isRunning;
    }
    
    /**
     * Checks to see if the XSet instance specified was the actual XSet
     * responsible for submitting its associated job.
     *
     * @param xset The XSet instance that will be checked to see if it was
     *             the XSet responsible for submitting its assoc. job.
     * 
     * @return ReferenceJob returns the job associated with the specified XSet
     *                      if there is a job running, otherwise null
     *
     * @throws RRMException   
     */
    protected static ReferenceJob getJobSubmittedByXSet(ReferenceXSet xset)
                                                        throws XAMException {
        ReferenceJob job = null;
        boolean inMap = false;
        synchronized(JobMgr.class) {
            inMap = xsetInstanceMap.containsKey(xset);
            if (inMap) {
                job = jobMap.get(xsetInstanceMap.get(xset));
            }
        }
        // debugging information only
        if (debug) {
            if (inMap) {
                System.out.println("JobMgr::getJobSubmittedByXSet -- " + 
                    "xset, " + xset.toString() + ", is in xsetInstanceMap");
                if (job != null)
                    System.out.println("JobMgr::getJobSubmittedByXSet -- " + 
                                "Job is found in jobMap: " + job.getID());
            } else {
                System.out.println("JobMgr::getJobSubmittedByXSet -- xset " +
                    "is NOT in the xsetInstanceMap: " + xset.toString());
            }
        }
        return job;
    }
    
    /**
     * Checks to see if the job was submitted by the specified XSet instance
     * and if the job is running.
     *
     * @param xset The XSet instance that will be checked
     * 
     * @return boolean flag indicating if the job associated with the XSet
     *                 is still running and was submitted by the XSet instance
     *                 specified (true) or not (false)
     *
     * @throws RRMException   
     */
    public static boolean isJobSubmittedByXSetRunning(ReferenceXSet xset)
                                                        throws XAMException {
        boolean running = false;
        ReferenceJob refJob = JobMgr.getJobSubmittedByXSet(xset);
        if (refJob != null && !refJob.getJobStatus().isDone()) {
            running = true;
        }
        return running;
    }
    
    /**
     * Submits a job for execution.
     *  
     * @param job Job object which gets passed as a reference into a
     *                new thread and then run.
     *
     * @throws XAMException an exception is thrown if submission fails.
     */
    public static synchronized Thread submit(ReferenceJob job) throws XAMException {           
        String jobID = createJobID();
        if (debug)
            System.out.println(jobID + ": Starting thread");
        job.setJobID(jobID);
        // creates a new thread with the thread's name equal to the jobID value
        Thread jobThread = new Thread(job, jobID);
        jobThread.start();

        // stores thread and its identifier in the maps
        synchronized(JobMgr.class) {
            updateMap(job);
            ReferenceXSet refXSet = job.getXSet();
            /* The xsetInstanceMap is populated only when submit is called so
             it contains only those ReferenceXSet instances that have  
             submitted the job.  If a ReferenceXSet is created, populated with
             job fields and then committed, it will have a XUID.  When the job
             is submitted, the ReferenceXSet will get stored in the
             xsetInstanceMap since it was the XSet that submitted the job AND
             it would also be in the xuidMap since it was already committed.*/
            updateMap(refXSet, jobID);
            if (refXSet.containsField(XSet.XAM_XUID)) {
                updateMap(refXSet.getXUIDString(), refXSet);
            }
        }
        if (debug)
            System.out.println("Populated map(s): " + job.getXSet().toString());
        
        return jobThread;
    }
    
    /**
     * Determine the number of jobs submitted and running.
     * @return Returns the number of jobs submitted and running.
     */
    public static synchronized int numSubmittedAndRunning() {
        return jobMap.size();
    }
    
    /**
     * Removes the job info from maps.
     *
     * @param job The job to remove from the map(s) 
     * @return Returns true if the job has been successfully removed from
     * the map(s), otherwise false.
     */
    public static synchronized boolean removeFromMaps(ReferenceJob job) {
        boolean removed = false;
        if (null != job) {
            String id = job.getID();

            // remove thread from array
            if (jobMap.containsKey(id)) { 
                jobMap.remove(id);
                removed = true;
                
                if (debug)
                    System.out.println(id + ":  Removed job from job map.");
            }
            ReferenceXSet refXSet = job.getXSet();
            String xsetXUID = refXSet.getXUIDString();
            if (xsetInstanceMap.containsKey(refXSet)) {
                xsetInstanceMap.remove(refXSet);
            }
            if (xsetXUID.length() != 0 && xuidMap.containsKey(xsetXUID)) {
                    xuidMap.remove(xsetXUID);
            }
            if (debug)
                System.out.println("XSet removed from maps: " + 
                                                    job.getXSet().toString());
        }
        return removed;
    }
    
    /**
     * Calls the cancel method on the job specified by the jobID.
     *
     * @param jobId     string with the ID of the job that is to be cancelled.
     * @param wait      flag indicating whether or not this method should
     *                  wait for the job to be cancelled
     *
     * @return ReferenceJob   if the job has been successfully cancelled it
     *                        is returned, otherwise null is returned if:
     *                        a) the job hasn't been cancelled before the
     *                           timeout period
     *                        b) the job doesn't exist in the jobMap so it
     *                           cannot be cancelled
     *
     * @throws RRMException     If job cancellation fails.
     */
    private static ReferenceJob cancel(String jobId, boolean wait)
    throws XAMException {
        
        // current implementation will have one thread per job
        ReferenceJob job = null;

        /**
         * Calls cancel on the Job.
         * Each Job will set its local cancelRequested boolean to true.
         */
        synchronized(JobMgr.class) {
            if (jobMap.containsKey(jobId)) {
                // map contains Job --
                job = jobMap.get(jobId);
                job.cancel();
            }
        }
        
        if (job != null && wait) {
            long startTime = System.currentTimeMillis();
            while (!JobStatus.isDone(job.getJobStatus())) {
                try {
                    if (System.currentTimeMillis() - 
                                        job.getWaitTimeout() > startTime) {
                        throw new ObjectInUseException(
                                "Timed out waiting for job to halt.");
                    }
                    Thread.sleep(job.getWaitSleepTime());
                } catch (InterruptedException ex) {
                    // ignore
                }
            }
        }
        return job;
    }
    
    /**
     * Creates a unique job identifier
     * 
     * @return generated job identifier in the form of JOBID_NUM_GMTDateTime
     */
    private static String createJobID() {
        TimeZone zone = TimeZone.getTimeZone("GMT");
        Calendar c = Calendar.getInstance(zone);
        SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd_HH:mm:ssz");
        sdf.setTimeZone(zone);
        String timeStamp = sdf.format(c.getTime());
        return XAMConstants.JOB_ID_PREFIX + String.valueOf(++jobCounter) + 
                                                                "_" + timeStamp;
    }
    
    /**
     * Adds the job to the jobMap
     *
     * @param job The job to be added to the jobMap
     */
    private static void updateMap(ReferenceJob job) {
        String jobId = null;
        
        if (null != job) {
            jobId = job.getID();
            jobMap.put(jobId, job);
        }
    }
    
    /**
     * Adds the job to the xsetInstanceMap
     *
     * @param set The ReferenceXSet instance being added to the map
     */
    private static void updateMap(ReferenceXSet set, String jobid) {
        if (null != set && null != jobid) {
            xsetInstanceMap.put(set, jobid);
        }
    }
    
    /**
     * Adds the job to the xuidMap
     *
     * @param set The ReferenceXSet instance being added to the map
     */
    private static void updateMap(String xuid, ReferenceXSet set) {
        if (null != xuid && null != set) {
            xuidMap.put(xuid, set);
        }
    }


}
