/*
 * Copyright (c) 2008, Sun Microsystems, Inc.
 * Copyright (c) 2008, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import org.snia.xam.XSet;

/**
 *
 * Job Status  information. Work In Progress.
 * Not sure need all methods. Possibly only need the strings  and a few 
 * helper methods. Decided to not use enums because of issues with valueOf
 * method, which can not be overridden, and does not handle the case where
 * the enum string contains spaces (e.g. "SHUTTING DOWN").
 * 
 * Note: It might make sense to move this class to the XAM Library as it is
 * really general purpose and does not contain any Reference VIM specific 
 * code.
 */
public class JobStatus extends OperationalState {
    
 
    // Error message string
    private static final String INVALID_STATUS_STRING_MSG =
            "Invalid or unknown job status text";    

    private JobStatus(String jobStatusText) {
        super(jobStatusText);        
    }
    
    /**
     * Job was just submitted and had not yet been started by the system
     */
    public static final JobStatus NEW = new JobStatus(XSet.XAM_JOB_STATUS_NEW);
    /**
     * The job has been started but is still being initialized and is not yet
     * running.
     */
    public static final JobStatus STARTING = 
            new JobStatus(XSet.XAM_JOB_STATUS_STARTING);
    /**
     * The job is running.
     */
    public static final JobStatus RUNNING = 
            new JobStatus(XSet.XAM_JOB_STATUS_RUNNING);
    /**
     * The job is in the process of shutting down.
     */
    public static final JobStatus SHUTTING_DOWN = 
            new JobStatus(XSet.XAM_JOB_STATUS_SHUTTING_DOWN);
    /**
     * The job is no longer running. It has run to the end.
     */
    public static final JobStatus COMPLETE = 
            new JobStatus(XSet.XAM_JOB_STATUS_COMPLETE);
    /**
     * The job is no longer running. The system has temporarily ceased
     * processing; however it will resume.
     */
    public static final JobStatus SUSPENDED = 
            new JobStatus(XSet.XAM_JOB_STATUS_SUSPENDED);
    /**
     * The job is no longer running because the application stopped it.
     * Note: This is normally done by calling the haltJob() method.
     */
    public static final JobStatus HALTED = 
            new JobStatus(XSet.XAM_JOB_STATUS_HALTED);
    /**
     * The job is not longer running because the underlying storage system
     * stopped it.
     */
    public static final JobStatus KILLED = 
            new JobStatus(XSet.XAM_JOB_STATUS_KILLED);
    
    
    /**
     * Determines whether the job status text string matches the text string
     * for the current job status instance.
     * 
     * @param jobStatusString
     * @return true if the job status text string matches the current job
     * status instance; false otherwise
     */
    public boolean equals(String jobStatusString) {
        return jobStatusString == null ? false : jobStatusString.equals(
                getName());
    }
    
    /**
     * Use this method instead of valueOf which will not work correctly.
     * 
     * Retrieves the JobStatus instance for the job status text string passed
     * in as an argument. This method should be called to convert a text
     * job status string to an JobStatus instance.
     * 
     * @param jobStatusString the job status text 
     * @return the JobStatus instance that represents the job status text string
     * @throws java.lang.IllegalArgumentException if the job status string
     * is null or is not known
     */
    public static JobStatus jobStatusValueOf(String jobStatusString) 
            throws IllegalArgumentException {
        
        if (jobStatusString == null) {
            throw new IllegalArgumentException(
                    INVALID_STATUS_STRING_MSG);
        }
        
        if (RUNNING.equals(jobStatusString)) {
            return RUNNING;
        } else if (COMPLETE.equals(jobStatusString)) {
            return COMPLETE;
        } else if (NEW.equals(jobStatusString)) {
            return NEW;
        } else if (STARTING.equals(jobStatusString)) {
            return STARTING;
        } else if (HALTED.equals(jobStatusString)) {
            return HALTED;
        } else if (SHUTTING_DOWN.equals(jobStatusString)) {
            return SHUTTING_DOWN;
        } else if (SUSPENDED.equals(jobStatusString)) {
            return SUSPENDED;
        } else if (KILLED.equals(jobStatusString)) {
            return KILLED;
        } else {
            throw new IllegalArgumentException(
                    INVALID_STATUS_STRING_MSG);
        }
    }
    
    /**
     * Determines if the job has finished running. 
     * @return true if the job has finished running; false otherwise
     */
    public boolean isDone() {
        boolean jobDone = false;
        if (this == COMPLETE ||
                this == HALTED ||
                this == KILLED) {
            jobDone = true;
        }
        return jobDone;
    }
    
    /**
     * Determines if the job has finished running. 
     * @return true if the job has finished running; false otherwise
     */
    public static boolean isDone(JobStatus compareStatus) {
        boolean jobDone = false;
        if (compareStatus == COMPLETE ||
                compareStatus == HALTED ||
                compareStatus == KILLED) {
            jobDone = true;
        }
        return jobDone;
    }

    /**
     * Determines if the job has finished running. 
     * @return true if the job has finished running; false otherwise
     * 
     * @throws java.lang.IllegalArgumentException if the job status string
     * is null or is not known
     */
    public static boolean isDone(String jobStatusString) 
            throws IllegalArgumentException{
        JobStatus jobStatus = jobStatusValueOf(jobStatusString);
        return jobStatus.isDone();
    }
    
    /**
     * Determines if the job has finished running. 
     * @return true if the job has finished running; false otherwise
     */
    public boolean isDoneExecution() {
        boolean jobDone = false;
        if (this == SHUTTING_DOWN ||
                this == COMPLETE ||
                this == HALTED ||
                this == KILLED) {
            jobDone = true;
        }
        return jobDone;
    }
    
    /**
     * Determines if the job has finished executing (that is it is not running
     * or in the process of shutting down). 
     * 
     * @return true if the job has finished running or is in the process
     * of shutting down; false otherwise
     */
    public static boolean isDoneExecution(JobStatus compareStatus) {
        boolean jobDone = false;
        if (compareStatus == SHUTTING_DOWN ||
                compareStatus == COMPLETE ||
                compareStatus == HALTED ||
                compareStatus == KILLED) {
            jobDone = true;
        }
        return jobDone;
    }

    /**
     * Determines if the job has finished executing (that is it is not running
     * or in the process of shutting down). 
     * 
     * @return true if the job has finished running or is in the process
     * of shutting down; false otherwise
     * @throws java.lang.IllegalArgumentException if the job status string
     * is null or is not known
     */
    public static boolean isDoneExecution(String jobStatusString) 
            throws IllegalArgumentException{
        JobStatus jobStatus = jobStatusValueOf(jobStatusString);
        return jobStatus.isDoneExecution();
    }

    /**
     * Determines whether the job currently active. This means the job is
     * not done running and is not in the suspended state.
     * @return true if the job is active; false if the job is done or suspended
     */
    public boolean isActive() {
        return (this == SUSPENDED ||
                isDone()) ? false : true;
    }

    /**
     * Determines whether the job currently active. This means the job is
     * not done running and is not in the suspended state.
     * 
     * @return true if the job is active; false if the job is done or suspended
     * @throws java.lang.IllegalArgumentException if the job status string
     * is null or is not known
     */
    public static boolean isActive(String jobStatusString) 
            throws IllegalArgumentException{
        JobStatus jobStatus = jobStatusValueOf(jobStatusString);
        return jobStatus.isActive();
    }

    /**
     * Determines whether the job currently in the RUNNING state. 
     * @return true if the job is running; returns false otherwise
      */
    public boolean isRunning() {
        return (this == RUNNING ? true : false);
    }

    /**
     * Determines whether the job currently in the RUNNING state. 
     * 
     * @return true if the job is running; returns false otherwise
        * @throws java.lang.IllegalArgumentException if the job status string
     * is null or is not known
     */
    public static boolean isRunning(String jobStatusString) 
            throws IllegalArgumentException{
        JobStatus jobStatus = jobStatusValueOf(jobStatusString);
        return jobStatus.isRunning();
    }

}
