/*
 *  Copyright (c) 2008-2009, Sun Microsystems, Inc.
 *  Copyright (c) 2008-2009, The Storage Networking Industry Association.
 *  Copyright (c) 2008-2009 Hewlett-Packard Development Company, L.P.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED FBY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.ByteArrayOutputStream;
import java.io.RandomAccessFile;
import java.io.UnsupportedEncodingException;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Hashtable;
import org.snia.xam.AuthenticationException;
import org.snia.xam.AuthorizationException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.XAMException;
import org.snia.xam.XSystem;
import org.snia.xam.XSystemAbandonException;
import org.snia.xam.XSystemException;
import org.snia.xam.vim.reference.AuthenticationState;
import org.snia.xam.vim.reference.utils.ReferenceSaslUtils;



/**
 * Controls authentication for VIM Reference implementation.
 * Note that full authentication has not been implemented. The VIM is
 * performing the authentication. For SASL PLAIN a hardcoded username and 
 * password have been defined.
 */
public class ReferenceAuthenticationStatus {
       
    // Current authentication state
    private AuthenticationState authState = AuthenticationState.UNAUTHENTICATED;

    // Current SASL Authentication mechanism in use
    private String saslAuthMechanism = null;
    
    // Hardcoded user name and password for PLAIN authentication
    // This is used when no passwd file is specified.
    public static final String TEST_USERNAME = "testuser";
    public static final String TEST_PASSWORD = "testpasswd";
    
    // private because not really immutable
    private static byte[] SASL_MECHANISM_PLAIN_BYTES = 
            ReferenceSaslUtils.SASL_MECHANISM_PLAIN.getBytes();

    public static byte[] SASL_MECHANISM_ANONYMOUS_BYTES =
            ReferenceSaslUtils.SASL_MECHANISM_ANONYMOUS.getBytes();
          
    /** Default SASL authentication mechanism */
    protected static final String SASL_DEFAULT = 
            ReferenceSaslUtils.SASL_MECHANISM_PLAIN;
    
    // Exception/error messages
    public static final String ERR_MSG_SYSTEM_ALEADY_AUTHENTICATED =
            "Session is already authenticated";
    public static final String ERR_MSG_REJECTED_CRED =  
            "Authentication credentials are invalid. Session is " +
            "un-authenticated. Must complete successfull authentication in " +
            "order to access the system.";
    public static final String ERR_MSG_FAILED_AUTH =  
            "Session is in the failed state. Application must abandon and " +
            "then close the current session.";
    public static final String ERR_MSG_ABANDONED =
            "Session in abandoned state. Application must close" +
                    "the current session.";
    public static final String ERR_MSG_CLOSED =
            "Access not allowed the session is closed.";
    public static final String ERR_MSG_UNAUTHENTICATED = 
            "Application not authenticated.";
    
    private Hashtable<String,String> m_passwdMap = new Hashtable<String,String>();

    /* This value will be returned via "getAuthenticatedUsername" */
    private String m_authenticated_username = null;
      
    /**
     * Build the ReferenceAuthenticationStatus using an external passwd map. 
     * File is in the form of:
     *     username passwd
     * One line per user, no spaces allowed in user or password. Password is in cleartext!
     * @param passwdFile The file containing the username and passwords.
     * @throws XSystemException
     */
    public ReferenceAuthenticationStatus( String passwdFile ) throws XSystemException
    {
       loadAuthenticationInfo( passwdFile );
    }
    
    public ReferenceAuthenticationStatus()
    {
       m_passwdMap.put( TEST_USERNAME, TEST_PASSWORD );
    }
    
    /**
     * Load user information from the password file. Note that since this method may be
     * applied to a preexisting authStatus object, the old password information is wiped clean.
     * @param passwdFile
     * @throws XSystemException 
     */
    public void loadAuthenticationInfo(String passwdFile) throws XSystemException
    {
       m_passwdMap = new Hashtable<String,String>();
       try {
          RandomAccessFile raf = new RandomAccessFile( passwdFile, "r" );
          String line;
          
          while( (line = raf.readLine()) != null ) {
             StringTokenizer tok = new StringTokenizer(line);
             String user = tok.nextToken();
             String pass = tok.nextToken();
             m_passwdMap.put(user, pass);
          }
       }
       catch( Exception e )
       {
          throw new XSystemException( "Can't initialize the ReferenceVIM Passwd file :" + 
             passwdFile + " " + e.getMessage() );
       }
   }

    /**
     * Determines if current session has valid authentication credentials.
     * @return boolean true is current session is authenticated;
     *         false otherwise
     */
    boolean isAuthenticated() {
        boolean authenticated = false;
        if (authState == AuthenticationState.AUTHENTICATED ||
                authState == AuthenticationState.REAUTHENTICATING) {
            authenticated = true;
        }
        return authenticated;
    }
    
    /**
     * Determines if current session is in the process of re-authenticating.
     * Session is still considered to be authenticated while in this state.
     * 
     * @return boolean   true if current session is re-authenticating;
     *      false otherwise
     */
    boolean isReauthenticating() {
        return (authState == AuthenticationState.REAUTHENTICATING) ? 
            true : false;
    }
 
    /**
     * Determines if current session is in failed state. Only abandon() can
     * be called on session in failed state.
     * 
     * @return boolean  true if a fatal error was encountered and the session
     *  is now in the FAILED state. 
     */
    boolean isFailed() {
        return (authState == AuthenticationState.FAILED) ? true : false;
    }

    /**
     * Places session in the failed state. This is called after a fatal error
     * is encountered which renders the session unuseable.
     */
    void setFailed() {
        authState = AuthenticationState.FAILED;
    }


    /**
     * Completes successfully if session can legally perform authentication.
     * If current session is in failed, abandoned or closed state it is 
     * illegal to try and authenticate so an exception is thrown.
     * 
     * @throws org.snia.xam.XSystemException
     * @throws org.snia.xam.XSystemAbandonException
     */
    void checkAuthOperationValid() 
            throws XSystemException, XSystemAbandonException {
        if (authState == AuthenticationState.FAILED) {
            throw new XSystemException(ERR_MSG_FAILED_AUTH);
        }else if (authState == AuthenticationState.CLOSED) {
            throw new XSystemException(ERR_MSG_CLOSED);
        } else if (authState == AuthenticationState.ABANDONED) {
            throw new XSystemAbandonException(ERR_MSG_ABANDONED);
        }
    }
    
    /**
     * Determines if current session is has been abandoned. If session is
     * abandoned only the close method can be called.
     * 
     * @return boolean  true if session in abandoned state, false otherwise
     */
    boolean isAbandoned() {
        return (authState == AuthenticationState.ABANDONED ? true : false);
    } 

    /**
     * Sets the current session to abandoned.
     */
    void setAbandoned() {
        authState = AuthenticationState.ABANDONED;
    } 

    /**
     * Determines if the current session has been closed and is no longer
     * accessible.
     * 
     * @return boolean  true if session is closed.
     */
    boolean isClosed() {
        return (authState == AuthenticationState.CLOSED ? true : false);
    } 

    /**
     * Sets the session to closed. No further access will be allowed.
     */
    void setClosed() {
        authState = AuthenticationState.CLOSED;
    } 

    /**
     * Sets the session to unauthenticated. This is the initial state.
     */
    void setUnAuthenticated() {
        authState = AuthenticationState.UNAUTHENTICATED;
    } 

    /**
     * Retrieves the current authentication state.
     * 
     * @return AuthenticationState the sessions current authentication state
     */
    AuthenticationState getAuthState() {
        return authState;
    }
    
    /**
     * Determines the XSystem instance is accessible. Completes
     * successfully if can access the XSystem instance. Otherwise throws
     * an exception.
     * 
     * @throws org.snia.xam.AuthorizationException
     * @throws org.snia.xam.XSystemException
     * @throws org.snia.xam.XAMException
     */
    void checkAccessAllowed() throws AuthorizationException,
            XSystemAbandonException, XSystemException, XAMException {
        
        if (isAuthenticated()) {
            if (authState == AuthenticationState.CLOSED) {
                throw new XSystemException(ERR_MSG_CLOSED);
            } else { //OK
                    return;
            }
        } else if (authState == AuthenticationState.CLOSED) {
            throw new XSystemException(ERR_MSG_CLOSED);
        } else if (authState == AuthenticationState.FAILED) {
            throw new XSystemException(ERR_MSG_REJECTED_CRED);
        } else if (authState == AuthenticationState.ABANDONED) {
            throw new XSystemAbandonException(ERR_MSG_ABANDONED);
        } else {
            throw new AuthorizationException(ERR_MSG_UNAUTHENTICATED);
        }         
    }    

    /**
     * Completes successfully if session can legally perform reading
     * field information.
     * 
     * If current session is in failed, abandoned or closed state it is 
     * illegal so an exception is thrown.
     * 
     * @throws org.snia.xam.XSystemException
     * @throws org.snia.xam.XSystemAbandonException
     */
    void checkFieldReadAccessAllowed() 
            throws XSystemException, XSystemAbandonException {
        if (authState == AuthenticationState.FAILED) {
            throw new XSystemException(ERR_MSG_FAILED_AUTH);
        }else if (authState == AuthenticationState.CLOSED) {
            throw new XSystemException(ERR_MSG_CLOSED);
        } else if (authState == AuthenticationState.ABANDONED) {
            throw new XSystemAbandonException(ERR_MSG_ABANDONED);
        }
    }

    /**
     * Determines whether it's OK to close the XSystem.
     * 
     * @throws org.snia.xam.XSystemException
     */
    void checkCloseAllowed() throws XSystemException {
        if (isClosed()) {
            throw new XSystemException("Session is closed");
        }
        if (isFailed()) {
            throw new XSystemException("Session in the failed state." +
                    " Must call abandon before close.");
        }
        
    }
    
    /**
     * Authentication for the XSystem. Supports SASL ANONYMOUS and PLAIN
     * mechanisms.
     * For Anonymous the application will be required to send trace information
     * which can be an empty.
     * For Plain, the application must supply the username and the password.
     *
     * Note: This does not fully implement or enforce SASLprep: Stringprep 
     * profile for user names and passwords.
     * 
     * @param buffer the SASL message
     * @return byte[] Contains SASL authentication response: DONE, FAIL or
     *                empty string means more input is required.
     * @throws org.snia.xam.AuthenticationException
     * @throws org.snia.xam.XSystemException
     * @throws org.snia.xam.XAMException
     * @throws org.snia.xam.InvalidArgumentException
     */
    byte[] authenticate(byte[] buffer) 
            throws AuthenticationException, XSystemException, 
            XAMException, InvalidArgumentException {
        synchronized (this) {
            ByteArrayOutputStream respByteStream = new ByteArrayOutputStream();
            int retValue = this.authenticate(buffer, respByteStream);
            byte[] respBytes = respByteStream.toByteArray();
            return respBytes;
        }   // end synchronized

   }        

    /**
     * Determine if the buffer starts with specific bytes. Used by the authenticate method.
     * @param prefix The preifx bytes to test for. 
     * @param buffer The buffer to be examined.
     * @return
     */
    private int startsWith( byte[] prefix, byte[] buffer )
    {
       for( int i = 0; i < prefix.length; i++ )
          if( prefix[i] != buffer[i] )
             return -1;
       
       return prefix.length; 
    }
    
    /**
     * Authentication for the XSystem. Supports SASL ANONYMOUS and PLAIN
     * mechanisms.
     * For Anonymous the application will be required to send trace information
     * which can be an empty.
     * For Plain, the application must supply the username and the password.
     *
     * Note: This does not fully implement or enforce SASLprep: Stringprep 
     * profile for user names and passwords.
     * 
     * @param buffer the SASL message
     * @return byte[] Contains SASL authentication response: DONE, FAIL or
     *                empty string means more input is required.
     * @throws org.snia.xam.AuthenticationException
     * @throws org.snia.xam.XSystemException
     * @throws org.snia.xam.XAMException
     * @throws org.snia.xam.InvalidArgumentException
     */
    int authenticate(byte[] buffer, ByteArrayOutputStream response) 
            throws AuthenticationException, XSystemException, 
            XAMException, InvalidArgumentException {

        response.reset();
        byte[] respBytes = null;
        synchronized (this) {
            
            checkAuthOperationValid();

            // Authentication does not expire so don't allow re-authentication
            if (isAuthenticated()) {
                throw new XSystemException(ERR_MSG_SYSTEM_ALEADY_AUTHENTICATED);
            }
            
            // Check to see if buffer contains a SASL mechanism
            int start = 0;
            if (saslAuthMechanism == null) {
               start = -1;
               if( (start = startsWith(SASL_MECHANISM_PLAIN_BYTES, buffer ) ) > 0 ) {
                  saslAuthMechanism = ReferenceSaslUtils.SASL_MECHANISM_PLAIN;
               } else if( (start = startsWith(SASL_MECHANISM_ANONYMOUS_BYTES, buffer ) ) > 0 ) {
                  saslAuthMechanism = ReferenceSaslUtils.SASL_MECHANISM_ANONYMOUS;
               } else {
                  saslAuthMechanism = SASL_DEFAULT;
                  start = 0;
               }
               
               if( start >= buffer.length ) {
                  respBytes = ReferenceSaslUtils.SASL_RESPONSE_MORE.getBytes();
                  response.write(respBytes, 0, respBytes.length);
                  return XSystem.XAM_SASL_IN_PROCESS;
               }               
            }

            if (saslAuthMechanism.equals(ReferenceSaslUtils.
                    SASL_MECHANISM_ANONYMOUS)) {
                String trace = "";
                try {
                    if (buffer != null && buffer.length > 0) {
                        trace = new String(buffer, 
                                ReferenceSaslUtils.SASL_ENCODING_CHARSET);
                    }
                } catch (UnsupportedEncodingException ex) {
                    XAMException xex = new XAMException(
                            "UTF-8 encoding not supported.");
                    saslAuthMechanism = null;
                    throw(xex);                    
                }
                authState = AuthenticationState.AUTHENTICATED;
                
            } else if (saslAuthMechanism.equals(
                    ReferenceSaslUtils.SASL_MECHANISM_PLAIN)) {
                String[] strs = ReferenceSaslUtils.
                        decodeSASLPlain(buffer, start, buffer.length);

                // Validate username and password
                if (!validateSASLPlainCred(strs)) {
                    authState = AuthenticationState.UNAUTHENTICATED;
                    respBytes = ReferenceSaslUtils.SASL_FAILED.getBytes();
                    response.write(respBytes, 0, respBytes.length);
                    return XSystem.XAM_SASL_COMPLETE;
                }
                // Successfully authentication
                authState = AuthenticationState.AUTHENTICATED;
                
            } else {
                throw new AuthenticationException("SASL mechanism " + 
                        saslAuthMechanism + " unsupported");
            }
            
            respBytes = ReferenceSaslUtils.SASL_DONE.getBytes();
            response.write( respBytes, 0, respBytes.length );
            return XSystem.XAM_SASL_COMPLETE;
        }   // end synchronized

   }        

    /**
     * Determines if session is authenticated using SASL ANONYMOUS mechanism
     * 
     * @return boolean true if session authenticated using SASL ANONYMOUS
     */
    boolean isAuthAnonymous() {
        return (saslAuthMechanism != null && 
                saslAuthMechanism.equals(
                ReferenceSaslUtils.SASL_MECHANISM_ANONYMOUS)) ? 
            true : false;
    }

    /**
     * Determines if session is authenticated using SASL PLAIN mechanism
     * 
     * @return boolean true if session authenticated using SASL PLAIN
     */
    boolean isAuthPlain() {
        return (saslAuthMechanism != null && 
                saslAuthMechanism.equals(
                ReferenceSaslUtils.SASL_MECHANISM_PLAIN)) ? 
            true : false;
    }
	
 
    /*
     * Validates the user supplied authentication credentials for the
     * SASL PLAIN mechanism.
     * @param strs the username and password
     * @return boolean true if the supplied credentials match the hardcoded
     * username and password for SASL PLAIN.
     */
    private boolean validateSASLPlainCred(String[] strs) {
        boolean validCred = false;
        String passwd = m_passwdMap.get(strs[1]);
        if( (passwd != null) && (passwd.equals(strs[2])) ) {
           validCred = true;
           m_authenticated_username = strs[1];
        }
 
        return validCred;
    }
    
    /**
     * Return the authenticated username.
     * @return The username
     */
    public String getAuthenticatedUsername() {
       return m_authenticated_username;
    }
    
    /**
     * Return an enumeration of all the usernames defined for this system. Either from 
     * the password file at startup, or from the builtin username and password.
     * @return All the usernames for this XSystem.
     */
    public Enumeration<String> getUsernameEnumeration()
    {
       return m_passwdMap.keys();
    }
    
}
