/*
 * Copyright (c) 2008, Sun Microsystems, Inc.
 * Copyright (c) 2008, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Calendar;
import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.XAMException;
import org.snia.xam.XSetUnderHoldException;
import org.snia.xam.XStreamCorruptException;
import org.snia.xam.XStreamException;
import org.snia.xam.util.ISO8601Date;

/**
 *
 * Implements Export Stream  (Work in Progress)
 * 
 * Partial implementation - export stream for XSets with property fields only.
 * Can not export XSet with stream fields at this time.
 */
public class ReferenceExportXStream extends ReferenceXStreamBase {
    
    protected XStreamContents data = null;
    protected static boolean debug = false; // Used keep tmp files areound
    String filename = "";
    /*
     * MIME Boundary format
     * --SNIA_REFERENCE_VIM_timestamp.XZY_+
     */
    private static final String MIME_BOUNDARY_PREFIX = 
            "--SNIA_REFERENCE_VIM_MIME_BOUNDARY_";
    private static final String MIME_BOUNDARY_SUFFIX = ".XZY_+";
    private String mimeBoundary = "";
    private File outputFile = null;

    /**
     * Creates an instance of the export stream
     * @param system
     * @param xset
     * @throws org.snia.xam.XAMException
     */
    protected ReferenceExportXStream(ReferenceXSystem system,
            ReferenceXSet xset) throws XAMException {
        super(xset);
        parentXSet = xset;
        try {
            // Get list of Stream Fields
            //String fileName = parentXSet.getXMLFileName();
            state = XStreamState.EXPORT_READER;
            setMIMEBoundary();
            
            // Create tmp file to hold export stream contents
            // TODO:  needs to be moved to the persistence manager
            File reposDir = new File(system.getRepositoryLocation());
            outputFile = File.createTempFile(
                    "XSetExport_" + SniaXUID.convertToSafeString(xset.getXUID()),".tmp", 
                    reposDir);
            
            // Write out export format
            XSetXopWriter.outputXMLToTmpFile(outputFile.getAbsolutePath(),
                    xset, reposDir.getAbsolutePath(), mimeBoundary);

            // Setup channel for reading export stream
            input = new FileInputStream(
                    outputFile.getAbsoluteFile());
            channel = input.getChannel();
        } catch (IOException ex) {
            throw new XAMException("Error opening ExportXStream", ex);
        }

    }
    
    /**
     * Generates the MIME Boundary to use when writing out the export stream
     */
    private void setMIMEBoundary() {
        Calendar calendar = Calendar.getInstance();
        String isoDateString = ISO8601Date.toISO8601String(calendar);        
        mimeBoundary = MIME_BOUNDARY_PREFIX +
                isoDateString + MIME_BOUNDARY_SUFFIX;              
    }
    
    /**
     * @see ReferenceXStreamBase#closeStream() 
     * @throws org.snia.xam.AuthenticationExpiredException
     * @throws org.snia.xam.XSetUnderHoldException
     * @throws org.snia.xam.XStreamCorruptException
     * @throws org.snia.xam.XStreamException
     * @throws org.snia.xam.XAMException
     */
    public void closeStream() throws AuthenticationExpiredException, 
            XSetUnderHoldException, XStreamCorruptException, 
            XStreamException, XAMException {
        closeChannelAndStreams();
        
        if (outputFile != null) {
            try {
                if (!debug)
                    outputFile.delete();
            } catch (Exception ex) {
                // ignore
            }
        }
        outputFile = null;        
        parentXSet.removeExportStream(this);
    }
    
    /**
     * Sets the Export read stream to the corrupt state.
     */
    protected void setStateCorrupt() {
        state = XStreamState.EXPORT_READ_CORRUPT;
    }
        
    /**
     * Determines if method is allowed when exporting an XStream
     * @param methodName
     * @throws org.snia.xam.XStreamException
     */
    public void checkMethodAccess(String methodName) throws XStreamException {
        if (! (methodName.equals("read") 
                || methodName.equals("tell")
                || methodName.equals("abandon")
                || methodName.equals("close"))) {
            throw new XStreamException(
                    "Exporting XStream, operation is not allowed: " +
                    methodName);
        }
    }

}
