/*
 * Copyright (c) 2008-2009, Sun Microsystems, Inc.
 * Copyright (c) 2008-2009, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.XAMException;
import org.snia.xam.XSetUnderHoldException;
import org.snia.xam.XStreamCorruptException;
import org.snia.xam.XStreamException;

/**
 *
 * Implements Import XStream  
 * 
 */
public class ReferenceImportXStream extends ReferenceXStreamBase {
    
    private Logger logger = null;
    
    protected XStreamContents data = null;
    String filename = "";
    private File outputFile = null;
    
    // Set to true to leave tmp files; Also used by other import classes
    protected static boolean debug = false;

    /**
     * Creates an instance of the import stream
     * @param system
     * @param xset
     * @throws org.snia.xam.XAMException
     */
    protected ReferenceImportXStream(ReferenceXSystem system,
            ReferenceXSet xset) throws XAMException {
        super(xset);
        parentXSet = xset;
        logger = system.getLogger();
        try {
            // Create tmp file to hold import stream contents
            outputFile = File.createTempFile("XSet_Import_", 
                    ReferenceVIMPersistenceMgr.TMP_FILENAME_SUFFIX, null);
            state = XStreamState.IMPORT_WRITER;
            
            // Setup channel for reading import stream
            output = new FileOutputStream(
                    outputFile.getAbsoluteFile());
            channel = output.getChannel();
        } catch (IOException ex) {
            if (outputFile != null) {
                if (channel != null) {
                    try {
                        channel.close();
                    } catch (IOException ioex) { 
                        // log and continue
                        logger.log(Level.WARNING, "Could not close file " +
                                "channel on: " + outputFile.getAbsolutePath());
                    }
                }
                if (!debug) {
                    if (!outputFile.delete()) {
                        logger.log(Level.WARNING, "Could not delete temp " +
                                outputFile.getAbsolutePath());
                    }
                }
            }
            String msg = "Error opening ImportXStream";
            logger.log(Level.SEVERE, msg, ex);
            throw new XAMException(msg, ex);
        }

    }
     
    /**
     * Override to provide parsing and handling of import stream. 
     * 
     * @throws org.snia.xam.AuthenticationExpiredException
     * @throws org.snia.xam.XSetUnderHoldException
     * @throws org.snia.xam.XStreamCorruptException
     * @throws org.snia.xam.XStreamException
     * @throws org.snia.xam.XAMException
     */
    public void close() throws AuthenticationExpiredException, 
            XSetUnderHoldException, XStreamCorruptException, 
            XStreamException, XAMException {
        
        try {
            // parse & verify the imported XSet
            closeStream();
            
            state = XStreamState.CLOSED;
            parentXSet.removeImportStream(this);
            if (outputFile != null) {
                 if (!debug)
                   if (!outputFile.delete()) {
                        logger.log(Level.WARNING, "Could not delete temp " +
                                outputFile.getAbsolutePath());
                    }
             }
        } catch (XStreamCorruptException cex) {
            state = XStreamState.IMPORT_WRITE_CORRUPT;
        } catch (XStreamException ex) {
            state = XStreamState.IMPORT_WRITE_CORRUPT;
            throw ex;
        }
    }

    /**
     * Sets the Import write stream to the corrupt state.
     */
    protected void setStateCorrupt() {
        state = XStreamState.IMPORT_WRITE_CORRUPT;
    }
    /**
     * @see ReferenceXStreamBase#closeStream() 
     * @throws org.snia.xam.AuthenticationExpiredException
     * @throws org.snia.xam.XSetUnderHoldException
     * @throws org.snia.xam.XStreamCorruptException
     * @throws org.snia.xam.XStreamException
     * @throws org.snia.xam.XAMException
     */
    public void closeStream() throws AuthenticationExpiredException, 
            XSetUnderHoldException, XStreamCorruptException, 
            XStreamException, XAMException {
        
         closeChannelAndStreams();
         try {
            // parse imported XStream
             Vector<XStreamXopData> xstreamsDataList = new Vector<XStreamXopData>();
             String xsetXML = XSetXopReader.parse(
                     outputFile.getAbsolutePath(), xstreamsDataList, logger);
             XStreamXopData[] xstreamsData = new XStreamXopData[5];
             xstreamsData = xstreamsDataList.toArray(xstreamsData);
             parentXSet.loadImportedXSet(xsetXML, xstreamsData);
             
             if( !debug ) {
                outputFile.delete();
             }
             
         } catch (XAMException ex) {
              throw ex;
         } 
     
    }
       
    /**
     * Determines if method is allowed when importing an XStream
     * @param methodName
     * @throws org.snia.xam.XStreamException
     */
    public void checkMethodAccess(String methodName) throws XStreamException {
        if (! (methodName.equals("write") 
                || methodName.equals("tell")
                || methodName.equals("abandon")
                || methodName.equals("close"))) {
           String msg = "Importing XStream, operation is not allowed: " + methodName;
           logger.warning(msg);
           throw new XStreamException(msg);
        }
    }

}