/* Copyright (c) 2008-2009, Sun Microsystems, Inc.
 * Copyright (c) 2008-2009, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidFieldNameException;
import org.snia.xam.JobCommandException;
import org.snia.xam.JobUnsupportedException;
import org.snia.xam.XAMException;
import org.snia.xam.XSet;

/**
 * This is an abstract base class that implements the Job interface and
 * from which all specific jobs extend.  It contains data common to all 
 * types of jobs and provides the means to retrieve the job's id and XSet.
 */
public abstract class ReferenceJob implements Job {
    
    private static Logger logger = null;
    
    protected static final String XAM_JOB_ERROR_NOT_A_JOB = 
                                                    "org.snia.xam::not_a_job";
    protected static final String XAM_JOB_ERROR_UNSPECIFIED_CMD = 
                                            "org.snia.xam::unspecified_command";
    protected final static String XAM_JOB_ERROR_EXECUTION_FAILED = 
                    XAMConstants.SNIA_REFVIM_PREFIX + "xam::execution_failed";
    protected static final String JOB_DELAY_FIELD_IN_MILLISEC = 
                               XAMConstants.SNIA_REFVIM_PREFIX + "job.delay";
    protected static final long MAX_JOB_DELAY_IN_MILLISEC = 60000;
    
    protected static final String[] JOB_XSET_FIELDS = {
                                    XSet.XAM_JOB_COMMAND,
                                    XSet.XAM_JOB_ERROR,
                                    XSet.XAM_JOB_ERRORHEALTH,
                                    XSet.XAM_JOB_STATUS};
    
    /**
     * Identifier for this job in the form of "JOBID_generatedUniqueID".  
     * The job ID is also the name of the Thread object that's responsible for 
     * invoking the ReferenceJob's run() method.
     */
    protected String jobID = XAMConstants.NO_JOB_ID;
    /**
     * The actual XSet being used to store the job information
     */
    protected ReferenceXSet xset = null;
    /**
     * The default value in milliseconds to wait for a running job to respond
     */
    protected long timeout = XAMConstants.MILLISEC_PER_SECOND;
    /**
     * The default value in milliseconds to sleep before checking to see if 
     * a running job has finished
     */
    protected long waitSleep = XAMConstants.MILLISEC_PER_SECOND/2;
    /**
     * Indicates whether or not this service has received a request to cancel.
     * A value of <code>true</code> indicates the service should try and cancel
     * A value of <code>false</code> indicates the service should keep running
     */
    protected boolean cancelRequested = false;
    /**
     * Indicates when this job service has completed its processing,
     * either successfully or unsuccessfully.
     * A value of <code>true</code> indicates that the service is complete.
     * A value of <code>false</code> indicates that the service
     * has not yet completed.
     */
    protected boolean done = false;
    /**
     * The status of this job, which can be one of the following:
     *      NEW, STARTING, RUNNING, SHUTTING_DOWN, 
     *          COMPLETE, SUSPENDED, HALTED, KILLED
     */
    protected JobStatus jobStatus = JobStatus.NEW;
    /**
     * For debugging purposes ONLY
     */
    protected boolean debug = false;
    
    /**
     * Constructor for ReferenceJob -- 
     * @param xset the ReferenceXSet containing the job fields/information
     * @throws InvalidArgumentException an exception if the xset is null
     */
    protected ReferenceJob(ReferenceXSet xset) 
            throws InvalidArgumentException, JobCommandException, XAMException {
        if (xset == null) {
            throw new InvalidArgumentException("XSet can not be null");
        }
        logger = xset.getXSystem().getLogger();
        this.xset = xset;
        checkJobFields();
        setJobStatus(JobStatus.STARTING);
    }
    /**
     * This method is invoked by the Thread created in JobMgr::submit in order
     * to perform the actual work.
     */
    public void run() {
        try {
            // check if need to move this into checkJobFields or not.
            clearJobError();
            logger.log(Level.INFO, this.getID() + ": Running job.");
            if (debug)
                System.out.println(this.getID() + ": Running job.");
            
            /* Note that the subclassing job must set the status to RUNNING.
             * The JobStatus needs to be set after all the appropriate job 
             * fields have been set to ReadOnly, and just prior to actual 
             * execution.
             */
            execute();
            setJobStatus(JobStatus.SHUTTING_DOWN);
            if (debug)
                System.out.println("ReferenceJob.run - finished execute()");
            
        } catch (Exception e) {
            try {
                setJobError(XAM_JOB_ERROR_EXECUTION_FAILED);
            } catch (Exception ex) { /* ignore */ }
            String errMsg = "Error occured while executing the job. " +
                    "Error: " + e.getMessage();
            logger.log(Level.SEVERE, errMsg);
            e.printStackTrace();
        } finally {
            try {
                xset.setReadOnlyField(XSet.XAM_JOB_COMMAND, false);
                this.close();
                JobMgr.removeFromMaps(this);
                // Do not want to set job status to complete until the
                // job has been closed and removed from the maps
                if (!cancelRequested) {
                    setJobStatus(JobStatus.COMPLETE);
                } else {
                    setJobStatus(JobStatus.HALTED);
                }
                xset.setReadOnlyField(XSet.XAM_JOB_COMMAND, false);

            } catch (XAMException xe) {
                try {
                    xset.setReadOnlyField(XSet.XAM_JOB_COMMAND, false);
                    if (!cancelRequested) {
                        setJobStatus(JobStatus.COMPLETE);
                    } else {
                        setJobStatus(JobStatus.HALTED);
                    }
                    xset.setReadOnlyField(XSet.XAM_JOB_COMMAND, false);
                } catch (Exception ex) { /* ignore */ }

                logger.log(Level.SEVERE, "Unable to clean up job resources.", xe);

            }
        }
    }
    /**
     * Accessor for the time to wait for a running job to respond
     * @return long the duration to wait in milliseconds
     */
    public long getWaitTimeout() {
        return this.timeout;
    }
    /**
     * Accessor for the time to sleep before checking to see if a running job
     * is done.
     * @return long the duration to sleep in milliseconds
     */
    public long getWaitSleepTime() {
        return this.waitSleep;
    }
    /**
     * Accessor for the job's identifier
     * @return String the job ID
     */
    public String getID() {
        return this.jobID;
    }
    /**
     * Accessor for the XSet, which the job utilizes to get job information and
     * perform the work.
     * @return ReferenceXSet the XSet serving as a job information repository
     */
    public ReferenceXSet getXSet() {
        return this.xset;
    }
    
    /**
     * Determines if this process has been requested to be canceled --
     * This method is called and the value checked when the process may be
     * able to stop execution.
     * @return Returns true if the service should stop running, otherwise false
     */
    public boolean isCancelRequested() {
        return cancelRequested;
    }

    /**
     * Determines if this process has finished processing/preforming its work.
     * @return Returns true if the job has completed its work.
     */
    public boolean isExecuteDone() {
        if (debug)
            System.out.println(Thread.currentThread().getName() +  
                    ":   ReferenceJob.isDone(): " + done);
        return this.done;
    }   
    
    /**
     * Set the cancelRequested boolean to true.
     * @throws XAMException When an error is detected canceling this job.
     */
    public void cancel() throws XAMException
    {
        cancelRequested = true;
        setJobStatus(JobStatus.SHUTTING_DOWN);
    }
    protected void setJobID(String id) {
        this.jobID = id;
    }
    
    /**
     * Gets the current job status
     * 
     * @return the current JobStatus object
     */
    public JobStatus getJobStatus() {
        synchronized (jobStatus) {
            return jobStatus;
        }
    }
    
    /**
     * Sets the job status.
     * 
     * @param status the status of the job
     * @throws XAMException
     */
    protected void setJobStatus(JobStatus status) throws XAMException {
        synchronized (jobStatus) {
            jobStatus = status;
            boolean readonly = (status.isDone() ? false : true);
            xset.setJobField(
                    XSet.XAM_JOB_STATUS, false, readonly, status.getName());
        }
    }
    /**
     * Sets the job error.  Upon completion the call should terminate 
     * processing of the job.
     * @param errorCode the error code
     */
    protected void setJobError(String errorCode) throws XAMException {
        if (errorCode == null)
            throw new NullPointerException();
        logger.log(Level.WARNING, "Setting job error to " + errorCode);
        xset.setJobField(XSet.XAM_JOB_ERRORHEALTH, false, true, XSet.XAM_JOB_ERRORHEALTH_ERROR);
        xset.setJobField(XSet.XAM_JOB_ERROR, false, true, errorCode);
        setJobStatus(JobStatus.KILLED);
    }
    
    /**
     * Clears the error fields. The error and health fields are only
     * created when a job has an error. Must clear them out when starting
     * to run a job.
     * 
     * @throws org.snia.xam.XAMException
     */
    protected void clearJobError() throws XAMException {
         if (xset.containsField(XSet.XAM_JOB_ERRORHEALTH)) {
            if (xset.containsField(XSet.XAM_JOB_ERRORHEALTH)) {
                xset.deleteField(XSet.XAM_JOB_ERRORHEALTH);
            }
            if (xset.containsField(XSet.XAM_JOB_ERROR)) {
                xset.deleteField(XSet.XAM_JOB_ERROR);
            }
        }

    }
    /**
     * Checks the fields in the submitted job to make sure the job command field
     * is set to read-only so it can't be changed while the job is running and
     * to make sure the .xam.job.errorhealth field is removed before running
     * 
     * @throws JobUnsupportedException, JobCommandException, 
     *          InvalidFieldNameException, XAMException
     */
    protected void checkJobFields() throws JobCommandException, 
                                        InvalidFieldNameException,
                                        XAMException {
        
        // ensure the job command field exists, otherwise this is not a job
        if (!xset.containsField(XSet.XAM_JOB_COMMAND)) {
            setJobError(XAM_JOB_ERROR_NOT_A_JOB);
            String msg = "The org.snia.xam.job.command field is missing from the job.";
            logger.warning(msg);
            throw new JobCommandException();
        }
        xset.setReadOnlyField(XSet.XAM_JOB_COMMAND, true);
        
    }

    /**
     * Sets the readonly value of the base job fields.
     * Job fields are readonly when the job is running.
     * @param xset
     * @param readOnly
     */
    protected static void setJobFieldsReadOnly(
            ReferenceXSet xset, boolean readOnly) {
        
        if (xset.isJob()) {
            // For all job fields set them to readonly
            for (String fieldName : JOB_XSET_FIELDS) {
                try {
                    xset.setReadOnlyField(fieldName, readOnly);
                } catch (FieldDoesNotExistException nex) {
                    // ignore
                } catch (XAMException ex) {
                    // log exception and continue
                    logger.warning("Error setting job field readonly value." +
                            " Field name: " + fieldName + "Error: " +
                            ex.getMessage());
                            
                }
            }
        }
        
    }
    
    /**
     * Returns the value of org.snia.xam.job.command field that was specified
     * when submitting the job.
     * 
     * @return the value of the org.snia.xam.job.command field.
     */
    public abstract String getJobType();
}
