/*
 * Copyright (c) 2008-2009, Sun Microsystems, Inc.
 * Copyright (c) 2008-2009, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.File;
import java.io.FilenameFilter;
import java.io.StringBufferInputStream;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.FieldExistsException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidFieldNameException;
import org.snia.xam.JobCommandException;
import org.snia.xam.JobUnsupportedException;
import org.snia.xam.QueryException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XSet;
import org.snia.xam.XStream;
import org.snia.xam.XSystem;
import org.snia.xam.XUID;
import org.snia.xam.vim.reference.query.ParseException;
import org.snia.xam.vim.reference.query.ParsedFactorNode;
import org.snia.xam.vim.reference.query.ParsedQuery;
import org.snia.xam.vim.reference.query.WhereNode;
import org.snia.xam.vim.reference.query.XamQlParser;

/**
 * This job does the query job processing.
 */
public class ReferenceQueryJob extends ReferenceJob {
    
    private static Logger logger = null;
    
    protected static final String XAM_JOB_ERROR_INVALID_QUERY_CMD = 
                                           "xam.job.query::empty_query_command";
    public static final String JOB_LEVEL_1 = "org.snia.xam.job.query.level.1";
    public static final String JOB_LEVEL_2 = "org.snia.xam.job.query.level.2";
    protected static final String[] QUERY_JOB_XSET_FIELDS = {
                                    XSet.XAM_JOB_QUERY_COMMAND,
                                    XSet.XAM_JOB_QUERY_RESULTS,
                                    XSet.XAM_JOB_QUERY_RESULTS_COUNT,
                                    XSet.XAM_JOB_QUERY_LEVEL};
    

    /**
     * The XStream field that stores the result information
     */
    protected XStream resultStream = null;
    /**
     * The XAM QL command contained in the xam.job.query.command XStream field
     * within the encapsulated XSet reference.
     */
    protected String xamqlCommand = XAMConstants.EMPTY_STRING;
    
    /**
     * The output of the parser.
     */
    ParsedQuery m_parsed_query = null;
    
    /**
     * Constructor for ReferenceQueryJob --
     * @param xset the ReferenceXSet containing the job fields/information
     * @throws InvalidArgumentException, XAMException
     */
    protected ReferenceQueryJob(ReferenceXSet xset) 
                                throws InvalidArgumentException, XAMException {
        super(xset);
        logger = xset.getXSystem().getLogger();
        verifyQueryJobType();
        createQueryJobFields();
    }
    
    /**
     * This performs the work of the job -- if any exceptions are thrown
     * they are logged since they are programming errors.
     */
    public void execute() throws XAMException {
       long startTime = System.currentTimeMillis();
       try {
            // while running set fields to readonly
            setJobFieldsReadOnly(xset, true);
            setJobStatus(JobStatus.RUNNING);
            
            ReferenceXSystem xsys = xset.getXSystem();
            DBManager        dbm  = xsys.getDBManager();
            
            Vector<String> readPolicies = xsys.getAccessManager().getReadPolicies();
            ResultSet rs = dbm.executeXamQuery( this, this.getID(), m_parsed_query, readPolicies );
            while( !cancelRequested && rs.next() ) {
               String xuid_str = rs.getString(1);
               XUID xuid = xsys.createXUID(xuid_str);
               storeResult(xuid);
            }
            
            // Delay implemented for debugging purposes
            if (xset.containsField(JOB_DELAY_FIELD_IN_MILLISEC)) {
                long delay = xset.getLong(JOB_DELAY_FIELD_IN_MILLISEC);
                if( delay > MAX_JOB_DELAY_IN_MILLISEC )
                   delay = MAX_JOB_DELAY_IN_MILLISEC;
                if (delay > 0 ) {                    
                    logger.log(Level.FINEST, "Delaying end of query job for xuid, " + 
                                xset.getXUIDString() + ", by " +
                                    String.valueOf(delay/1000) + " seconds.");
                    Thread.sleep(delay);
                }
            }
        } catch( XAMException xe ) {
           // Don't wrap XAM Exceptions, just pass them along
           throw xe;
        } catch (Exception ex) {
            throw new XAMException(ex);
        } finally {
           done = true;
           // job not running so set fields to read/write
           setJobFieldsReadOnly(xset, false);
            
           long endTime = System.currentTimeMillis();
           double deltaSeconds = (endTime - startTime) / 1000.0;
           logger.log(Level.INFO, "Query ran in " + deltaSeconds + " seconds : " + xamqlCommand );
        }
    }
    
   public void close() throws XAMException {
       // Close the resources used by this job
       if (isExecuteDone()) {
           setJobFieldsReadOnly(xset, false);
           if (resultStream != null) {
               resultStream.close();
           }
       }
   }
   
    /*
     * Method to store each XUID to the stream as it is returned from the query
     * and updates the result count -- 
     * @param xuid the XUID which meets the specified query criteria
     * @throws XAMException
     */
    protected void storeResult(XUID xuid) throws XAMException {
        byte[] xuidBytes = ((SniaXUID)xuid).toPaddedBytes();
        int written = 0;
        while (written < xuidBytes.length) {
                written += resultStream.write(xuidBytes, written, 
                                                    xuidBytes.length-written);
        }
        long count = xset.getLong(XSet.XAM_JOB_QUERY_RESULTS_COUNT);
        xset.setJobProperty(XSet.XAM_JOB_QUERY_RESULTS_COUNT, ++count);
    }
    /**
     * Verifies that this is a query job. Throws an exception if not a query job.
     * 
     * @throws JobUnsupportedException, JobCommandException, 
     *          InvalidFieldNameException, XAMException
     */
    protected void verifyQueryJobType() throws JobUnsupportedException, 
                                        JobCommandException, 
                                        InvalidFieldNameException,
                                        XAMException {
        if (!xset.getString(XSet.XAM_JOB_COMMAND).equals(XSet.XAM_JOB_QUERY)) {
            setJobError(XAM_JOB_ERROR_UNSPECIFIED_CMD);
            String msg = "The job submitted is not a query job.";
            logger.warning(msg);
            throw new JobUnsupportedException(msg);
        }
        if (!xset.containsField(XSet.XAM_JOB_QUERY_COMMAND)) {
           setJobError(XAM_JOB_ERROR_NOT_A_JOB);
           String msg = "The xam.job.query.command field is missing from the query job.";
           logger.warning(msg);
           throw new JobCommandException(msg);
        }
    }
    /**
     * Creates the query job-specific fields in the XSet
     * 
     * @throws org.snia.xam.XAMException
     */
    protected void createQueryJobFields() 
                            throws JobCommandException, XAMException {
        boolean error = false;
        XStream cmdStream = null;
        long buffer_size = xset.getFieldLength( XSet.XAM_JOB_QUERY_COMMAND );
        if( buffer_size > Integer.MAX_VALUE ) {
           String msg = "Query is too large for XSystem: " + buffer_size;
           logger.severe(msg);
           throw new QueryException( msg );
        }
        
        byte[] buffer = new byte[(int)buffer_size];
        try {
            cmdStream = xset.openXStream(XSet.XAM_JOB_QUERY_COMMAND,
                XStream.MODE_READ_ONLY);
            long numOfBytesRead = cmdStream.read(buffer);
            if (numOfBytesRead == XAMConstants.END_OF_STREAM) {
                setJobError(XAM_JOB_ERROR_INVALID_QUERY_CMD);
                String msg = "Query command stream is empty -- end of stream.";
                logger.warning(msg);
                throw new XAMException(msg);
            }
            while (numOfBytesRead < buffer.length) {
                long bytes_read = cmdStream.read(buffer, numOfBytesRead, 
                                                buffer.length - numOfBytesRead);
                if (bytes_read == XAMConstants.END_OF_STREAM) {
                    break;
                }
                numOfBytesRead += bytes_read;
            }
            xamqlCommand = new String(buffer).trim();        
            parseQuery( xamqlCommand );
            
            // if this job has already been run once then the results stream
            // field will already exist
            try {
                resultStream = xset.createXStream(XSet.XAM_JOB_QUERY_RESULTS, 
                                        false, XAMLibrary.XUID_LIST_MIME_TYPE);
                xset.setReadOnlyField( XSet.XAM_JOB_QUERY_RESULTS, true );
            } catch (FieldExistsException fe) {
                // clear out existing results
                resultStream = xset.openXStream(XSet.XAM_JOB_QUERY_RESULTS, 
                        XStream.MODE_WRITE_TRUNCATE);
            }
            xset.setJobField(XSet.XAM_JOB_QUERY_RESULTS_COUNT, false, 
                    false, new Long(0));
            setQueryLevel();            
        } catch (XAMException e) {
            error = true;
            throw e;
        } finally {
            if (cmdStream != null) {
                cmdStream.close();
            }
            if (error && resultStream != null) {
                resultStream.close();
            }
        }
    }

    /**
     * Create a query parser, and feed it the query string.
     */ 
    private void parseQuery(String xamqlCommand ) throws QueryException
    {
       try {
          XamQlParser xql = new XamQlParser(new StringBufferInputStream(
             xamqlCommand ));

          ParsedQuery q = xql.Start();
          m_parsed_query = q;
       }
       catch( Exception pe ) {
          String msg = "Parse error on submitted query";
          logger.log( Level.WARNING, msg, pe );
          throw new QueryException( msg, pe );
       }
    }

    /**
     * Set the level of the query, in the XSet, based on the output of the parser. 
     * Also evaluate if the query is within supported bounds of the XSystem.
     * @throws JobCommandException
     * @throws XAMException
     */
   protected void setQueryLevel() throws JobCommandException, XAMException {
        String queryCmdLevel = JOB_LEVEL_1;

        // verify query command level falls within the query level supported
        boolean levelOneSupported = xset.getXSystem().getBoolean(
                            XSystem.XAM_XSYSTEM_JOB_QUERY_LEVEL1_SUPPORTED);
        boolean levelTwoSupported = xset.getXSystem().getBoolean(
                            XSystem.XAM_XSYSTEM_JOB_QUERY_LEVEL2_SUPPORTED);
        if (levelOneSupported && !levelTwoSupported && 
            (m_parsed_query.getLevel() != 1) ) {
            // XSystem supports level 1 queries only and the query command
            // level is at level 2 so set the error and throw an exception
            setJobError(XSet.XAM_JOB_ERROR_LEVEL_NOT_SUPPORTED);
            String msg = "Only level 1 queries are supported.";
            logger.warning(msg);
            throw new JobCommandException(msg);
        }
        xset.setJobField(XSet.XAM_JOB_QUERY_LEVEL, false, true, queryCmdLevel);
    }
    
    /**
     * Sets the readonly value of the query job fields.
     * Job fields are readonly when the job is running.
     * @param xset
     * @param readOnly
     */
    protected static void setJobFieldsReadOnly(
            ReferenceXSet xset, boolean readOnly) {
        
        if (xset.isJob()) {
            // call base class
            ReferenceJob.setJobFieldsReadOnly(xset, readOnly);
            // For all job fields set them to readonly
            for (String fieldName : QUERY_JOB_XSET_FIELDS) {
                try {
                    xset.setReadOnlyField(fieldName, readOnly);
                } catch (FieldDoesNotExistException nex) {
                    // can ignore these
                    //logger.warning("Job field does not exist, can't set readonly value." +
                     //       " Field name: " + fieldName);
                } catch (XAMException ex) {
                    // log exception and continue
                   logger.warning("Error setting job field readonly value." +
                      " Field name: " + fieldName + "Error: " +
                      ex.getMessage());
                }
            }
        }
        
    }

    /**
     * Return the Query job string used when the job was submitted. Since
     * this is a query job can return the defined constant.
     * 
     * @return the query job string
     */
    public String getJobType() {
        return XSet.XAM_JOB_QUERY;
    }
}
