/*
 * Copyright (c) 2008, Sun Microsystems, Inc.
 * Copyright (c) 2008, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.util.LinkedList;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.activation.MimeType;
import javax.activation.MimeTypeParseException;

import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidFieldTypeException;
import org.snia.xam.ObjectInUseException;
import org.snia.xam.XAMException;
import org.snia.xam.XStreamAbandonException;
import org.snia.xam.XStreamCorruptException;
import org.snia.xam.XStreamException;
import org.snia.xam.base.StreamField;
import org.snia.xam.util.XAMIOException;

/**
 *  Note: Work in Progress - not final implementation
 *
 * Stream Field implementation for ReferenceXSet container
 */
public class ReferenceStreamField extends StreamField {
    
    private Logger logger = null;

    // Location of data on disk; filename
    private String location;
    
    // Object that handles manipulating stream contents
    private XStreamContents data = null;

    // List of streams open against this field
    private LinkedList<ReferenceXStream> streamList = 
            new LinkedList<ReferenceXStream>();


    /**
     * Constructs a new ReferenceStreamField with the specified name,
     * binding state and mime-type.
     * 
     * @param name The stream name
     * @param binding The stream binding state
     * @param type The stream mime-type
     */
    public ReferenceStreamField(String name,
            boolean binding,
            String type,
            Logger externLogger )
            throws InvalidArgumentException,
            InvalidFieldTypeException,
            XAMException
    {

        this(name, type, 0, binding, false, "", externLogger );
    }
    
    /**
     * Constructs a new ReferenceStreamField with the specified name,
     * mime-type, length in bytes, binding state and read state.
     * 
     * @param name The stream name
     * @param type The stream mime-type
     * @param length The stream length in bytes
     * @param binding The stream binding state
     * @param readOnly the stream readonly state. TODO: think this has to
     * be false
     * @param location indicator as to where the actual field
     * contents are located 
     */
    public ReferenceStreamField(String name,
                    String type,
                    long length,
                    boolean binding,
                    boolean readOnly,
                    String location,
                    Logger  externLogger ) 
                    throws InvalidArgumentException,
                    InvalidFieldTypeException,
                    XAMException {

        super(name, type, length, binding, readOnly);
        validateType(type);
        this.location = location;
        logger = externLogger;
        data = new XStreamContents(location, name, length, logger);
    }
        
    protected void setLocation(String location)
            throws XAMException
    {
        this.location = location;
    }
    
    /**
     * Gets the persisted location of the stream field  contents.
     * @return the absolute path to the persisted stream field value/contents
     * @throws org.snia.xam.XAMException
     */
    protected String getAbsoluteLocation() throws XAMException   {
        return data.getAbsoluteLocation();
    }

    protected String getLocationName() throws XAMException {
        return data.getLocationName();
    }
    /**
     * Returns the name of the file the XStream can use for read/write access.
     * 
     * @param openMode
     * @return name of the working file
     * @throws org.snia.xam.InvalidArgumentException
     * @throws org.snia.xam.XAMException
     */
    protected String getWorkingFilename(String openMode) 
            throws InvalidArgumentException, XAMException {
        return data.getWorkingFilename(openMode);
    }
        

    protected XStreamContents getContents() {
        return data;
    }
    
    protected void setContents(XStreamContents contents) 
            throws XStreamException {
        data = contents;
        
        
    }

    /**
     * Commits the changes to the Stream field. Causes the payload data
     * to be persisted. At this point any streams associated with this
     * field will be closed.
     * 
     * @param dirname directory where file stored
     * @throws org.snia.xam.XAMException
     */
    protected void commit(String dirname) throws XAMException {
        long length = data.commit(dirname);
        setLength(length);
    }
    
    /**
     * Validates the mime type (not the subtype)
     * @param type
     * @throws org.snia.xam.InvalidFieldTypeException
     */
    protected void validateType(String type)
            throws InvalidFieldTypeException	{

       try { MimeType mType = new MimeType( type ); }
       catch( MimeTypeParseException mpe ) { throw new InvalidFieldTypeException(type); }
       
    }  
    
    /**
     * Adds the stream to the list of streams for this field. Should only be 
     * called when opening or creating an XStream for a field. 
     * @param stream the stream to add to the list of streams for this field
     */
    protected void addStream(ReferenceXStream stream) {
        streamList.add(stream);
    }
    
    /**
     * If the stream is closed it is removed from the list of current streams
     * for this field.
     * 
     * @param stream the stream to remove
     * @return true if the stream was removed from this field's stream list;
     *         false otherwise
     */
    private boolean removeStream(ReferenceXStream stream) {
        boolean removed = false;
        if (stream.isClosed()) {
            removed = streamList.remove(stream);
        } else {
            logger.log(Level.SEVERE, "XStream is still open, cannot remove stream" +
                    " from active list: " + this.getName());
        }
        return removed;
    }

    /**
     * Closes the specified stream and removes it from the list.
     * @param stream
     */
    protected void closeStream(ReferenceXStream stream) {
        // remove stream from persisted field contents list also
        removeStream(stream);
    }
    
    /**
     * Abandons the specified stream and sets to corrupt all the other
     * streams open on this field.
     * 
     * @throws org.snia.xam.XStreamException
     * @throws org.snia.xam.XStreamCorruptException
     * @throws org.snia.xam.XStreamAbandonException
     * @throws org.snia.xam.XAMException
     */
    protected void abandonStreams(ReferenceXStream abandonStream) 
            throws XStreamException, XStreamCorruptException,
            XStreamAbandonException, XAMException {

        for (int i = 0; i < this.streamList.size(); i++) {
            ReferenceXStream stream = streamList.get(i);
            if (stream != abandonStream) {
                if (!stream.isAbandoned() && !stream.isClosed()) {
                    stream.setStateCorrupt();
                }
            } else if (!stream.isClosed()){
                stream.setStateAbandoned();
            }
        }
          
    }
 
    protected void delete() throws XAMException, XAMIOException {
        if (!this.streamList.isEmpty()) {
           String msg = "XStreams open during XStream.delete";
           logger.warning( msg );
           throw new ObjectInUseException(msg);
        }
        this.data.delete();
    }
}
