/*
 * Copyright (c) 2008, Sun Microsystems, Inc.
 * Copyright (c) 2008, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.channels.FileChannel;
import org.snia.xam.XAMException;

/**
 *
 * Small utility class to group together the files used when exporting
 * XStreams (i.e. stream fields). Also used to write out the MIME Content
 * Headers that appear before the actual exported binary stream data.
 * 
 * There are two files. One is a tmp file containing the header information 
 * the other is the actual XStream data file. This is the actual persisted
 * XStream data.
 */
public class ReferenceStreamFieldExport {
    File[] dataList = null;
    File mimeHdrFile = null;
    File contentsFile = null;
    
    ReferenceStreamField streamField = null;
    long streamSize = 0;
    String mimeBoundary = "";
    
    protected ReferenceStreamFieldExport(
            String mimeBoundary, ReferenceStreamField field) 
        throws XAMException {
        
        // get the location and the size of data
        streamField = field;        
        String filename = streamField.getAbsoluteLocation();
        streamSize = streamField.getLength();
        contentsFile = new File(filename);
        // Write out the MIME headers
        writeMIMEHeaders(mimeBoundary);
        dataList = new File[2];
        dataList[0] = mimeHdrFile;
        dataList[1] = contentsFile;
        
    }

    protected File[] getFileList() {
        return dataList;
    }
    
    protected File getMimeHeaderFile() {
        return mimeHdrFile;
    }
    
    protected File getStreamContentsFile() {
        return contentsFile;
    }
    
    /**
     * Cleans up tmp files used to export the XStream Field contents
     *  
     * @return true if file cleanup was successful; false otherwise.
     */
    protected boolean cleanup() {
        boolean success = true;
        for (File file : dataList) {
            if (!ReferenceExportXStream.debug) {
                file.delete();
            }
        }
        return success;
    }
    /**
     * Writes out the MIME Content Headers for the exported XStream
     * @param mimeBoundary
     * @throws org.snia.xam.XAMException
     */
    private void writeMIMEHeaders(String mimeBoundary) throws XAMException {
        FileOutputStream fos = null;
        FileChannel channel = null;
        Writer out = null;
        try {
            // create temp file
            mimeHdrFile = File.createTempFile("XStreamExportHdrs_", ".tmp");
            fos = new FileOutputStream(mimeHdrFile);
            channel = fos.getChannel();
   
            out = new BufferedWriter(
                    new OutputStreamWriter(fos, XAMConstants.UTF_8));

            // write out mime boundary
            XSetXopWriter.writeMimeBoundary(
                    out, channel, mimeBoundary, false);
            
            // write out headers
            XSetXopWriter.writeLineTerminator(out, channel);
            // Write out content headers
            out.write(XmlConstants.CONTENT_TYPE_HEADER);
            out.write(streamField.getType() + "\n");
            out.write(XmlConstants.CONTENT_TRANSER_ENCODING_HEADER);
            out.write(XmlConstants.TRANSFER_TYPE_BINARY);
            out.write("\n");
            out.write(XmlConstants.CONTENT_ID_HEADER);
            out.write("<" + streamField.getName() + XAMConstants.AT_SNIA_ORG +
                    ">");
            XSetXopWriter.writeLineTerminator(out, channel);
            XSetXopWriter.writeLineTerminator(out, channel);
            
        } catch (IOException ioex) {
            throw new XAMException("Error creating temp files for " +
                    "exporting XStreams", ioex);
        } finally {
            if (out != null) {
                try {
                    out.flush();
                    out.close();
                } catch (IOException ioex) { /* ignore */ }
            }
        }
    }
}
