/*
 * Copyright (c) 2008, Sun Microsystems, Inc.
 * Copyright (c) 2008, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.File;
import java.io.FileOutputStream;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.TimeZone;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.snia.xam.XAMException;
import org.snia.xam.XUID;

/**
 * This will become the Persistence Manager for the Reference VIM.
 * The intention is to consolidate all the file, directory manipulation
 * routines into this class.
 * 
 */
public class ReferenceVIMPersistenceMgr {
    private String repositoryLocation = null;
    private ReferenceXSystem system = null;
    private static String tmpDir = null;
    private static String fileSeparator = "";
    // Constants used to create tmp and persisted files
    protected static final String TMP_FILENAME_SUFFIX = ".tmp";
    protected static final String XSET_FILE_PREFIX = "XSet_";
    protected static final String XSET_FILE_SUFFIX = ".xml";
    protected static final String OBSOLETE = "OBSOLETE_";
    private static Logger logger = null;
    
    static {
        fileSeparator = System.getProperty("file.separator");

        tmpDir = System.getProperty("java.io.tmpdir");
        if (tmpDir == null) {
            tmpDir = fileSeparator + "tmp";
        }
        
         if ( !(tmpDir.endsWith((fileSeparator)))) {
            tmpDir = tmpDir + fileSeparator;
        }

        
    }
    
    protected ReferenceVIMPersistenceMgr(ReferenceXSystem system, 
                                        String            dirName ) 
       throws Exception {
        
        this.system = system;
        logger = system.getLogger();
        if (dirName == null)
            dirName = System.getProperty("java.io.tmpdir");
        
        if (dirName == null) {
            throw new Exception("Unable to determine repository location.");
        }   
        
        if ( !(dirName.endsWith((fileSeparator)))) {
            dirName = dirName + fileSeparator;
        }
        File resposDir = new File(dirName);
        if (!resposDir.exists()) {
           String msg = "Repository location specified does not exist: " + dirName;
           logger.severe(msg);
           throw new XAMException(msg);
        }
        repositoryLocation = dirName;
        
    }
    
    protected static String getTmpDir() {
        return tmpDir;
    }
 
    /**
     * Determines whether the file is a Reference VIM temporary file
     * @param filename
     * @return true if the file is a temp file; false otherwise
     */
    protected static boolean isTmpFile(String filename) {
        boolean isTmp = false;
        filename = filename.trim();
        int tmpPos = filename.indexOf(tmpDir);
        if (tmpPos == 0) {
            String name = filename.substring(tmpDir.length());
            if (name.indexOf(fileSeparator) == -1 &&
                    name.endsWith(TMP_FILENAME_SUFFIX)) {
                isTmp = true;
            }
        }
        
        return isTmp;
    }
    
    /**
     * Prepares the XSet_ storage directory. This is called before committing
     * an XSet. If the directory does not exist it is created. If the XSet 
     * is being imported then the clearExistingContents parameter should be
     * set to true so that any existing files in the directory will be deleted.
     *  
     * @param dirName storage directory location
     * @param clearExistingContents true if any existing files stored in the
     * directory should be deleted; false otherwise
     */
    protected static void prepareStorageLocation(String dirName, 
            boolean clearExistingContents) {
        // Double check to see if this is an XSet related directory so don't
        // delete something important by mistake
        if (dirName.indexOf(XSET_FILE_PREFIX) != -1) {            
            File storageDir = new File(dirName);
            if (!storageDir.exists()) {
                storageDir.mkdir();
            } else {
                if (clearExistingContents) {
                    File files[] = storageDir.listFiles();
                    if (files != null) {
                        for (File contentFile : files) {
                            // Only delete XStream files
                            if (contentFile.getName().startsWith(
                                XStreamContents.XSTREAM_FILE_PREFIX)) {
                                if (!contentFile.delete()) {
                                    logger.log(Level.WARNING,
                                        "Failed to delete file: " +
                                        contentFile.getAbsolutePath());
                                }
                            }
                        }   // end for
                    }
                } 
            }     // end else if directory exists
        }
    }
    
    /**
     * Create the XSet directory used to store xstreams from the xuid.
     * Note: This needs to be moved to persistence mgr class.
     * @param xuid XUID of XSet
     * @return String the xset directory name for this XUID
     */ 
    protected String getXSetDirName(XUID xuid) {
        StringBuffer buf = new StringBuffer();
        if (repositoryLocation != null && repositoryLocation.length() > 0) {
            buf.append(repositoryLocation);
            if (buf.lastIndexOf(File.separator) != buf.length() -1) {
                buf.append(File.separator);
            }
        }
        buf.append(XSET_FILE_PREFIX);
        buf.append(SniaXUID.convertToSafeString(xuid));
        return buf.toString();
    }
    
    /**
     * Create the XML file name from the xuid. Used to presist XSets
     * @param xuid XUID of XSet
     * @return String the filename for this XUID
     */
    protected String getXMLFileName(XUID xuid) {
        StringBuffer buf = new StringBuffer();
        if (repositoryLocation != null && repositoryLocation.length() > 0) {
            buf.append(repositoryLocation);
            buf.append(File.separator);
        }
        buf.append(XSET_FILE_PREFIX);
        buf.append(SniaXUID.convertToSafeString(xuid) );
        buf.append(XSET_FILE_SUFFIX);
        return buf.toString();
    }

    /**
     * Returns the last part of the persisted XML file name
     * TODO:  Currently not used; decide if should remove
     * @param xuid of the XSet
     * @return the filename for the XSet
     */
    protected String getXSetFileName(XUID xuid) {
        String filename =  getXMLFileName(xuid);
        if (repositoryLocation != null & repositoryLocation.length() > 0) {
            if (filename.startsWith(repositoryLocation)) {
                File file = new File(filename);
                filename = file.getName();
            }
        }
        return filename;
    }
    
    /**
     * Gets handle to File for persisted XSet
     * @param xuid the XUID of the persisted XSet
     * @return File instance of persisted XSet
     */
    protected File getPersistedXSetFile(XUID xuid) {
        String fileName = getXMLFileName(xuid);
        File file = new File(fileName);
        return file;
    }

    /**
     * Creates the name of the stream field's persisted file
     * @param xsetXUID
     * @param streamFilename
     * @return absolute path to stream field's persisted data contents
     */
    protected String getStreamFieldAbsolutePath(
            XUID xsetXUID, String streamFilename) {
        String absolutePath = this.getXSetDirName(xsetXUID);
        absolutePath += fileSeparator;
        absolutePath += streamFilename;
        return absolutePath;
    }
    /**
     * Retrieves current repository directory
     * @return full path to directory where files are stored
     */
    protected String getRepositoryLocation() {
        return repositoryLocation;
    }
    
    /**
     * Clears out the repository value when the XSystem is closed
     * @return true if repository location was reset on close.
     */
    protected boolean resetRepositoryLocation() {
        boolean reset = false;
        if (this.system.isClosed()) {
            repositoryLocation = "";
            reset = true;
        }
        
        return reset;
    }
    /**
     * Removes the files associated with the xuid specified.
     * The XSet is removed and then the files under the XSet_ storage 
     * directory are deleted one at a time
     * 
     * This implementation implements shred by writing zeros to the file.
     *  
     * @param xuid identifier for the XSet to delete
     * @param shred Shred setting sent from higher levels. 
     */
    protected synchronized void deleteXSet(XUID xuid, boolean shred) throws XAMException {
        // remove the XSet/XML file now
        File xmlFile = getPersistedXSetFile(xuid);
        
        boolean deleted = false;
        if (xmlFile.exists()) {
           if( shred )
              deleted = shredFile(xmlFile);
           else
              deleted = xmlFile.delete();
            if (!deleted) {
                logger.log(Level.SEVERE, "Failed to remove the XSet's XML file", 
                                                            xmlFile.getName());
                throw new XAMException("Failed to delete the XSet, " +
                                                           xuid.toString());
            }
        }
        
        // Double check to see if this is an XSet related directory so don't
        // delete something important by mistake
        String dirName = getXSetDirName(xuid);
        
        /*
         * Remove the path up to the XSet directory and use the XSet directory
         * name to build up a new directory name.  In case an error occurs or
         * deleting the directory and its files fails, the new XSet directory
         * name flags it as part of a failed XSet delete.  Renaming the XSet dir
         * will make it easier to clean up the obsolete files and/or directory
         * later.  If XSet delete is completely successful, the XSet XML file
         * and its corresponding directory will be removed.  
         */ 
        int idx = dirName.lastIndexOf(File.separator);
        String xsetName = dirName.substring(++idx);
        String xsetPath = dirName.substring(0, idx);
        // XSet dir exists so rename directory to ensure proper cleanup
        File storageDir = new File(dirName);

        String newDirName = xsetPath + OBSOLETE + xsetName;
        String timeStamp = createTimeStamp();
        File renamedDir = new File(newDirName + "_" + timeStamp);

        // rename the dir if possible and then delete its files and itself
        File targetDir = storageDir.renameTo(renamedDir) ? renamedDir : storageDir;
        if (targetDir.exists()) {
            File files[] = targetDir.listFiles();
            if (files != null) {
                for (File contentFile : files) {
                   if( shred )
                      deleted = shredFile(contentFile);
                   else
                      deleted = contentFile.delete();
                    if (!deleted) {
                        logger.log(Level.SEVERE,
                                "Failed to remove the XSet directory file: " +
                                contentFile.getAbsolutePath() + "\npertaining to " +
                                "the XSet, " + xuid.toString() + 
                                "\nXUID as a hex string: " + SniaXUID.convertToSafeString(xuid));
                        throw new XAMException("Failed to delete the file," +
                                contentFile.getAbsolutePath() + "\npertaining to " +
                                "the XSet, " + xuid.toString() +
                                "\nXUID as a hex string: " + SniaXUID.convertToSafeString(xuid));
                    }
                }
            }
            deleted = targetDir.delete();
            if (!deleted) {
                logger.log(Level.SEVERE,
                        "Failed to remove the XSet directory",
                        targetDir.getName());
                throw new XAMException("Failed to delete the XSet dir, " +
                        xuid.toString());
            }
        }      
    }
    
    /**
     * "Shreds" a file by writing zeros to it. Not intended to be secure, but does
     * show a "best effort". This is a hokey little implementation that is not efficient,
     * and in fact should actually impose some penalty for marking objects to be deleted with 
     * a shred function.
     * 
     * WARNING: Do not rely upon the reference VIM to provide secure storage. This is
     * a teaching and development example.
     * @param f The file to "shred"
     */
    private boolean shredFile( File f )
    {
       final int SIZE = 1024;
       byte zeros[] = new byte[SIZE];
       
       for( int i = 0; i < SIZE; i++ )
          zeros[i] = 0;
       
       try {
          long fLength = f.length();
          FileOutputStream fos = new FileOutputStream( f );
          while( fLength > 0 ) {
             fos.write( zeros );
             fLength -= SIZE;
          }
          fos.close();
          f.delete();
       }
       catch( Exception e )
       {
          logger.log( Level.WARNING, "Shreding of file failed: " + f.getAbsolutePath() +
                        "   error == " + e.getMessage() );
          return false;
       }
       return true;
    }
    
    protected String createTimeStamp() {
        TimeZone zone = TimeZone.getTimeZone("GMT");
        Calendar c = Calendar.getInstance(zone);
        SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd_HH:mm:ssz");
        sdf.setTimeZone(zone);
        return sdf.format(c.getTime());
    }
}

