/*
 *  Copyright (c) 2009, Sun Microsystems, Inc.
 *  Copyright (c) 2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.util.logging.Logger;

import org.snia.xam.AsyncHaltedException;
import org.snia.xam.AsyncPendingException;
import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.FieldContainer;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XSet;
import org.snia.xam.XStream;
import org.snia.xam.XUID;

/**
 * The ReferenceVIM's implementation of the XAsync request object.
 */
public class ReferenceXAsync
   implements XAsync, Runnable
{
   protected enum XA_OP { NONE,
                          XSYSTEM_OPEN_XSET,
                          XSYSTEM_COPY_XSET,
                          XSET_COMMIT,
                          XSTREAM_WRITE,
                          XSTREAM_READ,
                          XSTREAM_CLOSE,
                          XSTREAM_OPEN
   };
   
   protected enum XA_STATE { NEW,
                             RUNNING,
                             HALTED,
                             ERROR,
                             COMPLETE
   };

   // XAsync state variables
   private XA_OP          m_op            = XA_OP.NONE;
   private XA_STATE       m_state         = XA_STATE.NEW;
   private long           m_xopid         = 0;
   private XUID           m_xuid          = null;
   private XSet           m_xset          = null;
   private XStream        m_xstream       = null;
   private long           m_bytes_read    = 0;
   private long           m_bytes_written = 0;
   private XAMException   m_exception     = null;
   private XAsyncListener m_listener      = null;
   
   // Operational parameters
   private ReferenceXSystem m_xsystem        = null;
   private Logger           m_logger         = null;
   private String           m_xset_mode      = null;
   private byte             m_buffer[]       = null;
   private long             m_bytes_to_read  = 0;
   private long             m_bytes_to_write = 0;

   // Used for openXStream
   private FieldContainer   m_fieldContainer = null;  
   private String           m_stream_name    = null;
   private String           m_xstream_mode   = null;

   /**
    * Build an async request to be used for XSystem operations.
    * @param xsystem The XSystem to be operated upon.
    * @param op The operation tag
    * @param opid The application supplied XOPID of this request
    */
   public ReferenceXAsync( ReferenceXSystem xsystem,
                           XA_OP            op,
                           long             opid
                           )
      throws XAMException
   {
      m_xsystem = xsystem;
      m_logger  = xsystem.getLogger();
      m_xopid   = opid;
      m_op      = op;
   }
   
   /**
    * Build an async request to be used for XSystem operations.
    * @param fc The FieldContainer to be operated upon.
    * @param op The operation tag
    * @param opid The application supplied XOPID of this request
    */
   public ReferenceXAsync( FieldContainer   fc,
                           XA_OP            op,
                           long             opid,
                           Logger           logger           
                           )
      throws XAMException
   {
      m_fieldContainer = fc;
      m_logger  = logger;
      m_xopid   = opid;
      m_op      = op;
   }
   
   /**
    * Build an async request to be used for XSet operations.
    * @param xset The XSet to be operated upon
    * @param op The operation tag
    * @param opid The application supplied XOPID of this request
    */
   public ReferenceXAsync( ReferenceXSet  xset,
                           XA_OP op,
                           long  opid )
      throws XAMException
   {
      m_xset    = xset;
      m_logger  = xset.getXSystem().getLogger();
      m_xopid   = opid;
      m_op      = op;
   }
   
   /**
    * Build an async request to be used for XStream operations.
    * @param xstream The XStream to be operated upon.
    * @param op The operation tag
    * @param opid The application supplied XOPID of this request
    */
   public ReferenceXAsync( XStream  xstream,
                           XA_OP    op,
                           long     opid,
                           Logger   extLogger )
      throws XAMException
   {
      m_xstream = xstream;
      m_logger  = extLogger;
      m_xopid   = opid;
      m_op      = op;
   }
   
   /**
    * Release any resources being consumed by this request.
    */
   public void close()
      throws AuthenticationExpiredException, XAMException
   {
      if( ! isComplete() ) {
         String msg = "Can not close XAsync, operation is not complete: XOPID == " + m_xopid;
         m_logger.warning(msg);
         throw new AsyncPendingException( msg );
      }
   }

   /** Return the number of bytes read in the XStream.asyncRead operation.
    */
   public long getBytesRead()
      throws AuthenticationExpiredException, AsyncHaltedException, XAMException
   {
      return m_bytes_read;
   }

   /**
    * Record the bytes read in the async XStream read operation.
    * @param read
    */
   private void setBytesRead(long read)
   {
      m_bytes_read = read;
   }

   /** Return the number of bytes wrutten i the XStream.asyncWrite operation.
    */
   public long getBytesWritten()
      throws AuthenticationExpiredException, AsyncHaltedException, XAMException
   {
      return m_bytes_written;
   }

   /**
    * Record the bytes written in the async XStream write operation.
    * @param read
    */
   private void setBytesWritten(long write)
   {
      m_bytes_written = write;
   }

   /** Set the field name to be used for the openXStream. */
   protected void setStreamName( String name )
   {
      m_stream_name = name;
   }
   
   /** Set the XStream open mode to be used for the openXStream. */
   protected void setStreamMode( String mode )
   {
      m_xstream_mode = mode;
   }
   
   /**
    * If the operation did not encounter an error, otherwise throw the encountered exception.
    * @throws The exception encountered in the XAsync operation.
    */
   public void getStatus()
      throws AuthenticationExpiredException, AsyncHaltedException, XAMException
   {
      if( m_exception != null )
         throw m_exception;
   }

   /**
    * Get the application supplied XOPID value.
    */
   public long getXOPID()
      throws AuthenticationExpiredException, AsyncHaltedException, XAMException
   {
      return m_xopid;
   }
   
   /**
    * Return the async operation code, usually used for debugging.
    * @return The operation code.
    */
   protected XA_OP getOp() { return m_op; }

   /**
    * Return XSet that was opened or copied by the operation.
    */
   public XSet getXSet()
      throws AuthenticationExpiredException, AsyncHaltedException, XAMException
   {
      return m_xset;
   }
   
   /**
    * Record the XSet which is a result of open or copy operation.
    * @param xset
    */
   protected void setXSet( XSet xset )
   {
      m_xset = xset;
   }

   /**
    * Get the XStream associated with the XAsync operation.
    */
   public XStream getXStream()
      throws AuthenticationExpiredException, AsyncHaltedException, XAMException
   {
      return m_xstream;
   }

   /**
    * @param stream The XStream to be operated upon.
    */
   void setXStream(XStream stream)
   {
      m_xstream = stream;
   }

   /**
    * Get the XUID, if any, associated with the XAsync operation.
    */
   public XUID getXUID()
      throws AuthenticationExpiredException, AsyncHaltedException, XAMException
   {
      return m_xuid;
   }
   
   /**
    * Set the XUID to be used for the pending operation.
    * @param inXUID
    */
   protected void setXUID( XUID inXUID )
   {
      m_xuid = inXUID;
   }

   /**
    * Set the state of this operation to halted, notify any threads which might be blocked on this
    * operation. 
    */
   public void halt()
      throws AuthenticationExpiredException, XAMException
   {
      synchronized(this) 
      {
         m_state = XA_STATE.HALTED;
         this.notifyAll();
      }
   }

   /** Indicate if the XAsync is completed, or pending. If the operation is new or running
    * this method will return false, otherwise the method returns true.
    */
   public boolean isComplete()
      throws AuthenticationExpiredException, XAMException
   {
      return (m_state == XA_STATE.COMPLETE) || (m_state == XA_STATE.ERROR) || (m_state == XA_STATE.HALTED);
   }

   // Operational specific parameters
   /**
    * Add the XAsyncListener to the operation.
    * @param listener
    */
   protected void setListener(XAsyncListener listener)
   {
      m_listener = listener;
   }

   /**
    * Set the mode used for open/copy XSet.
    * @param mode The mode from XSet interface.
    */
   public void setOpenMode(String mode)
   {
      m_xset_mode = mode;
   }

   /**
    * Record the requested number of bytes to read.
    * @param bytes_to_read
    */
   public void setBytesToRead(long bytes_to_read)
   {
      m_bytes_to_read = bytes_to_read;
   }
   
   /**
    * Record the requested number of bytes to write.
    * @param bytes_to_write
    */
   public void setBytesToWrite(long bytes_to_write)
   {
      m_bytes_to_write = bytes_to_write;
   }

   public void setBuffer(byte[] buffer)
   {
      m_buffer = buffer;
   }

   /**
    * The core run method, activated from the manager's thread. Grab an element from the 
    * request queue, execute it.
    */
   public void run()
   {
      try
      {
         switch( m_op ) {
            case NONE:
               complete( new ReferenceXAsyncException( "Illegal operation code: NONE" ));
               break;
            
            case XSYSTEM_OPEN_XSET:
               XSet xset_open = this.m_xsystem.openXSet( m_xuid, m_xset_mode );
               setXSet(xset_open);
               complete();
               break;
            
            case XSYSTEM_COPY_XSET:
               XSet xset_copy = this.m_xsystem.copyXSet( m_xuid, m_xset_mode );
               setXSet(xset_copy);
               complete();
               break;
            
            case XSET_COMMIT:
               setXUID( m_xset.commit() );
               complete();
               break;
               
            case XSTREAM_OPEN:
               XStream stream = this.m_fieldContainer.openXStream( m_stream_name, m_xstream_mode );
               setXStream(stream);
               complete();
               break;
               
            case XSTREAM_WRITE:
               setBytesWritten(m_xstream.write( m_buffer, m_bytes_to_write ));
               complete();
               break;
               
            case XSTREAM_READ:
               setBytesRead(m_xstream.read( m_buffer, m_bytes_to_read ));
               complete();
               break;
               
            case XSTREAM_CLOSE:
               m_xstream.close();
               complete();
               break;

            default:
               complete( new ReferenceXAsyncException( "Unimplemented operation code: " + m_op ));
         }
      }
      catch( Throwable t )
      {
         complete( t );
      }
      
   }

   /**
    * Finish off the operation because of an error.
    * @param Throwable
    */
   private void complete( Throwable e )
   {
      synchronized(this) 
      {
         if( e instanceof XAMException )
            m_exception = (XAMException) e;
         else
            m_exception = new ReferenceXAsyncException( "XAsync operation error due to undeclared exception: " + e.getMessage(), e );
         m_state = XA_STATE.ERROR;
         this.notifyAll();
         
         // Notify listeners, and protect from ill behaved clients.
         try 
         {
            if( m_listener != null )
               m_listener.XAsyncCallback(this);
         }
         catch( Throwable t )
         {
            m_logger.severe( "Unexpected exception occured in XASyncCallback: " + t );
         }     
      }
   }
   
   /**
    * This method records the successful competion of the asynch operation.
    */
   private void complete()
   {
      synchronized(this)
      {
         m_exception = null;
         m_state = XA_STATE.COMPLETE;
         this.notifyAll();
      }
      
      // Notify listeners, and protect from ill behaved clients.
      try 
      {
         if( m_listener != null )
            m_listener.XAsyncCallback(this);
      }
      catch( Throwable t )
      {
         m_logger.severe( "Unexpected exception occured in XASyncCallback: " + t );
      }
   }
   
   public String toString()
   {
      return "Reference XAsync: " + m_op + ": " + m_xopid;
   }
}
