/*
 *  Copyright (c) 2009, Sun Microsystems, Inc.
 *  Copyright (c) 2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.util.Vector;
import java.util.logging.Logger;
import java.util.logging.Level;

import org.snia.xam.FieldContainer;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XStream;
import org.snia.xam.XUID;

/** Take an asynchronous operation request, queues it up for execution, and when a worker 
 * thread is available, it will execute it.
 */
public class ReferenceXAsyncManager
   implements Runnable
{
   Vector<ReferenceXAsync> m_request_queue = new Vector<ReferenceXAsync>();
   
   private Logger            logger   = null;
   private ReferenceXSystem xsystem   = null;
   private boolean          m_running = true;

   public ReferenceXAsyncManager( ReferenceXSystem system, int threadCount )
      throws XAMException
   {
      xsystem = system;
      logger  = system.getLogger();
      for( int tIdx = 0; tIdx < threadCount; tIdx++ )
      {
         Thread t = new Thread( this );
         t.setName("ReferenceVIM Asynchronous Thread (" +this.hashCode() + ") " + tIdx );
         t.setDaemon(true);
         t.start();
      }
   }

   /**
    * Initiate an asynchronous openXSet operation.
    * @param referenceXSystem The XSystem used to open
    * @param inXuid The XUID of the XSet to open
    * @param mode The mode in which to open
    * @param xopid The application supplied operation ID
    * @param listener The listener to inform when the operation completes
    * @return The XAsync object representing the operation
    */
   XAsync openXSet(ReferenceXSystem xsystem,
                   XUID inXuid,
                   String mode,
                   long xopid,
                   XAsyncListener listener)
      throws XAMException
   {
      ReferenceXAsync async = new ReferenceXAsync( xsystem, ReferenceXAsync.XA_OP.XSYSTEM_OPEN_XSET, xopid );
      async.setXUID(inXuid);
      async.setOpenMode(mode);
      async.setListener( listener );
      
      synchronized( m_request_queue )
      {
         m_request_queue.add(async);
         m_request_queue.notifyAll();
      }
      return async;
   }

   /**
    * Initiate an asynchronous copyXSet operation.
    * @param referenceXSystem The XSystem used to open
    * @param inXuid The XUID of the XSet to open
    * @param mode The mode in which to open
    * @param xopid The application supplied operation ID
    * @param listener The listener to inform when the operation completes
    * @return The XAsync object representing the operation
    */
   XAsync copyXSet(ReferenceXSystem xsystem,
                   XUID inXuid,
                   String mode,
                   long xopid,
                   XAsyncListener listener)
      throws XAMException
   {
      ReferenceXAsync async = new ReferenceXAsync( xsystem, ReferenceXAsync.XA_OP.XSYSTEM_COPY_XSET, xopid );
      async.setXUID(inXuid);
      async.setOpenMode(mode);
      async.setListener( listener );
      
      synchronized( m_request_queue )
      {
         m_request_queue.add(async);
         m_request_queue.notifyAll();
      }
      return async;
   }


   /**
    * Initiate an asynchronous XSet commit operation
    * @param xset The XSet to be committed
    * @param xopid The application supplied operation ID
    * @param listener The listener to inform when the operation completes
    * @return The XAsync object representing the operation
    */
   XAsync commitXSet(ReferenceXSet  xset ,
                     long           xopid,
                     XAsyncListener listener )
      throws XAMException
   {
      ReferenceXAsync async = new ReferenceXAsync( xset, 
                                                   ReferenceXAsync.XA_OP.XSET_COMMIT, 
                                                   xopid );
      async.setXSet(xset);
      async.setListener( listener );
      
      synchronized( m_request_queue )
      {
         m_request_queue.add(async);
         m_request_queue.notifyAll();
      }
      return async;
   }

   /**
    * Initiate an asynchronous openXStream operation.
    * @param fc The fieldContainer containing the stream to open
    * @param name The stream name
    * @param mode The mode in which to open
    * @param xopid The application supplied operation ID
    * @param listener The listener to inform when the operation completes
    * @return The XAsync object representing the operation
    */
   XAsync openXSstream( FieldContainer fc,
                        String name,
                        String mode,
                        long   xopid,
                        XAsyncListener listener)
      throws XAMException
   {
      ReferenceXAsync async = new ReferenceXAsync( fc, ReferenceXAsync.XA_OP.XSTREAM_OPEN, xopid, logger );
      async.setStreamName(name);
      async.setStreamMode(mode);
      async.setListener( listener );
      
      synchronized( m_request_queue )
      {
         m_request_queue.add(async);
         m_request_queue.notifyAll();
      }
      return async;
   }

   /**
    * Initiate an asynchronous XSet commit operation
    * @param xset The XStream to be read
    * @param xopid The application supplied operation ID
    * @param listener The listener to inform when the operation completes
    * @return The XAsync object representing the operation
    */
   XAsync readStream(XStream stream,
                     byte    buffer[],
                     long    bytes_to_read,
                     long    xopid,
                     XAsyncListener   listener,
                     Logger  extLogger )
      throws XAMException
   {
      ReferenceXAsync async = new ReferenceXAsync( stream, 
                                                   ReferenceXAsync.XA_OP.XSTREAM_READ, 
                                                   xopid,
                                                   extLogger );
      async.setXStream(stream);
      async.setBuffer(buffer);
      async.setBytesToRead(bytes_to_read);
      async.setListener( listener );
      
      synchronized( m_request_queue )
      {
         m_request_queue.add(async);
         m_request_queue.notifyAll();
      }
      return async;
   }


   /**
    * Initiate an asynchronous XStream write operation
    * @param stream The XStream to write to.
    * @param buffer The byte buffer to be written.
    * @param bufferLength The number of bytes to write. Must be less than or equal to the length of buffer.
    * @param xopid The application supplied operation ID
    * @param listener The listener to inform when the operation completes
    * @return The XAsync object representing the operation
    */
   public XAsync writeStream(XStream stream,
                             byte[]           buffer,
                             long             bufferLength,
                             long             xopid,
                             XAsyncListener   listener,
                             Logger           extLogger )
      throws XAMException
   {
      ReferenceXAsync async = new ReferenceXAsync( stream, 
                                                   ReferenceXAsync.XA_OP.XSTREAM_WRITE, 
                                                   xopid, extLogger );
      async.setXStream(stream);
      async.setBuffer(buffer);
      async.setBytesToWrite(bufferLength);
      async.setListener( listener );

      synchronized( m_request_queue )
      {
         m_request_queue.add(async);
         m_request_queue.notifyAll();
      }
      return async;
   }


   /**
    * Initiate an asynchronous XStream close operation
    * @param stream The XStream to be closed
    * @param xopid The application supplied operation ID
    * @param listener The listener to inform when the operation completes
    * @return The XAsync object representing the operation
    */
   public XAsync closeStream(XStream stream,
                             long             xopid,
                             XAsyncListener   listener,
                             Logger           extLogger )
      throws XAMException
   {
      ReferenceXAsync async = new ReferenceXAsync( stream,
                                                   ReferenceXAsync.XA_OP.XSTREAM_CLOSE, 
                                                   xopid, extLogger );
      async.setXStream(stream);
      async.setListener(listener);
      synchronized (m_request_queue)
      {
         m_request_queue.add(async);
         m_request_queue.notifyAll();
      }
      return async;
   }

   /**
    * Continue to consume events from the queue, executing them as needed.
    * Note that all the bullet proofing around the logger is because the XAsync manager 
    * is created BEFORE the XSystem is finished with it's initialization phase. The logger is
    * not available until after initialization.
    */
   public void run()
   {
      while( m_running )
      {
         synchronized(m_request_queue)
         {
            try { 
               logger = xsystem.getLogger();
            }
            catch( Throwable t ) { /* sigh, ignore it */ }
            
            if( m_request_queue.size() > 0 )
            {
               ReferenceXAsync async = m_request_queue.get(0);
               m_request_queue.remove(0);
               boolean complete = true;
               try { complete = async.isComplete(); }
               catch( Throwable t ) 
               {
                  logger.log( Level.SEVERE, "Error determining if the op is complete. "+ async, t );
               }
               if( complete )
                  continue;
               
               logger.log( Level.FINEST, "Async Thread picked up job " + "(" + logger.getLevel() + ")" + async);
               async.run();
            }
            else
            {
               try { m_request_queue.wait(); } catch( InterruptedException ie ) { /* ignore */ }
            }
         }
      }
   }

   /**
    * Typically called by the owning ReferenceXSystem, when shutting down.
    */
   public void stopAsyncManager()
   {
      synchronized(m_request_queue)
      {
         m_running = false;
         m_request_queue.notifyAll();
      }
   }
}
