/*
 *  Copyright (c) 2008-2009, Sun Microsystems, Inc.
 *  Copyright (c) 2008-2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.StringReader;
import java.lang.reflect.Proxy;
import java.sql.SQLException;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.parsers.SAXParser;
import org.snia.xam.AuthenticationException;
import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.AuthorizationException;
import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.FieldExistsException;
import org.snia.xam.FieldReadOnlyException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidFieldNameException;
import org.snia.xam.InvalidFieldTypeException;
import org.snia.xam.InvalidXSetModeException;
import org.snia.xam.JobCommandException;
import org.snia.xam.JobRunningException;
import org.snia.xam.JobUnsupportedException;
import org.snia.xam.MaximumFieldException;
import org.snia.xam.ObjectInUseException;
import org.snia.xam.PolicyNameException;
import org.snia.xam.RetentionValueException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XIterator;
import org.snia.xam.XSet;
import org.snia.xam.XSetAbandonException;
import org.snia.xam.XSetCorruptException;
import org.snia.xam.XSetException;
import org.snia.xam.XSetInaccessibleException;
import org.snia.xam.XSetUnderHoldException;
import org.snia.xam.XStream;
import org.snia.xam.XStreamAbandonException;
import org.snia.xam.XStreamCorruptException;
import org.snia.xam.XStreamException;
import org.snia.xam.XSystem;
import org.snia.xam.XSystemAbandonException;
import org.snia.xam.XSystemCorruptException;
import org.snia.xam.XUID;
import org.snia.xam.FieldContainer;
import org.snia.xam.base.AbstractFieldContainer;
import org.snia.xam.base.DefaultField;
import org.snia.xam.base.Field;
import org.snia.xam.base.Property;
import org.snia.xam.base.StreamField;
import org.snia.xam.util.ISO8601Date;
import org.snia.xam.util.XAMIOException;
import org.xml.sax.InputSource;

/**
 *  Reference implementation of XSet interface. 
 * Basic implementation of single user mode functionality. No async methods
 * are implemented. Also support for default system policy only.
 * 
 * @see XSet
 */
public class ReferenceXSet extends AbstractFieldContainer
        implements XSet {
    
    private Logger logger = null;
    private String s_logClass = ReferenceXSet.class.getName();
        
    public static final String XSET_FILE_PREFIX = "XSet_";
    public static final String XSET_FILE_SUFFIX = ".xml";

    // controls access to submitting jobs 
    private Object jobLock = new Object();
    
    private XSetState xsetState = XSetState.UNRESTRICTED_WRITER;
    
    // If the XSet was filled via import information, this gets set and
    // special commit time operations occur. See commit()
    private boolean   wasImported = false;
    
    private XUID xuid = null;
    private String openMode = XSet.MODE_UNRESTRICTED;

    // Determines if opening existing XSet. Need to be able to create
    // properties/streams and not set dirty field
    private boolean loadingXSet = false;
    
    // XSystem that created this XSet
    private ReferenceXSystem xsystem = null;

    // Contains open stream fields
    private LinkedList<ReferenceXStreamBase> streamList = 
            new LinkedList<ReferenceXStreamBase>();
    
    private LinkedList<ReferenceStreamField> deletedStreamFields =
            new LinkedList<ReferenceStreamField>();
    
    private ReferenceExportXStream xsetExportStream     = null;
    private ReferenceImportXStream xsetImportStream     = null;
    
    /*
     * Object which contains retention management logic and
     * contains retention helper methods
     */
    private RetentionMgr retentionMgr = null;
    
    /**
     * Used to verify policy settings when importing an XSet to make sure that
     * existing policies don't conflict, or the result will be to reduce retention.
     * If this value is null, the check does not take place.
     */
    HashMap<String,RetentionPolicyReference> m_retention_policy_references = null;

    /* Empty XSet */ 
    ReferenceXSet() {        
        xsetState = XSetState.UNRESTRICTED_WRITER;    
    }

    /**
     * Creates a new XSet
     * @param parentXSystem ReferenceXSystem the XSystem requesting the XSet open
     * @param createMode The creation XSet.MODE_READ_ONLY, XSet.MODE_RESTRICTED or
     *            XSet.MODE_UNRESTRICTED
     * @throws XAMException
     * @see XSet#MODE_READ_ONLY
     * @see XSet#MODE_RESTRICTED
     * @see XSet#MODE_UNRESTRICTED
     * @throws org.snia.xam.XAMException
     */
    protected ReferenceXSet(ReferenceXSystem parentXSystem, String createMode) 
    throws XAMException {
        this(parentXSystem, createMode, false);
    }
    
    /**
     * Creates a new XSet
     * @param parentXSystem the XSystem requesting the XSet open
     * @param createMode The creation XSet.MODE_READ_ONLY, XSet.MODE_RESTRICTED
     *            or XSet.MODE_UNRESTRICTED
     * @param existing true if the XSet exists and is being opened, false if
     *      a new XSet is being created.
     * @throws XAMException
     * @see XSet#MODE_READ_ONLY
     * @see XSet#MODE_RESTRICTED
     * @see XSet#MODE_UNRESTRICTED
     */
    protected ReferenceXSet( ReferenceXSystem parentXSystem, 
                             String           createMode, 
                             boolean          existing) 
    throws XAMException {
        
       xsystem      = parentXSystem;
       openMode     = createMode;
       retentionMgr = new RetentionMgr(this);

       // NOTE:  Do NOT create any management value properties when 
        // instantiating an XSet -- mgmt properties should be created at
        // commit time if needed/required.
        if (parentXSystem == null) {
            throw new InvalidArgumentException("XSystem can not be null");
        }
        
        logger       = parentXSystem.getLogger();
        super.setLogger(logger);
        super.setVerbosity(parentXSystem.m_enterExit);

        if ((createMode == null) ||
               !(createMode.equals(XSet.MODE_READ_ONLY) || 
               createMode.equals(XSet.MODE_RESTRICTED) ||
               createMode.equals(XSet.MODE_UNRESTRICTED))) {
           String msg = "Unrecognized XSet mode: +" + createMode;
           logger.warning(msg);
           throw new InvalidXSetModeException(msg);
        }
        
        Calendar calendar = Calendar.getInstance();
        xsetState = XSetState.getXSetState(XSet.MODE_UNRESTRICTED);

        if (!existing) {
            addField(new Property(
                    XAM_TIME_CREATION,
                    XAMLibrary.STYPE_DATETIME_MIME_TYPE,
                    calendar,
                    true, 
                    true));
            addField(new Property(
                    XAM_TIME_ACCESS,
                    XAMLibrary.STYPE_DATETIME_MIME_TYPE,
                    calendar,
                    false, 
                    true));
        }

        xsetState = XSetState.getXSetState(openMode);
    }
    
    
   /**
     * Creates a new XSet instance based on an already persisted XSet
     * @param parentXSystem the XSystem requesting the XSet open
     * @param xuid The XUID of the XSet to open
     * @param openMode The creation XSet.MODE_READ_ONLY, XSet.MODE_RESTRICTED or
     *            XSet.MODE_UNRESTRICTED
     * @throws XAMException
     * @see XSet#MODE_READ_ONLY
     * @see XSet#MODE_RESTRICTED
     * @see XSet#MODE_UNRESTRICTED
     */
    protected ReferenceXSet(ReferenceXSystem parentXSystem, 
            XUID xuid, String openMode)
    throws XAMException {
        

        this(parentXSystem, openMode, true);
        
        // Load persisted XSet information
        // set intial mode so can add all fields
        //String fileName = xsystem.getPersistenceMgr().getXMLFileName(xuid);
        this.openMode = XSet.MODE_UNRESTRICTED;
        xsetState = XSetState.getXSetState(XSet.MODE_UNRESTRICTED);
        XSetXmlReader handler = new XSetXmlReader(this);
        this.xuid = xuid;
        loadingXSet = true;

        try {
            File file = xsystem.getPersistenceMgr().getPersistedXSetFile(xuid);
            SAXParser parser = XSetXmlReader.factory.newSAXParser();
            parser.parse(file, handler);
            
            // The parser can not add the system fields because it is
            // does not have the right access privs. So load them now.
            processSystemFields(xuid, handler.getSystemFields());
     
            // set all properties isModified to false
            setPropsUnmodified();
            
            persistAccessTime();
            
            // Set correct mode after load all properties
            this.openMode = openMode;
            xsetState = XSetState.getXSetState(openMode);
            
        } catch( FileNotFoundException fne ) {
           // NOTE: It is important not to include any information about why it's not available 
           // This would be considered a security leak and applications could "probe" to see if
           // XSets are really there, or permissions prevent reading.
           throw new XSetInaccessibleException();
        } catch (Exception e) {
            System.err.println("Error: " + e);
            e.printStackTrace();
            XAMException xex = new XAMException("Error opening XSet", e);
            throw(xex);
        } finally {
            loadingXSet = false;
        }
    }

    /**
     * Creates a new XSet instance based on an already persisted XSet. This method creates an 
     * XSet which will never change any setting in the XSet. Typically this might be used 
     * by the RefereceXSystem when grabbing some value from the XSet.
     * @param parentXSystem the XSystem requesting the XSet open
     * @param xuid The XUID of the XSet to open
     * @throws XAMException
     */
    protected ReferenceXSet(ReferenceXSystem parentXSystem, XUID xuid )
    throws XAMException {
        

        this(parentXSystem, XSet.MODE_READ_ONLY, true);
        
        // Load persisted XSet information
        this.openMode = XSet.MODE_READ_ONLY;

        XSetXmlReader handler = new XSetXmlReader(this);

        this.xuid = xuid;
        loadingXSet = true;

        try {
            xsetState = XSetState.getXSetState(XSet.MODE_UNRESTRICTED);
            this.openMode = XSet.MODE_UNRESTRICTED; // Allows parser to load props
            File file = xsystem.getPersistenceMgr().getPersistedXSetFile(xuid);
            SAXParser parser = XSetXmlReader.factory.newSAXParser();
            parser.parse(file, handler);
            
            // The parser can not add the system fields because it is
            // does not have the right access privs. So load them now.
            processSystemFields(xuid, handler.getSystemFields());
     
            // set all properties isModified to false
            setPropsUnmodified();
            
        } catch (Exception e) {
            XAMException xex = new XAMException("Error opening XSet", e);
            throw(xex);
        } finally {
            loadingXSet = false;
            this.openMode = XSet.MODE_READ_ONLY;
            xsetState = XSetState.getXSetState(XSet.MODE_READ_ONLY);
        }
    }

    /**
     * This method is available for use by opening the XSet to update the access time
     * and to make the change persistent. This method should only be called early in 
     * openXSet path (normal constructor with mode). When this is called early in 
     * the open path, only the AccessTime is modified, and the commit time is not 
     * changed.
     * 
     * @throws XAMException 
     * @throws ObjectInUseException 
     */
    private void persistAccessTime() throws XAMException 
    
    {
       // This probably needs to be removed.
        if (this.isJob()) {
            if (JobMgr.isJobRunningForXSet(this)) {
               String msg = "Can not commit job when it is still running";
               logger.warning(msg);
                throw new ObjectInUseException(msg);
            }
            // Make sure job fields are set to read/write
            ReferenceQueryJob.setJobFieldsReadOnly(this, false);
        }
        // If this is a new XSet, we shouldn't try to persist this.
        // In fact, this method should not have been called.
        if (! containsField(XAM_XUID))
           return;
         
        Calendar now = Calendar.getInstance();
        
        // update the times for these fields
        Field f = getField( XAM_TIME_ACCESS );
        removeField(f);
        addField(new Property(
                XAM_TIME_ACCESS,
                XAMLibrary.STYPE_DATETIME_MIME_TYPE,
                now,
                false, 
                true));

        // Persist this
        try {
            XSetXmlWriter.outputXMLToFile(this);
            String nowStr = ISO8601Date.toISO8601String(now);
            xsystem.getDBManager().storeField( XSet.XAM_TIME_ACCESS,
                                              XAMLibrary.STYPE_DATETIME_MIME_TYPE,
                                              nowStr,
                                              0.0,
                                              0,
                                              false, /* binding */
                                              true,  /* readonly */
                                              nowStr.getBytes("utf-8").length,
                                              getXUID() );
                        
        } catch (XAMException ex) {
            xsetState = XSetState.CORRUPT;
            String msg = "Error persisting XSet. Cannot open or create file";
            logger.log( Level.SEVERE, msg, ex);
            throw new XSetCorruptException(msg, ex);
        } catch( SQLException seq) {
           xsetState = XSetState.CORRUPT;
           String msg = "SQL error persisting access time, db error";
           logger.log(Level.SEVERE, msg, seq );
           throw new XSetCorruptException( msg, seq );
        } catch( Exception e ) {
           xsetState = XSetState.CORRUPT;
           String msg = "Unknown exception corrupted XSet";
           logger.log( Level.SEVERE, msg, e );
           throw new XSetException( msg, e );
        }
    }

    /**
     * Processes the system fields when loading an existing XSet from
     * persistent storage. 
     * 
     * @param xuid the xuid of the XSet
     * @param fieldList the list of fields to process
     * @throws org.snia.xam.XAMException
     */
    private void processSystemFields(XUID xuid, Map fieldList) 
            throws XAMException {
        Collection list = fieldList.values();
        Iterator fields = list.iterator();
        String xuidValue = XAMConstants.EMPTY_STRING;

        XUID xuidInstance = null;
        while (fields.hasNext()) {
            Field field = (Field) fields.next();
            String fieldName = field.getName();
            if (xuidInstance == null && fieldName.equals(XAM_XUID)) {
                if (field instanceof Property) {
                    Property prop = (Property) field;
                    xuidInstance = (XUID) prop.getValue();
                    xuidValue = xuidInstance.toString();                    
                }
            }
            retentionMgr.checkFieldAsOtherRetention(field);
        }
        if (xuid != null) {
            String expectedXUID = xuid.toString();
            if (!expectedXUID.equals(xuidValue)) {
                xsetState = XSetState.CORRUPT;
                String msg = "Error loading XSet: Expected XUID: " + xuid.toString() + "  Found XUID: " + xuidValue;
                logger.severe(msg);
                throw new XSetCorruptException(msg);
            }
            
        } else {
            this.xuid = xuidInstance;
        }
        
    }

    /**
     * Retrieves the parent XSystem for this XSet
     * @return
     */
    ReferenceXSystem getXSystem() {
        return this.xsystem;
    }
    /**
     * @see XSet#applyAccessPolicy(boolean, java.lang.String)
     */
    public void applyAccessPolicy(boolean binding, String policyName)
       throws AuthenticationExpiredException, MaximumFieldException, 
              PolicyNameException, ObjectInUseException, 
              XSetAbandonException, XSetCorruptException, XAMException {
       final String method = "applyAccessPolicy";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );
       
       /*
        * Determine if the passed in policy name is known to the XSystem. This
        * is done by getting all the published access policies in the XSystem. 
        * Note: If this was a production system, a more efficient mechanism 
        * should be used. 
        */
       XIterator policies = xsystem.openFieldIterator(ReferenceXSystem.XAM_XSYSTEM_ACCESS_POLICY_LIST);
       
       while( policies.hasNext() ) {
          String fieldName = (String) policies.next();
          
          if( fieldName.endsWith( policyName )) {
             addField(new Property(
                XSet.XAM_ACCESS_POLICY,
                XAMLibrary.STYPE_STRING_MIME_TYPE,
                policyName,
                binding,
                false));
             if( xsystem.m_enterExit )
                logger.exiting( s_logClass, method );
             return;
          }
       }
       String msg = "Access policy name not known to XSystem";
       logger.warning(msg);
       throw new PolicyNameException( msg );
    }
    
    /**
     * @see XSet#resetAccessFields() 
     */
    public void resetAccessFields()
       throws AuthenticationExpiredException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException, XAMException {
       final String method = "resetAccessFields";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        xsystem.checkAccessAllowed();
        if( containsField( XSet.XAM_ACCESS_POLICY ) ) {
           Property p = (Property) getField( XSet.XAM_ACCESS_POLICY);
           boolean bound = p.getBinding();
           // Don't call the superclass "deleteField" since it requires foolishness
           // the "protected" name space setting. While it would be possible to 
           // reset the protected setting for the duration of the call, it may 
           // make the XSet non thread safe.
           Field f = getField(XSet.XAM_ACCESS_POLICY);
           removeField(f);
           this.setXSetDirty(XSet.XAM_ACCESS_POLICY, bound);
        }
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }
    
    /**
     * @see XSet#applyManagementPolicy(boolean, java.lang.String) 
     */
    public void applyManagementPolicy(boolean binding, String policyName)
       throws MaximumFieldException, PolicyNameException, 
              ObjectInUseException, XSetAbandonException, 
              XSetCorruptException, XAMException {
       final String method = "applyManagementPolicy";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        xsystem.checkAccessAllowed();
        String listName = XSystem.XAM_MANAGEMENT_POLICY_LIST + policyName;
        
        XSystem xsys = this.getXSystem();
        if( ! xsys.containsField(listName) )
           throw new PolicyNameException( "Management policy " + policyName + " not known to system");
       
        boolean bindingChange = false;
        if( containsField(XSet.XAM_MANAGEMENT_POLICY) ) {
           Property p = (Property) getField(XSet.XAM_MANAGEMENT_POLICY);
           bindingChange = binding ^ p.getBinding();
           deleteField(XSet.XAM_MANAGEMENT_POLICY);
        }
           
        addField(new Property(
           XSet.XAM_MANAGEMENT_POLICY,
           XAMLibrary.STYPE_STRING_MIME_TYPE,
           policyName,
           binding,
           false));
        this.setXSetDirty(XSet.XAM_MANAGEMENT_POLICY, bindingChange);
        
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }
    
    /**
     * @see XSet#resetManagementFields() 
     */
    public void resetManagementFields() 
       throws AuthenticationExpiredException, ObjectInUseException, 
              XSetAbandonException, XSetCorruptException, XAMException {
       final String method = "resetManagementFields";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

       xsystem.checkAccessAllowed();
       if( containsField( XSet.XAM_MANAGEMENT_POLICY) ) {
          Property p = (Property) getField( XSet.XAM_MANAGEMENT_POLICY );
          boolean bound = p.getBinding();
          Field f = getField(XSet.XAM_MANAGEMENT_POLICY);
          this.setXSetDirty(XSet.XAM_MANAGEMENT_POLICY, bound);
          removeField(f);
       }
       
       XIterator retentions = openFieldIterator( XSet.XAM_RETENTION_LIST );
       while( retentions.hasNext() )
       {
          String listPropName = retentions.next();
          String retention_id = getString(listPropName);
          String prefix       = ".xset.retention." + retention_id;
          boolean bound       = false;
          
          Field f = getField( prefix + ".enabled" );
          if( f != null ) 
          {
             bound = f.getBinding();
             setXSetDirty( f.getName(), bound );
             removeField(f);
          }

          f = getField( prefix + ".enabled.policy" );
          if( f != null )
          {
             bound = f.getBinding();
             setXSetDirty( f.getName(), bound );
             removeField(f);
          }

          f = getField( prefix + ".duration" );
          if( f != null ) 
          {
             bound = f.getBinding();
             setXSetDirty( f.getName(), bound );
             removeField(f);
          }

          f = getField( prefix + ".duration.policy" );
          if( f != null )
          {
             bound = f.getBinding();
             setXSetDirty( f.getName(), bound );
             removeField(f);
          }

          f = getField( listPropName );
          if( f != null )
          {
             bound = f.getBinding();
             setXSetDirty( f.getName(), bound );
             removeField(f);
          }
       }
       if( xsystem.m_enterExit )
          logger.exiting( s_logClass, method );
    }
    
    /**
     * @see XSet#createRetention(boolean, java.lang.String) 
     */
    public void createRetention(boolean binding, String  retentionID)
       throws AuthenticationExpiredException, InvalidArgumentException, 
              InvalidFieldNameException, ObjectInUseException, 
              XSetAbandonException, XSetCorruptException, XAMException {
       final String method = "createRetention";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        RetentionMgr.validateRetentionID(retentionID);
        retentionMgr.createRetention(retentionID, binding);
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }
    
    /**
     * @see XSet#setRetentionEnabledFlag(java.lang.String, boolean, boolean)  
     */
    public void setRetentionEnabledFlag(String retentionID, 
                                        boolean binding, 
                                        boolean enabled)
            throws AuthenticationExpiredException, InvalidArgumentException,
            InvalidFieldNameException, ObjectInUseException,
            XSetAbandonException, XSetCorruptException, XAMException {
       final String method = "setRetentionEnabledFlag";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );
        
        RetentionMgr.validateRetentionID(retentionID);
        retentionMgr.setEnabled(retentionID, binding, enabled);
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }

    /**
     * @see XSet#applyRetentionEnabledPolicy(java.lang.String, boolean, java.lang.String) 
     */
    public void applyRetentionEnabledPolicy(String retentionID, boolean binding, String policyName)
       throws AuthenticationExpiredException, InvalidArgumentException,
              InvalidFieldNameException, PolicyNameException, 
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XAMException {
       
       final String method = "applyRetentionEnabledPolicy";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

       xsystem.checkAccessAllowed();
       RetentionMgr.validateRetentionID(retentionID);
       
       boolean oldEnabled = this.getActualRetentionEnabled(retentionID);
       RetentionPolicy policy = RetentionMgr.GetPolicy(policyName);
       if( policy == null ) {
          String msg = "Retention enabled policy '" + policyName + "' not known to XSystem";
          logger.warning(msg);
          throw new PolicyNameException( msg );
       }

       if( oldEnabled && ! policy.getEnabled() ) {
          String msg = "Retention policy " + policyName + " may not disable existing retention";
          logger.warning(msg);
          throw new RetentionValueException( msg );
       }
        
       String fieldName = ".xset.retention." + retentionID + ".enabled.policy";
       addField(new Property(
          fieldName,
          XAMLibrary.STYPE_STRING_MIME_TYPE,
          policyName,
          binding,
          false));
       
       this.setXSetDirty(fieldName, binding);
       if( xsystem.m_enterExit )
          logger.exiting( s_logClass, method );
       return;
    }
    
    /**
     * @see XSet#getActualRetentionEnabled(java.lang.String) 
     */
    public boolean getActualRetentionEnabled(String retentionID)
       throws AuthenticationExpiredException, InvalidArgumentException,
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XAMException {
       final String method = "getActualRetentionEnabled";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        xsystem.checkAccessAllowed();
        
        boolean enabled = false;
        String enabledField = ".xset.retention." + retentionID + ".enabled";
        if (containsField(enabledField)) {
           enabled = getBoolean(enabledField);
        } else {       
           String policyFieldName = enabledField + ".policy";
           if( ! enabled && containsField(policyFieldName)) {
              String policyName = getString(policyFieldName);
              RetentionPolicy policy = RetentionMgr.GetPolicy(policyName);
              enabled = policy.getEnabled();
           }
        }
        
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
        return enabled;
    }
    
    /**
     * @see XSet#setRetentionDuration(java.lang.String, boolean, long) 
     */
    public void setRetentionDuration(String retentionID, boolean binding, long duration)
       throws AuthenticationExpiredException, InvalidArgumentException,
              InvalidFieldNameException, RetentionValueException,
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XAMException {
               
       final String method = "setRetentionDuration";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        RetentionMgr.validateRetentionID(retentionID);
        if (!RetentionMgr.isDurationValid(duration)) {
            throw new InvalidArgumentException("Retention duration specified " +
                "for \"" + retentionID + "\" must be a non-negative number " +
                "of milliseconds or \"forever\", " + "which is denoted by -1.");
        }
        retentionMgr.setDuration(retentionID, binding, duration);
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }
    
    /**
     * @see XSet#applyRetentionDurationPolicy(java.lang.String, boolean, java.lang.String) 
     */
    public void applyRetentionDurationPolicy(String retentionID,
                                              boolean binding,
                                              String  policyName) 
       throws AuthenticationExpiredException, InvalidArgumentException,
              InvalidFieldNameException, PolicyNameException,
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XAMException {
       
       final String method = "applyRetentionDurationPolicy";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

       xsystem.checkAccessAllowed();
       RetentionMgr.validateRetentionID(retentionID);
       
       long oldDuration = this.getActualRetentionDuration(retentionID);
       RetentionPolicy policy = RetentionMgr.GetPolicy(policyName);
       if( policy == null ) {
          String msg = "Retention policy '" + policyName + "' not known to XSystem";
          logger.warning(msg);
          throw new PolicyNameException( msg );
       }

       if( policy.getDuration() < oldDuration ) {
          String msg = "Policy " + policyName + " has duration less than current retention policy";
          logger.warning(msg);
          throw new RetentionValueException( msg );
       }
        
       String fieldName = ".xset.retention." + retentionID + ".duration.policy";
       addField(new Property(
          fieldName,
          XAMLibrary.STYPE_STRING_MIME_TYPE,
          policyName,
          binding,
          false));
       this.setXSetDirty(fieldName, binding);
       
       if( xsystem.m_enterExit )
          logger.exiting( s_logClass, method );
       return;
    }
    
    /**
     * @see XSet#getActualRetentionDuration(java.lang.String) 
     */
    public long getActualRetentionDuration(String retentionID)
       throws AuthenticationExpiredException, InvalidArgumentException,
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XAMException {
       final String method = "getActualRetentionDuration";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

       xsystem.checkAccessAllowed();
       long duration = 0;
       String durationField = ".xset.retention." + retentionID + ".duration";
       if (containsField(durationField)) {
          duration = getLong(durationField);
       } else {
          String policyFieldName = durationField + ".policy";
          if( containsField(policyFieldName)) {
             String policyName = getString(policyFieldName);
             RetentionPolicy policy = RetentionMgr.GetPolicy(policyName);
             duration = policy.getDuration();
          }
       }
       
       if( xsystem.m_enterExit )
          logger.exiting( s_logClass, method );
       return duration;
    }
    
    /**
     * @see XSet#setRetentionStarttime(java.lang.String, boolean) 
     */
    public void setRetentionStarttime(String retentionID, boolean binding) 
       throws AuthenticationExpiredException, FieldExistsException,
              InvalidArgumentException, InvalidFieldNameException,
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XAMException {
        final String method = "setRetentionStarttime";
        if( xsystem.m_enterExit )
           logger.entering( s_logClass, method );

        RetentionMgr.validateRetentionID(retentionID);
        retentionMgr.setStarttime(retentionID, binding);
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }
    
    /**
     * @see XSet#setBaseRetention(boolean, long) 
     */
    public void setBaseRetention(boolean binding, long duration)
       throws AuthenticationExpiredException, RetentionValueException, 
              ObjectInUseException, XSetAbandonException, 
              XSetCorruptException, XAMException {
       final String method = "setBaseRetention";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        if (!RetentionMgr.isDurationValid(duration)) {
           String msg = "Retention duration specified " +
           "for the " + RetentionMgr.BASE + " retention identifier must " +
           "be a positive value or \"forever\", which is denoted by -1.";
           logger.warning(msg);
           throw new RetentionValueException(msg);
        }
        retentionMgr.setBase(binding, duration);
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }  
    
    /**
     * @see XSet#applyBaseRetentionPolicy(boolean, java.lang.String) 
     */
    public void applyBaseRetentionPolicy(boolean binding, String policyName)
       throws AuthenticationExpiredException, InvalidArgumentException,
              InvalidFieldNameException, PolicyNameException, 
              ObjectInUseException, XSetAbandonException, 
              XSetCorruptException, XAMException {
       final String method = "applyBaseRetentionPolicy";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

       xsystem.checkAccessAllowed();
       String id = "base";
       
       long oldDuration = this.getActualRetentionDuration(id);
       boolean oldEnabled = this.getActualRetentionEnabled(id);

       RetentionPolicy policy = RetentionMgr.GetPolicy(policyName);
       if( policy == null ) {
          String msg = "Retention policy '" + policyName + "' not known to XSystem";
          logger.warning(msg);
          throw new PolicyNameException( msg );
       }

       if( policy.getDuration() < oldDuration ) {
          String msg = "Policy " + policyName + " has duration less than current retention policy";
          logger.warning(msg);
          throw new RetentionValueException( msg );
       }
                
       if( oldEnabled && ! policy.getEnabled() ) {
          String msg = "Retention policy " + policyName + " may not disable existing retention";
          logger.warning(msg);
          throw new RetentionValueException( msg );
       }
       
       String fieldName1 = ".xset.retention." + id + ".duration.policy";
       addField(new Property(
          fieldName1,
          XAMLibrary.STYPE_STRING_MIME_TYPE,
          policyName,
          binding,
          false));

       String fieldName2 = ".xset.retention." + id + ".enabled.policy";
       addField(new Property(
          fieldName2,
          XAMLibrary.STYPE_STRING_MIME_TYPE,
          policyName,
          binding,
          false));
       this.setXSetDirty(fieldName1, binding);
       if( xsystem.m_enterExit )
          logger.exiting( s_logClass, method );
    }    
    
    /**
     * @see XSet#setAutoDelete(boolean, boolean) 
     */
    public void setAutoDelete(boolean binding, boolean autoDelete)
       throws AuthenticationExpiredException, MaximumFieldException, 
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XAMException {
       final String method = "setAutoDelete";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

       xsystem.checkAccessAllowed();
       addField(new Property(
          XAM_DELETION_AUTODELETE,
          XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
          Boolean.valueOf(autoDelete),
          binding,
          false));
       if( xsystem.m_enterExit )
          logger.exiting( s_logClass, method );
    }
    
    /**
     * @see XSet#applyAutoDeletePolicy(boolean, java.lang.String) 
     */
    public void applyAutoDeletePolicy(boolean binding, String policyName)
       throws AuthenticationExpiredException, PolicyNameException,
              ObjectInUseException, XSetAbandonException, 
              XSetCorruptException, XAMException {
       final String method = "applyAutoDeletePolicy";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        xsystem.checkAccessAllowed();
        addField(new Property(
                XAM_DELETION_AUTODELETE_POLICY,
                XAMLibrary.STYPE_STRING_MIME_TYPE,
                policyName,
                binding,
                false));
        
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }
    
    /**
     * @see XSet#getActualAutoDelete() 
     */
    public boolean getActualAutoDelete() 
       throws AuthenticationExpiredException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException, XAMException {
       final String method = "getActualAutoDelete";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

       xsystem.checkAccessAllowed();
       try { return getBoolean(XSet.XAM_DELETION_AUTODELETE); }
       catch( FieldDoesNotExistException fdne ) { /* ignore */ }
       
       boolean autodel = false;
       String delPolicy = null;
       try { 
          delPolicy = getString(XSet.XAM_DELETION_AUTODELETE_POLICY);
          autodel = DispositionPolicy.isAutodelete(delPolicy);
       } catch( FieldDoesNotExistException fdne2 ) { /* ignore */ }

       if( xsystem.m_enterExit )
          logger.exiting( s_logClass, method );

       return autodel;
    }

    /**
     * @see XSet#setShred(boolean, boolean) 
     */
    public void setShred(boolean binding, boolean shred)
       throws AuthenticationExpiredException, MaximumFieldException, 
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XAMException {
       final String method = "setShred";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

       xsystem.checkAccessAllowed();
       addField(new Property(
          XAM_DELETION_SHRED,
          XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
          Boolean.valueOf(shred),
          binding,
          false));
       if( xsystem.m_enterExit )
          logger.exiting( s_logClass, method );
    }
    
    /**
     * @see XSet#applyShredPolicy(boolean, java.lang.String) 
     */
    public void applyShredPolicy(boolean binding, String policyName)
       throws AuthenticationExpiredException, MaximumFieldException,
              PolicyNameException, ObjectInUseException, 
              XSetAbandonException, XSetCorruptException, XAMException {
       final String method = "applyShredPolicy";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        xsystem.checkAccessAllowed();
        addField(new Property(
                XAM_DELETION_SHRED_POLICY,
                XAMLibrary.STYPE_STRING_MIME_TYPE,
                policyName,
                binding,
                false));
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }
    
    /**
     * @see XSet#getActualShred() 
     */
    public boolean getActualShred() 
       throws AuthenticationExpiredException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException, XAMException {
       final String method = "getActualShred";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        xsystem.checkAccessAllowed();
        try { return getBoolean(XSet.XAM_DELETION_SHRED); }
        catch( FieldDoesNotExistException fdne ) { /* ignore */ }

        boolean shred = false;
        String shredPolicy = null;
        try { 
           shredPolicy = getString(XSet.XAM_DELETION_SHRED_POLICY);
           shred = DispositionPolicy.isAutodelete(shredPolicy);
        } catch( FieldDoesNotExistException fdne2 ) { /* ignore */ }

        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
        return shred;
    }  
    
    /**
     * @see XSet#applyStoragePolicy(boolean, java.lang.String) 
     */
    public void applyStoragePolicy(boolean binding, String policyName)
       throws AuthenticationExpiredException, MaximumFieldException, 
              PolicyNameException, ObjectInUseException, 
              XSetAbandonException, XSetCorruptException, XAMException {
       final String method = "applyStoragePolicy";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        xsystem.checkAccessAllowed();
        addField(new Property(
                XAM_STORAGE_POLICY,
                XAMLibrary.STYPE_STRING_MIME_TYPE,
                policyName,
                binding,
                false));
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }
    
    // == XSet Operations ==
    /**
     * @see XSet#commit()
     */
    public XUID commit() 
    throws AuthenticationExpiredException, AuthorizationException, 
              JobRunningException, XSetUnderHoldException, 
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XAMException {
        
       final String method = "commit";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        XUID xuidTmp = null;
        checkAccessAllowed(XSetOperation.COMMIT);
        if (this.isJob()) {
            if (JobMgr.isJobRunningForXSet(this)) {
               String msg = "Can not commit job when it is still running";
               logger.warning(msg);
                throw new ObjectInUseException(msg);
            }
            // Make sure job fields are set to read/write
            ReferenceQueryJob.setJobFieldsReadOnly(this, false);
        }
        // Check all XStreams are closed
        checkXStreamsClosed();
        
        if( wasImported )
        {
           checkBindingChanges();
           this.retentionMgr.validateRetentionPeriods();
        }
        
        if (containsField(XAM_XUID))
            xuidTmp = getXUID(XAM_XUID);
        boolean createXUID = false;
         
        // If there is no XUID for this XSet create it
        if (xuidTmp == null) {
            createXUID = true;
        } else if (hasBindingFieldChanged()) {
            // If a binding field has changed then we have to update the XUID. 
            createXUID = true;
        }
        // Check to see if the commit time has changed between the persisted
        // xset and this xset instance
        if (!createXUID && !wasImported &&
                hasCommitTimeChanged()) {
            xsetState = XSetState.CORRUPT;
            String msg = "XSet has been changed since being originally opened.";
            logger.warning( msg );
            throw new XSetCorruptException(msg);
        }

        // make sure the following management fields exist for each XSet:
        // .xset.management.policy, .xset.retention.list.event, 
        // .xset.retention.list.base and .xset.retention.base.enabled,
        // .xset.retention.base.starttime and .xset.hold
        Calendar now = Calendar.getInstance();
        setManagementFields(now);

        validateAllPolicies();
        
        if (createXUID) {
            xuidTmp = SniaXUID.generateXUID();
            this.xuid = xuidTmp;
            addField(new Property(
                XAM_XUID,
                XAMLibrary.STYPE_XUID_MIME_TYPE,
                xuidTmp,
                true,
                true));
            addField(new Property(
                XAM_TIME_XUID,
                XAMLibrary.STYPE_DATETIME_MIME_TYPE,
                now,
                false, 
                true));
            addField(new Property(
               XAM_TIME_RESIDENCY,
               XAMLibrary.STYPE_DATETIME_MIME_TYPE,
               now,
               false,
               true ));
        }
        // update the times for these fields
        addField(new Property(
                XAM_TIME_COMMIT,
                XAMLibrary.STYPE_DATETIME_MIME_TYPE,
                now,
                false, 
                true));
        
        addField(new Property(
                XAM_TIME_ACCESS,
                XAMLibrary.STYPE_DATETIME_MIME_TYPE,
                now,
                false, 
                true));
        
        /* Committing XSet now in Clean state.
         * Remove dirty field
         */
        try {
            Property p = (Property) getField(XAM_DIRTY);
            if (p != null) {
                removeField(p);
            }
        } catch (XAMException ex) {}
                
        String dirname = xsystem.getPersistenceMgr().getXSetDirName(this.xuid);
        
        try {
            boolean clearExistingContent = 
                    wasImported ? true : false;
            ReferenceVIMPersistenceMgr.prepareStorageLocation(dirname, 
                    clearExistingContent);

            deleteStreamFields(dirname);
            commitXStreamContentChanges(dirname);
            XSetXmlWriter.outputXMLToFile(this);
            
            // Send to the database
            xsystem.getDBManager().storeXSet( this );
            
        } catch (XAMException ex) {
            xsetState = XSetState.CORRUPT;
            String msg = "Error persisting XSet. Cannot open or create file";
            logger.log( Level.SEVERE, msg, ex );
            throw new XSetCorruptException( msg, ex);
        }
        
        // sets all properties to unmodified
        this.setPropsUnmodified();

        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
        return xuidTmp;
    }
    
    /**
     * Called if this XSet was a result of import. Checks to see if an XSet with the same XUID
     * exists in the store. If it does, check each field to make sure that the binding field
     * attributes are the same. 
    * @throws XAMException 
    * @throws InvalidArgumentException 
    * @throws AuthorizationException 
    * @throws AuthenticationExpiredException 
    * @throws XSystemCorruptException 
    * @throws XSystemAbandonException 
    * @throws InvalidXSetModeException 
     */
    private void checkBindingChanges() throws InvalidXSetModeException, XSystemAbandonException, XSystemCorruptException, AuthenticationExpiredException, AuthorizationException, InvalidArgumentException, XAMException
    {
       if( xuid == null )
          return;
       XSet oldXSet = null;
       
       try {
          oldXSet = xsystem.openXSet(xuid, XSet.MODE_READ_ONLY);
       } catch( XSetInaccessibleException xie ) {
          return;
       }
       
       try {
          // Record all the old binding fields
          HashMap<String,Boolean> bindingMap = new HashMap<String,Boolean>();
          XIterator oldIter = oldXSet.openFieldIterator("");
          while( oldIter.hasNext() ) { 
             String fieldName = (String) oldIter.next();
             if( oldXSet.getFieldBinding(fieldName)) 
                bindingMap.put(fieldName, Boolean.TRUE);
          }
          oldIter.close();
       
          // Check against the current set.
          boolean bindingChange = false;
          XIterator newIter = openFieldIterator("");
          while( newIter.hasNext()) {
             String fieldName = (String) newIter.next();
             boolean newBinding = getFieldBinding(fieldName);
             Boolean mapBinding = bindingMap.get(fieldName);
             boolean oldBinding = (mapBinding == null?false:mapBinding.booleanValue());
             if( oldBinding != newBinding )
                bindingChange = true;          
          }
          newIter.close();
      
          // If there was a binding change, require a new XUID
          if (bindingChange && containsField(XAM_XUID)) {
             Field f = getField(XAM_XUID);
             removeField(f);
          }
       } finally {
          oldXSet.close();
       }
    }

   /**
     * Called when committing an XSet, make sure all policies are valid
     * @throws XAMException
     */
    private void validateAllPolicies() throws XAMException
    {
       final String PREFIX = ".xset.retention.";
       final String MSG = "Unknown retention policy: ";
       final String TYPE_MSG = "Ilegal field type for policy: " ;
       final String AUTO_MSG = "Unknown autodelete policy: ";
       final String SHRED_MSG = "Unknown shred policy: ";
       
       // Check actual retention fields, don't rely upon the retention list to be consistent,
       // so look for all retention fields ending in .policy.
       //
       // A nicer implementation would look for a policy which matches the one specified, 
       // but it is permissable to fail here. 
       XIterator iter = openFieldIterator( PREFIX );
       while( iter.hasNext() )
       {
          String policyName = null;
          String fieldName = (String) iter.next();
          if( ! fieldName.endsWith(".policy" ))
             continue;
          
          // Check the duration/enabled policy
          if( containsField( fieldName ))
          {
             Property policy = (Property) getField( fieldName );
             if( ! policy.getType().equals(XAMLibrary.STYPE_STRING_MIME_TYPE) ) {
                String msg = TYPE_MSG + fieldName;
                logger.warning(msg);
                throw new PolicyNameException( msg );
             }
             
             policyName = (String) policy.getValue();
             if( RetentionMgr.GetPolicy( policyName ) == null ) {
                String msg = MSG + policyName;
                logger.warning(msg);
                throw new PolicyNameException( msg );
             }
             
             // Verify that the retentionID is in the retention list.
             int lastLoc = fieldName.indexOf(".duration.policy");
             if( lastLoc < 0 )
                lastLoc = fieldName.indexOf(".enabled.policy");
             String retentionID = fieldName.substring( PREFIX.length(), lastLoc );
             if( ! containsField( XSet.XAM_RETENTION_LIST + retentionID ) ) {
                String msg = "Retention ID '" + retentionID + "' not in retention list";
                logger.warning(msg);
                throw new RetentionValueException( msg );
             }
          }     
       }
       
       // Check disposition policies
       if( containsField( XSet.XAM_DELETION_AUTODELETE_POLICY ))
       {
          String fieldName = XSet.XAM_DELETION_AUTODELETE_POLICY;
          Property policy = (Property) getField( fieldName );
          if( ! policy.getType().equals(XAMLibrary.STYPE_STRING_MIME_TYPE) ) {
             String msg = TYPE_MSG + fieldName;
             logger.warning(msg);
             throw new PolicyNameException( msg );
          }
          
          String policyName = (String) policy.getValue();
          if( DispositionPolicy.GetPolicy(policyName) == null ) {
             String msg = AUTO_MSG + policyName;
             logger.warning(msg);
             throw new PolicyNameException( msg ); 
          }
       }

       // Check the shred policies
       if( containsField( XSet.XAM_DELETION_SHRED_POLICY ))
       {
          String fieldName = XSet.XAM_DELETION_SHRED_POLICY;
          Property policy = (Property) getField( fieldName );
          if( ! policy.getType().equals(XAMLibrary.STYPE_STRING_MIME_TYPE) ) {
             String msg = TYPE_MSG + fieldName;
             logger.warning(msg);
             throw new PolicyNameException( msg );
          }
          
          String policyName = (String) policy.getValue();
          if( DispositionPolicy.GetPolicy(policyName) == null ) {
             String msg = SHRED_MSG + policyName;
             logger.warning(msg);
             throw new PolicyNameException( msg ); 
          }
       }
    }

   /**
     * @see XSet#close()
     */
    public void close() 
       throws AuthenticationExpiredException, ObjectInUseException,
              XSetCorruptException, XAMException {
        
       final String method = "close";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        checkAccessAllowed(XSetOperation.CLOSE);
        // Can not close if this instance is associated with a running job
        if (JobMgr.isJobSubmittedByXSetRunning(this)) {
           String msg = "Can not close XSet until stop running job";
           logger.warning(msg);
            throw new ObjectInUseException( msg );
        }
        // verify XStreams are closed
        checkXStreamsClosed();
        
        // Closing uncommited XSet, do cleanup
        if (isXSetDirty()) {
            cleanupUncommittedXStreamContentChanges();
        }
        xuid = null;
        xsetState = XSetState.CLOSED;
        openMode = XSet.MODE_READ_ONLY;
        clearFields();
        retentionMgr.clearRetentionList();
        xsystem.xsetClosed(this);
        
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }
    
    /**
     * Sets corrupt state
     * TODO:  double check that this is all that is needed when have XStreams, may need to corrupt open XSrteams
     */
    protected void setCorrupt() throws AuthenticationExpiredException, XAMException {
        xuid = null;
        xsetState = XSetState.CORRUPT;
    }

    /**
     * @see XSet#abandon()
     */
    public void abandon() throws AuthenticationExpiredException, XAMException {
       final String method = "abandon";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        checkAccessAllowed(XSetOperation.ABANDON);
        xuid = null;
        if (JobMgr.isJobSubmittedByXSetRunning(this)) {
            this.haltJob();
        }
        abandonStreams();
        xsetState = XSetState.ABANDONED;
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }

    // == Job Methods ==
    /**
     * @see XSet#submitJob()
     */
    public void submitJob() 
       throws AuthenticationExpiredException, AuthorizationException,
              JobCommandException, JobUnsupportedException, 
              XSetAbandonException, XSetCorruptException, XAMException {
        
        final String method = "submitJob";
        if( xsystem.m_enterExit )
           logger.entering( s_logClass, method );

        synchronized(jobLock) {
            if (JobMgr.isJobRunningForXSet(this)) {
                // only one job per XSet is allowed
               String msg = "Can't submit a job.  Current job is still active";
               logger.warning(msg);
                throw new XAMException(msg);
            }
            ReferenceQueryJob job = new ReferenceQueryJob(this);
            JobMgr.submit(job);
        }
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }
              
    /**
     * @see XSet#haltJob
     */
    public void haltJob() 
       throws AuthorizationException, AuthenticationException, 
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XAMException {
       final String method = "haltJob";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        JobMgr.cancelJobForXSet(this, true);
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }
    
    /**
     * @see XSet#openExportXStream() 
     */
    public XStream openExportXStream() 
       throws AuthenticationExpiredException, ObjectInUseException,
       XSetAbandonException, XSetCorruptException, XAMException {
       final String method = "openExportXStream";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        checkAccessAllowed(XSetOperation.OPEN_EXPORT_XSTREAM);
        // Opening a XStream instance in write mode automatically sets the
        // XSet state to dirty so this will catch XStreams open for write
        if (isXSetDirty()) {
            throw new XAMException("XSet must be clean with no pending " +
                    "changes before can open export stream");
        }
        
        // Can not export if job is currently running
        if (JobMgr.isJobSubmittedByXSetRunning(this)) {
           String msg = "Can not export XSet until stop running job";
           logger.warning(msg);
            throw new ObjectInUseException(msg);
        }

        if (!streamList.isEmpty()) {
           String msg = "Export of XSets with open XStreams fields is not supported";
           logger.warning(msg);
           throw new ObjectInUseException(msg);
        }
        
        // Set to exporting
        xsetState = XSetState.EXPORT;
        xsetExportStream = 
                new ReferenceExportXStream(xsystem, this);
        
        try { 
           return (XStream) Proxy.newProxyInstance(
              xsetExportStream.getClass().getClassLoader(),
              getAllInterfaces(xsetExportStream.getClass()),
              new XStreamClientInvocationHandler(xsetExportStream)); 
        } finally {
           if( xsystem.m_enterExit )
              logger.exiting( s_logClass, method );
        }
    }
    
    /**
     * @see XSet#openImportXStream() 
     */
    public XStream openImportXStream() 
       throws AuthenticationExpiredException, ObjectInUseException,
              XSetAbandonException, XSetCorruptException, XAMException {
       final String method = "openImportXStream";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        // Will also catch XStream instances open for write because opening 
        // a write mode XStream instance automatically sets the dirty bit
        if (isXSetDirty()) {
           String msg = "XSet must be clean with no pending changes before can open import stream";
           logger.warning(msg);
           throw new XAMException(msg);
        }
        checkAccessAllowed(XSetOperation.OPEN_IMPORT_XSTREAM);
        
        // If the XSet isn't new, do not proceed.
        if( containsField(XSet.XAM_XUID) || containsField(XSet.XAM_DIRTY) ) {
           String msg = "Import XStream may only be opened on a non dirty, no XUID XSet.";
           logger.warning(msg);
           throw new XSetException( msg );
        }
        
        // Can not export if job is currently running
        // Shouldn't need this check as XSet should be dirty if job is running
        if (JobMgr.isJobSubmittedByXSetRunning(this)) {
           String msg = "Can not import XSet until stop running job";
           logger.warning(msg);
            throw new ObjectInUseException(msg);
        }

        xsetImportStream = 
                new ReferenceImportXStream(xsystem, this);
        
        xsetState = XSetState.IMPORT;
        wasImported = true;
        try {
           return (XStream) Proxy.newProxyInstance(
              xsetImportStream.getClass().getClassLoader(),
              getAllInterfaces(xsetImportStream.getClass()),
              new XStreamClientInvocationHandler(xsetImportStream));
        } finally {
           if( xsystem.m_enterExit )
              logger.exiting( s_logClass, method );
        }
    }
    
    /**
     * @see XSet#asyncCommit(long, org.snia.xam.XAsyncListener) 
     */
    public XAsync asyncCommit(long           xopid,
                               XAsyncListener listener)
       throws AuthenticationExpiredException, InvalidArgumentException, 
              ObjectInUseException, XSetAbandonException, 
              XSetCorruptException, XAMException {
       final String method = "asyncCommit";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

       try { 
          return xsystem.getXAsyncManager().commitXSet( this, xopid, listener );
       } finally {
          if( xsystem.m_enterExit )
             logger.exiting( s_logClass, method );
       }
    }
    
    /**
     * @see XSet#asyncOpenXStream(
     *  java.lang.String, java.lang.String, long, org.snia.xam.XAsyncListener) 
     */
    public XAsync asyncOpenXStream( String         name,
                                    String         mode,
                                    long           xopid,
                                    XAsyncListener listener )
       throws AuthenticationExpiredException, InvalidArgumentException,
              InvalidFieldNameException, FieldDoesNotExistException,
              ObjectInUseException, XSetAbandonException,
              XSetCorruptException, XSystemAbandonException,
              XSystemCorruptException, XAMException {
       final String method = "asyncOpenXStream";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

       try { 
          return xsystem.getXAsyncManager().openXSstream(this, name, mode, xopid, listener);
       } finally {
          if( xsystem.m_enterExit )
             logger.exiting( s_logClass, method );
       }
    }
    
    
    /**
     * @see XSet#createXStream(java.lang.String, boolean, java.lang.String) 
     */
    public XStream createXStream(String fieldName,
            boolean binding,
            String mimeType)
            throws AuthenticationExpiredException, InvalidArgumentException,
                InvalidFieldNameException,
                FieldExistsException, MaximumFieldException, 
                ObjectInUseException, XSetAbandonException, 
                XSetCorruptException, XSystemAbandonException,
                XSystemCorruptException, XAMException {
        
       final String method = "createXStream";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        XSetOperation oper =(binding) ?
            XSetOperation.OPEN_XSTREAM_WRITE_BINDING :
            XSetOperation.OPEN_XSTREAM_WRITE_NONBINDING;
       
        XSetState.validateOperation(xsetState, oper);
        checkFieldExists(fieldName);
        // Create field        
        ReferenceStreamField field = new ReferenceStreamField(
                fieldName, mimeType, 0, binding, false, XAMConstants.EMPTY_STRING, logger );
        addField(field);
        
        ReferenceXStream stream = new ReferenceXStream(field, this, 
            XStream.MODE_WRITE_TRUNCATE);
        addStream(stream);
        setXSetDirty(fieldName, binding);

        try { 
           return (XStream) Proxy.newProxyInstance(
                stream.getClass().getClassLoader(),
                getAllInterfaces(stream.getClass()),
                new XStreamClientInvocationHandler(stream));   
        } finally {
           if( xsystem.m_enterExit )
              logger.exiting( s_logClass, method );
        }
    }


    /**
     * @see FieldContainer#openXStream(java.lang.String, java.lang.String) 
     * @param fieldName the XStream field name
     * @param mode open mode for XStream
     * @return the newly opened XStream instance
     * 
     * @throws org.snia.xam.AuthenticationExpiredException
     * @throws org.snia.xam.InvalidFieldNameException
     * @throws org.snia.xam.FieldDoesNotExistException
     * @throws org.snia.xam.ObjectInUseException
     * @throws org.snia.xam.XSetAbandonException
     * @throws org.snia.xam.XSetCorruptException
     * @throws org.snia.xam.XSystemAbandonException
     * @throws org.snia.xam.XSystemCorruptException
     * @throws org.snia.xam.XAMException
     */
    public XStream openXStream(String fieldName, String mode)
            throws AuthenticationExpiredException, InvalidFieldNameException,
                FieldDoesNotExistException, ObjectInUseException, 
                XSetAbandonException, XSetCorruptException, 
                XSystemAbandonException, XSystemCorruptException,
                XAMException {
        final String method = "openXStream";
        if( xsystem.m_enterExit )
           logger.entering( s_logClass, method );

        XSetOperation oper = XSetOperation.OPEN_XSTREAM_WRITE_BINDING;
        Field field =  getFieldNoChecking(fieldName);
        if (!(field instanceof ReferenceStreamField)) {
           String msg = "Not a stream field: " + fieldName;
           logger.warning(msg);
           throw new XAMException(msg);
        }
        
        boolean binding = false;
        boolean setModified = false;
        if (mode.equals(XStream.MODE_READ_ONLY)) {
            oper = XSetOperation.OPEN_XSTREAM_READONLY;
        } else {
            setModified = true;
            if (mode.equals(XStream.MODE_WRITE_APPEND) || 
                    mode.equals(XStream.MODE_WRITE_TRUNCATE)) {
                binding = field.getBinding();
               if (!binding) {
                   oper = XSetOperation.OPEN_XSTREAM_WRITE_NONBINDING;
               }
            } else {
               String msg = "Invalid open mode specified: " + mode;
               logger.warning(msg);
               throw new XAMException(msg);
            }
        }
        
        XSetState.validateOperation(xsetState, oper);
        // check if other XStreams are open against this stream field
        checkXStreamAccessAllowed(fieldName, mode);
        ReferenceStreamField rField = (ReferenceStreamField) field;
        ReferenceXStream stream = new ReferenceXStream(rField, this,
                mode);
        if (setModified) {
            rField.setModified(true);
            this.setXSetDirty(fieldName, binding);
        }
        addStream(stream);

        try { 
           return (XStream) Proxy.newProxyInstance(
                    stream.getClass().getClassLoader(),
                    getAllInterfaces(stream.getClass()),
                    new XStreamClientInvocationHandler(stream));   
        } finally {
           if( xsystem.m_enterExit )
              logger.exiting( s_logClass, method );
        }
    }
    
    // Helper methods
    
    protected Class[] getAllInterfaces(Class classInst) {

        Class[] allInterfacesList = null;
        Vector<Class> tmpList = new Vector<Class>();
        
        Class currentClass = classInst;
        while (currentClass != null &&  
                !currentClass.getSimpleName().contains("Object")) {
            Class[] itfList = currentClass.getInterfaces();
            if (itfList != null) {
                for (Class c : itfList) {
                    tmpList.add(c);
                }
            }
            currentClass = currentClass.getSuperclass();                       
        }
        allInterfacesList = new Class[tmpList.size()];
        return  tmpList.toArray(allInterfacesList);
    }

    protected Class[] getProxyInterfaces(Class classInst) {

        Class[] instInterfaceList = classInst.getInterfaces();
        int totalSize = instInterfaceList.length;
        Class intClass = classInst.getSuperclass();
        Class[] baseInterfaceList =null;
        if (intClass != null) {
            baseInterfaceList = intClass.getInterfaces();
            totalSize += baseInterfaceList.length;
        }        

        Class[] allInterfaceList = null;
        int startpos = 0;
        allInterfaceList =  new Class[totalSize];
        
        if (baseInterfaceList != null && baseInterfaceList.length > 0) {
            System.arraycopy(baseInterfaceList, 0, allInterfaceList, 0, baseInterfaceList.length);
            startpos += baseInterfaceList.length;
        }
        
        if (instInterfaceList != null && instInterfaceList.length > 0) {
            System.arraycopy(instInterfaceList, 0, allInterfaceList, 
                    baseInterfaceList.length, instInterfaceList.length);             
        }     
        return allInterfaceList;
        
    }
    /**
     * Walk through all the properties in this xset to determine if
     * any of the fields that are binding have changed. 
     * 
     * Note that this does not detect when a binding field has been changed
     * to non-binding which does qualify as a change to a binding field.
     * 
     * @return boolean true if the value of a binding field for this xset
     * has changed, false otherwise.
     */
    private boolean hasBindingFieldChanged() {
        XIterator iterator = fieldIterator(XAMConstants.EMPTY_STRING);
        while (iterator.hasNext()) {
            String fieldName = (String) iterator.next();
            try {
                Field field = getField(fieldName);
                if (field.isModified() && field.getBinding()) {
                     return true;
                }
            } catch (XAMException ex) {}
        }
        return false;
    }
    
    /**
     * Gets the string representation of the current XUID. If not xuid
     * exists (binding change made and XSet has not been committed yet) then
     * an empty string is returned.
     * @return String representation of the current xset xuid
     */
    protected String getXUIDString() {
        String xuidString = XAMConstants.EMPTY_STRING;
        if (xuid != null) {
            xuidString = xuid.toString();
        }
        return xuidString;
    }
    
    /**
     * Retrieves the current xuid instance
     * @return XUID of XSet; may be null
     */
    protected XUID getXUID() {
        return xuid;
    }

    /**
     * Returns the current XML File name (if one exists)
     * @return String the filename for this XUID
     */
    protected String getXMLFileName() {
        return xsystem.getPersistenceMgr().getXMLFileName(xuid);
    }

    protected void addStream(ReferenceXStreamBase stream) {
        streamList.add(stream);
    }
    
    protected boolean removeStream(ReferenceXStreamBase stream) {
        boolean removed = false;
        if (stream.isClosed()) {
            removed = streamList.remove(stream);
        } else {
            logger.log(Level.WARNING, "Could not remove XStream. " +
                    stream.filename);
        }
        return removed;
    }
    
    /**
     * Removes the reference to the export stream (if it's closed) and
     * also sets the XSet instance to the correct state.
     * 
     * @param stream the export stream
     * @return true if the stream was removed; false otherwise
     */
    protected boolean removeExportStream(ReferenceExportXStream stream) {
        boolean removed = false;
        if (stream.isClosed()) {
            this.xsetExportStream = null;
            try {
                this.xsetState = XSetState.getXSetState(openMode);
            } catch (InvalidArgumentException ex) {
                //ignore; Have already validated openMode previously
            }
        }
        return removed;
    }

    /**
     * Removes the reference to the import stream (if it's closed) and
     * also sets the XSet instance to the correct state.
     * 
     * @param stream the export stream
     * @return true if the stream was removed; false otherwise
    * @throws XAMException 
     */
    protected boolean removeImportStream(ReferenceImportXStream stream) throws XAMException {
        boolean removed = false;
        if (stream.isClosed()) {
            xsetImportStream = null;
            try {
               this.xsetState = XSetState.getXSetState(openMode);
           } catch (InvalidArgumentException ex) {
               //ignore; Have already validated openMode previously
           }
           
           try {
              retentionMgr.validateImportRetentionSettings();
           } catch( XAMException e ) {
              this.setCorrupt();
              throw e;
           }
        }
        return removed;
    }

    /*
     * Overridding AbstractFieldContainer methods to add access checking
     * and setting/clearing of xset.dirty
     */

    /**
     * Sets the specified field to a binding field. 
     * If a field has changed from binding to non-binding or non-binding to 
     * binding a new XUID will be generated on commit.
     * 
     * @param fieldName name of field to set to binding = true
     * @throws org.snia.xam.XAMException
     */
    public void setFieldAsBinding(String fieldName)
		throws XAMException
    {
       final String method = "setFieldAsBinding";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

       checkAccessAllowed(XSetOperation.CHANGE_FIELD_BINDING);
       synchronized(this) {
          DefaultField field = (DefaultField)getField(fieldName);
          boolean binding = field.getBinding();
          if (binding)
             return;
                
          try {
             super.setFieldAsBinding(fieldName);
          } catch (FieldReadOnlyException ex) {
             throw(ex);
          } catch (XAMException x) {
             xsetState = XSetState.CORRUPT;
             throw(x);
          }
            
          this.setXSetDirty(fieldName, true);
       }
       if( xsystem.m_enterExit )
          logger.exiting( s_logClass, method );
	}
	
    /**
     * Sets the specified field to a non-binding field. 
     * If a field has changed from binding to non-binding or non-binding to 
     * binding a new XUID will be generated on commit.
     * 
     * @param fieldName name of field to set to binding = true
     * @throws org.snia.xam.XAMException
     */
	public void setFieldAsNonbinding(String fieldName)
		throws XAMException
	{
	   final String method = "setFieldAsNonbinding";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

       checkAccessAllowed(XSetOperation.CHANGE_FIELD_BINDING);
       synchronized(this) {
          DefaultField field = (DefaultField)getField(fieldName);
          boolean binding = field.getBinding();
          if (!binding)
             return;
                
          try {
             super.setFieldAsNonbinding(fieldName);
          } catch (FieldReadOnlyException ex) {
             throw(ex);
          } catch (XAMException x) {
             xsetState = XSetState.CORRUPT;
             throw(x);
          }
          
          this.setXSetDirty(fieldName, true);
       }
       if( xsystem.m_enterExit )
          logger.exiting( s_logClass, method );
	}
        
    /**
     * Modifies the specified property to have the new value and type.
     * Exceptions will be thrown if the field does not exist, is not
     * a property or is readonly.
     *  
     * @param fieldName The property name
     * @param value The new value
     * @param type The new mime type
     * @throws XAMException If the field does not exist, isn't a property or 
     * is read only.
     */
    protected void setProperty(String fieldName, Object value, String type)
            throws XAMException
    {
       final String method = "setProperty";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        checkPropModify(fieldName);
        
        super.setProperty(fieldName, value, type);
        setXSetDirty(fieldName, false);

        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }
    
   /**
    * Modifies the specified property to have the new value and type. Exceptions
    * will be thrown if the field does not exist, is not a property or is
    * readonly.
    * 
    * @param fieldName The property name
    * @param value The new value
    * @throws XAMException If the field does not exist, isn't a property or is
    *            read only.
    */
   protected void setJobProperty(String fieldName, long value)
      throws XAMException
   {
      synchronized (this)
      {
         checkFieldNameWellFormed(fieldName);
         Property prop = getPropertyField(fieldName);
         //checkFieldReadOnly(prop);
         prop.setValue(new Long(value), XAMLibrary.STYPE_INT_MIME_TYPE);
         prop.setModified(true);
      }
   }

    /**
     * @see AbstractFieldContainer#getProperty(
     *  java.lang.String, java.lang.String) 
     * @param fieldName The property name
     * @param type The property mime type
     * @return The named property
     * @throws XAMException If the property does not exist, is a stream or 
     * does not match the requested mime type.
     */
    protected Property getProperty(String fieldName, String type)
            throws XAMException
    {
        checkAccessAllowed(XSetOperation.GET_PROPERTY);
        return super.getProperty(fieldName, type);
    }

    
    /**
     * This method is called by the public createProperty method specified by
     * FieldContainer interface. Added access checking before calling the
     * base implementation of createProperty. After property creation will
     * set the dirty property if needed.
     * 
     * Adds the specified {@link Property} to the FieldContainer after
     * checking to see a field with the same name exists and that the
     * property name is not in the protected name space.
     * 
     * @param prop The new property
     * @throws XAMException If a field with the same name exists or the
     * property name is in the protected name space. 
     */
    protected void createProperty(Property prop)
            throws XAMException
    {
       final String method = "createProperty";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        checkPropCreate(prop);
        super.createProperty(prop);
        // Unless loading an existing XSet, set dirty flag to indicate
        // uncomitted changes
        if (!loadingXSet) {
            boolean binding = prop.getBinding();
            setXSetDirty(prop.getName(), binding);
        }
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }

    /**
    * Removes the {@link Field} from the {@link FieldContainer}. Note, this
    * does not delete the field from the system. 
    * 
    * @param field
    */
    protected void removeField(Field field)
    {
        synchronized(this) {
            super.removeField(field);
        }
    }
    
    /**
     * Deletes the field.
     * @param fieldName
     * @throws org.snia.xam.XAMException
     */
    public void deleteField(String fieldName)
            throws XAMException
    {
       final String method = "deleteField";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        synchronized(this) {
            checkPropDelete(fieldName);
            Field field = this.getField(fieldName);
            if (field instanceof StreamField) {
               ReferenceStreamField rField = (ReferenceStreamField) field;
               try {
                  rField.delete();
                  this.deletedStreamFields.add(rField);
               } catch (XAMIOException ex) {
                  String msg = "Exception deleting a field: " + fieldName;
                  logger.log( Level.SEVERE, msg, ex );
                  throw new XAMException(msg, ex);
               }
            }
            boolean binding = field.getBinding();           
            setXSetDirty(fieldName, binding);
            super.deleteField(fieldName);
        }
        
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }
    
    /**
     * @see AbstractFieldContainer#getField(java.lang.String)
     */
    protected Field getField(String fieldName)
    throws XAMException {
        checkAccessAllowed(XSetOperation.GET_FIELD_INFO);
        return super.getField(fieldName);
    }

    /**
     * @see AbstractFieldContainer#getField(java.lang.String)
     */
    protected Field getFieldNoChecking(String fieldName)
    throws XAMException {
        return super.getField(fieldName);
    }

    /**
     * Adds a field to the list
     * @param field
     */
    protected void addField(Field field) {
        super.addField(field);
    }
    /**
     * Adds a field to the list
     * @param field
     */
    protected void addFieldSetDirty(Field field) {
        addField(field);    
        if (!loadingXSet) {
            boolean binding = field.getBinding();
            try {
                setXSetDirty(field.getName(), binding);
            } catch (XAMException ex) {
                // ignore -- not throwing
            }
        }
    }
  
    /*
     * Access check methods
     */
    protected void checkPropModify(String fieldName) throws XAMException {   
        
        Field field = null;
        boolean binding = true;
        try {
            field = getField(fieldName);
            binding = field.getBinding();
        } catch (Exception FieldDoesNotExistException) {
            /* Several reasons for this; One is that the XSystem/XSet
             * has been close. The exception will be reported correctly
             * later so ignore for now.
             */
        }
        XSetOperation op = binding ? 
            XSetOperation.SET_PROPERTY_BINDING :
            XSetOperation.SET_PROPERTY_NONBINDING;
        
        checkAccessAllowed(op);
    }

    protected void checkPropCreate(Property prop) throws XAMException {        
        XSetOperation op = prop.getBinding() ? 
            XSetOperation.CREATE_PROPERTY_BINDING :
            XSetOperation.CREATE_PROPERTY_NONBINDING;
        checkAccessAllowed(op);
    }
    
    protected void checkPropDelete(String fieldName) throws XAMException {        
        Field field = this.getField(fieldName);
        XSetOperation op = field.getBinding() ? 
            XSetOperation.DELETE_FIELD_BINDING :
            XSetOperation.DELETE_FIELD_NONBINDING;
        
        checkAccessAllowed(op);
    }
    
    /*
     * Determines if XSet is dirty or not
     */
    private boolean isXSetDirty() {
        boolean dirty = false;
        try {
            dirty = containsField(XAM_DIRTY);
        } catch (XAMException ex) { /* ignore */ }
        return dirty;
    }
    /*
     * A field change, binding change or field addition/deletion requires 
     * setting the dirty property. In addition changes to binding fields or
     * changing a field from binding to non-binding or non-binding to binding
     * requires that the XSet be placed into the NO XUID state immediately.
     * A new XUID will be generated on commit.
     * 
     * @param fieldName name of the field that's changing
     * @param bindingChange fields new binding property value
     * @throws org.snia.xam.XAMException
     */
    private void setXSetDirty(String fieldName, boolean bindingChange)
            throws XAMException {
        
         addField(new Property(
                XAM_DIRTY,
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false,
                true  ));

        /* Also considered a binding change if field was changed from
         * binding to non-binding.
         */
        if (!bindingChange) {
            Field field = this.getField(fieldName);
            bindingChange = field.getBinding();
        }
         
        /* If a binding field changed or field changed binding type then
         * have to get rid of current xuid. New one will be generated on
         * commit
         */
        synchronized (this) {
            if (!loadingXSet && bindingChange) {
                if (containsField(XAM_XUID)) {
                    Field f = getField(XAM_XUID);
                    removeField(f);
                    this.xuid = null;
                }
                if (containsField(XAM_TIME_XUID)) {
                    Field f = getField(XAM_TIME_XUID);
                    removeField(f);
                }
            }
        }
    }

    /**
     * Determines whether xset is corrupt
     * @return boolean  true if xset is corrupt; false otherwise
     */
    protected boolean isCorrupt() {
        return (xsetState.equals(XSetState.CORRUPT) ?
            true : false);
    }

    /**
     * Determines whether xset is closed
     * @return boolean  true if xset is closed; false otherwise
     */
    protected boolean isClosed() {
        return (xsetState.equals(XSetState.CLOSED) ?
            true : false);
    }

    /**
     * Determines whether the xset has been abandoned
     * @return boolean  true if the xset has been abandoned; false otherwise
     */
    protected boolean isAbandoned() {
        return (xsetState.equals(XSetState.ABANDONED) ?
            true : false);
    }


    /**
     * Wrapper routine to perform all checks needed before allowing
     * access to a method/operation.
     * 
     * @param operation the operation or method to check access
     * 
     * @throws org.snia.xam.AuthenticationExpiredException
     * @throws org.snia.xam.XSetAbandonException
     * @throws org.snia.xam.XSetCorruptException
     * @throws org.snia.xam.XAMException
     */
    private void checkAccessAllowed(XSetOperation operation)
            throws AuthenticationExpiredException, XSetAbandonException,
            XSetCorruptException, XAMException {

        XSetState.validateOperation(xsetState, operation);

    }
    /**
     * Method used to retrieve the application/client (i.e. "other") specified
     * retention criteria identifiers (i.e. <retention id>).
     * 
     * @return Set "Other" retention criteria identifiers in the XSet.
     * @throws org.snia.xam.XAMException
     */
    protected Set<String> getOtherRetentionCriteria() throws XAMException {
        return retentionMgr.getRetentionList();
    }
    /**
     * Method used to check if the persisted XSet contains any 
     * application/client specified (i.e. "other) retention criteria identifiers 
     * 
     * @return boolean Flag indicates whether or not the current XSet instance
     *                 contains retention criteria identifiers other than
     *                 the expected "base" and "event" retention value
     *                 management properties.  Returns true if it
     *                 has retention ID's other than "base" and "event",
     *                 otherwise, false.
     * @throws org.snia.xam.XAMException
     */
    protected boolean hasOtherRetentionCriteria() throws XAMException {
        if (retentionMgr.getRetentionList().isEmpty()) {
            return false;
        }
        return true;
    }
    /**
     * Routine to perform checks for XSet management properties
     * These fields must be present in the XSet at the time it is
     * committed.
     * @param xuidTime the XUID creation time, .xset.time.xuid  
     * @throws org.snia.xam.XAMException
     */
    private void setManagementFields(Calendar xuidTime) throws XAMException {
        if (!containsField(XAM_MANAGEMENT_POLICY)) {
            // get the system default since XSet's management policy wasn't set          
            addFieldSetDirty(new Property(
                    XAM_MANAGEMENT_POLICY,
                    XAMLibrary.STYPE_STRING_MIME_TYPE,
                    ReferenceXSystem.REF_VIM_DEFAULT_POLICY,
                    true, 
                    true));    
        }
        if (!containsField(XSET_HOLD)) {          
            addFieldSetDirty(new Property(
                    XSET_HOLD,
                    XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                    Boolean.FALSE,
                    false, 
                    true));    
        }
        retentionMgr.setRetentionFieldsAtCommit(xuidTime);
    }
    
    /**
     * Checks that all XStreams are closed. This routine is double checking
     * things as if the stream is closed it should not be in the
     * streamList list..
     * 
     * @throws org.snia.xam.XStreamException
     * @throws org.snia.xam.XStreamCorruptException
     * @throws org.snia.xam.XStreamAbandonException
     * @throws org.snia.xam.XAMException
     */
    private void checkXStreamsClosed() 
            throws XStreamException, XStreamCorruptException,
            XStreamAbandonException, XAMException {

        while(streamList.size() > 0) {
            ReferenceXStreamBase stream = streamList.get(0);
            if (stream.isAbandoned()) {
                stream.close();
            }
            if (!stream.isClosed()) {
               String msg = "All XStreams must be closed";
               logger.warning(msg);
               throw new XStreamException(msg);
            }
        }

        return;
          
    }
    
    /**
     * Verifies that no other XStream instance is open for this stream
     * field by this XSet. The exception is the internal xstream opened to
     * to write the job results.
     * 
     * @param streamFieldName
     * @param mode
     * @throws org.snia.xam.InvalidArgumentException
     * @throws org.snia.xam.XAMException
     */
    private void checkXStreamAccessAllowed(String streamFieldName,
                                            String mode)
            throws InvalidArgumentException, XAMException {

        boolean writeAccess = 
                (mode.equals(XStream.MODE_READ_ONLY) ? false : true);
        for (ReferenceXStreamBase baseXStream : this.streamList) {
            if (baseXStream instanceof ReferenceXStream) {
                ReferenceXStream xstream = (ReferenceXStream) baseXStream;
                String fieldName = xstream.getFieldName();
                if (fieldName.equals(streamFieldName)) {
                    // Write access - must be only open XStream instance
                    if (writeAccess) {
                       String msg = "Can not open XStream for writing. " +
                             "A XStream instance is already open on the " +
                             "stream field. ";
                       logger.warning(msg);
                       throw new InvalidArgumentException(msg);
                    }
                    
                    if (xstream.getOpenMode().equals(XStream.MODE_READ_ONLY)) {
                        break; // don't need to check any more because all
                               // will be open for reading
                    } else {
                        // allow reader if this is a running job where the
                        // query results are poplulated by the VIM
                        if (!fieldName.equals(XSet.XAM_JOB_QUERY_RESULTS) ||
                            !JobMgr.isJobRunningForXSet(this)) {
                           String msg = "Can not open XStream for reading. " +
                                  "A XStream instance is already open in on the" +
                                  "the stream field for writing.";
                           logger.warning(msg);
                           throw new InvalidArgumentException(msg);
                        }
                    }
                }
            }
        }

    }

    /**
     * Abandons all open streams: XStreams and ImportXStreams and ExportXStreams
     * 
     * Only abandon streams if they are not already abandoned or closed.
     * 
     * FYI: If there is an open export stream there will not be any other
     *      open streams (including import). Likewise for an open import stream.
     * 
     * @throws org.snia.xam.XStreamException
     * @throws org.snia.xam.XStreamCorruptException
     * @throws org.snia.xam.XStreamAbandonException
     * @throws org.snia.xam.XAMException
     */
    private void abandonStreams() 
            throws XStreamException, XStreamCorruptException,
            XStreamAbandonException, XAMException {

        // Abandon any open the XStreams; these are the streams based
        // on stream fields. 
        for (int i = 0; i < streamList.size(); i++) {
            ReferenceXStreamBase stream = streamList.get(i);
            if (!stream.isAbandoned() && !stream.isClosed()) {
                stream.abandon();
            }
        }
        
        // Abandon open export stream - if exists
        // If open export stream there will not be any other open streams
        if (xsetExportStream != null && !xsetExportStream.isAbandoned() &&
                !xsetExportStream.isClosed()) {
            xsetExportStream.abandon();
        }

        // Abandon open import stream - if exists
        // If open import stream there will not be any other open streams
        if (xsetImportStream != null && !xsetImportStream.isAbandoned() &&
                !xsetImportStream.isClosed()) {
            xsetImportStream.abandon();
        }
        return;
          
    }
    
    /**
     * If the field already exists on the XSet, this method removes the field
     * and then recreates it with the new values.
     * @param fieldName the name of the Field to set
     * @param binding flag indicating whether or not the Field should be 
     *                binding (true) or non-binding (false)
     * @param readOnly flag indicating whether or not the Field should be
     *                 readonly (true) or read/write (false)
     * @param value the field value -- valid Objects are:
     *                Boolean, String, Long, Double, XUID and Calendar
     * @throws XAMException
     */
    protected void setJobField(String fieldName, boolean binding, 
                                        boolean readOnly, Object value) 
                                        throws XAMException{
        /* Need to be able to add job fields to the XSet even if they are 
         * in the reserved namespace.  Check to see if the field name starts
         * with ".xam.job." and if it does, then add the field.  The addField
         * method takes care of creating the field if it doesn't exist or
         * overwriting the field and its values if it does exist.  If the job
         * field is not a part of the reserved namespace, then its deleted (if
         * exists) and then created.
         */
        try {
            checkProtectedNameSpace(fieldName);
        } catch (XAMException xe) {
            if (fieldName.startsWith(XAMConstants.XAM_JOB)) {
                addField(new Property(
                    fieldName,
                    XAMLibrary.STYPE_STRING_MIME_TYPE,
                    value,
                    binding,
                    readOnly));
                return;
            } else {
                throw xe;
            }
        }
        if (containsField(fieldName)) {
            Field f = getField(fieldName);
            if (f.getReadOnly()) {
                // can't call deleteField on readonly field
                checkFieldNameWellFormed(fieldName);
                removeField(f);
            } else {
                deleteField(f.getName());
            }
        }
        if (value instanceof Boolean) {
            createProperty(fieldName, binding, ((Boolean)value).booleanValue());
        } else if (value instanceof String) {
            createProperty(fieldName, binding, (String)value);
        } else if (value instanceof Long) {
            createProperty(fieldName, binding, ((Long)value).longValue());
        } else if (value instanceof Double) {
            createProperty(fieldName, binding, ((Double)value).doubleValue());
        } else if (value instanceof XUID) {
            createProperty(fieldName, binding, ((XUID)value));
        } else if (value instanceof Calendar) {
            createProperty(fieldName, binding, ((Calendar)value));
        } else {
            XAMException ex = new XAMException("Unsupported field value type");
            logger.log( Level.SEVERE, ex.getMessage(), ex );
            throw ex;
        }
        setReadOnlyField(fieldName, readOnly);
    }
    /**
     * Sets the read-only field attribute to true for the specified field
     * @param fieldName the name of the Field to set
     * @throws InvalidFieldNameException, FieldDoesNotExistException,
     *          XAMException
     */
    protected void setReadOnlyField(String fieldName, boolean readOnly) 
            throws InvalidFieldNameException, FieldDoesNotExistException, 
                                                                XAMException {
        checkFieldNameWellFormed(fieldName);
        if (!containsField(fieldName)) {
           String msg = fieldName + " does not exist.";
           logger.warning(msg);
           throw new FieldDoesNotExistException(msg);
        }
        DefaultField field = (DefaultField) getField(fieldName);
        field.setReadonly(readOnly);
        field.setModified(true);
    }

    private void commitXStreamContentChanges(String dirname) 
            throws XStreamException, XStreamCorruptException, XAMException {

        XIterator iterator = this.openFieldIterator("");
        ReferenceStreamField rField = null;
        while (iterator.hasNext()) {
            String fieldName = (String)iterator.next();
            
            Field field = this.getField(fieldName);
            if (field instanceof ReferenceStreamField) {
                rField = (ReferenceStreamField) field;
                rField.commit(dirname);
            }
        }   // end while
        return;
          
    }
    private void deleteStreamFields(String dirname) 
            throws XStreamException, XStreamCorruptException, XAMException {

        ListIterator<ReferenceStreamField> iterator =
                deletedStreamFields.listIterator();
        while (iterator.hasNext()) {
            ReferenceStreamField rField = iterator.next();
            rField.commit(dirname);
        }
        deletedStreamFields.clear();
        return;
          
    }
    
    /*
     * Called only when closing an XSet withoug committing the changes. 
     * Will cleanup any tmp files used for stream fields.
     */
    private void cleanupUncommittedXStreamContentChanges() 
            throws XStreamException, XStreamCorruptException, XAMException {

        XIterator iterator = this.openFieldIterator("");
        ReferenceStreamField rField = null;
        while (iterator.hasNext()) {
            String fieldName = (String)iterator.next();
            
            Field field = this.getFieldNoChecking(fieldName);
            if (field instanceof ReferenceStreamField) {
                rField = (ReferenceStreamField) field;
                XStreamContents contents = rField.getContents();
                contents.cleanupUncommitedChanges();
            }
        }   // end while
        return;
          
    } 

    /**
     * Method used to check if the persisted XSet's .xset.time.commit has
     * changed relative to the current XSet instance.
     * 
     * @return boolean Flag indicates whether or not the current XSet instance
     *                 has a commit time that differs from the persisted XSet.
     *                 Returns true if it has changed, otherwise false.
     * @throws org.snia.xam.XAMException
     */
    protected boolean hasCommitTimeChanged() throws XAMException {

        // current commit time of persisted xset -- need to retrieve new 
        // instance to compare latest commit time value
        Date pcTime = null; 
        // commit time of this set
        Date cTime = null;
        
        if (xuid == null) {
            throw new XAMException("XUID is null!");
        }
        ReferenceXSet xset = null;
        try {
            xset = new ReferenceXSet(this.xsystem, 
                    this.xuid, XSet.MODE_READ_ONLY);        
            Field persistedField = xset.getField(XSet.XAM_TIME_COMMIT);
            Calendar pcCalendar = (Calendar) ((Property) persistedField).getValue();
            pcTime = new Date(pcCalendar.getTimeInMillis());
            // commit time of this particular set when it was opened
            Field field = this.getField(XSet.XAM_TIME_COMMIT);
            cTime = new Date(((Calendar) ((Property) field).
                                                getValue()).getTimeInMillis());
            if (pcTime.after(cTime)) {
                return true;
            } else {
                // times are equal
                return false;
            }
        } catch (Exception e) {
            throw new XAMException(e);
        } finally {
            xset.close();
        } 
    }

    private void setPropsUnmodified() throws XAMException {
        // Set all props to not modified
        XIterator fields = this.openFieldIterator("");
        while (fields.hasNext()) {
            String fieldName = (String)fields.next();
            try {
                Field field = this.getField(fieldName);
                if (field instanceof ReferenceStreamField) {
                    ReferenceStreamField rField = (ReferenceStreamField) field;
                    rField.setModified(false);
                } else {
                    Property prop = getPropertyField(fieldName);
                    prop.setModified(false);
                }
            } catch (FieldDoesNotExistException ex) {
                // skip
            } catch (InvalidFieldTypeException fe) {
                //skip
            } catch (XAMException xex) {
                this.xsetState = XSetState.CORRUPT;
                throw(xex);
            }
        }
        
    }

    /**
     * Parses, validates and loads the imported XSet definition. This method
     * is called only when the client application has called close on the
     * imported XStream instance.
     * 
     * @param xsetDefinition String containing the imported XSet XML that
     * needs to be validated and loaded into the system.
     * 
     * @throws org.snia.xam.XAMException
     */
    protected void loadImportedXSet(String xsetDefinition, 
            XStreamXopData[] xstreamsData) throws XAMException {
        
        if (!xsetState.equals(XSetState.IMPORT)) {
           String msg = "Must be importing an XSet to load an XSet definition: " + xsetState.getName();
           logger.warning(msg);
           throw new XAMException(msg);
        }
        
        // Set state so can set field values when loading XSet definition
        XSetState saveXSetState = xsetState;
        xsetState = XSetState.UNRESTRICTED_WRITER;
        loadingXSet = true;
        XSetXmlReader handler = new XSetXmlReader(this, xstreamsData);
        xuid = null;
        Calendar timestamp = Calendar.getInstance();
        StringReader reader = null;
        InputSource is = null;
        try {
            reader = new StringReader(xsetDefinition);
            is = new InputSource(reader);
            SAXParser parser = XSetXmlReader.factory.newSAXParser();
            parser.parse(is, handler);
            
            // The parser can not add the system fields because it is
            // does not have the right access privs. So load them now.
            processSystemFields(xuid, handler.getSystemFields());
     
            // set all properties isModified to false
            setPropsUnmodified();
            
            // The access time has changed --
            // add the new access time back to the XSet instance
            Field f = getField(XAM_TIME_ACCESS);
            removeField(f);
            addField(new Property(
                    XAM_TIME_ACCESS,
                    XAMLibrary.STYPE_DATETIME_MIME_TYPE,
                    timestamp,
                    false, 
                    true));  
            
            // Reset mode after xsetState = saveOpenState;load all properties
             XSetState.getXSetState(openMode);
        } catch (XSetCorruptException cex) {
            xsetState = XSetState.CORRUPT;
            throw cex;
        } catch (AuthenticationExpiredException aex) {
            xsetState = saveXSetState;
            throw aex;
        } catch (XAMException e) {
            xsetState = XSetState.CORRUPT;
           XSetCorruptException xex = new XSetCorruptException(
                    "Error parsing imported XSet", e);
            throw(xex);
        } catch (Exception ex) {
            XSetCorruptException cex = new XSetCorruptException(
                    "Error parsing/validating imported XSet.", ex);
            xsetState = XSetState.CORRUPT;
            throw cex;
        } finally {
            loadingXSet = false;
            if (reader != null) {
                reader.close();
                reader = null;
            }
            is = null;

        }        
    }
    
    /**
     * Usually called by the XML reader and sets the policy references. Used during the import 
     * process via validateRetention().
     * @param ref The retention policy references.
     */
    protected void setRetentionPolicyReferences( HashMap<String,RetentionPolicyReference> ref )
    {
       m_retention_policy_references = ref;
    }
    
    /**
     * Get the policy references. This map was produced by the import process.
     * @return The policy HashMap.
     */
    protected HashMap<String,RetentionPolicyReference> getImportRetentionPolicyReferences()
    {
       return m_retention_policy_references;
    }
    
    /**
     * Clear the policy references, usually after the policy references have been validated.
     */
    protected void clearRetentionPolicyReferences()
    {
       m_retention_policy_references = null;
    }
    
    /**
     * Convenience method to determine if XSet contains job command field
     * @return true if this XSet represents a job; that is has job command 
     *         specified.
     */
    protected boolean isJob() {
        boolean jobXSet = false;
        // check if this XSet represents a job (only query jobs supported)
        try {
            if (containsField(XSet.XAM_JOB_COMMAND)) {
                jobXSet = true;
            }
        } catch (XAMException ex) { /* ignore */ }

        return jobXSet;
    }
   
    protected boolean sameXUID(XUID compareXUID) {
        boolean sameXUIDValue = false;
        if (compareXUID != null && xuid != null) {
            sameXUIDValue = xuid.equals(compareXUID);
        }
        return sameXUIDValue;
    }
    
    protected boolean sameXUID(ReferenceXSet compareXSet) {
        boolean sameXUIDValue = false;
        XUID compareXUID = compareXSet.xuid;
        if (compareXUID != null && xuid != null) {
            sameXUIDValue = xuid.equals(compareXUID);
        }
        return sameXUIDValue;
    }
}
