/*
 * Copyright (c) 2008, Sun Microsystems, Inc.
 * Copyright (c) 2008, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.logging.Level;

import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.FieldContainer;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.XAMException;
import org.snia.xam.XSet;
import org.snia.xam.XSetUnderHoldException;
import org.snia.xam.XStream;
import org.snia.xam.XStreamAbandonException;
import org.snia.xam.XStreamCorruptException;
import org.snia.xam.XStreamException;

/**
 * This class implements the final level of functionality for XStreams in the Reference VIM.
 */
public class ReferenceXStream extends ReferenceXStreamBase {
    // Field properties
    ReferenceStreamField field = null;
    protected boolean resultsField = false;
    
    /**
     * Creating an XStream object for existing XStream (do not open)
     * @param field
     * @param fieldContainer
     */
    private ReferenceXStream(ReferenceStreamField field, 
            FieldContainer fieldContainer) throws XAMException {
        super(fieldContainer);
        this.field = field;
        this.fieldContainer = fieldContainer;
        if (fieldContainer instanceof ReferenceXSet) {
            this.parentXSet = (ReferenceXSet) fieldContainer;
            if (field.getName().equals(XSet.XAM_JOB_QUERY_RESULTS)) {
                resultsField = true;
            }
        } else {
           String msg = "Invalid type of XSet class. " + fieldContainer.getClass().getName() + 
              "ReferenceXSet instance required.";
           logger.severe(msg);
           throw new XAMException(msg);
        }
        
        field.addStream(this);

    }

    protected ReferenceXStream(ReferenceStreamField field,
            FieldContainer fieldContainer, String openMode)
            throws XAMException {
        this(field, fieldContainer);
        this.openMode = openMode;
        setState(openMode);
        try {
            this.checkFieldAccess(openMode, field);
        } catch (XAMException ex) {
            this.state = XStreamState.CLOSED;
            field.closeStream(this);
            throw ex;
        }
        openFileChannel();
        
    }
 
    
    public String getFieldName() {
        return field.getName();
    }
    
    protected void setState(String openMode) {
        if (openMode.equals(XStream.MODE_READ_ONLY)) {
            this.state = XStreamState.READER;
        } else {
            this.state = XStreamState.WRITER;
        }        
    }
    
    public void abandon() throws AuthenticationExpiredException, XAMException {
       final String method = "abandon";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        // If not read only stream then must abandon or set to corrupt other
        // streams open on the same field
        if (!openMode.equals(XStream.MODE_READ_ONLY)) {
            field.abandonStreams(this);
        }

        state = XStreamState.ABANDONED;
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }
    
    protected void setStateCorrupt() {
        if (openMode.equals(XStream.MODE_READ_ONLY)) {
            state = XStreamState.READ_CORRUPT;
        } else {
            state = XStreamState.WRITE_CORRUPT;
        }
    }
    
    protected void setStateAbandoned() {
        state  = XStreamState.ABANDONED;
    }
    
    /**
     * Opens a file channel for the stream. Need to make sure the file being 
     * used is the current location of the data contents of the stream field.
     * @return true if the channel was opened successfully
     * @throws org.snia.xam.XAMException
     */
    protected boolean openFileChannel() throws XAMException {
        String checkFilename = "";
        
        try {
            // Get current file name. Might be source file or tmp/working file
            checkFilename = field.getWorkingFilename(openMode);
            if (channel != null && channel.isOpen()) {
                return false;
            }
            // open or reopen channel if file has changed.
            if ((checkFilename.compareToIgnoreCase(filename) != 0) ||
                    filename.length() == 0) {
                filename = checkFilename;
                file = new File(filename);

                if (openMode.equals(XStream.MODE_READ_ONLY)) {
                    input = new FileInputStream(filename);
                    channel = input.getChannel();
                    state = XStreamState.READER;
                } else if (openMode.equals(XStream.MODE_WRITE_APPEND)) {
                    output = new 
                            FileOutputStream(filename, true);
                    channel = output.getChannel();
                    state = XStreamState.WRITER;
                } else if (openMode.equals(XStream.MODE_WRITE_TRUNCATE)) {
                    output = new FileOutputStream(filename);
                    channel = output.getChannel();
                    state = XStreamState.WRITER;
                } else {
                    throw new InvalidArgumentException("Unsupported mode of " + 
                            openMode + " specified.");
                }
            }
        }
        catch (IOException io) {
           logger.log( Level.SEVERE, "Exception opening stream", io );
           throw new XStreamException(io);
        }
        
        // Initialize stream and field lengths to be the same
        streamLength = file.length();        
        field.setLength(streamLength);
        return true;
    }

    /**
     * 
     * @see ReferenceXStreamBase#closeStream() 
     * 
     * @throws org.snia.xam.AuthenticationExpiredException
     * @throws org.snia.xam.XSetUnderHoldException
     * @throws org.snia.xam.XStreamCorruptException
     * @throws org.snia.xam.XStreamException
     * @throws org.snia.xam.XAMException
     */
    protected void closeStream() throws AuthenticationExpiredException, XSetUnderHoldException, XStreamCorruptException, XStreamException, XAMException {
        field.closeStream(this);
    }

    /**
     * @see ReferenceXStreamBase#checkMethodAccess(String)
     * 
     * @param methodName
     * @throws org.snia.xam.XStreamException
     */
    public void checkMethodAccess(String methodName) throws XStreamException {
        // Check read/write access allowed
        if (methodName.equals("write") ||
                methodName.indexOf("Write") != -1) {
                if (! (state == XStreamState.WRITER)) {
                   String msg = "Write access is not allowed on XStream.";
                   logger.warning(msg);
                    throw new XStreamException(msg);
                }
        } else if (methodName.equals("read") ||
            methodName.equals("seek") ||
            methodName.indexOf("Read") != -1) {
            if (! (state == XStreamState.READER)) {
               String msg = "Read access is not allowed on XStream.";
                throw new XStreamException( msg );
            }
        }

    }
    /**
     * Checks open type is allowed based on field access mode
     * 
     * @param openMode the XStream open mode
     * @param field the stream field object
     * @throws org.snia.xam.XStreamException
     */
    public void checkFieldAccess(String openMode, ReferenceStreamField field) 
            throws XStreamException {
        
        // Check write access allowed
        if (!openMode.equals(XStream.MODE_READ_ONLY) && field.getReadOnly()) {
           String msg = "Can not open XStream for write access on read only field";
           logger.warning(msg);
           throw new XStreamException(msg);
        }

    }
    
    /**
     * @see ReferenceXStreamBase#write(byte[], long, long)
     * Need to overrride so can update the field length after the write
     * 
     * @param buffer
     * @param offset
     * @param bytesToWrite
     * @return the number of bytes written
     * @throws org.snia.xam.AuthenticationExpiredException
     * @throws org.snia.xam.InvalidArgumentException
     * @throws org.snia.xam.XStreamAbandonException
     * @throws org.snia.xam.XStreamCorruptException
     * @throws org.snia.xam.XStreamException
     * @throws org.snia.xam.XSetUnderHoldException
     * @throws org.snia.xam.XAMException
     */
    public long write(byte[] buffer, long offset, long bytesToWrite) 
            throws AuthenticationExpiredException, InvalidArgumentException, 
            XStreamAbandonException, XStreamCorruptException, XStreamException, 
            XSetUnderHoldException, XAMException {
        final String method = "write";
        if( xsystem.m_enterExit )
           logger.entering( s_logClass, method );

        long numBytes = super.write(buffer,offset, bytesToWrite);
        field.setLength(getLength());
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
        return numBytes;       
    }
    
    /**
     * @see ReferenceXStreamBase#read(byte[], long, long)
     * Need to overrride so can check if file location has changed and reopen
     * channel.
     */
    public long read(byte[] buffer, long offset, long bytesToRead) 
            throws AuthenticationExpiredException, InvalidArgumentException, 
                XStreamAbandonException, XStreamCorruptException, XAMException {
        
       final String method = "read";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        String checkFilename = field.getWorkingFilename(openMode);
        if (checkFilename.compareToIgnoreCase(filename) != 0 && 
            filename.length() != 0) {
            try {
                // close existing channel and reopen
                long savePosition = channel.position();
                closeChannelAndStreams();
                filename = "";
                openFileChannel();
                channel.position(savePosition);
             } catch (IOException iox) {
                String msg = "Error reading from XStream. Could not locate stream contents.";
                logger.log(Level.SEVERE, msg, iox);
                throw new XAMException(msg, iox);
            }
        }
        
        long bytesRead = super.read(buffer, offset, bytesToRead);
        // Only the XSet instance that submitted the job can see the results
        // while the job is running.
        if (resultsField && bytesRead == XAMConstants.END_OF_STREAM &&
                JobMgr.isJobSubmittedByXSetRunning(parentXSet)) {
            bytesRead = 0;
        }
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
        return bytesRead;
    }
    
}
