/*
 * Copyright (c) 2008, Sun Microsystems, Inc.
 * Copyright (c) 2008, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.FieldContainer;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XSetUnderHoldException;
import org.snia.xam.XStream;
import org.snia.xam.XStreamAbandonException;
import org.snia.xam.XStreamCorruptException;
import org.snia.xam.XStreamException;

/**
 *
 * Base Class for Reference Implementation of XStreams 
 */
abstract public class ReferenceXStreamBase implements XStream {

    protected ReferenceXSystem xsystem   = null;
    protected Logger          logger     = null;
    protected String          s_logClass = ReferenceXStreamBase.class.getName();

    protected XStreamState state = null;
    
    // Not needed protected boolean modified = false;
    protected String openMode = null; 
    
    protected File tmpCopy = null;
    // The field container would be the parent XSet
    protected ReferenceXSet parentXSet = null;
    protected FieldContainer fieldContainer = null;
    //protected XStreamContents data = null;
    
    
    protected long position = 0;
    FileChannel channel = null;
    FileInputStream input = null;
    FileOutputStream output = null;
    File file = null;
    String filename = "";
    protected long streamLength = 0;
    
    /**
     * Creating an XStream object for existing XStream (do not open)
     * @param fieldContainer the parent field container for this stream field
     */
    protected ReferenceXStreamBase( 
            FieldContainer fieldContainer) throws XAMException {
        this.fieldContainer = fieldContainer;
        if (fieldContainer instanceof ReferenceXSet) {
            this.parentXSet = (ReferenceXSet) fieldContainer;
            xsystem = this.parentXSet.getXSystem();
         } else if(fieldContainer instanceof ReferenceXSystem) {
          xsystem = ((ReferenceXSystem) fieldContainer);
        } else {
           String msg = "Invalid type of XSet class. " + fieldContainer.getClass().getName() +  
              "ReferenceXSet instance required.";
           logger.severe( msg );
           throw new XAMException(msg);
        }
        logger = xsystem.getLogger();

    }

    protected ReferenceXStreamBase(
            FieldContainer fieldContainer, String openMode)
            throws XAMException {
        this.openMode = openMode;
    }
    
    /**
     * @see XStream#tell() 
     * @return the current position in the file
     * @throws org.snia.xam.AuthenticationExpiredException
     * @throws org.snia.xam.XStreamAbandonException
     * @throws org.snia.xam.XStreamCorruptException
     * @throws org.snia.xam.XAMException
     */
    public long tell() throws AuthenticationExpiredException, 
            XStreamAbandonException, XStreamCorruptException, XAMException {
       final String method = "tell";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        try {
            return channel.position();
        } catch (IOException iox) {
           String msg = "Could not get current stream position";
           logger.log( Level.SEVERE, msg, iox );
            throw new XAMException( msg, iox);
        } finally {
           if( xsystem.m_enterExit )
              logger.exiting( s_logClass, method );
        }
    }

    public void seek(long offset, long whence) 
            throws AuthenticationExpiredException, InvalidArgumentException, 
            XStreamAbandonException, XStreamCorruptException, XStreamException,
            XAMException {
       final String method = "seek";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        long position = 0; //initialize to start of file
        try {
            if (whence == XStream.SEEK_CUR) {
                position = channel.position();
            } else if (whence == XStream.SEEK_END) {
                position = channel.size();
            } // otherwise from start
            
            if (position + offset > channel.size()) {
               String msg = "Seek position is outside the boundaries of the XStream.";
               logger.warning(msg);
               throw new XStreamException(msg);
            }
            position += offset;
            channel.position(position);
        } catch (IOException iox) {
           String msg = "Could not get current stream position";
           logger.severe(msg);
           throw new XAMException(msg, iox);
        } finally {
           if( xsystem.m_enterExit )
              logger.exiting( s_logClass, method );
        }
    }

    /**
	 * @see java.io.OutputStream#write(byte[], int, int)
	 */
     public long write(byte[] buffer) throws AuthenticationExpiredException, InvalidArgumentException, XStreamAbandonException, XStreamCorruptException, XStreamException, XSetUnderHoldException, XAMException {
            return write(buffer, 0, buffer.length);
	}
        
    public long write(byte[] buffer, long bytesToWrite) throws AuthenticationExpiredException, InvalidArgumentException, XStreamAbandonException, XStreamCorruptException, XStreamException, XSetUnderHoldException, XAMException {
        return write(buffer, 0, bytesToWrite);
    }

    public long write(byte[] buffer, long offset, long bytesToWrite) throws AuthenticationExpiredException, InvalidArgumentException, XStreamAbandonException, XStreamCorruptException, XStreamException, XSetUnderHoldException, XAMException {
        long numBytes = -1;
        int writeOffset = 0;
        int writeSize = 0;
        
        if (offset > Integer.MAX_VALUE || bytesToWrite > Integer.MAX_VALUE) {
           String msg = "Specify smaller buffer to write";
           logger.warning("Specify smaller buffer to write");
           throw new InvalidArgumentException(msg);
        }
        writeOffset = (int) offset;
        writeSize = (int) bytesToWrite;
        
            try {
                ByteBuffer bb = ByteBuffer.wrap(buffer, writeOffset, writeSize);
                numBytes = channel.write(bb);
                streamLength = channel.size();
                position = channel.position();
            } catch (IOException ex) {
                state = XStreamState.WRITE_CORRUPT;
                parentXSet.setCorrupt();
                XAMException ioex = new XAMException(
                        "Error writing to stream", ex);
                throw ioex;
            }
            return numBytes;
    }

    public long read(byte[] buffer, long offset, long bytesToRead) 
            throws AuthenticationExpiredException, InvalidArgumentException, 
                XStreamAbandonException, XStreamCorruptException, XAMException {
        long numBytes = -1;
        int readOffset = 0;
        int readSize = 0;
        if (offset > Integer.MAX_VALUE || bytesToRead > Integer.MAX_VALUE) {
           String msg = "Specify smaller buffer to read";
           logger.warning(msg);
           throw new InvalidArgumentException(msg);
        }
        readOffset = (int) offset;
        readSize = (int) bytesToRead;
        try {
            ByteBuffer bb =ByteBuffer.wrap(buffer, readOffset, readSize);
            numBytes = channel.read(bb);
            position = channel.position();
        } catch (IOException ex) {
            this.state = XStreamState.READ_CORRUPT;
            XAMException ioex = new XAMException(
                    "Error writing to stream", ex);
            throw ioex;
        }
        return numBytes;
    }

    public long read(byte[] buffer, long bytesToRead) 
            throws AuthenticationExpiredException, InvalidArgumentException, 
            XStreamAbandonException, XStreamCorruptException, XAMException {
        return read(buffer, 0, bytesToRead);
    }

    public long read(byte[] buffer) 
       throws AuthenticationExpiredException, InvalidArgumentException, XStreamAbandonException, XStreamCorruptException, XStreamException, XAMException {
        
        return read(buffer, 0, buffer.length);
    }

    public void close() 
       throws AuthenticationExpiredException, XSetUnderHoldException, 
              XStreamCorruptException, XStreamException, XAMException {
        final String method = "close";
        if( xsystem.m_enterExit )
           logger.entering( s_logClass, method );

        try {
             // Close all underlying file io streams and the channel
            // if not already closed
            closeChannelAndStreams();
            
            state = XStreamState.CLOSED;
            closeStream();
        } catch (XAMException ex) {
            logger.log(Level.WARNING, "Errors encountered when closing XStream instance.", ex);
        }
        

        if (!parentXSet.removeStream(this)) {
            logger.log(Level.WARNING, "Could not remove the XStream instance from the list of streams." +
                    " This is an expected result for import/export streams." + this.filename);
        }
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
    }

    abstract protected void closeStream() throws AuthenticationExpiredException, XSetUnderHoldException, XStreamCorruptException, XStreamException, XAMException;
    
    public void abandon() throws AuthenticationExpiredException, XAMException {
        state = XStreamState.ABANDONED;
    }

    public XAsync asyncRead(byte[] buffer, long bufferLength, long xopid, XAsyncListener listener) 
       throws AuthenticationExpiredException, InvalidArgumentException, 
              XStreamAbandonException, XStreamCorruptException, 
              XStreamException, XAMException {
       final String method = "asyncRead";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

        XAsync async = parentXSet.
           getXSystem().
              getXAsyncManager().
                 readStream( this, buffer, bufferLength, xopid, listener, logger );
        if( xsystem.m_enterExit )
           logger.exiting( s_logClass, method );
        return async;
    }

    public XAsync asyncWrite(byte[] buffer, long bufferLength, long xopid, XAsyncListener listener) 
       throws AuthenticationExpiredException, InvalidArgumentException, 
              XStreamAbandonException, XStreamCorruptException, 
              XStreamException, XAMException {
       final String method = "asyncWrite";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );

       XAsync async = parentXSet.
       getXSystem().
          getXAsyncManager().
             writeStream( this, buffer, bufferLength, xopid, listener, logger );

       if( xsystem.m_enterExit )
          logger.exiting( s_logClass, method );
       return async;
    }

    public XAsync asyncClose(long xopid, XAsyncListener listener) 
       throws AuthenticationExpiredException, InvalidArgumentException, 
          XStreamCorruptException, XStreamException, XAMException {
       final String method = "asyncClose";
       if( xsystem.m_enterExit )
          logger.entering( s_logClass, method );
       
       XAsync async = parentXSet.
       getXSystem().
          getXAsyncManager().
             closeStream( this, xopid, listener, logger );
       if( xsystem.m_enterExit )
          logger.exiting( s_logClass, method );
       return async;
    }
    
    public boolean isClosed() {
        return state == XStreamState.CLOSED;
    }
    
    abstract protected void setStateCorrupt();

    public boolean isAbandoned() {
        return state == XStreamState.ABANDONED;
    }
    
    protected void setStateAbandoned() {
        state = XStreamState.ABANDONED;
    }
    
    public boolean isCorrupt() {
        return state.isCorrupt();
    }
    
    public boolean isExport() {
        return state.isExport();
    }
  
    public boolean isImport() {
        return state.isImport();
    }

    public String getOpenMode() {
        return openMode;
    }    
    /**
     * Determines whether the call to the method is allowed.
     * 
     * @param methodName the name of the method that will be called
     * @throws org.snia.xam.XStreamException if access to method is not
     * allowed
     */
    abstract public void checkMethodAccess(String methodName) 
            throws XStreamException;
        
    protected ReferenceXSet getParentXSet() {
        return this.parentXSet;
    }

    protected XStreamState getState() {
        return state;
    }
    
    protected long getLength() {
        return streamLength;
    }
    
    /**
     * Closes the channel and any existing file streams
     * Note: Should only have to close the channel but close the input/output
     *       streams just in case.
     */
    protected void closeChannelAndStreams() {
        if (channel != null) {
            try {
                channel.close();
            } catch (IOException iox) {
                logger.log(Level.WARNING, "Errors encountered when closing file channel for XStream.", iox);
             }
            channel = null;
        }

        if (input != null) {
            try {
            input.close();
            } catch (IOException iox) {
                logger.log(Level.WARNING, "Errors encountered when closing file input stream for XStream.", iox);
            }
            input = null;
        }
        
        if (output != null) {
            try {
            output.close();
            } catch (IOException iox) {
               logger.log(Level.WARNING, "Errors encountered when closing file ouput stream for XStream.", iox);
             }
            output = null;
        }
        
    }
}
