/*
 *  Copyright (c) 2008-2009, Sun Microsystems, Inc.
 *  Copyright (c) 2008-2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.io.ByteArrayOutputStream;
import java.lang.reflect.Proxy;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Calendar;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;
import java.util.logging.FileHandler;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.snia.xam.AuthenticationException;
import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.AuthorizationException;
import org.snia.xam.FieldContainer;
import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.FieldExistsException;
import org.snia.xam.HoldIdException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidFieldNameException;
import org.snia.xam.InvalidOperationException;
import org.snia.xam.InvalidXSetModeException;
import org.snia.xam.InvalidXUIDException;
import org.snia.xam.MaximumFieldException;
import org.snia.xam.ObjectInUseException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XIterator;
import org.snia.xam.XSet;
import org.snia.xam.XSetAbandonException;
import org.snia.xam.XSetCorruptException;
import org.snia.xam.XSetInaccessibleException;
import org.snia.xam.XSetUnderHoldException;
import org.snia.xam.XSetUnderRetentionException;
import org.snia.xam.XStream;
import org.snia.xam.XSystem;
import org.snia.xam.XSystemAbandonException;
import org.snia.xam.XSystemCorruptException;
import org.snia.xam.XSystemException;
import org.snia.xam.XUID;
import org.snia.xam.base.AbstractFieldContainer;
import org.snia.xam.base.Field;
import org.snia.xam.base.Property;
import org.snia.xam.base.XAMImplementation;
import org.snia.xam.base.XRI;
import org.snia.xam.vim.reference.JobMgr;
import org.snia.xam.vim.reference.utils.ReferenceSaslUtils;
import org.snia.xam.util.LogManager;

/**
 * Reference implementation of XSystem. 
 * Basic implementation for single user. Plain and Simple authentication
 * supported to username and password. If the Java property "org.snia.xam.reference.passwd"
 * is defined, the ReferenceAuthenticationStatus will read username/password entries from 
 * the file referred to by the property. See ReferenceAuthenticationStatus for more info.
 * 
 * Todo:  the async methods are not implemented at this time.
 * 
 * Data will be persisted in the tmp directory defined by java.io.tmpdir
 * unless a location is specified via the xri parameter on connect.
 * 
 * @see XSystem
 */
public class ReferenceXSystem
        extends AbstractFieldContainer
        implements XSystem {
    
    public static final String REF_VIM_DEFAULT_POLICY = 
                                "org.snia.refvim.default.mgmt.policy";
    public static final String REF_VIM_DEFAULT_POLICY_ENABLED = 
                                "org.snia.refvim.default.mgmt.policy.enabled";
    public static final String REF_VIM_DEFAULT_POLICY_DURATION = 
                                "org.snia.refvim.default.mgmt.policy.duration";
    public static final String REF_VIM_ASYNC_THREAD_COUNT = 
                                "org.snia.refvim.async.thread.count";

    private static final String ERROR_XUID_IS_NULL = "XUID Value is null";;
    
    // Authorization granules, would be nice if these were in the interfaces.
    private static final String GRANULE_READ              = "read";
    private static final String GRANULE_WRITE_APPLICATION = "write-application";
    private static final Object GRANULE_WRITE_SYSTEM      = "write-system";
    private static final Object GRANULE_CREATE            = "create";
    private static final Object GRANULE_DELETE            = "delete";
    private static final Object GRANULE_JOB               = "job";
    private static final Object GRANULE_JOB_COMMIT        = "job-commit";
    private static final Object GRANULE_HOLD              = "hold";
    private static final Object GRANULE_RETENTION_EVENT   = "retention-event";
    
    // Log defaults, as specified
    private static final long    DEFAULT_LOG_SIZE  = 1024 * 1024 * 1024; // 1 GB as required by specification
    private static final long    DEFAULT_LOG_COUNT = 1;
    private static final String  DEFAULT_LOG_PATH  = "SNIA-XAM.log";
    private static final long    DEFAULT_LOG_VERBOSITY = 1;
    private static final long    DEFAULT_LOG_LEVEL = XAMLibrary.XAM_LOG_INFO;
    
    // Logging parameters
    private Logger      logger          = null;
    private FileHandler m_logHandler    = null;
            boolean     m_enterExit     = false; // Verbosity > 1
    private static String s_logClass = "org.snia.xam.reference.ReferenceXSystem";

    /**
     * Manages the location of the repository where all the data associated 
     * with this XSystem are stored on the file system
     */
    private ReferenceVIMPersistenceMgr persistenceMgr = null;
    
    /** Authentication State **/
    final static public String PROPERTY_PASSWD = "org.snia.xam.reference.passwd";

    final public static String PROPERTY_AUTODELETE_PERIOD = "org.snia.xam.reference.autodelete.period";

    private ReferenceAuthenticationStatus authStatus = new ReferenceAuthenticationStatus();    

    /* List of XSets open for this XSystem */
    private Vector<ReferenceXSet> xsetList = new Vector<ReferenceXSet>();

    /* Manage and maintain access policies */
    private AccessPolicyManager m_access_policy_mgr;
    
    /** The Database where all the property information is stored */
    private DBManager m_database_manager;
    
    /** Disposition manager, supports auto delete daemon */
    @SuppressWarnings("unused")
    private DispositionManager m_disposition_manager;
    
    /** The async manager executes the operations. */
    private ReferenceXAsyncManager m_async_manager;
    
    /**
     * Creates a new ReferenceXSystem instance.
     * 
     * @throws org.snia.xam.XAMException
     */
    protected ReferenceXSystem() throws XAMException {
        loadDefaults();
    }

    // Public methods part of XSystem interface
    
    /**
     * @see XSystem#asyncOpenXStream(java.lang.String, java.lang.String, long, 
     *  org.snia.xam.XAsyncListener) 
     */
    public XAsync asyncOpenXStream(String name,
            String mode,
            long xopid,
            XAsyncListener listener)
            throws AuthenticationExpiredException, InvalidArgumentException,
            InvalidFieldNameException, FieldDoesNotExistException,
            ObjectInUseException, XSetAbandonException, XSetCorruptException,
            XSystemAbandonException, XSystemCorruptException, XAMException {
        final String method = "asyncOpenXStream";
        if( m_enterExit )
           logger.entering( s_logClass, method );
        
        try {
           throw new InvalidOperationException();
        } finally {
           if( m_enterExit )
              logger.entering( s_logClass, method );
        }
    }
    
    /**
     * @see XSystem#createXStream(java.lang.String, boolean, java.lang.String) 
     */
    public XStream createXStream(String fieldName,
            boolean binding,
            String mimeType)
            throws AuthenticationExpiredException, InvalidArgumentException,
                InvalidFieldNameException, InvalidOperationException,
                FieldExistsException, MaximumFieldException, 
                ObjectInUseException, XSetAbandonException, 
                XSetCorruptException, XSystemAbandonException,
                XSystemCorruptException, XAMException {
        
        throw new InvalidOperationException();
    }

    /**
     * @see XSystem#openXStream(java.lang.String, java.lang.String) 
     */
    public XStream openXStream(String fieldName, String mode)
            throws AuthenticationExpiredException, InvalidFieldNameException,
                FieldDoesNotExistException, ObjectInUseException, 
                XSetAbandonException, XSetCorruptException, 
                XSystemAbandonException, XSystemCorruptException,
                XAMException {
        
       final String method = "openXStream";
       if( m_enterExit )
          logger.entering( s_logClass, method );
       
       try {
          throw new InvalidOperationException();
       } finally {
          if( m_enterExit )
             logger.exiting( s_logClass, method ); 
       }
    }

    /**
     * Note: This method should complete even if the system is not
     * authenticated.
     * 
     * @see XSystem#abandon() 
     */
    public void abandon()
            throws AuthenticationExpiredException, XAMException {
        
       final String method = "abandon";
       
       if( m_enterExit )
          logger.entering(s_logClass, method );
        synchronized(this) {
            if (authStatus.isClosed()) {
               String msg = "XSystem is closed";
               logger.warning( msg );
               throw new XSystemException(msg);
            }
            if( m_async_manager != null )
               m_async_manager.stopAsyncManager();
            abandonXSets();
            authStatus.setAbandoned();
        }
        if( m_enterExit )
           logger.exiting(s_logClass, method );
    }
    
    /**
     * @see XSystem#createXUID( String inBase64XUID )
     */
    public XUID createXUID( String inBase64XUID ) 
	throws AuthenticationExpiredException, InvalidXUIDException, XAMException
	{
       final String method = "createXUID(base64)";
       if( m_enterExit )
          logger.entering( s_logClass, method );
      
       try {
          return SniaXUID.getInstance(inBase64XUID);
       } finally {      
          if( m_enterExit )
             logger.entering( s_logClass, method );
       }
	}

    /**
     * @see XSystem#createXUID(byte[]) 
     */
    public XUID createXUID( byte inBinaryXUID[] ) 
	throws AuthenticationExpiredException, InvalidXUIDException, XAMException
	{
       final String method = "createXUID(byte[])";
       if( m_enterExit )
          logger.entering( s_logClass, method );
       
       try {
          return SniaXUID.getInstance(inBinaryXUID);
       } finally {
          if( m_enterExit )
             logger.exiting( s_logClass, method );
       }
	}
    
    /**
     * @see XSystem#accessXSet(org.snia.xam.XUID, long) 
     * This method will not allow any checks if the XSet is not available (either
     * it does not exist on the XSystem, or it is not readable by the app).
     */
    public boolean accessXSet(XUID xsetid, long mode)
            throws AuthenticationExpiredException, InvalidArgumentException, 
                XSystemAbandonException, XSystemCorruptException, 
                XAMException {
       
       final String method = "accessXSet";
       if( m_enterExit )
          logger.entering(s_logClass, method );

       if (xsetid == null) {
          throw new InvalidArgumentException(ERROR_XUID_IS_NULL);
       }

       // Note: Use the special non-AccessTime modifying constructor
       ReferenceXSet xset = new ReferenceXSet( this, xsetid );
       try {
          String accessPolicy = null;
          String user = this.getAuthenticatedUsername();
          
          try { accessPolicy = xset.getString( XSet.XAM_ACCESS_POLICY ); }
          catch( FieldDoesNotExistException fdne ) { /* ignore */ }
          
          if( ! m_access_policy_mgr.allowReadAccess( accessPolicy, user ))
             throw new XSetInaccessibleException();
          
          boolean writeAllowed = m_access_policy_mgr.allowWriteAccess(accessPolicy, user );
         
          // This doesn't make much sense on a particular XSet, but still ...
          if( mode ==  XSystem.ACCESS_CREATE_OK )
             return getBoolean(XAM_XSYSTEM_AUTH_GRANULE_LIST + GRANULE_CREATE);
          
          else if( mode == XSystem.ACCESS_WRITE_APPLICATION_OK)
             return writeAllowed && getBoolean(XAM_XSYSTEM_AUTH_GRANULE_LIST + GRANULE_WRITE_APPLICATION );
          
          else if( mode == XSystem.ACCESS_WRITE_SYSTEM_OK) 
             return getBoolean(XAM_XSYSTEM_AUTH_GRANULE_LIST + GRANULE_WRITE_SYSTEM );
          
          else if( mode == XSystem.ACCESS_DELETE_OK)
                return writeAllowed && getBoolean(XAM_XSYSTEM_AUTH_GRANULE_LIST + GRANULE_DELETE);
          
          else if( (mode == XSystem.ACCESS_READ_OK ))
             return getBoolean(XAM_XSYSTEM_AUTH_GRANULE_LIST + GRANULE_READ );
          
          else if( mode == XSystem.ACCESS_HOLD_OK) 
             return writeAllowed && getBoolean(XAM_XSYSTEM_AUTH_GRANULE_LIST + GRANULE_HOLD );
          
          else if( mode == XSystem.ACCESS_RETENTION_EVENT_OK)
             return writeAllowed && getBoolean(XAM_XSYSTEM_AUTH_GRANULE_LIST + GRANULE_RETENTION_EVENT);
          
          else if( mode == XSystem.ACCESS_JOB_OK )
             return writeAllowed && getBoolean(XAM_XSYSTEM_AUTH_GRANULE_LIST + GRANULE_JOB);
          
          else if( mode == XSystem.ACCESS_JOB_COMMIT_OK)
             return writeAllowed && getBoolean(XAM_XSYSTEM_AUTH_GRANULE_LIST + GRANULE_JOB_COMMIT);
             
          else {
             String msg = "Unknown mode : 0x" + Long.toHexString(mode);
             logger.warning(msg);
             throw new InvalidArgumentException( msg );
          }
             
       }
       finally {
          if( xset != null )
             xset.close();
          if( m_enterExit )
             logger.exiting(s_logClass, method );
       }
    }

    /**
     * @see XSystem#asyncCopyXSet(org.snia.xam.XUID, java.lang.String, long, 
     *  org.snia.xam.XAsyncListener) 
     */
    public XAsync asyncCopyXSet(XUID inXuid, String mode, long xopid, 
            XAsyncListener listener)
            throws AuthenticationExpiredException, AuthorizationException, 
                InvalidArgumentException, XSetInaccessibleException, 
                XSystemAbandonException, XSystemCorruptException, 
                XAMException {
      
       final String method = "asyncCopyXSet";
       if( m_enterExit )
          logger.entering(s_logClass, method );

       try {
          return m_async_manager.copyXSet( this,
             inXuid,
             mode,
             xopid,
             listener );
       } finally {
          if( m_enterExit )
             logger.exiting(s_logClass, method );
       }
       
    }

    /**
     * @see XSystem#asyncOpenXSet(org.snia.xam.XUID, java.lang.String, long, 
     *  org.snia.xam.XAsyncListener) 
     */
    public XAsync asyncOpenXSet(XUID inXuid, String mode, long xopid, 
            XAsyncListener listener)
            throws AuthorizationException, AuthenticationExpiredException, 
                InvalidArgumentException, XSetInaccessibleException, 
                XSystemAbandonException, XSystemCorruptException, 
                XAMException {
       
       final String method = "asyncOpenXSet";
       if( m_enterExit )
          logger.entering(s_logClass, method );

       try {
          return m_async_manager.openXSet( this,
                                           inXuid,
                                           mode,
                                           xopid,
                                           listener );
       } finally {
          if( m_enterExit )
             logger.exiting(s_logClass, method );
       }
    }

    /**
     * Authentication for the TestXSystem. This XSystem will support ANONYMOUS and PLAIN.
     * For Plain to work out, the application must supply the username "test". The password
     * is ignored.
     */
    public int authenticate(byte[] buffer, ByteArrayOutputStream response)
      throws AuthenticationException, InvalidArgumentException, XAMException
    {
       final String method = "authenticate";
       if( m_enterExit )
          logger.entering(s_logClass, method );
       
       // Loading authentication information is delayed until it is time to authenticate
       // This allows the XAMLibrary a chance to copy fields over from initializing,
       // including the all important "org.snia.xam.reference.passwd.
       loadAuthentication();

       int returnValue = XSystem.XAM_SASL_COMPLETE;
        synchronized (this) {
            returnValue = authStatus.authenticate(buffer, response);
            boolean authenticated = authStatus.isAuthenticated();
            
            // Based on the authentication adjust the options allowed for the
            // XSystem
            if (authenticated) {
                if (authStatus.isAuthAnonymous()) {
                    setAnonymousPrivileges();
                } else if (authStatus.isAuthPlain()) {
                    setFullPrivileges();
                }
                m_access_policy_mgr = new AccessPolicyManager( this, authStatus );
            }
            
            if( m_enterExit )
               logger.exiting(s_logClass, method );
            return returnValue;
        }
    }

    /**
     * Authentication for the ReferenceXSystem. This XSystem will support 
     * ANONYMOUS and PLAIN.  
     * For Anonymous the application will be required to send trace information
     * which can be an empty.
     * For Plain, the application must supply the username "testuser" and
     * the password "testpasswd".
     * @deprecated XSystem#authenticate(byte[], ByteArrayOutputStream)
     */
    public byte[] authenticate(byte[] buffer)
            throws AuthenticationException, InvalidArgumentException, 
                XAMException {
       final String method = "authenticate";
       if( m_enterExit )
          logger.entering(s_logClass, method );

        ByteArrayOutputStream respByteStream = new ByteArrayOutputStream();
        synchronized (this) {
            authenticate(buffer, respByteStream);
            byte[] respBytes = respByteStream.toByteArray();
 
            if( m_enterExit )
               logger.exiting(s_logClass, method );
            return respBytes;
        }
    }
    
    /**
     * @see XSystem#close()
      */
    public void close()
            throws AuthenticationExpiredException, ObjectInUseException, 
                XAMException {

       final String method = "close";
       if( m_enterExit )
          logger.entering(s_logClass, method );

        synchronized (this) {
            authStatus.checkCloseAllowed();
            // If XSets are closed also know that there are no running jobs
            verifyJobsCompleted();
            verifyXSetsClosed();
            clearXSetList();
            authStatus.setClosed();
            if( persistenceMgr != null )
               persistenceMgr.resetRepositoryLocation();
            
            if( m_async_manager != null )
               m_async_manager.stopAsyncManager();

            if( logger != null )
               logger.log(Level.INFO, "Closing the Java XAM Library" );
            
            if( m_logHandler != null )
            {
               m_logHandler.close();
               logger.removeHandler(m_logHandler);
            }
            
            if( m_enterExit )
               logger.exiting(s_logClass, method );

            m_logHandler = null;
            logger = null;
        }
    }
    
    /**
     * @see XSystem#connect(java.lang.String) 
     */
    public void connect(String xriString)
            throws XAMException {
       final String method = "connect";
       if( m_enterExit )
          logger.entering(s_logClass, method );
     
        /*
         * The only thing we need is the location of the repository
         * we are working with on the local host
         * if none are specified set to java.io.tmpdir 
         * (if not explicitly defined usually points to /var/tmp on Unix
         * and c:\temp on windows
         */
        XRI xri = XRI.valueOf(xriString);
        String dirParam = 
                xri.getParameterValue(XAMConstants.XRI_DIR_PARAM_NAME);
        try {
            persistenceMgr = 
                        new ReferenceVIMPersistenceMgr(this, dirParam );
        } catch (Exception ex) {
           String msg = "Unable to connect to persistence manager.";
           logger.log( Level.SEVERE, msg, ex );
        }
        
        try {
           m_database_manager = DBManager.GetInstance(this);
           Property p = getPropertyField( XAM_XSYSTEM_JOB_QUERY_LEVEL2_SUPPORTED );
           if( m_database_manager.supportsLevel2() )
              p.setValue( Boolean.TRUE, XAMLibrary.STYPE_BOOLEAN_MIME_TYPE );
        }
        catch( Exception e ) { 
           String msg = "Unable to instantiate the Reference VIM Database manager";
           logger.log( Level.SEVERE, msg, e );
           throw new XSystemException( msg, e );
        }
        
        try {
           m_disposition_manager = DispositionManager.GetInstance(this);
        } catch( Exception e ) {
           String msg = "Unabled to load Reference VIM Disposition Manager";
           logger.log( Level.SEVERE, msg, e );
           throw new XSystemException( msg, e );
        }
        authStatus.setUnAuthenticated();
        
        if( m_enterExit )
           logger.exiting(s_logClass, method );
    }

    /**
     * Perform post initialization operations.
     * @throws XAMException 
     */
    private void completeInitialization() throws XAMException
    {
       if( logger == null )
          updateLogger(false);
          
       // Create the asynchronous manager
       int threadCount = 2;
       if( containsField(REF_VIM_ASYNC_THREAD_COUNT))
          threadCount = (int) getLong(REF_VIM_ASYNC_THREAD_COUNT);
       if( m_async_manager != null )
          m_async_manager.stopAsyncManager();
       m_async_manager = new ReferenceXAsyncManager(this, threadCount);
    }

    /**
     * Updates the XAM Library logger to match the values set in the library
     * properties.
     * @param force Caused the logger to be updated regardless of the initialization state.
     */
    protected void updateLogger(boolean force)
       throws XAMException
    {
       boolean initializing = false;
       try { initializing = getBoolean( XSystem.XAM_XSYSTEM_INITIALIZING ); }
       catch( FieldDoesNotExistException fdne ) { };
       
       if( initializing && ! force )
          return;
       
       // Read the log configuration from the property files.
       String path   = null;
       long   lsize  = DEFAULT_LOG_SIZE;
       long   lcount = DEFAULT_LOG_COUNT;
       long   level  = XAMLibrary.XAM_LOG_FATAL;
      
       path = getString(XAMLibrary.XAM_LOG_PATH); 
       level = getLong(XAMLibrary.XAM_LOG_LEVEL);
       long verbosity = getLong(XAMLibrary.XAM_LOG_VERBOSITY);
       m_enterExit  = verbosity >= 200;
       
       try 
       { 
          lsize = getLong( XAMLibrary.XAM_LOG_MAX_SIZE ); 
          if( (lsize > Integer.MAX_VALUE) || (lsize < 0) ) {
             String msg = "Log file size is too large, or less than zero";
             logger.warning(msg);
             throw new InvalidArgumentException( msg );
          }
       }
       catch( FieldDoesNotExistException x ) { }
       
       try 
       { 
          lcount = getLong( XAMLibrary.XAM_LOG_MAX_ROLLOVERS ); 
          if( (lcount > Integer.MAX_VALUE) || (lcount < 0) ) {
             String msg = "Log count is too large, or less than zero";
             logger.warning(msg);
             throw new InvalidArgumentException( msg );
          }
       }
       catch(FieldDoesNotExistException x) { }
       
       boolean append = getBoolean( XAMImplementation.XAM_LOG_APPEND );
       logger = LogManager.GetLogger(path,(int) lsize, (int) lcount, append);
       logger.setLevel(LogManager.mapLogLevelToJavaLevel(level));
       
       super.setLogger(logger);
       super.setVerbosity(m_enterExit);
    }

    /**
     * @return The configured logger for this XSystem instance.
     * @throws XAMException
     */
    Logger getLogger()
       throws XAMException
    {
       if( logger == null )
          updateLogger(false);
       
       return logger;
    }
    
     /**
     * Override setting property so we can intercept logging properties and react accordingly.
     */
    protected synchronized void setProperty(String fieldName,
                                            Object value,
                                            String type)
       throws XAMException
    {
       final String method = "setProperty";
       if( m_enterExit )
          logger.entering( s_logClass, method );
       
       super.setProperty(fieldName, value, type);
       if (XAMLibrary.XAM_LOG_LEVEL.equals(fieldName))
       {
          updateLogger(false);
       }
       else if (XAMLibrary.XAM_LOG_VERBOSITY.equals(fieldName))
       {
          updateLogger(false);
       }
       else if (XAMLibrary.XAM_LOG_PATH.equals(fieldName))
       {
          updateLogger(false);
       }

       if( m_enterExit )
          logger.exiting( s_logClass, method );
    }
    
    /**
     * Override this method so the ReferenceXSystem can perform post initialization functions.
     * @see FieldContainer#deleteField(java.lang.String)
     * @throws XAMException 
     */
    public synchronized void deleteField( String fieldName ) throws XAMException
    {
       final String method = "deleteField";
       if( m_enterExit )
          logger.entering( s_logClass, method );
       
       super.deleteField(fieldName);
       if( fieldName.equals(XSystem.XAM_XSYSTEM_INITIALIZING) )
          completeInitialization();

       if( m_enterExit )
          logger.exiting( s_logClass, method );
    }

   /**
     * @see XSystem#copyXSet(org.snia.xam.XUID, java.lang.String) 
     */
    public XSet copyXSet(XUID xsetid, String copyMode)
            throws AuthenticationExpiredException, AuthorizationException, 
            InvalidArgumentException, InvalidXSetModeException, 
            XSetInaccessibleException, XSystemAbandonException, 
            XSystemCorruptException, XAMException {
       final String method = "copyXSet";
       if( m_enterExit )
          logger.entering( s_logClass, method );
       
        authStatus.checkAccessAllowed();
        
        if (xsetid == null) {
           throw new InvalidArgumentException(ERROR_XUID_IS_NULL);
        }

        ReferenceXSet copy = new ReferenceXSet( this, xsetid, copyMode );
        try
        {
           // Remove all items in the hold list, and set hold to false.
           XIterator iter = copy.openFieldIterator(XSet.XSET_HOLD_LIST);
           while (iter.hasNext())
           {
              Field f = copy.getField((String) iter.next());
              copy.removeField(f);
           }
         
           // Set hold to to false.
           Property hold = (Property) copy.getField(XSet.XSET_HOLD);
           hold.setValue(Boolean.FALSE,
              XAMLibrary.STYPE_BOOLEAN_MIME_TYPE);
         
           // Remove fields required by the specification (section V1.0 8.2.4)
           Field f = copy.getField(XSet.XAM_TIME_XUID);
           copy.removeField(f);
         
           f = copy.getField(XSet.XAM_TIME_COMMIT);
           copy.removeField(f);
         
           f = copy.getField(XSet.XAM_TIME_ACCESS);
           copy.removeField(f);
         
           f = copy.getField(XSet.XAM_TIME_RESIDENCY);
           copy.removeField(f);
         
           f = copy.getField(XSet.XAM_XUID);  // See Arch Spec V1.0 Section 8.5
           copy.removeField(f);
         
           // Set dirty flag, see Arch Spec V1.0 8.5.3.2.2
           copy.addField(new Property(
              XSet.XAM_DIRTY,
              XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
              Boolean.TRUE,
              false, 
              true));
        }
        catch (XAMException xe)
        {
           copy.close();
           throw xe;
        }
        finally {
           if( m_enterExit )
              logger.exiting( s_logClass, method );
        }
      
        this.addXSet(copy);
        return (XSet) Proxy.newProxyInstance(copy.getClass().getClassLoader(),
           copy.getClass().getInterfaces(), 
           new XSetClientInvocationHandler(copy));        
    }
    
    /**
     * @see XSystem#createXSet(java.lang.String) 
     */
    public XSet createXSet(String createMode)
            throws AuthenticationExpiredException, AuthorizationException, 
            InvalidArgumentException, XSystemAbandonException, 
            XSystemCorruptException, XAMException {
        final String method = "createXSet";
        if( m_enterExit )
           logger.entering( s_logClass, method );
        
        try {
           authStatus.checkAccessAllowed();
           ReferenceXSet xset = 
              new ReferenceXSet(this, createMode);
           addXSet(xset);
        
           return (XSet) Proxy.newProxyInstance(xset.getClass().getClassLoader(),
              xset.getClass().getInterfaces(), 
              new XSetClientInvocationHandler(xset));    
        } finally {        
           if( m_enterExit )
              logger.exiting( s_logClass, method );
        }
    }

    
    /**
     * @see XSystem#deleteXSet(org.snia.xam.XUID) 
     */
    public void deleteXSet(XUID xsetid)
            throws AuthorizationException, AuthenticationExpiredException, 
            XSetUnderRetentionException, XSetUnderHoldException, 
            InvalidArgumentException, XSetInaccessibleException, 
            XSystemAbandonException, XSystemCorruptException, 
            XAMException {
       final String method = "deleteXSet";
       if( m_enterExit )
          logger.entering( s_logClass, method );
       
        if (xsetid == null) {
            throw new InvalidArgumentException(ERROR_XUID_IS_NULL);
        }
        
        m_database_manager.deleteXSet( this, xsetid );
        
        // Note: Use the special non-AccessTime modifying constructor.
        ReferenceXSet xset = new ReferenceXSet(this, xsetid );
        checkDeleteAllowed(xset);
        persistenceMgr.deleteXSet(xsetid, xset.getActualShred() );
        xset.close();
        
        if( m_enterExit )
           logger.exiting( s_logClass, method );
    }

    
    /**
     * @see XSystem#getXSetAccessTime(org.snia.xam.XUID) 
     */
    public Calendar getXSetAccessTime(XUID xsetid)
            throws AuthenticationExpiredException, InvalidArgumentException, 
            XSetInaccessibleException, XSystemAbandonException, 
            XSetCorruptException, XAMException {
       final String method = "getXSetAccessTime";
       if( m_enterExit )
          logger.entering( s_logClass, method );
       
       if (xsetid == null) {
          throw new InvalidArgumentException(ERROR_XUID_IS_NULL);
       }

       authStatus.checkAccessAllowed();
       ReferenceXSet xset = null;
       try {
          // Note: use the special non-AccessTime modifying constructor
          xset = new ReferenceXSet(this, xsetid );
          String accessPolicy = "";
          try
          {
             Property ap = xset.getProperty( XSet.XAM_ACCESS_POLICY, XAMLibrary.STYPE_STRING_MIME_TYPE );
             accessPolicy = (String) ap.getValue();
          }
          catch( FieldDoesNotExistException fdne ) { /* can ignore it for this segment */ }
          if( m_access_policy_mgr.allowReadAccess(accessPolicy, getAuthenticatedUsername()) )
          {
             Property at = xset.getProperty( XSet.XAM_TIME_ACCESS, XAMLibrary.STYPE_DATETIME_MIME_TYPE );
             return (Calendar) at.getValue();
          }
          else throw new XSetInaccessibleException();
       }
       finally {
          if( xset != null )
             xset.close();
          if( m_enterExit )
             logger.exiting( s_logClass, method );          
       }
    }

    
    /**
     * @see XSystem#holdXSet(org.snia.xam.XUID, java.lang.String) 
     */
    public void holdXSet(XUID xsetid, String holdID)
            throws AuthenticationExpiredException, AuthorizationException, 
                InvalidFieldNameException, HoldIdException, 
                InvalidArgumentException, XSetInaccessibleException, 
                XSystemAbandonException, XSystemCorruptException, 
                XAMException {
       final String method = "holdXSet";
       if( m_enterExit )
          logger.entering( s_logClass, method );
       
       if (xsetid == null) {
          throw new InvalidArgumentException(ERROR_XUID_IS_NULL);
       }

       authStatus.checkAccessAllowed();
        
       ReferenceXSet xset = new ReferenceXSet( this, xsetid, XSet.MODE_UNRESTRICTED);
       
       // Add the ID to the hold list and set the hold boolean.
       try {
          if( ! xset.getBoolean( XSet.XSET_HOLD ))
          {
             xset.addField(new Property( XSet.XSET_HOLD,
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,  /* prop value */
                false,  /* binding */
                true)); /* readonly */
          }
          xset.addField(new Property( XSet.XSET_HOLD_LIST+holdID,
             XAMLibrary.STYPE_STRING_MIME_TYPE,
             holdID,  /* prop value */
             false,  /* binding */
             true)); /* readonly */
          xset.commit();
       }
       finally {
          xset.close();
          if( m_enterExit )
             logger.exiting( s_logClass, method );         
       }   
    }

    /**
     * @see XSystem#isXSetRetained(org.snia.xam.XUID) 
     */
    public boolean isXSetRetained(XUID xsetid)
            throws AuthenticationExpiredException, InvalidArgumentException, 
                XSetInaccessibleException, XSystemAbandonException, 
                XSystemCorruptException, XAMException {
       final String method = "isXSetRetained";
       if( m_enterExit )
          logger.entering( s_logClass, method );
       
        if (xsetid == null) {
            throw new InvalidArgumentException(ERROR_XUID_IS_NULL);
        }
        
        boolean isRetained = false;
        
        // Note: Use the special non-AccessTime modifying constructor
        ReferenceXSet xset = new ReferenceXSet( this, xsetid);
        try {
           isRetained = RetentionMgr.isRetained(xset);
        }
        finally {
           xset.close();
           if( m_enterExit )
              logger.exiting( s_logClass, method );         
        }
        return isRetained;
    }
    
    /**
     * @see XSystem#openXSet(org.snia.xam.XUID, java.lang.String) 
     */
    public XSet openXSet(XUID xsetid, String openMode)
            throws AuthenticationExpiredException, AuthorizationException, 
                InvalidArgumentException, InvalidXSetModeException, 
                XSetInaccessibleException, XSystemAbandonException, 
                XSystemCorruptException, XAMException {
       final String method = "openXSet";
       if( m_enterExit )
          logger.entering( s_logClass, method );
               
          if (xsetid == null) {
             throw new InvalidArgumentException(ERROR_XUID_IS_NULL);
          }

          ReferenceXSet xset = 
                new ReferenceXSet(this, xsetid, openMode);
        
        String accessPolicy = null;
        try { accessPolicy = xset.getString( XSet.XAM_ACCESS_POLICY ); }
        catch( FieldDoesNotExistException fdne ) { ; /* ignorable here */ }
        
        if( ! m_access_policy_mgr.allowAccess( openMode, accessPolicy, authStatus.getAuthenticatedUsername() ))
           throw new XSetInaccessibleException();
        
        if( (openMode.equals( XSet.MODE_RESTRICTED) || 
             openMode.equals( XSet.MODE_UNRESTRICTED) ) &&
           this.isXSetUnderHold(xsetid)) {
           String msg = "XSet it under hold, can not open for modify";
           logger.warning(msg);
           throw new XSetUnderHoldException( msg );
        }
        
        addXSet(xset);
        XSet rvalue =  (XSet) Proxy.newProxyInstance(xset.getClass().getClassLoader(),
                xset.getClass().getInterfaces(), 
                new XSetClientInvocationHandler(xset));
        if( m_enterExit )
           logger.exiting( s_logClass, method );
        return rvalue;
    }

    /**
     * @see XSystem#releaseXSet(org.snia.xam.XUID, java.lang.String) 
     */
    public void releaseXSet(XUID xsetid, String holdID)
            throws AuthenticationExpiredException, AuthorizationException, 
                HoldIdException, InvalidArgumentException, 
                XSetInaccessibleException, XSystemAbandonException, 
                XSystemCorruptException, XAMException {
       
       final String method = "releaseXSet";
       if( m_enterExit )
          logger.entering( s_logClass, method );
       
       if (xsetid == null) {
          throw new InvalidArgumentException(ERROR_XUID_IS_NULL);
       }
       
       authStatus.checkAccessAllowed();
       ReferenceXSet xset = new ReferenceXSet( this, xsetid, XSet.MODE_UNRESTRICTED);
       
       // Add the ID to the hold list and set the hold boolean.
       String hold_list_prop = XSet.XSET_HOLD_LIST + holdID ;
       try{  
          xset.getString( hold_list_prop ); // Make sure it's under hold for this ID
          Field f = xset.getField(hold_list_prop);
          xset.removeField(f);
          XIterator iter = xset.openFieldIterator(XSet.XSET_HOLD_LIST);
          if( ! iter.hasNext() )
          {
             Property hold = (Property) xset.getField(XSet.XSET_HOLD);
             hold.setValue( Boolean.FALSE, XAMLibrary.STYPE_BOOLEAN_MIME_TYPE );
          }
          xset.commit();
       }
       catch( FieldDoesNotExistException fdne ) {
          String msg = "The XSet is not being held with id '" + holdID + "'";
          logger.warning(msg);
          throw new HoldIdException( msg );
       }
       finally {
          xset.close();
          if( m_enterExit )
             logger.exiting( s_logClass, method );          
       }
    }
    
    public void setFieldAsBinding( String fieldName )
    throws InvalidOperationException
    {
       throw new InvalidOperationException( "ReferenceVIM does not support binding fields" );
    }

    public void setFieldAsNonbinding( String fieldName )
    throws InvalidOperationException
    {
       throw new InvalidOperationException( "ReferenceVIM does not support binding fields" );
    }

    // Helper methods
    
    /**
     * Determines whether the XSystem is initializing
     * 
     * @return true if XSystem currently being initialized; false otherwise
     */
    protected boolean isInitializing() {
      boolean initializing = false;
      try {
         if (containsField(XSystem.XAM_XSYSTEM_INITIALIZING)) {
            if (getBoolean(XSystem.XAM_XSYSTEM_INITIALIZING))  {
               initializing = true;
            }
         }
      }
      catch (XAMException ex) { /* ignore */ }

      return initializing;
    }
    
    // Methods to help track open XSets; manage xsetList
    private void addXSet(ReferenceXSet xset) {
        xsetList.add(xset);
    }
    
    private void removeXSet(ReferenceXSet xset) {
        xsetList.remove(xset);
    }
    
    private void clearXSetList() {
        xsetList.clear();
    }
    
    private boolean xsetOpen(XUID xuid) {
        boolean open = false;
        Iterator<ReferenceXSet> iterator = xsetList.iterator();
        while (iterator.hasNext()) {
            ReferenceXSet chkXSet = iterator.next();
            if (xuid.equals(chkXSet.getXUID())) {
                if (!chkXSet.isClosed()) {
                    open = true;
                }
                break;
            }
        }
        return open;
    }

    /**
     * XSet is closed so remove it from the interal list
     * @param xset
     * @throws org.snia.xam.XAMException
     */
    protected void xsetClosed(ReferenceXSet xset) throws XAMException {
        if (xset.isClosed()) {
            removeXSet(xset);
        }        
    }
    
    /**
     * Verifies XSets are either closed or abandoned. If XSets are in
     * abandoned state then it will close them which removes them from
     * <code>xsetList</code>.
     * 
     * @throws org.snia.xam.XAMException
     */
    private void verifyXSetsClosed() throws XAMException {
        
        // Note: xsetList elements removed by call to xset.close() command
        for (int i = xsetList.size() - 1; i >= 0; i--) {
            ReferenceXSet xset = xsetList.get(i);
            if (xset.isAbandoned()) {
                    xset.close();
            } else if (!xset.isClosed()) {
                // XSets should not be in the list if closed but it's checked
                // before throw exception
               String msg = "Must close open XSets before closing XSystem";
               logger.warning(msg);
               throw new ObjectInUseException(msg);
            }         
        }
    }
   
    /**
     * Abandon all open XSets. This is called only when the XSystem is being
     * abandoned.
     * @throws org.snia.xam.XAMException
     */
    private void abandonXSets() throws XAMException {
        
        Iterator<ReferenceXSet> iterator = xsetList.iterator();
        while (iterator.hasNext()) {
            ReferenceXSet xset = iterator.next();
            if (!xset.isClosed() && !xset.isAbandoned()) {
                xset.abandon();
            }
        }
    }
    
    /**
     * Verifies XSystem has no running jobs. If there is a running job, then 
     * the system cannot close.  
     * If the XSets are in the abandoned state, then the XSet will be closed.
     * 
     * @throws org.snia.xam.XAMException
     */
    private void verifyJobsCompleted() throws XAMException {
       // Previous versions of this code used to check the JobMgr count of running jobs. 
       // This is going to work since the JobMgr is static, and retains jobs across multiple 
       // XSystem instances. This code must check for the XSets pertaining to this XSystem instance.
       Enumeration<ReferenceXSet> x = xsetList.elements();
       while( x.hasMoreElements() )
       {
          ReferenceXSet xset = x.nextElement();
          if( ! xset.isAbandoned() && JobMgr.isJobRunningForXSet(xset) )
          {
             String msg = "Can not close XSystem when jobs still running.";
             logger.warning(msg);
             throw new XAMException(msg + " " + JobMgr.numSubmittedAndRunning());
          }
       }
    }

    /**
     * Set up the initial System properties and there default values
     */
    protected void loadDefaults() 
    throws XAMException {
        String hostname = "localHost";
        try {
            hostname = InetAddress.getLocalHost().getHostName();
        }
        catch (UnknownHostException ignore) {}
       addField(new Property(XAM_XSYSTEM_INITIALIZING,
           XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
           Boolean.TRUE,
           false,
           false  ));

       addField(new Property(
                XAM_XSYSTEM_IDENTITY,
                XAMLibrary.STYPE_STRING_MIME_TYPE,
                new StringBuffer(hostname)
                    .append(" File System").toString(),
                false,
                true));
        addField(new Property(
                XAM_XSYSTEM_TIME,
                XAMLibrary.STYPE_DATETIME_MIME_TYPE,
                Calendar.getInstance(),
                false, 
                true));
        addField(new Property(
                XAM_XSYSTEM_LIMITS_MAX_FIELDS,
                XAMLibrary.STYPE_INT_MIME_TYPE,
                new Long(10000),
                false,
                true));
        addField(new Property(
                XAM_XSYSTEM_LIMITS_MAX_STREAM_SIZE,
                XAMLibrary.STYPE_INT_MIME_TYPE,
                XAMConstants.XSTREAM_MAX_SIZE,
                false, 
                true));
        
        // Authentication properties
        // These properties are required to be in the un authenticated system
        addField(new Property(XAM_XSYSTEM_AUTH_SASL_LIST + 
                ReferenceSaslUtils.SASL_MECHANISM_ANONYMOUS,
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false,
                true));
        addField(new Property(XAM_XSYSTEM_AUTH_SASL_LIST + 
                ReferenceSaslUtils.SASL_MECHANISM_PLAIN,
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false,
                true));
        
        addField(new Property(XAM_XSYSTEM_AUTH_SASL_DEFAULT,
                XAMLibrary.STYPE_STRING_MIME_TYPE,
                ReferenceAuthenticationStatus.SASL_DEFAULT,
                false,
                true));
        addField(new Property(
                XAM_XSYSTEM_AUTH_EXPIRATION,
                XAMLibrary.STYPE_INT_MIME_TYPE,
                new Long(-1),
                false, 
                true));
        
        // Auth granules have been removed 2009-02-16, will be added in authentcation state.
        
        addField(new Property(
                XAM_XSYSTEM_AUTH_IDENTITY_AUTHENTICATION,
                XAMLibrary.STYPE_STRING_MIME_TYPE,
                XAMConstants.EMPTY_STRING,
                false,
                true));
               
        addField(new Property(
                XAM_XSYSTEM_AUTH_IDENTITY_AUTHORIZATION,
                XAMLibrary.STYPE_STRING_MIME_TYPE,
                XAMConstants.EMPTY_STRING,
                false,
                true));
        addField(new Property(
                XAM_XSYSTEM_ACCESS,
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false,
                true));
        
        addField(new Property(
                XAM_XSYSTEM_JOB_COMMIT_SUPPORTED,
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.FALSE,
                false, 
                true));
        addField(new Property(
                XAM_XSYSTEM_JOB_LIST_QUERY,
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false, 
                true));
        addField(new Property(
                XAM_XSYSTEM_JOB_QUERY_CONTINUANCE_SUPPORTED,
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.FALSE,
                false, 
                true));
        addField(new Property(
                XAM_XSYSTEM_JOB_QUERY_LEVEL1_SUPPORTED,
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false,
                true));
        addField(new Property(
                XAM_XSYSTEM_JOB_QUERY_LEVEL2_SUPPORTED,
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.FALSE,
                false,
                true));
        addField(new Property(
                XAM_DELETION_AUTODELETE,
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false,
                true));
        addField(new Property(
                XAM_DELETION_SHRED,
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false,
                true));
        
        // Disposition Policies
        // Audodelete
        addField(new Property(
           XAM_DELETION_AUTODELETE_POLICY_LIST + DispositionPolicy.DISPOSITION_AUTODELETE,
           XAMLibrary.STYPE_STRING_MIME_TYPE,
           DispositionPolicy.DISPOSITION_AUTODELETE,
           false,
           true));
        
        addField(new Property(
           XAM_DELETION_AUTODELETE_POLICY_LIST + DispositionPolicy.DISPOSITION_SHRED,
           XAMLibrary.STYPE_STRING_MIME_TYPE,
           DispositionPolicy.DISPOSITION_SHRED,
           false,
           true));
        
        addField(new Property(
           XAM_DELETION_AUTODELETE_POLICY_LIST + DispositionPolicy.DISPOSITION_AUTODELETE_AND_SHRED,
           XAMLibrary.STYPE_STRING_MIME_TYPE,
           DispositionPolicy.DISPOSITION_AUTODELETE_AND_SHRED,
           false,
           true));
        
        // Shred
        addField(new Property(
           XAM_DELETION_SHRED_POLICY_LIST + DispositionPolicy.DISPOSITION_AUTODELETE,
           XAMLibrary.STYPE_STRING_MIME_TYPE,
           DispositionPolicy.DISPOSITION_AUTODELETE,
           false,
           true));
        
        addField(new Property(
           XAM_DELETION_SHRED_POLICY_LIST + DispositionPolicy.DISPOSITION_SHRED,
           XAMLibrary.STYPE_STRING_MIME_TYPE,
           DispositionPolicy.DISPOSITION_SHRED,
           false,
           true));
        
        addField(new Property(
           XAM_DELETION_SHRED_POLICY_LIST + DispositionPolicy.DISPOSITION_AUTODELETE_AND_SHRED,
           XAMLibrary.STYPE_STRING_MIME_TYPE,
           DispositionPolicy.DISPOSITION_AUTODELETE_AND_SHRED,
           false,
           true));
        
        addField(new Property(
                new StringBuffer(XAM_MANAGEMENT_POLICY_LIST)
                    .append(REF_VIM_DEFAULT_POLICY).toString(),
                XAMLibrary.STYPE_STRING_MIME_TYPE,
                REF_VIM_DEFAULT_POLICY,
                false, 
                true));
        addField(new Property(
                XAM_MANAGEMENT_POLICY_DEFAULT,
                XAMLibrary.STYPE_STRING_MIME_TYPE,
                REF_VIM_DEFAULT_POLICY,
                false, 
                true));
        
        addRetentionPolicies();
        
        // Default logging parameters
        addField(new Property( XAMLibrary.XAM_LOG_PATH,
                               XAMLibrary.STYPE_STRING_MIME_TYPE,
                               DEFAULT_LOG_PATH,
                               false, 
                               false ));
        addField(new Property( XAMLibrary.XAM_LOG_LEVEL,
                               XAMLibrary.STYPE_INT_MIME_TYPE,
                               DEFAULT_LOG_LEVEL,
                               false, 
                               false));
        addField(new Property( XAMLibrary.XAM_LOG_VERBOSITY,
                               XAMLibrary.STYPE_INT_MIME_TYPE,
                               DEFAULT_LOG_VERBOSITY,
                               false, 
                               false));
        addField(new Property( XAMLibrary.XAM_LOG_APPEND,
                               XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                               false,
                               false, 
                               false));
        addField(new Property( XAMLibrary.XAM_LOG_MAX_SIZE,
                               XAMLibrary.STYPE_INT_MIME_TYPE,
                               DEFAULT_LOG_SIZE,
                               false, 
                               false));
        addField(new Property( XAMLibrary.XAM_LOG_MAX_ROLLOVERS,
                               XAMLibrary.STYPE_INT_MIME_TYPE,
                               DEFAULT_LOG_COUNT,
                               false, 
                               false));
        this.updateLogger(true);
        
        /* Set up a default, in case the initialization protocol has been botched. This has been observed */
        m_async_manager = new ReferenceXAsyncManager(this, 2);
    }

    private void addRetentionPolicies() throws XAMException
    {
       Vector<String> durationPolicies = RetentionMgr.GetDurationPolicies();
       Enumeration<String> e = durationPolicies.elements();
       while( e.hasMoreElements() )
       {
          String name = e.nextElement();
          addField(new Property(
             XSystem.XAM_RETENTION_DURATION_POLICY_LIST + name,
             XAMLibrary.STYPE_STRING_MIME_TYPE,
             name,
             false, 
             true));
       }

       Vector<String> enabledPolicies = RetentionMgr.GetEnabledPolicies();
       e = enabledPolicies.elements();
       while( e.hasMoreElements() )
       {
          String name = (String) e.nextElement();
          addField(new Property(
             XSystem.XAM_RETENTION_ENABLED_POLICY_LIST + name,
             XAMLibrary.STYPE_STRING_MIME_TYPE,
             name,
             false, 
             true));
       }
    }

   /**
     * This package access method is available for worker packages to be able to publish functionality
     * via properties in the XSystem.
     * @param name     The property name.
     * @param readonly If this value is readonly
     * @param value    The property value.
     * @throws XAMException
     */
    void createXSystemProperty( String name, boolean readonly, String value ) throws XAMException
    {
       addField(new Property( name, XAMLibrary.STYPE_STRING_MIME_TYPE,
                              value, false /* binding */, readonly ));
    }
    
    /**
     * Create the authentication status, and potentially load the usernames from a file.
     * When done, create the access policies for each user, plus the two other global 
     * access policies.
     * @throws XAMException 
     */
    private void loadAuthentication() throws XAMException
    {
       String passwd_path = null;
       
       try { passwd_path = this.getString(PROPERTY_PASSWD); }
       catch( FieldDoesNotExistException fdne ) { ; /* ignore it */ }
       
       if( (passwd_path != null) && (! passwd_path.equals("")) )
          authStatus.loadAuthenticationInfo(passwd_path);
    }
    /**
     * 
     * @param fieldName
     * @throws org.snia.xam.XAMException
     */
    protected void checkProtectedNameSpace(String fieldName)
            throws XAMException {
        if (!isInitializing()) {
            super.checkProtectedNameSpace(fieldName);
        }
    }

    /**
     * Get the currently authenticated username.
     * @return The username currently authenticated.
     */
    protected String getAuthenticatedUsername() {
       return authStatus.getAuthenticatedUsername();
    }
    
    /**
     * Get the access manager in use by the XSystem.
     * @return The access manager.
     */
    AccessPolicyManager getAccessManager() { return this.m_access_policy_mgr; }

    /**
     * We have an anonymous user.  Only allow read like operations.  
     */
    private void setAnonymousPrivileges() 
        throws XAMException {
       addField(new Property(
         new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST).append(GRANULE_READ).toString(), 
         XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, 
         Boolean.TRUE,
         false, 
         true));
      addField(new Property(
         new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST).append(GRANULE_WRITE_APPLICATION).toString(), 
         XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, 
         Boolean.FALSE,
         false, 
         true));
      addField(new Property(
         new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST).append(GRANULE_WRITE_SYSTEM).toString(), 
         XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, 
         Boolean.FALSE,
         false, 
         true));
      addField(new Property(
         new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST).append(GRANULE_CREATE).toString(), 
         XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, 
         Boolean.FALSE,
         false, 
         true));
      addField(new Property(
         new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST).append(GRANULE_DELETE).toString(), 
         XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, 
         Boolean.FALSE,
         false, 
         true));
      addField(new Property(
         new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST).append(GRANULE_JOB).toString(),
         XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, 
         Boolean.TRUE,
         false, 
         true));
      addField(new Property(
         new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST).append(GRANULE_JOB_COMMIT).toString(), 
         XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, 
         Boolean.FALSE,
         false, 
         true));
      addField(new Property(
         new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST).append(GRANULE_HOLD).toString(), 
         XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, 
         Boolean.FALSE,
         false, 
         true));
      addField(new Property(
         new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST).append(GRANULE_RETENTION_EVENT).toString(), 
         XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, 
         Boolean.FALSE,
         false, 
         true));
      
      Property authentication = this.getPropertyField(XAM_XSYSTEM_AUTH_IDENTITY_AUTHENTICATION);
      authentication.setValue("anonymous", XAMLibrary.STYPE_STRING_MIME_TYPE );
      Property authorization = this.getPropertyField(XAM_XSYSTEM_AUTH_IDENTITY_AUTHORIZATION);
      authorization.setValue( "", XAMLibrary.STYPE_STRING_MIME_TYPE );
    }
    
    /**
       * Allow user to perform any operation
       */
    private void setFullPrivileges() 
        throws XAMException {
        addField(new Property(
                new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST)
                    .append(GRANULE_READ).toString(),
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false, 
                true));
        addField(new Property(
                new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST)
                    .append(GRANULE_WRITE_APPLICATION).toString(),
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false, 
                true));
        addField(new Property(
                new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST)
                    .append(GRANULE_WRITE_SYSTEM).toString(),
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false, 
                true));
        addField(new Property(
                new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST)
                    .append(GRANULE_CREATE).toString(),
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false, 
                true));
        addField(new Property(
                new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST)
                    .append(GRANULE_DELETE).toString(),
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false, 
                true));
        addField(new Property(
                new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST)
                    .append(GRANULE_JOB).toString(),
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false, 
                true));
        addField(new Property(
                new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST)
                    .append(GRANULE_JOB_COMMIT).toString(),
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false, 
                true));
        addField(new Property(
                new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST)
                    .append(GRANULE_HOLD).toString(),
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false, 
                true));
        addField(new Property(
                new StringBuffer(XAM_XSYSTEM_AUTH_GRANULE_LIST)
                    .append(GRANULE_RETENTION_EVENT).toString(),
                XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                Boolean.TRUE,
                false, 
                true));
 
        Property authentication = this.getPropertyField(XAM_XSYSTEM_AUTH_IDENTITY_AUTHENTICATION);
        authentication.setValue(this.getAuthenticatedUsername(), XAMLibrary.STYPE_STRING_MIME_TYPE );
        Property authorization = this.getPropertyField(XAM_XSYSTEM_AUTH_IDENTITY_AUTHORIZATION);
        authorization.setValue(this.getAuthenticatedUsername(), XAMLibrary.STYPE_STRING_MIME_TYPE );
    }
    /*
     * Determine if it is OK to delete the specified XSet
     */
    protected void checkDeleteAllowed(ReferenceXSet xset) 
                                        throws XSystemException, XAMException {
        XUID xsetid = xset.getXUID();
        /*
         * Checking to see if the XSet is open covers a variety of illegal
         * deletion scenarios, other than the XSet just being open.
         * For instance, in order to run a job, the XSet being used to run the
         * job must be open.  Since an open XSet cannot be deleted, then an XSet
         * which has a running job against it cannot be deleted either.
         */
        if (this.xsetOpen(xsetid)) {
           String msg = "XSet is currently open. Close before deleting";
           logger.warning(msg);
           throw new XAMException(msg);
        }
        // check to see if the XSet is retained
        if (this.isXSetRetained(xsetid)) {
           String msg = "Unable to delete the XSet when it is retained.";
           logger.warning(msg);
           throw new XSetUnderRetentionException();
        }       
        if (this.isXSetUnderHold(xsetid)) {
           String msg = "Unable to delete the XSet, currently under hold";
           logger.warning(msg);
           throw new XSetUnderHoldException( msg );
        }
        
        // Check to make sure no job running
        if (JobMgr.isJobRunningForXSet(xset)) {
           String msg = "XSet has a job running. Wait until job completes before deleting.";
           logger.warning(msg);
           throw new XAMException(msg);
        }
    }

    /**
     * Determines if the XSet is under hold. Simple check, just verify that .xset.hold is true.
     * @param xsetid
     * @return
     * @throws XAMException 
     */
    boolean isXSetUnderHold(XUID xsetid) throws XAMException
    {
       if (xsetid == null) {
          throw new InvalidArgumentException(ERROR_XUID_IS_NULL);
       }

       // NOTE: Use the special non-AccessTime modifying constructor
       ReferenceXSet xset = new ReferenceXSet( this, xsetid );
       boolean hold = xset.getBoolean(XSet.XSET_HOLD);
       xset.close();
       return hold;
    }

    /*
     * May need to perform more checks so leave this wrapper routine for
     * the time being.
     */
    protected void checkAccessAllowed() throws XSystemException, XAMException {
        authStatus.checkAccessAllowed();
    }
    
    /*
     * May need to perform more checks so leave this wrapper routine for
     * the time being. 
     */
    protected void checkFieldReadAccessAllowed() throws 
            XSystemException, XAMException {
        authStatus.checkFieldReadAccessAllowed();
    }

    /**
     * Returns the name of the system's management policy, which is
     * .xsystem.management.policy.default.
     * @return String the name of the default policy
     * @throws FieldDoesNotExistException error if the field does not exist
     */
    protected String getDefaultPolicy() throws XAMException {
        Field field = (Field)this.getField(XAM_MANAGEMENT_POLICY_DEFAULT);
        return (String)((Property)field).getValue();
    }

    /**
     * Return the location of the repository.  This is where all the data for
     * all the XSets associated with this XSystem are stored.
     * @return String the location of the repository
     */
    protected String getRepositoryLocation() {
        return persistenceMgr.getRepositoryLocation();
    }

    /**
     * Retrieves the persistence manager instance for this XSystem
     * @return persistence manager
     */
    protected ReferenceVIMPersistenceMgr getPersistenceMgr() {
        return persistenceMgr;
    }
    
    /**
     * Retrieves the database manager instance for this XSystem.
     * @return The database manager instance
     */
    DBManager getDBManager() {
       return m_database_manager;
    }
    
    /**
     * Retrieve the asynchronous operation manager.
     * @return The async manager
     */
    ReferenceXAsyncManager getXAsyncManager() {
       return m_async_manager;
    }

    protected boolean isClosed() {
        return this.authStatus.isClosed();
    }
}
