/*
 * Copyright (c) 2008, Hewlett-Packard Development Company, L.P. 
 * Copyright (c) 2008-2009, The Storage Networking Industry Association.
 * Copyright (c) 2008-2009, Sun Microsystems, Inc.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.util.Calendar;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Set;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.snia.xam.FieldExistsException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.PolicyNameException;
import org.snia.xam.RetentionValueException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XIterator;
import org.snia.xam.XSet;
import org.snia.xam.XSetException;
import org.snia.xam.XSetInaccessibleException;
import org.snia.xam.XSystem;
import org.snia.xam.base.Field;
import org.snia.xam.base.Property;

/**
 * Class designed to manage retention for the XSet. An instance of the retention manager will be
 * created for each XSet instance.
 *
 */
public class RetentionMgr {
    
    private Logger logger = null;
    
    public static final String POLICY_NONE        = "org.snia.refvim.retention.none";
    public static final String POLICY_ONE_SECONDS = "org.snia.refvim.retention.one.second"; // For testing!
    public static final String POLICY_ONE_DAY     = "org.snia.refvim.retention.one.day";
    public static final String POLICY_THIRTY_DAYS = "org.snia.refvim.retention.thirty.days";
    public static final String POLICY_ONE_YEAR    = "org.snia.refvim.retention.one.year";
    
    private static final long  ONE_SECOND = 1000; // mS in one second
    private static final long  ONE_MINUTE = 60 * ONE_SECOND;
    private static final long  ONE_HOUR = 60 * ONE_MINUTE;
    private static final long  ONE_DAY = 24 * ONE_HOUR;
    private static final long  THIRTY_DAYS = 30 * ONE_DAY;
    private static final long  ONE_YEAR = (long) (365.25 * ONE_DAY);
    
    public static final String BASE = "base";
    public static final String EVENT = "event";
    public static final String ENABLED = "enabled";
    public static final String STARTTIME = "starttime";
    public static final String DURATION = "duration";
    public static final String XAM_RETENTION_PREFIX = ".xset.retention.";

    private static final String DURATION_SUFFIX  = ".duration.policy";
    private static final String ENABLED_SUFFIX   = ".enabled.policy";
    
    private ReferenceXSet xset = null;
    
    private static Hashtable<String,RetentionPolicy> s_policies = null;
    private static Boolean   s_lock = new Boolean(true);
    
    /*
     * Set of application/client specified retention criteria identifiers
     * (i.e. <retention id>'s) for the encapsulated XSet
     */  
    private Set<String> retentionList = new HashSet<String>();
     
    protected RetentionMgr(ReferenceXSet set) 
       throws XAMException
    {
        if (set != null) {
            xset = set;
            logger = set.getXSystem().getLogger();
        } else {
            xset = new ReferenceXSet();
        }
    }
    
    /**
     * Routine to check the retentionID value
     * @param retentionID the retention identifier to validate
     * @throws org.snia.xam.InvalidArgumentException
     */
    public static void validateRetentionID(String retentionID) 
            throws InvalidArgumentException {
        if (retentionID == null) {
            throw new InvalidArgumentException(
                                            "Retention ID specified is null.");
        }
        if (retentionID.length() == 0) {
            throw new InvalidArgumentException(
                                    "Retention ID cannot be an empty string.");
        }
        if (retentionID.equals(BASE)) {
            throw new InvalidArgumentException("The retention ID must be a " +
                    "string other than " + BASE);
        }
    }
    
    /**
     * Routine to determine if the duration time value is valid.
     * @param duration the time in milliseconds the XSet should be retained
     *                 from the starttime
     * @return boolean flag indicating if the duration value is valid 
     *                 (true if greater than or equal to zero or equal to -1),
     *                 otherwise false.
     */
    public static boolean isDurationValid(long duration) {
        return (duration > 0 || isForeverRetained(duration) || 
                                                    isDurationZero(duration));
    }
    
    /**
     * Routine to determine if the XSet should be retained forever
     * @param duration the time in milliseconds the XSet should be retained
     *                 from the starttime
     * @return boolean flag indicating if the XSet is retained forever (true)
     *                 or not (false)
     */
    public static boolean isForeverRetained(long duration) {
        return (duration == -1);
    }
    
    /**
     * Routine to determine if the duration is zero and thus, the XSet is
     * not retained.
     * @param duration the time in milliseconds the XSet should be retained
     *                 from the starttime
     * @return boolean flag indicating if duration is zero (true) or 
     *                 non-zero (false)
     */
    public static boolean isDurationZero(long duration) {
        return (duration == 0);
    }
    
    /**
     * Routine to determine if the new duration value will increase the current
     * duration value.  Changing duration to the same value is also allowed.
     * @param currentValue duration currently set
     * @param newValue new duration value to be set
     * @return boolean flag indicating if the new duration value will increase
     *                 the current duration (true), otherwise false
     */
    public static boolean isDurationSameOrIncreasing(
                                            long currentValue, long newValue) {
        return (Long.valueOf(currentValue).compareTo(
                                            Long.valueOf(newValue)) <= 0 ||
                                                   isForeverRetained(newValue));
    }
    
    /**
     * Helper method to build the .xset.retention.list.<retentionID> string
     * @param retentionID the retention identifier to include
     * @return String .xset.retention.list.<retentionID> constructed string
     */
    public static String getListString(String retentionID) {
        return XSet.XAM_RETENTION_LIST + retentionID;
    }
    
    /**
     * Helper method to build the .xset.retention.<retentionID>.enabled string
     * @param retentionID the retention identifier to include
     * @return String .xset.retention.<retentionID>.enabled constructed string
     */
    public static String getEnabledString(String retentionID) {
        return XAM_RETENTION_PREFIX + retentionID + "." + ENABLED;
    }
    
    /**
     * Helper method to build the .xset.retention.<retentionID>.duration string
     * @param retentionID the retention identifier to include
     * @return String .xset.retention.<retentionID>.duration constructed string
     */
    public static String getDurationString(String retentionID) {
        return XAM_RETENTION_PREFIX + retentionID + "." + DURATION;
    }
    
    /**
     * Helper method to build the .xset.retention.<retentionID>.starttime string
     * @param retentionID the retention identifier to include
     * @return String .xset.retention.<retentionID>.starttime constructed string
     */
    public static String getStarttimeString(String retentionID) {
        return XAM_RETENTION_PREFIX + retentionID + "." + STARTTIME;
    }
    
    /**
     * Determines whether or not at least one of the retention identifiers for
     * the specified XSet has not met its retention criteria.  If so, then the
     * XSet is retained and a value of true will be returned.  If all 
     * retention criteria has been met, then this method will return false.
     * @param xset reference to a ReferenceXSet to check its retention
     * @return RetentionState the retention state
     */
    protected static boolean isRetained(ReferenceXSet xset) throws XAMException {  
        /* evaluate all retention criteria for the specified xset for
         * all retention criteria assoc. with "base", "event" and any other
         * retention ID's. */
        long now = System.currentTimeMillis();
 
        // evaluate base and event retention criteria
        if (isRetained(xset, BASE, now) || isRetained(xset, EVENT, now)) {
            return true;
        }
        /* "other" retention criteria -- if the application/client has
         * created other retention criteria via the createRetention method,
         * evaluate the XSet's other retention identifiers (i.e. <retention id>)
         */ 
        HashSet<String> otherRTC = null;
        if (xset.hasOtherRetentionCriteria()) {
            otherRTC = (HashSet<String>)xset.getOtherRetentionCriteria();
        }
        if (otherRTC != null && otherRTC.size() > 0) {
            for(String id: otherRTC) {
                if (isRetained(xset, id, now)) {
                    return true;
                }
            }
        }
        return false;
    }
    
    private static boolean isRetained(ReferenceXSet xset,
                                String retID,
                                long currentTime)
                                throws XAMException {
        String rtcStartTime = getStarttimeString(retID);
        Field persistedField = null;
        Calendar startTime = null;
        long duration = 0;
        boolean enabled = xset.getActualRetentionEnabled(retID);
        
        if (enabled) {
            if (xset.containsField(rtcStartTime)) {
                persistedField = xset.getField(rtcStartTime);
                startTime = (Calendar) ((Property) persistedField).getValue();
            }
            if (startTime == null) {
               return true; // Indefinite period
            }
            else {
               duration = xset.getActualRetentionDuration(retID);
                long criteriaTime = startTime.getTimeInMillis() + duration; 
                if (currentTime < criteriaTime) {
                    // retention time criteria (RTC) hasn't been met
                    return true;
                }
            }
        }
        return false;
    }
    
    /**
     * Accessor method to get the list of application/client created
     * retention identifiers for the XSet.  These are also referred to as
     * "other" retention.
     * @return Set<String> set of other retention identifier strings, which
     *                     excludes the .xset.retention.list. prefix
     *                   (e.g. .xset.retention.list.foo ends up as "foo" in set)
     */
    protected Set<String> getRetentionList() {
        return retentionList;
    }
    
    /**
     * Clears the set containing the application/client specified retention id's
     */
    protected void clearRetentionList() {
        retentionList.clear();
    }
    
    /*
     * Method checks to see if the field specified is an "other" retention
     * identifier, which starts with the .xset.retention.list. prefix and has
     * a retention identifier other than "base" or "event".  If the field meets
     * this criteria, it is added to the retention list for easy lookup.
     * @param field the XSet field to check for .xset.retention.list prefix
     */
    protected void checkFieldAsOtherRetention(Field field) {
        String fieldName = field.getName();
        // populate local vector with application/client (i.e. other) specified
        // retention criteria identifiers -- excludes "event" and "base"
        if (fieldName.startsWith(XSet.XAM_RETENTION_LIST)) {
            int initPos = XSet.XAM_RETENTION_LIST.length() - 1;
            // strip the field name's prefix of ".xset.retention.list."
            // and leave only the retention identifier (i.e. <retention id>)
            String retentionID = fieldName.substring(++initPos);
            if (!(retentionID.equals(EVENT) || retentionID.equals(BASE))) {
                    retentionList.add(retentionID);
            }   
        }
        xset.addField(field);
    }
    
    /*
     * Method creates retention for the specified retention id if retention
     * doesn't already exist for specified identifier.
     * @param retentionID retention identifier to create
     * @param binding flag indicating whether or not the retention id created
     *                is binding (true) or non-binding (false).
     * @throws XAMException
     */
    protected void createRetention(String retentionID, boolean binding) 
                                                        throws XAMException {
        // check .xset.retention.list.<retentionID> field  -- if the field
        // already exists, then notify application the retention ID is invalid
        String newRetID = getListString(retentionID);
        if (xset.containsField(newRetID)) {
           String msg = "The retention identifier, \"" +
              retentionID + "\", is already in use.";
              logger.warning(msg);
           throw new InvalidArgumentException(msg);
        }
        xset.addFieldSetDirty(new Property(newRetID,
                                            XAMLibrary.STYPE_STRING_MIME_TYPE,
                                            retentionID,
                                            new Boolean(binding), 
                                            true));                                         
        // keep track of "other" retention identifiers created for this XSet
        retentionList.add(retentionID);
    }
    
    /*
     * Method sets the retention enabled flag for the specified retention id.
     * It checks to see if the .xset.retention.list.<retentionID> field exists -
     * if so, it sets .xset.retention.<retentionID>.enabled only to TRUE IFF 
     * the current enabled value is false.  If the current enabled value is
     * true, it cannot be set to FALSE.  If the <retentionID>.enabled does not
     * exist, then create it and set it to the value specified.
     * @param retentionID retention identifier for which enabled flag is set
     * @param binding flag indicating whether or not the enabled field
     *                created for the specified retention identifier 
     *                is binding (true) or non-binding (false).
     * @param enabled flag indicating whether or not the enabled value is true
     *                or false
     * @throws XAMException
     */
    protected void setEnabled(String retentionID, boolean binding, boolean enabled)
            throws XAMException {

          boolean currentlyEnabled = xset.getActualRetentionEnabled(retentionID);
          if( currentlyEnabled && ! enabled ) {
             String msg = "Retention may not be disabled once it has been enabled. Retention ID = " 
                + retentionID;
             logger.warning(msg);
             throw new RetentionValueException( msg );
          }
          
        String xamRetentionID = getListString(retentionID);
        String xamRetentionEnabled = getEnabledString(retentionID);
        if (!xset.containsField(xamRetentionID)) {
           String msg = "The retention identifier, \"" +
                retentionID + "\", does not exist.";
           logger.warning(msg);
           throw new InvalidArgumentException(msg);
        }
        
        if (!xset.containsField(xamRetentionEnabled)) {
            // create enabled field for the first time
            xset.addFieldSetDirty(new Property(
                                            xamRetentionEnabled,
                                            XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                                            enabled,
                                            binding,
                                            true));
        } else {
            /* .xset.retention.<retentionID>.enabled exists -- if value is
             * FALSE and enabled = TRUE, then change its value.  Otherwise, this
             * field cannot be changed from TRUE to FALSE. */
            Field field = xset.getField(xamRetentionEnabled);
            Boolean fieldValue = (Boolean) ((Property) field).getValue();
            if (fieldValue.equals(Boolean.TRUE) && !enabled) {
               String msg = "Retention is already enabled for the retention identifier, \"" + retentionID + "\".";
               logger.warning(msg);
               throw new RetentionValueException(msg);
            }
            
            if (fieldValue.equals(Boolean.FALSE) && enabled) {
                xset.addFieldSetDirty(new Property(
                                            xamRetentionEnabled,
                                            XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                                            Boolean.TRUE,
                                            binding,
                                            true));
            }
        }
    }
    
    /*
     * Method sets the retention duration for the specified retention id.
     * It checks to see if the .xset.retention.list.<retentionID> field exists
     * and <retentionID> enabled is TRUE or FALSE - if so, it checks to see if
     * .xset.retention.<retentionID>.duration exists.  If duration field exists,
     * the duration specified must INCREASE its present value.  If the field
     * exists and has a retention value of "forever" (i.e. -1), then duration 
     * cannot be set again.  If the .xset.retention.<retentionID>.duration
     * does NOT exist, then it is created and set to the value specified.
     * @param retentionID retention identifier for which duration is set
     * @param binding flag indicating whether or not the duration field
     *                created for the specified retention identifier 
     *                is binding (true) or non-binding (false).
     * @param duration value of duration in milliseconds
     * @throws XAMException
     */
    protected void setDuration(String retentionID, boolean binding, long duration) 
                                                           throws XAMException {
       long oldDuration = xset.getActualRetentionDuration(retentionID);
       if( (duration < oldDuration) &&(duration != -1) ) {
          String msg = "Retention duration may not be less than existing";
          logger.warning(msg);
          throw new RetentionValueException( msg );
       }
       
        String xamRetentionID = getListString(retentionID);
        String xamRetentionEnabled = getEnabledString(retentionID);
        String xamRetentionDuration = getDurationString(retentionID);
        if (!xset.containsField(xamRetentionID)) {
            throw new InvalidArgumentException("The retention identifier, \"" +
                    retentionID + "\", does not exist.");
        }
        if (!xset.containsField(xamRetentionEnabled)) {
            logger.log(Level.INFO, "The enabled flag is not set for \"" +
                retentionID + "\".  Until the retention id is enabled it will have no effect.");
        } 
        Long newDuration = Long.valueOf(duration);
        if (xset.containsField(xamRetentionDuration)) {
            // .xset.retention.<retentionID>.duration exists -- 
            Field field = xset.getField(xamRetentionDuration);
            long fieldValue = ((Long) ((Property) field).getValue()).longValue();
            if (isForeverRetained(fieldValue)) {
               String msg = "Unable to set the duration time to " +
                  String.valueOf(duration) + ".  Retention duration is " +
                  "already set to \"forever\" for \"" + retentionID + "\".";
               logger.warning(msg);
               throw new RetentionValueException(msg);
            }
            if (isDurationSameOrIncreasing(fieldValue, duration)) {
                xset.addFieldSetDirty(new Property(
                    xamRetentionDuration,
                    XAMLibrary.STYPE_INT_MIME_TYPE,
                    newDuration,
                    binding,
                    true));
            } else {
               String msg = "Unable to set the duration time to " +
                  newDuration.toString() + ".  Retention duration for " +
                  "\"" + retentionID + "\" is " + String.valueOf(fieldValue) +
                  ", which is not increased by " + String.valueOf(duration) + ".";
               logger.warning(msg);
               throw new RetentionValueException(msg);
            }
        } else {
            // .xset.retention.<retentionID>.duration doesn't exist so create it
            xset.addFieldSetDirty(new Property(
                xamRetentionDuration,
                XAMLibrary.STYPE_INT_MIME_TYPE,
                newDuration,
                binding,
                true));
        }
    }
    
    /*
     * Method sets the starttime for the specified retention id.
     * It checks to see if the .xset.retention.list.<retentionID> field exists - 
     * if so, it determines which enabled and duration <retentionID> values are
     * being used (i.e. value mgmt property or policy).  It checks that the
     * enabled value is TRUE and the duration value is -1 or a non-negative
     * value.  If this criteria is met, then the starttime is created
     * with the current system time and specified binding value.  It cannot
     * be set again once created.
     * @param retentionID retention identifier for which starttime is set
     * @param binding flag indicating whether or not the starttime field
     *                created for the specified retention identifier 
     *                is binding (true) or non-binding (false).
     * @throws XAMException
     */
    protected void setStarttime(String retentionID, boolean binding) 
                                                        throws XAMException {
        String xamRetentionID = getListString(retentionID);
        String xamRetentionStarttime = getStarttimeString(retentionID);
        if (!xset.containsField(xamRetentionID)) {
           String msg = "The retention identifier, \"" +
              retentionID + "\", does not exist.";
           logger.warning(msg);
           throw new InvalidArgumentException(msg);
        }
        if (xset.containsField(xamRetentionStarttime)) {
           String msg = STARTTIME + " already " + "exists for the retention identifier, \"" +
              retentionID + "\", and cannot be set again.";
           logger.warning(msg);
           throw new FieldExistsException(msg);
        }
        boolean enabled = xset.getActualRetentionEnabled(retentionID);
        if (!enabled) {
           String msg = "Unable to set the starttime for retention identifier, \"" + retentionID + "\"" +
              ".  Retention is not enabled.";
           logger.warning(msg);
           throw new XAMException(msg);
        }
        xset.addFieldSetDirty(new Property(
            xamRetentionStarttime,
            XAMLibrary.STYPE_DATETIME_MIME_TYPE,
            Calendar.getInstance(),
            binding,
            true));
    }
    
    /*
     * Method sets the base retention duration.  It creates the required
     * base retention fields, .xset.retention.list.base and
     * .xset.retention.base.enabled, if they don't exist.
     * @param binding flag indicating whether or not the base duration
     *                is binding (true) or non-binding (false).
     * @param duration duration value in milliseconds
     * @throws XAMException
     */
    protected void setBase(boolean binding, long duration) throws XAMException {
        // creates .xset.retention.list.base IFF it doesn't exist
        if (!xset.containsField(XSet.XAM_RETENTION_LIST_BASE)) {
            xset.addFieldSetDirty(new Property(
                    XSet.XAM_RETENTION_LIST_BASE,
                    XAMLibrary.STYPE_STRING_MIME_TYPE,
                    BASE,
                    true, 
                    true));             
        }
        // creates .xset.retention.base.enabled = TRUE IFF it doesn't exist
        if (!xset.containsField(XSet.XAM_RETENTION_BASE_ENABLED)) {
            xset.addFieldSetDirty(new Property(
                    XSet.XAM_RETENTION_BASE_ENABLED,
                    XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                    Boolean.TRUE,
                    true, 
                    true));
        }
        if (!xset.containsField(XSet.XAM_RETENTION_BASE_DURATION)) {
            // creates .xset.retention.base.duration IFF it doesn't exist.
            xset.addFieldSetDirty(new Property(
                    XSet.XAM_RETENTION_BASE_DURATION,
                    XAMLibrary.STYPE_INT_MIME_TYPE,
                    duration,
                    binding, 
                    true));
        } else {
            Field field = xset.getField(XSet.XAM_RETENTION_BASE_DURATION);
            long fieldValue = ((Long) ((Property) field).getValue()).longValue();
            if (isForeverRetained(fieldValue)) {
               String msg = "Unable to set the duration for the \"base\" retention identifier.  " + 
                  "Retention duration is already set to -1.";
               logger.warning(msg);
                throw new RetentionValueException(msg);
            } else if (!isDurationSameOrIncreasing(fieldValue, duration)) {
               String msg = "Unable to set the " +
                  "duration for the \"base\" retention identifier.  " + 
                  "The specified duration of " + String.valueOf(duration) +
                  " does not increase the current duration value, " + 
                  String.valueOf(fieldValue) + ".";
               logger.warning(msg);
               throw new RetentionValueException( msg );
            } else {
                xset.addFieldSetDirty(new Property(
                    XSet.XAM_RETENTION_BASE_DURATION,
                    XAMLibrary.STYPE_INT_MIME_TYPE,
                    duration,
                    binding, 
                    true));
            }
        }    
    }
    
    protected void setRetentionFieldsAtCommit(Calendar time) throws XAMException {
        if (!xset.containsField(XSet.XAM_RETENTION_LIST_BASE)) {
            xset.addFieldSetDirty(new Property(
                    XSet.XAM_RETENTION_LIST_BASE,
                    XAMLibrary.STYPE_STRING_MIME_TYPE,
                    RetentionMgr.BASE,
                    true, 
                    true));             
        }
        if (!xset.containsField(XSet.XAM_RETENTION_LIST_EVENT)) {
            xset.addFieldSetDirty(new Property(
                    XSet.XAM_RETENTION_LIST_EVENT,
                    XAMLibrary.STYPE_STRING_MIME_TYPE,
                    RetentionMgr.EVENT,
                    true, 
                    true));
        }
        if (!xset.containsField(XSet.XAM_RETENTION_BASE_ENABLED)) {
            xset.addFieldSetDirty(new Property(
                    XSet.XAM_RETENTION_BASE_ENABLED,
                    XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                    Boolean.TRUE,
                    true, 
                    true));
        }
        // set base starttime with the value of XAM_TIME_XUID only if 
        // XAM_TIME_XUID exists and XAM_RETENTION_BASE_STARTTIME does NOT exist.
        if (!(xset.containsField(XSet.XAM_RETENTION_BASE_STARTTIME))) {
            if (xset.containsField(XSet.XAM_TIME_XUID)) {
                // xuid exists so use the .xset.time.xuid value
                Field field = xset.getField(XSet.XAM_TIME_XUID);
                Calendar tmpTime = (Calendar) ((Property) field).getValue();
                xset.addFieldSetDirty(new Property(
                    XSet.XAM_RETENTION_BASE_STARTTIME,
                    XAMLibrary.STYPE_DATETIME_MIME_TYPE,
                    tmpTime,
                    true, 
                    true));
            } else {
                // new XUID being created so use time passed into method
                xset.addFieldSetDirty(new Property(
                        XSet.XAM_RETENTION_BASE_STARTTIME,
                        XAMLibrary.STYPE_DATETIME_MIME_TYPE,
                        time,
                        true, 
                        true));
            }
        }
    }

    /**
     * Get the policy referred to by name.
     * @param name The name of the policy.
     * @return The named retention policy, or null if a policy with that name does not exist.
     */
    static public RetentionPolicy GetPolicy( String name )
    {
       if( s_policies == null )
          synchronized( s_lock )
          {
             if( s_policies == null )
                s_policies = new Hashtable<String,RetentionPolicy>();
             
             s_policies.put( POLICY_NONE,        new RetentionPolicy( POLICY_NONE, 0, false ) );
             s_policies.put( POLICY_ONE_SECONDS, new RetentionPolicy( POLICY_ONE_SECONDS, ONE_SECOND, true ));
             s_policies.put( POLICY_ONE_DAY,     new RetentionPolicy( POLICY_ONE_DAY, ONE_DAY, true )); 
             s_policies.put( POLICY_THIRTY_DAYS, new RetentionPolicy( POLICY_THIRTY_DAYS, THIRTY_DAYS, true )); 
             s_policies.put( POLICY_ONE_YEAR,    new RetentionPolicy( POLICY_ONE_YEAR, ONE_YEAR, true )); 
          }
       
       return (RetentionPolicy) s_policies.get(name);
    }
    
    /**
     * Return a list of policies suitable for determining duration.
     */
    static public Vector<String> GetDurationPolicies()
    {
       // Side effect: will create s_policies if needed.
       @SuppressWarnings("unused")
      RetentionPolicy policy = GetPolicy( POLICY_NONE ); 
       
       Enumeration<String> e = s_policies.keys();
       Vector<String> list = new Vector<String>();
       
       while( e.hasMoreElements() )
       {
          String policyName = e.nextElement();
          list.add(policyName);
       }
       return list;
    }
    
    /**
     * Return a list of policies suitable for determining enabled.
     */
     static public Vector<String> GetEnabledPolicies()
     {
        // The current implementation of the retention manager, policies are good for both.
       return GetDurationPolicies(); 
     }
     
     /**
      * Construct a list of all policies being referenced by the XSet. Each policy reference will also
      * contain a list of the actual retention properties referring to the policy.
      * @param xset
      * @return The list of the policy reference list.
      * @throws XAMException 
      */
     static public Collection<RetentionPolicyReference> GetPolicyReferences( ReferenceXSet xset ) throws XAMException
     {
        XSystem xsys = xset.getXSystem();
        Hashtable<String,RetentionPolicyReference> policies = new Hashtable<String,RetentionPolicyReference>();
        XIterator iter = xset.openFieldIterator( XSet.XAM_RETENTION_LIST );
        while( iter.hasNext() )
        {
           String fieldName = (String) iter.next();
           String retentionID = xset.getString(fieldName);
           String durationRetentionPolicy = XAM_RETENTION_PREFIX + retentionID + DURATION_SUFFIX;
           String enabledRetentionPolicy  = XAM_RETENTION_PREFIX + retentionID + ENABLED_SUFFIX;
           
           if( xset.containsField( durationRetentionPolicy ))
           {          
              String policyName = xset.getString( durationRetentionPolicy );
              String policyRef = XSystem.XAM_RETENTION_DURATION_POLICY_LIST + policyName;
              RetentionPolicyReference ref = policies.get(policyRef);
              if( ref == null )
              {
                 ref = new RetentionPolicyReference( policyRef,
                    XAMLibrary.STYPE_STRING_MIME_TYPE,
                    policyName,
                    xsys.getFieldBinding(policyRef),
                    xsys.getFieldReadOnly(policyRef) );
                 policies.put(policyRef,ref);
              }
              
              RetentionPolicy policy = GetPolicy(policyName);
              Property value = new Property( XAM_RETENTION_PREFIX + retentionID + ".duration",
                                             XAMLibrary.STYPE_INT_MIME_TYPE,
                                             policy.getDuration(),
                                             false,
                                             false );
              ref.addReference( value );
           }
           
           if( xset.containsField( enabledRetentionPolicy ))
           {
              String policyName = xset.getString( enabledRetentionPolicy );
              String policyRef = XSystem.XAM_RETENTION_ENABLED_POLICY_LIST + policyName;
              RetentionPolicyReference ref = policies.get(policyRef);
              if( ref == null )
              {
                 ref = new RetentionPolicyReference( policyRef,
                    XAMLibrary.STYPE_STRING_MIME_TYPE,
                    policyName,
                    xsys.getFieldBinding(policyRef),
                    xsys.getFieldReadOnly(policyRef) );
                 policies.put(policyRef,ref);
              }
              
              RetentionPolicy policy = GetPolicy(policyName);
              Property value = new Property( XAM_RETENTION_PREFIX + retentionID + ".enabled",
                                             XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
                                             policy.getEnabled(),
                                             false,
                                             false );
              ref.addReference( value );
           }
        }
        return policies.values();
     }

     /**
      * Validate all the policy settings in the XSet against the exported list of policies in the 
      * canonical file (read in via XMLReader). If the settings would reduce retention settings, 
      * this validation process should fail.
      * @throws RetentionValueException If committing the XSet would violate the retention settings of the XSet (as imported).
      */
     protected void validateImportRetentionSettings()
     throws RetentionValueException, XAMException
     {
        HashMap<String,RetentionPolicyReference> policy_references = xset.getImportRetentionPolicyReferences();
        
        // If the policy references are not present, that means they don't need to be verified.
        if( policy_references == null )
           return;
        
        XIterator retentionIter = xset.openFieldIterator(XSet.XAM_RETENTION_LIST);
        
        try { 
           while( retentionIter.hasNext()) {
              String listId = (String) retentionIter.next();
              String id = xset.getString(listId);
              String durationField = XAM_RETENTION_PREFIX + id + DURATION_SUFFIX;
              String enabledField  = XAM_RETENTION_PREFIX + id + ENABLED_SUFFIX;
           
              if( xset.containsField(durationField) ) {
                 String policyName = xset.getString(durationField);
                 RetentionPolicy actualPolicy = GetPolicy(policyName);
                 if( actualPolicy == null ) {
                    String msg = "Reference XSystem does not support retention policy '"  +
                       policyName + "'";
                    logger.warning(msg);
                    throw new PolicyNameException( msg );
                 }
              
                 String policyContext = XSystem.XAM_RETENTION_DURATION_POLICY_LIST + policyName;
                 RetentionPolicyReference policyRef = policy_references.get( policyContext );
                 Property value = policyRef.getReferences().get(XAM_RETENTION_PREFIX + id + ".duration" );
                 if( actualPolicy.getDuration() < ((Long) value.getValue()).longValue() )
                    throw new RetentionValueException( "Imported XSet not compatible with duration of policy '" +
                       policyName + "'" );
              }
           
              if( xset.containsField(enabledField) ) {
                 String policyName = xset.getString(enabledField);
                 RetentionPolicy actualPolicy = GetPolicy(policyName);
                 if( actualPolicy == null )
                    throw new PolicyNameException( "Reference XSystem does not support retention policy '"  +
                       policyName + "'" );
              
                 String policyContext = XSystem.XAM_RETENTION_ENABLED_POLICY_LIST + policyName;
                 RetentionPolicyReference policyRef = policy_references.get( policyContext );
                 Property value = policyRef.getReferences().get(XAM_RETENTION_PREFIX + id + ".enabled" );
                 if( actualPolicy.getEnabled() != ((Boolean) value.getValue()).booleanValue() ) {
                    String msg = "Imported XSet not compatible with enabled state of policy '" +
                       policyName + "'";
                    logger.warning(msg);
                    throw new RetentionValueException( msg );
                 }
              }
           }
        } catch( RetentionValueException rve ) {
           throw rve ;
        } catch( PolicyNameException pne ) {
           throw pne ;
        } catch ( Exception e ) {
           String msg = "Unabled to validate imported policies " + e.getMessage();
           logger.log( Level.SEVERE, msg, e );
           throw new RetentionValueException( msg, e );
        }
        xset.clearRetentionPolicyReferences();
     }

     /**
      * Make sure that the new XSet does not reduce the import settings of the previously stored XSet.
      * This is typically called for imported XSets, but it should not hurt (other than time) for 
      * other XSets.
      * @throws XAMException Typically throws a RetentionValueException.
      */
     protected void validateRetentionPeriods() throws XAMException
     {
        XSet oldXSet = null;
        if( xset.getXUID() == null )
           return;
        
        try {
           oldXSet = xset.getXSystem().openXSet( xset.getXUID(), XSet.MODE_READ_ONLY);
        } catch( XSetInaccessibleException xie ) {
           return;
        } catch( Exception e ) { 
           String msg = "Unabled to validate retention of imported XSet";
           logger.warning(msg);
           throw new XSetException( msg, e );
        }
      
        try {
           XIterator iter = oldXSet.openFieldIterator( XSet.XAM_RETENTION_LIST );
           while( iter.hasNext() ) {
              String fieldName = (String) iter.next();
              String retentionID = oldXSet.getString(fieldName);
              if( oldXSet.getActualRetentionDuration(retentionID) > xset.getActualRetentionDuration(retentionID))
                 throw new RetentionValueException("Imported XSet violated retention duration of previously stored XSet, id = " +
                    retentionID );
           
              if( oldXSet.getActualRetentionEnabled(retentionID) && ! xset.getActualRetentionEnabled(retentionID))
                 throw new RetentionValueException("Imported XSet disabled retention duration of previously stored XSet, id = " +
                    retentionID );
           }
        } finally {
           oldXSet.close();
        }
     }
}
