/*
 *  Copyright (c) 2008-2009, Sun Microsystems, Inc.
 *  Copyright (c) 2008-2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.math.BigInteger;
import java.util.Calendar;
import java.util.UUID;
import org.snia.xam.InvalidXUIDException;
import org.snia.xam.XAMException;
import org.snia.xam.XUID;
import org.snia.xam.base.XAMCrc;
import org.snia.xam.base.Base64;

/**
 *  Note: Work in Progress - not final implementation
 * 
 *  Need to change algorithm for generating XUID to ensure uniqueness.
 *
 */
class SniaXUID implements XUID
{

    /**
     * Length of SNIA XUID indentifing info for xuid 
     */
    protected static final int XUID_PAYLOAD_LEN = XUID.MIN_LENGTH - 1;
    
    /**
     * The vendor identifier as identified via 
     * http://www.iana.org/assignments/enterprise-numbers
     * 
     * SNIA # is 14851
     */
    private static final int VENDOR_ID = 14851;
    private static final byte SNIA_RESERVED = 0;
    protected static final int XUID_LENGTH_INDEX = 5;
    protected static final int CRC_INDEX = 6;
    
    private byte[] data;
    private static int offset = 0;
    
    private static XAMCrc crc16 = new XAMCrc();
    private static byte[] vendorIdBytes;
    static {
        vendorIdBytes = intToByteArray(VENDOR_ID);
    }

    /**
     * Creates empty instance of SniaXUID
     */
    private SniaXUID() {
        
    }
    
    /**
     * Creates a new instance of SniaXUID. The xuid value passed is assumed
     * to be valid.
     * 
     * @param xuidBytes the xuid value
     */
    private SniaXUID(byte[] xuidBytes, int length)
            throws InvalidXUIDException
    {
        data = new byte[length];
        System.arraycopy(xuidBytes, 0, data, 0, length);
    }

    /**
     * Creates a new instance of SniaXUID based on the identifier passed in.
     * 
     * @param identifier the unique identifier for the record stored by this
     * system.
     * @throws org.snia.xam.InvalidXUIDException
     */
    private static byte[] createXUID(byte[] identifier) throws InvalidXUIDException
    {
        if (identifier.length > (XUID.MAX_LENGTH - XUID_PAYLOAD_LEN)) {
            throw new InvalidXUIDException("Specified identifier is too long.");
        }

        int length = identifier.length + XUID_PAYLOAD_LEN;
        byte[] xuidBytes = new byte[length];
        xuidBytes[0] = SNIA_RESERVED;
        // Vendor identifier bytes 1,2,3
        xuidBytes[1] = vendorIdBytes[1];
        xuidBytes[2] = vendorIdBytes[2];
        xuidBytes[3] = vendorIdBytes[3];
        xuidBytes[4] = SNIA_RESERVED;
        // CRC bytes 6,7 - Must be zero for the CRC calculation
        xuidBytes[6] = 0;
        xuidBytes[7] = 0;
        System.arraycopy(identifier, 0, xuidBytes, XUID_PAYLOAD_LEN, 
                identifier.length);
        // Set the actual length
        xuidBytes[XUID_LENGTH_INDEX] = 
                (byte)(XUID_PAYLOAD_LEN + identifier.length); 
        // Calculate the CRC
        int crc_value = crc16.getCrc(xuidBytes);
        // Insert CRC value into bytes 6,7
        xuidBytes[6] = (byte) ((crc_value & 0xff00) >> 8);
        xuidBytes[7] = (byte) (crc_value & 0xff);
        checkCrc(xuidBytes);    // double check crc
        
        return xuidBytes;
    	
    }
       
    private static byte[] intToByteArray(int value)
    {
        byte[] b = new byte[4];
        for (int i = 0; i < 4; i++) {
            int offset = (b.length - 1 - i) * 8;
            b[i] = (byte) ((value >>> offset) & 0xFF);
        }
        return b;
    }
    
    private static byte[] longToByteArray(long value)
    {
        byte[] b = new byte[8];
        for (int i = 0; i < 8; i++) {
            int offset = (b.length - 1 - i) * 8;
            b[i] = (byte) ((value >>> offset) & 0xFF);
        }
        return b;
    }
    
    /**
     * @see XUID#toBytes()
     * @return
     */
    public byte[] toBytes() {
        byte[] byteArray = new byte[data.length];
        System.arraycopy(data, 0, byteArray, 0, byteArray.length);
        return byteArray;
    }

    /**
     * @see XUID#toString()
     */
    public String toString()
    {
        return Base64.base64Encode(data); 
    }
    
    /**
     * Encodes a XUID as a base64 (RFC4648 table 2 variant) string.
     * @return Filename and URI safe string.
     */
    public String toSafeString()
    {
       return Base64.safe64Encode(data);
    }
    
    static XUID createFromSafeString( String safeString )
       throws InvalidXUIDException
    {
       byte values[] = Base64.safe64Decode(safeString);
       return new SniaXUID(values, values.length);
    }
    
    /**
     * Utility function to convert XUIDs to safe strings.
     * @param x
     * @return
     */
    static String convertToSafeString( XUID x )
    {
       return Base64.safe64Encode(x.toBytes());
    }
    
    /*
     * Pads the XUID byte array with zeros to ensure the XUID length is always
     * 80 bytes (i.e. XUID_MAX_LEN) 
     */
    protected byte[] toPaddedBytes() {
        byte[] xuidMaxLengthBuffer = new byte[XUID.MAX_LENGTH];
        byte[] xuidBytes = this.toBytes();
        // gets the length of the XUID from the 6th byte of the XUID
        int xuidLength = 0xff & xuidBytes[XUID_LENGTH_INDEX];
        // copy xuid bytes into MAX_XUID_LEN buffer
        System.arraycopy(xuidBytes, 0, xuidMaxLengthBuffer, 0, xuidLength);
        return xuidMaxLengthBuffer;
    }


    /**
     * @see XUID#equals(java.lang.Object)
     */
    public boolean equals(Object obj)
    {
        if (obj instanceof XUID) {
            return toString().equals(((XUID)obj).toString());
        }
        return false;
    }
    
    /**
     * Generate a "unique" XUID.
     * Because of this implementation, there is no guarantee that a duplicate XUID value
     * might not be generated elsewhere, but it is likely to be universally unique.
     * 
     * @return XUID a unique XUID
     * @throws org.snia.xam.XAMException
     */
    public synchronized static XUID generateXUID()
    throws XAMException {
        long time = 0;
        long randNum = 0;
        long randNum2 = 0;

        synchronized (SniaXUID.class) {
            Calendar calendar = Calendar.getInstance();
            time = calendar.getTimeInMillis();
            // To help with time stamp collisions
            offset = (offset < Integer.MAX_VALUE) ? (offset + 1) : 1;
            UUID id = UUID.randomUUID();
            randNum  = id.getMostSignificantBits();
            randNum2 = id.getLeastSignificantBits();
        } // end synchronized
        
        byte[] timeBytes = Long.toString(time).getBytes();
        int timeLength = timeBytes.length;
        // Use this method to eliminate leading zeros in the offset
        byte[] offsetBytes = BigInteger.valueOf(offset).toByteArray();
        int offsetLength = offsetBytes.length;
        
        byte[] randBytes = longToByteArray(randNum);
        int randLength = randBytes.length;
        
        byte[] randBytes2 = longToByteArray(randNum2);
        int randLength2 = randBytes2.length;
        
        SniaXUID xuid = new SniaXUID();
        byte[] hashBytes = intToByteArray(
                System.identityHashCode(xuid));
        int hashLength = hashBytes.length;
        
        byte[] identifier = new byte[
                timeLength + offsetLength + randLength + randLength2 + hashLength];
        int pos = 0;
        System.arraycopy(timeBytes, 0, identifier, pos, timeLength);
        pos += timeLength;
        System.arraycopy(
                offsetBytes, 0, identifier, pos, offsetLength);
        pos += offsetLength;
        
        System.arraycopy(
                randBytes, 0, identifier, pos, randLength);
        pos += randLength;
        
        System.arraycopy(
           randBytes2, 0, identifier, pos, randLength2);
        pos += randLength;
        
        System.arraycopy(
                hashBytes, 0, identifier, pos, hashLength);
        xuid.data = createXUID(identifier);
        return xuid;
    }
    

    
    /**
     * Create a XUID instance given the XUID string. The String is a
     * valid XUID that has been encoded int base64.
     * TODO:  need to validate the xuid passed in?
     * @param xuidString valid base64 encoded XUID
     * @return XUID a XUID instance representing the XUID value passed in.
     * @throws org.snia.xam.XAMException
     */
    protected synchronized static XUID getInstance(String xuidString)
    throws InvalidXUIDException
    {
        try {
            byte[] decodedXUID =  Base64.base64Decode(xuidString);
            return getInstance(decodedXUID);
        } catch (InvalidXUIDException xex) {
            throw xex;
        } catch (Exception ex) {
            throw new InvalidXUIDException("Invalid XUID found for XSet.", ex);
        }
    }

    /**
     * Create a XUID instance that represents the XUID value passed in.
     * TODO:  need to validate the xuid passed in?
     * @return XUID a XUID instance for the XUID value passed in.
     * @throws org.snia.xam.XAMException
     */
    protected synchronized static XUID getInstance(byte[] xuidBytes)
    throws InvalidXUIDException
    {
        try {
            if (xuidBytes.length > (XUID.MAX_LENGTH)) {
                throw new InvalidXUIDException("Specified xuid is too long.");
            }
            if (xuidBytes.length < XUID.MIN_LENGTH) {
                throw new InvalidXUIDException("Specified xuid is too short." +
                    " Must be at least " + XUID.MIN_LENGTH + " bytes long");
            }
            
            int length = xuidBytes[XUID_LENGTH_INDEX];
            if (length > xuidBytes.length || length > XUID.MAX_LENGTH) {
                throw new InvalidXUIDException(
                        "XUID length is incorrect. Internal: " + length +
                        "  Array length: " + xuidBytes.length);
            }
            
            checkCrc(xuidBytes);
            SniaXUID xuidInstance = new SniaXUID(xuidBytes, length);
            return xuidInstance;
            
        } catch (Exception ex) {
            throw new InvalidXUIDException("Invalid XUID found for XSet.", ex);
        }
    }

//    /**
//     *  Converts the Hex String to a XUID.
//     *  The hex string is used in filenames to avoid
//     *  problems with illegal file name characters that
//     *  are possible given the characterset used by the
//     *  base64 encoder when generating XUIDs.
//     * 
//     *  @param hex hex string to map to a XUID value
//     *  @return XUID the xuid value
//     *  @throws XAMException
//     */
//    protected static XUID mapHexToXUID(String hex) throws XAMException {
//        int byteIdx = 0;
//        byte[] bytes = new byte[hex.length()/2];
//        for (int idx = 0; idx < hex.length(); idx += 2, byteIdx++) {
//            String tmp = hex.substring(idx, idx + 2);
//            bytes[byteIdx] = Byte.parseByte(tmp, 16);
//        }
//        return getInstance(new String(bytes));
//    }
//    
//    /**
//     *  Converts the XUID to a Hex String.
//     *  This is string is used in filenames to avoid
//     *  problems with illegal file name characters that
//     *  are possible givent the characterset used by the
//     *  base64 encoder when generating XUIDs.
//     * 
//     *  @param xuid XUID value to map to a hex string
//     *  @return String the xuid value as a string of hex digits
//     */
//    protected static String mapXUIDToHex(XUID xuid) {
//        byte[] xuidBytes = xuid.toString().getBytes();
//        StringBuffer hexString = new StringBuffer();
//        for (byte b : xuidBytes) {
//            hexString.append(String.format("%02X", b));
//        }
//        return hexString.toString();
//    }
//    
    protected static void checkCrc(byte[] bytes)
        throws InvalidXUIDException {
        byte msb = bytes[CRC_INDEX];
        byte lsb = bytes[CRC_INDEX + 1];
        int crc = ((msb & 0xff) << 8) | (lsb & 0xff);

        byte length = bytes[XUID_LENGTH_INDEX];

        try {
            bytes[CRC_INDEX] = 0;
            bytes[CRC_INDEX + 1] = 0;
        if (crc != crc16.getCrc(bytes, 0, length))
            throw new InvalidXUIDException("CRC check failed");
        }
        finally {
            bytes[CRC_INDEX] = msb;
            bytes[CRC_INDEX + 1] = lsb;
        }
    }
   
    public static void main(String[] args) {
        try {
           for( int i = 0; i < 10; i++ )
           {
              XUID x1 = generateXUID();
              System.out.println("Gen XUID: " + x1.toString() + " " + ( (SniaXUID) x1).toSafeString() );
           }
        } catch (XAMException ex) {
            ex.printStackTrace();
        }
        
    }
}
