/*
 *  Copyright (c) 2008-2009, Sun Microsystems, Inc.
 *  Copyright (c) 2008-2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import org.snia.xam.InvalidArgumentException;
import org.snia.xam.ObjectInUseException;
import org.snia.xam.XAMException;
import org.snia.xam.XSet;
import org.snia.xam.XSetAbandonException;
import org.snia.xam.XSetCorruptException;
import org.snia.xam.XSetException;

enum XSetOperation {
    ABANDON,
    COMMIT,
    CLOSE,
    
    CREATE_XSET,
    CREATE_XSTREAM,
    OPEN_EXPORT_XSTREAM,
    OPEN_IMPORT_XSTREAM,
    OPEN_XSTREAM_READONLY,
    OPEN_XSTREAM_WRITE_BINDING,
    OPEN_XSTREAM_WRITE_NONBINDING,
    CREATE_PROPERTY_BINDING,
    CREATE_PROPERTY_NONBINDING,
    SET_PROPERTY_BINDING,
    SET_PROPERTY_NONBINDING,
    GET_PROPERTY,
    
    DELETE_FIELD_BINDING,
    DELETE_FIELD_NONBINDING,
    GET_FIELD_INFO,
    CHANGE_FIELD_BINDING,
    SET_FIELD_NONBINDING
 
}
/**
 * Typesafe enum class that can be supported by JDK 1.4.
 * This class construct can be replaced in JDK 1.5 by the Java
 * Enum calss
 */
public class XSetState extends OperationalState {
    /**
     * Opened in Read only mode
     */
    public static final XSetState READER = new XSetState("READER");
    
    /**
     * Opened in Unrestricted mode
     */
    public static final XSetState UNRESTRICTED_WRITER = 
            new XSetState("UNRESTRICTED_WRITER");
    
    /**
     * Opened in Restricted mode
     */
    public static final XSetState RESTRICTED_WRITER = 
            new XSetState("RESTRICTED_WRITER");
        
    /**
     * The state that indicates the XSet is being exported.
     */
    public static final XSetState EXPORT = new XSetState("EXPORT");
    
    /**
     * The state the indicates the XSet is being imported.
     */
    public static final XSetState IMPORT = new XSetState("IMPORT");
    
    /**
     * The state that indicates the stream is corrupt
     */
    public static final XSetState CORRUPT = new XSetState("CORRUPT");
    
    /**
     * The state that indicates the stream has been abandoned
     */
    public static final XSetState ABANDONED = new XSetState("ABANDONED");
    
    /**
     * The state that indicates the stream has been closed.  No further 
     * action on the stream is possible
     */
    public static final XSetState CLOSED = new XSetState("CLOSED");

    /**
     * Creates new XSetState object. Private access. Do not allow creation
     * of other XSetState object instances.
     * 
     * @param state
     */
    private XSetState(String state) {
        super(state);
    }
    
    /**
     * Returns the correct XSetState object based on the open mode for
     * the XSet.
     * 
     * @param openMode
     * @return the XSetState object that represents the specified openMode
     * @throws InvalidArgumentException
     */
    public static XSetState getXSetState(String openMode)
        throws InvalidArgumentException {
        
        XSetState state = null;
        if (openMode.equals(XSet.MODE_READ_ONLY)) {
            state = READER;
        } else if (openMode.equals(XSet.MODE_RESTRICTED)) {
            state = RESTRICTED_WRITER;
        } else if (openMode.equals(XSet.MODE_UNRESTRICTED)) {
            state = UNRESTRICTED_WRITER;
        } else {
            throw new InvalidArgumentException("Unrecognized open mode: " +
                    openMode);
        }
        return state;
    }
    /**
     * Completes successfully if operation is allowed given current state
     * of XSet. Throws exception if operation is not allowed.
     * 
     * @param currentState State of XSet
     * @param operation Operation to be attempted
     * @throws org.snia.xam.XAMException
     */
    protected static void validateOperation(XSetState currentState, 
            XSetOperation operation) 
            throws XSetException, XSetCorruptException,
                ObjectInUseException, XAMException {
        String errorMsg = "";
       
        /*
         * Check basic XSet state to see if any access is allowd
         */
        if (currentState == null) {
            if (operation.equals(XSetOperation.CREATE_XSET)) {
                return;
            } else {
                errorMsg = "Invalid operation, XSet does not exist";
            }
        } else if (currentState.equals(CLOSED)) {
            throw new XSetException(
                    "XSet is closed, no access allowed");
        } else if (currentState.equals(CORRUPT)) {
            if (operation.equals(XSetOperation.ABANDON)) {
                return;
            } else {
                throw new XSetCorruptException(
                        "Invalid operation specified on corrupt XSet.");
            }
        } else if (currentState.equals(ABANDONED)) {
            if (operation.equals(XSetOperation.CLOSE)) {
                return;
            } else {
                throw new XSetAbandonException(
                        "Invalid operation specified on abandoned XSet.");
            }
        }

        /*
         * If reach here know that the state is not abandoned, closed or
         * corrupt
         */
        switch (operation) {
            case COMMIT:
                if (currentState.equals(READER)) {
                    errorMsg = "XSet is open for read only access, " +
                            "cannot commit XSet.";
                } else {
                    return;
                }
                break;
                
            case ABANDON :
                return;

            case CLOSE :
                // check for open XStreams else where
                return;
                
            case OPEN_EXPORT_XSTREAM :
            case OPEN_IMPORT_XSTREAM :
                if (currentState.equals(EXPORT) || 
                        currentState.equals(IMPORT)) {
                    throw new ObjectInUseException("XSet has open import or " +
                            "export xstream.");
                }
                
                if (operation == XSetOperation.OPEN_IMPORT_XSTREAM) {
                    if (currentState.equals("READER")) {
                        errorMsg = "Must open XSet for write access to import " +
                            "XSet";
                    } else {
                        return;
                    }
                } else {
                    return;
                }
                break;
                
            case OPEN_XSTREAM_READONLY :
            case GET_PROPERTY :
            case GET_FIELD_INFO :
                if (currentState.equals(READER) ||
                        currentState.equals(RESTRICTED_WRITER) ||
                        currentState.equals(UNRESTRICTED_WRITER) ||
                        currentState.equals(IMPORT)) {
                    return;
                } else {
                    errorMsg = "Read access is not allowed on XSet.";
                }
                break;
            case OPEN_XSTREAM_WRITE_BINDING :
            case CREATE_PROPERTY_BINDING :
            case SET_PROPERTY_BINDING :
            case CHANGE_FIELD_BINDING :
            case DELETE_FIELD_BINDING :
                if (currentState.equals(UNRESTRICTED_WRITER)) {
                    return;
                } else {
                    errorMsg = 
                            "XSet is open in read only or restricted mode, " +
                            "adding, deleting or changing binding fields is " +
                            "not allowed";
                }
                break; 
            case OPEN_XSTREAM_WRITE_NONBINDING :
            case CREATE_PROPERTY_NONBINDING :
            case SET_PROPERTY_NONBINDING :
            case DELETE_FIELD_NONBINDING :
                if (currentState.equals(READER)) {
                    errorMsg = "XSet is open for read only, adding, deleting " +
                            "or changing fields is not allowed";
                } else {
                    return;
                }
                break;
            default :
                errorMsg = "XSet access does not allow operation";
        } // end switch

        throw new XSetException(errorMsg);
    }
}
