/*
 *  Copyright (c) 2008, Sun Microsystems, Inc.
 *  Copyright (c) 2008, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.io.File;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;

import javax.xml.parsers.SAXParser;
import org.snia.xam.XUID;
import org.snia.xam.base.Field;
import org.snia.xam.base.Property;
import org.snia.xam.util.ISO8601Date;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import javax.xml.parsers.SAXParserFactory;
import org.xml.sax.helpers.DefaultHandler;


import org.snia.xam.XAMLibrary;
import org.snia.xam.XSet;
import org.snia.xam.XAMException;

/**
 * Parses and loads persisted XSet for Reference implementation.
 * Used to parse & load both XSets that already exist in the XSystem and
 * XSets being imported. This is possible because the Reference VIM persists
 * XSets in the XAM defined XSet interchange format.
 */
public class XSetXmlReader extends DefaultHandler
{

    final static SAXParserFactory factory = SAXParserFactory.newInstance();
    // parser state
    boolean inPolicies = false;
    boolean inPolicy = false;
    boolean inXSets = false;
    boolean inXSet = false;
    boolean inXStreams = false;
    boolean inXStream = false;
    boolean inProperties = false;
    boolean inProperty = false;
    boolean inString = false;
    boolean inDatetime = false;
    boolean inBoolean = false;
    boolean inInteger = false;
    boolean inDouble = false;
    boolean inXUID = false;
    boolean inXStreamData = false;
    boolean inCharacters = false;
    boolean inXStreamXOPInclude = false;
    
    // Record the property information until the closing /property
    String name      = null;
    String propType  = null;
    boolean binding  = false;
    boolean readOnly = false;
    
    // Record policy information until the closing /policy
    RetentionPolicyReference                 policy_reference;
    HashMap<String,RetentionPolicyReference> policy_reference_map = null;
    
    // Record property information until the closing /property
    Property current_property;
    
    // Record values to be used by the closing of properties, or policies.
    Object   value_object;
    String   value_mime_type;
    
    String xopHrefCID = "";
    long length = 0;
    boolean debug = false;
    
    // Can not create system fields outside a FieldContainer so store
    // them here so the parent XSet can create them later
    Map<String, Field> systemFields = new HashMap<String, Field>();

    // Keeps track of elements and character data. Need to accumulate all
    // the characters for a particular element before processing.
    // key is the element name; value is the character data
    HashMap<String, StringBuffer> elementValues = new HashMap<String, StringBuffer>();
    String currentElement = null;

    ReferenceXSet xset = null;
    
    // Contains data on stream contents when loading imported XSet
    // Info from the TOC & Stream Binary Data sections which are read
    // and loaded before processing the XSet XML. 
    XStreamXopData[] xstreamsImportData = null;
    // true if parsing imported XSet 
    boolean importingXSet = false;
    
    XUID xuid = null; // XUID for XSet

    private Logger logger = null;
    
    private XSetXmlReader() {
        xset = new ReferenceXSet();
        
    }

    
    public XSetXmlReader(ReferenceXSet curXSet)
       throws XAMException 
    {
        xset = curXSet;
        logger = curXSet.getXSystem().getLogger();
    }
    
    /**
     * Creates a new XSet XML parser instance.
     * 
     * @param curXSet   the XSet instance to populate with the objects
     * created when parsing the XSet XML
     * @param importing     <code>true</code> if importing an XSet; 
     * <code>false</code> if opening an XSet an existing XSet.
     */
    public XSetXmlReader(ReferenceXSet curXSet, boolean importing) 
       throws XAMException
    {
        this(curXSet);
        importingXSet = importing;
    }

    /**
     * Creates a new XSet XML parser instance. Should only be called when
     * importing an XSet.
     * @param curXSet   the XSet instance to populate with the objects
     * created when parsing the XSet XML 
     * @param xstreamsImportData    imported TOC & streams data
     */
    public XSetXmlReader(ReferenceXSet curXSet, 
            XStreamXopData[] xstreamsImportData) 
       throws XAMException
    {
        this(curXSet, true);
        this.xstreamsImportData = xstreamsImportData;        
    }

    public XSet getXSet() {
        return xset;
    }
    
    public Map getSystemFields() {
        return systemFields;
    }
    
    public void setImportingXSet(boolean importing) {
        importingXSet = importing;
    }
    
    // Helper methods 
    private void addSystemField(Field field) {
        systemFields.put(field.getName(), field);
    }

    private void requireXSets(String tag)
    {
        if (!inXSets) {
            throw new MalformedXmlException(tag +
                    " tag detected outside of " + XmlConstants.TAG_XSETS);
        }
    }

    private void requireXSet(String tag)
    {
        requireXSets(tag);
        if (!inXSet) {
            throw new MalformedXmlException(tag +
                    " tag detected outside of " + XmlConstants.TAG_XSET);
        }
    }

    private void requireXStreams(String tag)
    {
        requireXSet(tag);
        if (!inXStreams) {
            throw new MalformedXmlException(tag +
                    " tag detected outside of " + XmlConstants.TAG_XSTREAMS);
        }
    }

    private void requireXStream(String tag)
    {
        requireXStreams(tag);
        if (!inXStream) {
            throw new MalformedXmlException(tag +
                    " tag detected outside of " + XmlConstants.TAG_XSTREAM);
        }
    }

    private void requireProperties(String tag)
    {
        requireXSet(tag);
        if (!inProperties) {
            throw new MalformedXmlException(tag +
                    " tag detected outside of " + XmlConstants.TAG_PROPERTIES);
        }
    }

    private void requirePropertyOrPolicies(String tag)
    {
       if( !inProperty && !inPolicy )
          throw new MalformedXmlException( tag + " tag detected outside of " +
             XmlConstants.TAG_POLICY + " or " + XmlConstants.TAG_PROPERTY );
    }

    private void requirePropertiesOrPolicies(String tag)
    {
       if( !inProperties && !inPolicy )
          throw new MalformedXmlException( tag + " tag detected outside of " +
             XmlConstants.TAG_POLICY + " or " + XmlConstants.TAG_PROPERTIES );
    }

    public void startElement(String uri,
                             String localName,
                             String qName,
                             Attributes attributes)
    throws SAXException
    {
       try
      {
         if (debug) {
            System.out.println("starting: " + qName);
         }
         currentElement = qName;
         elementValues.put(currentElement, new StringBuffer());
         if (XmlConstants.TAG_XSETS.equals(qName)) {
            if (inXSets) {
               throw new MalformedXmlException("Nested "
                  + XmlConstants.TAG_XSETS + " tag detected");
            }
            inXSets = true;
         } else if (XmlConstants.TAG_POLICIES.equals(qName)) {
            inPolicies = true;
            policy_reference_map = new HashMap<String,RetentionPolicyReference>();
         } else if (XmlConstants.TAG_POLICY.equals(qName)) {
            inPolicy = true;
            policy_reference = new RetentionPolicyReference(
               attributes.getValue(XmlConstants.ATTRIBUTE_NAME),
               attributes.getValue(XmlConstants.ATTRIBUTE_TYPE),
               "", // Fill this in processing the string element
               Boolean.valueOf(attributes.getValue(XmlConstants.ATTRIBUTE_BINDING)),
               Boolean.valueOf(attributes.getValue(XmlConstants.ATTRIBUTE_READONLY)));
         }
         else if (XmlConstants.TAG_STRING.equals(qName)) {
            requirePropertyOrPolicies(qName);
            if (debug) System.out.println("String value: " + qName);
            inString = true;
         }
         else if (XmlConstants.TAG_BOOLEAN.equals(qName)) {
            requirePropertyOrPolicies(qName);
            inBoolean = true;
            if (debug) System.out.println("Boolean value: " + qName);
         }
         else if (XmlConstants.TAG_DATE.equals(qName))  {
            requirePropertyOrPolicies(qName);
            if (debug) System.out.println("Datetime value: " + qName);
            inDatetime = true;
         }
         else if (XmlConstants.TAG_INTEGER.equals(qName)) {
            requirePropertyOrPolicies(qName);
            inInteger = true;
            if (debug) System.out.println("Integer value: " + qName);
         }
         else if (XmlConstants.TAG_DOUBLE.equals(qName))  {
            requirePropertyOrPolicies(qName);
            inDouble = true;
            if (debug) System.out.println("Double value: " + qName);
         }
         else if (XmlConstants.TAG_XUID.equals(qName)) {
            requirePropertyOrPolicies(qName);
            inXUID = true;
            if (debug) System.out.println("XUID value: " + qName);
         }
         else if (XmlConstants.TAG_PROPERTY.equals(qName))  {
            requirePropertiesOrPolicies(qName);
            inProperty = true;
            binding = Boolean.parseBoolean(attributes.getValue(XmlConstants.ATTRIBUTE_BINDING));
            name = attributes.getValue(XmlConstants.ATTRIBUTE_NAME);
            propType = attributes.getValue(XmlConstants.ATTRIBUTE_TYPE);
            readOnly = Boolean.parseBoolean(attributes.getValue(XmlConstants.ATTRIBUTE_READONLY));
            current_property = new Property( attributes.getValue(XmlConstants.ATTRIBUTE_NAME),
                                             XAMLibrary.STYPE_STRING_MIME_TYPE, // type to come later
                                             "",                                // value to come later
                                             Boolean.valueOf(attributes.getValue(XmlConstants.ATTRIBUTE_BINDING)),
                                             Boolean.valueOf(attributes.getValue(XmlConstants.ATTRIBUTE_READONLY)) );
            if (debug) System.out.println("Property name: " + name);
         }
         
         if (!inPolicies)  {
            if (XmlConstants.TAG_XSET.equals(qName))  {
               requireXSets(qName);
               if (inXSet) {
                  throw new MalformedXmlException("Nested "
                     + XmlConstants.TAG_XSET + " tag detected");
               }
               inXSet = true;
            }  
            else if (XmlConstants.TAG_XSTREAMS.equals(qName)) {
               requireXSet(qName);
               inXStreams = true;
            }  
            else if (XmlConstants.TAG_XSTREAM.equals(qName)) {
               requireXStreams(qName);
               inXStream = true;
               binding = Boolean.parseBoolean(attributes.getValue(XmlConstants.ATTRIBUTE_BINDING));
               name = attributes.getValue(XmlConstants.ATTRIBUTE_NAME);
               propType = attributes.getValue(XmlConstants.ATTRIBUTE_TYPE);
               readOnly = Boolean.parseBoolean(attributes.getValue(XmlConstants.ATTRIBUTE_READONLY));
               length = Long.parseLong(attributes.getValue(XmlConstants.ATTRIBUTE_LENGTH));
               if (debug) System.out.println("Stream Field name: " + name);
            }  
            else if (XmlConstants.TAG_PROPERTIES.equals(qName)) {
               requireXSet(qName);
               inProperties = true;
            } 
            else if (XmlConstants.TAG_XSTREAM_DATA.equals(qName)) {
               requireXStream(qName);
               inXStreamData = true;
            }
            else if (XmlConstants.TAG_XSTREAM_XOP_INCLUDE.equals(qName)) {
               requireXStream(qName);
               inXStreamXOPInclude = true;
               xopHrefCID = attributes.getValue(XmlConstants.ATTRIBUTE_HREF);
            }
         }
      }
      catch (Exception e)
      {
         throw new SAXException(
            "Unexpected exception while processing start tag: " + qName, e);
      }
   }

    public void characters(char[] ch, int start, int length)
    {
        inCharacters = true;
        if (debug) {
            System.out.println("\"" + new String(ch, start, length) + "\"");
        }
        
        String value = new String(ch, start, length);
        if( !inString && value.trim().equals("")) {
            // ignore
        } else {

            /* Add characters to the current element. Processing will occur
             * when the end element tag is reached. Need to do this because
             * the parser delivers only a chunk of characters at a time. 
             * All characters for a particular element may be in one chunk
             * or split into several chunks.
             */
           ((StringBuffer)(elementValues.get(currentElement))).append(value);
        }

    }

    public void endElement(String uri,
                           String localName,
                           String qName)
    throws SAXException
    {
        if (debug) {
            System.out.println("leaving element:" + qName);
        }
        
        try {
        String value = ((StringBuffer) (elementValues.get(qName))).toString();
         elementValues.remove(qName);
         /*
          * We do not reset the current element here as the startElement method
          * should do that at the start of an element
          * 
          */
         // Process character data now have hit the end element tag
         if (inCharacters)  {
            processCharacters(qName, value);
            inCharacters = false;
         }
         if (XmlConstants.TAG_XSETS.equals(qName))  {
            requireXSets(qName);
            inXSets = false;
         }
         else if (XmlConstants.TAG_XSET.equals(qName)) {
            requireXSet(qName);
            inXSet = false;
         }
         else if (XmlConstants.TAG_POLICIES.equals(qName)) {
            inPolicies = false;
            xset.setRetentionPolicyReferences(policy_reference_map);
            policy_reference = null;
         }
         else if (XmlConstants.TAG_POLICY.equals(qName)) {
            inPolicy = false;
            policy_reference_map.put( policy_reference.getName(), policy_reference );
         }
         else if (XmlConstants.TAG_XSTREAMS.equals(qName)) {
            inXStreams = false;
         }
         else if (XmlConstants.TAG_XSTREAM.equals(qName)) {
            name = null;
            inXStream = false;
         }
         else if (XmlConstants.TAG_PROPERTIES.equals(qName))  {
            inProperties = false;
         }
         else if (XmlConstants.TAG_PROPERTY.equals(qName)) {
            name = null;
            inProperty = false;
         }
         else if (XmlConstants.TAG_BOOLEAN.equals(qName)) {
            inBoolean = false;
            completeValue();
         }
         else if (XmlConstants.TAG_STRING.equals(qName)) {
            inString = false;
            completeValue();
         }
         else if (XmlConstants.TAG_DATE.equals(qName)) {
            inDatetime = false;
            completeValue();
         }
         else if (XmlConstants.TAG_DOUBLE.equals(qName)) {
            inDouble = false;
            completeValue();
         }
         else if (XmlConstants.TAG_INTEGER.equals(qName)) {
            inInteger = false;
            completeValue();
         }
         else if (XmlConstants.TAG_XUID.equals(qName)) {
            inXUID = false;
            completeValue();
        }
         else if (XmlConstants.TAG_XSTREAM_DATA.equals(qName)) {
            inXStreamData = false;
         }
         else if (XmlConstants.TAG_XSTREAM_XOP_INCLUDE.equals(qName)) {
            inXStreamXOPInclude = false;
         }
      }
      catch (Exception e)
      {
         throw new SAXException("Error closing element " + qName, e);
      }       
    }

    /**
     * Completes the defintion of a value. It might be a value for a property or a policy.
    * @throws XAMException 
     */
    private void completeValue() throws XAMException
    {
       if( inPolicy && !inProperty )
          policy_reference.setValue(value_object, value_mime_type);
       else if( inPolicy && inProperty )
       {
          current_property.setValue(value_object, value_mime_type);
          policy_reference.addReference(current_property);
          current_property = null;
       }
       else if (inProperty)
          addProperty( value_object, value_mime_type );
    }


   /**
     * Called when it is appropriate to add a property.
     * @param value The boxed value (String, Long, Double, etc).
     * @param mime_type The MIME type of the last value
     * @throws XAMException
     */
    private void addProperty(Object value, String mime_type) throws XAMException
    {
       current_property.setValue(value, mime_type);
       boolean systemField = current_property.getName().startsWith(".");
       if(systemField)
          addSystemField(current_property);
       else
          xset.addField(current_property);
       
       // As a side effect, make sure this gets recorded.
       if( current_property.getName().equals(XSet.XAM_XUID))
          xuid = (XUID) current_property.getValue();
       
       current_property = null;
    }


   /**
     * Processes the character data received. This method should only be
     * called once the end tag is reached for an element and there has been
     * character data received for that element.
     * 
     * @param qName element name
     * @param value character data to process
     */
    protected void processCharacters(String qName, String value) {
        try {
           boolean systemField = (name == null || name.startsWith(".") ?
                        true : false);
            if (inBoolean) {
               value_object = Boolean.valueOf(value);
               value_mime_type = XAMLibrary.STYPE_BOOLEAN_MIME_TYPE;

            } else if (inDatetime) {
               value_object= ISO8601Date.toCalendar(value);
               value_mime_type = XAMLibrary.STYPE_DATETIME_MIME_TYPE;

            } else if (inString) {
               value_object = value;
               value_mime_type = XAMLibrary.STYPE_STRING_MIME_TYPE;
               if( inPolicy ) {
                  if (debug)
                     System.out.println("Setting policy reference: " + value );
                  policy_reference.setValue( value, policy_reference.getType() );
               }
           } else if (inInteger) {
               value_object = Long.valueOf(value);
               value_mime_type = XAMLibrary.STYPE_INT_MIME_TYPE;

            } else if (inDouble) {
               value_object = Double.valueOf(value);
               value_mime_type = XAMLibrary.STYPE_DOUBLE_MIME_TYPE;

           } else if (inXUID) {
              value_object = SniaXUID.getInstance(value);
              value_mime_type = XAMLibrary.STYPE_XUID_MIME_TYPE;

           }
           else if (inXStreamData && !importingXSet) {
              String fileLocation = xset.getXSystem().getPersistenceMgr().
                                        getStreamFieldAbsolutePath(xuid, value);
              if (debug)
                 System.out.println("Stream Name: " + fileLocation);
           
              ReferenceStreamField rField =
                   new ReferenceStreamField(
                   name, propType, length, binding, readOnly, 
                   fileLocation, logger);
              xset.addField(rField);

           } else if (inXStreamXOPInclude && importingXSet) {
              String dataLocation = getXStreamDataLocation(name);
              ReferenceStreamField rField =
                   new ReferenceStreamField(
                   name, propType, length, binding, readOnly, dataLocation, logger );
              xset.addField(rField);
           } 
        } catch (NumberFormatException ex) {
          throw new RuntimeException("Failed to set value " +
             value +
             " for " + name + ": " + ex, ex);
       } catch (XAMException xe) {
          throw new RuntimeException("Failed to set value " +
             value +
             " for " + name + ": " + xe, xe);
       } catch (ParseException pe ) {
          throw new RuntimeException("Failed to parse date " +
             value + " for " + name + ": " + pe, pe );
       }
    }
    
    private String getXStreamDataLocation(String fieldName) 
            throws XAMException {
        String xstreamLocation = "";
        if (importingXSet) {
            int index = XStreamXopData.findFieldEntry(fieldName, 
                    xstreamsImportData);
            
            if (index != -1) {
                xstreamLocation = xstreamsImportData[index].tmpFilename;
            } else {
                throw new XAMException("Error parsing XSet XML. " +
                        "Missing data for XStream field: " + fieldName);
            }
        } 
        return xstreamLocation;
    }
}
