/*
 *  Copyright (c) 2008-2009, Sun Microsystems, Inc.
 *  Copyright (c) 2008-2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;

import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XSystem;
import org.snia.xam.XIterator;
import org.snia.xam.XSet;
import org.snia.xam.base.Field;
import org.snia.xam.base.Property;
import org.snia.xam.util.ISO8601Date;
import org.xml.sax.helpers.DefaultHandler;

/**
 *  Note: Work in Progress - not final implementation
 * 
 * Persists the XSet as an XML file.
 */
public class XSetXmlWriter extends DefaultHandler {
    
    private final static String INDENT = "   ";
    
    public static void outputXMLToFile(ReferenceXSet xset) throws XAMException {
        ReferenceVIMPersistenceMgr persistenceMgr = 
                xset.getXSystem().getPersistenceMgr();
        String filename = persistenceMgr.getXMLFileName(xset.getXUID());
        String dirname = persistenceMgr.getXSetDirName(xset.getXUID());
        outputXMLToFile(filename, xset, dirname);
    }
    
    private static void outputXMLToFile(String filename, ReferenceXSet xset,
            String dirname)
    throws XAMException {
        try {
            Writer out
                = new BufferedWriter(
                    new OutputStreamWriter(
                        new FileOutputStream(filename), XAMConstants.UTF_8));
            out.write(XmlConstants.XML_HEADER);
            out.write("\n");
            startElement(out, XmlConstants.TAG_XSETS);
            writeAttribute(out, "xsi:schemaLocation",
                    "http://www.snia.org/2007/xam/export/XAMCanonicalXSetDefinition.xsd");
            writeAttribute(out, "xmlns", "http://www.snia.org/2007/xam/export");
            writeAttribute(out, "xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance");
            writeAttribute(out, "xmlns:xop", "http://www.w3.org/2004/08/xop/include");
            //endElement(out);
            out.write(">\n");
            startElement(out, XmlConstants.TAG_VERSION);
            out.write(">");
            out.write("1.0.0");
            endElement(out, XmlConstants.TAG_VERSION);
            
            startElement(out, XmlConstants.TAG_POLICIES);
            out.write(">\n");
            processXSetPolicies(out, INDENT, xset);
            endElement(out, XmlConstants.TAG_POLICIES);
            
            startElement(out, XmlConstants.TAG_XSET);
            out.write(">\n");
            
            startElement(out, XmlConstants.TAG_PROPERTIES);
            out.write(">\n"); //endElement(out);
            processXSetProperties(out, INDENT, xset);
            endElement(out, XmlConstants.TAG_PROPERTIES);
            
            startElement(out,XmlConstants.TAG_XSTREAMS);
            out.write(">\n");
            processXSetXStreams(out, xset, dirname);
            endElement(out,XmlConstants.TAG_XSTREAMS);

            endElement(out, XmlConstants.TAG_XSET);
            endElement(out, XmlConstants.TAG_XSETS);
            out.flush();
            out.close();
        }
        catch (IOException io) {
            throw new XAMException(io);
        }
    }
    
    /**
     * Write each policy element out.
     * @param out
    * @param indent 
     * @param xset
     * @throws IOException 
     * @throws XAMException 
     */
    private static void processXSetPolicies(Writer out, String indent, ReferenceXSet xset) throws IOException, XAMException
    {
       
       // Top level management property
       Property mgmtPolicyField = (Property) xset.getField( XSet.XAM_MANAGEMENT_POLICY );
       String context = XSystem.XAM_MANAGEMENT_POLICY_LIST + mgmtPolicyField.getValue();
       writePolicy( out, indent, context, mgmtPolicyField );
       
       // Retention Policies
       Iterator<RetentionPolicyReference> policyIter = RetentionMgr.GetPolicyReferences(xset).iterator();
       while( policyIter.hasNext() )
       {
          RetentionPolicyReference policyRef = (RetentionPolicyReference) policyIter.next();
          writePolicy( out, indent, policyRef );
       }
       
       // Disposition Policies
       if( xset.containsField( XSet.XAM_DELETION_AUTODELETE_POLICY ))
       {
          Property autodeleteProperty = (Property) xset.getField( XSet.XAM_DELETION_AUTODELETE_POLICY );
          context = XSystem.XAM_DELETION_AUTODELETE_POLICY_LIST + autodeleteProperty.getValue();
          writePolicy( out, indent, context, mgmtPolicyField );
       }

       if( xset.containsField( XSet.XAM_DELETION_SHRED_POLICY ))
       {
          Property autodeleteProperty = (Property) xset.getField( XSet.XAM_DELETION_SHRED_POLICY );
          context = XSystem.XAM_DELETION_SHRED_POLICY_LIST + autodeleteProperty.getValue();
          writePolicy( out, indent, context, mgmtPolicyField );
       }
    }

    /**
     * Write out a single policy
     * @param out Where to write
    * @param context 
     * @param prop The list property referring to the property.
     * @throws IOException
     */
    private static void writePolicy(Writer out, String indent, String context, Property prop ) throws IOException
    {
       String indent2 = indent + indent;
       
       out.write( indent );
       startElement( out, XmlConstants.TAG_POLICY );
       writeAttribute( out, XmlConstants.ATTRIBUTE_NAME,     context );
       writeAttribute( out, XmlConstants.ATTRIBUTE_TYPE,     prop.getType() );
       writeAttribute( out, XmlConstants.ATTRIBUTE_READONLY, Boolean.toString(prop.getReadOnly()) );
       writeAttribute( out, XmlConstants.ATTRIBUTE_BINDING,  Boolean.toString(prop.getBinding()) );
       writeAttribute( out, XmlConstants.ATTRIBUTE_LENGTH,   Long.toString(prop.getLength()) );
       out.write( ">\n" );
      
       out.write( indent2 );
       startElement( out, XmlConstants.TAG_STRING );
       out.write( ">" );
       out.write( prop.getValue().toString() );
       endElement( out, XmlConstants.TAG_STRING );
      
       out.write( indent );
       endElement( out, XmlConstants.TAG_POLICY );
    }

    /**
     * Write out a single policy, with referring properties.
     * @param out Where to write
     * @param policy The list property referring to the property.
     * @throws IOException
     */
    private static void writePolicy(Writer out, String indent, RetentionPolicyReference policy ) throws IOException
    {
       String indent2 = indent + indent;
       
       out.write( indent );
       startElement( out, XmlConstants.TAG_POLICY );
       writeAttribute( out, XmlConstants.ATTRIBUTE_NAME,     policy.getName() );
       writeAttribute( out, XmlConstants.ATTRIBUTE_TYPE,     policy.getType() );
       writeAttribute( out, XmlConstants.ATTRIBUTE_READONLY, Boolean.toString(policy.getReadOnly()) );
       writeAttribute( out, XmlConstants.ATTRIBUTE_BINDING,  Boolean.toString(policy.getBinding()) );
       writeAttribute( out, XmlConstants.ATTRIBUTE_LENGTH,   Long.toString(policy.getLength()) );
       out.write( ">\n" );
      
       out.write( indent2 );
       startElement( out, XmlConstants.TAG_STRING );
       out.write( ">" );
       out.write( policy.getValue().toString() );
       endElement( out, XmlConstants.TAG_STRING );
      
       // Write out the referring properties
       HashMap<String,Property> map = policy.getReferences();
       Iterator<String> propIter = map.keySet().iterator();
       while( propIter.hasNext() )
       {
          String propName = propIter.next();
          Property prop = map.get(propName);
          writeProperty( out,
             indent2,
             prop.getName(), 
             prop.getType(), 
             prop.getBinding(), 
             prop.getReadOnly(), 
             prop.getLength(), 
             prop.getValue() );
       }
       out.write( indent );
       endElement( out, XmlConstants.TAG_POLICY );
    }

   /**
     * Locates occurrences of special XML characters (like '<', '&', etc.) 
     * in a string and replaces them by sequences of the form &X...X;
     *
     * In addition, if replaceWhitechars parameter is true, '\n', '\r' and '\t' 
     * are replaced by &#NN; sequences. 
     */
    private static String replaceSpecialChars(String s, boolean replaceWhitechars)
    {
        int length = s.length();
        char[] chars = new char[6 * length];
        int n = 0;
        for (int x = 0; x < length; x++) {
            char c = s.charAt(x);
            switch (c) {
                case '&':
                    chars[n] = '&';
                    n++;
                    chars[n] = 'a';
                    n++;
                    chars[n] = 'm';
                    n++;
                    chars[n] = 'p';
                    n++;
                    chars[n] = ';';
                    n++;
                    break;
                case '\'':
                    chars[n] = '&';
                    n++;
                    chars[n] = 'a';
                    n++;
                    chars[n] = 'p';
                    n++;
                    chars[n] = 'o';
                    n++;
                    chars[n] = 's';
                    n++;
                    chars[n] = ';';
                    n++;
                    break;
                case '\"':
                    chars[n] = '&';
                    n++;
                    chars[n] = 'q';
                    n++;
                    chars[n] = 'u';
                    n++;
                    chars[n] = 'o';
                    n++;
                    chars[n] = 't';
                    n++;
                    chars[n] = ';';
                    n++;
                    break;
                case '<':
                    chars[n] = '&';
                    n++;
                    chars[n] = 'l';
                    n++;
                    chars[n] = 't';
                    n++;
                    chars[n] = ';';
                    n++;
                    break;
                case '>':
                    chars[n] = '&';
                    n++;
                    chars[n] = 'g';
                    n++;
                    chars[n] = 't';
                    n++;
                    chars[n] = ';';
                    n++;
                    break;
                default:
                    if (replaceWhitechars) {
                        switch (c) {
                            case '\n':
                                chars[n] = '&';
                                n++;
                                chars[n] = '#';
                                n++;
                                chars[n] = '1';
                                n++;
                                chars[n] = '0';
                                n++;
                                chars[n] = ';';
                                n++;
                                break;
                            case '\r':
                                chars[n] = '&';
                                n++;
                                chars[n] = '#';
                                n++;
                                chars[n] = '1';
                                n++;
                                chars[n] = '3';
                                n++;
                                chars[n] = ';';
                                n++;
                                break;
                            case '\t':
                                chars[n] = '&';
                                n++;
                                chars[n] = '#';
                                n++;
                                chars[n] = '9';
                                n++;
                                chars[n] = ';';
                                n++;
                                break;
                            default:
                                chars[n] = c;
                                n++;
                        }
                    } else {
                        chars[n] = c;
                        n++;
                    }
            } // switch

        } // for

        return new String(chars, 0, n);
    }

    
    private final static void startElement(Writer writer, String name)
    throws IOException {
        writer.write("<");
        writer.write(name);
    }
    
    private final static void endElement(Writer writer, String name)
    throws IOException {
        writer.write("</");
        writer.write(name);
        writer.write(">\n");
    }
    
    private final static void endElement(Writer writer) 
    throws IOException {
        writer.write("/>\n");
    }
    
    
    private final static void writeAttributeValue(
            Writer writer, String type, String value)
    throws IOException {
        writer.write("<");
        writer.write(type);
        writer.write(">");
        writer.write(value);
        endElement(writer, type);
    }
    
    /**
     * Writes an attribute of the currently written XML element to the output writer.
     *
     * @param name attribute name
     * @param value value of the attribute
     */
    private final static void writeAttribute(Writer writer,
                                           String name,
                                           String value)
    throws IOException {
        writer.write(" ");
        writer.write(name);
        writer.write("=\"");
        writer.write(value);
        writer.write("\"");
    }
    
    /**
     * Write the properties from the XSet.
     * @param writer Where to write
    * @param indent 
     * @param xset The source of the properties.
     * @throws XAMException
     * @throws IOException
     */
    private static void processXSetProperties(Writer writer, String indent, ReferenceXSet xset)
    throws XAMException, IOException {
        XIterator iterator = xset.openFieldIterator("");

        while (iterator.hasNext()) {
            String fieldName = (String)iterator.next();
            
            //TODO:  get list of fields should not write out and skip
            if (fieldName.contains(XSet.XAM_DIRTY))
                continue;
            
            Field field = xset.getField(fieldName);
            if (field instanceof ReferenceStreamField) {
                // skip
                continue;
            }
            Property prop = (Property) field;
            writeProperty( writer, 
                           indent,
                           prop.getName(), 
                           prop.getType(), 
                           prop.getBinding(), 
                           prop.getReadOnly(), 
                           prop.getLength(), 
                           prop.getValue() );
        }   // end while
    }
    
    /** 
     * Write a property value. Can be used for properties in the XSet or in the policies section.
     * @param writer Where to write the expression
     * @param name The property name
     * @param type The stype MIME string
    * @param string 
     * @param binding True if binding, false otherwise
     * @param readOnly True if readonly, false otherwise
     * @param length The length of the property value
     * @param value The boxed property value.
     * @throws IOException
     */
    private static void writeProperty( Writer writer,
                                       String  indent,
                                       String  name,
                                       String  type,
                                       boolean binding,
                                       boolean readOnly,
                                       long    length,
                                       Object  value) throws IOException
    {
       String indent2 = indent + indent;
       writer.write(indent);
       startElement(writer, XmlConstants.TAG_PROPERTY);
       writeAttribute(writer, XmlConstants.ATTRIBUTE_NAME,     name );
       writeAttribute(writer, XmlConstants.ATTRIBUTE_TYPE,     type );
       writeAttribute(writer, XmlConstants.ATTRIBUTE_BINDING,  Boolean.toString(binding) );
       writeAttribute(writer, XmlConstants.ATTRIBUTE_READONLY, Boolean.toString(readOnly) );
       writeAttribute(writer, XmlConstants.ATTRIBUTE_LENGTH,   Long.toString(length) );
       writer.write(">\n");
       //endElement(writer);
       writer.write(indent2);

       if (type.equals(XAMLibrary.STYPE_BOOLEAN_MIME_TYPE)) {
           writeAttributeValue(writer, XmlConstants.TAG_BOOLEAN, 
                   value.toString() );
       } else if (type.equals(XAMLibrary.STYPE_DATETIME_MIME_TYPE)) {
           writeAttributeValue(writer, XmlConstants.TAG_DATE,
                   ISO8601Date.toISO8601String( (Calendar) value ) );
       } else if (type.equals(XAMLibrary.STYPE_DOUBLE_MIME_TYPE)) {
           writeAttributeValue(writer, XmlConstants.TAG_DOUBLE,
                   value.toString() );
       } else if (type.equals(XAMLibrary.STYPE_INT_MIME_TYPE)) {
           writeAttributeValue(writer, XmlConstants.TAG_INTEGER,
                   value.toString() );
       } else if (type.equals(XAMLibrary.STYPE_STRING_MIME_TYPE)) {
           writeAttributeValue(writer, XmlConstants.TAG_STRING, 
                   replaceSpecialChars(value.toString(), true ) );
       } else if (type.equals(XAMLibrary.STYPE_XUID_MIME_TYPE)) {
           writeAttributeValue(writer, XmlConstants.TAG_XUID, 
                   value.toString() );
       } 
       writer.write(indent);
       endElement(writer, XmlConstants.TAG_PROPERTY);
    }

   private static void processXSetXStreams(Writer writer, ReferenceXSet xset,
            String dirname)
    throws XAMException, IOException {
        XIterator iterator = xset.openFieldIterator("");
        while (iterator.hasNext()) {
            String fieldName = (String)iterator.next();
            
            Field field = xset.getField(fieldName);
            if (field instanceof ReferenceStreamField) {
                startElement(writer, XmlConstants.TAG_XSTREAM);
                writeAttribute(writer, XmlConstants.ATTRIBUTE_NAME, field.getName());
                writeAttribute(writer, XmlConstants.ATTRIBUTE_TYPE, field.getType());
                writeAttribute(writer, XmlConstants.ATTRIBUTE_BINDING, Boolean.toString(field.getBinding()));
                writeAttribute(writer, XmlConstants.ATTRIBUTE_READONLY, Boolean.toString(field.getReadOnly()));
                writeAttribute(writer, XmlConstants.ATTRIBUTE_LENGTH, Long.toString(field.getLength()));
                writer.write(">\n");
                writer.write("\t");
                ReferenceStreamField sfield = (ReferenceStreamField) field;
                writeAttributeValue(writer, XmlConstants.TAG_XSTREAM_DATA,
                        sfield.getLocationName());
                startElement(writer, XmlConstants.TAG_XSTREAM_XOP_INCLUDE);
                String streamCID = XmlConstants.ATTRIBUTE_CID +
                        sfield.getName() + XAMConstants.AT_SNIA_ORG;
                writeAttribute(writer, XmlConstants.ATTRIBUTE_HREF,
                        streamCID);
                writeAttribute(writer, XmlConstants.ATTRIBUTE_XMLNS_XOP,
                        XmlConstants.XMLNS_XOP_VALUE);
                endElement(writer);
                endElement(writer, XmlConstants.TAG_XSTREAM);
            }
        } // end while       
    }
    
    public static void main(String[] args) throws Exception {
        ReferenceXSet xset = new ReferenceXSet();
        xset.createProperty("TestBoolean", true, true);
        xset.createProperty("TestCalendar", true, Calendar.getInstance());
        XSetXmlWriter.outputXMLToFile("/tmp/abc", xset, "/var/tmp");
    }
}
