/*
 * Copyright (c) 2008-2009, Sun Microsystems, Inc.
 * Copyright (c) 2008-2009, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the fieldName of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.Properties;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.mail.BodyPart;
import javax.mail.Header;
import javax.mail.MessagingException;
import javax.mail.Session;
import javax.mail.internet.*;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMultipart;
import javax.activation.FileDataSource;
import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.XAMException;
import org.snia.xam.XStreamCorruptException;
import org.snia.xam.vim.reference.SniaXUID;

/**
 *
 * Class to process the XOP Packaging format used to store exported XSets.
 * 
  * Note: Using the java mail api to help parse. Tried to pick something
 * that did not require too many external jar files. To compile and run
 * the mailapi.jar file and the activation.jar (if don't have java 1.6) are
 * required.
 */
public class XSetXopReader  {
    public static boolean debug = false;
    
    XSetXopReader() {
        
    }
       
    /**
     * On import parses the XSet's streams which are stored in the XOP export
     * format.
     * @param filename of the XSet being imported
     * @param xstreamsData used to output the a vector of xstream
     * @return the parsed XSet XML
     * 
     * @throws org.snia.xam.AuthenticationExpiredException
     * @throws org.snia.xam.XStreamCorruptException
     * @throws org.snia.xam.XAMException
     */
    public static String parse(
            String                 filename, 
            Vector<XStreamXopData> xstreamsData,
            Logger                 logger )
            throws AuthenticationExpiredException, XStreamCorruptException,
            XAMException {
        
        // TODO:  clean up data
        String xsetXML = "";
        XStreamXopData[] dataOffsets = null;
        FileDataSource fData = new FileDataSource(filename);
        String contentType = fData.getContentType();

        //MimeMultipart mm = null;
        Session session = Session.getDefaultInstance(new Properties());
        InputStream is = null;
        try {
            is = fData.getInputStream();
            MimeMessage mMsg = new MimeMessage(session, is);
            String ctString = mMsg.getContentType();
            ContentType ct = new ContentType(ctString);
            
            Object obj = mMsg.getContent();
            if (obj instanceof MimeMultipart) {
                MimeMultipart mp = (MimeMultipart) obj;
                
                // Get the TOC and Stream content values parsed and loaded first
                dataOffsets = parseTOCAndXStreamXopData(mp);
                
                // Process each mime body part
                for (int i = 0; i < mp.getCount(); i++) {  
                    BodyPart mimeBP = mp.getBodyPart(i);
                    InputStream isBodyPart = mimeBP.getInputStream();
                    MimeBodyPart mmBP = new MimeBodyPart(isBodyPart);
                    String mimeBodyPartContentString = mmBP.getContentType();
                    String mimeContentId = mmBP.getContentID();
                    String encoding = mmBP.getEncoding();
                    ContentType partCT = new ContentType(
                            mimeBodyPartContentString);
                    
                    if (mimeBP instanceof MimeBodyPart) {
                        MimeBodyPart mbp = (MimeBodyPart) mimeBP;
                        Object objBP = mbp.getContent();
                        if (objBP.getClass().getName().indexOf("String")
                                > -1) {
                            ContentType partContentType = 
                                    new ContentType(mimeBodyPartContentString);
                            String baseType = partContentType.getPrimaryType();
                            String subType = partContentType.getSubType();
                            
                            // Parse the XSet XML
                            if (mimeContentId.equals(
                                    XmlConstants.XSET_CONTENT_ID_VALUE)) {
                                 xsetXML = getXSetXML(mmBP, baseType, subType,
                                         partContentType);
                            // Parse the TOC
                            } else if (mimeContentId.equals(
                                    XmlConstants.TOC_CONTENT_ID_VALUE)) {
                                // skip already processed
                            } else {
                                // skip
                            }
                        }
                    }
               }   // end for
            }
            xstreamsData.clear();
            if( dataOffsets != null )
               for (int i = 0; i < dataOffsets.length; i++) {
                  xstreamsData.add(dataOffsets[i]);
               }
        } catch (MessagingException ex) {
            XStreamCorruptException cex = 
                    new XStreamCorruptException(
                    "Error importing XStream. Invalid " +
                    "import/export format. XStream is corrupt.", ex);
            throw cex;
        } catch (IOException ex) {
            XStreamCorruptException cex =
                    new XStreamCorruptException(
                    "Error importing XStream. Invalid " +
                    "import/export format. XStream is corrupt.", ex);
        } finally {
            try {
                if (is != null)
                    is.close();
            } catch (IOException iox) {
                // log and continue
                logger.log(Level.WARNING, "Unable to close input stream");
            }
            is = null;
            fData = null;
        }
        return xsetXML;
    }
    
    private static XStreamXopData[] parseTOCAndXStreamXopData(
            MimeMultipart multiPart)
            throws XAMException, MessagingException, IOException {
        
        XStreamXopData[] dataOffsets = null;
        // Process the TOC and XStream parts first
        for (int i = 0; i < multiPart.getCount(); i++) {  
            BodyPart mimeBP = multiPart.getBodyPart(i);
            InputStream isBodyPart = mimeBP.getInputStream();
            MimeBodyPart mmBP = new MimeBodyPart(isBodyPart);
            String mimeBodyPartContentString = mmBP.getContentType();
            String mimeContentId = mmBP.getContentID();
            String encoding = mmBP.getEncoding();
 
            if (mimeBP instanceof MimeBodyPart) {
                MimeBodyPart mbp = (MimeBodyPart) mimeBP;
                Object objBP = mbp.getContent();
               if (objBP.getClass().getName().indexOf("String")
                        > -1) {
                    ContentType partContentType = 
                            new ContentType(mimeBodyPartContentString);
                    String baseType = partContentType.getPrimaryType();
                    String subType = partContentType.getSubType();
                   // Parse the XSet XML
                    if (mimeContentId.equals(
                            XmlConstants.XSET_CONTENT_ID_VALUE)) {
                        // skip over; will parse later
                    // Parse the TOC
                    } else if (mimeContentId.equals(
                            XmlConstants.TOC_CONTENT_ID_VALUE)) {
                        dataOffsets = parseTOC( mmBP, baseType, 
                                subType, encoding);
                    } else {
                        if (dataOffsets != null) {
                            parseXStream(mmBP, encoding, dataOffsets);
                        }
                    }
                }
            }
        }   // end for
        
    return dataOffsets;
        
    }
    /**
     * Locates the MIME part that contains the XSet XML and returns it
     * as a String.
     * 
     * @param xset  current XSet instance
     * @param mimePart the MIME body part instance containing the XSet XML
     * 
     * @throws org.snia.xam.XAMException
     * @throws javax.mail.MessagingException
     * @throws java.io.IOException
     */
    private static String getXSetXML( 
            MimeBodyPart mimePart, String baseType, String subType,
            ContentType partContentType)
            throws XAMException, MessagingException, IOException {
      
       String xsetXML = "";
       String charset = partContentType.getParameter(
                XmlConstants.ATTRIBUTE_CHARSET);
        String bodyType = partContentType.getParameter(
                XmlConstants.ATTRIBUTE_TYPE);
        if (!bodyType.equals(
                XmlConstants.CONTENT_SUBTYPE_TEXT + "/"
                + XmlConstants.CONTENT_SUBTYPE_XML) ||
            !charset.equals(XAMConstants.UTF_8)) {
            throw new XAMException(
                "Error parsing XSet XML. Only charset " 
                + "of UTF-8 and type of text/xml is "
                + "currently supported.");
        }                                
        if (baseType.indexOf(
            XmlConstants.CONTENT_TYPE_APPLICATION) != -1
            && subType.equals(
            XmlConstants.CONTENT_SUBTYPE_XOP)) {
            Enumeration hdrList = mimePart.getAllHeaders();
            while (hdrList.hasMoreElements()) {
                Header hdr = (Header) hdrList.nextElement();
                mimePart.removeHeader(hdr.getName());
            }

            Object objBP = mimePart.getContent();
            if (objBP.getClass().getName().indexOf("String") > -1) {
                xsetXML = (String) objBP;
            }
        } else {
            throw new XAMException("Error parsing XSet XML. Only content type" +
                " of application/xop+xml is currently supported.");
        }
        return xsetXML;
    }
    
    /**
     * Parses the TOC header and offset entries (if any)
     * @param xset
     * @param mimePart
     * @param baseType
     * @param subType
     * @param partContentType
     * @return XStreamXopData[] containing list of offsets to binary content for
     *      stream fields.
     * @throws org.snia.xam.XAMException
     * @throws javax.mail.MessagingException
     * @throws java.io.IOException
     */
    private static XStreamXopData[]  parseTOC(
            MimeBodyPart mimePart, String baseType, String subType,
            String encoding)
            throws XAMException, MessagingException, IOException {
        
        String baseTOCErrorMsg = "Error parsing XSet XOP package.";
        
        // Verify correct type and encoding values
        if (!encoding.equals(XmlConstants.CONTENT_TYPE_TEXT)) {
            throw new XAMException(baseTOCErrorMsg +
                    "Only <TOC> encoding type of text is currently supported.");
        }
        if (!baseType.equals(XmlConstants.CONTENT_TYPE_TEXT) ||
                !subType.equals(XmlConstants.CONTENT_SUBTYPE_TEXT)) {
            throw new XAMException(baseTOCErrorMsg +
                    "Only <TOC> content type of " + 
                    XmlConstants.CONTENT_TYPE_TEXT + "/" + 
                    XmlConstants.CONTENT_SUBTYPE_TEXT +
                    " is currently supported.");
 
        }
        
        Enumeration hdrList = mimePart.getAllHeaders();
        Vector<XStreamXopData> offsets = new Vector<XStreamXopData>();
        while (hdrList.hasMoreElements()) {
            Header hdr = (Header) hdrList.nextElement();
            String name = hdr.getName();
            XStreamXopData entry = null;

            if (name.startsWith(XmlConstants.TOC_OFFSET_TEXT)) {            
                entry = new XStreamXopData();
                String[] tokens = hdr.getValue().split(":");
                if (tokens == null || tokens.length < 2) {
                    throw new XAMException(baseTOCErrorMsg +
                            "Invalid Offset line: " + hdr.toString());
                }
                entry.xuidString = name.substring(
                        XmlConstants.TOC_OFFSET_TEXT.length());
                String cid = tokens[0];
                int index = cid.indexOf("@");
                entry.cid = cid;
                if (index != -1) {
                    entry.fieldName = cid.substring(0, index);
                }
                entry.xuid = SniaXUID.getInstance(entry.xuidString);
                try {
                    tokens[1] = tokens[1].trim();
                    entry.offset = Long.parseLong(tokens[1]);
                } catch (NumberFormatException ex) {
                    throw new XAMException(baseTOCErrorMsg + "Invalid offset: " +
                            hdr.toString(), ex);
                }
            }
            if (entry != null) {
                offsets.add(entry);
            }
        }   // end while
            
        return offsets.toArray(new XStreamXopData[offsets.size()]);

    }

    /**
     * On import parses XStream export format and creates a tmp file containing 
     * the stream contents. 
     * @param mimePart
     * @param encoding
     * @param xstreamData
     * @throws org.snia.xam.XAMException
     * @throws javax.mail.MessagingException
     * @throws java.io.IOException
     */
    private static void parseXStream( 
            MimeBodyPart mimePart, 
            String encoding, XStreamXopData[] xstreamData) 
            throws XAMException, MessagingException, IOException {
        boolean ignore = false;
        XStreamXopData xstream = null;
        String baseXStreamErrorMsg = "Error parsing XSet XOP package.";
        if (!(encoding.equals(XmlConstants.TRANSFER_TYPE_BINARY) ||
                encoding.equals(XmlConstants.TRANSER_TYPE_TEXT) ||
                encoding.equals(XmlConstants.TRANSFER_TYPE_8BIT))) {
            throw new XAMException(baseXStreamErrorMsg +
                    "Only tranfer encoding of binary or 8bit text is currently " +
                    "supported.");
        }
        
        String cid = mimePart.getContentID();
        int index = XStreamXopData.findFieldEntry(cid, xstreamData);
        File tempFile = File.createTempFile("ImportStreamField_", ".tmp");
        OutputStream os = new FileOutputStream(tempFile);
        if (index == -1) {
            // TODO:  log this
            System.out.print("Ignoring XStream data. " +
                    "No matching xstream field found in TOC for: " +
                    cid);
            ignore = true;
            os.close();
        } else {
        
            InputStream is = null;
            try {
                xstream = xstreamData[index];
                XSetXopReader.removeAllHeaders(mimePart);
                //InputStream is = mimePart.getInputStream();
                is = mimePart.getRawInputStream();
                byte[] dataBytes =  new byte[4000];
                int numBytes = is.read(dataBytes);
                while (numBytes > 0) {
                    os.write(dataBytes, 0, numBytes);
                    numBytes = is.read(dataBytes);
                }
                xstream.tmpFilename = tempFile.getAbsolutePath();
            } finally {
                os.close();
                is.close();
            }
        }
        
    }

    private static void removeAllHeaders(MimeBodyPart mimeBodyPart) 
            throws MessagingException {
        Enumeration hdrList = mimeBodyPart.getAllHeaders();
        while (hdrList.hasMoreElements()) {
            Header hdr = (Header) hdrList.nextElement();
            mimeBodyPart.removeHeader(hdr.getName());
        }

    }    
}
