/*
 * Copyright (c) 2008, Sun Microsystems, Inc.
 * Copyright (c) 2008, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;


import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.SequenceInputStream;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.nio.channels.FileChannel;
import java.util.Vector;
import org.snia.xam.XAMException;
import org.snia.xam.XIterator;
import org.snia.xam.XUID;
import org.snia.xam.base.Field;
import org.xml.sax.helpers.DefaultHandler;

/**
 *
 * 
 */
public class XSetXopWriter extends DefaultHandler {

    private static final int TRANSFER_SIZE = 5000;
    
    /**
     * Writes out a temp file containing the XOP Package for the exported XSet.
     * 
     * Note that this is a Work In Progress. Right now all the content to
     * be exported is rewritten out to a tmp file. This is not needed as
     * the actual stream contents are already stored in a format that is
     * directly acessible by the client ExportXStream. Also the stream contents
     * can also be quite large so it would be better to not make another copy.
     * It was easier for debugging to create a single file for the first pass
     * and it is less likely that another client will modify the stream contents
     * while this client is trying to perform an export.
     * 
     * Future versions might use the SequenceInputStream class to provided
     * ordered access to the tmp and stream content files without the need
     * for rewriting the stream contents yet again.
     * 
     * @param filename tmp filename; exported xset xop& xml is written to this 
     * file
     * @param xset reference to the xset being exported
     * @param dirname repository directory name - currently not used
     * @param mimeBoundary the MIME boundary that will be used in the export
     * @throws org.snia.xam.XAMException
     */
    public static void outputXMLToTmpFile(String filename, ReferenceXSet xset,
            String dirname, String mimeBoundary)
    throws XAMException {
        String mimeBoundaryOutput = "--" + mimeBoundary;
        try {
            FileOutputStream fos = new FileOutputStream(filename);
            FileChannel channel = fos.getChannel();
   
            Writer out
                = new BufferedWriter(
                    new OutputStreamWriter(fos, XAMConstants.UTF_8));
            out.write("Content-Type: Multipart/Related;boundary=\"");
            out.write(mimeBoundary);
            out.write("\";");
            out.write(XmlConstants.ATTRIBUTE_TYPE + 
                    "=\"" + XmlConstants.CONTENT_TYPE_APPLICATION_XOP +
                    "\";");
            out.write(XmlConstants.ATTRIBUTE_START + 
                    "=\"" + XmlConstants.XSET_CONTENT_ID_VALUE + "\";");
            out.write(XmlConstants.ATTRIBUTE_START_INFO +
                    "=\"text/xml\"\n");
            out.flush();
            out.write(XmlConstants.CONTENT_DESC_HEADER + 
                    XmlConstants.DESC_TEXT + xset.getXUIDString());
            writeLineTerminator(out, channel);
            writeMimeBoundary(out, channel, mimeBoundaryOutput, false);
            out.flush();
            
            // Write out XSet XML
            writeXSetXML(out, channel, xset, fos, mimeBoundaryOutput);
            
            ReferenceStreamFieldExport[] exportStreams =
                    getXStreamsOutputFiles(xset, mimeBoundaryOutput);
            
            // Write out TOC
            writeTOC(out, channel, xset, exportStreams);
            
            // Write out XStreams
            writeXStreams(out, channel, exportStreams);

            //writeLineTerminator(out, channel);
            writeMimeBoundary(out, channel, mimeBoundaryOutput, true);
            
            out.write("\n");
            out.close();
        }
        catch (IOException io) {
            throw new XAMException(io);
        }
    }
    
    /*
     * Writes out the Table of Contents information
     */
    private static void writeTOC(Writer out, FileChannel channel, 
            ReferenceXSet xset,  ReferenceStreamFieldExport[] streams)
            throws IOException {
        
        writeLineTerminator(out, channel);
        // Write out content headers
        out.write(XmlConstants.CONTENT_TYPE_HEADER);
        out.write("text/text\n");
        out.write(XmlConstants.CONTENT_TRANSER_ENCODING_HEADER);
        out.write(XmlConstants.TRANSER_TYPE_TEXT);
        out.write("\n");
        out.write(XmlConstants.CONTENT_ID_HEADER);
        out.write(XmlConstants.TOC_CONTENT_ID_VALUE);
        out.flush();
        
        if (streams.length > 0) {
            writeLineTerminator(out, channel);
            long startPosition = channel.position();
            long tocLength = 0;
            StringBuffer[] toc = new StringBuffer[streams.length];
            long[] offsets = new long[streams.length];
            for (int i = 0; i < streams.length; i++) {
                StringBuffer buffer = 
                        new StringBuffer(XmlConstants.TOC_OFFSET_TEXT);
                buffer.append(xset.getXUIDString());
                buffer.append(":");
                buffer.append(streams[i].streamField.getName());
                buffer.append(XAMConstants.AT_SNIA_ORG);
                buffer.append(":");
                
                //String offsetString = 
                String offsetString = String.format("%1$-20d", new Long(0));
                buffer.append(offsetString);
                buffer.append("\n");
                String tmp = buffer.toString();
                toc[i] = buffer;
                byte[] bytes = tmp.getBytes(XAMConstants.UTF_8);
                tocLength += bytes.length;
            }
            
            long currentPosition = startPosition + tocLength;
            
            for (int i = 0; i < streams.length; i++) {
                offsets[i] = currentPosition;
                int pos = toc[i].lastIndexOf(":");
                String offsetString = String.format("%1$-20d",
                        new Long(currentPosition));
                offsetString += "\n";
                toc[i].replace(pos+1, toc[i].length(), offsetString);
                
                out.write(toc[i].toString());
                out.flush();
                out.flush();
                currentPosition += 
                        streams[i].getMimeHeaderFile().length();
                currentPosition += 
                        streams[i].getStreamContentsFile().length();
                // or can do currentPostion = channel.position();
            }          
          
        }   // end if stream fields        
        out.flush();
    }
    
    /*
     * Formats and gets a list of stream field data contents ready for
     * export.
     */
    private static ReferenceStreamFieldExport[]
            getXStreamsOutputFiles(ReferenceXSet xset, 
            String mimeBoundaryOutput) throws IOException, XAMException {
        
        XIterator fields = xset.openFieldIterator("");
        Vector<ReferenceStreamFieldExport> exportStreamList =
                new Vector<ReferenceStreamFieldExport>();
        ReferenceStreamFieldExport[] exportStreams = 
                new ReferenceStreamFieldExport[0];;
        
        while (fields.hasNext()) {
            String name = (String) fields.next();
            Field field = xset.getFieldNoChecking(name);
            if (field instanceof ReferenceStreamField) {
                ReferenceStreamField streamField = (ReferenceStreamField) field;
                ReferenceStreamFieldExport exportStream =
                        new ReferenceStreamFieldExport(mimeBoundaryOutput,
                        streamField);
               exportStreamList.add(exportStream); 
            }
        }
        int num = exportStreamList.size();
        if (num > 0) {
            exportStreams = new ReferenceStreamFieldExport[num];
            exportStreamList.toArray(exportStreams);
        }
        
        return exportStreams;

    }
    
    /*
     * Writes out the XStream information 
     */
    private static void writeXStreams(Writer out, FileChannel channel,
            ReferenceStreamFieldExport[] exportStreams) 
            throws IOException, XAMException {
               
        for (ReferenceStreamFieldExport stream : exportStreams) {
            File[] exportData = stream.getFileList();
            for (File file : exportData) {
                FileInputStream fis = new 
                        FileInputStream(file);
                FileChannel inputChannel = fis.getChannel();
                long numBytes = inputChannel.size();
                long currentPosition = channel.position();
                long num = channel.transferFrom(
                        inputChannel, channel.position(), numBytes);
                inputChannel.close();
                channel.position(currentPosition + num);
                out.flush();
            }
            
            if (!ReferenceExportXStream.debug)
                stream.mimeHdrFile.delete();
        }
    }

    /**
     * 
     * @param out
     * @param xset
     * @param fos
     * @param mimeBoundaryOutput
     * @throws java.io.IOException
     */
    private static void writeXSetXML(Writer out, FileChannel channel, 
            ReferenceXSet xset, FileOutputStream fos, 
            String mimeBoundaryOutput) throws IOException {
        
        // Write out content headers
        writeLineTerminator(out, channel);
        out.write(XmlConstants.CONTENT_TYPE_HEADER);
        out.write(XmlConstants.CONTENT_TYPE_APPLICATION_XOP);
        out.write(";    ");
        out.write(XmlConstants.ATTRIBUTE_CHARSET);
        out.write("=\"" + XAMConstants.UTF_8 + "\";    ");
        out.write(XmlConstants.ATTRIBUTE_TYPE);
        out.write("=\"text/xml\"\n");
        out.write(XmlConstants.CONTENT_TRANSER_ENCODING_HEADER);
        out.write(XmlConstants.TRANSFER_TYPE_8BIT);
        out.write("\n");
        out.write(XmlConstants.CONTENT_ID_HEADER);
        out.write(XmlConstants.XSET_CONTENT_ID_VALUE);
        writeLineTerminator(out, channel);
        writeLineTerminator(out, channel);
        
        // Now write out the XSet XML
        XUID xuid = xset.getXUID();
        String filename = xset.getXSystem().getPersistenceMgr()
                .getXMLFileName(xuid);
        File xmlFile = new File(filename);
        FileInputStream fis = new FileInputStream(xmlFile);
        FileChannel inChannel = fis.getChannel();
        FileChannel outChannel = fos.getChannel();
        long numTransfer = inChannel.transferTo(0, TRANSFER_SIZE, outChannel);
        long position = 0;
        while (numTransfer > 0) {
            position += numTransfer;
            numTransfer = inChannel.transferTo(position, TRANSFER_SIZE,
                    outChannel);
        }
        fis.close();
        writeMimeBoundary(out, channel, mimeBoundaryOutput, false);
   }
    
    /**
     * 
     * Writes out the MIME Boundary and the associate CRLF's needed
     * 
     * @param out
     * @param mimeBoundary
     * @param finalBoundary
     * @throws java.io.IOException
     */
    protected static void writeMimeBoundary(Writer out, FileChannel channel, 
            String mimeBoundary, boolean finalBoundary) throws IOException {
        writeLineTerminator(out, channel);
        out.write(mimeBoundary);
        if (finalBoundary) {
            out.write("--");
        }
        writeLineTerminator(out, channel);               
    }
    
    protected static void writeLineTerminator(Writer out, FileChannel channel) 
            throws IOException {
        out.flush();
        out.write("\n");
        out.flush();
    }

}
