/*
 * Copyright (c) 2008, Sun Microsystems, Inc.
 * Copyright (c) 2008, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.logging.Logger;

import org.snia.xam.XAMException;
import org.snia.xam.XSet;
import org.snia.xam.XSetAbandonException;
import org.snia.xam.XSetException;
import org.snia.xam.XStream;
import org.snia.xam.XStreamAbandonException;
import org.snia.xam.XStreamCorruptException;
import org.snia.xam.XStreamException;
import org.snia.xam.XSystem;

/**
 *  Note: Work in Progress - not final implementation
 *
 * Proxy class to control access to XStream objects
 * 
 */
public class XStreamClientInvocationHandler implements InvocationHandler {
    private ReferenceXSystem refXSystem = null;
    private ReferenceXSet    refXSet    = null;
    private Logger           logger     = null;
    private ReferenceXStreamBase refXStream = null;

    
    XStreamClientInvocationHandler(XStream xstream) throws XAMException {
        if (xstream instanceof ReferenceXStreamBase) {
            refXStream = (ReferenceXStreamBase) xstream;
            refXSet = refXStream.getParentXSet();
            refXSystem = refXSet.getXSystem();
            logger = refXSystem.getLogger();
        } else {
            // change to runtime exception?
           String msg = "XStream class type not supported: " + xstream.getClass().getName();
           logger.severe(msg);
           throw new XSetException(msg);
        }
    }
    
    /**
     * @see InvocationHandler#invoke(Object, Method, Object[]) 
     * 
     * @param proxy
     * @param method
     * @param args
     * @return the actual object used to invoke the method
     * @throws java.lang.Throwable
     */
    public Object invoke(Object proxy, Method method, Object[] args) 
            throws Throwable {
            String methodName = method.getName();
            Object returnObj = null;
            
            try {
                if (refXStream.isCorrupt() && 
                        !(methodName.equals("abandon"))) {
                   String msg = "XStream is corrupt and must be abandoned.";
                   logger.warning(msg);
                   throw new XStreamCorruptException(msg);
                }
                 /*
                 * All these methods can call without being authenticated
                 */
                if (methodName.equals("abandon") ||
                        methodName.equals("close")) {
                    if (refXSet.isClosed()) {
                       String msg = "XSet is closed and is not accessible.";
                       logger.warning(msg);
                       throw new XSetException(msg);
                    }
                    if (refXStream.isClosed()) {
                       String msg = "XStream is closed and is not accessible.";
                       logger.warning(msg);
                       throw new XStreamException(msg);
                    }
                    
                    return method.invoke(refXStream, args);
                }

                // Otherwise check access is allowed
                refXSystem.checkAccessAllowed();
              
                if (refXSet.isClosed()) {
                   String msg = "XSet is closed and is not accessible.";
                   logger.warning(msg);
                   throw new XSetException(msg);
                }
                if (refXSet.isAbandoned()) {
                   String msg = "XSet is abandoned and must be closed.";
                   logger.warning(msg);
                   throw new XSetAbandonException(msg);
                }
                
                if (refXStream.isClosed()) {
                    if (!methodName.equals("openXStream")) {
                       String msg = "XStream is closed and is not accessible";
                       logger.warning(msg);
                       throw new XStreamException(msg);
                    }
                }
                
                if (refXStream.isAbandoned() &&
                        !methodName.equals("close")) {
                   String msg = "XStream has been abandoned and must be closed.";
                   logger.warning(msg);
                   throw new XStreamAbandonException(msg);
                }

                // Check to see if calling method is allowd
                refXStream.checkMethodAccess(methodName);

                // finally call the method
                returnObj = method.invoke(refXStream, args);
                
            } catch (InvocationTargetException ex) {
                Throwable th = ex.getCause();
                if (th != null && th instanceof XAMException) {
                    throw th;
                } else {
                    throw ex;
                }
            }
            return returnObj;
    }
}
