/*
 * Copyright (c) 2008-2009, Sun Microsystems, Inc.
 * Copyright (c) 2008-2009, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.File;

import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.XAMException;
import org.snia.xam.XStream;
import org.snia.xam.XStreamException;
import org.snia.xam.util.XAMIOException;

/**
 *
 *  Note: Work in Progress - not final implementation
 * 
 *  Handles writing stream field contents to disk and managing temp files
 *  for Reference VIM.
 * 
 * This class was originally handling all access for the XStream, now that has
 * been moved to the ReferenceXStream class. However this class still needs to
 * be cleaned up to remove those methods and close streams when not in use.
 */
public class XStreamContents {
    
    public static final String XSTREAM_FILE_PREFIX = "XStream_";
    public static final String XSTREAM_FILE_SUFFIX = ".data";

    private static final int BUFFER_SIZE = 1000;
    protected String srcFileName = "";
    protected String fileName = "";
    protected String fieldName = "";
    protected String openMode = "";
    protected boolean modified = false;
    protected boolean deleted = false;
    protected boolean corrupt = false;
    protected boolean abandoned = false;
    private long length = 0;
    
    private Logger logger = null;
    
    private boolean closed = true;
    File workingFile;
    File srcFile;
    private FileWriter fw;
    private FileReader fr;
    private FileOutputStream fos;
    private FileInputStream fis;
    private FileChannel fc;
    
    
    /**
     * Creates new XStreamContents which manages the payload of an XStream field
     * 
     * @param fileName  source file name (blank if new XStream)
     * @param fieldName field name
     * @param length    initial length of XStream; zero if new XStream
     * 
     * @throws org.snia.xam.InvalidArgumentException
     * @throws org.snia.xam.XAMException
     */
     protected XStreamContents(String fileName, String fieldName, long length, Logger externLogger ) 
            throws InvalidArgumentException, XAMException  {
        srcFileName = fileName;
        this.fieldName = fieldName;
        this.length = length;
        this.logger = externLogger;
        
    }
   
    /*
     * Creates new XStreamContents which manages the payload of an XStream field
     *  Used for testing only.
     * 
     * @param fileName  source file name (blank if new XStream)
     * @param fieldName field name
     * @param openMode  mode to open XStream
     * @param length    initial length of XStream; zero if new XStream
     * 
     * @throws org.snia.xam.InvalidArgumentException
     * @throws org.snia.xam.XAMException
     */
    private XStreamContents(String fileName, String fieldName, 
            String openMode, long length) 
            throws InvalidArgumentException, XAMException  {
        srcFileName = fileName;
        this.fieldName = fieldName;
        this.length = length;
        
        if (openMode != null) {
            open(openMode);
        }
        
    }
    
    /**
     * Determines the file that contains the latest field contents and opens
     * the file if not already open. For write mode access this will create
     * a tmp file. For truncate mode the tmp file is empty. If open in append
     * mode then the original data contents are copied to the tmp file before
     * this method returns.
     * 
     * @param openMode
     * @return the name of the file that contains the latest field contents
     * @throws org.snia.xam.InvalidArgumentException
     * @throws org.snia.xam.XAMException
     */
    protected String getWorkingFilename(String openMode) 
            throws InvalidArgumentException, XAMException {
        
        String fileName = "";
        if (workingFile != null) {
            fileName = workingFile.getAbsolutePath();
        }
        if (!modified) {
            if (openMode.equals(XStream.MODE_READ_ONLY)) {
                if (closed) {
                    openRead();
                }
                fileName = srcFileName;

            } else if (openMode.equals(XStream.MODE_WRITE_APPEND)) {
                openAppend();
                fileName = this.workingFile.getAbsolutePath();

            } else if (openMode.equals(XStream.MODE_WRITE_TRUNCATE))  {
                // Open new empty file
                openTrunc();
                fileName = this.workingFile.getAbsolutePath();

            } else {
               String msg = "Invalid open mode: " + openMode;
               logger.warning(msg);
               throw new InvalidArgumentException(msg);
            }
            this.openMode = openMode;
        }
        return fileName;
    }
        
    private void open(String openMode) 
            throws InvalidArgumentException, XAMException {
        
        if (!modified) {
            if (openMode.equals(XStream.MODE_READ_ONLY)) {
                openRead();

            } else if (openMode.equals(XStream.MODE_WRITE_APPEND)) {
                openAppend();

            } else if (openMode.equals(XStream.MODE_WRITE_TRUNCATE))  {
                // Open new empty file
                openTrunc();

            } else {
               String msg = "Invalid open mode: " + openMode;
               logger.warning(msg);
               throw new InvalidArgumentException(msg);
            }
            this.openMode = openMode;
        }
    }
    
    /*
     *  Open for reading 
     * @throws org.snia.xam.XAMException
     */
    private void openRead() throws XAMException {
        if (deleted) {
           String msg = "XStream has been deleted.";
           logger.warning(msg);
           throw new XAMException(msg);
        }
        try {
            // calling close
            close();
            // reopen
            srcFile = new File(srcFileName);
            fr = new FileReader(srcFile);
            fis = new FileInputStream(srcFile);
            fc = fis.getChannel();
        } catch (IOException ex) {
            corrupt = true;
            String msg = "Could not open the XStream: " + srcFileName;
            logger.log(Level.SEVERE, msg, ex);
            throw new XStreamException(msg, ex);
        }
        
        closed = false;
        
    }

    /**
     * Set the contents/file for deletion. This means the stream field is
     * being deleted. The file will not actually be deleted until the
     * xset changes are committed.
     * 
     * @throws org.snia.xam.XAMException
     * @throws org.snia.xam.util.XAMIOException
     */
    protected void delete() throws XAMException, XAMIOException{
        if (deleted) {
           String msg = "XStream is open or has already been deleted.";
           logger.warning(msg);
           throw new XAMException(msg);
        }
        
        if (modified) {
            try {
                // delete the working copy at this time
                if (fileName != null && fileName.length() > 0) {
                    close();
                    deleteFile(fileName);
                }
                modified = true;
            } catch (XAMException ex) {
                throw ex;
            } catch (Exception iox) {
                corrupt = true;
                throw new XAMException("Error deleting temp XStream file",
                        iox);
            }
        }
        // Delete permament copy on commit
        deleted = true;

    }
    
    /*
     *  Deletes the file
     *  
     * @param fname
     * @throws org.snia.xam.XAMException
     */
    private synchronized void deleteFile(String fname) throws XAMException {
        File tmpFile  = new File(fname);
        if (!tmpFile.delete()) {
            if (tmpFile.exists()) {
                corrupt = true;
                String msg = "Unable to delete temp file for XStream: " + fname;
                logger.severe(msg);
                throw new XAMException(msg);
            }
        }
    }
    

    /**
     * Commits the changes.
     * @param dirname name of the directory storing info for XSet
     * @return long the number of bytes stored in this stream file
     * 
     * @throws org.snia.xam.XAMException
     */
    protected long commit(String dirname ) throws XAMException {
        
        synchronized(this) {
            this.close();

            if (!dirname.endsWith(File.separator)) {
                dirname = dirname + File.separator;
            }

            if (deleted) {
                if (srcFileName != null && srcFileName.length() > 0 &&
                        !srcFileName.equals(fileName)) {
                    // if persisting to new XUID/new dir name then don't
                    // delete existing copy cause belongs to different XUID
                    if (srcFileName.indexOf(dirname) != -1) {
                        deleteFile(srcFileName);
                    }
                }
            } else {
                int pos = srcFileName.indexOf(dirname);
                boolean isTmpFile = ReferenceVIMPersistenceMgr.isTmpFile(
                        srcFileName);                       
                boolean srcFileExists = false;
                File file = null;
                if (srcFileName != null && srcFileName.length() > 0 &&
                        !srcFileName.equals(fileName)) {
                    file = new File(srcFileName);
                    srcFileExists =  file.exists();
                    
                }
                // Writing file back to same directory also
                // means that the xuid has not changed
                if (pos != -1) {
                    // remove old copy
                    if (modified) {
                        if (srcFileExists) {
                            // delete old
                            if (!file.delete()) {
                               String msg = "Failed to delete file. XSet is corrupt.";
                               logger.warning(msg);
                               throw new XAMException(msg);
                            }
                        }
                        // rename file
                        File tmpFile = new File(fileName);
                        String persistedFileName = file.getAbsolutePath();
                        cloneContents(fileName, persistedFileName);
                        if (!tmpFile.delete()) {
                            logger.log(Level.WARNING, 
                                    "Failed to delete temp file: " +
                                    tmpFile.getAbsolutePath());
                        }
                    }
                } else {    // new xuid; copy file
                    if (!modified) {
                        File outDir = new File(dirname);
                        File newFile = null;
                        try {
                            newFile = File.createTempFile(
                                XSTREAM_FILE_PREFIX, XSTREAM_FILE_SUFFIX, 
                                outDir);
                        } catch (IOException iox) {
                           String msg = "Could not persist XStream.";
                           logger.log( Level.SEVERE, msg, iox );
                            throw new XAMException( msg, iox);
                        }
                        String newName = newFile.getAbsolutePath();
                        cloneContents(srcFileName, newName);
                        // Normally happens when importing an XSet with
                        // XStreams
                        if (isTmpFile) {
                            File tmpFile = new File(srcFileName);
                            if (!ReferenceImportXStream.debug) {
                                if (!tmpFile.delete()) {
                                    logger.log(Level.WARNING, 
                                            "Failed to delete file: " +
                                            tmpFile.getAbsolutePath());
                                }
                            }
                        }
                        srcFileName = newName;
                        fileName = null;
                    } else {
                        // file has been modified
                        file = new File(fileName);
                        String newName = file.getName();
                        pos = newName.indexOf(
                                ReferenceVIMPersistenceMgr.TMP_FILENAME_SUFFIX);                                                
                        newName = dirname + newName.substring(0, pos) + 
                                XSTREAM_FILE_SUFFIX;
                        cloneContents(fileName, newName);
                        if (!file.delete()) {
                            logger.log(Level.WARNING, 
                                    "Failed to delete file: " +
                                    file.getAbsolutePath());
                        }
                        this.srcFileName = newName;
                        fileName = null;
                    }
                }
            }
            workingFile = null;
            modified = false;
            File tmpFile = new File(srcFileName);
            length = tmpFile.length();
            return length;
        }  // end synchronized
        
    }

    /**
     * Clean up tmp file if changes were made to the stream and they are not
     * being committed.
     * Note: This is only called by parent XSet when closing the xset without
     * committing changes.
     */
    protected void cleanupUncommitedChanges()  {
        // Close - should really already be closed
        close();
        
        if (modified) {
            File file = new File(workingFile.getAbsolutePath());
            if (!file.delete()) {
                logger.log(Level.WARNING, "Failed to cleanup temp file: " +
                        workingFile.getAbsolutePath());
            }
        }
    }
    
    /*
     * Open stream contents file for write-append mode. This means opening
     * a new tmp file and copying the existing contents to the tmp file.
     * (Remember changes are not persisted until client executes commit on
     * the xset instance.)
     *  
     * @throws org.snia.xam.XAMException
     */
    private void openAppend() throws XAMException {
        if (deleted) {
           String msg = "XStream is already open or has been deleted.";
           logger.warning(msg);
           throw new XAMException(msg);
        }
            
        FileChannel fic = null;
        try {
            if (fileName == null || fileName.length() == 0) {
                // Copy contents to new working copy;
                openTrunc();
                fr = new FileReader(srcFileName);
                fis = new FileInputStream(srcFileName);
                fic = fis.getChannel();
                byte[] b = new byte[BUFFER_SIZE];
                ByteBuffer bb = ByteBuffer.wrap(b);
                int numRead = fic.read(bb);
                while (numRead > 0) {
                    ByteBuffer bbOut = ByteBuffer.wrap(b, 0, numRead);
                    fc.write(bbOut);
                    bb.clear();
                    numRead = fic.read(bb);
                }

                fc.close();
                fw.close();
                fos.close();
                fr.close();
                fic.close();
                fis.close();
            }
            
           fw = new FileWriter(workingFile, true);
           fos = new FileOutputStream(workingFile, true);
           fc = fos.getChannel();
           
        } catch (IOException ex) {
            corrupt = true;
            try {
                if (fic != null) {
                    fic.close();
                    fic = null;
                }
                if (fis != null) {
                    fis.close();
                    fis = null;
                }
                if (fw != null) {
                    fw.close();
                    fw = null;
                }
                if (fos != null) {
                    fos.close();
                    fos = null;
                }
                if (fc != null) {
                    fc.close();
                    fc = null;
                }
            } catch (Exception exc) { 
                logger.log(Level.WARNING, "Failed to close all io streams for" +
                        " file: " + workingFile);
            }
            
            String msg = "Could not open the XStream.";
            logger.log( Level.SEVERE, msg, ex );
            throw new XStreamException( msg, ex );
        }
        closed = false;
        modified = true;
        
    }
    
    /*
     * Open stream contents file for write-trunc mode. 
     * It creates a new empty tmp file. (Remember changes are not persisted 
     * until client executes commit on the xset instance.)
     *  
     * NOTE: Sets the this.workingFile member to the newly created temp file. 
     * @throws org.snia.xam.XAMException
     */
    private void openTrunc() throws XAMException {
        if (deleted) {
           String msg = "XStream is already open or deleted.";
           logger.warning(msg);
           throw new XAMException(msg);
        }

        try {
            workingFile = File.createTempFile(XSTREAM_FILE_PREFIX, ".tmp");
            fileName = workingFile.getAbsolutePath();
            if (srcFileName == null || srcFileName.length() == 0) {
                srcFileName = fileName;
            }
            fw = new FileWriter(workingFile);
            fos = new FileOutputStream(workingFile, false);

            fc = fos.getChannel();
            fc.truncate(0);
            length = 0;
        } catch (Exception ex) {
            corrupt = true;
            throw new XStreamException("Could not create XStream. " +
                    "Error opening file.", ex);
        }
        closed = false;
        modified = true;
    }
    
    
    /*
     * Copy contents from one file to another file (usually the working copy/
     * tmp file).
     *  
     * @param src file to copy contents from
     * @param dest file to copy contents to
     * @throws org.snia.xam.XAMException
     */
    private void cloneContents(String src, String dest) throws XAMException {
        //workingFile = new File(srcFileName);
        if (!closed || deleted) {
           String msg = "XStream is already open or has been deleted.";
           logger.warning(msg);
           throw new XAMException(msg);
        }
            
        FileChannel fic = null;
        FileChannel foc = null;
        File srcFile = null;
        File destFile = null;
        FileOutputStream destfos = null;
        FileReader srcfr = null;
        FileInputStream srcfis = null;
        try {
            srcFile = new File(src);
            destFile = new File(dest);
            destFile.createNewFile();
            destfos = new FileOutputStream(dest, true);

            foc = destfos.getChannel();
            foc.truncate(0);
  
            srcfr = new FileReader(srcFile);
            srcfis = new FileInputStream(srcFile);
            fic = srcfis.getChannel();
            byte[] b = new byte[BUFFER_SIZE];
            ByteBuffer bb = ByteBuffer.wrap(b);
            int numRead = fic.read(bb);
            while (numRead > 0) {
                ByteBuffer bbOut = ByteBuffer.wrap(b, 0, numRead);
                int numWrite = foc.write(bbOut);
                destfos.flush();
                bb.clear();
                numRead = fic.read(bb);
            }

            bb = null;            
        } catch (IOException ex) {
           String msg = "Could not copy the XStream data.";
           logger.log( Level.SEVERE, msg, ex );
           throw new XStreamException( msg, ex);
        }
        finally {
            // Must match order in switch statement
            String[] names = {"input file channel", "source input stream",
                            "destination output stream", "output file channel",
                            "source file reader"};
            int i = 0;
            while(i < names.length) {
                try {
                    switch (i) {
                        case 0:
                            i++;
                            if (fic != null) 
                                fic.close();
                        case 1:
                            i++;
                            if (srcfis != null)
                                srcfis.close();
                        case 2:
                            i++;
                             if (destfos != null)
                                destfos.close();
                        case 3:
                            i++;
                            if (foc != null)
                                foc.close();
                        case 4:
                            i++;
                            if (srcfr != null)
                                srcfr.close();
                    }   // end switch

                } catch (Exception exc) { 
                    /* log and continue */ 
                    logger.log(Level.WARNING, 
                            "Unable to close " + names[i], exc);
                }
            }   // end while
            
            fic = null;
            srcfis = null;
            foc = null;
            destfos = null;
            srcfr = null;
            // Make sure everything closed on source file
            this.close();
        }
    }
    

    /**
     * Makes sure everything is closed.
     */
    private void close() {
        closed = true;
        // Must match order in switch statement
        String[] names = {"input stream", "file channel", "file writer",
                        "output stream", "file reader"};
        int i = 0;
        while(i < names.length) {
            try {
                switch (i) {
                    case 0:
                        i++;
                        if (fis != null) 
                            fis.close();
                    case 1:
                        i++;
                        if (fc != null)
                            fc.close();
                    case 2:
                        i++;
                         if (fw != null)
                            fw.close();
                    case 3:
                        i++;
                        if (fos != null)
                            fos.close();
                    case 4:
                        i++;
                        if (fr != null)
                            fr.close();
                }   // end switch

            } catch (Exception exc) { 
                /* log and continue */ 
                logger.log(Level.WARNING, "Unable to close " + names[i], exc);
            }
        }   // end while
        // Make sure set to null
        fc = null;
        fr = null;
        fw = null;
        fis = null;
        fos  = null;
        
    }
    
    // These getStatus methods are currently not used; the status fields are
    // not guaranteed to be set correctly
    protected boolean isModified() {
        return modified;
    }
    
    protected boolean isClosed() {
        return closed;
    }
    
    protected boolean isAbandoned() {
        return abandoned;
    }
    
    protected boolean isCorrupt() {
        return corrupt;
    }
    
    protected long getLength() {
        return length;
    }
    
    /**
     * Retrieves the current persisted file for data contents
     * @return the persisted filename 
     */
    protected String getAbsoluteLocation() {
        return srcFileName;
    }
    
    /**
     * Retrieves the current persisted file name for data contents
     * @return the persisted filename (last name in the pathname's sequence)
     */
    protected String getLocationName() {
        String name = "";
        if (srcFileName != null) {
            File file = new File(srcFileName);
            name = file.getName();
        }
        return name;
    }



}
