/*
 *  Copyright (c) 2008, Sun Microsystems, Inc.
 *  Copyright (c) 2008, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.util.Vector;

/**
 * Typesafe enum class that can be supported by JDK 1.4.
 * This class construct can be replaced in JDK 1.5 by the Java
 * Enum calss
 */
public class XStreamState extends OperationalState {
    
    /**
     * List of states
     */
    protected static Vector<XStreamState> states = new Vector<XStreamState>();

    /**
     * The state that indicates the stream has been abandoned
     */
    public static final XStreamState ABANDONED = new XStreamState("ABANDONED");
    
    /**
     * The state that indicates the stream has been closed.  No further 
     * action on the stream is possible
     */
    public static final XStreamState CLOSED = new XStreamState("CLOSED");
        
    /**
     * Reader State
     */
    public static final XStreamState READER = new XStreamState("READER");

    /**
     * The state that indicates the read stream is corrupt
     */
    public static final XStreamState READ_CORRUPT = 
            new XStreamState("READ_CORRUPT");

    /**
     * Export Reader State
     */
    public static final XStreamState EXPORT_READER = 
            new XStreamState("EXPORT_READER");

    /**
     * The state that indicates the export read stream is corrupt
     */
    public static final XStreamState EXPORT_READ_CORRUPT = 
            new XStreamState("EXPORT_READ_CORRUPT");

    /**
     * Writer State
     */
    public static final XStreamState WRITER = new XStreamState("WRITER");
    
    /**
     * The state that indicates the write stream is corrupt
     */
    public static final XStreamState WRITE_CORRUPT = 
            new XStreamState("WRITE_CORRUPT");
        
    /**
     * Import Writer State
     */
    public static final XStreamState IMPORT_WRITER = 
            new XStreamState("IMPORT_WRITER");
    
    /**
     * The state that indicates the import write stream is corrupt
     */
    public static final XStreamState IMPORT_WRITE_CORRUPT = 
            new XStreamState("IMPORT_WRITE_CORRUPT");
    
    private XStreamState(String state) {
        super(state);
        addState(this);
    }

    private void addState(XStreamState newState) {
        states.add(newState);
    }
    
    static protected Vector<XStreamState> getValidStates()  {
        return (Vector<XStreamState>) states.clone();
    }
    
    
    public boolean isAbandoned() {
        return (this.equals(ABANDONED)) ? true : false;
    }
    
    public boolean isClosed() {
        return (this.equals(CLOSED)) ? true : false;
    }

    /**
     * Returns true if XStream is open for exporting.
     * @return true if XStream open for exporting
     */
    public boolean isExport() {
        return (this.equals(EXPORT_READER) 
                || this.equals(EXPORT_READ_CORRUPT)) ? true : false;
    }
    
    public boolean isExportReader() {
        return (this.equals(EXPORT_READER)) ? true : false;
    }
    
    public boolean isReader() {
        return (this.equals(READER)) ? true : false;
    }
    
    public boolean isOpenReadMode() {
        boolean openRead = false;
        if (isExport() ||
            isReader() ||
            this.equals(READ_CORRUPT)) {
            openRead = true;
        }
        return openRead;
    }

    public boolean isImportWriter() {
        return (this.equals(IMPORT_WRITER)) ? true : false;
    }
    
    /**
     * Determines if XStream was open for importing.
     * @return true if XStream open for importing
     */
    public boolean isImport() {
        return (this.equals(IMPORT_WRITER)
                || this.equals(IMPORT_WRITE_CORRUPT)) ? true : false;
    }
    
    public boolean isWriter() {
        return (this.equals(WRITER)) ? true : false;
    }
    
    public boolean isOpenWriteMode() {
        boolean openWrite = false;
        if (isImport() ||
            isWriter() ||
            this.equals(WRITE_CORRUPT)) {
            openWrite = true;
        }
        return openWrite;
    }
    
    public boolean isOpen() {
        return (this.equals(CLOSED)) ? false : true;
    }
    /**
     * Determines if the XStream is in a corrupted state and must be
     * abandoned and closed. 
     * 
     * @return true if stream state is equal to the corrupt state
     */
    public boolean isCorrupt() {
        return (this.equals(READ_CORRUPT) 
                || this.equals(WRITE_CORRUPT)
                || this.equals(EXPORT_READ_CORRUPT) 
                || this.equals(IMPORT_WRITE_CORRUPT)) ?
            true : false;
    }

    /**
     * Retrieves the correct corrupt state for the current XStream state.
     * If the XStream is already corrupt, abandoned or closed then the same
     * state is returned. Do not want to transition from closed or abandoned
     * to corrupt.
     * @return the corresponding corrupt XStream state
     */
    public XStreamState getCorruptState() {
        XStreamState corruptState = this;
        if (isReader()) {
            corruptState = XStreamState.READ_CORRUPT;
        } else if (isWriter()) {
            corruptState = XStreamState.WRITE_CORRUPT;
        } else if (isImportWriter()) {
            corruptState = XStreamState.IMPORT_WRITE_CORRUPT;
        } else if (isExportReader()) {
            corruptState = XStreamState.EXPORT_READ_CORRUPT;
        }
        return corruptState;
    }
}
