/* Copyright (c) 2009, The Storage Networking Industry Association.
   (c) Copyright 2009 Hewlett-Packard Development Company, L.P.
   
 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference.query;

import org.snia.xam.QueryException;

/**
 * FactorNodes contain leaf expressions in the where expression tree.
 * Most common leaf nodes will be of the form "prop_name > 42" (or some such).
 * Some expressions contain attribute expressions which will refer to 
 * sub-attributes of a property, e.g. "typeof(prop_name) like 'text/csv'". 
 * In the case of a property attribute, the above expression would be formed
 * with two ParsedFactorNodes, as in the following:
 */
public class ParsedFactorNode
   implements FactorNode
{
   protected String           m_prop;

   protected int              m_value_ref; // What value to get from property

   protected int              m_op; // FUNCTION

   protected AnnotatedLiteral m_literal; // value for expression, null for function

   static int                 seq = 0;

   protected boolean          m_isParenthesis;

   public ParsedFactorNode()
   {
      super();
      seq++;
   }

   /**
    * Built when a attribute level function is referenced. 
    */
   public ParsedFactorNode(int valueOf, String prop)
   {
      this();
      m_prop = prop;
      m_value_ref = valueOf;
   }

   public int getValueRef()
   {
      return m_value_ref;
   }

   public int getOperator()
   {
      return m_op;
   }

   public String getProperty()
   {
      return m_prop;
   }

   public AnnotatedLiteral getExpression()
   {
      return m_literal;
   }

   // This is a bit of a hack to determine unary functions.
   public boolean isOperatorUnary()
   {
      return m_value_ref >= EXISTS;
   }

   public boolean isTerm()
   {
      return false;
   }

   public boolean isFactor()
   {
      return true;
   }

   public boolean isParenthesis()
   {
      return false;
   }

   public void setProperty(String property)
   {
      m_prop = property;
   }

   public void setOperator(int op)
   {
      m_op = op;
   }

   public void setValueRef(int op)
   {
      m_value_ref = op;
   }

   public void setExpression(AnnotatedLiteral expression)
   {
      m_literal = expression;
   }

   /**
    * Validate if the comparison operator is legal for the literal, e.g. no ">=" with boolean.
    * @throws QueryException
    */
   public void checkComparison() throws QueryException
   {
      if( m_literal.getType() == TypeCode.TYPE_BOOLEAN )
      {
         if( (m_op != EQUALS) && (m_op != NOT_EQUALS ))
            throw new QueryException( "Comparison operator not allowed with XAM_BOOLEAN: " + 
               operatorToString() );
      }

      if( m_literal.getType() == TypeCode.TYPE_XUID )
      {
         if( (m_op != EQUALS) && (m_op != NOT_EQUALS ))
            throw new QueryException( "Comparison operator not allowed with XAM_XUID: " + 
               operatorToString() );
      }
}
   
   /**
    * Also for debugging
    */
   String operatorToString()
   {
      switch (m_op)
      {
         case 0 :
            return "";
         case EQUALS :
            return "=";
         case NOT_EQUALS :
            return "!=";
         case LESS_THAN :
            return "<";
         case LESS_THAN_EQ :
            return "<=";
         case GREATER_THAN :
            return ">";
         case GREATER_THAN_EQ :
            return ">=";
         case LIKE :
            return "like";
         case CONTAINS:
            return "contains";
         case BEFORE:
            return "before";
         case AFTER:
            return "after";
         case WITHIN:
            return "within";
         default :
            return "?" + m_op;
      }
   }

   /** Return the value of expression. Not sure if it is worth using 
    * a String buffer for these expressions.
    */
   String valueOfToString()
   {
      switch (m_value_ref)
      {
         case VALUE_OF :
            return m_prop;
         case VALUE_TYPEOF :
            return "typeof(" + m_prop + ")";
         case VALUE_BINDING :
            return "binding(" + m_prop + ")";
         case VALUE_LENGTH :
            return "length(" + m_prop + ")";
         case VALUE_READONLY:
            return "readonly(" + m_prop + ")";
         case EXISTS :
            return "exists(" + m_prop + ")";
         default :
            return "?" + m_value_ref + "(" + m_prop + ")";
      }
   }

   /**
    * Mostly for Debugging.
    */
   public String toString()
   {
      StringBuffer sb = new StringBuffer("F(");
      sb.append(valueOfToString());
      if (m_op != 0)
      {
         sb.append(" ");
         sb.append(operatorToString());
         sb.append(" ");
         sb.append(m_literal);
      }
      sb.append(")");
      return sb.toString();
   }
}
