/* Copyright (c) 2009, The Storage Networking Industry Association.
   (c) Copyright 2009 Hewlett-Packard Development Company, L.P.
   
 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference.query;

import java.util.List;
import java.util.Properties;
import java.util.Vector;
import java.util.UUID;

/**
 * This class is emitted from the query parser and contains all the semantic parse tree 
 * as well as some of the other pieces of information required to execute the query.
 */
public class ParsedQuery
{
   protected UUID           m_id;

   protected Vector<String> m_props;
   
   // Where properties and exists properties are computed by SQL (in DBSqlTranslator)
   // So these structures simply record the field name and the value required (TypeCode).
   protected Properties     m_where_properties;
   protected Properties     m_exists_properties;  
   
   // Level 2 fields are the actual ParsedFactorNodes for each level 2 
   // subexpression. Because the expressions are computed externally,
   // they are handled differently than the where/exists properties. 
   // See DBSqlTranslator for more information. 
   protected Properties     m_level2_fields;     // These are handled differently than properties.

   protected String         m_query;

   protected long           m_start;

   protected WhereNode      m_where;

   protected int            m_level;

   public ParsedQuery(String query)
   {
      super();
      m_start = System.currentTimeMillis();
      m_id = UUID.randomUUID();
      m_query = query;
   }

   /**
    * Each time the query is parsed, it is supplied with a new UUID.
    * @return The UUID of this parsed query
    */
   public UUID getId()
   {
      return m_id;
   }

   /**
    * Get the selected properties for this query.
    * @return A list structure of the selected properties for this query.
    */
   public List<String> getSelectedPropertyList()
   {
      return m_props;
   }
   
   /**
    * Get the list of properties that are involved in the where clause. 
    * @return The properties construct with all properties involves in the
    * the where clause (except for those in exists functions).
    */
   public Properties getWherePropertyList()
   {
      return m_where_properties;
   }
   
   /** Get the list of properties that are present in an exists sub-clause.
    * @return The properties construct with all exists properties
    */
   public Properties getExistsPropertyList()
   {
      return m_exists_properties;
   }

   /** Get the list of level 2 expressions in the where clause.
    * @return The level 2 expressions used in the where clause.
    */
   public Properties getLevel2Fields()
   {
      return m_level2_fields;
   }
   
   /**
    * Return the where clause sub tree.
    * @return The where clause 
    */
   public WhereNode getWhereExpression()
   {
      return m_where;
   }

   /**
    * Get the XAM QL level of the query.
    * @return 1 if the syntax used only level 1 constructs, otherwise level 2.
    */
   public int getLevel()
   {
      return m_level;
   }

   /**
    * Return the time of the parse
    * @return Time in system long format.
    */
   public long getStartTime()
   {
      return m_start;
   }

   /**
    * Set the list of property names in the select clause
    * @param props Vector<String> of property names
    */
   protected void setSelectedPropertyList(Vector<String> props)
   {
      m_props = props;
   }
   
   /**
    * Set the list of property names in the where clause.
     */
   protected void setWherePropertyList(Properties props)
   {
      m_where_properties = props;
   }
   
   /**
    * Set the list of property names being used in the exists clauses.
    * @param props The Properties of exists properties.
    */
   protected void setExistsPropertyList(Properties props)
   {
      m_exists_properties = props;
   }
   
   /**
    * Set the list of level 2 expressions being used in the where clause.
    * @param fields The Fields used in the level 2 expression factor nodes
    */
   protected void setLevel2Fields(Properties fields)
   {
      m_level2_fields = fields;
   }

   /**
    * Sets  the where expression.
    * @param root Base node of the where clause, may be null.
    */
   public void setWhereExpression(WhereNode root)
   {
      m_where = root;
   }

   /**
    * Set the syntax level of the query.
    * @param level
    */
   protected void setLevel(int level)
   {
      m_level = level;
   }

   /**
    * Take the parsed query and reconstruct a string. Used mainly in debugging the parser.
    */
   public String toString()
   {
      StringBuffer buf = new StringBuffer(256);
      buf.append("SELECT ");
      int size = (m_props == null ? 0 : m_props.size());
      for (int i = 0; i < size; i++)
      {
         if (i > 0) buf.append(", ");
         buf.append((String) m_props.get(i));
      }

      if (m_where != null)
      {
         buf.append(" WHERE");
         toString(buf, m_where);
      }

      return buf.toString();
      //return m_query;
   }

   /**
    * Used for debuggging, provides the string version of a where node.
    * @param buf
    * @param node
    */
   private void toString(StringBuffer buf, WhereNode node)
   {
      if (node instanceof TermNode)
      {
         buf.append(" ");
         TermNode term = (TermNode) node;
         if (term.isNotExpression()) buf.append("NOT ");
         toString(buf, term.getLeftNode());
         switch (term.getOperator())
         {
            case TermNode.OPERATOR_AND :
               buf.append(" AND");
               break;
            case TermNode.OPERATOR_OR :
               buf.append(" OR");
               break;
         }
         node = term.getRightNode();
         if (node != null) toString(buf, node);
      }
      else if (node instanceof FactorNode)
      {
         FactorNode factor = (FactorNode) node;
         if (factor.isOperatorUnary())
         {
            toString(buf, factor.getOperator());
            buf.append(factor.getProperty());
         }
         else
         {
            buf.append(factor.getProperty());
            toString(buf, factor.getOperator());
            buf.append(factor.getExpression());
         }
      }
   }

   /**
    * Used for debugging, provides the string version of a factor.
    * @param buf
    * @param factorOp
    */
   private void toString(StringBuffer buf, int factorOp)
   {
      switch (factorOp)
      {
         case FactorNode.EXISTS :
            buf.append("EXISTS ");
            break;
         case FactorNode.EQUALS :
            buf.append(" = ");
            break;
         case FactorNode.NOT_EQUALS :
            buf.append(" != ");
            break;
         case FactorNode.GREATER_THAN :
            buf.append(" > ");
            break;
         case FactorNode.GREATER_THAN_EQ :
            buf.append(" >= ");
            break;
         case FactorNode.LESS_THAN :
            buf.append(" < ");
            break;
         case FactorNode.LESS_THAN_EQ :
            buf.append(" <= ");
            break;
         case FactorNode.LIKE :
            buf.append(" like ");
            break;
      }
   }
}
