/* Copyright (c) 2009, The Storage Networking Industry Association.
   (c) Copyright 2009 Hewlett-Packard Development Company, L.P.
   
 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference.query;

import java.util.Calendar;

import org.snia.xam.XAMLibrary;
import org.snia.xam.XUID;

/**
 * This enum is used by the Reference VIM to relate types to properties. 
 * This object can also be annotated by the explicit literal which was 
 * parsed from the query.
 * 
 * Generally this is used in AnnotatedLiteral to decorate the literal instance
 * and allows the SQL generator to decide which column the value should be retrieved from.
 */
public enum TypeCode {
   // Type codes, used for quick switch statements.
      TYPE_STRING  (XAMLibrary.STYPE_STRING_MIME_TYPE),  
      TYPE_DATETIME(XAMLibrary.STYPE_DATETIME_MIME_TYPE),
      TYPE_XUID    (XAMLibrary.STYPE_XUID_MIME_TYPE), 
      TYPE_INTEGER (XAMLibrary.STYPE_INT_MIME_TYPE), 
      TYPE_DOUBLE  (XAMLibrary.STYPE_DOUBLE_MIME_TYPE), 
      TYPE_BOOLEAN (XAMLibrary.STYPE_BOOLEAN_MIME_TYPE),
      
      // The following are pseudo types which will not appear as literals, but
      // serve to communicate to the sql generator which special column the value
      // should be retrieved from.
      // The string versions are used when creating property aliases so these values 
      // can be distinguished from regular property values.
      TYPE_TYPEOF  ("_TYP"),
      TYPE_LENGTH  ("_LEN"),
      TYPE_READONLY("_RO"),
      TYPE_BINDING ("_BND"),
      TYPE_EXISTS  ("_EXT"),
      
      // Level two expressions
      TYPE_CONTAINS("_CON"),
      TYPE_BEFORE  ("_BEF"),
      TYPE_AFTER   ("_AFT"),
      TYPE_WITHIN  ("_WTH"),
      
      TYPE_OTHER   ("?");   // Non Stype
      
      // Constants used to iterate
      public static final TypeCode FIRST = TYPE_STRING;
      public static final TypeCode LAST  = TYPE_EXISTS;
      
      public static final TypeCode LIST[] = { 
         TYPE_STRING,
         TYPE_DATETIME,
         TYPE_XUID,
         TYPE_INTEGER,
         TYPE_DOUBLE,
         TYPE_BOOLEAN,
         TYPE_TYPEOF,
         TYPE_LENGTH,
         TYPE_READONLY,
         TYPE_BINDING,
         TYPE_EXISTS };
      
      private String m_mime;
      
      /** built in constructor, not available to others. */
      private TypeCode( String m ) { m_mime = m; } 
      
      public String toString() { return m_mime; }
      
      /** 
       * Given an object, determine what type code represents the type of 
       * the object. 
       * @param o The object to be inspected.
       * @return The TypeCode
       */
      public static TypeCode valueOf( Object o ) {
         if( o instanceof String )
            return TYPE_STRING;
         
         else if( o instanceof Calendar )
            return TYPE_DATETIME;
         
         else if( o instanceof XUID )
            return TYPE_XUID;
         
         else if( o instanceof Long )
            return TYPE_INTEGER;
         
         else if( o instanceof Double )
            return TYPE_DOUBLE;
         
         else if( o instanceof Boolean )
            return TYPE_BOOLEAN;
         
         else 
            return TYPE_OTHER;
      }
      
      /**
       * Determines if the TypeCode is an attribute of a field.
       */
      public boolean isAttribute()
      {
         return this.equals(TypeCode.TYPE_BINDING) || this.equals(TypeCode.TYPE_READONLY) ||
                this.equals(TypeCode.TYPE_EXISTS)  || this.equals(TypeCode.TYPE_LENGTH )  ||
                this.equals(TypeCode.TYPE_TYPEOF);
      }
   };
   
