/*
 * Copyright (c) 2008, Hewlett-Packard Development Company, L.P. 
 * Copyright (c) 2008, The Storage Networking Industry Association.
 * Copyright (c) 2008, Sun Microsystems, Inc.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference.utils;


import java.io.ByteArrayOutputStream;
import java.io.IOException;

import java.io.UnsupportedEncodingException;
import org.snia.xam.XAMException;
import org.snia.xam.XSystem;

/**
 * SASL utilities useable by both the XAM library as well as
 * client applications for encoding and decoding SASL byte[]
 * communication. Included are two convenience methods for
 * autheticating either anonymously or using a username password
 * against an unauthenticated XSystem.
 */
public class ReferenceSaslUtils
{
    public static final byte   SASL_NULL_SEP = (byte)0;
    
    protected String saslAuthMechanism;
    public static final String SASL_MECHANISM_ANONYMOUS = "ANONYMOUS";
    public static final String SASL_MECHANISM_PLAIN =  "PLAIN";
    
    // private because not really immutable
    private static byte[] SASL_MECHANISM_PLAIN_BYTES =
            SASL_MECHANISM_PLAIN.getBytes();
    
    private static byte[] SASL_MECHANISM_ANONYMOUS_BYTES =
            SASL_MECHANISM_ANONYMOUS.getBytes();

    /* SASL Response codes */
    // little concerned about more response code being "" 
    // perhaps should change to "+"
    public static final String SASL_RESPONSE_MORE = "+";
    //public static final String SASL_MORE = "+";
    // Also would rather have this be something like DONE
    public static final String SASL_SUCCESS = "DONE";
    public static final String SASL_DONE = "DONE";
    public static final String SASL_FAILED = "FAIL";
    
    public static final String SASL_ENCODING_CHARSET = "UTF-8";
	

        /**
         * Convenience method for performing a SASL ANONYMOUS authentication
         * against an unauthenticated XSystem. If any error occurs during 
         * authentication a XAMException will be returned. The returned byte[] 
         * will contain the SASL success string and is only returned if 
         * authentication succeeds. It may be required, in the future, that we 
         * return failure information in the byte[] rather than throwing an
         * Exception.
         * 
         * @param system An unauthenticated XSystem
         * @param id The authentication id
         * @return SASL success string as a byte[]
         * @throws XAMException If any error occurs during authentication
         */
        public static byte[] authenticateAnonymous(XSystem system,
                        String id)
                throws XAMException {	
            system.authenticate(SASL_MECHANISM_ANONYMOUS_BYTES);
            try {
                byte[] creds = encodeSASLAnonymous(id);
                return system.authenticate(creds);
            } catch (UnsupportedEncodingException ex) {
                XAMException xex = new XAMException(
                        "UTF-8 encoding not supported.", ex);
                throw(xex);
            }
        }
        
        /**
         * Encodes a String id into a byte[] that can be passed to the
         * XSystem.authenticate(byte[], int) method during ANONYMOUS
         * authentication.
         * 
         * @param id The authentication id
         * @return A SASL encoded byte[] containing the authentication id
         * @throws UnsupportedEncodingException
         */
        public static byte[] encodeSASLAnonymous(String id) 
                throws UnsupportedEncodingException {
                if (id != null)
                        return id.getBytes(SASL_ENCODING_CHARSET);

                return new byte[0];
        }
	
        /**
         * Convenience method for performing a SASL PLAIN authentication
         * against an unauthenticated XSystem. If any error occurs during 
         * authentication a XAMException will be returned. The returned byte[] 
         * will contain the SASL success string and is only returned if 
         * authentication succeeds. It may be required, in the future, that we 
         * return failure information in the byte[] rather than throwing an
         * Exception.
         * 
         * @param system An unauthenticated XSystem
         * @param username The authentication username
         * @param password The authentication password
         * @return SASL success string as a byte[]
         * @throws XAMException If any error occurs during authentication
         */
        public static byte[] authenticatePlain(XSystem system,
                        String username, String password)
                throws XAMException   {
                system.authenticate(SASL_MECHANISM_PLAIN_BYTES);
                byte[] creds = encodeSASLPlain(null, username, password);
                return system.authenticate(creds);
        }

        /**
         * Encodes an authorization id, authentication username and
         * authentication password into a byte[] that can be passed to
         * the XSystem.authenticate(byte[], int) method during PLAIN
         * authentication.
         * 
         * Important: This method does not follow the SASLPrep profile 
         * fully when encoding strings.
         * 
         * @param authorizationId
         * @param authenticationId
         * @param password
         * @return SASL encoded byte array with parameter info
         */
        public static byte[] encodeSASLPlain(String authorizationId,
                        String authenticationId, String password) 
                    throws XAMException {
            ByteArrayOutputStream os = new ByteArrayOutputStream();
            try {
                if (authorizationId != null) {
                    os.write(authorizationId.getBytes(SASL_ENCODING_CHARSET));
                }
                os.write(SASL_NULL_SEP);

                if (authenticationId != null) {
                    os.write(authenticationId.getBytes(SASL_ENCODING_CHARSET));
                }
                os.write(SASL_NULL_SEP);

                if (password != null) {
                    os.write(password.getBytes(
                            SASL_ENCODING_CHARSET));
                }

                return os.toByteArray();

            } catch (UnsupportedEncodingException ex) {
                XAMException xex = new XAMException(
                        "UTF-8 encoding not supported.");
                throw(xex);                    
            }
            catch (IOException ex)
            {
                XAMException xex = new XAMException(
                        "IOException encountered.", ex);
                throw(xex);
            }

        }
	
        /**
         * Decodes a SASL PLAIN encoded byte[] into a String[]
         * containing the authorization id, authentication username
         * and password in that order.
         * 
         * @param bytes A SASL PLAIN encoded byte[]
         * @param offset Where the PLAIN auth information starts.
         * @param length The number of bytes encoded
         * @return A String[] containing the authorization id, 
         * authentication username and password
         * @throws XAMException If the byte[] is improperly encoded
         */
        public static String[] decodeSASLPlain(byte[] bytes, int offset, int length)
            throws XAMException    
        {
           int ptr_authzId  = offset;
           int ptr_username = -1;
           int ptr_password = -1;
           for( int i = bytes.length-1; i >= offset; i--)
           {
              if( bytes[i] == SASL_NULL_SEP ) {
                 if( ptr_password == -1 )
                    ptr_password = i + 1;
                 else if( ptr_username == -1 )
                    ptr_username = i + 1;
                 else if( ptr_authzId == 0 )
                 {
                    ptr_authzId = i + 1;
                    break;
                 }
              }
           }

           if ( (ptr_username == -1) || (ptr_password == -1))
              throw new XAMException("Invalid PLAIN authentication data");

           String[] strs = new String[3];
           try {
              strs[0] = new String(bytes, ptr_authzId, ptr_username-ptr_authzId-1, 
                        SASL_ENCODING_CHARSET);
              strs[1] = new String(bytes, ptr_username, ptr_password-ptr_username-1, 
                        SASL_ENCODING_CHARSET);
              strs[2] = new String(bytes, ptr_password, length-ptr_password, 
                        SASL_ENCODING_CHARSET);
           } catch (UnsupportedEncodingException ex) {
              XAMException xex = new XAMException(
              "UTF-8 encoding not supported.");
              throw(xex);                                
           }
           return strs;
        } 
        
        /**
         * Retrieves the SASL response code that indicates more data is needed.
         * 
         * @return byte[] contains SASL more response code
         */
        public byte[] getSASLResponseMore() {
            return SASL_RESPONSE_MORE.getBytes();
        }
}
