 /*
 *  Copyright (c) 2009, Sun Microsystems, Inc.
 *  Copyright (c) 2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED FBY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import org.snia.xam.PolicyNameException;
import org.snia.xam.XSet;
import org.snia.xam.XSetInaccessibleException;
import org.snia.xam.XSystem;
import org.snia.xam.XUID;

import junit.framework.Test;
import junit.framework.TestSuite;

/**
 * This test will test the access policies for application to read/write/delete
 * activities on XSets. Test of access policies with Level 1 query will be done
 * as part of the query test.
 * 
 * Note that this test relies upon knowledge of how access policies are created and used
 * inside of the ReferenceVIM. It is highly unlikely that these this test will be generally
 * useful in testing other XAM compliant systems.
 */
public class AccessPolicyTest
   extends ReferenceXAMTest
{
   public AccessPolicyTest(String name)
   {
      super(name);
   }

   public static void main(String[] args)
   {
      System.exit(runTest(args, suite()));
   }

   public static Test suite()
   {
      TestSuite suite = new TestSuite();
      suite.addTest(new AccessPolicyTest("testNoAccessPolicy"));
      suite.addTest(new AccessPolicyTest("test2AccessPolicy"));
      suite.addTest(new AccessPolicyTest("testSNIAAccessPolicy"));
      suite.addTest(new AccessPolicyTest("testIllegalAccessPolicy"));
      return suite;
   }

   /**
    * Create several XSets with out access policies (from the users), and make sure they can 
    * be applied to the XSet. 
    */
   public void testNoAccessPolicy()
      throws Exception
   {
      final String PROPERTY_NAME     = "com.example.access.test";
      final String TEST_VALUE1        = "foo1";
      final String TEST_VALUE2        = "foo2";
      
      XSystem xsys1 = getAuthenticatedXSystem(PASSWD_INFO[0][0], PASSWD_INFO[0][1] );
      XSet xset1 = xsys1.createXSet(XSet.MODE_UNRESTRICTED);
      xset1.createProperty(PROPERTY_NAME, false, TEST_VALUE1);
      XUID xset1XUID = xset1.commit();
      
      XSystem xsys2 = getAuthenticatedXSystem(PASSWD_INFO[1][0], PASSWD_INFO[1][1] );
      XSet xset2 = xsys2.createXSet(XSet.MODE_UNRESTRICTED);
      xset2.createProperty(PROPERTY_NAME, false, TEST_VALUE2);
      XUID xset2XUID = xset2.commit();
      
      // See if the XSets can be cross opened.
      XSet readable_xset2 = xsys1.openXSet(xset2XUID, XSet.MODE_UNRESTRICTED);
      assertEquals("TestXSet2 Example property not returned correctly",
         TEST_VALUE2,
         readable_xset2.getString(PROPERTY_NAME));
      
      XSet readable_xset1 = xsys2.openXSet(xset1XUID, XSet.MODE_UNRESTRICTED);
      assertEquals("TestXSet1 Example property not returned correctly",
         TEST_VALUE1,
         readable_xset1.getString(PROPERTY_NAME));
   }

   /**
    * Create several XSets with different access policies (from the users), and
    * make sure they can be applied to the XSet. Make sure that any policy names
    * not in access policy list should generate an error. These examples are
    * using the "user only" access policies.
    */
   public void test2AccessPolicy()
      throws Exception
   {
      final String PROPERTY_NAME      = "com.example.access.test";
      final String TEST_VALUE1        = "foo1";
      final String TEST_VALUE2        = "foo2";
      
      XSystem xsys1 = getAuthenticatedXSystem( PASSWD_INFO[0][0], PASSWD_INFO[0][1]  );
      XSet xset1 = xsys1.createXSet(XSet.MODE_UNRESTRICTED);
      xset1.createProperty( PROPERTY_NAME, false, TEST_VALUE1 );
      xset1.applyAccessPolicy( false, PASSWD_INFO[0][0] );
      XUID xset1XUID = xset1.commit();
      xset1.close();
      
      XSystem xsys2 = getAuthenticatedXSystem( PASSWD_INFO[1][0], PASSWD_INFO[1][1]  );
      XSet xset2 = xsys2.createXSet(XSet.MODE_UNRESTRICTED);
      xset2.createProperty( PROPERTY_NAME, false, TEST_VALUE2 );
      xset2.applyAccessPolicy( false, PASSWD_INFO[1][0] );
      XUID xset2XUID = xset2.commit();
      xset2.close();
      
      // See if the XSets can be cross opened.
      // Each should fail appropriately for all modes.
      try {
         XSet readable_xset2 = xsys1.openXSet( xset2XUID, XSet.MODE_UNRESTRICTED);
         fail( "XSet2 should not be openable from XSystem1 " + readable_xset2.getXUID(XSet.XAM_XUID) );
      } catch( XSetInaccessibleException xie ) { ; /* expected */ }
      
      try {
         XSet readable_xset1 = xsys2.openXSet( xset1XUID, XSet.MODE_UNRESTRICTED);
         fail( "XSet1 should not be openable from XSystem2 " + readable_xset1.getXUID(XSet.XAM_XUID));
      } catch( XSetInaccessibleException xie2 ) { ; /* expected */ }
   }

   /**
    * Create several XSets with different access policies (from the users), and make sure they can 
    * be applied to the XSet. One of the XSets need to have the predefined SniaReadAll or SniaReadWrite
    * access policies.
    */
   public void testSNIAAccessPolicy()
      throws Exception
   {
      final String PROPERTY_NAME     = "com.example.access.test";
      final String TEST_VALUE1        = "foo1";
      final String TEST_VALUE2        = "foo2";
      
      XSystem xsys1 = getAuthenticatedXSystem( PASSWD_INFO[0][0], PASSWD_INFO[0][1] );
      XSet xset1 = xsys1.createXSet(XSet.MODE_UNRESTRICTED);
      xset1.createProperty( PROPERTY_NAME, false, TEST_VALUE1 );
      xset1.applyAccessPolicy( false, AccessPolicyManager.STR_ACCESS_POLICY_READ_WRITE_ALL );
      XUID xset1XUID = xset1.commit();
      xset1.close();
      
      XSystem xsys_ra = getAuthenticatedXSystem( PASSWD_INFO[0][0], PASSWD_INFO[0][1] );
      XSet xset_ra = xsys_ra.createXSet(XSet.MODE_UNRESTRICTED);
      xset_ra.applyAccessPolicy( false, AccessPolicyManager.STR_ACCESS_POLICY_READ_ALL );
      XUID xset_raXUID = xset_ra.commit();
      xset_ra.close();
      
      XSystem xsys2 = getAuthenticatedXSystem( PASSWD_INFO[1][0], PASSWD_INFO[1][1] );
      XSet xset2 = xsys2.createXSet(XSet.MODE_UNRESTRICTED);
      xset2.createProperty( PROPERTY_NAME, false, TEST_VALUE2 );
      xset2.applyAccessPolicy( false, PASSWD_INFO[1][0] );
      XUID xset2XUID = xset2.commit();
      xset2.close();
      
      // See if the XSets can be cross opened by a third party
      XSystem xsys3 = getAuthenticatedXSystem( PASSWD_INFO[2][0], PASSWD_INFO[2][1] );

      // this should fail 
      try {
         XSet readable_xset2 = xsys3.openXSet( xset2XUID, XSet.MODE_UNRESTRICTED);
         fail( "XSet2 should not be openable from XSystem3 " + readable_xset2.getXUID(XSet.XAM_XUID));
      } catch( XSetInaccessibleException xie ) { ; /* expected */ }
      
      // XSet1 is set up for SniaReadWrite, so a foreign user should be able to see it in all modes.
      XSet readable_xset1_A = xsys3.openXSet( xset1XUID, XSet.MODE_UNRESTRICTED);
      readable_xset1_A.close();
      
      XSet readable_xset1_B = xsys3.openXSet( xset1XUID, XSet.MODE_RESTRICTED);
      readable_xset1_B.close();

      XSet readable_xset1_C = xsys3.openXSet( xset1XUID, XSet.MODE_READ_ONLY);
      readable_xset1_C.close();

      // XSet_ra is set up for SniaReadWrite, so a foreign user should be
      // able only read it.
      XSet readable_xset_ra_A = null;
      try {
         readable_xset_ra_A = xsys3.openXSet(xset_raXUID, XSet.MODE_UNRESTRICTED);
         fail("XSet_ra should not be openable from XSystem3 w/ unrestricted mode " + readable_xset_ra_A.getXUID(XSet.XAM_XUID));
      } catch (XSetInaccessibleException xie) { ; /* expected */ }
      finally{  
         if( readable_xset_ra_A != null )
            readable_xset_ra_A.close(); 
      }
      
      XSet readable_xset_ra_B = null;
      try {
         readable_xset_ra_B = xsys3.openXSet(xset_raXUID, XSet.MODE_RESTRICTED);
         fail("XSet_ra should not be openable from XSystem3 w/ restricted mode " + readable_xset_ra_B.getXUID(XSet.XAM_XUID));
      } catch (XSetInaccessibleException xie) { ; /* expected */ }
      finally {
         if( readable_xset_ra_B != null )
            readable_xset_ra_B.close();
      }

      XSet readable_xset_ra_C = xsys3.openXSet(xset_raXUID, XSet.MODE_READ_ONLY);
      readable_xset_ra_C.close();
   }

   /**
    * Create several XSets with out  access policies (from the users), and make sure they can 
    * be applied to the XSet. 
    */
   public void testIllegalAccessPolicy()
      throws Exception
   {
      XSystem xsys1 = getAuthenticatedXSystem( PASSWD_INFO[0][0], PASSWD_INFO[0][1] );
      XSet xset1 = xsys1.createXSet(XSet.MODE_UNRESTRICTED);
      try {
         xset1.applyAccessPolicy(false, "POLICY NAME NOT IN USER LIST" );
         fail( "Exected an exception during apply access policy" );
      }
      catch( PolicyNameException ipne ) { ; /* expected */ }
   }
}
