/*
 *  Copyright (c) 2009, Sun Microsystems, Inc.
 *  Copyright (c) 2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED FBY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XSet;
import org.snia.xam.XStream;
import org.snia.xam.XSystem;
import org.snia.xam.XUID;

import junit.framework.Test;
import junit.framework.TestSuite;

/**
 * This test will test the access policies for application to read/write/delete
 * activities on XSets. Test of access policies with Level 1 query will be done
 * as part of the query test.
 * 
 * Note that this test relies upon knowledge of how access policies are created and used
 * inside of the ReferenceVIM. It is highly unlikely that these this test will be generally
 * useful in testing other XAM compliant systems.
 */
public class AsynchronousOperationTest
   extends ReferenceXAMTest
{
   
   class xaListener implements XAsyncListener
   {
      long      xal_xopid;
      boolean   xal_complete;
      XSet      xal_xset;
      XStream   xal_xstream;
      XUID      xal_xuid;
      long      xal_bytes_written;
      long      xal_bytes_read;
      Throwable xal_exception;
      Throwable xal_fail;  // If set, the callback failed miserably.
      
      public void XAsyncCallback(XAsync async)
      {
         try
         {
            xal_xopid    = async.getXOPID();
            xal_complete = async.isComplete();
            xal_xset     = async.getXSet();
            xal_xstream  = async.getXStream();
            xal_xuid     = async.getXUID();
            xal_bytes_written = async.getBytesWritten();
            xal_bytes_read    = async.getBytesRead();
            try { async.getStatus(); } 
            catch( Throwable t ) { xal_exception = t; }
         }
         catch( Throwable t )
         {
            xal_fail = t;
         }  
      }
      
   }
   
   public AsynchronousOperationTest(String name)
   {
      super(name);
   }

   public static void main(String[] args)
   {
      System.exit(runTest(args, suite()));
   }

   public static Test suite()
   {
      TestSuite suite = new TestSuite();
      suite.addTest(new AsynchronousOperationTest("testXSystemXSetOps"));
      suite.addTest(new AsynchronousOperationTest("testXSetCommit"));
      suite.addTest(new AsynchronousOperationTest("testAsyncXStream"));
      return suite;
   }

   /**
    * Test the XSystem open/copy XSet asynchronous operations. Tests the polling versions, listeners tested 
    * using the xaListener. 
    */
   public void testXSystemXSetOps()
      throws Exception
   {
      final int    LIMIT         = 20; /* 2 seconds ought to be long enough! */
      final String PROPERTY_NAME = "com.example.async.open/copy";
      final String TEST_VALUE1   = "foo1";
      xaListener   openListener = new xaListener();
      xaListener   copyListener = new xaListener();
      
      s_xam.createProperty( ReferenceXSystem.REF_VIM_ASYNC_THREAD_COUNT, false, 2 );
      XSystem xsys1 = this.connectAndAuthenticate(true);
      XSet xset1 = xsys1.createXSet(XSet.MODE_UNRESTRICTED);
      xset1.createProperty(PROPERTY_NAME, false, TEST_VALUE1);
      XUID xset1XUID = xset1.commit();
      
      // Open the XSet.
      XAsync async = xsys1.asyncOpenXSet( xset1XUID, XSet.MODE_READ_ONLY, 1, openListener);
      int count = 0;
      while( ! async.isComplete() )
      {
         count ++;
         try { Thread.sleep(100); } catch( InterruptedException ie ) { /* ignore */ }
         if ( count > LIMIT )
            fail( "Timed out waiting for XAsync open XSet to complete" );
      }
      async.getStatus();
      assertEquals( "Asynchronous XUID should be equal to previous XUID", xset1XUID, async.getXUID() );
      assertTrue( "Asynchronous XSet didn't have the appropriate property",
                  async.getXSet().getString(PROPERTY_NAME).equals(TEST_VALUE1) );
      assertTrue( "Expected asynchronously open XSet to be different instance than the original",
                 xset1 != async.getXSet() );
      assertEquals( "Listener had bad XOPID", 1, openListener.xal_xopid );
      async = null;
      
      // Copy the XSet
      XAsync async2 = xsys1.asyncCopyXSet( xset1XUID, XSet.MODE_READ_ONLY, 2, copyListener);
      count = 0;
      while( ! async2.isComplete() )
      {
         count ++;
         try { Thread.sleep(100); } catch( InterruptedException ie ) { /* ignore */ }
         if ( count > LIMIT )
            fail( "Timed out waiting for XAsync open XSet to complete" );
      }
      XUID testXUID = null;
      try {
         testXUID = async2.getXSet().getXUID(XSet.XAM_XUID);
      }
      catch( FieldDoesNotExistException fdne ) { /* ignore */ }
      async2.getStatus();
      assertTrue( "Asynchronous XUID should NOT be be present in new XSet", 
         null == testXUID );
      assertTrue( "Asynchronous XSet didn't have the appropriate property",
                  async2.getXSet().getString(PROPERTY_NAME).equals(TEST_VALUE1) );
      assertTrue( "Expected asynchronously open XSet to be different instance than the original",
                 xset1 != async2.getXSet() );   
      assertEquals( "Listener had bad XOPID", 2, copyListener.xal_xopid );
   }

   /**
    * Test the XSet asyncCommit. Tests the polling versions, listeners tested 
    * using the xaListener. 
    */
   public void testXSetCommit()
      throws Exception
   {
      final int    LIMIT         = 20; /* 2 seconds ought to be long enough! */
      final String PROPERTY_NAME = "com.example.async.commit-xset";
      final String TEST_VALUE1   = "foo1";
      xaListener   commitListener = new xaListener();
      
      XSystem xsys1 = this.connectAndAuthenticate(true);
      XSet xset1 = xsys1.createXSet(XSet.MODE_UNRESTRICTED);
      xset1.createProperty(PROPERTY_NAME, false, TEST_VALUE1);
      
      // Commit
      XAsync async = xset1.asyncCommit(3, commitListener);
      int count = 0;
      while( ! async.isComplete() )
      {
         count ++;
         try { Thread.sleep(100); } catch( InterruptedException ie ) { /* ignore */ }
         if ( count > LIMIT )
            fail( "Timed out waiting for XAsync XSet commit to complete" );
      }
      async.getStatus();
      assertNotNull( "Asynchronous XUID should have been returned", async.getXUID() );
      XSet xset2 = xsys1.openXSet(async.getXUID(), XSet.MODE_READ_ONLY);
      assertTrue( "Asynchronous XSet didn't have the appropriate property",
                 xset2.getString(PROPERTY_NAME).equals(TEST_VALUE1) );
      assertEquals( "Listener had bad XOPID", 3, commitListener.xal_xopid );
      async = null;
   }

   /**
    * Test the openXStream call from XSet (won't work from XSystem), tests the polling version 
    * then tests using the xaListener. 
    */
   public void testAsyncXStream()
      throws Exception
   {
      final String PROPERTY_NAME = "com.example.async.open-xstream";
      final String TEST_VALUE1   = "foo1";
      final String STREAM_NAME   = "com.example.stream";
      final int    SIZE          = 42;
      xaListener   commitListener = new xaListener();
      xaListener   streamListener = new xaListener();
      xaListener   readListener   = new xaListener();
      xaListener   closeListener  = new xaListener();
      xaListener   openListener   = new xaListener();
      
      XSystem xsys1 = this.connectAndAuthenticate(true);
      XSet xset1 = xsys1.createXSet(XSet.MODE_UNRESTRICTED);
      xset1.createProperty(PROPERTY_NAME, false, TEST_VALUE1);
      XStream testStream = xset1.createXStream(STREAM_NAME, false, XAMLibrary.DEFAULT_MIME_TYPE);
      byte buffer[] = new byte[SIZE];
      for( byte i = 0; i < SIZE; i++ )
         buffer[i] = i;
      
      // Write the stream
      XAsync async = testStream.asyncWrite( buffer, SIZE, 4, streamListener );
      int count = 0;
      waitForXA( async, "Timed out waiting for XAsync XStream.write to complete" );

      async.getStatus();
      assertEquals( "Not enough bytes written in XStream.asyncWrite", 
                    SIZE, streamListener.xal_bytes_written );
      async.getStatus(); // Should just return
      async = null;
      
      // close the XSetream
      XAsync async1_5 = testStream.asyncClose( 15, closeListener);
      count = 0;
      waitForXA( async1_5, "Timed out waiting for XAsync XSet commit to complete" );

      async1_5.getStatus(); // Should just return
      assertEquals( "Listener had bad XOPID", 15, closeListener.xal_xopid );
      async1_5 = null;
      
      // Repoen the XStream with an asynchronous operation
      
      XAsync async1_6 = xset1.asyncOpenXStream(STREAM_NAME, XStream.MODE_READ_ONLY, 16, openListener );
      waitForXA( async1_6, "Timed out waiting for XAsync open XStream to complete" );
      async1_6.getStatus();  // Should return 
      
      byte bits[] = new byte[10];
      XStream stream1_6 = openListener.xal_xstream;
      stream1_6.read(bits);
      for( int b = 0; b < bits.length; b++ )
         assertEquals( "Read wrong bits from asyncrhonously opened XStream", b, bits[b] );
      stream1_6.close();
      
      // Commit the XSet.
      XAsync async2 = xset1.asyncCommit( 5, commitListener);
      count = 0;
      waitForXA( async2, "Timed out waiting for XAsync XSet commit to complete" );

      async2.getStatus(); // Should just return
      assertNotNull( "Asynchronous XUID should have been returned", async2.getXUID() );
      XSet xset2 = xsys1.openXSet(async2.getXUID(), XSet.MODE_READ_ONLY);
      assertTrue( "Asynchronous XSet didn't have the appropriate property",
                 xset2.getString(PROPERTY_NAME).equals(TEST_VALUE1) );
      assertEquals( "Listener had bad XOPID", 5, commitListener.xal_xopid );
      async2 = null;
      
      // Read the bytes from the stream
      XStream readStream = xset2.openXStream(STREAM_NAME, XStream.MODE_READ_ONLY);
      final int SIZE2 = 1024;
      byte buffer2[] = new byte[SIZE2];
      XAsync async3 = readStream.asyncRead(buffer2, SIZE2, 6, readListener);
      count = 0;
      waitForXA( async3, "Timed out waiting for XAsync XSet commit to complete" );

      async3.getStatus(); // Should just return
      assertEquals( "Didn't read enough bytes from the stream", SIZE, readListener.xal_bytes_read );
      for( byte x = 0; x < SIZE; x++ )
         assertEquals( "Buffer value read incorrectly from stream", x, buffer2[x] );
   }
   
   private void waitForXA( XAsync xa, String msg ) throws XAMException
   {
      final int LIMIT = 20; /* 2 seconds ought to be long enough! */
      
      int count = 0;
      while( ! xa.isComplete() )
      {
         count ++;
         try { Thread.sleep(100); } catch( InterruptedException ie ) { /* ignore */ }
         if ( count > LIMIT )
            fail( msg );
      }
   }
}
