/*
 *  Copyright (c) 2008, Sun Microsystems, Inc.
 *  Copyright (c) 2008, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.util.Calendar;
import junit.framework.Test;
import junit.framework.TestSuite;
import org.snia.xam.XAMException;
import org.snia.xam.XSet;
import org.snia.xam.XStream;
import org.snia.xam.XSystem;
import org.snia.xam.XUID;

public class DeleteXSetTest extends ReferenceXAMTest {

    public DeleteXSetTest(String name) {
        super(name);
    }

    public static void main(String[] args) {
        System.exit(runTest(args, suite()));
    }

    public static Test suite() {
        TestSuite suite = new TestSuite();
        suite.addTest(new DeleteXSetTest("testBasicDelete"));
        suite.addTest(new DeleteXSetTest("testDeleteWithJobRunning"));
        suite.addTest(new DeleteXSetTest("testDeleteWithRetention"));
        return suite;
    }
    /**
     * Tests deleting an XSet that is open (fails) as well as closed (succeeds) 
     * @throws java.lang.Exception
     */
    public void testBasicDelete() throws Exception {
        announce("DeleteXSetTest.testBasicDelete");
        XSystem xsystem = this.connectAndAuthenticate(true);
        XSet xset1 = xsystem.createXSet(XSet.MODE_UNRESTRICTED);
        xset1.createProperty("basic.delete.test.property", false, Calendar.getInstance());
        XStream xstream = ReferenceTestUtilities.createXStream(
            xset1, "basic.delete.test.stream", true, "image/jpg", (byte) 'c', 2000);
        xstream.close();
        XUID xuid1 = xset1.commit();
        try {
            xsystem.deleteXSet(xuid1);
            fail("XSet is open so deleting it should not be possible.");
        } catch (Exception e) {
//            announce("XSet is open on delete -- expected exception");
        }
        xset1.close();
        xsystem.deleteXSet(xuid1);
        xsystem.close();
    }
    /**
     * Tests deleting an XSet used for job submission in which the job is
     * still running (fails) and when the job finishes (succeeds).
     * @throws java.lang.Exception
     */
    public void testDeleteWithJobRunning() throws Exception {
        announce("DeleteXSetTest.testDeleteWithJobRunning");
        XSystem xsystem = this.connectAndAuthenticate(false);
        XSet xset = xsystem.createXSet(XSet.MODE_UNRESTRICTED);
        xset.createProperty(XSet.XAM_JOB_COMMAND, false, XSet.XAM_JOB_QUERY);
        // introduce a 2 sec. delay to ensure the job runs long enough to
        // try and delete the corresponding XSet
        xset.createProperty(ReferenceJob.JOB_DELAY_FIELD_IN_MILLISEC, false, 2000);
        XStream stream = xset.createXStream(XSet.XAM_JOB_QUERY_COMMAND,
                                                   false, "text/plain");
        byte[] bytes = ReferenceQueryTest.SELECT_ALL_QUERY.getBytes();
        int wrote = 0;
        while (wrote < bytes.length) {
            wrote += stream.write(bytes, wrote, bytes.length - wrote);
        }
        stream.close();
        stream = null;
        XUID xuid = xset.commit();
        
        xset.submitJob();
        // wait for job to start
        // wait until job has started running
        int numTries = 0;
        String jobStatusString = xset.getString(XSet.XAM_JOB_STATUS);
        while (numTries < 250 &&
               !JobStatus.isDoneExecution(jobStatusString) &&
               !JobStatus.isRunning(jobStatusString)) {
            Thread.sleep(1);
            numTries++;
            jobStatusString = xset.getString(XSet.XAM_JOB_STATUS);
        }
        
        if (!JobStatus.isDone(xset.getString(XSet.XAM_JOB_STATUS))) {
            try {
                // Attempting to delete XSet with running job
                xsystem.deleteXSet(xuid);
                fail("Job is running against XSet so deleting it " +
                        "should not be possible.");
            } catch (XAMException ex) {
                /* announce("Expected exception when trying to " +
                        "delete an XSet for which a job is running."); */
            }
        }
        // wait for job to complete
        do {
            // after sleeping for one second, check if job has completed
            Thread.sleep(XAMConstants.MILLISEC_PER_SECOND);
        } while (!JobStatus.isDone(xset.getString(XSet.XAM_JOB_STATUS)));
        
        // job has completed but wait a another second
        Thread.sleep(XAMConstants.MILLISEC_PER_SECOND);
        
        // Close with no commit means the results will be lost
        xset.close();
        xset = null;
        Thread.sleep(XAMConstants.MILLISEC_PER_SECOND);
        xsystem.deleteXSet(xuid);
        xsystem.close(); 
    }
    /**
     * Tests deleting an XSet that is retained (fails) and not
     * retained (succeeds) 
     * @throws java.lang.Exception
     */
    public void testDeleteWithRetention() throws Exception {
        announce("DeleteXSetTest.testDeleteWithRetention");
        XSystem xsystem = this.connectAndAuthenticate(true);
        XSet xset = xsystem.createXSet(XSet.MODE_UNRESTRICTED);
        // setting the duration value to a positive integer (5 seconds)
        xset.setBaseRetention(false, 2000);
        XUID xuid = xset.commit();
        assertEquals(xsystem.isXSetRetained(xuid), true);
        xset.close();
        try {
            xsystem.deleteXSet(xuid);
            fail("XSet is retained so deleting it should not be possible.");
        } catch (Exception ex) {
//            announce("Expected exception upon set deletion since " +
//                    "the set is under retention -- ");
        }
        try {
            Thread.sleep(2000);
        } catch (InterruptedException ex) {
            // announce("Interrupted Exception thrown:  " + ex.getMessage());
        }
        assertEquals(xsystem.isXSetRetained(xuid), false);
        xsystem.deleteXSet(xuid);
        xsystem.close(); 
    }
    
}
