/*
 *  Copyright (c) 2009, Sun Microsystems, Inc.
 *  Copyright (c) 2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED FBY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.io.File;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;

import org.snia.xam.XSet;
import org.snia.xam.XSetInaccessibleException;
import org.snia.xam.XSystem;
import org.snia.xam.XUID;

import junit.framework.Test;
import junit.framework.TestSuite;

/**
 * This test will verifies the delete behavior of XSets in the system. Basic delete has been
 * tested and retention protect has been tested elsewhere. This tests the autodelete and shred
 * settings. 
 * 
 * NOTE: It's possible to tell if autodelete has worked, but shred isn't testable at the
 * API, only the settings of the values. A low level unit test on XSet/XStream is needed to 
 * verify shred behavior. It's also important to realize that the "shred" implemented in
 * this ReferenceVIM is available for exercising the API, not to implement any sort of 
 * secure data erase.
 */
public class DispositionTest
   extends ReferenceXAMTest
{
   // Used for the fast autodelete daemon
   static String           m_fast_store_path = null;
   static XSystem          m_fast_xsys       = null;
   
   String m_oldXRI;
   
   public DispositionTest(String name)
   {
      super(name);
   }

   public static void main(String[] args)
   {
      System.exit(runTest(args, suite()));
   }

   public static Test suite()
   {
      // Make sure that the last test in the list executes tearDownSubStore
      TestSuite suite = new TestSuite();
      suite.addTest(new DispositionTest("testXSystemSettings"));
      suite.addTest(new DispositionTest("testXSetSettings"));
      suite.addTest(new DispositionTest("testAutodelete"));
      suite.addTest(new DispositionTest("testAutodeleteWithRetention"));
      suite.addTest(new DispositionTest("testAutodeleteWithHold"));
      suite.addTest(new DispositionTest("testAutodeletePolicy"));
      suite.addTest(new DispositionTest("testAutodeleteWithEventRetention"));
      return suite;
   }

   /**
    * Create several XSets with out  access policies (from the users), and make sure they can 
    * be applied to the XSet. 
    */
   public void testXSystemSettings()
      throws Exception
   {
      XSystem xsys = this.doConnection(false);
      assertTrue( "Expected to see autodelete supported on the ReferenceXSystem",
         xsys.getBoolean( XSystem.XAM_DELETION_AUTODELETE ));
      
      assertTrue( "Expected to see shred supported on the ReferenceXSystem",
         xsys.getBoolean( XSystem.XAM_DELETION_SHRED ));
   }
   
   /**
    * Tests the ability to set the autodelete and shred settings on an XSet.
    * @throws Exception 
    */
   public void testXSetSettings() throws Exception
   {
      XSystem xsys = this.getAuthenticatedXSystem( PASSWD_INFO[0][0], PASSWD_INFO[0][1] );
      XSet    xset = xsys.createXSet( XSet.MODE_UNRESTRICTED );
      xset.setAutoDelete( false, true );
      xset.setShred( false, true );
      xset.commit();
      
      assertTrue( "Expected autodelete to be set on xset", xset.getBoolean(XSet.XAM_DELETION_AUTODELETE));

      assertTrue( "Expected shred to be set on xset", xset.getBoolean(XSet.XAM_DELETION_SHRED));
      xset.close();
   }
   
   /**
    * Test to make sure the disposition manager does the right thing. Create an XSet and wait for the
    * Disposition manager autodelete daemon to kick in and delete it. This is a tricky problem because 
    * all of the other tests have been running using the same directory. The autodelete daemons are 
    * created as a single instance per store directory, and it's likely (OK, guaranteed) that an 
    * instance exists, so the period is already set. 
    * 
    * Also set the shred setting. If a fatal error occurs during disposition, the test should catch
    * it. This doesn't test that the file is truly shredded though.
    * 
    * This test uses a special instance of a store, configured with a fast autodelete daemon.
    * Because this test created a unique daemon, the routine tries to test as many conditions in this 
    * one test. If the tests aren't packed together, it could result in lots of autodelete daemons 
    * left lying around, consuming resources from the test runner's VM. 
    * 
    * To test this, we'll use a specific subdirectory of the /tmp, which forces a new instance of
    * the auto delete daemon. 
    * 
    * @throws Exception 
    */
   public void testAutodelete() throws Exception
   {
      String newStore = setupSubStore(m_fast_store_path);
      announce( "testAutodelete @ " + newStore );
      
      try {
         s_xam.createProperty( ReferenceXSystem.PROPERTY_AUTODELETE_PERIOD, false, 1 );
         if( m_fast_xsys == null )
            m_fast_xsys = this.getAuthenticatedXSystem( PASSWD_INFO[0][0], PASSWD_INFO[0][1] );
         
         // SUB TEST 1: Simple XSet, no retention, no hold.
         // Basic XSet, should autodelete quickly.
         XSet    xset = m_fast_xsys.createXSet( XSet.MODE_UNRESTRICTED );
         xset.setAutoDelete( false, true );
         xset.setShred( false, true );
         XUID xuid = xset.commit();
      
         assertTrue( "Expected autodelete to be set on xst", xset.getBoolean(XSet.XAM_DELETION_AUTODELETE));
         xset.close();
      
         try { Thread.sleep(2000); } 
         catch( InterruptedException ie ) { /* ignore it */ }

         try {
            XSet test = m_fast_xsys.openXSet(xuid, XSet.MODE_READ_ONLY);
            test.close();
            fail( "Autodelete should have deleted the XSet " + xuid.toString() );
         }
         catch( XSetInaccessibleException xie ) { /* ignore it */ }
         
      }
      finally {
         s_xri = m_oldXRI;
         s_xam.deleteField( ReferenceXSystem.PROPERTY_AUTODELETE_PERIOD );
      }
   }

   /**
    * Test to make sure that autodelete does not dispose the XSet under retention protection.
    *    
    * This test uses a special instance of a store, configured with a fast autodelete daemon.
    * Because this test created a unique daemon, the routine tries to test as many conditions in this 
    * one test. If the tests aren't packed together, it could result in lots of autodelete daemons 
    * left lying around, comsuming resources from the test runner's VM. 
    * 
    * To test this, we'll use a specific subdirectory of the /tmp, which forces a new instance of
    * the auto delete daemon. 
    * 
    * @throws Exception
    */
   public void testAutodeleteWithRetention() throws Exception
   {
      String newStore = setupSubStore(m_fast_store_path);
      announce( "testAutodeleteWithRetention @ " + newStore );

      try
      {
         s_xam.createProperty( ReferenceXSystem.PROPERTY_AUTODELETE_PERIOD, false, 1 );
         if( m_fast_xsys == null )
            m_fast_xsys = this.getAuthenticatedXSystem( PASSWD_INFO[0][0], PASSWD_INFO[0][1] );

         // SUB TEST 2 : Retention
         // XSet with retention
         XSet xset2 = m_fast_xsys.createXSet( XSet.MODE_UNRESTRICTED );
         xset2.setBaseRetention( false, 1000 );
         xset2.setAutoDelete( false, true );
         XUID xuid2 = xset2.commit();

         try { Thread.sleep(750); }
         catch( InterruptedException ie ) { /* ignore */ }
         XSet xsettmp = m_fast_xsys.openXSet( xuid2, XSet.MODE_READ_ONLY );
         // If it gets here, the XSet still exists.
         xsettmp.close();

         try { Thread.sleep(1500); }
         catch( InterruptedException ie ) { /* ignore */ }

         // Should be deleted at this point.
         try { 
            XSet xsettmp2 = m_fast_xsys.openXSet( xuid2, XSet.MODE_UNRESTRICTED ); 
            fail( "XSet2 in testAutodelete should have expired and been autodeleted." );
            xsettmp2.close();
         }
         catch( XSetInaccessibleException xie ) { /* Expected exception */ }
      }
      finally
      {
         s_xri = m_oldXRI;
         s_xam.deleteField( ReferenceXSystem.PROPERTY_AUTODELETE_PERIOD );
      }
   }
   
   /**
    * Test to make sure that autodelete will be stopped if the XSet is on hold.
    * 
    * This test uses a special instance of a store, configured with a fast autodelete daemon.
    * Because this test created a unique daemon, the routine tries to test as many conditions in this 
    * one test. If the tests aren't packed together, it could result in lots of autodelete daemons 
    * left lying around, comsuming resources from the test runner's VM. 
    * 
    * To test this, we'll use a specific subdirectory of the /tmp, which forces a new instance of
    * the auto delete daemon. 
    * 
    * @throws Exception
    */
   public void testAutodeleteWithHold() throws Exception
   {
      String newStore = setupSubStore(m_fast_store_path);
      announce( "testAutodeleteWithHold @ " + newStore );

      try
      {
         s_xam.createProperty( ReferenceXSystem.PROPERTY_AUTODELETE_PERIOD, false, 1 );
         if( m_fast_xsys == null )
            m_fast_xsys = this.getAuthenticatedXSystem( PASSWD_INFO[0][0], PASSWD_INFO[0][1] );

         // SUB TEST 3: Retention extended via hold.
         // XSet with retention + hold (past retention time)
         XSet xset3 = m_fast_xsys.createXSet( XSet.MODE_UNRESTRICTED );
         xset3.setBaseRetention( false, 1000 );
         xset3.setAutoDelete( false, true );
         XUID xuid3 = xset3.commit();
         
         try { Thread.sleep(750); }
         catch( InterruptedException ie ) { /* ignore */ }
         XSet xsettmp3 = m_fast_xsys.openXSet( xuid3, XSet.MODE_READ_ONLY );
         // If it gets here, the XSet still exists.
         xsettmp3.close();
         m_fast_xsys.holdXSet( xuid3, "testAutodelete") ;
         
         try { Thread.sleep(1000); }
         catch( InterruptedException ie ) { /* ignore */ }
         
         // Should NOT be deleted at this point.
         xsettmp3 = m_fast_xsys.openXSet( xuid3, XSet.MODE_READ_ONLY ); 
         xsettmp3.close();
         
         // release then see what happens.
         m_fast_xsys.releaseXSet(xuid3, "testAutodelete" );
         try { Thread.sleep(1500); }
         catch( InterruptedException ie ) { /* ignore */ }
         
         try { 
            xsettmp3 = m_fast_xsys.openXSet( xuid3, XSet.MODE_UNRESTRICTED ); 
            fail( "XSet2 in testAutodelete should have expired and been autodeleted." );
            xsettmp3.close();
         }
         catch( XSetInaccessibleException xie ) { /* Expected exception */ }
      }
      finally
      {
         s_xri = m_oldXRI;
         s_xam.deleteField( ReferenceXSystem.PROPERTY_AUTODELETE_PERIOD );
      }
   }
   
   /**
    * Create an XSet, apply an autodelete policy, test to make sure it will dispose properly.
    * @throws Exception
    */
   public void testAutodeletePolicy() throws Exception
   {
      String newStore = setupSubStore(m_fast_store_path);
      announce( "testAutodeletePolicy @ " + newStore );
      
      try {
         s_xam.createProperty( ReferenceXSystem.PROPERTY_AUTODELETE_PERIOD, false, 1 );
         if( m_fast_xsys == null )
            m_fast_xsys = this.getAuthenticatedXSystem( PASSWD_INFO[0][0], PASSWD_INFO[0][1] );
         
         // Basic XSet, should autodelete quickly.
         XSet    xset = m_fast_xsys.createXSet( XSet.MODE_UNRESTRICTED );
         xset.applyAutoDeletePolicy( false, DispositionPolicy.DISPOSITION_AUTODELETE );
         XUID xuid = xset.commit();
      
         assertEquals( "Expected autodelete policy to be set on xst", 
            DispositionPolicy.DISPOSITION_AUTODELETE,
            xset.getString(XSet.XAM_DELETION_AUTODELETE_POLICY));
         xset.close();
      
         try { Thread.sleep(2000); } 
         catch( InterruptedException ie ) { /* ignore it */ }

         try {
            XSet test = m_fast_xsys.openXSet(xuid, XSet.MODE_READ_ONLY);
            test.close();
            fail( "Autodelete should have deleted the XSet " + xuid.toString() );
         }
         catch( XSetInaccessibleException xie ) { /* ignore it */ }
         
      }
      finally {
         s_xri = m_oldXRI;
         s_xam.deleteField( ReferenceXSystem.PROPERTY_AUTODELETE_PERIOD );
      }
   }
   
   /**
    * Test to make sure that autodelete does not dispose the XSet under retention protection.
    * Autodelete should only occur once the event has been signaled.
    *    
    * To test this, we'll use a specific subdirectory of the /tmp, which forces a new instance of
    * the auto delete daemon. 
    * 
    * @throws Exception
    */
   public void testAutodeleteWithEventRetention() throws Exception
   {
      final String EVENT_RET = "autodelete.retention";
         
      String newStore = setupSubStore(m_fast_store_path);
      announce( "testAutodeleteWithRetention @ " + newStore );

      try
      {
         s_xam.createProperty( ReferenceXSystem.PROPERTY_AUTODELETE_PERIOD, false, 1 );
         if( m_fast_xsys == null )
            m_fast_xsys = this.getAuthenticatedXSystem( PASSWD_INFO[0][0], PASSWD_INFO[0][1] );

         // SUB TEST 2 : Retention
         // XSet with retention
         XSet xset2 = m_fast_xsys.createXSet( XSet.MODE_UNRESTRICTED );
         xset2.createRetention( false, EVENT_RET );
         xset2.setRetentionDuration(EVENT_RET, false, 10);
         xset2.setRetentionEnabledFlag(EVENT_RET, false, true );
         xset2.setAutoDelete( false, true );
         XUID xuid2 = xset2.commit();

         try { Thread.sleep(3000); }
         catch( InterruptedException ie ) { /* ignore */ }
         XSet xsettmp = m_fast_xsys.openXSet( xuid2, XSet.MODE_READ_ONLY );
         // If it gets here, the XSet still exists.
         xsettmp.close();
         
         // This should start the retention, and THEN it should be auto deleted.
         xset2.setRetentionStarttime(EVENT_RET, false);
         xset2.commit();
         
         try { Thread.sleep(1000); }
         catch( InterruptedException ie ) { /* ignore */ }

         // Should be deleted at this point.
         try { 
            XSet xsettmp2 = m_fast_xsys.openXSet( xuid2, XSet.MODE_UNRESTRICTED ); 
            fail( "XSet2 in testAutodelete should have expired and been autodeleted." );
            xsettmp2.close();
         }
         catch( XSetInaccessibleException xie ) { /* Expected exception */ }
      }
      finally
      {
         s_xri = m_oldXRI;
         s_xam.deleteField( ReferenceXSystem.PROPERTY_AUTODELETE_PERIOD );
         tearDownSubStore(m_fast_store_path);
      }
   }
   
   /**
    * Create a randomized (sort of) sub directory of the current tmp directory, which will
    * be used for a brand new Store for the autodelete test. This will force a new instance
    * of the autodelete daemon.
    * @return
    * @throws UnsupportedEncodingException 
    */
   private String setupSubStore( String oldSubStore ) throws UnsupportedEncodingException
   {
      String subStoreString = null;
      
      if( oldSubStore != null ) 
         subStoreString = oldSubStore;
      else
      {
         String fileSeparator = System.getProperty("file.separator");
         String tmpDir        = System.getProperty("java.io.tmpdir");
      
            if (tmpDir == null) {
               tmpDir = fileSeparator + "tmp";
            }
      
            if ( !(tmpDir.endsWith((fileSeparator)))) {
               tmpDir = tmpDir + fileSeparator;   
            }

            File subStore = new File( tmpDir + "tmp" + System.currentTimeMillis() );
            subStore.mkdir();
            subStoreString = subStore.getAbsolutePath();
            if( subStoreString.endsWith(fileSeparator))
               subStoreString = subStoreString + fileSeparator;
     }
      subStoreString = URLEncoder.encode(subStoreString, "utf-8" );
       
      m_fast_store_path = subStoreString;
      m_oldXRI = s_xri;
      s_xri = "snia-xam://SNIA_Reference_VIM!localhost?dir=" + subStoreString;
      return subStoreString;
   }

   /**
    * Called by the final test in this suite, deletes the fast store directory.
    * @param The 
    */
   private void tearDownSubStore(String path )
   {
      this.deleteDirectory( new File(path) );
   }


   // TODO: Test autodelete with event retention
}

