/*
 * Copyright (c) 2008-2009, Sun Microsystems, Inc.
 * Copyright (c) 2008-2009, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import junit.framework.Test;
import junit.framework.TestSuite;

import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.AuthorizationException;
import org.snia.xam.FieldContainer;
import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.PolicyNameException;
import org.snia.xam.RetentionValueException;
import org.snia.xam.XAMException;
import org.snia.xam.XIterator;
import org.snia.xam.XSet;
import org.snia.xam.XSetCorruptException;
import org.snia.xam.XSetException;
import org.snia.xam.XStream;
import org.snia.xam.XSystem;
import org.snia.xam.XSystemAbandonException;
import org.snia.xam.XSystemCorruptException;
import org.snia.xam.XUID;

/**
 *  Test suite for Exporting/Importing XSet
 *
 */
public class ExportImportTest extends ReferenceXAMTest {
    
   // XSet is exported out of the system, and later re-imported.
   static private String exportedFilename = "";
   static private String exportedStreamsFilename = "";
   static private String mungedPolicyExport = "";
   static private String badPolicyFile = "";
   static private String badRetentionIDFile = "";
   static private String dispositionPolicyExport = "";
   static private String badDispositionPolicyExport = "";
   static private String badShredPolicyFile = "";
   static private String exportedSimpleFilename = "";
   static private String badDurationPolicyFile = "";
   static private String badEnabledPolicyFile = "";
   static private String bindingExportFile = "";
   static private String changedBindingFileName = "";
   static private String reducedExportFilename = "";
   static private String reducedImportFilename = "";
    
    public ExportImportTest(String name) {
        super(name);
    }

    public static void main(String[] args) {
        System.exit(runTest(args, suite()));
    }
    
    public static Test suite() {
    	TestSuite suite = new TestSuite();
    	suite.addTest(new ExportImportTest("testSimpleExport"));
        suite.addTest(new ExportImportTest("testSimpleImport"));
        suite.addTest(new ExportImportTest("testSimpleStreamExport"));
        suite.addTest(new ExportImportTest("testStreamsExport"));
        suite.addTest(new ExportImportTest("testSimpleStreamImport"));
//        suite.addTest(new ExportImportTest("testContributedImports"));
        suite.addTest(new ExportImportTest("testImportErrorChecks"));
        suite.addTest(new ExportImportTest("testRetentionPolicies"));
        suite.addTest(new ExportImportTest("testBindingChange"));
        suite.addTest(new ExportImportTest("testRetentionReduction"));
        suite.addTest(new ExportImportTest("testDispositionPolicies"));
        return suite;
    }
    
    /**
     * Creates a simple XSet and exports it. Doesn't verify that the canonical format is correct, this needs to be done
     * by inspection. The exported file is later used to test import.
     * @throws Exception
     */
    public void testSimpleExport() throws Exception {
        final String PREFIX = "test.simple.export.";

        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        FieldContainer fc = (FieldContainer) xset;
        xset.createProperty("testBoolean", false, true);
        xset.createProperty("testInteger", true, 125);
        xset.createProperty("testDouble", true, (double) 1.4);
        
        // commit 
        XUID xuid1 = xset.commit();
        
        // Now export
        File exportFile = File.createTempFile("TestSimpleExport_", ".data");
        exportedSimpleFilename = exportFile.getAbsolutePath();
        writeExportFile( xset.openExportXStream(), exportedSimpleFilename );
        
        xset.close();
        system.close();
        // Display file name
        System.out.println("Export File: " + exportFile.getAbsolutePath());

    }
    
   /** 
     * Creates an XSet with properties and streams, then exports them. This exported XSet will be used later
     * to import and test the XSet.
     * @throws Exception
     */
    public void testSimpleStreamExport() throws Exception {
        final String PREFIX = "test.simple.stream.export.";
        System.out.println("ExportImportTest.testSimpleStreamExport" +
                " - NOT Fully Implemented" +
                "\nYou must verify the exported xml by hand.");
        int bufferSize = 100;
        
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        FieldContainer fc = (FieldContainer) xset;
        xset.createProperty("testBoolean", false, true);
        xset.createProperty("testInteger", true, 125);
        xset.createProperty("testDouble", true, (double) 1.4);
        createStream(xset, "stream.field", true, "image/jpeg", 
                (byte)'b', 100);
        // commit 
        XUID xuid1 = xset.commit();
        
        // Now export
        File exportFile = File.createTempFile("TestSimpleStreamExport_", ".data");
        exportedStreamsFilename = exportFile.getAbsolutePath();
        writeExportFile( xset.openExportXStream(), exportedStreamsFilename );
        xset.close();
        system.close();
        // Display file name
        System.out.println("Export File: " + exportFile.getAbsolutePath());
    }
       
    public void testStreamsExport() throws Exception {
        final String PREFIX = "test.streams.export.";
        System.out.println("ExportImportTest.testStreamsExport" +
                " - NOT Fully Implemented" +
                "\nYou must verify the exported xml by hand.");
        int bufferSize = 100;
        
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        FieldContainer fc = (FieldContainer) xset;
        xset.createProperty("testBoolean", false, true);
        xset.createProperty("testInteger", true, 125);
        xset.createProperty("testDouble", true, (double) 1.4);
        //createStreamFromFile(xset,"stream.switch.field", true, "image/jpeg", "/var/tmp/testswitch.jpg");
        createStream(xset, "stream.field", true, "image/jpeg", 
                (byte)'b', 100);
        createStream(xset, "stream2.field", false, "image/jpeg",
                (byte) 'c', 200);
        // commit 
        XUID xuid1 = xset.commit();
        
        // Now export
        File exportFile = File.createTempFile("TestStreamExport_", ".data");
        exportedFilename = exportFile.getAbsolutePath();
        writeExportFile( xset.openExportXStream(), exportedFilename );
        xset.close();
        system.close();
        // Display file name
        System.out.println("Export File: " + exportFile.getAbsolutePath());

    }

    private long createStream(XSet xset, String fieldName, boolean binding,
            String type, byte data, int numBytes)
            throws Exception {
        XStream xstream = xset.createXStream(fieldName, binding, type);
        byte[] buffer = new byte[numBytes];
        for (int i = 0; i < buffer.length; i++) {
            buffer[i] = data;
        }
        long numBytesWritten = xstream.write(buffer);
        assertTrue("Number of bytes written does not match number sent",
                numBytesWritten == buffer.length);
        xstream.close();
        return numBytesWritten;

    }

    /**
     * Import an XSet from a previously exported file, check to make sure that the properties were persisted
     * correctly, and to make sure that the system fields are preserved.
     * @throws Exception
     */
    public void testSimpleImport() throws Exception {
        final String PREFIX = "test.simple.import.";
        System.out.println("ExportImportTest.testSimpleImport" +
                " - NOT Fully Implemented" +
                "\nJust reads from import stream - no verification or" +
                "loading of XSet.");

        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        XStream importStream = xset.openImportXStream();
        //File importFile = new File("/var/tmp/SimpleExport.data");
        File importFile = new File(exportedSimpleFilename);
        FileInputStream fis = new FileInputStream(importFile);
        byte[] importBytes = new byte[2000];
        long numBytes = fis.read(importBytes);
        while (numBytes > 0) {
            importStream.write(importBytes, numBytes);
            numBytes = fis.read(importBytes);
        }
        
        fis.close();
        importStream.close();
        XUID xuid = xset.commit();
        xset.close();
        
        xset = system.openXSet(xuid, XSet.MODE_READ_ONLY);
        long value = xset.getLong("testInteger");
        assertTrue("Integer field not persisted with correct value.",
                value == 125);
        xset.close();
        system.close();
        
    }

    /**
     * Imports a previously exported XSet and makes sure that the streams still exist.
     * @throws Exception
     */
    public void testSimpleStreamImport() throws Exception {
        final String PREFIX = "test.simple.stream.import.";
        System.out.println("ExportImportTest.testSimpleStreamImport" +
                " - NOT Fully Implemented" +
                "\nJust reads from import stream - no verification or" +
                "loading of XSet.");

        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        XStream importStream = xset.openImportXStream();
        File importFile = new File(exportedStreamsFilename);
        //File importFile = new File("/var/tmp/SimpleStreamExport.data");
        FileInputStream fis = new FileInputStream(importFile);
        byte[] importBytes = new byte[2000];
        long numBytes = fis.read(importBytes);
        while (numBytes > 0) {
            importStream.write(importBytes, numBytes);
            numBytes = fis.read(importBytes);
        }
        
        fis.close();
        importStream.close();
        xset.commit();
        xset.close();
        system.close();
        
   }
    
    /**
     * Import all the XSets from vendors (AKA contributed). This checks to make sure a manuall added
     * property matches expectations.
     * @throws Exceptions
     */
    public void testContributedImports() throws Exception
    {
       final String TEST_PROP = "org.snia.xam.import_unit_test.salt";
       
       XSystem system = connectAndAuthenticate(true);
       File dataDirectory = new File( "test/data" );
       if( ! dataDirectory.isDirectory() )
          fail( "test/data/ does not exist" );
       
       String list[] = dataDirectory.list();
       
       for( int i = 0; i < list.length; i++ )
       {
          XSet xset = system.createXSet( XSet.MODE_UNRESTRICTED );
          XStream importXStream = xset.openImportXStream();
          writeImportData( dataDirectory.getAbsolutePath() + File.separator + list[i], importXStream );
          importXStream.close();
          xset.commit();
          long test = xset.getLong( TEST_PROP );
          
          // The 114968 is a randomly chosen number, essentially meaningless.
          assertEquals( "Expected to see manually salted number in each contributed XSet " + TEST_PROP,
             114968, test );
          xset.close();
       }
       system.close();
    }
    
    /**
     * Tests import to make sure that it tests binding changes and policy compliance 
     * the way specified in the XAM FCAS 1.0.1 specification.
     * @throws Exception
     */
    public void testImportErrorChecks() throws Exception
    {
       XSystem xsys = connectAndAuthenticate(true);
       try
       {
          // Make sure that import only accepts a new unused XSet instance.
          // If the system gets as far as opening the import stream, this sub test passes.
          XSet    x1   = xsys.createXSet( XSet.MODE_UNRESTRICTED );
          XStream importStream = x1.openImportXStream();
          writeImportData( exportedFilename, importStream );
          importStream.close();
          XUID xuid1 = x1.commit();
          
          // Now, this should fail.
          try { 
             importStream = x1.openImportXStream();
             fail( "Import should have failed on a committed XSet" );
          }
          catch( XSetException xe ) {
             /* Ignore, it is expected here. */
          }
          x1.close();
          
          // Let's import the same thing, again, it should succeed.
          XSet x2 = xsys.createXSet(XSet.MODE_UNRESTRICTED);
          importStream = x2.openImportXStream();
          writeImportData( exportedFilename, importStream );
          importStream.close();
          XUID xuid2 = x2.commit();
          
          assertEquals( "Reimporting an XSet must result in the same XUID", xuid1, xuid2 );
          x2.close();
          
          long count = countXSets( xsys, "select \".xset.xuid\" where \".xset.xuid\"=xuid('" + 
             xuid2 + "')");
          assertEquals( "Should not see more than one XSet", 1, count );
          
          // Import it and then make a few non-binding changes.
          final String TEST_PROP = "test.import.property";
          XSet x3 = xsys.createXSet(XSet.MODE_RESTRICTED);
          importStream = x3.openImportXStream();
          writeImportData( exportedFilename, importStream );
          importStream.close();
          x3.createProperty( TEST_PROP, false, "test.import.value" );
          x3.commit();
          
          // Now see if import will reset the non-binding changes.
          XSet x4 = xsys.createXSet(XSet.MODE_RESTRICTED);
          importStream = x4.openImportXStream();
          writeImportData( exportedFilename, importStream );
          importStream.close();
          try { 
             String value = x4.getString( TEST_PROP );
             fail( "XSet should have been reset as a result of reimporting it. " + value);
          }
          catch( FieldDoesNotExistException fdne ) { /*ignore */ }
       }
       finally
       {
          xsys.abandon();
          xsys.close();
        }
    }
    
    /**
     * Checks to make sure that exceptions are thrown when non existant policies are in the 
     * import file, or a retention ID listed in a retention policy doesn't exist in the 
     * retention list.
     * @throws Exception
     */
    public void testRetentionPolicies() throws Exception
    {
       final String RET_ID  = "import.retention.unit.test";
       final String RET_ID2 = "retention-id-2";
       final String BAD_POLICY_FILE = "TestExportImportBadPolicy_";
       
       XSystem xsys = connectAndAuthenticate(true);

          File exportFile = File.createTempFile( BAD_POLICY_FILE, ".data");
          mungedPolicyExport = exportFile.getAbsolutePath();
          
          // Create a basic XSet using a policy for retention
          XSet xset = xsys.createXSet( XSet.MODE_UNRESTRICTED );
          xset.createRetention( false, RET_ID );
          xset.applyRetentionDurationPolicy(RET_ID, false, RetentionMgr.POLICY_ONE_DAY);
          xset.applyRetentionEnabledPolicy (RET_ID, false, RetentionMgr.POLICY_ONE_DAY);
          xset.createRetention( false, RET_ID2 );
          xset.applyRetentionDurationPolicy(RET_ID2, false, RetentionMgr.POLICY_ONE_DAY);
          xset.applyRetentionEnabledPolicy (RET_ID2, false, RetentionMgr.POLICY_ONE_DAY);
          xset.applyAutoDeletePolicy( false, DispositionPolicy.DISPOSITION_AUTODELETE );
          xset.applyShredPolicy( false, DispositionPolicy.DISPOSITION_SHRED );
          XUID xuid1 = xset.commit();
          
          // Count the properties on the new XSet.
          XIterator counter = xset.openFieldIterator("");
          int count = 0;
          while( counter.hasNext() ) {
             count++;
             counter.next();
          }
          counter.close();
          
          XStream exportStream = xset.openExportXStream();
          writeExportFile( exportStream, mungedPolicyExport );

          // Straightforward import
          XSet okXSet = xsys.createXSet(XSet.MODE_UNRESTRICTED);
          XStream okImportStream = okXSet.openImportXStream();
          writeImportData( mungedPolicyExport, okImportStream );
          okImportStream.close();
          
          XIterator testCounter = xset.openFieldIterator("");
          int testCount = 0;
          while( testCounter.hasNext() ) {
             testCount++;
             testCounter.next();
          }
          testCounter.close();
          
          assertEquals( "Differing number of fields in imported file from original file", count, testCount );

          // Do a bad import
          File badPolicy = File.createTempFile( BAD_POLICY_FILE + "_badPol_", ".data" );
          badPolicyFile = badPolicy.getAbsolutePath();
         
          // Modify the file to change the string value of the policy.
          editExportFile( exportFile, badPolicy,
                          RetentionMgr.POLICY_ONE_DAY, "foo" );
          xset.close();
                    
          // Try the import, it should fail
          XSet xset2 = xsys.createXSet( XSet.MODE_UNRESTRICTED );
          XStream importStream = xset2.openImportXStream();
          writeImportData( badPolicyFile, importStream );
          try {
             importStream.close();
             XUID xuid2 = xset2.commit();
             fail( "Should have thrown an exception because of illegal retention policy" );
          }
          catch( PolicyNameException rve ) {
             /* ignore */
          }
          
          // Change the retentionID to something not in the retentionList
          File badRetentionID = File.createTempFile( BAD_POLICY_FILE + "_badRetID_", ".data" );
          badRetentionIDFile = badRetentionID.getAbsolutePath();
          editExportFile( exportFile, badRetentionID,
                          ".xset.retention." + RET_ID + ".duration.policy", 
                          ".xset.retention.bad.retention.id.duration.policy" );
          // Try the import, it should fail
          XSet xset3 = xsys.createXSet( XSet.MODE_UNRESTRICTED );
          importStream = xset3.openImportXStream();
          writeImportData( badRetentionIDFile, importStream );
          importStream.close();
          try {
             XUID xuid3 = xset3.commit();
             fail( "Should have thrown an exception because of illegal retention ID" );
          }
          catch( RetentionValueException rve ) {
             /* ignore */
          }

          // Change the retention duration to something greater than the actual policy will support
          // This will cause import to fail.
          File badDurationPolicy = File.createTempFile( BAD_POLICY_FILE + "_badRetDuration_", ".data" );
          badDurationPolicyFile = badDurationPolicy.getAbsolutePath();
          // Make export file require a longer duration
          editExportFile( exportFile, badDurationPolicy,
                          "86400000",   // The one day value 
                          "86400001" ); // One day plus one millisecond
          // Try the import, it should fail
          XSet xset4 = xsys.createXSet( XSet.MODE_UNRESTRICTED );
          importStream = xset4.openImportXStream();
          writeImportData( badDurationPolicyFile, importStream );
          try {
             importStream.close();
             fail( "Should have thrown an exception because of policy value failures" );
          }
          catch( RetentionValueException rve ) {
             /* ignore */
          }
          
          // Change the retention enabled to false, which would result in a reduction of retention
          // This will cause import to fail.
          File badEnabledPolicy = File.createTempFile( BAD_POLICY_FILE + "_badRetEnabled_", ".data" );
          badEnabledPolicyFile = badEnabledPolicy.getAbsolutePath();
          // Make export file require disabled retention, differing from the policy setting
          // Apologies for the ugly edit, but changing "true" to "false" would have been too broad, and more
          // context is required to make sure that the correct instance of the boolean value is edited.
          editExportFile( exportFile, badEnabledPolicy,
             "<property name=\".xset.retention.import.retention.unit.test.enabled\" type=\"application/vnd.snia.xam.boolean\" binding=\"false\" readOnly=\"false\" length=\"1\">\n            <boolean>true",   // Enabled = true 
             "<property name=\".xset.retention.import.retention.unit.test.enabled\" type=\"application/vnd.snia.xam.boolean\" binding=\"false\" readOnly=\"false\" length=\"1\">\n            <boolean>false");   // Enabled = true 
          // Try the import, it should fail
          XSet xset5 = xsys.createXSet( XSet.MODE_UNRESTRICTED );
          importStream = xset5.openImportXStream();
          writeImportData( badEnabledPolicyFile, importStream );
          try {
             importStream.close();
             fail( "Should have thrown an exception because of policy value failures" );
          }
          catch( RetentionValueException rve ) {
             /* ignore */
          }

    }

    /**
     * Checks to make sure that a binding change to the exported file will result in a new XUID being
     * when the imported XSet is committed. Also checks to make sure that the previous XSet has not changed.
     * @throws Exception
     */
    public void testBindingChange() throws Exception
    {
       final String RET_ID  = "import-binding-change";
       final String BINDING_FILE = "TestExportImportBindingChange_";
       final String PROP = "com.example.property";
       
       XSystem xsys = connectAndAuthenticate(true);

          File exportFile = File.createTempFile( BINDING_FILE, ".data");
          bindingExportFile = exportFile.getAbsolutePath();
          
          // Create a basic XSet using a policy for retention
          XSet xset = xsys.createXSet( XSet.MODE_UNRESTRICTED );
          xset.createRetention( false, RET_ID );
          xset.applyRetentionDurationPolicy(RET_ID, false, RetentionMgr.POLICY_ONE_DAY);
          xset.applyRetentionEnabledPolicy (RET_ID, false, RetentionMgr.POLICY_ONE_DAY);
          xset.applyAutoDeletePolicy( false, DispositionPolicy.DISPOSITION_AUTODELETE );
          xset.applyShredPolicy( false, DispositionPolicy.DISPOSITION_SHRED );
          xset.createProperty( PROP, true, RET_ID );
          XUID xuid1 = xset.commit();
          
          // Count the properties on the new XSet.
          XIterator counter = xset.openFieldIterator("");
          int count = 0;
          while( counter.hasNext() ) {
             count++;
             counter.next();
          }
          counter.close();
          
          XStream exportStream = xset.openExportXStream();
          writeExportFile( exportStream, bindingExportFile );

          // Import the file
          File changedBindingFile = File.createTempFile( BINDING_FILE + "_bindingChange_", ".data" );
          changedBindingFileName = changedBindingFile.getAbsolutePath();
          editExportFile( exportFile, changedBindingFile,
             "<property name=\"com.example.property\" type=\"application/vnd.snia.xam.string\" binding=\"true\"",     // binding = true 
             "<property name=\"com.example.property\" type=\"application/vnd.snia.xam.string\" binding=\"false\"");   // binding = false 
          XSet newXSet = xsys.createXSet(XSet.MODE_UNRESTRICTED);
          XStream okImportStream = newXSet.openImportXStream();
          writeImportData( changedBindingFileName, okImportStream );
          okImportStream.close();
          
          XIterator testCounter = xset.openFieldIterator("");
          int testCount = 0;
          while( testCounter.hasNext() ) {
             testCount++;
             testCounter.next();
          }
          testCounter.close();
          
          assertEquals( "Differing number of fields in imported file from original file", count, testCount );
          
          XUID xuid2 = newXSet.commit();
          
          assertFalse( "Expected new xuid to be generated by binding change import", xuid1.equals(xuid2) );
    }

    /**
     * Checks to make sure that a binding change to the exported file will result in a new XUID being
     * when the imported XSet is committed. Also checks to make sure that the previous XSet has not changed.
     * @throws Exception
     */
    public void testRetentionReduction() throws Exception
    {
       final String RET_ID  = "import-retention-reduction";
       final String REDUCTION_FILE = "TestExportImportReduction_";
       
       XSystem xsys = connectAndAuthenticate(true);

          File exportFile = File.createTempFile( REDUCTION_FILE, ".data");
          reducedExportFilename = exportFile.getAbsolutePath();
          
          // Create a basic XSet using a policy for retention
          XSet xset = xsys.createXSet( XSet.MODE_UNRESTRICTED );
          xset.createRetention( false, RET_ID );
          xset.setRetentionDuration    (RET_ID, false, 1234 );
          xset.setRetentionEnabledFlag (RET_ID, false, true);
          xset.applyAutoDeletePolicy( false, DispositionPolicy.DISPOSITION_AUTODELETE );
          xset.applyShredPolicy( false, DispositionPolicy.DISPOSITION_SHRED );
          XUID xuid1 = xset.commit();
                    
          XStream exportStream = xset.openExportXStream();
          writeExportFile( exportStream, reducedExportFilename );

          // Import the file
          File reducedImportFile = File.createTempFile( REDUCTION_FILE + "_reduced_", ".data" );
          reducedImportFilename = reducedImportFile.getAbsolutePath();
          editExportFile( exportFile, reducedImportFile,
             "<integer>1234</integer>",     // binding = true 
             "<integer>1233</integer>");   // binding = false 
          XSet newXSet = xsys.createXSet(XSet.MODE_UNRESTRICTED);
          XStream okImportStream = newXSet.openImportXStream();
          writeImportData( reducedImportFilename, okImportStream );
          okImportStream.close();
          
          try {
             XUID xuid2 = newXSet.commit();
             fail( "Should have thrown a RetentionValueException to prevent commit()" );
          }
          catch( RetentionValueException rve ) { /* ignore */ }
    }

    /**
     * Checks to make sure that exceptions are thrown when non existant disposition policies 
     * are in the import file.
     * @throws Exception
     */
    public void testDispositionPolicies() throws Exception
    {
       final String BAD_POLICY_FILE = "TestExportImportBadDispositionPolicy_";
       
       XSystem xsys = connectAndAuthenticate(true);
       try
       {
          File exportFile = File.createTempFile( BAD_POLICY_FILE, ".data");
          dispositionPolicyExport = exportFile.getAbsolutePath();
          
          // Create a basic XSet using a policy for retention
          XSet xset = xsys.createXSet( XSet.MODE_UNRESTRICTED );
          xset.applyAutoDeletePolicy( false, DispositionPolicy.DISPOSITION_AUTODELETE );
          xset.applyShredPolicy( false, DispositionPolicy.DISPOSITION_SHRED );
          XUID xuid1 = xset.commit();
          XStream exportStream = xset.openExportXStream();
          writeExportFile( exportStream, dispositionPolicyExport );

          File badPolicy = File.createTempFile( BAD_POLICY_FILE + "_badDispPol_", ".data" );
          badDispositionPolicyExport = badPolicy.getAbsolutePath();
         
          // Modify the file to change the string value of the policy.
          editExportFile( exportFile, badPolicy,
             DispositionPolicy.DISPOSITION_AUTODELETE, "foo" );
          xset.close();
                    
          // Try the import, it should fail
          XSet xset2 = xsys.createXSet( XSet.MODE_UNRESTRICTED );
          XStream importStream = xset2.openImportXStream();
          writeImportData( badDispositionPolicyExport, importStream );
          importStream.close();
          String auto_policy = xset2.getString(XSet.XAM_DELETION_AUTODELETE_POLICY);
          
          try {
             XUID xuid2 = xset2.commit();
             fail( "Should have thrown an exception because of illegal autodelete policy " + auto_policy);
          }
          catch( PolicyNameException rve ) {
             /* ignore */
          }
          
          // Change the shred policy to an illegal policy
          File badShredPolicy = File.createTempFile( BAD_POLICY_FILE + "_badRetID_", ".data" );
          badShredPolicyFile = badShredPolicy.getAbsolutePath();
          editExportFile( exportFile, badShredPolicy,
                          DispositionPolicy.DISPOSITION_SHRED, 
                          "foo" );
          // Try the import, it should fail
          XSet xset3 = xsys.createXSet( XSet.MODE_UNRESTRICTED );
          importStream = xset3.openImportXStream();
          writeImportData( badShredPolicyFile, importStream );
          importStream.close();
          try {
             XUID xuid3 = xset3.commit();
             fail( "Should have thrown an exception because of illegal shred policy" );
          }
          catch( PolicyNameException pne ) {
             /* ignore */
          }
       }
       finally
       {
          cleanUp();
       }
    }

    // This is a hack, since generally attempting to read an entire file into memory
    // is not scalable. Simple export files are not very large, so this should be doable.
    private void editExportFile(File   srcFile,
                                File   dstFile,
                                String srcStr,
                                String dstStr)
    throws IOException
   {
       int size = (int) srcFile.length();
       byte buffer[] = new byte[size];
       FileInputStream fis = new FileInputStream(srcFile);
       long numBytes = fis.read(buffer);
       String tmpStr = new String(buffer);
       fis.close();
       StringBuffer sBuff = new StringBuffer(tmpStr);
       int loc;
       
       while( (loc = sBuff.indexOf(srcStr)) >= 0 )
       {
          sBuff.replace(loc, loc+srcStr.length(), dstStr );
       }
       FileOutputStream fos = new FileOutputStream(dstFile);
       fos.write(sBuff.toString().getBytes());
       fos.close(); 
   }

   /**
     * Run a query, return the number of results returned.
     * @param string The query needed by the test
     * @return
    * @throws XAMException 
     */
    private long countXSets(XSystem system,
                            String  xamql ) throws XSystemAbandonException, XSystemCorruptException, AuthenticationExpiredException, AuthorizationException, InvalidArgumentException, XAMException
    {
       XSet xset = null;
       if (system != null && xamql != null) {
          xset = system.createXSet(XSet.MODE_UNRESTRICTED);
          xset.createProperty(XSet.XAM_JOB_COMMAND, false, XSet.XAM_JOB_QUERY);
          XStream stream = xset.createXStream(XSet.XAM_JOB_QUERY_COMMAND,
             false, "text/plain;charset=utf-8");
          byte[] bytes = xamql.getBytes();
          int wrote = 0;
          while (wrote < bytes.length) {
             wrote += stream.write(bytes, wrote, bytes.length - wrote);
          }
          
          stream.close();
          xset.submitJob();
       }
       do {
          try {
              Thread.sleep(100);
          } catch (InterruptedException ex) {
              // announce("Interrupted Exception thrown:  " + ex.getMessage());
          }
      } while (!JobStatus.isDone(xset.getString(XSet.XAM_JOB_STATUS)));

       return xset.getLong(XSet.XAM_JOB_QUERY_RESULTS_COUNT);
    }

   /**
     * Writes the data from an export file into a (presumed) import stream.
     * @param fileName
     * @param importStream
     * @throws IOException
     * @throws XAMException
     */
   private void writeImportData(String fileName, XStream importStream)
      throws IOException, XAMException
   {
      File importFile = new File(fileName);
      FileInputStream fis = new FileInputStream(importFile);
      byte[] importBytes = new byte[2000];
      long numBytes = fis.read(importBytes);
      while (numBytes > 0) {
          importStream.write(importBytes, numBytes);
          numBytes = fis.read(importBytes);
      } 
      fis.close();
   }
   
   private void writeExportFile( XStream exportXStream,
                                 String  filename ) throws Exception
  {
      FileOutputStream fos = new FileOutputStream(new File(filename), true);
      FileChannel channel = fos.getChannel();

      //File exportFile = new File()
      byte[] exportBytes = new byte[2000];
      //ByteBuffer byteBuffer = ByteBuffer.wrap(exportBytes);
      long numBytes = exportXStream.read(exportBytes);
      // Write out export file
      while (numBytes > 0) {
          ByteBuffer byteBuffer = ByteBuffer.wrap(exportBytes, 0, 
                  (int) numBytes);
          channel.write(byteBuffer);
          byteBuffer.position(0);
          numBytes = exportXStream.read(exportBytes);
      }
      channel.close();
      fos.close();
      exportXStream.close();
  }

   /**
    * Try to be a good citizen and delete temporary files used to export.
    * This should be called from the very last test, there doesn't seem to be a tearDown for the
    * entire suite.
    * @see org.snia.xam.vim.reference.XAMTest#tearDown()
    */
   public void cleanUp()
   {
      File f1 = new File( exportedFilename );
      f1.delete();
      File f2 = new File( exportedStreamsFilename );
      f2.delete();
      File f3 = new File( mungedPolicyExport );
      f3.delete();
      File f4 = new File( badPolicyFile );
      f4.delete();
      File f5 = new File( badShredPolicyFile );
      f5.delete();
      File f6 = new File( badDispositionPolicyExport );
      f6.delete();
      File f7 = new File( dispositionPolicyExport );
      f7.delete();
      File f8 = new File( badRetentionIDFile );
      f8.delete();
      File f9 = new File( exportedSimpleFilename );
      f9.delete();
      File f10 = new File( badDurationPolicyFile );
      f10.delete();
      File f11 = new File( badEnabledPolicyFile );
      f11.delete();
      File f12 = new File( bindingExportFile );
      f12.delete();
      File f13 = new File( changedBindingFileName );
      f13.delete();
      File f14 = new File( reducedExportFilename );
      f14.delete();
      File f15 = new File( reducedImportFilename );
      f15.delete();
   }
}
