/*
 *  Copyright (c) 2009, Sun Microsystems, Inc.
 *  Copyright (c) 2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED FBY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Properties;

import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.PolicyNameException;
import org.snia.xam.XAMException;
import org.snia.xam.XIterator;
import org.snia.xam.XSet;
import org.snia.xam.XSetCorruptException;
import org.snia.xam.XSetInaccessibleException;
import org.snia.xam.XSetUnderHoldException;
import org.snia.xam.XSystem;
import org.snia.xam.XUID;
import org.snia.xam.vim.reference.utils.ReferenceSaslUtils;

import junit.framework.Test;
import junit.framework.TestSuite;

/**
 * Exercises the litigation hold functionality in the XAM API. 
 */
public class LitigationHoldTest
   extends ReferenceXAMTest
{
   public LitigationHoldTest(String name)
   {
      super(name);
   }

   public static void main(String[] args)
   {
      System.exit(runTest(args, suite()));
   }

   public static Test suite()
   {
      TestSuite suite = new TestSuite();
      suite.addTest(new LitigationHoldTest("testHoldRelease"));
      suite.addTest(new LitigationHoldTest("testDeleteUnderHold"));
      suite.addTest(new LitigationHoldTest("testModifyUnderHold"));
      suite.addTest(new LitigationHoldTest("testCopyXSet"));
      return suite;
   }

   /**
    * Create several XSets with out  access policies (from the users), and make sure they can 
    * be applied to the XSet. 
    */
   public void testHoldRelease()
      throws Exception
   {
      final String HOLD_ID = "litigationHoldRelease test ID";
      XSystem xsys = null;
      XSet    xset = null;
      XIterator iter = null;
      
      try
      {
         xsys = connectAndAuthenticate(false);
         xset = xsys.createXSet(XSet.MODE_UNRESTRICTED);
         xset.createProperty("org.snia.xam.reference.test.litigationHold.testHoldRelease",
            false,
            "testing hold an release.");
         XUID xuid = xset.commit();
         
         // Make sure it's not current under hold.
         boolean onHold = xset.getBoolean(XSet.XSET_HOLD);
         assertEquals("New XSet should never be on hold", false, onHold);
         
         // Also check to make sure the hold ID list is non existant.
         iter = xset.openFieldIterator(XSet.XSET_HOLD_LIST);
         assertEquals("Hold list must be empty.", false, iter.hasNext());
         iter.close();
         
         // Place XSet under a specific hold, then check to make sure the ID
         // shows up
         // in the hold id list.
         xsys.holdXSet(xuid, HOLD_ID);
         XSet xset2 = xsys.openXSet(xuid, XSet.MODE_READ_ONLY );
         assertEquals("XSet.hold property incorrect,",
            true,
            xset2.getBoolean(XSet.XSET_HOLD));
         
         iter = xset2.openFieldIterator(XSet.XSET_HOLD_LIST);
         String holdId = (String) iter.next();
         
         assertEquals( "Wrong Hold ID", XSet.XSET_HOLD_LIST+HOLD_ID, holdId );
         
         // Remove the hold and make sure the properties are removed.
         xsys.releaseXSet( xuid, HOLD_ID );
         XSet xset3 = xsys.openXSet( xuid, XSet.MODE_READ_ONLY );
         assertEquals( "Hold must be set to false", false, xset.getBoolean( XSet.XSET_HOLD) );
         
         iter = xset.openFieldIterator(XSet.XSET_HOLD_LIST);
         assertEquals( "There should be no additional holds on the xset", false, iter.hasNext());
      }
      finally {
         if( iter != null ) {
            iter.close();
         }
         
         if( xset != null ) {
            xset.abandon();
            xset.close();
         }
         
         if( xsys != null ) {
            xsys.abandon();
            xsys.close();
         }
      }
   }


   /**
    * Attempt to delete an XSet under hold. 
    */
   public void testDeleteUnderHold()
      throws Exception
   {
      final String HOLD_ID = "litigationHoldRelease test ID";
      XSystem xsys = null;
      XSet    xset = null;
      
      try
      {
         xsys = connectAndAuthenticate(false);
         xset = xsys.createXSet(XSet.MODE_UNRESTRICTED);
         xset.createProperty("org.snia.xam.reference.test.litigationHold.testDelete",
            false,
            "testing delete while held.");
         XUID xuid = xset.commit();
         xset.close();
         
         // Place XSet under a specific hold, then check to make sure the ID
         // shows up
         // in the hold id list.
         xsys.holdXSet(xuid, HOLD_ID);
         try { 
            xsys.deleteXSet(xuid); 
            fail( "XSystem.deleteXSet should have thrown a XSetUnderHoldException");
         }
         catch( XSetUnderHoldException xuhe ) { /* ignore, it's good. */ }
         
         xsys.releaseXSet(xuid, HOLD_ID);
      }
      finally {
         if( xsys != null ) {
            xsys.abandon();
            xsys.close();
         }
      }
   }

   /**
    * Attempt to delete an XSet under hold. 
    */
   public void testModifyUnderHold()
      throws Exception
   {
      final String HOLD_ID = "litigationHoldModify test ID";
      XSystem xsys = null;
      XSet    xset = null;
      
      try
      {
         xsys = connectAndAuthenticate(false);
         xset = xsys.createXSet(XSet.MODE_UNRESTRICTED);
         xset.createProperty("org.snia.xam.reference.test.litigationHold.testModify",
            false,
            "testing modify while held.");
         XUID xuid = xset.commit();
         
         // Place XSet under a specific hold, then check to make sure the ID
         // shows up
         // in the hold id list.
         xsys.holdXSet(xuid, HOLD_ID);
         try { 
            xsys.openXSet(xuid, XSet.MODE_RESTRICTED ); 
            fail( "XSystem.openXSet(RESTRICTED) should have thrown a XSetUnderHoldException");
         }
         catch( XSetUnderHoldException xuhe ) { /* ignore, it's good. */ }
         
         try { 
            xsys.openXSet(xuid, XSet.MODE_UNRESTRICTED ); 
            fail( "XSystem.openXSet(UNRESTRICTED) should have thrown a XSetUnderHoldException");
         }
         catch( XSetUnderHoldException xuhe ) { /* ignore, it's good. */ }
         
         // See if the open XSet can be modified.
         
         xset.createProperty( "com.example.foo", false, "foo" );
         try { xset.commit(); }
         catch( XSetCorruptException xce ) 
         { 
            xset.abandon();
            xset.close();
         }
         
         // This should allow access
         xsys.releaseXSet(xuid, HOLD_ID);
         xset = xsys.openXSet( xuid, XSet.MODE_UNRESTRICTED );
         xset.createProperty( "com.example.foo", false, "foo2" );
         xset.commit();
         xset.close();
      }
      finally {
         if( xsys != null ) {
            xsys.abandon();
            xsys.close();
         }
      }
   }
   /**
    * Verify that copyXSet will clear the hold conditions. 
    */
   public void testCopyXSet()
      throws Exception
   {
      final String HOLD_ID = "litigationHoldModify test ID";
      XSystem xsys = null;
      XSet    xset = null;
      
      try
      {
         xsys = connectAndAuthenticate(false);
         xset = xsys.createXSet(XSet.MODE_UNRESTRICTED);
         xset.createProperty("org.snia.xam.reference.test.litigationHold.testModify",
            false,
            "testing modify while held.");
         XUID xuid = xset.commit();
         
         // Place XSet under a specific hold, then copy it.
         xsys.holdXSet(xuid, HOLD_ID);
         XSet xset2 = xsys.copyXSet(xuid, XSet.MODE_UNRESTRICTED );
      }
      finally {
         if( xsys != null ) {
            xsys.abandon();
            xsys.close();
         }
      }
   }

}
