/*
 *  Copyright (c) 2008-2009, Sun Microsystems, Inc.
 *  Copyright (c) 2008-2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.PrintWriter;
import java.io.RandomAccessFile;
import java.util.Properties;

import junit.framework.Test;
import junit.framework.TestSuite;

import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XSet;
import org.snia.xam.XSystemException;
import org.snia.xam.vim.reference.utils.ReferenceSaslUtils;

import org.snia.xam.XSystem;
import org.snia.xam.vim.reference.ReferenceAuthenticationStatus;
import org.snia.xam.vim.reference.ReferenceXSystem;

/**
 * Basic authentication tests for SASL PLAIN and ANONYMOUS
 * 
 */
public class ReferenceAuthenticationTest extends ReferenceXAMTest {
    
    protected static final String[] PRE_AUTH_PROPS = {
        XSystem.XAM_XSYSTEM_AUTH_SASL_LIST + 
                ReferenceSaslUtils.SASL_MECHANISM_ANONYMOUS,
        XSystem.XAM_XSYSTEM_AUTH_SASL_LIST + 
                ReferenceSaslUtils.SASL_MECHANISM_PLAIN,
        XSystem.XAM_XSYSTEM_AUTH_SASL_DEFAULT
    };
    protected static final String[] PRE_AUTH_TYPES = {
        XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
        XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
        XAMLibrary.STYPE_STRING_MIME_TYPE
    };
    protected static final String[] POST_AUTH_PROPS = {
        XSystem.XAM_DELETION_AUTODELETE,
        XSystem.XAM_XSYSTEM_IDENTITY,
        XSystem.XAM_XSYSTEM_JOB_COMMIT_SUPPORTED,
        XSystem.XAM_XSYSTEM_JOB_LIST_QUERY,
        XSystem.XAM_XSYSTEM_JOB_QUERY_CONTINUANCE_SUPPORTED,
        XSystem.XAM_XSYSTEM_JOB_QUERY_LEVEL1_SUPPORTED,
        XSystem.XAM_XSYSTEM_JOB_QUERY_LEVEL2_SUPPORTED,
        XSystem.XAM_XSYSTEM_LIMITS_MAX_FIELDS,
        XSystem.XAM_XSYSTEM_LIMITS_MAX_STREAM_SIZE,
        XSystem.XAM_XSYSTEM_TIME
    };
    protected static final String[] POST_AUTH_TYPES = {
        XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, /* autodelete */
        XAMLibrary.STYPE_STRING_MIME_TYPE, /* identity */
        XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, /* commit supported */
        XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, /* job list query */
        XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, /* query continuance supported */
        XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, /* query level 1 */
        XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, /* query level 2 */
        XAMLibrary.STYPE_INT_MIME_TYPE, /* max fields */
        XAMLibrary.STYPE_INT_MIME_TYPE, /* max stream size */
        XAMLibrary.STYPE_DATETIME_MIME_TYPE /* system time */

    };

    public ReferenceAuthenticationTest(String name) {
        super(name);
    }
    
    public static void main(String[] args)
    {
        System.exit(runTest(args, suite()));
    }
    
    public static Test suite()
    {
    	TestSuite suite = new TestSuite();
    	suite.addTest(new ReferenceAuthenticationTest("testAnonymous"));
        suite.addTest(new ReferenceAuthenticationTest("testAnonymousRelogin"));
        suite.addTest(new ReferenceAuthenticationTest("testLoadAuthDefaults"));
        suite.addTest(new ReferenceAuthenticationTest("testDefault"));
        suite.addTest(new ReferenceAuthenticationTest("testAuthRetry"));
        suite.addTest(new ReferenceAuthenticationTest("testPlain"));
        suite.addTest(new ReferenceAuthenticationTest("testPlainInvalidPassword"));
        suite.addTest(new ReferenceAuthenticationTest("testPlainViaPasswd"));
        suite.addTest(new ReferenceAuthenticationTest("testProperties"));
    	return suite;
    }

    /**
     * Test of loadDefaults method, of class ReferenceXSystem.
     */
    public void testLoadAuthDefaults() throws Exception {
        ReferenceXSystem instance = new ReferenceXSystem();
        instance.loadDefaults();
       // Authentication properties
        assert(instance.getBoolean(XSystem.XAM_XSYSTEM_AUTH_SASL_LIST +
                ReferenceSaslUtils.SASL_MECHANISM_ANONYMOUS) == true);
        assert(instance.getBoolean(XSystem.XAM_XSYSTEM_AUTH_SASL_LIST +
                ReferenceSaslUtils.SASL_MECHANISM_PLAIN) == true);
        
        assertNotNull(XSystem.XAM_XSYSTEM_AUTH_SASL_DEFAULT);
        assertNotNull(XSystem.XAM_XSYSTEM_AUTH_EXPIRATION);

    }
    
    /**
     * Test basic plain authentication. 
     * @throws Exception
     */
    public void testDefault() throws Exception {
        announce("ReferenceAuthenticationTest.testDefault");
        ByteArrayOutputStream byteOutput = new ByteArrayOutputStream(200);
        byte[] inputData = null;
        int retValue = XSystem.XAM_SASL_IN_PROCESS;
        XSystem system = doConnection(true);
        String defMech = system.getString(XSystem.XAM_XSYSTEM_AUTH_SASL_DEFAULT);
        if (defMech.equals(ReferenceSaslUtils.SASL_MECHANISM_ANONYMOUS)) {
            retValue = system.authenticate(inputData, byteOutput);
        } else if (defMech.equals(ReferenceSaslUtils.SASL_MECHANISM_PLAIN)) {
            byte[] creds = ReferenceSaslUtils.encodeSASLPlain(null,
                    ReferenceAuthenticationStatus.TEST_USERNAME,
                    ReferenceAuthenticationStatus.TEST_PASSWORD);
            retValue = system.authenticate(creds, byteOutput);
         } else {
            fail("Unknown default SASL mechanism " + defMech);
        }
        verifyAuthResponse(ReferenceSaslUtils.SASL_DONE.getBytes(), byteOutput);
        // Validate return status value
        assertTrue("Test Failed. System should NOT request more SASL data.",
                retValue == XSystem.XAM_SASL_COMPLETE);

        system.close();
    }

    /**
     * Test basic anonymous authentication.
     * @throws Exception
     */
    public void testAnonymous() throws Exception  {
        announce("ReferenceAuthenticationTest.testAnonymous");
        XSystem xsystem = doConnection(false);
        ByteArrayOutputStream response = new ByteArrayOutputStream(200);
        int retValue = xsystem.authenticate(ReferenceSaslUtils.
                SASL_MECHANISM_ANONYMOUS.getBytes(), response);//
        assertTrue("Test Failed. System should request more SASL data.",
                retValue == XSystem.XAM_SASL_IN_PROCESS);

        verifyAuthResponse(ReferenceSaslUtils.
                SASL_RESPONSE_MORE.getBytes(), response);
        byte[] trace = ReferenceSaslUtils.encodeSASLAnonymous(
                "myemail@mycompany.com");
        response.reset();
        retValue = xsystem.authenticate(trace, response);
        assertTrue("Test Failed. System should NOT request more SASL data.",
                retValue == XSystem.XAM_SASL_COMPLETE);
        xsystem.close();
    }
    
    public void testAnonymousRelogin() throws Exception  {
        announce("ReferenceAuthenticationTest.testAnonymousRelogin");
        XSystem xsystem = doConnection(false);
        ByteArrayOutputStream response = new ByteArrayOutputStream(200);
        int retValue = xsystem.authenticate(
                ReferenceSaslUtils.SASL_MECHANISM_ANONYMOUS.getBytes(),
                response);
        assertTrue("Test Failed. System should request more SASL data.",
                retValue == XSystem.XAM_SASL_IN_PROCESS);
        verifyAuthResponse(ReferenceSaslUtils.
                SASL_RESPONSE_MORE.getBytes(), response);
        response.reset();
        retValue = xsystem.authenticate(null, response);
        assertTrue("Test Failed. System should NOT request more SASL data.",
                retValue == XSystem.XAM_SASL_COMPLETE);
        verifyAuthResponse(ReferenceSaslUtils.SASL_DONE.getBytes(), response);
        
        // TODO:  right now this is coded to fail but will allow re-authentication
        // in future
        //String msg = "Test Failed. Second Authentication succeeded.";
        String msg = "";
        //resp = null;
        try {
            response.reset();
            retValue = xsystem.authenticate(
                    ReferenceSaslUtils.SASL_MECHANISM_ANONYMOUS.getBytes(),
                    response);
            verifyAuthResponse(ReferenceSaslUtils.
                    SASL_RESPONSE_MORE.getBytes(), response);
            assertTrue("Test Failed. System should request more SASL data.",
                retValue == XSystem.XAM_SASL_IN_PROCESS);

            response.reset();
            retValue = xsystem.authenticate(null, response);
            assertTrue("Test Failed. System should NOT request more SASL data.",
                retValue == XSystem.XAM_SASL_COMPLETE);
            verifyAuthResponse(
                    ReferenceSaslUtils.SASL_DONE.getBytes(), response);
            
        } catch (XSystemException ex) {
            msg = ex.getMessage();
        }
        
        assertEquals("Test Failed. Second Authentication should have failed.",
               msg, ReferenceAuthenticationStatus.
                    ERR_MSG_SYSTEM_ALEADY_AUTHENTICATED);               
        
        xsystem.close();
    }

    /**
     * Test standard username/password authentication. This uses the builtin username/passwords
     * @throws Exception
     */
    public void testPlain() throws Exception  {
        announce("ReferenceAuthenticationTest.testPlain");
        
        // Testing deprecated authenticate method until it's removed
        XSystem xsystem = doConnection(false);
        byte[] resp = xsystem.authenticate(ReferenceSaslUtils.
                SASL_MECHANISM_PLAIN.getBytes());
        verifyAuthResponse(ReferenceSaslUtils.
                SASL_RESPONSE_MORE.getBytes(), resp);
        
        
        byte[] authInfo = ReferenceSaslUtils.encodeSASLPlain(
                "junkAuthorzId", ReferenceAuthenticationStatus.TEST_USERNAME,
                ReferenceAuthenticationStatus.TEST_PASSWORD);
        resp = xsystem.authenticate(authInfo);
        verifyAuthResponse(ReferenceSaslUtils.SASL_DONE.getBytes(), 
                resp);
        xsystem.close();
        
        // Test the new authenticate method
        xsystem = doConnection(false);
        ByteArrayOutputStream respByteStream = new ByteArrayOutputStream();        
        int respValue = xsystem.authenticate(ReferenceSaslUtils.
                SASL_MECHANISM_PLAIN.getBytes(), respByteStream);
        assertEquals("Test Failed. System should request more SASL data.",
                respValue, XSystem.XAM_SASL_IN_PROCESS);
        verifyAuthResponse(ReferenceSaslUtils.SASL_RESPONSE_MORE.getBytes(), 
                respByteStream);
        
        respByteStream.reset();
        respValue = xsystem.authenticate(authInfo, respByteStream);
        assertEquals("Test Failed. System should not request more SASL data.",
                respValue, XSystem.XAM_SASL_COMPLETE);
        verifyAuthResponse(
                ReferenceSaslUtils.SASL_DONE.getBytes(), respByteStream);
        xsystem.close();
    }
    
    /**
     * Test standard username/password authentication. This uses the invalid username/passwords
     * @throws Exception
     */
    public void testPlainInvalidPassword() throws Exception  {
        announce("ReferenceAuthenticationTest.testPlainInvalidPassword");
        XSystem xsystem = doConnection(false);
        // test deprecated authenticate method
        byte[] resp = xsystem.authenticate(
                ReferenceSaslUtils.SASL_MECHANISM_PLAIN.getBytes());
        verifyAuthResponse(ReferenceSaslUtils.
                SASL_RESPONSE_MORE.getBytes(), resp);
                
        byte[] authInfo = ReferenceSaslUtils.encodeSASLPlain(
                "junkAuthorzId", ReferenceAuthenticationStatus.TEST_USERNAME,
                "badPassword");
        resp = xsystem.authenticate(authInfo);
        verifyAuthResponse(ReferenceSaslUtils.SASL_FAILED.getBytes(), 
                resp);
        try {
            xsystem.close();
        } catch (XSystemException se) {
            fail("Error should be able to call close after failed authentication.");
        }
        
        // Test new authenticate method
        xsystem = doConnection(false);
        ByteArrayOutputStream respByteStream = new ByteArrayOutputStream();        
        int respValue = xsystem.authenticate(
                ReferenceSaslUtils.SASL_MECHANISM_PLAIN.getBytes(),
                respByteStream);
        assertEquals("Test Failed. System should request more SASL data.",
                respValue, XSystem.XAM_SASL_IN_PROCESS);
        verifyAuthResponse(ReferenceSaslUtils.SASL_RESPONSE_MORE.getBytes(), 
                respByteStream);
        
        respByteStream.reset();
        respValue = xsystem.authenticate(authInfo, respByteStream);
        assertEquals("Test Failed. System should not request more SASL data.",
                respValue, XSystem.XAM_SASL_COMPLETE);
        verifyAuthResponse(
                ReferenceSaslUtils.SASL_FAILED.getBytes(), respByteStream);
        try {
            xsystem.close();
        } catch (XSystemException se) {
            fail("Error should be able to call close after failed authentication.");
        }
        
    }

    /**
     * This test will make sure that the Reference VIM can authenticate as several users using
     * the <i>passwd</i> file extension to the ReferenceAuthentication manager,
     * 
     * The test creates a temporary passwd file, sets the system property to point to the file,
     * and connects and authenticates.
     * 
     * This test does not test via the deprecated authenticate method. This test assumes that the 
     * deprecated method calls the same underlying mechanism in the XSystem and thus the deprecated
     * method has been tested thoroughly. This test is testing the underlying mechanism via the 
     * current authenticate method.
     * @throws Exception
     */
    public void testPlainViaPasswd() throws Exception  {
       announce("ReferenceAuthenticationTest.testPlainViaPasswd");
       
       // Write the temporary passwd file
       String passwd_info[][] = { 
          { "user1", "pass1"},
          { "user2", "pass2"},
          { "user3", "pass3"},
          { "user4", "pass4"},
          { "xyzzy", "plugh"} };
       File passwd_file = File.createTempFile( "passwd", "txt" );
       PrintWriter passwd_writer = new PrintWriter(passwd_file);
       for( int p = 0; p < passwd_info.length; p++ ) {
          passwd_writer.println( passwd_info[p][0] + " " + passwd_info[p][1] );
       }
       passwd_writer.close();
       
       // Grab the old value of the org.snia.xam.reference.passwd property so we can restore it later.
       String oldPasswdProperty = null;
       try {
          oldPasswdProperty = s_xam.getString(ReferenceXSystem.PROPERTY_PASSWD);
       }
       catch( FieldDoesNotExistException fdne ) { ; /* ignore */ }

       if( oldPasswdProperty == null )
          s_xam.createProperty(ReferenceXSystem.PROPERTY_PASSWD, false, passwd_file.getAbsolutePath() );
       else
          s_xam.setProperty( ReferenceXSystem.PROPERTY_PASSWD, passwd_file.getAbsolutePath() );

       // Setup done, time to test
       try {
          // Try each of the usernames and passwords in the file, each of which should succeed.
          for( int p = 0; p < passwd_info.length; p++ ) {
           
            byte[] authInfo = ReferenceSaslUtils.encodeSASLPlain("junkAuthorzId",
               passwd_info[p][0], passwd_info[p][1] );
            
            XSystem xsystem = doConnection(true);
            ByteArrayOutputStream respByteStream = new ByteArrayOutputStream();
            int respValue = xsystem.authenticate(ReferenceSaslUtils.SASL_MECHANISM_PLAIN.getBytes(),
               respByteStream);
            assertEquals("Test Failed. System should request more SASL data.",
               respValue,
               XSystem.XAM_SASL_IN_PROCESS);
            verifyAuthResponse(ReferenceSaslUtils.SASL_RESPONSE_MORE.getBytes(),
               respByteStream);
            respByteStream.reset();
            respValue = xsystem.authenticate(authInfo, respByteStream);
            assertEquals("Test Failed. System should not request more SASL data.",
               respValue,
               XSystem.XAM_SASL_COMPLETE);
            verifyAuthResponse(ReferenceSaslUtils.SASL_DONE.getBytes(),
               respByteStream);
            xsystem.close();
         }

          // Try authenticating with the default username and password.
          // Knowing that these are NOT in the passwd file, this should 
          // not succeed.
          byte[] authInfo = ReferenceSaslUtils.encodeSASLPlain("junkAuthorzId",
             ReferenceAuthenticationStatus.TEST_USERNAME,
             ReferenceAuthenticationStatus.TEST_PASSWORD );
          
          XSystem xsystem = doConnection(true);
          ByteArrayOutputStream respByteStream = new ByteArrayOutputStream();
          int respValue = xsystem.authenticate(ReferenceSaslUtils.SASL_MECHANISM_PLAIN.getBytes(),
             respByteStream);
          assertEquals("Test Failed. System should request more SASL data.",
             respValue,
             XSystem.XAM_SASL_IN_PROCESS);
          verifyAuthResponse(ReferenceSaslUtils.SASL_RESPONSE_MORE.getBytes(),
             respByteStream);
          respByteStream.reset();
          respValue = xsystem.authenticate(authInfo, respByteStream);
          assertEquals("Test Failed. System should not request more SASL data.",
             respValue,
             XSystem.XAM_SASL_COMPLETE);
          verifyAuthResponse(ReferenceSaslUtils.SASL_FAILED.getBytes(),
             respByteStream);
          xsystem.close();

       }
      finally {
         passwd_file.delete();
         if( oldPasswdProperty != null )
            s_xam.setProperty( ReferenceXSystem.PROPERTY_PASSWD, oldPasswdProperty );
         else
            s_xam.deleteField( ReferenceXSystem.PROPERTY_PASSWD );
       }
   }
   
    /**
     * Test to make sure that the XSystem can not authenticate more than once. 
     * The first authentication is via anonymous, the second is via plain.
     * @throws Exception
     */
    public void testAuthRetry() throws Exception  {
        announce("ReferenceAuthenticationTest.testAuthRetry");
        XSystem xsystem = doConnection(false);
        ByteArrayOutputStream response = new ByteArrayOutputStream(200);
        int retValue = xsystem.authenticate(ReferenceSaslUtils.
                SASL_MECHANISM_PLAIN.getBytes(), response);
        assertEquals("Test Failed. System should request more SASL data.",
                retValue, XSystem.XAM_SASL_IN_PROCESS);
        verifyAuthResponse(ReferenceSaslUtils.
                SASL_RESPONSE_MORE.getBytes(), response);
        
        byte[] authInfo = ReferenceSaslUtils.encodeSASLPlain(
                "junkAuthorzId", ReferenceAuthenticationStatus.TEST_USERNAME,
                "badPassword");
        response.reset();
        retValue = xsystem.authenticate(authInfo, response);
        assertEquals("Test Failed. System should NOT request more SASL data.",
                retValue, XSystem.XAM_SASL_COMPLETE);
        verifyAuthResponse(ReferenceSaslUtils.SASL_FAILED.getBytes(), 
                response);
        
        // Attempt to authenticate with valid credentials using same XSystem.
        // Once fail login should not be able to do anything with XSystem.
        authInfo = ReferenceSaslUtils.encodeSASLPlain(
                "junkAuthorzId", ReferenceAuthenticationStatus.TEST_USERNAME,
                ReferenceAuthenticationStatus.TEST_PASSWORD);
        try {
            response.reset();
            retValue = xsystem.authenticate(authInfo, response);
            assertEquals(
                "Test Failed. System should NOT request more SASL data.",
                retValue, XSystem.XAM_SASL_COMPLETE);
            verifyAuthResponse(ReferenceSaslUtils.SASL_DONE.getBytes(), 
                response);
        } catch (XSystemException ex) {
            fail("Should be able to re-authenticated after failed " +
                    "authentication.");
        }
        
        try {
            XSet xset = xsystem.createXSet(XSet.MODE_UNRESTRICTED);
         } catch (XSystemException ex) {
            fail("Should be able to access system after authentication.");
        }
        xsystem.abandon();
        xsystem.close();
    }
    
    /**
     * Basic test for properties before and after authentication
     * 
     * @throws java.lang.Exception
     */
    public void testProperties() throws Exception {
        announce("ReferenceAuthenticationTest.testProperties");
        XSystem system = doConnection(true);
        checkProperties(system, PRE_AUTH_PROPS, PRE_AUTH_TYPES);
        ByteArrayOutputStream response = new ByteArrayOutputStream(200);
        int retValue = system.authenticate(ReferenceSaslUtils.
                SASL_MECHANISM_PLAIN.getBytes(), response);
        assertEquals("Test Failed. System should request more SASL data.",
                retValue, XSystem.XAM_SASL_IN_PROCESS);
        verifyAuthResponse(ReferenceSaslUtils.SASL_RESPONSE_MORE.getBytes(), 
                response);
        byte[] creds = ReferenceSaslUtils.encodeSASLPlain(
                null, s_user, s_pass);
        response.reset();
        retValue = system.authenticate(creds, response);
        assertEquals("Test Failed. System should NOT request more SASL data.",
                retValue, XSystem.XAM_SASL_COMPLETE);
        verifyAuthResponse(
                ReferenceSaslUtils.SASL_SUCCESS.getBytes(), response);
        checkProperties(system, POST_AUTH_PROPS, POST_AUTH_TYPES);
        system.close();
    }
}
