/*
 *  Copyright (c) 2008-2009, Sun Microsystems, Inc.
 *  Copyright (c) 2008-2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.io.UnsupportedEncodingException;
import java.util.Calendar;
import java.util.Random;

import junit.framework.Test;
import junit.framework.TestSuite;

import org.snia.xam.QueryException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XSet;

import org.snia.xam.XStream;
import org.snia.xam.XSystem;
import org.snia.xam.XUID;
import org.snia.xam.util.ISO8601Date;
import org.snia.xam.util.XUIDIterator;

/**
 * Retention Time Criteria (RTC) tests
 * 
 */
public class ReferenceQuerySemanticTest extends ReferenceXAMTest {

   private static final String XAM_STRING   = ".xam_string";
   private static final String XAM_DATETIME = ".xam_datetime";
   private static final String XAM_XUID     = ".xam_xuid";
   private static final String XAM_INT      = ".xam_int";
   private static final String XAM_DOUBLE   = ".xam_double";
   private static final String XAM_BOOLEAN  = ".xam_boolean";
   private static final String XAM_EVEN     = ".xam_even";
   private static final String XAM_ODD      = ".xam_odd";
   private static final String XAM_PROMOTE  = ".xam_promote";
   private static final String TROUBLE      = ".trouble";
   
   private static final String STREAM_CONTENT = ".streamContent";
   
   private static final String OP_EQUAL     = "=";
   private static final String OP_NE        = "<>";
   private static final String OP_GT        = ">";
   private static final String OP_GE        = ">=";
   private static final String OP_LT        = "<";
   private static final String OP_LE        = "<=";
   private static final String OP_LIKE      = " like ";
   
   static final String COM_EXAMPLE_PROPERTY = "com.example.property";
   static       String PROPERTY_NAME ;
   
   static final int COUNT = 10;
   static final String   VALUE_S = "testSimpleWhere.value";
   static final Calendar VALUE_C = Calendar.getInstance();
   static       XUID     VALUE_X ;
   
    public ReferenceQuerySemanticTest(String name) {
        super(name);
        
        try { 
           VALUE_X = SniaXUID.generateXUID();      
        } catch( Exception e ) { /* Should never happen. */ }
    }

    public static void main(String[] args) {
        System.exit(runTest(args, suite()));
    }

    public static Test suite() {
        TestSuite suite = new TestSuite();
        suite.addTest(new ReferenceQuerySemanticTest("testSimpleWhere"));
        suite.addTest(new ReferenceQuerySemanticTest("testSimpleWhereOps"));
        suite.addTest(new ReferenceQuerySemanticTest("testStringOps"));
        suite.addTest(new ReferenceQuerySemanticTest("testAttributeFunctions"));
        suite.addTest(new ReferenceQuerySemanticTest("testLogicalWhereOps"));
        suite.addTest(new ReferenceQuerySemanticTest("testNonExistantProperties"));
        suite.addTest(new ReferenceQuerySemanticTest("testGeneralizedAttributes"));
        suite.addTest(new ReferenceQuerySemanticTest("testGeneralizedFieldAttributes"));
        suite.addTest(new ReferenceQuerySemanticTest("testExistFunction"));
        suite.addTest(new ReferenceQuerySemanticTest("testCannonicalQueries"));
        suite.addTest(new ReferenceQuerySemanticTest("testComplexTypePromotion"));
        suite.addTest(new ReferenceQuerySemanticTest("test754Exceptions"));
        suite.addTest(new ReferenceQuerySemanticTest("test754ExceptionsNonOptimized"));
        suite.addTest(new ReferenceQuerySemanticTest("testAccessRestrictions"));
        suite.addTest(new ReferenceQuerySemanticTest("testLevel2"));
        suite.addTest(new ReferenceQuerySemanticTest("testTroublesomeCharacters"));
        return suite;
    }
    
    
    /**
     * Test a query with a simple where clause.
     * @throws java.lang.Exception
     */
    public void testSimpleWhere() throws Exception {
        announce("ReferenceQuerySemanticTest.testSimpleWhere");

        XSystem system = connectAndAuthenticate(true);

        PROPERTY_NAME = saltStore( system, COUNT, 
           VALUE_S, VALUE_C, VALUE_X );;
        
        runAndTestQuery( system, PROPERTY_NAME, OP_EQUAL, VALUE_S, COUNT );
        runAndTestQuery( system, PROPERTY_NAME, OP_EQUAL, VALUE_C, COUNT );
        runAndTestQuery( system, PROPERTY_NAME, OP_EQUAL, VALUE_X, COUNT );
        runAndTestQuery( system, PROPERTY_NAME, OP_EQUAL, Long.valueOf(0), 1 );
        runAndTestQuery( system, PROPERTY_NAME, OP_EQUAL, Double.valueOf(0), 1 );
        runAndTestQuery( system, PROPERTY_NAME, OP_EQUAL, Boolean.valueOf(true), COUNT/2 );
        
        system.close();
    }
    
    /**
     * Run simple where statements with a variety of binary operators.
     * @throws Exception
     */
    public void testSimpleWhereOps() throws Exception {
       announce( "ReferenceQUerySemanticsTest.testSimpleWhereOps" );
       XSystem system = connectAndAuthenticate(true);

       runAndTestQuery( system, PROPERTY_NAME, OP_NE, VALUE_S, 0 );
       
       runAndTestQuery( system, PROPERTY_NAME, OP_GT, Long.valueOf(0), COUNT-1 );
       runAndTestQuery( system, PROPERTY_NAME, OP_GE, Long.valueOf(0), COUNT );
       runAndTestQuery( system, PROPERTY_NAME, OP_LT, Long.valueOf(COUNT-1), COUNT-1 );
       runAndTestQuery( system, PROPERTY_NAME, OP_LE, Long.valueOf(COUNT), COUNT );
       
       runAndTestQuery( system, PROPERTY_NAME, OP_GT, Double.valueOf((COUNT/2)/100.0), COUNT/2-1 );
       runAndTestQuery( system, PROPERTY_NAME, OP_GE, Double.valueOf(0.01), COUNT -1 );
       runAndTestQuery( system, PROPERTY_NAME, OP_LT, Double.valueOf((COUNT-1) / 100.0), COUNT-1 );
       runAndTestQuery( system, PROPERTY_NAME, OP_LE, Double.valueOf((COUNT-1) / 100.0), COUNT );
       
       runAndTestQuery( system, PROPERTY_NAME, OP_NE, VALUE_X, 0 );

       runAndTestQuery( system, PROPERTY_NAME, OP_NE, VALUE_C, 0 );
       
       // Test to make sure that prohibited operator / operand comparisons are rejected.
       // See the architecture specification, section 10.4
       // Prohibited Boolean Operations
       try { 
          runAndTestQuery( system, PROPERTY_NAME, OP_LT, Boolean.FALSE, 0 );
          fail( "Query should not accept < for boolean comparison" );
       } catch( QueryException e ) { /* Expected exception */ }

       try { 
          runAndTestQuery( system, PROPERTY_NAME, OP_LE, Boolean.FALSE, 0 );
          fail( "Query should not accept < for boolean comparison" );
       } catch( QueryException e ) { /* Expected exception */ }

       try { 
          runAndTestQuery( system, PROPERTY_NAME, OP_GT, Boolean.FALSE, 0 );
          fail( "Query should not accept < for boolean comparison" );
       } catch( QueryException e ) { /* Expected exception */ }

       try { 
          runAndTestQuery( system, PROPERTY_NAME, OP_GE, Boolean.FALSE, 0 );
          fail( "Query should not accept < for boolean comparison" );
       } catch( QueryException e ) { /* Expected exception */ }

       // Prohibited XUID operators
       try { 
          runAndTestQuery( system, PROPERTY_NAME, OP_LT, VALUE_X, 0 );
          fail( "Query should not accept < for XUID comparison" );
       } catch( QueryException e ) { /* Expected exception */ }

       try { 
          runAndTestQuery( system, PROPERTY_NAME, OP_LE, VALUE_X, 0 );
          fail( "Query should not accept < for XUID comparison" );
       } catch( QueryException e ) { /* Expected exception */ }

       try { 
          runAndTestQuery( system, PROPERTY_NAME, OP_GT, VALUE_X, 0 );
          fail( "Query should not accept < for XUID comparison" );
       } catch( QueryException e ) { /* Expected exception */ }

       try { 
          runAndTestQuery( system, PROPERTY_NAME, OP_GE, VALUE_X, 0 );
          fail( "Query should not accept < for XUID comparison" );
       } catch( QueryException e ) { /* Expected exception */ }

       system.abandon();
       system.close();
    }
    
    /**
     * Test basic string "like" operator.
     * @throws Exception
     */
    public void testStringOps() throws Exception
    {
       announce( "ReferenceQUerySemanticsTest.testSimpleWhereOps" );
       XSystem system = connectAndAuthenticate(true);
       
       runAndTestQuery( system, PROPERTY_NAME, OP_LIKE, "%" + VALUE_S + "%", COUNT );

       system.abandon();
       system.close();
    }

    /**
     * Test length(), binding(), readonly()
     * @throws Exception
     */
    public void testAttributeFunctions() throws Exception
    {
       announce( "ReferenceQuerySemanticsTest.testAttributeFunctions");
       XSystem system = connectAndAuthenticate(true);
       
       // Test "length()" function
       String boolPropName = PROPERTY_NAME + XAM_BOOLEAN;
       String xamql = "select \"" + XSet.XAM_XUID + "\" where length(\""
         + boolPropName + "\") = 1" ;
       runAndTestQuery( system, xamql, COUNT );   
       
       // Test "binding()" function
       String bindingEquals = "select \"" + XSet.XAM_XUID + "\" where binding(\""
       + boolPropName + "\") = true";
       runAndTestQuery( system, bindingEquals, COUNT / 2 );
       
       String bindingBareFunction = "select \"" + XSet.XAM_XUID + "\" where binding(\""
       + boolPropName + "\")";
       runAndTestQuery( system, bindingBareFunction, COUNT / 2 );
       
       // Test "readonly() function
       long totalXSetsInStore = runAndTestQuery( system, "select \".xset.xuid\"", -1 );
       runAndTestQuery( system,  
          "select \".xset.xuid\" where readonly(\".xset.xuid\")", (int) totalXSetsInStore );
       
       // Test "typeof()" function
       String typeOfQuery = "select \"" + XSet.XAM_XUID + "\" where typeof(\""
           + boolPropName + "\") = '" + XAMLibrary.STYPE_BOOLEAN_MIME_TYPE + "'";
       runAndTestQuery( system, typeOfQuery, COUNT );
       
       system.close();
    }
    
    /**
     * Test the basic form of the "and", "or", "not" operators which will produce more
     * complex where clauses. This also causes the query processor to generate an 
     * intermediate table. 
     * @throws Exception
     */
    public void testLogicalWhereOps() throws Exception
    {
       announce( "ReferenceQuerySemanticsTest.testAttributeFunctions");
       XSystem system = connectAndAuthenticate(true);
       
       String boolPropName    = PROPERTY_NAME + XAM_BOOLEAN;
       String intPropName     = PROPERTY_NAME + XAM_INT;
       String doublePropName  = PROPERTY_NAME + XAM_DOUBLE;
       String stringPropName  = PROPERTY_NAME + XAM_STRING;
       String xuidPropName    = PROPERTY_NAME + XAM_XUID;
       
       // Simple combinations of a single property
       String xamql = "select \"" + XSet.XAM_XUID + "\" where (\""
         + boolPropName + "\" = true) or (\"" + boolPropName + "\" = false)" ;
       runAndTestQuery( system, xamql, COUNT );   
       
       xamql = "select \"" + XSet.XAM_XUID + "\" where \""
         + boolPropName + "\" or  not \"" + boolPropName + "\"" ;
       runAndTestQuery( system, xamql, COUNT );
       
       // Combinations of two properties
       xamql = "select \"" + XSet.XAM_XUID + "\" where \""
          + boolPropName + "\" and \"" + stringPropName + 
          "\" like '" + VALUE_S + "'";
       runAndTestQuery( system, xamql, COUNT/2 );
      
       xamql = "select \"" + XSet.XAM_XUID + "\" where \""
       + boolPropName + "\" and \"" + intPropName + 
       "\" > " + COUNT/2 + "";
       runAndTestQuery( system, xamql, COUNT/4 );
       
       // Pick out even values, the hard way.
       xamql = "select \"" + XSet.XAM_XUID + "\" where \"" + 
       intPropName + "\" = 0 or \"" +
       intPropName + "\" = 2 or \"" +
       intPropName + "\" = 4 or \"" +
       intPropName + "\" = 6 or \"" +
       intPropName + "\" = 8";
       runAndTestQuery( system, xamql, 5 );
       
       // Try more than two properties
       xamql = "select \"" + XSet.XAM_XUID + "\" where \"" + 
       intPropName + "\" > 0 or \"" +
       boolPropName + "\" = TRUE or \"" +
       stringPropName + "\" like '" + VALUE_S + "' or \"" +
       doublePropName + "\"> 0.0 or \"" +
       xuidPropName + "\" = xuid('" + VALUE_X + "')";
       runAndTestQuery( system, xamql, COUNT );
   
      system.close();
    }
    
    /**
     * Test the basic form of the "and", "or", "not" operators which will produce more
     * complex where clauses. This also causes the query processor to generate an 
     * intermediate table. 
     * @throws Exception
     */
    public void testNonExistantProperties() throws Exception
    {
       Random rand = new Random();
       
       announce( "ReferenceQuerySemanticsTest.testAttributeFunctions");
       XSystem system = connectAndAuthenticate(true);
              
       // single property
       String xamql = "select \"" + XSet.XAM_XUID + "\" where (\""
         + "dkdknfvffkkndikjw[pcklm\" = 3) " ;
       runAndTestQuery( system, xamql, 0 );   
   
       // Lots of non existent property names. Syntactically legal, but had BETTER return zero!
      StringBuffer tmp = new StringBuffer("select \".xset.xuid\" where ");
      for( int i = 0; i < 10; i++ )
      {
         if( i > 0 )
            tmp.append( " or " );
         tmp.append( "\"" );
         tmp.append(rand.nextInt());
         tmp.append("\" = " );
         tmp.append( rand.nextDouble() );
      }
      runAndTestQuery( system, tmp.toString(), 0 );   
      system.close();
    }
    
    /**
     * Tests the generalized parser, and the presence of attribute functons.
     * @throws Exception
     */
    public void testGeneralizedAttributes() throws Exception
    {
       announce( "ReferenceQuerySemanticsTest.testGeneralizedAttribute");
       XSystem system = connectAndAuthenticate(true);
              
       String boolPropName    = PROPERTY_NAME + XAM_BOOLEAN;
       String intPropName     = PROPERTY_NAME + XAM_INT;
       String doublePropName  = PROPERTY_NAME + XAM_DOUBLE;
       String stringPropName  = PROPERTY_NAME + XAM_STRING;
       String xuidPropName    = PROPERTY_NAME + XAM_XUID;
       String datePropName    = PROPERTY_NAME + XAM_DATETIME;

       // Two properties with typeof() and length()
       // Ex: select ".xset.xuid" where binding(".xset.xuid") and 
       //                               not readonly("com.example.property.1236278955186.xam_int") 

       String xamql = "select \"" + XSet.XAM_XUID + "\" where ( length(\""
         + intPropName + "\") = 8 and typeof(\"" + intPropName + "\") = '"
         + XAMLibrary.STYPE_INT_MIME_TYPE + "') or ( length(\"" 
         + doublePropName + "\") = 8 and typeof(\"" + doublePropName +"\") = '" 
         + XAMLibrary.STYPE_DOUBLE_MIME_TYPE + "')";
       runAndTestQuery( system, xamql, COUNT );   
       
       // Two properties and binding() and readonly()
       xamql = "select \"" + XSet.XAM_XUID + "\" where binding(\""
       + XSet.XAM_XUID + "\") and not readonly(\"" + intPropName + "\") ";
   
      runAndTestQuery( system, xamql, COUNT );   

      // Four attribute and four  or more property values.
      // Ex: select ".xset.xuid" where binding(".xset.xuid") and 
      //                         not readonly("com.example.property.1236278955186.xam_int") and 
      //                         length("com.example.property.1236278955186.xam_double") = 8 and 
      //                         typeof("com.example.property.1236278955186.xam_string") = 'application/vnd.snia.xam.string' and 
      //                         ("com.example.property.1236278955186.xam_int" > 0 or 
      //                          "com.example.property.1236278955186.xam_double" > 0 or 
      //                          "com.example.property.1236278955186.xam_string" = 'testSimpleWhere.value' or 
      //                          "com.example.property.1236278955186.xam_xuid" = xuid('AAA6AwAe0PkxMjM2Mjc4OTUzNDkwB03Xl+YAE1gt' or 
      //                          "com.example.property.1236278955186.xam_datetime" = date('2009-03-05T10:49:13.472-08:00'))

      xamql = "select \"" + XSet.XAM_XUID + "\" where binding(\""
      + XSet.XAM_XUID + "\") and not readonly(\"" + intPropName + "\") and "
      + "length(\"" + doublePropName + "\") = 8 and typeof(\"" + stringPropName 
      + "\") = '" + XAMLibrary.STYPE_STRING_MIME_TYPE + "' and (\""
      + intPropName + "\" > 0 or \"" + doublePropName + "\" > 0 or \"" 
      + stringPropName + "\" = '" + VALUE_S + "' or \"" + xuidPropName + 
      "\" = xuid('" + VALUE_X + "') or \"" + datePropName + "\" = date('" + 
      ISO8601Date.toISO8601String(VALUE_C) + "'))";
  
     runAndTestQuery( system, xamql, COUNT );   
     system.close();
    }
    
    /**
     * Tests the generalized parser, and the presence of attribute functons.
     * @throws Exception
     */
    public void testGeneralizedFieldAttributes() throws Exception
    {
       announce( "ReferenceQuerySemanticsTest.testGeneralizedFieldAttribute");
       XSystem system = connectAndAuthenticate(true);
              
       String streamName = PROPERTY_NAME + ".testStream";

       // One properties with typeof()  

       String xamql = "select \"" + XSet.XAM_XUID + "\" where typeof(\""
         + streamName + "\") like 'application/0' or typeof(\"" 
         + streamName + "\") like 'application/1' or typeof(\""
         + streamName + "\") like 'application/2' or typeof(\""
         + streamName + "\") like 'application/3' or typeof(\""
         + streamName + "\") like 'application/5' or typeof(\""
         + streamName + "\") like 'application/7'";
       runAndTestQuery( system, xamql, 6 );   

       // Check out the other attributes.
       xamql = "select \"" + XSet.XAM_XUID + "\" where length(\""
       + streamName + "\") > 100 and not readonly(\""
       + streamName + "\") and not binding(\""
       + streamName + "\")";
       runAndTestQuery( system, xamql, 10);   
       system.close();
    }
    
    /**
     * Test "exists()" function. Make sure that the presence of a property is detected, as
     * well as the absence.
     * @throws Exception
     */
    public void testExistFunction() throws Exception
    {
       announce( "ReferenceQuerySemanticsTest.testGeneralizedFieldAttribute");
       XSystem system = connectAndAuthenticate(true);
       
       String boolPropName    = PROPERTY_NAME + XAM_BOOLEAN;
       String intPropName     = PROPERTY_NAME + XAM_INT;
       String doublePropName  = PROPERTY_NAME + XAM_DOUBLE;
       String stringPropName  = PROPERTY_NAME + XAM_STRING;
       String xuidPropName    = PROPERTY_NAME + XAM_XUID;
       String datePropName    = PROPERTY_NAME + XAM_DATETIME;
       String evenPropName    = PROPERTY_NAME + XAM_EVEN;
       String oddPropName     = PROPERTY_NAME + XAM_ODD;

       String xamql = "select \"" + XSet.XAM_XUID + "\" where exists(\""
       + boolPropName + "\")";
       runAndTestQuery( system, xamql, COUNT);                 

       xamql = "select \"" + XSet.XAM_XUID + "\" where exists(\""
       + evenPropName + "\")";
       runAndTestQuery( system, xamql, COUNT / 2);   
              
       xamql = "select \"" + XSet.XAM_XUID + "\" where exists(\""
       + oddPropName + "\")";
       runAndTestQuery( system, xamql, COUNT / 2);   
             
       // Sum both even and odd properties
       xamql = "select \"" + XSet.XAM_XUID + "\" where exists(\""
       + oddPropName + "\") or exists(\"" + evenPropName + "\")";
       runAndTestQuery( system, xamql, COUNT);   

       // Make sure that "not exists" does the right thing
       long totalXSetsInStore = runAndTestQuery( system, "select \".xset.xuid\"", -1 );
       xamql = "select \"" + XSet.XAM_XUID + "\" where not exists(\""
          + evenPropName + "\")";
       runAndTestQuery( system, xamql, (int) totalXSetsInStore - COUNT/2);
       system.close();
    }
    
    /**
     * As closely as possible, run the cannonical queries as presented in section 10.8
     * of the XAM Architectural specification. Note that the property names are modified
     * with the addition of currentTimeMilis to avoid confusion with preexisting XSets in
     * the store. 
     * @throws Exception
     */
    public void testCannonicalQueries() throws Exception
    {
       announce( "ReferenceQuerySemanticsTest.testGeneralizedFieldAttribute");
       XSystem system = connectAndAuthenticate(true);
       
       String prefix = "com.example." + System.currentTimeMillis();
       String p1 = prefix + ".foo";
       String p2 = prefix + ".bar";
       String p3 = prefix + ".num";
       
       XSet xs = system.createXSet(XSet.MODE_UNRESTRICTED);
       xs.createProperty( p1, false, 1 );
       xs.createProperty( p2, false, "string" );
       xs.createProperty( p3, false, 123.55 );
       XUID xuid1 = xs.commit();
       xs.close();
       
       xs = system.createXSet(XSet.MODE_UNRESTRICTED);
       xs.createProperty( p1, false, 77 );
       xs.createProperty( p2, false, 42 );
       xs.createProperty( p3, false, 100 );
       XUID xuid2 = xs.commit();
       xs.close();
       
       xs = system.createXSet(XSet.MODE_UNRESTRICTED);
       xs.createProperty( p1, false, 6 );
       xs.createProperty( p3, false, 200 );
       XUID xuid3 = xs.commit();
       xs.close();
       
       // 10.8.2
       XUIDIterator iter_10_8_2 = runAndTestQuery( system,
          "select \".xset.xuid\" where (\""+ p1 + "\" > 0) and (\"" + p1 + "\" < 50)" );
       boolean have1 = false, have3 = false;
       while( iter_10_8_2.hasNext())
       {
          XUID test = (XUID) iter_10_8_2.next();
          if( xuid1.equals(test))
             have1 = true;
          else if( xuid3.equals(test))
             have3 = true;
          else
             fail("Unexpected result in cannonical test 10.8.2");
       }
       assertTrue( "Did not find XSET1 from cannonical test 10.8.2", have1 );
       assertTrue( "Did not find XSet2 from cannonical test 10.8.2", have3 );
       
       // 10.8.3
       XUIDIterator iter_10_8_3 = runAndTestQuery( system,
          "select \".xset.xuid\" where (\""+ p2 + "\" > 0) and (\"" + p2 + "\" < 100)" );
       boolean have2 = false;
       while( iter_10_8_3.hasNext())
       {
          XUID test = (XUID) iter_10_8_3.next();
          if( xuid2.equals(test))
             have2 = true;
          else
             fail("Unexpected result in cannonical test 10.8.3");
       }
       assertTrue( "Did not find XSET1 from cannonical test 10.8.3", have2 );
       
       // 10.8.4
       XUIDIterator iter_10_8_4 = runAndTestQuery( system,
          "select \".xset.xuid\" where exists(\""+ p2 + "\")" );
       have1 = have2 = false;
       while( iter_10_8_4.hasNext())
       {
          XUID test = (XUID) iter_10_8_4.next();
          if( xuid1.equals(test))
             have1 = true;
          else if( xuid2.equals(test))
             have2 = true;
          else
             fail("Unexpected result in cannonical test 10.8.4");
       }
       assertTrue( "Did not find XSET1 from cannonical test 10.8.4", have1 );
       assertTrue( "Did not find XSET1 from cannonical test 10.8.4", have2 );
       
       // 10.8.5
       XUIDIterator iter_10_8_5 = runAndTestQuery( system,
          "select \".xset.xuid\" where \"" + p2 + "\" like '%ing%'" );
       have1 = false;
       while( iter_10_8_5.hasNext())
       {
          XUID test = (XUID) iter_10_8_5.next();
          if( xuid1.equals(test))
             have1 = true;
          else
             fail("Unexpected result in cannonical test 10.8.5");
       }
       assertTrue( "Did not find XSET1 from cannonical test 10.8.5", have1 );
       
       // 10.8.6
       XUIDIterator iter_10_8_6 = runAndTestQuery( system,
          "select \".xset.xuid\" where \""+ p3 + "\" > 124" );
       have3 = false;
       while( iter_10_8_6.hasNext())
       {
          XUID test = (XUID) iter_10_8_6.next();
          if( xuid3.equals(test))
             have3 = true;
          else
             fail("Unexpected result in cannonical test 10.8.6");
       }
       assertTrue( "Did not find XSET3 from cannonical test 10.8.6", have3 );
       
       // 10.8.7 
       XUIDIterator iter_10_8_7 = runAndTestQuery( system,
          "select \".xset.xuid\" where \""+ p3 + "\" > 124.6" );
       have3 = false;
       while( iter_10_8_7.hasNext())
       {
          XUID test = (XUID) iter_10_8_7.next();
          if( xuid3.equals(test))
             have3 = true;
          else
             fail("Unexpected result in cannonical test 10.8.7");
       }
       assertTrue( "Did not find XSET3 from cannonical test 10.8.7", have3 );
       
       // 10.8.8
       XUIDIterator iter_10_8_8 = runAndTestQuery( system,
          "select \".xset.xuid\" where \""+ p3 + "\" > 123 and typeof(\""
          + p3 + "\") = 'application/vnd.snia.xam.int'" );
       have3 = false;
       while( iter_10_8_8.hasNext())
       {
          XUID test = (XUID) iter_10_8_8.next();
          if( xuid3.equals(test))
             have3 = true;
          else
             fail("Unexpected result in cannonical test 10.8.b");
       }
       assertTrue( "Did not find XSET3 from cannonical test 10.8.8", have3 );
       
       // 10.8.9
       // Test to check multiple types on a single field name
       // Should only return XSet1 and XSet2
       XUIDIterator iter_10_8_9 = runAndTestQuery( system,
          "select \".xset.xuid\" where \""+ p2 + "\" <100 or \""
          + p2 + "\" like 'string'" );
       have1 = have2 = have3 = false;
       while( iter_10_8_9.hasNext())
       {
          XUID test = (XUID) iter_10_8_9.next();
          if( xuid1.equals(test))
             have1 = true;
          else if( xuid2.equals(test))
             have2 = true;
          else
             fail("Unexpected result in cannonical test 10.8.extra");
       }
       assertTrue( "Did not find XSET1 from cannonical test 10.8.9", have1 );
       assertTrue( "Did not find XSET2 from cannonical test 10.8.9", have2 );

       system.abandon();
       system.close();
    }
    
    /**
     * This test will duplicate a canonical test (type promotion), but does so in
     * two ways. One will exercise the optimized sql generator, and the other will
     * use the generalized test.
     * @throws Exception
     */
    public void testComplexTypePromotion() throws Exception
    {
       announce( "ReferenceQuerySemanticsTest.testGeneralizedFieldAttribute");
       XSystem system = connectAndAuthenticate(true);
       String p = PROPERTY_NAME + XAM_PROMOTE;
       
       // Optimized promotion
       String xamql = "select \".xset.xuid\" where \""+ p + "\" >= 0";      
       runAndTestQuery( system, xamql, COUNT);   

       // Promotion with the full SQL path
       xamql = "select \".xset.xuid\" where (\""+ p + "\" >= 0) and exists(\"" + p + "\")";      
       runAndTestQuery( system, xamql, COUNT);   

       system.abandon();
       system.close();
    }
    
    /**
     * This test Check to make sure that IEEE-754 constants are handled appropriately
     * @throws Exception
     */
    public void test754Exceptions() throws Exception
    {
       announce( "ReferenceQuerySemanticsTest.test754Exceptions");
       XSystem system = connectAndAuthenticate(true);
       String p = PROPERTY_NAME + XAM_DOUBLE;
       
       // >= Infinity, should be zero results
       String xamql = "select \".xset.xuid\" where \""+ p + "\" >= Inf";      
       runAndTestQuery( system, xamql, 0 );   

       // <= Infinity, should be COUNT results
       xamql = "select \".xset.xuid\" where \""+ p + "\" <= Inf";      
       runAndTestQuery( system, xamql, COUNT );   

       // >= -Infinity, should be COUNT results
       xamql = "select \".xset.xuid\" where \""+ p + "\" >= -Inf";      
       runAndTestQuery( system, xamql, COUNT );   

       // <= -Infinity, should be zero results
       xamql = "select \".xset.xuid\" where \""+ p + "\" <= -Inf";      
       runAndTestQuery( system, xamql, 0 );   

       // == -Infinity, should be zero results
       xamql = "select \".xset.xuid\" where \""+ p + "\" = Inf";      
       runAndTestQuery( system, xamql, 0 );   

       // == Nan, should be zero results
       xamql = "select \".xset.xuid\" where \""+ p + "\" = NaN";      
       runAndTestQuery( system, xamql, 0 );   

       system.abandon();
       system.close();
    }
    
    /**
     * This test Check to make sure that IEEE-754 constants are handled appropriately.
     * The test forces an exists() so the query runs non-optimized
     * @throws Exception
     */
    public void test754ExceptionsNonOptimized() throws Exception
    {
       announce( "ReferenceQuerySemanticsTest.test754Exceptions");
       XSystem system = connectAndAuthenticate(true);
       String p = PROPERTY_NAME + XAM_DOUBLE;
       String existsSuffix = " and exists(\"" + p + "\")";
       
       // >= Infinity, should be zero results
       String xamql = "select \".xset.xuid\" where \""+ p + "\" >= Inf" + existsSuffix;      
       runAndTestQuery( system, xamql, 0 );   

       // <= Infinity, should be COUNT results
       xamql = "select \".xset.xuid\" where \""+ p + "\" <= Inf" + existsSuffix;      
       runAndTestQuery( system, xamql, COUNT );   

       // >= -Infinity, should be COUNT results
       xamql = "select \".xset.xuid\" where \""+ p + "\" >= -Inf" + existsSuffix;      
       runAndTestQuery( system, xamql, COUNT );   

       // <= -Infinity, should be zero results
       xamql = "select \".xset.xuid\" where \""+ p + "\" <= -Inf" + existsSuffix;      
       runAndTestQuery( system, xamql, 0 );   

       // == -Infinity, should be zero results
       xamql = "select \".xset.xuid\" where \""+ p + "\" = Inf" + existsSuffix;      
       runAndTestQuery( system, xamql, 0 );   

       // == Nan, should be zero results
       xamql = "select \".xset.xuid\" where \""+ p + "\" = NaN" + existsSuffix;      
       runAndTestQuery( system, xamql, 0 );  
       
       system.abandon();
       system.close();
    }
    
    /**
     * This test checks to make sure that XSets created by one use, and restricted by access policies are
     * accessible as detailed in the access policy (included or not included in the query results) as appropriate.
     * <b>WARNING:</b> This test uses hard coded access policies for SniaAccessReadAll and SniaAccessReadWriteAll. 
     * If the AccessPolicyManager changes these names, this test will break. 
     * @throws Exception
     */
    public void testAccessRestrictions() throws Exception
    {
       announce( "ReferenceQuerySemanticsTest.testAccessRestrictions");
       XSystem system = connectAndAuthenticate(true);
       XSystem xsys2 = getAuthenticatedXSystem( PASSWD_INFO[0][0], PASSWD_INFO[0][1]  );

       String p = PROPERTY_NAME + XAM_DOUBLE;
       String p2 = PROPERTY_NAME + ".access.policy.test";
       
       // >= Infinity, should be zero results
       String xamql = "select \".xset.xuid\" where \""+ p + "\" > -1" ;      
       runAndTestQuery( system, xamql, COUNT );   
       runAndTestQuery( xsys2,  xamql, COUNT );
       
       // Create a couple of XSets in xsys2, isolate some, make others available.
       XSet xs = xsys2.createXSet( XSet.MODE_UNRESTRICTED );
       xs.applyAccessPolicy( false, PASSWD_INFO[1][0] ); // Restrict it just this user
       xs.createProperty(p2, false, p2);
       XUID xset1 = xs.commit();
       xs.close();
       
       xs = xsys2.createXSet( XSet.MODE_UNRESTRICTED );
       xs.applyAccessPolicy( false, AccessPolicyManager.STR_ACCESS_POLICY_READ_ALL );
       xs.createProperty(p2, false, p2);
       XUID xset2 = xs.commit();
       xs.close();
       
       xs = xsys2.createXSet( XSet.MODE_UNRESTRICTED);
       xs.applyAccessPolicy( false, AccessPolicyManager.STR_ACCESS_POLICY_READ_WRITE_ALL );
       xs.createProperty(p2, false, p2);
       XUID xset3 = xs.commit();
       xs.close();

       xsys2.abandon();
       xsys2.close();

       // Make sure the policies are enforced on query.
       // Should only see XSet2 and XSet2, since XSet 1 has user only access.
       xamql = "select \".xset.xuid\" where \""+ p2 + "\" like '%test%'" ;      
       XUIDIterator iter = runAndTestQuery( system, xamql );
       boolean have2 = false, have3 = false;
       while( iter.hasNext())
       {
          XUID test = (XUID) iter.next();
          if( xset2.equals(test))
             have2 = true;
          else if( xset3.equals(test))
             have3 = true;
          else if( xset1.equals(test))
             fail("XSet1 should not have been returned. " + test.toString() );
          else  
             fail("Unexpected result in query access policy test " + test.toString());
       }
       assertTrue( "Did not find XSet1 from query access policy test", have2 );
       assertTrue( "Did not find XSet2 from query access policy test", have3 );
       assertFalse( "Should be no other results in the query", iter.hasNext() );

       system.abandon();
       system.close();
    }
    
    /** 
     * Test the level 2 constructs.
     */
    public void testLevel2()
       throws Exception
    {
       XSystem system = connectAndAuthenticate(true);
       assertTrue( system.getBoolean( XSystem.XAM_XSYSTEM_JOB_QUERY_LEVEL2_SUPPORTED ));
       
       String p = PROPERTY_NAME + STREAM_CONTENT ;
       
       String xamql = "select \".xset.xuid\" where \""+ p + "\" contains('ipsum')";      
       runAndTestQuery( system, xamql, 10 );   
       
       // A bit more complicated
       xamql = "select \".xset.xuid\" where \""+ p + "\" contains('ipsum') and \"" +
                     p + "\" contains('tempus')";
       runAndTestQuery( system, xamql, 10 ); 
       
       // A bit more complicated
       xamql = "select \".xset.xuid\" where \""+ p + "\" contains('XSET0') or \"" +
                      p + "\" contains('XSET2') or \"" +
                      p + "\" contains('XSET4') or \"" +
                      p + "\" contains('XSET6') or \"" +
                      p + "\" contains('XSET8')";
       runAndTestQuery( system, xamql, 5 ); 
       
       // A bit more restrictive
       xamql = "select \".xset.xuid\" where \""+ p + "\" contains('ipsum') and \"" +
                     p + "\" contains('XSET0')";
       runAndTestQuery( system, xamql, 1 ); 
               
       // The before operator
       xamql = "select \".xset.xuid\" where \""+ p + "\" before('sagittis','suscipit') " ;
       runAndTestQuery( system, xamql, 10 ); 
       
       // The before operator, big distance between the two
       xamql = "select \".xset.xuid\" where \""+ p + "\" before('ipsum','venenatis') " ;
       runAndTestQuery( system, xamql, 10 ); 
       
       // The after operator
       xamql = "select \".xset.xuid\" where \""+ p + "\" after('luctus','vitae') " ;
       runAndTestQuery( system, xamql, 10 ); 
       
       // The after operator, big distance between the two.
       xamql = "select \".xset.xuid\" where \""+ p + "\" after('venenatis','ultricies') " ;
       runAndTestQuery( system, xamql, 10 ); 
       
       // The within operator
       xamql = "select \".xset.xuid\" where \""+ p + "\" within('blandit','pellentesque', 4) " ;
       runAndTestQuery( system, xamql, 10 ); 
       
       // The within operator, but with text reversed
       xamql = "select \".xset.xuid\" where \""+ p + "\" within('pellentesque','blandit', 4) " ;
       runAndTestQuery( system, xamql, 10 ); 
       
       // Make sure that xsets are removed from level 2 query structures.
       String sName = PROPERTY_NAME + ".delete_stream_test";
       
       XSet xset = system.createXSet( XSet.MODE_UNRESTRICTED );
       XStream stream = xset.createXStream( sName, false, "text/plain" );
       String str = "'I wish you wouldn't keep appearing and vanishing so suddenly; you make one quite giddy!'\n" + 
                    "'All right,' said the Cat; and this time it vanished quite slowly, beginning with the end of the tail, and ending with the grin, which remained some time after the rest of it had gone." + 
                    "'Well! I've often seen a cat without a grin,' thought Alice; 'but a grin without a cat! It's the most curious thing I ever saw in all my life!'";
       stream.write( str.getBytes("utf-8") );
       stream.close();
       xset.commit();
       XUID delXUID = xset.getXUID(XSet.XAM_XUID);
       xset.close();

       xamql = "select \".xset.xuid\" where \""+ sName + "\" before('grin','cat', 2 ) " ;
       runAndTestQuery( system, xamql, 1 );       
       system.deleteXSet(delXUID);
       xamql = "select \".xset.xuid\" where \""+ sName + "\" before('grin','cat', 2 ) " ;
       runAndTestQuery( system, xamql, 0 );     
    }
    
    /**
     * This test will test when troublesome characters are included in a string. 
     * These characters may cause problems during query parsing.
     * @throws Exception
     */
    public void testTroublesomeCharacters() throws Exception
    {
       announce( "ReferenceQuerySemanticsTest.testGeneralizedFieldAttribute");
       XSystem system = connectAndAuthenticate(true);
       String p = PROPERTY_NAME + TROUBLE;
       
       // Query where there are troublesome strings.
       String xamql = "select \".xset.xuid\" where \""+ p + "\" like '%trouble%' ";      
       runAndTestQuery( system, xamql, COUNT);   

       // Make sure XAM QL quotes properly
       String xamql2 = "select \".xset.xuid\" where \""+ p + "\" like '%\\\'quoted items\\\'%' ";
       runAndTestQuery( system, xamql2, COUNT);   
       
       system.abandon();
       system.close();
    }
    
   /**
     * Create a number of XSets all loaded with a property to be queried upon.
     * @param i The number of XSets to make
     * @param stringValue The string value to use creating the properties.
     * @param calValue The calendar value to use in creating the properties
     * @param longValue The long value to use in creating the properties
     * @param doubleValue The double value to use in creating the properties
     * @param booleanValue The boolean value to use in creating the properties
     * @throws XAMException 
     * @throws UnsupportedEncodingException 
     */
    private String saltStore( XSystem xsys, int i, 
                              String   stringValue, 
                              Calendar calValue,
                              XUID     xuidValue  ) throws XAMException, UnsupportedEncodingException
    {
       String PROPERTY_NAME = COM_EXAMPLE_PROPERTY + "." + System.currentTimeMillis();
       String PROPERTY_S = PROPERTY_NAME + XAM_STRING;
       String PROPERTY_C = PROPERTY_NAME + XAM_DATETIME;
       String PROPERTY_X = PROPERTY_NAME + XAM_XUID;
       String PROPERTY_L = PROPERTY_NAME + XAM_INT;
       String PROPERTY_D = PROPERTY_NAME + XAM_DOUBLE;
       String PROPERTY_B = PROPERTY_NAME + XAM_BOOLEAN;
       String PROPERTY_EVEN = PROPERTY_NAME + XAM_EVEN;
       String PROPERTY_ODD  = PROPERTY_NAME + XAM_ODD;
       String PROPERTY_PROMOTE = PROPERTY_NAME + XAM_PROMOTE;
       
       for( int x = 0; x < i; x++ )
       {
          XSet xset = xsys.createXSet( XSet.MODE_UNRESTRICTED );
          xset.createProperty( PROPERTY_S, false, stringValue );
          xset.createProperty( PROPERTY_C, false, calValue );
          xset.createProperty( PROPERTY_X, false, xuidValue );
          xset.createProperty( PROPERTY_L, false, x );
          xset.createProperty( PROPERTY_D, false, x/100.0 );
          xset.createProperty( PROPERTY_NAME + TROUBLE, false,
                               "Sometimes 'quoted items' cause trouble" );
          if( (x % 2) == 0 )
          {
             xset.createProperty( PROPERTY_B, true, true );
             xset.createProperty( PROPERTY_EVEN, false, x );
             xset.createProperty( PROPERTY_PROMOTE, false, x );
          } else
          {
             xset.createProperty( PROPERTY_B, false, false );
             xset.createProperty( PROPERTY_ODD, false, x );
             xset.createProperty( PROPERTY_PROMOTE, false, (double) x );
          }
          
          XStream stream = xset.createXStream(PROPERTY_NAME + ".testStream", false, "application/" + x );
          byte buffer[] = new byte[1024];
          stream.write(buffer);
          stream.write(buffer);
          stream.write(buffer);
          stream.close();
          stream = null;
          
          XStream stream2 = xset.createXStream( PROPERTY_NAME + STREAM_CONTENT, false,
                                               XAMLibrary.TEXT_PLAIN_MIME_TYPE );
          StringBuffer dataSrc = new StringBuffer();
          dataSrc.append("Lorem ipsum dolor sit amet, consectetur adipiscing elit.\n");
          dataSrc.append("Donec fringilla vehicula mattis.\n");
          dataSrc.append("Nam aliquam volutpat lectus ac tristique.\n");
          dataSrc.append("Ut ultricies cursus accumsan.\n");
          dataSrc.append("Sed blandit dolor sit amet nunc pellentesque tempus.\n");
          dataSrc.append("Cras congue sagittis suscipit. Sed et urna dui.\n");
          dataSrc.append("Nullam magna odio, euismod vitae luctus at, rhoncus eu felis.\n");
          dataSrc.append("Mauris enim sapien, faucibus eget faucibus quis, blandit ut nisl.\n");
          dataSrc.append("Quisque elementum auctor diam nec porttitor.\n");
          dataSrc.append("Vivamus ut orci in tellus placerat tincidunt scelerisque venenatis arcu.\n");
          dataSrc.append( "Content salt for XSet # ");
          dataSrc.append( "XSET" );
          dataSrc.append( x );
          dataSrc.append( " " );
          dataSrc.append(PROPERTY_NAME);
          String data = dataSrc.toString();

          buffer = data.getBytes("utf-8" );
          stream2.write( buffer );
          stream2.close();
          xset.commit();
          xset.close();
       }
       return PROPERTY_NAME;
    }

    /**
     * Used by the testSimeWhere to generate, run, and test results from the salted store.
     * @param propRoot Root portion of the property names.
     * @param value The object value being searched for.
     * @param expectedCount The expected number of XSets to be returned from the query
     * @throws Exception
     */
    private long runAndTestQuery( XSystem system,
                                  String  propRoot,
                                  String  operator,
                                  Object  value,
                                  int     expectedCount ) throws Exception
    {
       String suffix = null;
       String valueStr = value.toString();
       
       if( value instanceof String )
       {
          suffix = XAM_STRING;
          valueStr = "'" + value + "'";
       }
       else if( value instanceof Calendar )
       {
          suffix = XAM_DATETIME;
          valueStr = "date('" + ISO8601Date.toISO8601String((Calendar) value) + "')";
       }
       else if( value instanceof XUID )
       {
          suffix = XAM_XUID;
          valueStr = "xuid('" + value + "')";
       }
       else if( value instanceof Long )
          suffix = XAM_INT;
       else if( value instanceof Double )
          suffix = XAM_DOUBLE;
       else if( value instanceof Boolean )
          suffix = XAM_BOOLEAN;
       else 
          throw new QueryException( "Unknown test value type: " + value );
       
       String xamql = "select \"" + XSet.XAM_XUID + "\" where \"" +
          (propRoot + suffix) + "\"" + operator + " " + valueStr ;

       return runAndTestQuery( system, xamql, expectedCount );
    }
    
    /**
     * Used by the testSimeWhere to generate, run, and test results from the salted store.
     * @param propRoot Root portion of the property names.
     * @param value The object value being searched for.
     * @param expectedCount The expected number of XSets to be returned from the query
     * @throws Exception
     */
    private long runAndTestQuery( XSystem system,
                                  String  xamql,
                                  int     expectedCount ) throws Exception
    {
       XSet xset = createXSetForQueryJob(system, false, xamql);
       xset.submitJob();

       // Wait for query job completion.
       do {
          try {
              Thread.sleep(100);
          } catch (InterruptedException ex) {
              // announce("Interrupted Exception thrown:  " + ex.getMessage());
          }
       } while (!JobStatus.isDone(xset.getString(XSet.XAM_JOB_STATUS)));

       long num = xset.getLong(XSet.XAM_JOB_QUERY_RESULTS_COUNT);
       if( expectedCount >= 0) {
          assertEquals( "Wrong number of results returned from query: " + xamql, 
             expectedCount, num );
       }
       xset.close();
       return num;
    }
    
    /**
     * Used by the testSimeWhere to generate, run, and test results from the salted store.
     * @param propRoot Root portion of the property names.
     * @param value The object value being searched for.
     * @param expectedCount The expected number of XSets to be returned from the query
     * @throws Exception
     */
    private XUIDIterator runAndTestQuery( XSystem system,
                                          String  xamql ) throws Exception
    {
       XSet xset = createXSetForQueryJob(system, false, xamql);
       xset.submitJob();

       // Wait for query job completion.
       do {
          try {
              Thread.sleep(100);
          } catch (InterruptedException ex) {
              // announce("Interrupted Exception thrown:  " + ex.getMessage());
          }
       } while (!JobStatus.isDone(xset.getString(XSet.XAM_JOB_STATUS)));
       XStream results = xset.openXStream( XSet.XAM_JOB_QUERY_RESULTS, XStream.MODE_READ_ONLY );
       XUIDIterator iter = new XUIDIterator(results);
       return iter;
    }
    
   /** Create the XSet for the query job. */
    private XSet createXSetForQueryJob(XSystem system, 
                                       boolean binding, 
                                       String xamql) throws XAMException {
       XSet xset = null;
       if (system != null && xamql != null) {
           xset = system.createXSet(XSet.MODE_UNRESTRICTED);
           xset.createProperty(XSet.XAM_JOB_COMMAND, binding, XSet.XAM_JOB_QUERY);
           XStream stream = xset.createXStream(XSet.XAM_JOB_QUERY_COMMAND,
                                                      binding, "text/plain;charset=utf-8");
           byte[] bytes = xamql.getBytes();
           int wrote = 0;
           while (wrote < bytes.length) {
               wrote += stream.write(bytes, wrote, bytes.length - wrote);
           }
           stream.close();
       }
       return xset;
   }
}
