/*
 *  Copyright (c) 2008-2009, Sun Microsystems, Inc.
 *  Copyright (c) 2008-2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.io.File;
import java.io.FilenameFilter;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;

import junit.framework.Test;
import junit.framework.TestSuite;
import org.snia.xam.ObjectInUseException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XIterator;
import org.snia.xam.XSet;

import org.snia.xam.XStream;
import org.snia.xam.XSystem;
import org.snia.xam.XUID;
import org.snia.xam.util.XUIDIterator;

/**
 * Retention Time Criteria (RTC) tests
 * 
 */
public class ReferenceQueryTest extends ReferenceXAMTest {

    protected static final String SELECT_ALL_QUERY = 
                                            "select \"" + XSet.XAM_XUID + "\"";
    private static final int BUF_SIZE = 512;
    protected static final String[] QUERY_JOB_PROPS = {
		XSet.XAM_JOB_COMMAND,
		XSet.XAM_JOB_STATUS,
		XSet.XAM_JOB_QUERY_RESULTS_COUNT,
		XSet.XAM_JOB_QUERY_LEVEL
	};

	protected static final String[] QUERY_JOB_TYPES = {
		XAMLibrary.STYPE_STRING_MIME_TYPE,
		XAMLibrary.STYPE_STRING_MIME_TYPE,
		XAMLibrary.STYPE_INT_MIME_TYPE,
		XAMLibrary.STYPE_STRING_MIME_TYPE
	};

    public ReferenceQueryTest(String name) {
        super(name);
    }

    public static void main(String[] args) {
        System.exit(runTest(args, suite()));
    }

    public static Test suite() {
        TestSuite suite = new TestSuite();
        suite.addTest(new ReferenceQueryTest("testVerifyJobType"));
        suite.addTest(new ReferenceQueryTest("testSimpleQuery"));
        suite.addTest(new ReferenceQueryTest("testQueryResultsWhileRunning"));
        suite.addTest(new ReferenceQueryTest("testCommitQueryResults"));
        suite.addTest(new ReferenceQueryTest("testAbandonQuery"));
        suite.addTest(new ReferenceQueryTest("testResubmitCommittedJob"));
        suite.addTest(new ReferenceQueryTest("testModifyQueryCommand"));
        suite.addTest(new ReferenceQueryTest("testQueryJobFieldsReadonly"));
        return suite;
    }
    
    /**
     * Test submitting jobs when missing various required properties or invalid
     * job type specified.
     * @throws java.lang.Exception
     */
    public void testVerifyJobType() throws Exception {
        announce("\nReferenceQueryTest.testVerifyJobType");
        XSystem system = connectAndAuthenticate(true);
        XSet xset1 = system.createXSet(XSet.MODE_UNRESTRICTED);
        
        // Job Missing the XAM_JOB_COMMAND property
        try {
            xset1.submitJob();
            fail("Job sumit should fail due to missing " +
                    XSet.XAM_JOB_COMMAND);
        } catch (XAMException e) {
            /* announce("Expected Exception:  " + e.getLocalizedMessage()); */
        }
        
        // Job type is not a query job
        xset1.createProperty(XSet.XAM_JOB_COMMAND, false, "invalidJobType");
        XStream stream = xset1.createXStream(XSet.XAM_JOB_QUERY_COMMAND,
                                                           false, "text/plain");
        String xamql = SELECT_ALL_QUERY;
        byte[] bytes = xamql.getBytes();
        int wrote = 0;
        while (wrote < bytes.length) {
            wrote += stream.write(bytes, wrote, bytes.length - wrote);
        }
        stream.close();
        try {
            xset1.submitJob();
            fail("Job submit should fail because not a query job");
        } catch (Exception e) {
            /* announce("Expected Exception:  " + e.getLocalizedMessage()); */
        }
        xset1.close();
        
        // Missing the XAM_JOB_QUERY_COMMAND property
        XSet xset2 = system.createXSet(XSet.MODE_UNRESTRICTED);
        xset2.createProperty(XSet.XAM_JOB_COMMAND, false, XSet.XAM_JOB_QUERY);
        try {
            xset2.submitJob();
            fail("Job submit should fail, missing " +
                    XSet.XAM_JOB_QUERY_COMMAND + " property.");
        } catch (Exception e) {
            /* announce("Expected Exception:  " + e.getLocalizedMessage()); */
        }
        xset2.close();
        system.close();
    }
    
    /**
     * Tests simple query; all XUIDs selected (no where clause)
     * 
     * @throws java.lang.Exception
     */
    public void testSimpleQuery() throws Exception {
        announce("ReferenceQueryTest.testSimpleQuery");
        ReferenceXSystem system = (ReferenceXSystem) connectAndAuthenticate(false);
       
        XSet xset = createXSetForQueryJob(system, SELECT_ALL_QUERY);
        xset.submitJob();
        
        do {
            try {
                Thread.sleep(100);
            } catch (InterruptedException ex) {
                // announce("Interrupted Exception thrown:  " + ex.getMessage());
            }
        } while (!JobStatus.isDone(xset.getString(XSet.XAM_JOB_STATUS)));

        Hashtable<String,XUID> storeResults = getStoredXSetList(system);
        
        XStream applStream = xset.openXStream(ReferenceXSet.XAM_JOB_QUERY_RESULTS, 
           XStream.MODE_READ_ONLY);
        Iterator<XUID> iter = new XUIDIterator(applStream);

        // job has completed
        long num = xset.getLong(XSet.XAM_JOB_QUERY_RESULTS_COUNT);
        // announce("Total Records: " + String.valueOf(num));
       
        StringBuffer nonExistant = new StringBuffer();
        int iterCount = 0;
        while (iter.hasNext()) {
            XUID xuid = iter.next();
            String xuidHexString = SniaXUID.convertToSafeString(xuid);

            if (xuid != null) 
                iterCount++;

            if( storeResults.containsKey(xuidHexString) )
               storeResults.remove( xuidHexString );
            else
            {
               nonExistant.append( xuidHexString );
               nonExistant.append( "\n" );
            }
        }
        assertEquals("Actual number of results in query # equal to what was reported", num, iterCount);

        // The following check used to make sense before access restrictions were put into the Reference VIM
        // It is possible that some XSets in the store are not available to the query because they are not 
        // accessible to the currently logged in user. 
        // assertEquals("Number of XSets in persistent store, not returned in query", 
        //              0, storeResults.size() );

        String nonExistantString = nonExistant.toString();
        if( nonExistantString.length() > 0 )
           fail( "Query result contained non existant XSET: " + nonExistantString );
        applStream.close();
        xset.close();
        system.close();
    }
    
    /** Utility used for testing. This method just creates a list of all the XUID/XSets 
     * present in the system. Mostly used for query tests of the form "select .xset.xuid".
     * @param xsystem The XSystem which will be tested
     * @return A Vector of all the XUID (strings) in the system.
     * @throws XAMException
     */
    private Hashtable<String,XUID> getStoredXSetList( ReferenceXSystem xsystem ) throws XAMException
    {
       Hashtable<String,XUID> results = new Hashtable<String, XUID>();
       
       String reposDir = xsystem.getRepositoryLocation();
       File dir = new File(reposDir);
       FilenameFilter hexFilter = new FilenameFilter() {
           public boolean accept(File dir, String name) {
               return (name.startsWith(ReferenceVIMPersistenceMgr.XSET_FILE_PREFIX) &&
                       name.endsWith(ReferenceVIMPersistenceMgr.XSET_FILE_SUFFIX));
           }
       };
       XUID xuid = null;
       String[] files = dir.list(hexFilter); 
       for (int idx = 0; idx < files.length; idx++) {
           int startIdx = files[idx].indexOf('_');
           int endIdx = files[idx].indexOf('.');
           String xuidString = files[idx].substring(++startIdx, endIdx);
           xuid = SniaXUID.createFromSafeString(xuidString);
           results.put(xuidString, xuid);
       }
      return results;
   }

   /**
     * Tests retrieving XUIDs returned from query while the job is still
     * running. (The XSet is not committed before the job is run.) 
     * Mainly testing that a 0 is returned on a read instead of
     * the end-of-stream marker when retrieving results from a running job.
     * 
     * Note: For this test to function correctly need enough XSets
     * stored so the job actually runs long enough for the test to query
     * while the job is running.
     * 
     * @throws java.lang.Exception
     */
    public void testQueryResultsWhileRunning() throws Exception {
        announce("ReferenceQueryTest.testQueryResultsWhileRunning");
        XSystem system = connectAndAuthenticate(true);
        XSet xset = createXSetForQueryJob(system, SELECT_ALL_QUERY);
        // introduce a 5 sec. delay to ensure the job runs long enough to
        // try and delete the corresponding XSet
        xset.createProperty(ReferenceJob.JOB_DELAY_FIELD_IN_MILLISEC, false, 5000);
        xset.submitJob();
        
        // wait until job has started running
        int numTries = 0;
        String jobStatusString = xset.getString(XSet.XAM_JOB_STATUS);
        while (numTries < 250 &&
               !JobStatus.isDoneExecution(jobStatusString) &&
               !JobStatus.isRunning(jobStatusString)) {
            Thread.sleep(1);
            numTries++;
            jobStatusString = xset.getString(XSet.XAM_JOB_STATUS);
        }
        announce("JobStatus after waiting: " + jobStatusString);
        
        XStream applStream = xset.openXStream(ReferenceXSet.XAM_JOB_QUERY_RESULTS, 
                                                XStream.MODE_READ_ONLY);
        Iterator<XUID> iter = new XUIDIterator(applStream);

        int iterCount = 0;
        while (iter.hasNext()) {
            XUID xuid = iter.next();
            // TODO:  come up with some other way to verify XUIDs this prints
            // out way too many lines in the test output
            //System.out.println("Resulting valid XUID: " + xuid.toString());
            if (xuid != null) 
                iterCount++;
            if (iterCount == 1) {
                try {
                    xset.commit();
                    fail("Should NOT be able to commit a XSet running a job.");
                 } catch (ObjectInUseException ox) {
                    /* announce(
                            "Expected exception. Tried to commit XSet " +
                            "instance that is running job\n" +
                            ox.getLocalizedMessage()); */
                } catch (XAMException ex) {
                    if (!JobStatus.isDone(xset.getString(XSet.XAM_JOB_STATUS))) {
                        fail("Incorrect exception returned when commit on " +
                                "XSet with running job: " + 
                                ex.getLocalizedMessage());
                    }
                   announce(
                            "Not exact expected exception when commit XSet " +
                            "with a running job. Try re-running when " +
                            "more XSets persisted so job takes longer.\n" +
                            ex.getLocalizedMessage());
                }
                try {
                    xset.close();
                    fail("Should NOT be able to close a XSet running a job.");
                 } catch (ObjectInUseException ox) {
                    /* announce(
                            "Expected exception. Tried to close XSet " +
                            "instance that is running job\n" +
                            ox.getLocalizedMessage()); */
               } catch (XAMException ex) {
                    if (!JobStatus.isDone(xset.getString(XSet.XAM_JOB_STATUS))) {
                        fail("Incorrect exception returned when close on " +
                                "XSet with running job: " + 
                                ex.getLocalizedMessage());
                    }
                   announce(
                            "Not exact expected exception when close XSet " +
                            "with a running job. Try re-running when " +
                            "more XSets persisted so job takes longer.\n" +
                            ex.getLocalizedMessage());
                }
            
            }
        }
        
        Thread.sleep(25); // Short delay makes sure job status in XSet is set
        if (!JobStatus.isDone(xset.getString(XSet.XAM_JOB_STATUS))) {
            fail("Iterator should not return end of stream until job completed");
        }
         // job has completed
        long num = xset.getLong(XSet.XAM_JOB_QUERY_RESULTS_COUNT);
       
        assertEquals(num, iterCount);
        applStream.close();
        xset.close();
        system.close();
    } 

    /**
     * Tests retrieving job results while job is running. The XSet used
     * to run the job is committed before the job is submitted. A second
     * XSet instance is instantiated to return the job results. Also attempts
     * to commit the XSet while the job is running - which should fail.
     * 
     * @throws java.lang.Exception
     */
    public void testCommitQueryResults() throws Exception {
        announce( "ReferenceQueryTest.testCommitQueryResults");
        XSystem system = connectAndAuthenticate(true);
        XSet xset = createXSetForQueryJob(system, SELECT_ALL_QUERY);
        // Commit the XSet before running the job - OK
        XUID jobXUID = xset.commit();
        
        // Open another XSet instance used to return job results
        XSet xset2 = system.openXSet(jobXUID, XSet.MODE_UNRESTRICTED);
        // Run job
        xset.submitJob();
        
        // Get results while job is still running
        XStream applStream = xset.openXStream(ReferenceXSet.XAM_JOB_QUERY_RESULTS, 
                                                XStream.MODE_READ_ONLY);
        Iterator<XUID> iter = new XUIDIterator(applStream);

        int iterCount = 0;
        while (iter.hasNext()) {
            XUID xuid = iter.next();
            if (xuid != null) 
                iterCount++;
            if (iterCount == 1) {
                try {
                    // Try committing the XSet while the job is still executing
                    if (!JobStatus.isDone(
                            xset.getString(XSet.XAM_JOB_STATUS))) {
                        xset2.commit();
                        if (!JobStatus.isDone(
                                xset.getString(XSet.XAM_JOB_STATUS))) {
                            fail("Should NOT be able to commit a XSet on " +
                                 "a running a job even if it is a different " +
                                 "XSet instance varb.");
                        }
                    }
                } catch (ObjectInUseException ox) {
                    /* announce(
                        "Expected exception. Tried to commit a running " +
                        "job using a different XSet instance varb\n" +
                        ox.getLocalizedMessage()); */
                } catch (XAMException ex) {
                    String status = xset.getString(XSet.XAM_JOB_STATUS);
                    if (!JobStatus.isDone(status)) {
                        fail("Incorrect exception returned when commit on " +
                             "XSet with running job using diff XSet instance." +
                             " Job Status: " + status + " Exception: "
                              + ex.getLocalizedMessage());
                    }
                   announce(
                            "Not exact expected exception when commit XSet " +
                            "with a running job using different XSet instance. "
                            + " Try re-running when more XSets persisted so " +
                            "job takes longer.\n" + ex.getLocalizedMessage());
                }
            
            }
        }
        
        Thread.sleep(25); // Short delay makes sure job status in XSet is set
        if (!JobStatus.isDone(xset.getString(XSet.XAM_JOB_STATUS))) {
            fail("Iterator should not return end of stream until job completed");
        }
         // job has completed
        long num = xset.getLong(XSet.XAM_JOB_QUERY_RESULTS_COUNT);
       
        assertEquals(num, iterCount);
        applStream.close();
        
        // testing commit after job finishes -  OK
        XUID jobXUID2 = xset.commit();
        xset.close();
        xset2.close();
        assertTrue("Committing job results changed XUID",
                jobXUID.toString().compareTo(jobXUID2.toString()) == 0);
        
        // Reopen and get results from committed XSet
        xset2 = system.openXSet(jobXUID, XSet.MODE_UNRESTRICTED);
        
        applStream = xset2.openXStream(ReferenceXSet.XAM_JOB_QUERY_RESULTS, 
                                                XStream.MODE_READ_ONLY);
        iter = new XUIDIterator(applStream);

        iterCount = 0;
        while (iter.hasNext()) {
            XUID xuid = iter.next();
            if (xuid != null) 
                iterCount++;
        }
        assertEquals(num, iterCount);
        applStream.close();
        xset2.close();
        
        system.close();
    } 
       
    /**
     * Tests abandoning an XSet that is being used to run a job. The job should
     * stop running when XSet.abandon() is called.
     * @throws java.lang.Exception
     */
    public void testAbandonQuery() throws Exception {
        announce("ReferenceQueryTest.testAbandonQuery");
        XSystem system = connectAndAuthenticate(true);
        XSet xset = createXSetForQueryJob(system, SELECT_ALL_QUERY);
        xset.submitJob();
        // Abandon should stop the running job
        xset.abandon();
        xset.close();               
        system.close();
    }
    /**
     * Tests resubmitting the job after it has been committed.  The result set
     * should increase by one due to the first commit.
     * 
     * @throws java.lang.Exception
     */
    public void testResubmitCommittedJob() throws Exception {
       announce( "ReferenceQueryTest.testResubmitCommittedJob");
        XSystem system = connectAndAuthenticate(true);
        XSet xset = createXSetForQueryJob(system, SELECT_ALL_QUERY);
        // Run job
        xset.submitJob();
        do {
            // after sleeping for one second, check if job has completed
            Thread.sleep(XAMConstants.MILLISEC_PER_SECOND);
        } while (!JobStatus.isDone(xset.getString(XSet.XAM_JOB_STATUS)));
        
        // Get results now that job has finished
        long num = xset.getLong(XSet.XAM_JOB_QUERY_RESULTS_COUNT);
        
        // Commit the XSet after running the job
        xset.createProperty(
                ReferenceJob.JOB_DELAY_FIELD_IN_MILLISEC, false, 2000);
        XUID jobXUID = xset.commit();
        
        // Open another XSet instance used to return job results
        XSet xset2 = system.openXSet(jobXUID, XSet.MODE_UNRESTRICTED);
        int numFields = checkFieldList(xset2, -1, "First run of job.");
        
        // Resubmit job
        xset2.submitJob();
        boolean firstPass = true;
        do {
            if (firstPass) {
                // Hopefully catch it when the values have been reset
                long num2 = xset2.getLong(XSet.XAM_JOB_QUERY_RESULTS_COUNT);
                assertTrue("The job results were not reset to zero when " +
                        "resubmitting a job", num2 == 0);
                firstPass = false;
            }
            // after sleeping for one second, check if job has completed
            Thread.sleep(XAMConstants.MILLISEC_PER_SECOND);
        } while (!JobStatus.isDone(xset2.getString(XSet.XAM_JOB_STATUS)));
        
        // Get results now that job has finished
        long num2 = xset2.getLong(XSet.XAM_JOB_QUERY_RESULTS_COUNT);
        
        assertTrue("Incorrect number of results returned for re-submitting job." +
                " Expected: " + (num + 1) + "  Actual: " + num2, 
                num + 1 == num2);
        
        xset2.commit();
        int numFields2 = this.checkFieldList(xset, numFields, 
                "Invalid number of fields in second run of job.");
        
        assertTrue("Invalid number of fields in second run of job. " +
                "Expected: " + numFields + "  Actual: " + numFields2,
                numFields == numFields2);
        
        // testing commit after resubmitted job finishes
        xset.close();
        xset2.close();
        system.close();
    }
    /**
     * Tests modifying the query command stream field to read-only while the 
     * job runs and then changing it back to read/write upon job completion.  
     * 
     * @throws java.lang.Exception
     */
    public void testModifyQueryCommand() throws Exception {
        announce("ReferenceQueryTest.testModifyQueryCommand");
        XSystem system = connectAndAuthenticate(true);
 
        // non-binding
        XSet xset = createXSetForQueryJob(system, SELECT_ALL_QUERY);
        xset.createProperty(
                ReferenceJob.JOB_DELAY_FIELD_IN_MILLISEC, false, 500 );
        xset.submitJob();

        // wait until job has started running
        int numTries = 0;
        String jobStatusString = xset.getString(XSet.XAM_JOB_STATUS);
        while (numTries < 250 &&
               !JobStatus.isDoneExecution(jobStatusString) &&
               !JobStatus.isRunning(jobStatusString)) {
            Thread.sleep(1);
            numTries++;
            jobStatusString = xset.getString(XSet.XAM_JOB_STATUS);
        }
        
        if (!JobStatus.isDone(xset.getString(XSet.XAM_JOB_STATUS))) {
            XStream roCmdStream = null;
            XStream cmdStream = null;
            try {
                // openXStream should succeed since it's read-only
                roCmdStream = xset.openXStream(ReferenceXSet.XAM_JOB_QUERY_COMMAND,
                                                        XStream.MODE_READ_ONLY);
            } catch (XAMException xame) {
                fail("Should be able to open the query command stream field" +
                    " in READ-ONLY mode when a job is running.");
            } finally {
                if (roCmdStream != null)
                    roCmdStream.close();
                roCmdStream = null;
            }
            try {
                // openXStream should fail since it is read-only
                cmdStream = xset.openXStream(ReferenceXSet.XAM_JOB_QUERY_COMMAND,
                                                    XStream.MODE_WRITE_TRUNCATE);
                if (JobStatus.isDone(xset.getString(XSet.XAM_JOB_STATUS)))
                    announce("Job finished too quickly and the " +
                    "query command stream field is no longer read-only.  Create" +
                            " more XSets and retry the test.");
                else
                {
                   String stat = xset.getString(XSet.XAM_JOB_STATUS);
                   fail("Test failed: Opening an XStream on a read only" +
                   "field succeeded. " + " " + stat );
                }
            } catch (XAMException e) {
                /* announce( "Expected exception.\n" + e.getLocalizedMessage()); */
            } finally {
                if (cmdStream != null)
                    cmdStream.close();
                cmdStream = null;
            }
        } else {
            fail("Job didn't run long enough to test job field" +
                    "access during the running of the job. Increase delay");
        }
        
        // try and open query command stream with write-append access
        // once the job has finished
        XStream queryStream = null;
        do {
            // after sleeping for one second, check if job has completed
            Thread.sleep(XAMConstants.MILLISEC_PER_SECOND);
        } while (!JobStatus.isDone(xset.getString(XSet.XAM_JOB_STATUS)));
        
        // job has completed so test...
        try {
            queryStream = xset.openXStream(
                 ReferenceXSet.XAM_JOB_QUERY_COMMAND, XStream.MODE_READ_ONLY);
            byte[] buffer = new byte[BUF_SIZE];
            long numOfBytesRead = queryStream.read(buffer);
            if (numOfBytesRead == XAMConstants.END_OF_STREAM) {
                throw new XAMException("Query command stream is empty -- end of stream.");
            }
            while (numOfBytesRead < buffer.length) {
                long bytes_read = queryStream.read(buffer, numOfBytesRead, 
                                            buffer.length - numOfBytesRead);
                if (bytes_read == XAMConstants.END_OF_STREAM) {
                    break;
                }
                numOfBytesRead += bytes_read;
            }
            String queryString = new String(buffer).trim();
            assertEquals(queryString, SELECT_ALL_QUERY);
            int size = queryString.getBytes().length;
            queryStream.close();
            
            // modify the XAM_JOB_QUERY_COMMAND
            queryStream = xset.openXStream(
                 ReferenceXSet.XAM_JOB_QUERY_COMMAND, XStream.MODE_WRITE_TRUNCATE);
            long numBytesWritten = queryStream.write(
                                            queryString.getBytes(), 0, size);
            assertTrue("Number of bytes written does not match number sent",
                                                    numBytesWritten == size);
            queryStream.close();
            XUID jobXUID = xset.commit();
        } catch (XAMException xe) {
            xe.printStackTrace();
            fail("Should be able to open the query command stream field" +
                    " in any mode now that the job has completed.");
        }
        xset.close();
        system.close();
    }
    
    /**
     * Tests that the job fields are read/write when job is not running
     * but are set to readonly when job is running.
     * 
     * @throws java.lang.Exception
     */
    public void testQueryJobFieldsReadonly() throws Exception {
        announce( "ReferenceQueryTest.testQueryJobFieldsReadonly");
        XSystem system = connectAndAuthenticate(true);
        XSet xset = createXSetForQueryJob(system, SELECT_ALL_QUERY);
        // introduce a 2 sec. delay to ensure the job runs long enough to
        // try and delete the corresponding XSet
        xset.createProperty(ReferenceJob.JOB_DELAY_FIELD_IN_MILLISEC, false, 500 );

        // Commit the XSet before running the job - OK
        XUID jobXUID = xset.commit();
        
        // Job not running verify that all job fields are read/write
        String[] jobFields = {      XSet.XAM_JOB_COMMAND,
                                    XSet.XAM_JOB_ERROR,
                                    XSet.XAM_JOB_ERRORHEALTH,
                                    XSet.XAM_JOB_STATUS,
                                    XSet.XAM_JOB_QUERY_COMMAND,
                                    XSet.XAM_JOB_QUERY_RESULTS,
                                    XSet.XAM_JOB_QUERY_RESULTS_COUNT,
                                    XSet.XAM_JOB_QUERY_LEVEL};
        for (String fieldName : jobFields) {
            if (xset.containsField(fieldName)) {
                assertTrue("Test Failed: job is NOT running, job fields " +
                        "should be read/write. Field Name: " + fieldName,
                xset.getFieldReadOnly(fieldName) == false);
                
            }
        }
        
        // Run job
        xset.submitJob();
        
        // wait until job has started running
        int numTries = 0;
        String jobStatusString = xset.getString(XSet.XAM_JOB_STATUS);
        while (numTries < 250 &&
               !JobStatus.isDoneExecution(jobStatusString) &&
               !JobStatus.isRunning(jobStatusString)) {
            Thread.sleep(5);
            numTries++;
            jobStatusString = xset.getString(XSet.XAM_JOB_STATUS);
        }
        
        // Get results while job is still running
        for (String fieldName : jobFields) {
            if (!JobStatus.isDoneExecution(xset.getString(XSet.XAM_JOB_STATUS))) {
                if (xset.containsField(fieldName)) {
                    if (xset.getFieldReadOnly(fieldName) != true) {
                        // Double check job is still running
                        if (!JobStatus.isDoneExecution(
                                xset.getString(XSet.XAM_JOB_STATUS))) {
                            fail("Test Failed: job is running, job fields " +
                            "should be readonly. Field Name: " + fieldName);
                        } else {
                            announce("Job completed (retry) can not test for "
                             + "readonly field: " + fieldName);
                        }
                    }
                }

            } else {
                    announce("Job completed can not test for " +
                            "readonly field: " + fieldName);
                   break;
            }
        }   // end for
        
        // Wait for job to finish
        while (!JobStatus.isDone(xset.getString(XSet.XAM_JOB_STATUS))) {
            Thread.sleep(5);
        }

         // Verify job fields back to read/write
         for (String fieldName : jobFields) {
            if (xset.containsField(fieldName)) {
                assertTrue("Test Failed: job is NOT running, job fields " +
                        "should be read/write. Field Name: " + fieldName,
                xset.getFieldReadOnly(fieldName) == false);
                
            }
        }
        XUID jobXUID2 = xset.commit();
        xset.close();
        
        // Reopen verify fields again
        XSet xset2 = system.openXSet(jobXUID, XSet.MODE_UNRESTRICTED);
         // Verify job fields back to read/write
         for (String fieldName : jobFields) {
            if (xset2.containsField(fieldName)) {
                assertTrue("Test Failed: job is NOT running, job fields " +
                        "should be read/write. Field Name: " + fieldName,
                xset2.getFieldReadOnly(fieldName) == false);
                
            }
        }
        
        xset2.close();
        system.close();
    } 

    private XSet createXSetForQueryJob(XSystem system, 
                                        boolean binding, 
                                        String xamql) throws XAMException {
        XSet xset = null;
        if (system != null && xamql != null) {
            xset = system.createXSet(XSet.MODE_UNRESTRICTED);
            xset.createProperty(XSet.XAM_JOB_COMMAND, binding, XSet.XAM_JOB_QUERY);
            XStream stream = xset.createXStream(XSet.XAM_JOB_QUERY_COMMAND,
                                                       binding, "text/plain;charset=utf-8");
            byte[] bytes = xamql.getBytes();
            int wrote = 0;
            while (wrote < bytes.length) {
                wrote += stream.write(bytes, wrote, bytes.length - wrote);
            }

            stream.close();
        }
        return xset;
    }
    
    private XSet createXSetForQueryJob(XSystem system, String xamql) 
                                                        throws XAMException {
        return createXSetForQueryJob(system, false, xamql);
    }
    
    /**
     * Verifies the number of fields in an XSet. Check for duplicate field
     * names also.
     * 
     * @param xset
     * @param numExpectedFields if -1 then don't check the number of fields
     * @param errMsgPrefix
     * @return the number of fields found
     * @throws java.lang.Exception
     */
    private int checkFieldList(XSet xset, int numExpectedFields, 
            String errMsgPrefix) throws Exception {
        XIterator iterator = xset.openFieldIterator("");
        int numFields = 0;
        HashMap<String, String> fieldList = new HashMap<String, String>();
        while (iterator.hasNext()) {
            numFields++;
            String fieldName = (String)iterator.next();
            if (fieldList.containsKey(fieldName)) {
                fail(errMsgPrefix + 
                        " Field exists in committed job more than once: " +
                        fieldName);
            }
            fieldList.put(fieldName, fieldName);
        }   // end while
        
        if (numExpectedFields != -1) {
            assertTrue(errMsgPrefix + "Expected: " + numExpectedFields + 
                    "  Actual: " + numFields, numExpectedFields == numFields);
        }
        return numFields;
        

    }
}
