/*
 *  Copyright (c) 2008, Sun Microsystems, Inc.
 *  Copyright (c) 2008, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.util.logging.Level;
import java.util.logging.Logger;
import junit.framework.Test;
import junit.framework.TestSuite;

import org.snia.xam.RetentionValueException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XSet;

import org.snia.xam.XSystem;
import org.snia.xam.XUID;

/**
 * Retention Time Criteria (RTC) tests
 * 
 */
public class ReferenceRetentionMgmtTest extends ReferenceXAMTest {

    protected static final int MIN_IN_MS  = 60 * 1000;
    protected static final int HOUR_IN_MS = 60 * MIN_IN_MS;
    protected static final String[] RTC_REQ_PROPS = {
        XSet.XAM_RETENTION_LIST_BASE,
        XSet.XAM_RETENTION_BASE_ENABLED,
        XSet.XAM_RETENTION_LIST_EVENT,
        XSet.XAM_RETENTION_BASE_STARTTIME,
        XSet.XAM_MANAGEMENT_POLICY
    };
    protected static final String[] RTC_REQ_TYPES = {
        XAMLibrary.STYPE_STRING_MIME_TYPE, /* list base */
        XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, /* base enabled */
        XAMLibrary.STYPE_STRING_MIME_TYPE, /* list event */
        XAMLibrary.STYPE_DATETIME_MIME_TYPE, /* base starttime */
        XAMLibrary.STYPE_STRING_MIME_TYPE /* principal mgmt policy */        
    };

    public ReferenceRetentionMgmtTest(String name) {
        super(name);
    }

    public static void main(String[] args) {
        System.exit(runTest(args, suite()));
    }

    public static Test suite() {
        TestSuite suite = new TestSuite();
        suite.addTest(new ReferenceRetentionMgmtTest("testProperties"));
        suite.addTest(new ReferenceRetentionMgmtTest("testRetentionIDValidation"));
        suite.addTest(new ReferenceRetentionMgmtTest("testDefaultPolicy"));
        suite.addTest(new ReferenceRetentionMgmtTest("testCreate"));
        suite.addTest(new ReferenceRetentionMgmtTest("testEnableRetention"));
        suite.addTest(new ReferenceRetentionMgmtTest("testRetentionDuration"));
        suite.addTest(new ReferenceRetentionMgmtTest("testRetentionStarttime"));
        suite.addTest(new ReferenceRetentionMgmtTest("testIsXSetRetained"));
        suite.addTest(new ReferenceRetentionMgmtTest("testRetentionPolicy"));
        suite.addTest(new ReferenceRetentionMgmtTest("testRetentionPolicyOverrides"));
        suite.addTest(new ReferenceRetentionMgmtTest("testMultipleRetentionPolicy"));
        suite.addTest(new ReferenceRetentionMgmtTest("testBaseRetentionPolicy"));
        return suite;
    }
    
    public void testRetentionIDValidation() throws Exception {
        System.out.println("ReferenceRetentionMgmtTest.testRetentionIDValidation");
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        // Checking that reserved retention criteria identifier,
        // "base", is protected and can't be valid input params
        try {
            xset.createRetention(false, RetentionMgr.BASE);
        } catch (XAMException xe) {
            // System.out.println("Expected exception: " + xe.getMessage());
        }
        // this should be allowed because the XSet hasn't been committed
        // yet -- only upon commit, will the "event" identifier be created
        // automatically
        xset.createRetention(false, RetentionMgr.EVENT);
        try {
            xset.createRetention(false, "");
        } catch (XAMException xe) {
            // System.out.println("Expected exception: " + xe.getMessage());
        }
        xset.commit();
        try {
            // this should result in error since "event" has been created
            // at the time of commit
            xset.createRetention(false, RetentionMgr.EVENT);
        } catch (XAMException xe) {
            // System.out.println("Expected exception: " + xe.getMessage());
        }
        xset.close();
        system.close();
    }
    public void testDefaultPolicy() throws Exception {
        announce("ReferenceRetentionMgmtTest.testDefaultPolicy");
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        // no retention criteria is set so the XSet should contain the required
        // retention fields with the default policy values
        xset.commit();
        checkProperties(xset, RTC_REQ_PROPS, RTC_REQ_TYPES);
        xset.close();
        system.close();
    }
    public void testCreate() throws Exception {
        announce("ReferenceRetentionMgmtTest.testCreate");
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);

        // Checking to make sure that if a retention ID already exists, the 
        // same ID cannot be inputted again
        xset.createRetention(false, "testCreate");
        XUID xuid = xset.commit();
        assertFalse("The XSet, " + xuid.toString() + "is not retained.",
                system.isXSetRetained(xuid));
        try {
            xset.createRetention(false, "testCreate");
        } catch (XAMException xe) {
            // System.out.println("Expected exception: " + xe.getMessage());
        }
        // check and make sure that if binding retention criteria are created,
        // a new XUID is generated
        xset.createRetention(true, "testBindingCreate");
        XUID newXUID = xset.commit();
        assertFalse("XUID should change, added binding retention criteria: " +
                "xuid: " + xuid.toString() + "  newXUID: " + newXUID.toString(),
                xuid.equals(newXUID));
        xset.close();
        system.close();
    }
    public void testEnableRetention() throws Exception {
        announce("ReferenceRetentionMgmtTest.testEnableRetention");
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        retentionEnabledWork(xset, false);
        retentionEnabledWork(xset, true);
        xset.close();
        system.close();
    }
    public void testRetentionDuration() throws Exception {
        announce("ReferenceRetentionMgmtTest.testRetentionDuration");
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        retentionDurationWork(xset, false);
        retentionDurationWork(xset, true);
        xset.close();
        system.close();
    }
    public void testRetentionStarttime() throws Exception {
        announce("ReferenceRetentionMgmtTest.testRetentionStarttime");
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        retentionStarttimeWork(xset, system, false);
        retentionStarttimeWork(xset, system, true);
        xset.close();
        system.close();
    }

    /**
     * Basic test for properties before and after an XSet has been retained
     * 
     * @throws java.lang.Exception
     */
    public void testProperties() throws Exception {
        announce("ReferenceRetentionMgmtTest.testProperties");
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        xset.createRetention(true, "testProperties");
        XUID xuid = xset.commit();
        checkProperties(xset, RTC_REQ_PROPS, RTC_REQ_TYPES);
        xset.close();
        system.close();
    }
    public void testIsXSetRetained() throws Exception {
        announce("ReferenceRetentionMgmtTest.testIsXSetRetained...");
        System.out.print("...this will take ~1 minute to finsh -- please wait...");
        XSystem system = connectAndAuthenticate(true);
        // test XUID is null
        try {
            system.isXSetRetained(null);
        } catch (XAMException e) {
             // System.out.println("Expected exception: " + e.getMessage());
        }
        // test retention criteria where base is set ONLY -- since base retention
        // time criteria will carry over from the nonbinding to binding test,
        // the XSet must be newly created prior to testing binding
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        xset.commit();
        baseCriteriaTest(system, xset, false);
        xset.close();
        
        xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        xset.commit();
        baseCriteriaTest(system, xset, true);
        xset.close();
        
        // test retention criteria where other retention is set ONLY (no base)
        xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        xset.commit();
        otherCriteriaTest(system, xset, false);
        xset.close();

        xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        xset.commit();
        otherCriteriaTest(system, xset, true);
        xset.close();
        
        // test retention criteria where base AND other retention are set
        xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        xset.commit();
        // nonbinding
        baseAndOtherCriteriaTest(system, xset, false);
        xset.close();
        
        // binding
        xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        xset.commit();
        baseAndOtherCriteriaTest(system, xset, true);
        xset.close();
        system.close();
        System.out.println("DONE");
    }
    private void baseCriteriaTest(XSystem system, XSet set, boolean binding) 
        throws XAMException {
        // test base retention criteria
        String testID = "test_" + (binding ? "binding" : "nonbinding");
        XUID initXUID = set.getXUID(XSet.XAM_XUID);
        
        // using system default policy value for duration -- base retention
        // management values enabled and starttime are set when the XSet is
        // first committed.  Since the system default policy's duration value 
        // is set to zero, this XSet can't be retained
        assertEquals(system.isXSetRetained(initXUID), false);
        
        // setting the duration value to a positive integer to test base
        // retention criteria
        set.setBaseRetention(binding, 5000);
        XUID xuid = set.commit();
        assertEquals(system.isXSetRetained(xuid), true);
        if (binding) {
            assertFalse("XUID should change, added binding retention criteria: " +
             "old XUID: " + initXUID.toString() + "  new XUID: " + xuid.toString(),
             initXUID.equals(xuid));
        }
        try {
            Thread.sleep(7000);
        } catch (InterruptedException ex) {
            // System.out.println("Interrupted Exception thrown:  " + ex.getMessage());
        }
        assertEquals(system.isXSetRetained(xuid), false);
        
        // set duration value to a greater positive integer so the base 
        // retention criteria indicates this XSet is retained while other
        // retention criteria is also not met (i.e. retained)
        set.setBaseRetention(binding, 10000);
        XUID xuid2 = set.commit();
        assertEquals(system.isXSetRetained(xuid2), true);
        if (binding) {
            assertFalse("XUID should change, added binding retention criteria: " +
             "old XUID: " + xuid.toString() + "  new XUID: " + xuid2.toString(),
             xuid.equals(xuid2));
        }
    }
    private void otherCriteriaTest(XSystem system, XSet set, boolean binding) 
        throws XAMException {
        // test other retention criteria
        String retID = "test_" + (binding ? "binding" : "nonbinding");
        
        XUID initXUID = set.getXUID(XSet.XAM_XUID);
        set.createRetention(binding, retID);
        XUID xuid = set.commit();
        if (binding) {
            assertFalse("XUID should change, added binding retention criteria: " +
             "old XUID: " + initXUID.toString() + "  new XUID: " + xuid.toString(),
             initXUID.equals(xuid));
        }
        
        // using system default policy values for enabled and duration --
        // starttime hasn't been set yet for retention identifier (i.e. retID)
        // so this XSet can't possibly be retained
        assertEquals(system.isXSetRetained(xuid), false);
        
        // using system default policy values for enabled and duration --
        // setting starttime for retID but XSet is still not retained because
        // the system default policy enable is false
        try { set.setRetentionStarttime(retID, binding); }
        catch( XAMException xe ) { /*expected*/ }
        XUID xuid2 = set.commit();
        assertEquals(system.isXSetRetained(xuid2), false);
        
        // using system default policy values (false) for enabled --
        // setting duration value greater than zero (now < (start + duration))
        // so XSet is not retained
        set.setRetentionDuration(retID, binding, 3000);
        XUID xuid3 = set.commit();
        assertEquals(system.isXSetRetained(xuid3), false);
        if (binding) {
            assertFalse("XUID should change, added binding retention criteria: " +
             "old XUID: " + xuid2.toString() + "  new XUID: " + xuid3.toString(),
             xuid2.equals(xuid3));
        }
        try {
            Thread.sleep(4000);
        } catch (InterruptedException ex) {
            // System.out.println("Interrupted Exception thrown:  " + ex.getMessage());
        }
        assertEquals(system.isXSetRetained(xuid3), false);
        
        // no longer using system default policy values --
        // setting enabled retention management value to false so this XSet
        // cannot be retained until enabled is set to true for retID
        set.setRetentionEnabledFlag(retID, binding, false);
        XUID xuid4 = set.commit();
        assertEquals(system.isXSetRetained(xuid4), false);
        if (binding) {
            assertFalse("XUID should change, added binding retention criteria: " +
             "old XUID: " + xuid3.toString() + "  new XUID: " + xuid4.toString(),
             xuid3.equals(xuid4));
        }
        
        // setting the enabled retention mgmt value to true and increasing
        // the duration so this will be retained for ~15 seconds
        set.setRetentionEnabledFlag(retID, binding, true);
        set.setRetentionDuration(retID, binding, 8000);
        XUID xuid5 = set.commit();
        assertEquals(system.isXSetRetained(xuid5), true);
        if (binding) {
            assertFalse("XUID should change, added binding retention criteria: " +
             "old XUID: " + xuid4.toString() + "  new XUID: " + xuid5.toString(),
             xuid4.equals(xuid5));
        }
        try {
            Thread.sleep(5000);
        } catch (InterruptedException ex) {
            // System.out.println("Interrupted Exception thrown:  " + ex.getMessage());
        }
    }
    private void baseAndOtherCriteriaTest(XSystem system, XSet set, boolean binding) 
        throws XAMException {
        // test other retention criteria
        String retID = "test_" + (binding ? "binding" : "nonbinding");        
        baseCriteriaTest(system, set, binding);

        XUID initXUID = set.getXUID(XSet.XAM_XUID);
        set.createRetention(binding, retID);
        XUID xuid = set.commit();
        if (binding) {
            assertFalse("XUID should change, added binding retention criteria: " +
             "old XUID: " + initXUID.toString() + "  new XUID: " + xuid.toString(),
             initXUID.equals(xuid));
        }
        
        // although the management starttime value property hasn't been set for
        // retID, the XSet is still retained because of base
        assertEquals(system.isXSetRetained(xuid), true);
        set.setRetentionEnabledFlag(retID, false, true);
        set.setRetentionStarttime(retID, binding);
        set.setRetentionDuration(retID, binding, 5000);
        XUID xuid2 = set.commit();
        assertEquals(system.isXSetRetained(xuid2), true);
        if (binding) {
            assertFalse("XUID should change, added binding retention criteria: " +
             "old XUID: " + xuid.toString() + "  new XUID: " + xuid2.toString(),
             xuid.equals(xuid2));
        }
        try {
            Thread.sleep(8000);
        } catch (InterruptedException ex) {
            // System.out.println("Interrupted Exception thrown:  " + ex.getMessage());
        }
        // base retention and other time criteria should have been met
        assertEquals(system.isXSetRetained(xuid2), false);
    }
    private void retentionEnabledWork(XSet set, boolean binding) 
        throws XAMException {
        String retID = "test_" + (binding ? "binding" : "nonbinding");
        set.createRetention(binding, retID);
        
        set.setRetentionEnabledFlag(retID, binding, false);
        XUID xuid = set.commit();
        set.setRetentionEnabledFlag(retID, binding, true);
        XUID xuid2 = set.commit();
        try {
            // should not be able to disable retention (i.e. false) once
            // it is enabled (i.e. true)
            set.setRetentionEnabledFlag(retID, binding, false);
        } catch (XAMException xe) {
            // System.out.println("Expected exception: " + xe.getMessage());
        }
        if (binding) {
            assertFalse("XUID should change, added binding retention criteria: " +
             "old XUID: " + xuid.toString() + "  new XUID: " + xuid2.toString(),
             xuid.equals(xuid2));
        }
        // test where retention ID is missing
        try {
            set.setRetentionEnabledFlag("nonexistentRetentionID", binding, false);
        } catch (XAMException xe) {
            // System.out.println("Expected exception: " + xe.getMessage());
        }
    }
    private void retentionStarttimeWork(XSet set, XSystem sys, boolean binding) 
        throws XAMException {
        String retID = "test_" + (binding ? "binding" : "nonbinding");
        set.createRetention(binding, retID);
        
        // set to now plus an hour
        long initDuration = System.currentTimeMillis() + HOUR_IN_MS;
        
        // test where retention ID is missing
        try {
            set.setRetentionStarttime("nonexistent", binding);
        } catch (XAMException xe) {
            // System.out.println("Expected exception: " + xe.getMessage());
        }
        // test using valid policy values for duration and enablement -- 
        // duration = 0 and enabled = FALSE
        set.setRetentionEnabledFlag(retID, false, true);
        set.setRetentionStarttime(retID, binding);
        XUID xuid = set.commit();
        // unable to change the value of the starttime once it is set
        try {
            set.setRetentionStarttime(retID, binding);
        } catch (XAMException xe) {
            // System.out.println("Expected exception: " + xe.getMessage());
        }
        
        // need to create new XSet in order to test alternate
        // scenarios of setting the starttime
        XSet set2 = sys.createXSet(XSet.MODE_UNRESTRICTED);
        String newRetID = retID + "_2";
        set2.createRetention(binding, newRetID);
        // test where the value property enabled flag is set to false and
        // still using the policy's duration value of zero -- the starttime
        // should not successfully be set since retention is not enabled
        set2.setRetentionEnabledFlag(newRetID, binding, false);
        xuid = set2.commit();
        try {
            set2.setRetentionStarttime(newRetID, binding);
         } catch (XAMException xe) {
            // System.out.println("Expected exception: " + xe.getMessage());
        }
        // test where enabled flag is set to true and still refering to
        // the policy's duration value of zero -- starttime should be set
        set2.setRetentionEnabledFlag(newRetID, binding, true);
        set2.setRetentionStarttime(newRetID, binding);       
        XUID xuid2 = set2.commit();
        if (binding) {
            assertFalse("XUID should change, added binding retention criteria: " +
             "old XUID: " + xuid.toString() + "  new XUID: " + xuid2.toString(),
             xuid.equals(xuid2));
        }
        set2.close();
    }
    private void retentionDurationWork(XSet set, boolean binding) 
        throws XAMException {
        String retID = "test_" + (binding ? "binding" : "nonbinding");
        set.createRetention(binding, retID);
        
        // set to now plus an hour
        long initDuration = System.currentTimeMillis() + HOUR_IN_MS;
        
        // test where retention ID is missing
        try {
            set.setRetentionDuration("nonexistent", binding, initDuration);
        } catch (XAMException xe) {
            // System.out.println("Expected exception: " + xe.getMessage());
        }
        // test where enabled flag is missing
        try {
            set.setRetentionDuration(retID, binding, initDuration);
        } catch (XAMException xe) {
            // System.out.println("Expected exception: " + xe.getMessage());
        }
        // test where enabled flag is present
        set.setRetentionEnabledFlag(retID, binding, false);
        XUID xuid = set.commit();
        // test where duration is not there but is then created now that the
        // enabled flag has been created
        set.setRetentionDuration(retID, binding, initDuration);
        // test setting new duration to a negative value other than -1
        try {
            set.setRetentionDuration(retID, binding, -2);
        } catch (XAMException xe) {
            // System.out.println("Expected exception: " + xe.getMessage());
        }
        // test setting new duration to a value less than old duration
        try {
            long oneLessMillis = initDuration - 1;
            set.setRetentionDuration(retID, binding, oneLessMillis);
        } catch (XAMException xe) {
            // System.out.println("Expected exception: " + xe.getMessage());
        }
        // test setting new duration to a value greater than old duration     
        try {
            long oneMoreMillis = initDuration + 1;
            set.setRetentionDuration(retID, binding, oneMoreMillis);
        } catch (XAMException xe) {
            // System.out.println("Expected exception: " + xe.getMessage());
        }
        // test setting new duration to "forever" (i.e. -1)
        try {
            set.setRetentionDuration(retID, binding, -1);
        } catch (XAMException xe) {
            // System.out.println("Expected exception: " + xe.getMessage());
        }
        // test setting new duration to something other than "forever"
        try {
            set.setRetentionDuration(retID, binding, System.currentTimeMillis());
        } catch (XAMException xe) {
            // System.out.println("Expected exception: " + xe.getMessage());
        }
        XUID xuid2 = set.commit();
        if (binding) {
            assertFalse("XUID should change, added binding retention criteria: " +
             "old XUID: " + xuid.toString() + "  new XUID: " + xuid2.toString(),
             xuid.equals(xuid2));
        }
    }
    
    /**
     * Tests the policy retention methods.
     * @throws Exception
     */
    public void testRetentionPolicy() throws Exception
    {
       System.out.println("ReferenceRetentionMgmtTest.testRetentionPolicy...");
       final String policyName = RetentionMgr.POLICY_ONE_SECONDS;
       final String id = "unit_test.policyRetention";
       final String prefix = ".xset.retention.";
       
       // Required properties for retention
       final String RETENTION_POLICY_PROPS[] = {
          prefix + id + ".enabled.policy",
          prefix + id + ".duration.policy"
       };
       
       final String RETENTION_POLICY_TYPES[] = {
          XAMLibrary.STYPE_STRING_MIME_TYPE,
          XAMLibrary.STYPE_STRING_MIME_TYPE
       };
       
       XSystem system = connectAndAuthenticate(true);
       XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
       
       xset.createRetention( false, id );
       xset.applyRetentionDurationPolicy( id, false, policyName );
       xset.applyRetentionEnabledPolicy(  id, false, policyName );
       
       XUID xuid = xset.commit();
       checkProperties(xset, RETENTION_POLICY_PROPS, RETENTION_POLICY_TYPES);
       
       assertEquals( "Expected duration to be one second", 1000, xset.getActualRetentionDuration(id));
       assertEquals( "Expected enabled to be true", true, xset.getActualRetentionEnabled(id));
       
       assertTrue( "Expected XSet to be under retention", system.isXSetRetained(xuid));
       xset.setRetentionStarttime(id, false);
       xset.commit();
       
       Thread.sleep(1500);
       
       // At this point, the period should be expired
       assertFalse( "Expected XSet to be released from retention", system.isXSetRetained(xuid));
       
       xset.close();
       system.close();
    }

    /**
     * Tests the policy retention methods.
     * @throws Exception
     */
    public void testRetentionPolicyOverrides() throws Exception
    {
       System.out.println("ReferenceRetentionMgmtTest.testRetentionPolicyOverrides...");
       final String policyName = RetentionMgr.POLICY_ONE_SECONDS;
       final String nonePolicy = RetentionMgr.POLICY_NONE;
       final String id = "unit_test.policyRetentionOverride";
       final String prefix = ".xset.retention.";
       
       // Required properties for retention
       final String RETENTION_POLICY_PROPS[] = {
          prefix + id + ".enabled.policy",
          prefix + id + ".duration.policy"
       };
       
       final String RETENTION_POLICY_TYPES[] = {
          XAMLibrary.STYPE_STRING_MIME_TYPE,
          XAMLibrary.STYPE_STRING_MIME_TYPE
       };
       
       XSystem system = connectAndAuthenticate(true);
       XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
       
       xset.createRetention( false, id );
       xset.applyRetentionDurationPolicy( id, false, policyName );
       xset.applyRetentionEnabledPolicy(  id, false, policyName );
       
       XUID xuid = xset.commit();
       checkProperties(xset, RETENTION_POLICY_PROPS, RETENTION_POLICY_TYPES);
       
       assertTrue( "Expected XSet to be under retention", system.isXSetRetained(xuid));
       xset.setRetentionStarttime(id, false);
       xset.commit();
       
       Thread.sleep(1500);
       
       // At this point, the period should be expired
       assertFalse( "Expected XSet to be released from retention", system.isXSetRetained(xuid));
       
       // Check to make sure that reducing the duration is rejected
       try { 
          xset.setRetentionDuration(id, false, 100);
          fail( "Should have received an exception attempting to reduce retention duration" );
       }
       catch( RetentionValueException rve ) { /*expected*/ }
       
       // Should be allowed to increase the time
       xset.setRetentionDuration( id, false, 2000 );
       
       // Check to make sure that retention may not be disabled.
       try { 
          xset.setRetentionEnabledFlag( id, false, false );
          fail( "Should have received an exception attempting to disable a retention preiod" );
       }
       catch( RetentionValueException rve2 ) { /* expected */ }
       xset.close();
       
       // Make sure that XSets with long policies can not take a policy with shorter.
       xset = system.createXSet( XSet.MODE_UNRESTRICTED );
       xset.createRetention(false, id);
       xset.applyRetentionDurationPolicy( id, false, policyName );
       xset.applyRetentionEnabledPolicy(  id, false, policyName );
       xset.setAutoDelete(false, false); // so it will go away, someday
       xset.commit();
       
       try {
          xset.applyRetentionDurationPolicy( id, false, nonePolicy );
          fail( "Expected an exception putting in a shorter policy (1second changed to none" );
       } catch( RetentionValueException rtve3 ) { /*expected */ }
       
       try {
          xset.applyRetentionEnabledPolicy(  id, false, nonePolicy );
          fail( "Expected an exception putting in a disabled policy (1second changed to none" );
       } catch( RetentionValueException rtve4 ) { /*expected */ }
       xset.commit();
       
       // Policies may never violate existing criteria.
       xset.setRetentionDuration(id, false, 1001);
       xset.commit();
       try {
          xset.applyRetentionDurationPolicy( id, false, nonePolicy );
          fail( "Expected an exception putting in a shorter policy (1second changed to none" );
       } catch( RetentionValueException rtve5 ) { /*expected */ }

       xset.close();
       system.close();
    }

    /**
     * Tests the policy retention methods, with multiple policies.
     * @throws Exception
     */
    public void testMultipleRetentionPolicy() throws Exception
    {
       System.out.println("ReferenceRetentionMgmtTest.testMultipleRetentionPolicy...");
       final String policyName1 = RetentionMgr.POLICY_ONE_DAY;
       final String policyName2 = RetentionMgr.POLICY_ONE_SECONDS;
       final String id1 = "unit_test.multiplePolicyRetention.1";
       final String id2 = "unit_test.multiplePolicyRetention.2";
       final String prefix = ".xset.retention.";
       
       // Required properties for retention
       final String RETENTION_POLICY_PROPS[] = {
          prefix + id1 + ".enabled.policy",
          prefix + id1 + ".duration.policy",
          prefix + id2 + ".enabled.policy",
          prefix + id2 + ".duration.policy"
       };
       
       final String RETENTION_POLICY_TYPES[] = {
          XAMLibrary.STYPE_STRING_MIME_TYPE,
          XAMLibrary.STYPE_STRING_MIME_TYPE,
          XAMLibrary.STYPE_STRING_MIME_TYPE,
          XAMLibrary.STYPE_STRING_MIME_TYPE
       };
       
       XSystem system = connectAndAuthenticate(true);
       XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
       xset.setAutoDelete(false, false);  // Hopefully it will get scavenged tomorrow...
       
       // Set up one day retention
       xset.createRetention( false, id1 );
       xset.applyRetentionDurationPolicy( id1, false, policyName1 );
       xset.applyRetentionEnabledPolicy(  id1, false, policyName1 );

       // Set up one second retention, should not fail
       xset.createRetention( false, id2 );
       xset.applyRetentionDurationPolicy( id2, false, policyName2 );
       xset.applyRetentionEnabledPolicy(  id2, false, policyName2 );
       
       XUID xuid = xset.commit();
       checkProperties(xset, RETENTION_POLICY_PROPS, RETENTION_POLICY_TYPES);
       
       Thread.sleep(1500); // exceed the one second policy, but should still be retained
       
       assertTrue( "Expected XSet to still be retained", system.isXSetRetained(xuid));
       
       // Make sure a policy can be extended to infinity
       xset.setRetentionDuration(id2, false, -1);
       
       xset.close();
       system.close();
    }

    /**
     * Tests the policy retention methods.
     * @throws Exception
     */
    public void testBaseRetentionPolicy() throws Exception
    {
       System.out.println("ReferenceRetentionMgmtTest.testBaseRetentionPolicy...");
       final String policyName = RetentionMgr.POLICY_ONE_SECONDS;
       final String id = "base";
       final String prefix = ".xset.retention.";
       
       // Required properties for retention
       final String RETENTION_POLICY_PROPS[] = {
          prefix + id + ".enabled.policy",
          prefix + id + ".duration.policy"
       };
       
       final String RETENTION_POLICY_TYPES[] = {
          XAMLibrary.STYPE_STRING_MIME_TYPE,
          XAMLibrary.STYPE_STRING_MIME_TYPE
       };
       
       XSystem system = connectAndAuthenticate(true);
       XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);

       xset.applyBaseRetentionPolicy(false, policyName);
       
       XUID xuid = xset.commit();
       checkProperties(xset, RETENTION_POLICY_PROPS, RETENTION_POLICY_TYPES);
       
       assertEquals( "Expected duration to be one second", 1000, xset.getActualRetentionDuration(id));
       assertEquals( "Expected enabled to be true", true, xset.getActualRetentionEnabled(id));
       
       assertTrue( "Expected XSet to be under retention", system.isXSetRetained(xuid));
       xset.commit();
       
       Thread.sleep(1500);
       
       // At this point, the period should be expired
       assertFalse( "Expected XSet to be released from retention", system.isXSetRetained(xuid));
       
       // Make sure that base can be extended to infinity
       xset.setBaseRetention(false, -1);
       
       xset.close();
       system.close();
    }
}
