/*
 * Copyright (c) 2008, Sun Microsystems, Inc.
 * Copyright (c) 2008, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.Arrays;
import org.snia.xam.XAMException;
import org.snia.xam.XSet;
import org.snia.xam.XStream;
import org.snia.xam.XUID;
import org.snia.xam.vim.reference.SniaXUID;

/**
 * Utility routines for testing
 * 
 */
public class ReferenceTestUtilities {
    
    /** 
     * Creates an XStream (stream field) and populates it with the desired
     * characters (specified as byte argument). 
     * @param xset
     * @param fieldName
     * @param binding
     * @param mimeType
     * @param charValue
     * @param fieldLength
     * @return open XStream reference 
     * @throws org.snia.xam.XAMException
     */
    static XStream createXStream(XSet xset, String fieldName, 
            boolean binding, String mimeType, byte charValue, int fieldLength) 
        throws XAMException {
        XStream xstream = xset.createXStream(
                fieldName, binding, mimeType);
        byte[] buffer = new byte[fieldLength];
        for (int i = 0; i < buffer.length; i++) {
            buffer[i] = charValue;
        }
        xstream.write(buffer);
        return xstream;
        
    }

    static byte[] createByteArray(byte value, int arraySize) {
        byte[] byteArray = new byte[arraySize];
        for (int i = 0; i < arraySize; i++) {
            byteArray[i] = value;
        }
        
        return byteArray;
    }
    
    /**
     * Creates an XStream from a file
     * @param xset
     * @param fieldName
     * @param binding
     * @param type
     * @param filename
     * @return
     * @throws java.lang.Exception
     */
    protected long createStreamFromFile(XSet xset, String fieldName, 
            boolean binding,String type, String filename)
            throws Exception {
        File file = new File(filename);
        long fileSize = file.length();
        XStream xstream = xset.createXStream(fieldName, binding, type);
        byte[] buffer = new byte[4000];
        InputStream is = new FileInputStream(file);
        int numBytes = is.read(buffer);
        long totalBytes = 0;
        while (numBytes > 0) {
            totalBytes += numBytes;
            xstream.write(buffer, 0, numBytes);
            numBytes = is.read(buffer);
        }
        xstream.close();
        return totalBytes;

    }
    
    /**
     * Opens the stream field for reading. Reads in the stream contents and 
     * compares them against the expected data.
     * 
     * @param expectedData
     * @param expectedNum
     * @param xset
     * @param fieldName
     * @return
     */
    static boolean verifyStreamContents(byte[] expectedData, 
            int expectedNum, XSet xset, String fieldName) {
        boolean validData = false;
        int buffSize = expectedData.length;
        byte[] buffer = new byte[buffSize];
        long offset = 0;
        long totalBytesRead = 0;
        XStream stream = null;
        try {
            stream = xset.openXStream(fieldName, XStream.MODE_READ_ONLY);
            long num = stream.read(buffer, offset, buffSize-offset);
            while (num > 0) {
                totalBytesRead += num;
                offset += num;
                num = stream.read(buffer, offset, buffSize-offset);
            }
            if (totalBytesRead == expectedNum) {
                validData = true;
                for (int i = 0; i < (int) totalBytesRead; i++) {
                    if (expectedData[i] != buffer[i]) {
                        validData = false;
                        break;
                    }
                }
            }
            
        } catch (XAMException xex) {
            validData = false;
            xex.printStackTrace();
        } finally {
            if (stream != null) {
                try {
                    stream.abandon();
                    stream.close();
                } catch (XAMException xex) {
                    validData = false;
                    xex.printStackTrace();
                }
            }
        }
        
        return validData;
    }
    
    /**
     * Opens the stream field for reading. Reads in the stream contents and 
     * compares them against the expected data.
     * 
     * @param expectedData
     * @param expectedNum
     * @param xset
     * @param fieldName
     * @return
     */
    static boolean verifyStreamContents(byte expectedChar, 
            int expectedNum, XSet xset, String fieldName) {
        boolean validData = false;
        int buffSize = 5000;
        byte[] buffer = new byte[buffSize];
        long totalBytesRead = 0;
        XStream stream = null;
        try {
            stream = xset.openXStream(fieldName, XStream.MODE_READ_ONLY);
            long num = stream.read(buffer);
            while (num > 0) {
                totalBytesRead += num;
                for (int i = 0; i < num; i++) {
                    if (buffer[i] != expectedChar) {
                        System.out.println("Invalid data detected. " +
                            "Expected: " + (char) expectedChar +
                            "  Value: " + (char) buffer[i]);
                        return validData;
                    }
                }
                num = stream.read(buffer);
            }
            
            if (totalBytesRead == expectedNum) {
                validData = true;
            }
            
        } catch (XAMException xex) {
            validData = false;
            xex.printStackTrace();
        } finally {
            if (stream != null) {
                try {
                    stream.abandon();
                    stream.close();
                } catch (XAMException xex) {
                    validData = false;
                    xex.printStackTrace();
                }
            }
        }
        
        return validData;
    }

    /**
     * Opens the stream field for reading. Reads in the stream contents and 
     * compares them against the expected data.
     * 
     * @param expectedData
     * @param expectedNum
     * @param xset
     * @param fieldName
     * @return
     */
    static void verifyStreamData(byte[] data, 
            int expectedLength, byte expectedByteValue) throws Exception {
        int size = data.length;
        if (data.length != expectedLength) {
            throw new Exception("Stream is not correct size. Bytes Expected: " 
                    + expectedLength + "   Bytes read: " + size);
        }
        
        for (byte b : data) {
            if (b != expectedByteValue) {
                throw new Exception("Stream contains invalid output. " +
                        "Expected value: " + expectedByteValue +
                        "  Actual value: " + b);
            }
        }
    }
    
    /**
     * Verify the streams for an XSet. It is expected that each stream is
     * just a fixed byte repeated.
     * 
     * @param dirName the absolute pathname of the repository
     * @param xuid  XSet XUID
     * @param numStreams    number of expected streams (can be 0)
     * @param expectedSize  expected size of stream
     * @param expectedByteValue expected byte value for stream
     * 
     * @throws java.lang.Exception
     */
    static void verifyXSetStreams(String dirName, XUID xuid, 
            int numStreams, int expectedSize, byte expectedByteValue) 
            throws Exception {
        
        String xuid1DirName = dirName + File.separator + "XSet_" + 
                SniaXUID.convertToSafeString(xuid);
        File xuid1Dir = new File(xuid1DirName);
        if (!xuid1Dir.isDirectory()) {
            throw new Exception("Error no directory for xuid: " + xuid.toString() +
                    "\n" + SniaXUID.convertToSafeString(xuid));
        }
        
        File[] files = xuid1Dir.listFiles();
        if (files == null || files.length == 0) {
            if (numStreams > 0) {
                throw new Exception(
                    "The stream field was deleted from the orginal XSet");
            } else {
                return;
            }
        } else if (numStreams == 0) {
            throw new Exception("Incorrect number of streams exist for XSet. " + 
                    "Expected: " + numStreams + "  Actual: " + files.length +
                    "  Example filename: " + files[0].getAbsolutePath());
        }
        
        for (int i = 0; i < files.length; i++) {
            FileInputStream fis = new FileInputStream(files[i]);
            long size = files[i].length();
            if (size != expectedSize) {
                throw new Exception("Stream is not correct size. Bytes Written: " +
                        expectedSize + "   File size: " + size);
            }
            long pos = 0;
            if (size < 50000) {
                byte[] buffer = new byte[(int)size];
                FileChannel channel = fis.getChannel();
                ByteBuffer bb = ByteBuffer.wrap(buffer);
                int num = channel.read(bb, pos);
                while (num > 0) {
                    pos += num;
                    num = channel.read(bb, pos);
                }
                fis.close();
                verifyStreamData(buffer, expectedSize, expectedByteValue);
            }
        }   // end for

    }
}
