/*
 *  Copyright (c) 2008, Sun Microsystems, Inc.
 *  Copyright (c) 2008, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.InetAddress;
import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;

import junit.framework.Assert;

import junit.framework.Test;
import junit.framework.TestSuite;

import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.FieldExistsException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XSystem;
import org.snia.xam.base.XRI;
import org.snia.xam.vim.reference.utils.ReferenceSaslUtils;

/**
 *
 * 
 */
public class ReferenceXAMTest extends XAMTest {
    
   // Temporary passwd file stuff, so tests can create instances with other users.
   final String PASSWD_INFO[][] = { 
      { "user1", "pass1"},
      { "user2", "pass2"},
      { "user3", "pass3"},
      { "user4", "pass4"},
      { "xyzzy", "plugh"} };

   private static Logger logger = null;

   protected boolean isApplicationConnect = false;
    protected XRI xri = null;

    public static void main(String[] args)
    {
        System.exit(runTest(args, suite()));
    }

    public ReferenceXAMTest(String name)
    {
       super(name);
       
       initializeLogger();
    }

    /**
     * Initialize the logger by stepping through the basic operations to initialize the 
     * the XAMLibrary and create a ReferenceXSystem instance. Then ask the XSystem instance
     * for the logger that it has configured (according to the XAM Library settings).
     */
   private void initializeLogger()
   {
       // Grab the configured logger.
       try {
          // Set up s_xam to be able to retrieve logging settings.
          if( s_xam == null )
             initLibrary();
          String  path      = s_xam.getString(XAMLibrary.XAM_LOG_PATH);
          long    size      = s_xam.getLong(XAMLibrary.XAM_LOG_MAX_SIZE);
          boolean append    = s_xam.getBoolean(XAMLibrary.XAM_LOG_APPEND);
          long    verbosity = s_xam.getLong(XAMLibrary.XAM_LOG_VERBOSITY);
          long    level     = s_xam.getLong(XAMLibrary.XAM_LOG_LEVEL);
          ReferenceXSystem xsys = new ReferenceXSystem();
          try { xsys.createProperty(XAMLibrary.XAM_LOG_PATH, false, path ); }
          catch( FieldExistsException fe ) { xsys.setProperty(XAMLibrary.XAM_LOG_PATH, path ); }
          
          try { xsys.createProperty(XAMLibrary.XAM_LOG_MAX_SIZE, false, size ); }
          catch( FieldExistsException fe ) { xsys.setProperty(XAMLibrary.XAM_LOG_MAX_SIZE, size ); } 
          
          try { xsys.createProperty(XAMLibrary.XAM_LOG_APPEND, false, append ); }
          catch( FieldExistsException fe ) { xsys.setProperty(XAMLibrary.XAM_LOG_APPEND, append ); } 
          
          try { xsys.createProperty(XAMLibrary.XAM_LOG_VERBOSITY, false, verbosity ); }
          catch( FieldExistsException fe ) { xsys.setProperty(XAMLibrary.XAM_LOG_VERBOSITY, verbosity ); } 
          
          try { xsys.createProperty(XAMLibrary.XAM_LOG_LEVEL, false, level ); }
          catch( FieldExistsException fe ) { xsys.setProperty(XAMLibrary.XAM_LOG_LEVEL, level ); } 
          
          xsys.deleteField(XSystem.XAM_XSYSTEM_INITIALIZING);
          logger = xsys.getLogger();
          logger.setUseParentHandlers(false);
          xsys.close();
       }
       catch( Exception e ) {
          System.out.println( "ERROR Setting up the logger" );
          e.printStackTrace();
       }
   }

   public static Test suite()
    {
    	TestSuite suite = new TestSuite();
    	suite.addTest(new ReferenceXAMTest("testXAMLibConnect"));
        suite.addTest(new ReferenceXAMTest("testDirectConnect"));
    	return suite;
    }
    
    public XRI getXRI() throws Exception 
    {
        if (null == xri) {
            XRI xriObj = new XRI();
            xriObj.setVimInfo(TestConstants.VIM_NAME);
            String ip = "";
            try {
                ip = InetAddress.getLocalHost().getHostAddress();
            } catch (Exception ex) {
                throw new XAMException("Error getting local host address.",
                        ex);
            }
            xriObj.setSystem(ip);
            xri = xriObj;
        }
        return xri;
    }
    public XSystem connectAndAuthenticate(boolean useXAMLib) throws Exception
    {
        XSystem xs = doConnection(useXAMLib);
        doAuthentication(xs);
        return xs;
    }
    public XSystem doConnection(boolean useXAMLib) throws Exception 
    {
        XSystem xsystem = null;
        if (useXAMLib) {
           logger.fine( "-------- Connecting using XAMLibrary --------" );
           xsystem = s_xam.connect(s_xri);
        } else {
           logger.fine("-------- Connecting directly --------");
           xsystem = new ReferenceXSystem();
           
           // Transfer logging properties
           try { xsystem.createProperty( XAMLibrary.XAM_LOG_APPEND, false, true ); }
           catch( FieldExistsException fe ) { xsystem.setProperty( XAMLibrary.XAM_LOG_APPEND, true ); }
           
           try { xsystem.createProperty( XAMLibrary.XAM_LOG_LEVEL, false, 
                                   s_xam.getLong(XAMLibrary.XAM_LOG_LEVEL) ); }
           catch( FieldExistsException fe ) 
           { xsystem.setProperty( XAMLibrary.XAM_LOG_LEVEL, s_xam.getLong(XAMLibrary.XAM_LOG_LEVEL) ); } 
           
           try { xsystem.createProperty( XAMLibrary.XAM_LOG_VERBOSITY, false, 
                                         s_xam.getLong(XAMLibrary.XAM_LOG_VERBOSITY) ); }
           catch( FieldExistsException fe ) 
           { xsystem.setProperty(XAMLibrary.XAM_LOG_VERBOSITY,  s_xam.getLong(XAMLibrary.XAM_LOG_VERBOSITY)); } 

           try { xsystem.createProperty( XAMLibrary.XAM_LOG_PATH, false,
                                   s_xam.getString(XAMLibrary.XAM_LOG_PATH) ); }
           catch( FieldExistsException fe ) 
           { xsystem.setProperty( XAMLibrary.XAM_LOG_PATH, s_xam.getString(XAMLibrary.XAM_LOG_PATH)); } 

           try { xsystem.createProperty( XAMLibrary.XAM_LOG_MAX_SIZE, false,
                                   s_xam.getLong(XAMLibrary.XAM_LOG_MAX_SIZE) ); }
           catch( FieldExistsException fe ) 
           { xsystem.setProperty( XAMLibrary.XAM_LOG_MAX_SIZE, s_xam.getLong(XAMLibrary.XAM_LOG_MAX_SIZE)); } 

           try { xsystem.createProperty( XAMLibrary.XAM_LOG_MAX_ROLLOVERS, false,
                                   s_xam.getLong(XAMLibrary.XAM_LOG_MAX_ROLLOVERS) ); }
           catch( FieldExistsException fe ) 
           { xsystem.setProperty( XAMLibrary.XAM_LOG_MAX_ROLLOVERS, s_xam.getLong(XAMLibrary.XAM_LOG_MAX_ROLLOVERS)); } 
           
           // Initialization phase is done.
           xsystem.deleteField( XSystem.XAM_XSYSTEM_INITIALIZING );
           xsystem.connect(getXRI().toString());
        }
        Assert.assertTrue("Connect failed.   Null XSystem returned.", 
                                                        xsystem != null);
        return xsystem;
    }
    public void doAuthentication(XSystem system) throws Exception 
    {
        byte[] cred = ReferenceSaslUtils.SASL_MECHANISM_PLAIN.getBytes();
        ByteArrayOutputStream response = new ByteArrayOutputStream(200);
        int returnCode = system.authenticate(cred, response);
        // Validate return status value
        assertTrue("Test Failed. System should request more SASL data.",
                returnCode == XSystem.XAM_SASL_IN_PROCESS);

        ReferenceAuthenticationTest.verifyAuthResponse(
            ReferenceSaslUtils.SASL_RESPONSE_MORE.getBytes(), 
            response);       

        byte[] authInfo = ReferenceSaslUtils.encodeSASLPlain(
            "junkAuthorzId", ReferenceAuthenticationStatus.TEST_USERNAME,
            ReferenceAuthenticationStatus.TEST_PASSWORD);
        returnCode = system.authenticate(authInfo, response);
        assertTrue("Test Failed. System should NOT request more SASL data.",
                returnCode == XSystem.XAM_SASL_COMPLETE);

        verifyAuthResponse(ReferenceSaslUtils.SASL_DONE.getBytes(), response);
    }
    
    public static void verifyAuthResponse(byte[] expected, byte[] resp)
            throws Exception  {
        // Validate response       
        assertTrue("Auth Response array invalid. Expected: " + 
                Arrays.toString(expected) + "   Received: " +
                Arrays.toString(resp), Arrays.equals(expected, resp));


    }
    public static void verifyAuthResponse(
            byte[] expected, ByteArrayOutputStream respStream)
            throws Exception  {
        byte[] resp = respStream.toByteArray();
        
        String expected_msg = new String( expected );
        String resp_msg = new String( resp );
        
        // Validate response       
        assertTrue("Auth Response array invalid. Expected: " + 
                Arrays.toString(expected) + "/" + expected_msg + "   Received: " +
                Arrays.toString(resp) + "/" + resp_msg, Arrays.equals(expected, resp));
    }
    
    public void testXAMLibConnect() throws Exception
    {
        doConnection(true);
    }
    public void testDirectConnect() throws Exception
    {
        doConnection(false);
    }

    /**
     * Authenticate with the XSystem, using SASL plain, and the specified username password.
     * This routine will perform the setupPassword for your test. 
     * @param string
     * @param string2
     * @throws Exception 
     */
    public XSystem getAuthenticatedXSystem( String username, String password ) throws Exception
    {
       byte[] authInfo = ReferenceSaslUtils.encodeSASLPlain("junkAuthorzId",
          username, password );
       
       XSystem xsystem = null;
       File passwordFile = File.createTempFile( "tempPassword", ".dat" );
       String oldPasswd = setupPasswd( passwordFile );
       
       try {
          xsystem = doConnection(true);
          ByteArrayOutputStream respByteStream = new ByteArrayOutputStream();
          int respValue = xsystem.authenticate(ReferenceSaslUtils.SASL_MECHANISM_PLAIN.getBytes(),
             respByteStream);
          assertEquals("Test Failed. System should request more SASL data.",
             respValue,
             XSystem.XAM_SASL_IN_PROCESS);
          verifyAuthResponse(ReferenceSaslUtils.SASL_RESPONSE_MORE.getBytes(),
             respByteStream);
          respByteStream.reset();
          respValue = xsystem.authenticate(authInfo, respByteStream);
          assertEquals("Test Failed. System should not request more SASL data.",
             respValue,
             XSystem.XAM_SASL_COMPLETE);
          verifyAuthResponse(ReferenceSaslUtils.SASL_DONE.getBytes(),
             respByteStream);
       }
       finally {
          tearDownPasswd( passwordFile, oldPasswd );
       }
       
       return xsystem;
    }

    /**
     * Called when we are done with a test, it will delete the password file and reset the 
     * system property.
     * @param passwdPath The name of the file to create.
     * @param oldSetting The old setting of the org.snia.xam.reference.passwd
     */
    private void tearDownPasswd( File passwdFile, String oldSetting)
       throws XAMException
    {
       passwdFile.delete();
       
       if( oldSetting == null )
          s_xam.deleteField( ReferenceXSystem.PROPERTY_PASSWD );
       else
          s_xam.setProperty( ReferenceXSystem.PROPERTY_PASSWD, oldSetting );
    }

    /**
     * Set up the password file, with the username passwords from the constant PASSWD_INFO
     * @param passwdPath
     * @return The old system property setting for org.snia.xam.reference.passwd
     * @throws IOException
     */
    private String setupPasswd( File passwdFile ) throws IOException, XAMException
    {
       PrintWriter passwd_writer = new PrintWriter(passwdFile);
       for( int p = 0; p < PASSWD_INFO.length; p++ ) {
          passwd_writer.println( PASSWD_INFO[p][0] + " " + PASSWD_INFO[p][1] );
       }
       passwd_writer.close();
       
       // Grab the old value of the org.snia.xam.reference.passwd property so we can restore it later.
       String oldPasswdProperty = null;
       
       try {
          oldPasswdProperty = s_xam.getString(ReferenceXSystem.PROPERTY_PASSWD);
          s_xam.setProperty(ReferenceXSystem.PROPERTY_PASSWD, passwdFile.getAbsolutePath());
       }
       catch( FieldDoesNotExistException fdne ) {
          s_xam.createProperty(ReferenceXSystem.PROPERTY_PASSWD, false, passwdFile.getAbsolutePath());
       }
       
       return oldPasswdProperty;
    }
    
    /**
     * Return an anonymously authenticated XSystem.
     * @return
     * @throws Exception
     */
    public XSystem getAnonymousSystem()
       throws Exception
    {
       XSystem xsystem = doConnection(false);
      ByteArrayOutputStream response = new ByteArrayOutputStream(200);
      int retValue = xsystem.authenticate(ReferenceSaslUtils.SASL_MECHANISM_ANONYMOUS.getBytes(),
         response);
      assertTrue("Test Failed. System should request more SASL data.",
         retValue == XSystem.XAM_SASL_IN_PROCESS);
      verifyAuthResponse(ReferenceSaslUtils.SASL_RESPONSE_MORE.getBytes(),
         response);
      byte[] trace = ReferenceSaslUtils.encodeSASLAnonymous("myemail@mycompany.com");
      response.reset();
      retValue = xsystem.authenticate(trace, response);
      assertTrue("Test Failed. System should NOT request more SASL data.",
         retValue == XSystem.XAM_SASL_COMPLETE);
      return xsystem;
    }
    
    /**
     * This logs a message. (used to print on stdout) 
     */
    public void announce( String msg )
    {
       logger.log(Level.INFO, msg );
    }

    /**
     * Delete an old directory. Used in deleting portions of the old database 
     * or temporary directories built up during unit testing.
     * @param db
     */
    public void deleteDirectory(File dir)
    {
       File contents[] = dir.listFiles();
       
       if( contents != null )
       {
          for( int x = 0; x < contents.length; x++ )
             if( contents[x].isDirectory() )
                deleteDirectory(contents[x]);
             else
                contents[x].delete();
       }
       
       dir.delete();
    }
}
