/*
 *  Copyright 2008 Hewlett-Packard Development Company, L.P.
 *  Copyright (c) 2008-2009, Sun Microsystems, Inc.
 *  Copyright (c) 2008-2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.snia.xam.vim.reference;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;
import java.util.List;

import junit.framework.Test;
import junit.framework.TestSuite;

import org.snia.xam.FieldContainer;
import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.ObjectInUseException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XSet;
import org.snia.xam.XSetException;
import org.snia.xam.XSystem;
import org.snia.xam.XUID;
import org.snia.xam.base.DefaultXUID;
import org.snia.xam.util.ISO8601Date;

public class ReferenceXSetTest extends ReferenceXAMTest /* TestCase */ {
    protected static final String TEST_XUID_1 = "AAA6AwAVGUQxMjE2NDExNjg4Nzg5";
    protected static final String TEST_XUID_2 = "AAA6AwAVj4UxMjE2NDExNjg4Nzkw";
    protected static final String[] NEW_PROPS = {};
    protected static final String[] NEW_TYPES = {};
    protected static final String[] STORED_PROPS = {
        XSet.XAM_TIME_COMMIT,
        XSet.XAM_TIME_ACCESS,
        XSet.XAM_TIME_XUID,
        XSet.XAM_XUID,
        XSet.XAM_RETENTION_LIST_BASE,
        XSet.XAM_RETENTION_BASE_ENABLED,
        XSet.XAM_RETENTION_LIST_EVENT
    };
    protected static final String[] STORED_TYPES = {
        //XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
        XAMLibrary.STYPE_DATETIME_MIME_TYPE,
        XAMLibrary.STYPE_DATETIME_MIME_TYPE,
        XAMLibrary.STYPE_DATETIME_MIME_TYPE,
        XAMLibrary.STYPE_XUID_MIME_TYPE,
        XAMLibrary.STYPE_STRING_MIME_TYPE, /* list base */
        XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, /* base enabled */
        XAMLibrary.STYPE_STRING_MIME_TYPE /* list event */
    };
    protected static final String[] PROP_OBJ_NAMES = {
        "test.date",
        "test.xuid",
        "test.boolean",
        "test.float",
        "test.int",
        "test.string"
    };
    protected static final Object[] PROP_OBJ_VALS_1 = {
        null,
        null,
        Boolean.TRUE,
        new Double(123.45),
        new Long(12345),
        new String("testing....")
    };
    protected static final Object[] PROP_OBJ_VALS_2 = {
        null,
        null,
        Boolean.FALSE,
        new Double(123.45),
        new Long(12345),
        new String("testing....")
    };

    public static void main(String[] args) {
        System.exit(runTest(args, suite()));
    }

    public static Test suite() {
        TestSuite suite = new TestSuite();
        /*
         * files in tmpdir unless pass true to connectAndAuthenticate
         */
        suite.addTest(new ReferenceXSetTest("testXSetBasic"));
        suite.addTest(new ReferenceXSetTest("testXSetOpen"));
        suite.addTest(new ReferenceXSetTest("testCommitTimeChange"));
        suite.addTest(new ReferenceXSetTest("testPropertyObjects"));
        suite.addTest(new ReferenceXSetTest("testProperties"));
        suite.addTest(new ReferenceXSetTest("testTroublesomeStrings"));
        suite.addTest(new ReferenceXSetTest("testDeleteProperties"));
        suite.addTest(new ReferenceXSetTest("testXSet"));
        suite.addTest(new ReferenceXSetTest("testXSetModes"));
        suite.addTest(new ReferenceXSetTest("testAbandon"));
        suite.addTest(new ReferenceXSetTest("testClosed"));
        return suite;
    }

    public ReferenceXSetTest(String name) {
        super(name);
    }

    public void testPropertyObjects() throws Exception {
        final String PREFIX = "test.prop.obj.";

        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);

        announce( "Create XUID Property" );
        // XAM_XUID - createProperty
        XUID x = DefaultXUID.valueOf(TEST_XUID_1);
        xset.createProperty( PREFIX+"xuid", false, x );
        assertTrue( "Create value xuid", 
                    x.equals( xset.getXUID( PREFIX+"xuid")) );
        
        announce( "Create Datetime Property" );
        // XAM_DATETIME - createProperty
        Calendar c = Calendar.getInstance();
        xset.createProperty( PREFIX+"cal", false, c );
        assertTrue( "Create value calendar",
                    c.equals( xset.getDateTime(PREFIX+"cal")) );
        announce( "Create XUID2 Property" );
        // XAM_XUID - setProperty
    	XUID x2 = DefaultXUID.valueOf(TEST_XUID_2);
        xset.setProperty( PREFIX+"xuid", x2 );
        assertTrue( "Set value xuid", 
                    x2.equals( xset.getXUID( PREFIX+"xuid")) );

        announce( "Create Cal Property" );
       // XAM_XUID - setProperty
        Calendar c2 = Calendar.getInstance();
        xset.setProperty( PREFIX+"cal", c2 );
        assertTrue( "Create value calendar",
                    c2.equals( xset.getDateTime(PREFIX+"cal")) );
     	
    	xset.close();
    	system.close();
	}


    public void testProperties() throws Exception {
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        checkProperties(xset, NEW_PROPS, NEW_TYPES);
        // creating an XSet with binding system properties - XSet is not clean
        checkXSetDirty(xset, true);
    	
    	XUID xuid  = xset.commit();
    	checkProperties(xset, STORED_PROPS, STORED_TYPES); 
        checkXSetDirty(xset, true);
    	assertTrue("XSet xuid", xuid.equals(xset.getXUID(XSet.XAM_XUID)));
    	
    	xset.createProperty("test.boolean", false, true);
    	assertEquals("Boolean property", true, xset.getBoolean("test.boolean"));
        checkXSetDirty(xset, false);
    	xset.createProperty("test.int", false, 12345L);
    	assertEquals("Int property", 12345L, xset.getLong("test.int"));
        checkXSetDirty(xset, false);
    	xset.createProperty("test.float", false, 1234.5);
    	assertEquals("Float property", 1234.5, xset.getDouble("test.float"), 0.0);
    	Calendar cal = Calendar.getInstance();
    	xset.createProperty("test.date", false, cal);
    	assertTrue("Calendar property", cal.equals(xset.getDateTime("test.date")));
    	String str = "testing....";
    	xset.createProperty("test.string", false, str);
    	assertTrue("String property", str.equals(xset.getString("test.string")));
    	xset.createProperty("test.xuid", false, xuid);
    	assertTrue("XUID property", xuid.equals(xset.getXUID("test.xuid")));
    	xuid = xset.commit();
    	xset.close();
    	
    	// Make assure persisted values work correctly.
    	xset = system.openXSet( xuid, XSet.MODE_READ_ONLY );

        assertEquals("Boolean property", true, xset.getBoolean("test.boolean"));
        assertEquals("Int property", 12345L, xset.getLong("test.int"));
        assertEquals("Float property", 1234.5, xset.getDouble("test.float"), 0.0);
        String orig8601 = ISO8601Date.toISO8601String(cal);
        String new8601  = ISO8601Date.toISO8601String(xset.getDateTime("test.date"));
        assertEquals("Calendar property", orig8601, new8601);
        assertEquals("Calender millis", cal.getTimeInMillis(), xset.getDateTime("test.date").getTimeInMillis() );
        assertTrue("String property", str.equals(xset.getString("test.string")));
    	
    	xset.close();
    	
    	system.close();
	}

    /**
     * Test to make sure that XML characters don't screw up the XML writer.
     * @throws Exception
     */
    public void testTroublesomeStrings() throws Exception 
    {
       final String TEST_STR = "http://www.topcoder.com/tc?module=ProjectDetail&pj=30007306";
       final String PROP     = "com.example.prop";
       
       final String TEST_STR2 = "some stuff <b>";
       final String PROP2    = "com.example.prop2";
       
       final String TEST_STR3 = "whitepsace test \t <>><";
       final String PROP3    = "com.example.prop3";
       
       XSystem system = connectAndAuthenticate(true);
       XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
       
       xset.createProperty( PROP, false, TEST_STR );
       xset.createProperty( PROP2, false, TEST_STR2 );
       xset.createProperty( PROP3, false, TEST_STR3 );
       XUID xuid = xset.commit();
       xset.close();
       
       xset = system.openXSet( xuid, XSet.MODE_READ_ONLY );
       assertEquals( "Expected the property string to read back the same", TEST_STR, xset.getString(PROP));
       assertEquals( "Expected the property2 string to read back the same", TEST_STR2, xset.getString(PROP2));
       assertEquals( "Expected the property3 string to read back the same", TEST_STR3, xset.getString(PROP3));
       xset.close();
       
       system.close();
   }

    public void testCommitTimeChange() throws Exception {
        announce("ReferenceXSetTest.testCommitTimeChange");
        XSystem system = connectAndAuthenticate(true);
        XSet xset1 = system.createXSet(XSet.MODE_UNRESTRICTED);
        Calendar cal = Calendar.getInstance();
        xset1.createProperty("testDate", false, cal);
        assertTrue("Calendar property", cal.equals(xset1.getDateTime("testDate")));
        XUID xuid1 = xset1.commit();
        
        // TESTING OPENING XSET and MODIFYING FILE TIMESTAMP        
        // First instance of the XSet
        XSet xset2 = system.openXSet(xuid1, XSet.MODE_UNRESTRICTED);
        xset2.createProperty("testBoolean", false, true);
        
        // Assures long enough offset between each XSet instance being opened
        Thread.sleep(1000);
        
        // Second instance of the same XSet as above
        XSet xset3 = system.openXSet(xuid1, XSet.MODE_UNRESTRICTED);
        // at this point, commit time for the persisted XSet has NOT changed
        // at this point, timestamp for the persisted XSet's file HAS changed

        // TESTING COMMITING XSET AFTER PERSISTED XSET CHANGED
        // open a second instance of the XSet and commit changes -- then try
        // and commit the previous instance
        xset3.createProperty("testInteger", false, 200);
        XUID xuid2 = xset3.commit();
        assertTrue("XUID should not change upon commit: " +
              "xuid1: " + xuid1.toString() + "  xuidRW2: " + xuid2.toString(), 
                xuid1.equals(xuid2));
        try {
            XUID xuid3 = xset2.commit();
        } catch (XAMException e) {
            //announce("Expected exception: " + e.getMessage());
            xset2.abandon();
        }
        

        xset1.close();
        xset2.close();
        xset3.close();
        system.close();
    }
    public void testBindingChange() throws Exception
	{
    	XSystem system = connectAndAuthenticate(true);
    	XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
    	checkProperties(xset, NEW_PROPS, NEW_TYPES);
        checkXSetDirty(xset, true);
    	
    	XUID xuid  = xset.commit();
    	checkProperties(xset, STORED_PROPS, STORED_TYPES); 
        checkXSetDirty(xset, true);
    	assertTrue("XSet xuid", xuid.equals(xset.getXUID(XSet.XAM_XUID)));
    	
    	xset.createProperty("test.boolean", false, true);
    	assertEquals("Boolean property", true, xset.getBoolean("test.boolean"));
        checkXSetDirty(xset, false);
        XUID xuid2  = xset.commit();
        
    	xset.createProperty("test.int", false, 12345L);
    	assertEquals("Int property", 12345L, xset.getLong("test.int"));
        checkXSetDirty(xset, true);
    	xset.createProperty("test.float", false, 1234.5);
    	assertEquals("Float property", 1234.5, xset.getDouble("test.float"), 0.0);
    	Calendar cal = Calendar.getInstance();
    	xset.createProperty("test.date", false, cal);
    	assertTrue("Calendar property", cal.equals(xset.getDateTime("test.date")));
    	String str = "testing....";
    	xset.createProperty("test.string", false, str);
    	assertTrue("String property", str.equals(xset.getString("test.string")));
    	xset.createProperty("test.xuid", false, xuid);
    	assertTrue("XUID property", xuid.equals(xset.getXUID("test.xuid")));
    	xset.close();
    	
    	system.close();
	   }

    public void testDeleteProperties() throws Exception	{
    	XSystem system = connectAndAuthenticate(true);
    	XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
    	
    	xset.createProperty("test.boolean", false, true);
    	assertEquals("Boolean property", true, xset.getBoolean("test.boolean"));
        checkXSetDirty(xset, false);
        XUID xuid1  = xset.commit();
        
    	xset.createProperty("test.int", true, 12345L);
    	assertEquals("Int property", 12345L, xset.getLong("test.int"));
        checkXSetDirty(xset, false);
        
    	xset.createProperty("test.float", false, 1234.5);
    	assertEquals("Float property", 1234.5, xset.getDouble("test.float"), 
                123.45);
        XUID xuid2  = xset.commit();
        assertFalse("XUID should change, added binding properties: " +
                "xuid1: " + xuid1.toString() + "  xuid2: " + xuid2.toString(), 
                xuid1.toString().compareTo(xuid2.toString()) == 0);

        // Delete non-binding field
        xset.deleteField("test.float");
        assertFalse("Field test.float was deleted, it should not exist.", 
                xset.containsField("test.float"));
        XUID xuid3 = xset.commit();
        assertTrue("XUID should not change, deleted nonbinding properties: " +
                "xuid2: " + xuid2.toString() + "  xuid3: " + xuid3.toString(), 
                xuid2.equals(xuid3));
        // Delete binding field
        xset.deleteField("test.int");
        assertFalse("Field test.int was deleted, it should not exist.", 
                xset.containsField("test.int"));
        XUID xuid4 = xset.commit();
        assertFalse("XUID should change, deleted binding properties: " +
                "xuid3: " + xuid3.toString() + "  xuid4: " + xuid4.toString(), 
                xuid3.equals(xuid4));
        /*
         * Close and open again to make sure deleted properties were not
         * persisted.
         */
        xset.close();
    	XSet xset2 = system.openXSet(xuid4, XSet.MODE_UNRESTRICTED);
        assertFalse("test.int was deleted, it should not exist in " +
                    " the persisted XSet.", xset2.containsField("test.int"));
        assertFalse("test.float was deleted, it should not exist in " +
                    " the persisted XSet.", xset2.containsField("test.float"));
    	assertEquals("Boolean property", true, xset2.getBoolean("test.boolean"));
        xset2.close();
    	system.close();
    }

    // Either the XSet is dirty and the dirty property is set or
    // the XSet is clean and the XSet property should not exist
    private void checkXSetDirty(XSet xset, boolean xsetClean)
            throws XAMException {
        try {
            // get xset.dirty value
            boolean xsetDirtyValue = xset.getBoolean(XSet.XAM_DIRTY);
            // Only reach this line if got value for dirty field
            if (xsetClean) {
                fail("XSet is clean. The " + XSet.XAM_DIRTY + " field should not exist");
            }
            // Otherwise XSet has been modified and dirty field should be true
            assertEquals("XSet dirty should be true ", true, xsetDirtyValue);

        } catch (FieldDoesNotExistException ex) {
            if (!xsetClean) {
                fail("XSet is dirty but " + XSet.XAM_DIRTY + " does not exist");
            }

        }
    }

    public void testXSetBasic() throws Exception {
        announce("ReferenceXSetTest.testXSetBasic");
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        XUID xuid1 = xset.commit();
        // Verify that after commit everything is still the same
        XUID checkXUID = xset.getXUID(XSet.XAM_XUID);            
        assertTrue("XUID should not be changed after commit. xuid: " +
                xuid1.toString() + "    xuid after commit: " + 
                checkXUID.toString(), xuid1.equals(checkXUID));
        
        xset.createProperty("testBoolean", false, true);
        xset.createProperty("testInteger", true, 125);
        xset.createProperty("testDouble", true, (double) 1.4);
        
        // test commiting new binding properties
        XUID xuid2 = xset.commit();
        assertFalse("XUID should change, adding binding properties: " +
                "xuid1: " + xuid1.toString() + "  xuid2: " + xuid2.toString(), 
                xuid1.equals(xuid2));
        
        // Changing value on non-binding property, no new XUID generated
        xset.setProperty("testBoolean", true);
        XUID xuid3 = xset.commit();
        assertEquals("Failed to modify testBoolean value",
                xset.getBoolean("testBoolean"), true);
        assertTrue("XUID should not change, modifying value of " +
                "non-binding property", xuid3.equals(xuid2));
        
        // Changing value on binding property, new XUID generated
        xset.setProperty("testInteger", 225);
        XUID xuid4 = xset.commit();
        assertEquals("Failed to modify testInteger value",
                xset.getLong("testInteger"), 225);
        assertFalse("XUID should change, modifying value of " +
                "binding property", xuid4.equals(xuid3));

        // Change field from binding to non-binding should generate new XUID
        xset.setFieldAsNonbinding("testInteger");
        XUID xuid5 = xset.commit();
        assertFalse(
                "XUID should change, binding field changed to non-binding",
                xuid5.equals(xuid4));
        
        // Add non-binding field, XUID should not change
        xset.createProperty("testBoolean2", false, true);
        XUID xuid6 = xset.commit();
        assertTrue("XUID should not change, adding a  " +
                "non-binding field", xuid6.equals(xuid5));
        
        xset.close();
        system.close();
    }

    public void testXSetOpen() throws Exception {
        announce("ReferenceXSetTest.testXSetOpen");
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        XUID xuid1 = xset.commit();
        long integerValue = 125;
        double floatValue = 1.4;

        xset.createProperty("testBoolean", false, true);
        xset.createProperty("testInteger", true, 125);
        xset.createProperty("testDouble", true, (double) 1.4);

        xset.setProperty("testBoolean", true);
        XUID xuid2 = xset.commit();
        //assertTrue("XUID should not change", xuid3.equals(xuid2));
        assertEquals("testBoolean value",
                xset.getBoolean("testBoolean"), true);
        xset.close();
        XSet xset2 = system.openXSet(xuid2, XSet.MODE_READ_ONLY);
        assertEquals("testBoolean value",
                xset2.getBoolean("testBoolean"), true);
        assertEquals("testInteger", xset2.getLong("testInteger"), integerValue);
        assertEquals("testDouble", xset2.getDouble("testDouble"), floatValue);
        boolean modified = false;
        try {
            xset2.setProperty("testBoolean", true);
            modified = true;
        } catch (XSetException xe) {
            //announce("Expected exception: " + xe.getMessage());
        }
        assertEquals("Field modifed in read only XSet", false, modified);

        xset2.close();
        modified = false;
        try {
            xset2.setProperty("testBoolean", true);
            modified = true;
        } catch (XAMException xe) {
            //announce("Expected exception: " + xe.getMessage());
        }

        assertEquals("Field modifed in closed XSet", false, modified);

        system.close();
    }

    public void testXSet() throws Exception {
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        XUID xuid1 = xset.commit();
        xset.close();

        xset = system.openXSet(xuid1, XSet.MODE_UNRESTRICTED);
        xset.createProperty("test.boolean", false, true);
        XUID xuid2 = xset.commit();
        assertTrue("XUID should not change", xuid2.equals(xuid1));
        assertEquals("test.boolean value",
                xset.getBoolean("test.boolean"), true);
        xset.close();

        xset = system.openXSet(xuid1, XSet.MODE_UNRESTRICTED);
        assertEquals("test.boolean value",
                xset.getBoolean("test.boolean"), true);
        xset.setFieldAsBinding("test.boolean");
        xset.setProperty("test.boolean", false);
        XUID xuid3 = xset.commit();
        assertTrue("XUID should change", !xuid3.equals(xuid2));
        xset.close();

        xset = system.openXSet(xuid1, XSet.MODE_READ_ONLY);
        assertEquals("test.boolean value",
                xset.getBoolean("test.boolean"), true);
        xset.close();

        xset = system.openXSet(xuid3, XSet.MODE_READ_ONLY);
        assertEquals("test.boolean value",
                xset.getBoolean("test.boolean"), false);
        xset.close();

        system.close();
    }

    public void testXSetModes() throws Exception {
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        xset.createProperty("test.binding", true, true);
        xset.createProperty("test.nonbinding", false, true);
        XUID xuid = xset.commit();
        xset.close();

        xset = system.openXSet(xuid, XSet.MODE_READ_ONLY);
        xset.getBoolean("test.binding");
        xset.getBoolean("test.nonbinding");
        try {
            xset.setProperty("test.binding", false);
            fail("Should not be able to set a property on a read only xset");
        } catch (XSetException x) {
        }
        xset.close();

        xset = system.openXSet(xuid, XSet.MODE_RESTRICTED);
        xset.setProperty("test.nonbinding", false);
        try {
            xset.setProperty("test.binding", false);
            fail("Should not be able to set a binding property on a restricted xset");
        } catch (XSetException x) {
        }
        xset.close();

        xset = system.openXSet(xuid, XSet.MODE_UNRESTRICTED);
        xset.setProperty("test.binding", false);
        xset.setProperty("test.nonbinding", false);
        xset.close();

    //system.close();
    }

    public void testClosed() throws Exception {
        announce("ReferenceXSetTest.testClosed");
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        xset.close();

        try {
            xset.createProperty("test", false, true);
            fail("Operation should have failed, the xset is closed");
        } catch (XAMException ex) {
            // should get this exception
        }

        xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        try {
            system.close();
            fail("Operation should have failed, the open xsets");
        //xset.createProperty("test", false, true);
        //fail("Operation should have failed, the system is closed");
        } catch (ObjectInUseException x) {
        }
    /*    	catch (InvalidOperationException x)
    {
    }
     */
    }

    /**
     * Tests abandoning XSets
     * @throws java.lang.Exception
     */
    public void testAbandon() throws Exception {
        XSystem system = connectAndAuthenticate(true);
        // Abandon empty XSet
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        xset.abandon();

        try {
            xset.createProperty("test", false, true);
            fail("Operation should have failed, the xset is abandoned");
        } catch (XAMException ex) {
            // should get this exception
        }

        try {
            xset.close();
        } catch (XAMException x) {
            fail("Close operation on abandoned XSet should have succeeded");
        }

    }

    protected void checkProperties(FieldContainer ct,
            String[] expected, String[] types)
            throws Exception {
        List<String> props = new ArrayList<String>();
        for (int i = 0; i < expected.length; i++) {
            props.add(expected[i]);
        }
        Iterator iter = ct.openFieldIterator("");
        while (iter.hasNext()) {
            props.remove((String) iter.next());
        }
        int notFound = props.size();
        if (notFound > 0) {
            StringBuffer buf = new StringBuffer(256);
            buf.append("The following properties were not found: ");
            buf.append(props.get(0));
            for (int i = 1; i < notFound; i++) {
                buf.append(", ");
                buf.append(props.get(i));
            }
            fail(buf.toString());
        }

        for (int i = 0; i < expected.length; i++) {
            assertEquals(expected[i] + " type",
                    types[i],
                    ct.getFieldType(expected[i]));
        }
    }

}
