/*
 * Copyright (c) 2008-2009, Sun Microsystems, Inc.
 * Copyright (c) 2008-2009, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.File;
import junit.framework.Test;
import junit.framework.TestSuite;

import org.snia.xam.FieldContainer;
import org.snia.xam.FieldExistsException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidFieldTypeException;
import org.snia.xam.ObjectInUseException;
import org.snia.xam.XAMException;
import org.snia.xam.XSet;
import org.snia.xam.XSetException;
import org.snia.xam.XStream;
import org.snia.xam.XSystem;
import org.snia.xam.XUID;
import org.snia.xam.base.XRI;
import org.snia.xam.vim.reference.ReferenceVIMPersistenceMgr;
import org.snia.xam.vim.reference.utils.ReferenceSaslUtils;

public class ReferenceXStreamTest extends ReferenceXAMTest /* TestCase */ {

    public static void main(String[] args) {
        System.exit(runTest(args, suite()));
    }

    public static Test suite() {
        TestSuite suite = new TestSuite();
        suite.addTest(new ReferenceXStreamTest("testXStreamBasic"));
        suite.addTest(new ReferenceXStreamTest("testXStreamAccessAllowed"));
        suite.addTest(new ReferenceXStreamTest("testXStreamPersistMult"));
        suite.addTest(new ReferenceXStreamTest("testXStreamFieldExists"));
        suite.addTest(new ReferenceXStreamTest("testXStreamDeleteUncommitted"));
        suite.addTest(new ReferenceXStreamTest("testXStreamDeleteUncommittedMult"));
        suite.addTest(new ReferenceXStreamTest("testXStreamDelete"));
        suite.addTest(new ReferenceXStreamTest("testXStreamDeleteOpen"));
        suite.addTest(new ReferenceXStreamTest("testXStreamAccessCheck"));
        suite.addTest(new ReferenceXStreamTest("testXStreamPersist"));
        suite.addTest(new ReferenceXStreamTest("testAbandonStreams"));
        suite.addTest(new ReferenceXStreamTest("testAbandonStreamsAdvanced"));
        suite.addTest(new ReferenceXStreamTest("testMimeValidation"));
        return suite;
    }

    public ReferenceXStreamTest(String name) {
        super(name);
    }

    /**
     * Tests 
     * - Creating an XStream instance (i.e. a stream field) 
     * - Writing to the stream
     * - Getting current position using .tell();
     * - Closes the stream and commits the XSet
     * 
     * @throws java.lang.Exception
     */
    public void testXStreamBasic() throws Exception {
        announce("ReferenceXStreamTest.testXStreamBasic");

        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        
        XStream xstream = xset.createXStream(
                "stream.field", true, "image/jpeg");
        byte[] buffer = new byte[100];
        for (int i = 0; i < buffer.length; i++) {
            buffer[i] = 'b';
        }
        xstream.write(buffer);
        long position = xstream.tell();
        assertTrue( "Current file position is not at the end. Position: " +
                position + " ", 
                    position == buffer.length);
        
        xstream.close();
        XUID xuid = xset.commit();
        xset.close();
        system.close();
	}
    
    /**
     * - Tests creating, writing to and persisting an XStream (stream field).
     * Opens the stream field for read access and reads data back in to verify
     * contents against the data written.
     * - Tests appending to an existing XStream.
     * - Tests truncating an existing XStream
     * - Verify XStream data is persisted correctly when changes made to
     *   the XSet
     *   - Tests adding a non-binding property to XSet
     *   - Tests adding a binding property to the XSet
     * @throws java.lang.Exception
     */
    public void testXStreamPersist() throws Exception {
        int bufferSize = 200000;
        announce("ReferenceXStreamTest.testXStreamPersist");

        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        
        XStream xstream = xset.createXStream("stream.field", true, "image/jpeg");
        byte[] buffer = new byte[bufferSize];
        for (int i = 0; i < buffer.length; i++) {
            buffer[i] = 'b';
        }
        long numBytesWritten = xstream.write(buffer);
        assertTrue("Number of bytes written does not match number sent",
                numBytesWritten == buffer.length);
        xstream.close();
        XUID xuid = xset.commit();
        xset.close();

        xset = system.openXSet(xuid, XSet.MODE_UNRESTRICTED);
        
        assertTrue("Stream data read does not match data written",
                verifyStreamContents(buffer, buffer.length, xset, 
                    "stream.field"));
        
        // Read data
        XStream stream2 = xset.openXStream("stream.field", XStream.MODE_READ_ONLY);
        byte[] buffer2 = new byte[bufferSize];
        byte[] bufferSmall = new byte[1000];
        long num = stream2.read(bufferSmall);
        int outPos = 0;
        while (num > 0) {
            // can cast num to an int because it will never be larger than 1000
            System.arraycopy(bufferSmall, 0, buffer2, outPos, (int) num);
            outPos += num;
            num = stream2.read(bufferSmall);
        }
        
        stream2.close();
        
        // Test appending to XStream
        stream2 = xset.openXStream("stream.field", XStream.MODE_WRITE_APPEND);
        byte[] dataOutput = new byte[buffer.length *2];
        System.arraycopy(buffer, 0, dataOutput, 0, buffer.length);
        
        byte[] changeBuffer = new byte[10000];
        for (int i = 0; i < changeBuffer.length; i++) {
            changeBuffer[i] = 'a';
        }
        for (int i = 0; i < buffer.length/2; i++) {
            buffer[i] = 'a';
        }
        System.arraycopy(changeBuffer, 0, dataOutput, 
                buffer.length, changeBuffer.length);
        
        stream2.write(changeBuffer);
        stream2.close();
        XUID xuid2 = xset.commit();

        assertTrue("Changed bound stream field, XUIDs should be different",
                !xuid.equals(xuid2));
        assertTrue("Stream data read does not match data written",
                verifyStreamContents(dataOutput, buffer.length+changeBuffer.length,
                xset, "stream.field"));

        // Truncate stream field
        XStream stream3 = xset.openXStream("stream.field", 
                XStream.MODE_WRITE_TRUNCATE);
        for (int i = 0; i < 10; i++) {
            buffer[i] = 'c';
        }
        stream3.write(buffer, 0, 10);
        stream3.close();
        XUID xuid3 = xset.commit();

        assertTrue("Changed bound stream field, XUIDs should be different",
                !xuid3.equals(xuid2));
        System.arraycopy(buffer, 0, dataOutput, 0, buffer.length);
        assertTrue("Stream data read does not match data written",
                verifyStreamContents(dataOutput, 10, xset, "stream.field"));
        xset.close();
        
        // Add nonbinding field change
        xset = system.openXSet(xuid3, XSet.MODE_RESTRICTED);
        xset.createProperty("test.nb.boolean", false, false);
        XUID xuid4 = xset.commit();
        assertTrue("Changed unbound property field, XUIDs should be the same",
                xuid3.equals(xuid4));
        assertTrue("Stream data read does not match expected data",
                verifyStreamContents(dataOutput, 10, xset, "stream.field"));
        xset.close();
        
        // Add binding field change

        xset = system.openXSet(xuid4, XSet.MODE_UNRESTRICTED);
    	xset.createProperty("test.int", true, 12345L);
        XUID xuid5 = xset.commit();
        assertTrue("Changed bound property field, XUIDs should be different",
                !xuid5.equals(xuid4));
        assertTrue("Stream data read does not match expected data",
                verifyStreamContents(dataOutput, 10, xset, "stream.field"));
        xset.close();
        system.close();

    }


    /**
     * - Tests creating, writing to and persisting an XStream (stream field).
     * Opens the stream field for read access and reads data back in to verify
     * contents against the data written.
     * - Tests appending to an existing XStream.
     * - Tests truncating an existing XStream
     * - Verify XStream data is persisted correctly when changes made to
     *   the XSet
     *   - Tests adding a non-binding property to XSet
     *   - Tests adding a binding property to the XSet
     * @throws java.lang.Exception
     */
    public void testXStreamPersistMult() throws Exception {
        int bufferSize = 200000;
        byte byteChar = (byte) 'a';
        String streamFieldNameBase = "stream.field";
        int numStreamFields = 20;
        announce("ReferenceXStreamTest.testXStreamPersistMult");

        // Need to have a handle to a reference XSystem in order to get file
        // location used in test. So do manual authentication.
        ReferenceXSystem system = new ReferenceXSystem();
        system.connect("snia-xam://SNIA_Reference_VIM!localhost");
        byte[] resp = system.authenticate(ReferenceSaslUtils.
                SASL_MECHANISM_ANONYMOUS.getBytes());
        verifyAuthResponse(ReferenceSaslUtils.
                SASL_RESPONSE_MORE.getBytes(), resp);
        byte[] trace = ReferenceSaslUtils.encodeSASLAnonymous(
                "myemail@mycompany.com");
        resp = system.authenticate(trace);
                
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        //Create 20 XStreams
        for (int i = 0; i < numStreamFields; i++) {
            String fieldName = streamFieldNameBase + i;
            XStream xstream = xset.createXStream(fieldName, true, "image/jpeg");
            byte[] buffer = new byte[bufferSize];
            for (int j = 0; j < buffer.length; j++) {
                buffer[j] = byteChar;
            }
            long numBytesWritten = xstream.write(buffer);
            assertTrue("Number of bytes written does not match number sent",
                    numBytesWritten == buffer.length);
            byteChar++;
            xstream.close();
        }
        XUID xuid = xset.commit();
        xset.close();
        
        // Verify 20 stream files persisted
        ReferenceVIMPersistenceMgr persistMgr = system.getPersistenceMgr();
        String dirString = persistMgr.getXSetDirName(xuid);

        File dirName = new File(dirString);
        
        File[] filesList = dirName.listFiles();
        int streamCount = 0;
        for (int i = 0; i < filesList.length; i++) {
            String fileName = filesList[i].getName();
            if (fileName.startsWith(XStreamContents.XSTREAM_FILE_PREFIX)  && 
                fileName.endsWith(XStreamContents.XSTREAM_FILE_SUFFIX)) {
                streamCount++;
                assertTrue("Persisted XStream is incorrect size." + fileName,
                        bufferSize == filesList[i].length());
            }
        }
        assertTrue("All XStreams were not persisted.",
                streamCount == numStreamFields);
        xset = system.openXSet(xuid, XSet.MODE_UNRESTRICTED);
                
        // TODO:  Read data back in
        for (int i = 0; i < numStreamFields; i++) {
            String fieldName = streamFieldNameBase + i;
            XStream xstream = xset.openXStream(fieldName, XStream.MODE_READ_ONLY);
            //assertTrue("Stream data read does not match data written",
            //         ReferenceTestUtilities.verifyStreamData(data, streamCount, byteChar)
            xstream.close();
        }
        xset.close();
        system.close();

    }
    
    /**
     * - Tests creating an XStream (stream field) when the specified field
     *   already exists.
     * Opens the stream field for read access and reads data back in to verify
     * contents against the data written.
     * - Tests appending to an existing XStream.
     * - Tests truncating an existing XStream
     * - Verify XStream data is persisted correctly when changes made to
     *   the XSet
     *   - Tests adding a non-binding property to XSet
     *   - Tests adding a binding property to the XSet
     * @throws java.lang.Exception
     */
    public void testXStreamFieldExists() throws Exception {
        int bufferSize = 2000;
        byte byteChar = (byte) 'a';
        String streamFieldNameBase = "stream.field";
        int numStreamFields = 3;
        announce("ReferenceXStreamTest.testXStreamFieldExists");

        // Need to have a handle to a reference XSystem in order to get file
        // location used in test. So do manual authentication.
        ReferenceXSystem system = new ReferenceXSystem();
        system.connect("snia-xam://SNIA_Reference_VIM!localhost");
        byte[] resp = system.authenticate(ReferenceSaslUtils.
                SASL_MECHANISM_ANONYMOUS.getBytes());
        verifyAuthResponse(ReferenceSaslUtils.
                SASL_RESPONSE_MORE.getBytes(), resp);
        byte[] trace = ReferenceSaslUtils.encodeSASLAnonymous(
                "myemail@mycompany.com");
        resp = system.authenticate(trace);
                
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        
        //Create 2 XStreams with the same name
        for (int i = 0; i < numStreamFields; i++) {
            String fieldName = streamFieldNameBase;
            XStream xstream = null;
            try {
                xstream = 
                        xset.createXStream(fieldName, true, "image/jpeg");
                if (i != 0) {
                    fail("Should not be able to create second XStream with " +
                            "the same field name: " + fieldName);
                }
            } catch (FieldExistsException fe) {
                // If not first time creating XStream then ignore the 
                // expected exception
                if (i == 0) {
                    fail("Creating first XStream failed. Field Name: " +
                            fieldName);
                 }
                
            }
            if (xstream != null) {
                byte[] buffer = new byte[bufferSize];
                for (int j = 0; j < buffer.length; j++) {
                    buffer[j] = byteChar;
                }
                long numBytesWritten = xstream.write(buffer);
                assertTrue("Number of bytes written does not match number sent",
                        numBytesWritten == buffer.length);
                byteChar++;
                xstream.close();
            }
        }
        XUID xuid = xset.commit();
        xset.close();
        
        // Verify only 1 stream file persisted
        ReferenceVIMPersistenceMgr persistMgr = system.getPersistenceMgr();
        String dirString = persistMgr.getXSetDirName(xuid);

        File dirName = new File(dirString);
        
        File[] filesList = dirName.listFiles();
        int streamCount = 0;
        for (int i = 0; i < filesList.length; i++) {
            String fileName = filesList[i].getName();
            if (fileName.startsWith(XStreamContents.XSTREAM_FILE_PREFIX)  && 
                fileName.endsWith(XStreamContents.XSTREAM_FILE_SUFFIX)) {
                streamCount++;
                assertTrue("Persisted XStream is incorrect size." + fileName,
                        bufferSize == filesList[i].length());
            }
        }
        assertTrue("Incorrect number of XStreams persisted: " + streamCount,
                streamCount == 1);
        xset = system.openXSet(xuid, XSet.MODE_UNRESTRICTED);
                
        // TODO:  Read data back in
        for (int i = 0; i < streamCount; i++) {
            String fieldName = streamFieldNameBase;
            XStream xstream = xset.openXStream(fieldName, XStream.MODE_READ_ONLY);
            //assertTrue("Stream data read does not match data written",
            //         ReferenceTestUtilities.verifyStreamData(data, streamCount, byteChar)
            xstream.close();
        }
        xset.close();
        system.close();

    }
    
    
    /**
     * Deletes stream field before the XSet is committed.
     * 
     * @throws java.lang.Exception
     */
    public void testXStreamDeleteUncommitted() throws Exception {
        announce("ReferenceXStreamTest.testXStreamDeleteUncommitted");
        String streamField = "stream.field";
        String booleanField = "boolean.field";
        
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        
        // Delete a stream field before committing it to XSet
        XStream xstream = ReferenceTestUtilities.createXStream(
                xset, streamField, true, "image/jpg", (byte) 'a', 3000);
        xstream.close();
        
        xset.createProperty(booleanField, true, true);
         // Delete the stream field
        xset.deleteField(streamField);
                 
         XUID xuid = xset.commit();

        xset.close();
        xset = system.openXSet(xuid, XSet.MODE_UNRESTRICTED);
        try {
            xset.openXStream(streamField, XStream.MODE_READ_ONLY);
            fail("stream.field should not exist.");
        } catch (Exception ex) {
            /* announce("Expected exception, stream.field does not exist"); */
        }
        boolean value = xset.getBoolean(booleanField);
        assertTrue(booleanField + " value should be true", value == true);

        xset.close();
        system.close();

    }

   /**
     * Deletes stream field before the XSet is committed several times and then
     * creates a stream field of the same name and commits it. Test verifies
     * the correct data has been persisted.
     * 
     * @throws java.lang.Exception
     */
    public void testXStreamDeleteUncommittedMult() throws Exception {
        announce("ReferenceXStreamTest.testXStreamDeleteUncommittedMult");
        String streamField = "stream.field";
        String booleanField = "boolean.field";
        
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        
        // Delete a stream field before committing it to XSet
        XStream xstream = ReferenceTestUtilities.createXStream(
                xset, streamField, true, "image/jpg", (byte) 'a', 5000);
        xstream.close();
        
        xset.createProperty(booleanField, true, true);
         // Delete the stream field
        xset.deleteField(streamField);
                 
        // Delete a stream field before committing it to XSet
        xstream = ReferenceTestUtilities.createXStream(
                xset, streamField, true, "image/jpg", (byte) 'b', 4000);
        xstream.close();
         // Delete the stream field
        xset.deleteField(streamField);

         // Delete a stream field before committing it to XSet
        xstream = ReferenceTestUtilities.createXStream(
                xset, streamField, true, "image/jpg", (byte) 'c', 2000);
        xstream.close();
        XUID xuid = xset.commit();

        xset.close();
        xset = system.openXSet(xuid, XSet.MODE_UNRESTRICTED);
        try {
            byte[] expectedData = ReferenceTestUtilities.createByteArray(
                    (byte) 'c', 2000);
        assertTrue("Stream data read does not match data written",
                verifyStreamContents(expectedData, 2000, xset, streamField));
        } catch (Exception ex) {
            fail("Test failed, the " + streamField + " field does not exist");
        }
        boolean value = xset.getBoolean(booleanField);
        assertTrue(booleanField + " value should be true", value == true);

        xset.close();
        system.close();

    }

    /*
     * Tests deleting a binding stream field after it has been committed to
     * the XSet. So the original XSet will be unchanged and a new XSet (i.e.
     * with new XUID) will be created and it should NOT contain the stream
     * field.
     */
    public void testXStreamDelete() throws Exception {
        announce("ReferenceXStreamTest.testXStreamDelete");
        String streamField = "stream.field";
        int streamFieldSize = 3000;
        
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        
         XStream xstream = ReferenceTestUtilities.createXStream(
                 xset, streamField, true, "image/jpg", 
                 (byte) 'a', streamFieldSize);
        xstream.close();
        
        XUID xuid = xset.commit();

        xset.createProperty("boolean.field", true, true);
        // Delete a stream field after it was committed to the XSet
        // Because it's a binding field then the XUID should change and the
        // stream field should still exist in the original XSet
        xset.deleteField(streamField);
        XUID xuid2 = xset.commit();

        xset.close();
        xset = system.openXSet(xuid2, XSet.MODE_UNRESTRICTED);
        boolean value = xset.getBoolean("boolean.field");
        try {
            xset.openXStream("stream.field", XStream.MODE_READ_ONLY);
            fail("stream.field should not exist.");
        } catch (Exception ex) {
            /* announce("Expected exception, stream.field does not exist"); */
        }
        xset.close();
        system.close();
        String xuidString = null;
        try {
            XRI xri = XRI.valueOf(s_xri);
            String dirParam = 
                    xri.getParameterValue(XAMConstants.XRI_DIR_PARAM_NAME);
            if (dirParam == null || dirParam.length() == 0) {
                dirParam = System.getProperty("java.io.tmpdir");
            }
            xuidString = SniaXUID.convertToSafeString(xuid);
            ReferenceTestUtilities.verifyXSetStreams(dirParam,
                    xuid, 1, streamFieldSize, (byte) 'a');
            xuidString = SniaXUID.convertToSafeString(xuid2);
            ReferenceTestUtilities.verifyXSetStreams(dirParam,
                    xuid2, 0, streamFieldSize, (byte) 'a');
        } catch (Exception ex) {
            ex.printStackTrace();
            fail("For XSet_" + xuidString + "\n" + ex.getLocalizedMessage());
        }
    }
    
    /*
     * Tests deleting a stream field when it has open XStream on it.
     * 
     * So the original XSet will be unchanged and a new XSet (i.e.
     * with new XUID) will be created and it should NOT contain the stream
     * field.
     * 
     * 
     */
    public void testXStreamDeleteOpen() throws Exception {
        announce("ReferenceXStreamTest.testXStreamDeleteOpen");
        String streamField = "stream.field";
        int streamFieldSize = 3000;
        
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        
         XStream xstream = ReferenceTestUtilities.createXStream(
                 xset, streamField, true, "image/jpg", 
                 (byte) 'a', streamFieldSize);
        xstream.close();
        
        XUID xuid = xset.commit();

        xstream = xset.openXStream(streamField, XStream.MODE_READ_ONLY);
        // Should fail
        try {
            xset.deleteField(streamField);
            fail("Should not be able to delete stream field with open " +
                    " XStream instances");
        } catch (ObjectInUseException ex) {
            //expected exception
        }
        xstream.close();
        xset.deleteField(streamField);
        XUID xuid2 = xset.commit();

        xset.close();
        xset = system.openXSet(xuid2, XSet.MODE_UNRESTRICTED);
        try {
            xset.openXStream("stream.field", XStream.MODE_READ_ONLY);
            fail("stream.field should not exist.");
        } catch (Exception ex) {
            // expected exception
        }
        xset.close();
        system.close();
    }

    /**
     * Helper method. Opens the stream field for reading. Reads in the 
     * stream contents and compares them against the expected data.
     * 
     * @param expectedData
     * @param expectedNum
     * @param xset
     * @param fieldName
     * @return
     */
    private boolean verifyStreamContents(byte[] expectedData, 
            int expectedNum, XSet xset, String fieldName) {
        boolean validData = false;
        int buffSize = expectedData.length;
        byte[] buffer = new byte[buffSize];
        long offset = 0;
        long totalBytesRead = 0;
        XStream stream = null;
        try {
            stream = xset.openXStream(fieldName, XStream.MODE_READ_ONLY);
            long num = stream.read(buffer, offset, buffSize-offset);
            while (num > 0) {
                totalBytesRead += num;
                offset += num;
                num = stream.read(buffer, offset, buffSize-offset);
            }
            if (totalBytesRead == expectedNum) {
                validData = true;
                for (int i = 0; i < (int) totalBytesRead; i++) {
                    if (expectedData[i] != buffer[i]) {
                        validData = false;
                        break;
                    }
                }
            }
            
        } catch (XAMException xex) {
            validData = false;
            xex.printStackTrace();
        } finally {
            if (stream != null) {
                try {
                    stream.abandon();
                    stream.close();
                } catch (XAMException xex) {
                    validData = false;
                    xex.printStackTrace();
                }
            }
        }
        
        return validData;
    }
    
    /**
     * Tests allowed/disallowed method calls depending on open mode selected.
     * - Tests seek disallowed in write mode
     * - Tests can't open XStream if XSet closed
     * - Tests correct use of seek() method
     * - Tests abandoning XSet, should not be able to access XStream
     * - Tests write disallowed when XStream open in read only mode
     * - Tests read disallowed when XStream open in write mode
     * 
     * @throws java.lang.Exception
     */
    public void testXStreamAccessCheck() throws Exception {
        announce("ReferenceXStreamTest.testXStreamAccessCheck");
        long num = 0;

        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        
        XStream xstream = xset.createXStream("stream.field", true, "image/jpeg");
        byte[] buffer = new byte[100];
        for (int i = 0; i < buffer.length; i++) {
            buffer[i] = 'b';
        }
        long numBytes = xstream.write(buffer);
        
        // seek should be disallowed on write access
        try {
            xstream.seek(buffer.length/2, XStream.SEEK_SET);
            fail("Seek operation should be disallowed with Write access");
        } catch (XAMException ex) {
            //OK - ignore
        }
        xstream.close();
        XUID xuid = xset.commit();
        xset.close();
        
        // should not be able to open XStream when XSet is closed
        try {
            xset.openXStream("stream.field", XStream.MODE_READ_ONLY);
            fail("XSet is closed, should not be able to open the XStream.");
        } catch (XSetException ex) {
            // OK
        }

        xset = system.openXSet(xuid, XSet.MODE_UNRESTRICTED);
        XStream stream2 = xset.openXStream("stream.field", XStream.MODE_READ_ONLY);
        try {
            long position = stream2.tell();
            assertTrue("Should be at beginning of file. Current position: " +
                    position, position == 0);
        } catch (XAMException ex) {
            fail("Error getting current stream position. " +
                    ex.getMessage());
        }
        
        // Try seek
        byte[] buffer2 = new byte[100];
        try {
            long offset = numBytes/2;
            stream2.seek(offset, XStream.SEEK_SET);
            num = stream2.read(buffer2);
            StringBuffer charsRead = new StringBuffer(100);
            while (num > 0) {
                for (int i = 0; i < num; i++) {
                    charsRead.append((char)buffer2[i]);
                }
                num = stream2.read(buffer2);
            }
            assertTrue("Issued seek, should only read 1/2 of data. Num read: " 
                    + charsRead.length(), charsRead.length() == numBytes/2);
        } catch (XAMException ex) {
            fail("Error issuing seek command. " +
                    ex.getMessage());
        }
        
        // Verify can't access XStream after XSet has been abandoned
        xset.abandon();
        num = 0;
        try {
            num = stream2.read(buffer2);
            fail("XSet is abandon, should not be able to read from XStream");
        } catch (XAMException ex) {
            //Expected
        }
        stream2.close();
        xset.close();
        
        // Attempt write with read only access
        xset = system.openXSet(xuid, XSet.MODE_UNRESTRICTED);
        stream2 = xset.openXStream("stream.field", XStream.MODE_READ_ONLY);
        try {
            num = stream2.write(buffer);
            fail("XStream is open for read only access, should not be able to write");
        } catch (XAMException ex) {
            // OK
        }
        stream2.close();
        
        // Attempt read with write access
        stream2 = xset.openXStream("stream.field", XStream.MODE_WRITE_APPEND);
        try {
            num = stream2.read(buffer2);
            fail("XStream is open for write access, should not be able to read");
        } catch (XAMException ex) {
            // OK
        }
        stream2.close();
        xset.close();
        system.close();

	}
    
    /**
     * Tests that abandoning an XSet also abandons any open XStreams.
     * Tests that closing an abandon XSet closes any XStreams which are in the
     * abandon state.
     * @throws java.lang.Exception
     */
    public void testAbandonStreams() throws Exception {
        announce("ReferenceXStreamTest.testAbandonStreams");
        
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        XStream writeStream = null;
        XStream writeStream2 = null;
        XStream readStream = null;
        
        // Test abandon XSet with open streams
        try {
            xset.createProperty("test", false, true);
            // create and write to XStream
            writeStream = ReferenceTestUtilities.createXStream(
                    xset, "streamField", true, "image/jpeg", (byte) 'j', 500);
            writeStream.close();
            xset.commit();
            writeStream2 = ReferenceTestUtilities.createXStream(
                    xset, "streamField2", true, "image/jpeg", (byte) 'j', 500);
            // open a read XStream
            readStream = xset.openXStream("streamField", XStream.MODE_READ_ONLY);
        } catch (XAMException ex) {
            ex.printStackTrace();
            fail("Could not create XStream for field streamField " + ex);
         }

        try {
            xset.abandon();
            xset.close();
            system.close();
        } catch (XAMException x) {
            fail("Could not abandon or close XSet with open XStreams");
        }
 
    }
    
    /**
     * - Tests having read/write XStream instances open on the same stream field.
     *   The write stream is abandoned and the read stream should get set to
     *   the corrupt state.
     * - Tests calling abandon() on XStream instance and then closing the 
     *   XStreams and XSet.
     * 
     * - Tests abandoning XSystem with multiple open Stream instances
     * 
     * @throws java.lang.Exception
     */
    public void testAbandonStreamsAdvanced() throws Exception {
        announce("ReferenceXStreamTest.testAbandonStreamsAdvanced");
        
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        String fieldName = "streamField1";
        String fieldName2 = "streamField2";
        XStream writeStream = null;
        XStream writeStream2 = null;
        XStream readStream = null;
        XUID xuid = null;
        
        // Create xset with persisted stream field
        try {
            xset.createProperty("test", false, true);
            // create and write to XStream
            writeStream = ReferenceTestUtilities.createXStream(
                    xset, fieldName, true, "image/jpeg", (byte) 'j', 2000);
            writeStream2 = ReferenceTestUtilities.createXStream(
                    xset, fieldName2, true, "image/jpeg", (byte) 'j', 2000);
            writeStream.close();
            writeStream2.close();
            xuid = xset.commit();
        } catch (XAMException ex) {
            fail("Could not create XStream for field streamField and commit " + ex);
         }

        // Abandon xstream instance and attempt to write to same stream field
        // using a second instance
        try {
            int size = 250;
            byte[] addData = ReferenceTestUtilities.createByteArray(
                    (byte) 'a', size);
            byte[] readData = new byte[size];
            writeStream = xset.openXStream(fieldName, XStream.MODE_WRITE_APPEND);
            readStream = xset.openXStream(fieldName2, XStream.MODE_READ_ONLY);

            writeStream.write(addData);
            readStream.read(readData);
            writeStream.abandon();
            writeStream.write(addData);
            fail("Should not be able to write to stream in corrupt state.");
        } catch (InvalidArgumentException e) {
            fail("Unexpected error creating/opening multiple streams.");
        } catch (XAMException ex) {
            /* announce(
                    "Expected exception attempting to write to abandoned/corrupt stream."); */
        }
          
        // Abandon and close the XStreams and close the XSet
        try {
            writeStream.close();
            readStream.abandon();
            readStream.close();
            xset.close();      
        } catch (XAMException ex) {
            ex.printStackTrace();
            fail("Unable to abandon and close rest of XStreams and close XSet");
        }

        // XSystem abandon with multiple open streams
        try {
            int size = 250;
            byte[] addData = ReferenceTestUtilities.createByteArray(
                    (byte) 'a', size);
            byte[] readData = new byte[size];
            system = connectAndAuthenticate(true);
            xset = system.openXSet(xuid, XSet.MODE_UNRESTRICTED);
            
            writeStream = xset.openXStream(fieldName, XStream.MODE_WRITE_APPEND);
            writeStream2 = ReferenceTestUtilities.createXStream(
                    xset, "fieldName2a", true, "image/jpeg", (byte) 'j', 2000);
            readStream = xset.openXStream(fieldName2, XStream.MODE_READ_ONLY);
            
            writeStream.write(addData);
            long numBytes = readStream.read(readData);
            writeStream2.write(addData);
            system.abandon();
            system.close();
        } catch (InvalidArgumentException e) {
            fail("Unexpected error creating multiple streams for test.");
            
        } catch (XAMException x) {
            x.printStackTrace();
            fail("Test abandoning and closing XSystem with open streams failed.");
        }
    }
    /**
     * Tests 
     * - Creating an XStream instance (i.e. a stream field) 
     * - Writing to the stream
     * - Getting current position using .tell();
     * - Closes the stream and commits the XSet
     * 
     * @throws java.lang.Exception
     */
    public void testXStreamAccessAllowed() throws Exception {
        announce("ReferenceXStreamTest.testXStreamAccessAllowed");

        String fieldName = "stream.field";
        String fieldName2 = "stream.field2";
        
        XSystem system = connectAndAuthenticate(true);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        FieldContainer fc = (FieldContainer) xset;
        
        XStream xstream = xset.createXStream(
                fieldName, true, "image/jpeg");
        byte[] buffer = new byte[100];
        for (int i = 0; i < buffer.length; i++) {
            buffer[i] = 'b';
        }
        xstream.write(buffer);
        long position = xstream.tell();
        assertTrue( "Current file position is not at the end. Position: " +
                position + " ", 
                    position == buffer.length);
        
        xstream.close();
        XUID xuid = xset.commit();
        
        // open multiple XStreams for read access
        XStream xstream2 = null;
        try {
            xstream = xset.openXStream(fieldName, XStream.MODE_READ_ONLY);
            xstream2 = 
                    xset.openXStream(fieldName, XStream.MODE_READ_ONLY);
        } catch (XAMException ex) {
            fail("Should be able to open multiple XStreams for reading.");
        }
        
        // try opening write mode XStream when readers present - should fail
        XStream xstream3 = null;
        try {
            xstream3 = xset.openXStream(fieldName, XStream.MODE_WRITE_APPEND);
            fail("Should NOT be able to open write append XStream when " +
                    "readers present.");
        } catch (InvalidArgumentException ex) {
            // expected exception
        }
        
        try {
            xstream3 = xset.openXStream(fieldName, XStream.MODE_WRITE_TRUNCATE);
            fail("Should NOT be able to open write truncate XStream when " +
                    "readers present.");
        } catch (InvalidArgumentException ex) {
            // expected exception
        }
        xstream.close();
        xstream2.close();
        
        // Open writer first and then try opening reader XStream - should fail
        xstream = xset.openXStream(fieldName, XStream.MODE_WRITE_TRUNCATE);
        try {
            xstream2 = xset.openXStream(fieldName, XStream.MODE_READ_ONLY);
            fail("Should NOT be able to open read mode XStream when " +
                    "writer is present.");            
        } catch (InvalidArgumentException ex) {
            // expected 
        }
        
       // Create another stream field, uncommitted, and test trying to open
       // a reader
       xstream2 = xset.createXStream(fieldName2, true, "image/jpeg");
         try {
            xstream3 = xset.openXStream(fieldName, XStream.MODE_READ_ONLY);
            fail("Should NOT be able to open read mode XStream when " +
                    "writer is present (create).");            
        } catch (InvalidArgumentException ex) {
            // expected 
        }
        
        // Close first XStream and reopen with multiple readers
        xstream.close();
        try {
            xstream = xset.openXStream(fieldName, XStream.MODE_READ_ONLY);
            XStream xstream4 = xset.openXStream(fieldName, XStream.MODE_READ_ONLY);
            xstream.close();
            xstream4.close();
        } catch (XAMException ex) {
            fail("Should be able to open multiple readers on a different stream field.");
        }
        xstream2.close();
        xset.close();
        system.close();

	}
    
    /**
     * Tests to make sure that the MIME validation will catch ill formed MIME types. NOTE: It is not the 
     * purpose of validation to make sure that the MIME content type matches the bits in an XStream.
     * @throws Exception
     */
    public void testMimeValidation() throws Exception
    {
       final String FIELD_NAME = "org.snia.xam.refvim.unit.test.field";
       XSystem system = connectAndAuthenticate(true);
       XSet    xs     = system.createXSet( XSet.MODE_UNRESTRICTED );
       String  mime   = "";
       
       try { 
          mime = "image//jpeg";
          xs.createXStream(FIELD_NAME, false, mime); 
          fail( "Mime validation should have failed for: " + mime );
       }
       catch( InvalidFieldTypeException ifte ) { /* expected, so ignore it */ }
       
       try { 
          mime = "image";
          xs.createXStream(FIELD_NAME, false, mime); 
          fail( "Mime validation should have failed for: " + mime );
       }
       catch( InvalidFieldTypeException ifte ) { /* expected, so ignore it */ }
       
       try { 
          mime = "text/plain;charset";
          xs.createXStream(FIELD_NAME, false, mime); 
          fail( "Mime validation should have failed for: " + mime );
       }
       catch( InvalidFieldTypeException ifte ) { /* expected, so ignore it */ }
       
       try { 
          mime = "image/jpeg;charset==";
          xs.createXStream(FIELD_NAME, false, mime); 
          fail( "Mime validation should have failed for: " + mime );
       }
       catch( InvalidFieldTypeException ifte ) { /* expected, so ignore it */ }
              
       system.abandon();
       system.close();
    }
    
}
