/*
 *  Copyright (c) 2008-2009, Sun Microsystems, Inc.
 *  Copyright (c) 2008-2009, The Storage Networking Industry Association.
 * 
 *  Redistribution and use in source and binary forms, with or without modification, 
 *  are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *  *  Neither the name of The Storage Networking Industry Association (SNIA) nor 
 *     the names of its contributors may be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.ByteArrayOutputStream;
import java.net.InetAddress;
import java.util.Calendar;
import org.snia.xam.XAMException;
import org.snia.xam.XSystemException;
import org.snia.xam.base.Base64;

import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.FieldExistsException;
import org.snia.xam.PolicyNameException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XIterator;
import org.snia.xam.XSet;
import org.snia.xam.XSetInaccessibleException;
import org.snia.xam.XStream;
import org.snia.xam.XSystem;
import org.snia.xam.XUID;
import org.snia.xam.base.XRI;
import org.snia.xam.testvim.TestXUID;
import org.snia.xam.util.ISO8601Date;
import org.snia.xam.util.SASLUtils;

import junit.framework.Test;
import junit.framework.TestSuite;

/**
 *  Test functionality present in the ReferenceXSystem. 
 *  
 *  NOTE: Hold and Release are not tested here, @see LitigationHoldTest
 */
public class ReferenceXSystemTest extends ReferenceXAMTest {
   private static final String VALID_XUID[] = {
      // XUID values donated by Hewlett Packard Co.
      "AAAACwAb25cAAHhzZXQxMjE2MjI5ODgwNTAz",
      "AAAACwAb6dYAAHhzZXQxMjE2MjI5ODgwNTM0",
      "AAAACwAbitQAAHhzZXQxMjE2MjI5ODgwNTUw",
      "AAAACwAbjOQAAHhzZXQxMjE2MjI5ODgwNjU5",
      "AAAACwAb6eUAAHhzZXQxMjE2MjI5ODgwNjc1",
      "AAAACwAbiiEAAHhzZXQxMjE2MjI5ODgwNjkw",
      "AAAACwAbGPYAAHhzZXQxMjE2MjI5ODgwNzA2",
      "AAAACwAbGUYAAHhzZXQxMjE2MjI5ODgwODAw",
      "AAAACwAbiocAAHhzZXQxMjE2MjI5ODgwODE1",
      "AAAACwAbKYcAAHhzZXQxMjE2MjI5ODgwODMx",
   
      // XUID values from org.snia.xam.testvim.TestXUID
      "AAA6AwA0ymgwMDUwNTZDMDAwMDgtMjAwOC0wNy0xOFQxMV81OF8zMl80ODMtMDhfMDAtYQ==",
      "AAA6AwA0RmowMDUwNTZDMDAwMDgtMjAwOC0wNy0xOFQxMV81OV8wMl85MTUtMDhfMDAtMQ==",
      "AAA6AwA0RyowMDUwNTZDMDAwMDgtMjAwOC0wNy0xOFQxMV81OV8wMl85MTUtMDhfMDAtMg==",
      "AAA6AwA0h+swMDUwNTZDMDAwMDgtMjAwOC0wNy0xOFQxMV81OV8wMl85MTUtMDhfMDAtMw==",
      "AAA6AwA0RaowMDUwNTZDMDAwMDgtMjAwOC0wNy0xOFQxMV81OV8wMl85MTUtMDhfMDAtNA==",
      "AAA6AwA0hWswMDUwNTZDMDAwMDgtMjAwOC0wNy0xOFQxMV81OV8wMl85MTUtMDhfMDAtNQ==",
      "AAA6AwA0hCswMDUwNTZDMDAwMDgtMjAwOC0wNy0xOFQxMV81OV8wMl85MTUtMDhfMDAtNg==",
      "AAA6AwA03nQwMDUwNTZDMDAwMDgtMjAwOC0wNy0xOFQxMV81OV8wMl85MzAtMDhfMDAtNw==",
      "AAA6AwA02jQwMDUwNTZDMDAwMDgtMjAwOC0wNy0xOFQxMV81OV8wMl85MzAtMDhfMDAtOA==",
      "AAA6AwA0GvUwMDUwNTZDMDAwMDgtMjAwOC0wNy0xOFQxMV81OV8wMl85MzAtMDhfMDAtOQ==",
      "AAA6AwA04PQwMDUwNTZDMDAwMDgtMjAwOC0wNy0xOFQxMV81OV8wMl85MzAtMDhfMDAtYQ==",

      // XUID Values donated by EMC
      "AAAEcwA9sb8wTUdOTjE2ME1LRktUZUVKVVBQOEpFUEYwOUNHNDEzRjE1VjkzRTFKRlBGQkU4U1FLM040SQ==",
      "AAAEcwA9419GVk1LQ05PNlRPQlE0ZTVJUTc1UzNNSlQ2Q0pHNDEzRjE1VkcySDFLNjFHMFFUTVRLOFNWNw==",
      "AAAEcwA9PyM3TU5GTFFOMFNBSERGZUNTM09BVDExUjJERk9HNDEzRjE1VksxNzE2RkNLS0cxUUdLSlRRVg==",
      "AAAEcwA9S5NGNzYzRTJCUzZLT0xDZUJLSjFLMDVHRzhLOFVHNDEzRjE1Vk8xQjFLQTVCS0JSUFRJT01RTg==",
      "AAAEcwA9EF1GOUI4SjdTNERLMTlWZTBNREFVQ0lHSUk3MkJHNDEzRjE2MDIzRjE5Qk9PSE1DNk1CNTlSNQ=="
   };
   
    public ReferenceXSystemTest(String name) {
        super(name);
    }
    public static void main(String[] args)
    {
        System.exit(runTest(args, suite()));
    }
    public static Test suite()
    {
    	TestSuite suite = new TestSuite();
    	suite.addTest(new ReferenceXSystemTest("testConnect"));
    	suite.addTest(new ReferenceXSystemTest("testDefaultSASL"));
        suite.addTest(new ReferenceXSystemTest("testLoadDefaults"));
        suite.addTest(new ReferenceXSystemTest("testCreateXUID"));
        suite.addTest(new ReferenceXSystemTest("testCreateForeignXUID"));
        suite.addTest(new ReferenceXSystemTest("testXSet"));
        suite.addTest(new ReferenceXSystemTest("testCopyXSet"));
        suite.addTest(new ReferenceXSystemTest("testGetXSetAccessTime"));
        suite.addTest(new ReferenceXSystemTest("testAccessXSet"));
        suite.addTest(new ReferenceXSystemTest("testManagementPolicy"));
    	return suite;
    }

    /**
     * Test of loadDefaults method, of class ReferenceXSystem.
     */
    public void testLoadDefaults() throws Exception {
        announce("ReferenceXSystemTest.loadDefaults");
        ReferenceXSystem instance = new ReferenceXSystem();
        instance.loadDefaults();
        assert(instance.getBoolean(XSystem.XAM_XSYSTEM_JOB_QUERY_LEVEL1_SUPPORTED) == true);
        assertNotNull(instance.getString(XSystem.XAM_XSYSTEM_IDENTITY));
    }
    
    /**
     * Test of connect method, of class ReferenceXSystem.
     * NOTE: Because this is done outside of the XAM_Library, a portion of the initialization
     * protocol is duplicated here, mainly so logging will work.
     */
    public void testConnect() throws Exception {
        announce("ReferenceXSystemTest.connect");
        XRI xri = new XRI();
        xri.setVimInfo(TestConstants.VIM_NAME);
        String ip = InetAddress.getLocalHost().getHostAddress();
        xri.setSystem(ip);
        XSystem xsystem = new ReferenceXSystem();
        
        // Fake initialization protocol
        String  path      = s_xam.getString(XAMLibrary.XAM_LOG_PATH);
        long    size      = s_xam.getLong(XAMLibrary.XAM_LOG_MAX_SIZE);
        boolean append    = s_xam.getBoolean(XAMLibrary.XAM_LOG_APPEND);
        long    verbosity = s_xam.getLong(XAMLibrary.XAM_LOG_VERBOSITY);
        long    level     = s_xam.getLong(XAMLibrary.XAM_LOG_LEVEL);
        try { xsystem.createProperty(XAMLibrary.XAM_LOG_PATH, false, path ); } 
        catch( FieldExistsException fe ) { xsystem.setProperty( XAMLibrary.XAM_LOG_PATH, path ); }
        
        try { xsystem.createProperty(XAMLibrary.XAM_LOG_MAX_SIZE, false, size ); }
        catch( FieldExistsException fe ) {xsystem.setProperty( XAMLibrary.XAM_LOG_MAX_SIZE, size ); }
        
        try { xsystem.createProperty(XAMLibrary.XAM_LOG_APPEND, false, append ); }
        catch( FieldExistsException fe ) {xsystem.setProperty( XAMLibrary.XAM_LOG_APPEND, append ); }
        
        try { xsystem.createProperty(XAMLibrary.XAM_LOG_VERBOSITY, false, verbosity ); }
        catch( FieldExistsException fe ) {xsystem.setProperty( XAMLibrary.XAM_LOG_VERBOSITY, verbosity ); }
        
        try { xsystem.createProperty(XAMLibrary.XAM_LOG_LEVEL, false, level ); }
        catch( FieldExistsException fe ) {xsystem.setProperty( XAMLibrary.XAM_LOG_LEVEL, level ); }
        
        xsystem.deleteField(XSystem.XAM_XSYSTEM_INITIALIZING); 
         
        xsystem.connect(xri.toString());
        xsystem.close();
    }
        
    /**
     * Test to see if reference VIM will work with default SASL PLAIN.
     */
    public void testDefaultSASL() throws Exception {
        announce("ReferenceXSystemTest.testDefaultSASL");
        XSystem xsys = s_xam.connect(s_xri);
        String creds = "\000testuser\000testpasswd";
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        
        xsys.authenticate(creds.getBytes(), baos );
        
        String result = new String(baos.toByteArray());
        assertEquals( "Expected OK as reult from authtenticate", "DONE", result );
        xsys.close();
        xsys = null;
        
        // See if authentication can be done in a single buffer, with mechanism
        XSystem xsys2 = s_xam.connect(s_xri);
        String creds2 = "PLAIN\000\000testuser\000testpasswd";
        ByteArrayOutputStream baos2 = new ByteArrayOutputStream();     
        xsys2.authenticate(creds2.getBytes(), baos2 );
        String result2 = new String(baos2.toByteArray());
        assertEquals( "Expected OK as reult from authtenticate", "DONE", result2 );
        baos2 = null;
        xsys2.close();
        xsys2 = null;
        
        // Compatibility with older buffers, missing a required \000 after the mechanism
        XSystem xsys3 = s_xam.connect(s_xri);
        String creds3 = "PLAIN\000testuser\000testpasswd";
        ByteArrayOutputStream baos3 = new ByteArrayOutputStream();     
        xsys3.authenticate(creds3.getBytes(), baos3 );
        String result3 = new String(baos3.toByteArray());
        assertEquals( "Expected OK as reult from authtenticate", "DONE", result3 );
        xsys3.close();
        xsys3 = null;
        
        // Test with authorization ID in the string
        XSystem xsys4 = s_xam.connect(s_xri);
        String creds4 = "PLAIN\000authzID\000testuser\000testpasswd";
        ByteArrayOutputStream baos4 = new ByteArrayOutputStream();     
        xsys4.authenticate(creds4.getBytes(), baos4 );
        String result4 = new String(baos4.toByteArray());
        assertEquals( "Expected OK as reult from authtenticate", "DONE", result4 );
        baos4 = null;
        xsys4.close();
        xsys4 = null;

        // Test with authorization ID in the string, using default mechanism
        XSystem xsys5 = s_xam.connect(s_xri);
        String creds5 = "authzID\000testuser\000testpasswd";
        ByteArrayOutputStream baos5 = new ByteArrayOutputStream();     
        xsys5.authenticate(creds5.getBytes(), baos5 );
        String result5 = new String(baos5.toByteArray());
        assertEquals( "Expected OK as reult from authtenticate", "DONE", result5 );
        baos5 = null;
        xsys5.close();
        xsys5 = null;
    }

    /**
     * Basic XSet test: create, open and delete
     * @throws java.lang.Exception
     */
    public void testXSet() throws Exception {
        announce("ReferenceXSystemTest.testXSet");
        XSystem system = doConnection(true);
        doAuthentication(system);
        XSet xset = system.createXSet(XSet.MODE_UNRESTRICTED);
        XUID xuid = xset.commit();
        xset.close();

        xset = system.openXSet(xuid, XSet.MODE_READ_ONLY);
        xset.close();
        system.close();

        try {
            xset = system.createXSet(XSet.MODE_UNRESTRICTED);
            fail("XSet creation should have failed, the system is closed");
        } catch (XSystemException x) {
        }
    }
    
    /**
     * Basic XSet test: create, open and delete
     * @throws java.lang.Exception
     */
    public void testCreateXUID() throws Exception {
        announce("ReferenceXSystemTest.testCreateXUID");
        XSystem system = doConnection(true);
        doAuthentication(system);
        XUID refVIMXuid = SniaXUID.generateXUID();
        
        // Test Reference VIM Generate XUID
        checkXUID(system, refVIMXuid, "Testing RefVIM XUID:");
        
        // Test "foreign" XUID - should be able to except non-Ref VIM XUIDs
        TestXUID foreignXuid = TestXUID.newXUID();
        checkXUID(system, foreignXuid, "Testing Foreign XUID: ");
        system.close();

    }

    /**
     * Basic XSet test: create, open and delete
     * @throws java.lang.Exception
     */
    public void testCreateForeignXUID() throws Exception {
        announce("ReferenceXSystemTest.testCreateForeignXUID");
        XSystem system = doConnection(true);
        doAuthentication(system);
        for (String stringXUID : VALID_XUID) {
            XUID xuidFromString = system.createXUID(stringXUID);
            byte[] byteXUID = xuidFromString.toBytes();
            XUID xuidFromBytes = system.createXUID(byteXUID);
            assertTrue("Test failed. XUIDs created from string and byte[] " +
                    "should be equal.", xuidFromBytes.equals(xuidFromString));
            checkXUID(system, xuidFromString, 
                    "Testing Foreign XUID from String: ");
            checkXUID(system, xuidFromBytes,
                    "Testing Foreign XUID from byte[]: ");
            
        }
        system.close();
    }

    /**
     * Verifies that the XUID can be correctly accepted by the createXUID
     * methods and returns the same XUID.
     * 
     * @param system
     * @param xuid
     * @param failMsgPrefix 
     * @throws org.snia.xam.XAMException
     */
    private void checkXUID(XSystem system, XUID xuid, String failMsgPrefix)
            throws XAMException {
        
        // First test that encoding and decoding XUID returns the same info
        byte[] xsetXUIDBytes = xuid.toBytes();
        String xsetXUIDString = 
                Base64.base64Encode(xsetXUIDBytes);
        byte[] checkDecode = Base64.base64Decode(xsetXUIDString);
        if (checkDecode.length != xsetXUIDBytes.length) {
            fail(failMsgPrefix + " Encode/Decode does not generate the same size byte array.");
        }
        
        for (int i = 0; i < checkDecode.length; i++) {
            if (checkDecode[i] != xsetXUIDBytes[i]) {
                fail(failMsgPrefix + " Encode/Decode does not generate the same array.");
            }
        }
        XUID xuidBytes = system.createXUID(xsetXUIDBytes);
        XUID xuidString = system.createXUID(xsetXUIDString);
               
        if (!xuid.equals(xuidBytes)) {
            fail(failMsgPrefix + " XUID created using createXUID(byte[]) call does not match original XUID.\n" +
                    "Start XUID:  " + xuid.toString() + "\nCreate XUID: " + xuidBytes.toString());
        }
        if (!xuid.equals(xuidString)) {
            fail(failMsgPrefix + " XUID created using createXUID(String) call does not match original XUID.\n" +
                   "Start XUID:  " + xuid.toString() + "\nCreate XUID: " + xuidString.toString());
        }
        
    }
    
    /**
     * Test of asyncOpenXStream method, of class ReferenceXSystem.
     */
    public void TODOtestAsyncOpenXStream() throws Exception {
        announce("ReferenceXSystemTest.asyncOpenXStream");
        String name = "";
        String mode = "";
        long xopid = 0L;
        XAsyncListener listener = null;
        ReferenceXSystem instance = new ReferenceXSystem();
        XAsync expResult = null;
        XAsync result = instance.asyncOpenXStream(name, mode, xopid, listener);
        assertEquals(expResult, result);
        // TODO review the generated test code and remove the default call to fail.
        fail("The test case is a prototype.");
    }

    /**
     * Test of createXStream method, of class ReferenceXSystem.
     */
    public void TODOtestCreateXStream() throws Exception {
        announce("ReferenceXSystemTest.createXStream");
        String fieldName = "";
        boolean binding = false;
        String mimeType = "";
        ReferenceXSystem instance = new ReferenceXSystem();
        XStream expResult = null;
        XStream result = instance.createXStream(fieldName, binding, mimeType);
        assertEquals(expResult, result);
        // TODO review the generated test code and remove the default call to fail.
        fail("The test case is a prototype.");
    }

    /**
     * Test of openXStream method, of class ReferenceXSystem.
     */
    public void TODOtestOpenXStream() throws Exception {
        announce("ReferenceXSystemTest.openXStream");
        String fieldName = "";
        String mode = "";
        ReferenceXSystem instance = new ReferenceXSystem();
        XStream expResult = null;
        XStream result = instance.openXStream(fieldName, mode);
        assertEquals(expResult, result);
        // TODO review the generated test code and remove the default call to fail.
        fail("The test case is a prototype.");
    }

    /**
     * Test of abandon method, of class ReferenceXSystem.
     */
    public void TODOtestAbandon() throws Exception {
        announce("ReferenceXSystemTest.abandon");
        ReferenceXSystem instance = new ReferenceXSystem();
        instance.abandon();
        // TODO review the generated test code and remove the default call to fail.
        fail("The test case is a prototype.");
    }

    /**
     * Test of accessXSet method, of class ReferenceXSystem.
     */
    public void testAccessXSet() throws Exception {
        announce("ReferenceXSystemTest.accessXSet");
        XSystem xsys = this.connectAndAuthenticate(false);
        XSet    xset = xsys.createXSet( XSet.MODE_UNRESTRICTED );
        xset.createProperty( "com.example.prop1", false, "testing-accessXSet" );
        XUID xsetid = xset.commit();
        
        // Normal user checks.
        assertTrue( "Expect to be able to read owners XSet", 
           xsys.accessXSet( xsetid, XSystem.ACCESS_READ_OK ));
        
        assertTrue( "Expect to be able to delete owners XSet", 
           xsys.accessXSet( xsetid, XSystem.ACCESS_DELETE_OK ));
        
        assertTrue( "Expect to be able to write application owners XSet",
           xsys.accessXSet(xsetid, XSystem.ACCESS_WRITE_APPLICATION_OK ));

        assertTrue( "Expect to be able to write system owners XSet",
           xsys.accessXSet(xsetid, XSystem.ACCESS_WRITE_SYSTEM_OK ));

        assertTrue( "Expect to be able to hold owners XSet",
           xsys.accessXSet(xsetid, XSystem.ACCESS_HOLD_OK ));

        assertTrue( "Expect to be able to set retention event on owners XSet",
           xsys.accessXSet(xsetid, XSystem.ACCESS_RETENTION_EVENT_OK ));

        assertTrue( "Expect to be able to submit owners XSet as job",
           xsys.accessXSet(xsetid, XSystem.ACCESS_JOB_OK ));
        
        assertTrue( "Expect to be able to commit owners job XSet",
           xsys.accessXSet(xsetid, XSystem.ACCESS_JOB_COMMIT_OK ));
        
        assertTrue( "Expect to be able to create an XSet",
           xsys.accessXSet(xsetid, XSystem.ACCESS_CREATE_OK ));
        
        // Now try the anonymous
        XSystem xsys2 = this.getAnonymousSystem();
        assertTrue( "Expect to be able to anonymously read owners XSet", 
           xsys2.accessXSet( xsetid, XSystem.ACCESS_READ_OK ));
        
        assertFalse( "Expect to be denined delete owners XSet", 
           xsys2.accessXSet( xsetid, XSystem.ACCESS_DELETE_OK ));
        
        assertFalse( "Expect to be denied write application owners XSet",
           xsys2.accessXSet(xsetid, XSystem.ACCESS_WRITE_APPLICATION_OK ));

        assertFalse( "Expect to be denied write system owners XSet",
           xsys2.accessXSet(xsetid, XSystem.ACCESS_WRITE_SYSTEM_OK ));

        assertFalse( "Expect to be denied hold owners XSet",
           xsys2.accessXSet(xsetid, XSystem.ACCESS_HOLD_OK ));

        assertFalse( "Expect to be denied set retention event on owners XSet",
           xsys2.accessXSet(xsetid, XSystem.ACCESS_RETENTION_EVENT_OK ));

        assertTrue( "Expect to be able to anonymously submit owners XSet as job",
           xsys2.accessXSet(xsetid, XSystem.ACCESS_JOB_OK ));
        
        assertFalse( "Expect to be denied commit owners job XSet",
           xsys2.accessXSet(xsetid, XSystem.ACCESS_JOB_COMMIT_OK ));
        
        assertFalse( "Expect to be denied create an XSet",
           xsys2.accessXSet(xsetid, XSystem.ACCESS_CREATE_OK ));
        
        xsys2.abandon();
        xsys2.close();

        // Finally, set AccessPolicy on the xset and make sure that another user gets the proper exception
        xset.applyAccessPolicy(false, "testuser");
        xset.commit();
        xset.close();
        
        XSystem xsys3 = this.getAuthenticatedXSystem(PASSWD_INFO[0][0], PASSWD_INFO[0][1] );
        try { 
           xsys3.accessXSet(xsetid, XSystem.ACCESS_DELETE_OK);
           fail( "XSystem.accessXSet should have thrown XSetInaccessibleException");
        }
        catch( XSetInaccessibleException xie ) { /* expected */ }
        xsys.abandon();
        xsys.close();    
        xsys3.abandon();
        xsys3.close();
    }

    /**
     * Test of asyncCopyXSet method, of class ReferenceXSystem.
     */
    public void TODOtestAsyncCopyXSet() throws Exception {
        announce("ReferenceXSystemTest.asyncCopyXSet");
        XUID inXuid = null;
        String mode = "";
        long xopid = 0L;
        XAsyncListener listener = null;
        ReferenceXSystem instance = new ReferenceXSystem();
        XAsync expResult = null;
        XAsync result = instance.asyncCopyXSet(inXuid, mode, xopid, listener);
        assertEquals(expResult, result);
        // TODO review the generated test code and remove the default call to fail.
        fail("The test case is a prototype.");
    }

    /**
     * Test of asyncOpenXSet method, of class ReferenceXSystem.
     */
    public void TODOtestAsyncOpenXSet() throws Exception {
        announce("ReferenceXSystemTest.asyncOpenXSet");
        XUID inXuid = null;
        String mode = "";
        long xopid = 0L;
        XAsyncListener listener = null;
        ReferenceXSystem instance = new ReferenceXSystem();
        XAsync expResult = null;
        XAsync result = instance.asyncOpenXSet(inXuid, mode, xopid, listener);
        assertEquals(expResult, result);
        // TODO review the generated test code and remove the default call to fail.
        fail("The test case is a prototype.");
    }

    /**
     * Test of authenticate method, of class ReferenceXSystem.
     */
    public void TODOtestAuthenticate() throws Exception {
        announce("ReferenceXSystemTest.authenticate");
        byte[] buffer = null;
        ReferenceXSystem instance = new ReferenceXSystem();
        byte[] expResult = null;
        byte[] result = instance.authenticate(buffer);
        assertEquals(expResult, result);
        // TODO review the generated test code and remove the default call to fail.
        fail("The test case is a prototype.");
    }

    /**
     * Test of setProperty method, of class ReferenceXSystem.
     */
    public void TODOtestSetProperty() throws Exception {
        announce("ReferenceXSystemTest.setProperty");
        String fieldName = "";
        boolean value = false;
        ReferenceXSystem instance = new ReferenceXSystem();
        instance.setProperty(fieldName, value);
        // TODO review the generated test code and remove the default call to fail.
        fail("The test case is a prototype.");
    }

    /**
     * Test of close method, of class ReferenceXSystem.
     */
    public void TODOtestClose() throws Exception {
        announce("ReferenceXSystemTest.close");
        ReferenceXSystem instance = new ReferenceXSystem();
        instance.close();
        // TODO review the generated test code and remove the default call to fail.
        fail("The test case is a prototype.");
    }
   

    /**
     * Test of copyXSet method, of class ReferenceXSystem.
     */
    public void testCopyXSet() throws Exception {
       final String PROP1 = "com.example.prop1";
       final String VALUE1 = "value1";
       
        announce("ReferenceXSystemTest.copyXSet");
        String copyMode = XSet.MODE_UNRESTRICTED;
        XSystem xsys = this.connectAndAuthenticate(false);
        XSet firstXSet = xsys.createXSet(XSet.MODE_UNRESTRICTED);
        firstXSet.createProperty( PROP1, false, VALUE1 );
        XUID xsetid = firstXSet.commit();
        
        XSet result = xsys.copyXSet(xsetid, copyMode);
        // Make sure that all the fields in the copy are in the source.
        int count = 0;
        XIterator iter = result.openFieldIterator("");
        while( iter.hasNext() ) {
           String field_name = (String) iter.next();
           if( field_name.equals( XSet.XAM_DIRTY ))
              continue;
           
           count++;
          
           assertEquals( "Copy XSet mismatch readonly atribute between original and copy, ",
              firstXSet.getFieldReadOnly( field_name ),
              result.getFieldReadOnly( field_name ) );
           assertEquals( "Copy XSet mismatch binding atribute between original and copy, ",
              firstXSet.getFieldBinding( field_name),
              result.getFieldBinding( field_name ) );
           assertEquals( "Copy XSet mismatch type atribute between original and copy, ",
              firstXSet.getFieldType( field_name),
              result.getFieldType( field_name ) );
           assertEquals( "Copy XSet mismatch length atribute between original and copy, ",
              firstXSet.getFieldLength( field_name),
              result.getFieldLength( field_name ) );
        }

        // Make sure that copy removes the field it's supposed to.
        // TODO: Add a value test between source and copy.
        try{ 
           result.getFieldReadOnly( XSet.XAM_TIME_XUID); 
           fail( "Field should have been removed from copy of XSet" + XSet.XAM_TIME_XUID);
        }
        catch( FieldDoesNotExistException fdne ) { } 

        try{ 
           result.getFieldReadOnly( XSet.XAM_TIME_COMMIT); 
           fail( "Field should have been removed from copy of XSet" + XSet.XAM_TIME_COMMIT);
        }
        catch( FieldDoesNotExistException fdne ) { } 

        try{ 
           result.getFieldReadOnly( XSet.XAM_TIME_ACCESS); 
           fail( "Field should have been removed from copy of XSet" + XSet.XAM_TIME_ACCESS);
        }
        catch( FieldDoesNotExistException fdne ) { } 

        try{ 
           result.getFieldReadOnly( XSet.XAM_TIME_RESIDENCY); 
           fail( "Field should have been removed from copy of XSet" + XSet.XAM_TIME_RESIDENCY);
        }
        catch( FieldDoesNotExistException fdne ) { } 
        
        // Check sum, make sure the number of fields in the source match copy 
        // (less the previous 4, and .xset.dirty which is not in the source XSet)
        iter = firstXSet.openFieldIterator("");
        int sourceCount = 0;
        while( iter.hasNext() ) {
           sourceCount++;
           iter.next();
        }
        
        assertEquals( "Copy XSet doesn't have the correct number of fields, ", 
           (sourceCount - 5), count );
        
        // Make sure this commits the proper way.
        XUID resultXuid = result.commit();
        if( xsetid.equals(resultXuid ))
           fail( "New XUID from copy should not be the same as the original" );
        
        // Last checksum, committing the copy should have restored the time properties
        iter = result.openFieldIterator("");
        count = 0;
        while( iter.hasNext() ) {
           count++;
           iter.next();
        }
        assertEquals( "Field count on source and copied XSet are not the same", sourceCount, count );
    }

    /**
     * Test of createXSet method, of class ReferenceXSystem.
     */
    public void TODOtestCreateXSet() throws Exception {
        announce("ReferenceXSystemTest.createXSet");
        String createMode = "";
        ReferenceXSystem instance = new ReferenceXSystem();
        XSet expResult = null;
        XSet result = instance.createXSet(createMode);
        assertEquals(expResult, result);
        // TODO review the generated test code and remove the default call to fail.
        fail("The test case is a prototype.");
    }

    /**
     * Test of deleteXSet method, of class ReferenceXSystem.
     */
    public void TODOtestDeleteXSet() throws Exception {
        announce("ReferenceXSystemTest.deleteXSet");
        XUID xsetid = null;
        ReferenceXSystem instance = new ReferenceXSystem();
        instance.deleteXSet(xsetid);
        // TODO review the generated test code and remove the default call to fail.
        fail("The test case is a prototype.");
    }

    /**
     * Test of getXSetAccessTime method, of class ReferenceXSystem.
     */
    public void testGetXSetAccessTime() throws Exception {
        announce("ReferenceXSystemTest.getXSetAccessTime");

        XSystem xsys = this.connectAndAuthenticate(false);
        XSet xset = xsys.createXSet(XSet.MODE_UNRESTRICTED);
        xset.createProperty( "com.example.prop1", false, "testing-getAccessTime" );
        xset.applyAccessPolicy(false, "testuser");
        XUID xuid = xset.commit();
       
        Calendar commitTime = xset.getDateTime(XSet.XAM_TIME_COMMIT);
        String   commitTStr = ISO8601Date.toISO8601String(commitTime);
        
        Calendar result     = xsys.getXSetAccessTime(xuid);
        String   resultTStr = ISO8601Date.toISO8601String(result);
        
        assertEquals( "Access time incorrect", commitTStr, resultTStr);
        
        // Make sure that open will update the Access Time
        // Note that this test doesn't check to see if the 
        // access time is relatively close to "now" since there may be
        // significant delays opening the XSet. The trick here is just 
        // to make sure that the access time is equal to or later than 
        // the time of the open call. 
        long nowMillis = System.currentTimeMillis();
        XSet xset2 = xsys.openXSet( xuid, XSet.MODE_READ_ONLY );
        long accessMillis = xset2.getDateTime(XSet.XAM_TIME_ACCESS).getTimeInMillis();
        long delta = accessMillis - nowMillis;
        if( delta < 0 )
           fail( "Open READONLY did not update access time" ) ;
        xset2.close();
        // Check the persisted value
        long persistedAccessMills = xsys.getXSetAccessTime(xuid).getTimeInMillis();
        assertEquals( "openXSet(Readonly) didn't persist access time properly",
           accessMillis, persistedAccessMills );
        
        nowMillis = System.currentTimeMillis();
        XSet xset3 = xsys.openXSet( xuid, XSet.MODE_RESTRICTED );
        accessMillis = xset3.getDateTime(XSet.XAM_TIME_ACCESS).getTimeInMillis();
        delta = accessMillis - nowMillis;
        if( delta < 0 )
           fail( "Open RESTRICTED did not update access time" ) ;
        xset3.close();
        // Check the persisted value
        persistedAccessMills = xsys.getXSetAccessTime(xuid).getTimeInMillis();
        assertEquals( "openXSet(Restricted) didn't persist access time properly",
           accessMillis, persistedAccessMills );
        
        nowMillis = System.currentTimeMillis();
        XSet xset4 = xsys.openXSet( xuid, XSet.MODE_UNRESTRICTED );
        accessMillis = xset4.getDateTime(XSet.XAM_TIME_ACCESS).getTimeInMillis();
        delta = accessMillis - nowMillis;
        if( delta < 0 )
           fail( "Open UNRESTRICTED did not update access time" ) ;
        xset4.close();
        // Check the persisted value
        persistedAccessMills = xsys.getXSetAccessTime(xuid).getTimeInMillis();
        assertEquals( "openXSet(Unrestricted) didn't persist access time properly",
           accessMillis, persistedAccessMills );
        
        xsys.abandon();
        xsys.close();
        
        // Create a new XSystem reference, authenticated as a different user
        // and check to make sure getAccessTime throws the correct exception.
        XSystem xsys2 = this.getAuthenticatedXSystem( PASSWD_INFO[0][0], PASSWD_INFO[0][1] );
        try {
           Calendar access = xsys2.getXSetAccessTime(xuid);
           fail( "getXSetAccessTime should have thrown an exception" );
        }
        catch( XSetInaccessibleException xsie ) { /* good, now ignore it */ }

        xsys2.abandon();
        xsys2.close();
    }

    /**
     * Tests the path through to the ManagePolicy method
     * @throws Exception
     */
    public void testManagementPolicy() throws Exception
    {
       XSystem xsys = s_xam.connect(s_xri);
       SASLUtils.authenticatePlain( xsys, "testuser", "testpasswd" );
       XSet xs = xsys.createXSet(XSet.MODE_UNRESTRICTED);

       String policy = xsys.getString( XSystem.XAM_MANAGEMENT_POLICY_DEFAULT );
       xs.applyManagementPolicy( false, policy );
       
       assertEquals( "Wrong value for the management policy", policy, xs.getString(XSet.XAM_MANAGEMENT_POLICY));
       
       
       // Make sure a bad policy is rejected.
       // Assuming that the following garbage name has not been chosen as 
       // a valid name... :-)
       try 
       {
          xs.applyManagementPolicy( false, "dflnkvhjwqi34848rkcdjknfjwdfjk" );
          fail( "Management policy should have thrown an exception" );
       }
       catch( PolicyNameException pne )
       {
          // ignore
       }
       xs.close();
       xsys.close();
    }
    
    /**
     * Test of isXSetRetained method, of class ReferenceXSystem.
     */
    public void TODOtestIsXSetRetained() throws Exception {
        announce("ReferenceXSystemTest.isXSetRetained");
        XUID xsetid = null;
        ReferenceXSystem instance = new ReferenceXSystem();
        boolean expResult = false;
        boolean result = instance.isXSetRetained(xsetid);
        assertEquals(expResult, result);
        // TODO review the generated test code and remove the default call to fail.
        fail("The test case is a prototype.");
    }

    /**
     * Test of openXSet method, of class ReferenceXSystem.
     */
    public void TODOtestOpenXSet() throws Exception {
        announce("ReferenceXSystemTest.openXSet");
        XUID xsetid = null;
        String openMode = "";
        ReferenceXSystem instance = new ReferenceXSystem();
        XSet expResult = null;
        XSet result = instance.openXSet(xsetid, openMode);
        assertEquals(expResult, result);
        // TODO review the generated test code and remove the default call to fail.
        fail("The test case is a prototype.");
    }
}
