/*
 * Copyright (c) 2008, Sun Microsystems, Inc.
 * Copyright (c) 2008, The Storage Networking Industry Association.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import org.snia.xam.XAMException;
import static junit.framework.Assert.*;

import org.snia.xam.XUID;

import org.snia.xam.testvim.TestXUID;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestResult;
import junit.framework.TestSuite;


/**
 * Basic tests for SniaXUID class
 */
public class SniaXUIDTest extends ReferenceXAMTest {
   
    SniaXUIDTest(String name) {
        super(name);
    }
    public static void main(String[] args)
    {
        System.exit(runTest(args, suite()));
    }
    public static Test suite()
    {
    	TestSuite suite = new TestSuite();
        suite.addTest(new SniaXUIDTest("testCreateXUID"));
        suite.addTest(new SniaXUIDTest("testInvalidXUID"));
    	return suite;
    }
    
    /**
     * Test the basic create and tranformation methods in SniaXUID
     * @throws java.lang.Exception
     */
    public void testCreateXUID() throws Exception {
        announce("SniaXUIDTest.testCreateXUID");
        // Test RefVIM generated XUID
        XUID xuid = SniaXUID.generateXUID();
        checkXUIDMethods(xuid, "RefVIM XUID: ");
        
        // Test Foreign generated XUID
        TestXUID foreignXuid = TestXUID.newXUID();
        checkXUIDMethods(foreignXuid,"Foreign XUID: ");
        
        if (xuid.equals(foreignXuid)) {
            fail("RefVIM and Foreign XUIDs should not be equal.");
        }
        XUID xuid2 = SniaXUID.generateXUID();
        if (xuid2.equals(xuid)) {
            fail("Generated XUIDs should not be equal.");
        }
        
        byte[] paddedBytes = ((SniaXUID)xuid).toPaddedBytes();
        XUID xuidPadded = SniaXUID.getInstance(paddedBytes);
        // Should be the same
        if (!xuidPadded.equals(xuid)) {
            fail("XUID to/from padded bytes is not the same.");
        }
    }
    
    /**
     * Verify that XUID returned is the same when performing hex/string and
     * bytes/string transformations.
     * 
     * @param xuid
     * @param failMsgPrefix
     * @throws java.lang.Exception
     */
    private void checkXUIDMethods(XUID xuid, String failMsgPrefix) throws Exception {
        
        // Check the hex to string and back conversions
        String hexString = SniaXUID.convertToSafeString(xuid);
        XUID xuidFromHex = SniaXUID.createFromSafeString(hexString);
        
        if (!xuid.equals(xuidFromHex)) {
            fail(failMsgPrefix + 
                    " XUID to/from hex representation is not the same." +
                    "\nStart XUID: " + xuid.toString() + 
                    "\nHex XUID  : " + xuidFromHex.toString());
        }
        
        XUID byteXUID = SniaXUID.getInstance(xuid.toBytes());
        XUID stringXUID = SniaXUID.getInstance(xuid.toString());
        
        if (!byteXUID.equals(stringXUID)) {
            fail(failMsgPrefix +
                    " XUID to/from byte[]/String is not the same");
        }
        
    }
    
    /**
     * Test the passing invalid XUIDs to getInstance method.
     * @throws java.lang.Exception
     */
    public void testInvalidXUID() throws Exception {
        announce("SniaXUIDTest.testInvalidXUID");
        // Test RefVIM generated XUID
        XUID xuid = SniaXUID.generateXUID();
        String badXUID = xuid.toString();
        
        // Fewer than the minimum bytes required
        badXUID = badXUID.substring(0, XUID.MIN_LENGTH -1);
        XUID testXUID = null;
        try {
            testXUID= SniaXUID.getInstance(badXUID);
            fail("Test failed. Invalid XUID string passed to SniaXUID.getInstance not caught.");
        } catch (XAMException ex) {
            // Expected exception
        }
        
        // Length is > max array size allowd
        byte[] xuidBytes = xuid.toBytes();
        int length = xuidBytes[SniaXUID.XUID_LENGTH_INDEX];
        xuidBytes[SniaXUID.XUID_LENGTH_INDEX] = XUID.MAX_LENGTH + 1;
        try {
            testXUID= SniaXUID.getInstance(xuidBytes);
            fail("Test failed. Invalid length (> max) in XUID bytes passed to SniaXUID.getInstance not caught.");
        } catch (XAMException ex) {
            // Expected exception
        }
        
        // Internal length is larger than array size
        xuidBytes[SniaXUID.XUID_LENGTH_INDEX] = (byte) (length + 1);
        try {
            testXUID= SniaXUID.getInstance(xuidBytes);
            fail("Test failed. Invalid length (> array length) in XUID bytes passed to SniaXUID.getInstance not caught.");
        } catch (XAMException ex) {
            // Expected exception
        }
    }
        }
