/*
 * Copyright (c) 2008, Hewlett-Packard Development Company, L.P. 
 * Copyright (c) 2008, The Storage Networking Industry Association.
 * Copyright (c) 2008, Sun Microsystems, Inc.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  
 * Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *  
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 *  
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 * the names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.snia.xam.vim.reference;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import org.snia.xam.FieldContainer;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XSet;
import org.snia.xam.XSetUnderRetentionException;
import org.snia.xam.XStream;
import org.snia.xam.XSystem;
import org.snia.xam.XUID;
import org.snia.xam.base.XAMImplementation;
import org.snia.xam.vim.reference.utils.ReferenceSaslUtils;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestResult;
import junit.framework.TestSuite;

public class XAMTest extends TestCase
{
    /** Indicate where the XAM Library config file lives */
    public static final String CONFIG = "XAMImplementation.config";
    
	/** System property that specifies the test property file name */
	public static final String TEST_PROP_FILE = "xam.test.props";
	
	/** Test property that specifies the connection XRI */
	public static final String XRI_PROP  = "xam.test.xri";
	
	/** Test property that specifies the VIM file */
	public static final String CONFIG_PROP = "xam.test.vims";
	
	/** Test property that specifies the username */
	public static final String USER_PROP = "xam.test.username";
	
	/** Test property that specifies the username */
	public static final String PASS_PROP = "xam.test.password";
	
	protected static final String DEFAULT_USER = "test";
	protected static final String DEFAULT_PASS = "test";
	
	protected static final String[] DEF_PROPS = {
		XAMLibrary.XAM_API_LEVEL,
		XAMLibrary.XAM_LOG_LEVEL,
        XAMLibrary.XAM_LOG_VERBOSITY,
        XAMLibrary.XAM_LOG_PATH,
		XAMLibrary.XAM_IDENTITY,
	};
	
	protected static final String[] DEF_TYPES = {
		XAMLibrary.STYPE_STRING_MIME_TYPE,
		XAMLibrary.STYPE_INT_MIME_TYPE,
        XAMLibrary.STYPE_INT_MIME_TYPE,
        XAMLibrary.STYPE_STRING_MIME_TYPE,
		XAMLibrary.STYPE_STRING_MIME_TYPE,
	};

	protected static String     s_pass;
	protected static String     s_user;
	protected static XAMLibrary s_xam;
	protected static String     s_xri;

	private List<XUID> m_xuids;
	
    public static void main(String[] args)
	{
    	System.exit(runTest(args, suite()));
	}

    public XAMTest(String name)
    {
        super(name);
        m_xuids = new ArrayList<XUID>();
    }
    
    public static Test suite()
    {
    	TestSuite suite = new TestSuite();
    	suite.addTest(new XAMTest("testProperties"));
    	suite.addTest(new XAMTest("testLogging"));
    	return suite;
    }
    
    public void testProperties() throws Exception
    {
    	checkProperties(s_xam, DEF_PROPS, DEF_TYPES);
    	
    	s_xam.setProperty(XAMLibrary.XAM_LOG_LEVEL, 1L);
    	assertEquals(XAMLibrary.XAM_LOG_LEVEL + " value", 1L, 
    		s_xam.getLong(XAMLibrary.XAM_LOG_LEVEL));
    	s_xam.setProperty(XAMLibrary.XAM_LOG_LEVEL, XAMLibrary.XAM_LOG_ALL );
    	assertEquals(XAMLibrary.XAM_LOG_LEVEL + " value", XAMLibrary.XAM_LOG_ALL,
        		s_xam.getLong(XAMLibrary.XAM_LOG_LEVEL));
    }

    public void testLogging() throws Exception
    {
    	// Do something that gets logged
    	Properties vims = new Properties();
    	vims.setProperty(XAMLibrary.XAM_LOG_PATH, "log.test");
    	vims.setProperty("xam_int." + XAMLibrary.XAM_LOG_LEVEL, String.valueOf(XAMLibrary.XAM_LOG_ALL));
    	XAMImplementation xam = new XAMImplementation();
    	xam.loadVIMConfig(vims);
    	
    	// Now check the log for the appropriate string.
    	boolean found = false;
    	String path = xam.getString(XAMLibrary.XAM_LOG_PATH);
    	BufferedReader reader = new BufferedReader(new FileReader(path));
    	String line = null;
    	while (!found && ((line = reader.readLine()) != null))
    	{
    		found = line.indexOf("loadVIMConfig") > -1;
    	}
    	
    	assertTrue("Did not find log message", found);
    }

    protected void checkProperties(FieldContainer ct,
    		String[] expected, String[] types)
    	throws Exception
    {
    	List<String> props = new ArrayList<String>();
    	for (int i = 0; i < expected.length; i++)
    		props.add(expected[i]);
    	Iterator iter = ct.openFieldIterator("");
    	while (iter.hasNext())
    		props.remove((String)iter.next());
    	
    	int notFound = props.size();
    	if (notFound > 0)
    	{
    		StringBuffer buf = new StringBuffer(256);
    		buf.append("The following properties were not found: ");
    		buf.append(props.get(0));
    		for (int i = 1; i < notFound; i++)
    		{
    			buf.append(", ");
    			buf.append(props.get(i));
    		}
    		fail(buf.toString());
    	}

    	for (int i = 0; i < expected.length; i++)
    	{
    		assertEquals(expected[i] + " type",
    			types[i],
    			ct.getFieldType(expected[i]));
    	}
    }
    
	protected void setUp() throws Exception
	{
		initLibrary();
	}

	protected void tearDown() throws Exception
	{
		cleanUpTestData();
	}

	public static int runTest(String[] args, Test test)
	{
    	if (args.length == 1)
    		System.setProperty(TEST_PROP_FILE, args[0]);
        TestResult res = junit.textui.TestRunner.run(test);
		return Math.max(res.failureCount(), res.errorCount());
	}
	
	public static void initLibrary() throws Exception
	{
        if (s_xam == null)
        {
           Properties props = new Properties();
           String testPropFile = System.getProperty(TEST_PROP_FILE);
           if (testPropFile == null)
            testPropFile = TEST_PROP_FILE;
        props.load(new FileInputStream(testPropFile));
        
        s_xri = props.getProperty(XRI_PROP);
        s_user = props.getProperty(USER_PROP, DEFAULT_USER);
        s_pass = props.getProperty(PASS_PROP, DEFAULT_PASS);

        s_xam = new XAMImplementation(props.getProperty(CONFIG_PROP));
        }
	}
	
	public static XSystem connect() throws XAMException
	{
		XSystem system = s_xam.connect(s_xri);
		ReferenceSaslUtils.authenticatePlain(system, s_user, s_pass);
		return system;
	}
	
	public static void printFieldNames(String title, FieldContainer ct)
		throws XAMException
	{
		System.out.println(title);
		Iterator iter = ct.openFieldIterator(".*");
		while (iter.hasNext())
		{
			String field = (String)iter.next();
			System.out.print("  ");
			System.out.print(field);
			System.out.print('=');
			System.out.print(getValueFor(ct, field));
			System.out.println();
		}
		System.out.println("--");
	}
	
	public static String getValueFor(FieldContainer ct, String field)
		throws XAMException
	{
		String type = ct.getFieldType(field);
		if (type.equals(XAMLibrary.STYPE_BOOLEAN_MIME_TYPE))
			return Boolean.toString(ct.getBoolean(field));
		else if (type.equals(XAMLibrary.STYPE_DATETIME_MIME_TYPE))
			return ct.getDateTime(field).toString();
		else if (type.equals(XAMLibrary.STYPE_DOUBLE_MIME_TYPE))
			return Double.toString(ct.getDouble(field));
		else if (type.equals(XAMLibrary.STYPE_INT_MIME_TYPE))
			return Long.toString(ct.getLong(field));
		else if (type.equals(XAMLibrary.STYPE_STRING_MIME_TYPE))
			return ct.getString(field);
		else if (type.equals(XAMLibrary.STYPE_XUID_MIME_TYPE))
			return ct.getXUID(field).toString();
		return "";
	}
	
	protected void cleanUpTestData() throws Exception
	{
		try
		{
			int count = m_xuids.size();
			if (count > 0)
			{
				XSystem system = connect();
				for (int i = 0; i < count; i++)
					deleteXSet(system, m_xuids.get(i));
			}
		}
		catch (XAMException x)
		{
			System.out.println(getClass().getName()+"."+getName());
		}
	}
	
	protected XUID commitXSet(XSet xset)
		throws XAMException
	{
		XUID xuid = xset.commit();
		if (!m_xuids.contains(xuid))
			m_xuids.add(xuid);
		return xuid;
	}
	
	protected boolean deleteXSet(XSystem system, XUID xuid)
		throws XAMException
	{
		try
		{
			system.deleteXSet(xuid);
			return true;
		}
		catch (XSetUnderRetentionException x)
		{
			// This is a valid error and should not cause
			// a test to fail
		}
		// Return false if the XSet was not deleted because
		// of retention
		return false;
	}

	protected void writeBytes(XStream stream, byte[] buf, int offset, int length)
		throws XAMException
	{
		int written = 0;
		while (written < length)
			written += stream.write(buf, offset+written, length-written);
	}

	protected void readBytes(XStream stream, byte[] buf, int offset, int length)
		throws XAMException
	{
		int read = 0;
		while (read < buf.length)
			read += stream.read(buf, offset+read, length-read);
	}
	
	protected void fillBuffer(byte[] buf, byte value)
	{
		for (int i = 0; i < buf.length; i++)
			buf[i] = value;
	}
	
	protected void checkBuffer(byte[] buf, int offset, int length, byte value)
	{
		for (int i = 0; i < length; i++)
			assertEquals("Byte " + i, value, buf[offset+i]);
	}
}
