/* Copyright (c) 2008-2009, The Storage Networking Industry Association.

 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */
/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */
/* (c) Copyright 2009 Sun Microsystems                          */

package org.snia.xam.base;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.snia.xam.FieldContainer;
import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.FieldExistsException;
import org.snia.xam.FieldReadOnlyException;
import org.snia.xam.InvalidFieldNameException;
import org.snia.xam.InvalidFieldTypeException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XAMException;
import org.snia.xam.XIterator;
import org.snia.xam.XUID;

/**
 * Abstract base class for objects implementing the {@link FieldContainer}
 * interface. Fields are stored internally in a Map in a thread safe manner.
 * Creating, modifying and getting of properties is implemented with layered
 * method calls so that sub-classes can intercept the calls at various points to
 * provide custom behavior. Internal methods that are not thread safe are noted
 * and should be protected by the sub-class.
 */
public abstract class AbstractFieldContainer
   implements FieldContainer
{
   private static final String WILDCARD = ".*";
   protected static final String PROTECTED_NAMESPACE_PREFIX = ".";

   private Map<String,Object> m_fields;

   private boolean m_protectNamespace;
   
   private Logger  m_logger;
   private boolean m_highVerbosity;
   private static String s_logClass = AbstractFieldContainer.class.getName();

   /**
    * Protected constructor, sub-classes should provide their own public
    * constructors.
    */
   protected AbstractFieldContainer()
   {
      m_fields = new HashMap<String,Object>();
      setProtectNamespace(true);
   }

   protected void setLogger( Logger extLogger )
   {
      m_logger = extLogger;
   }
   
   protected void setVerbosity( boolean verbosity )
   {
      m_highVerbosity = verbosity;
   }
   
   public void createProperty(String fieldName, boolean binding, boolean value)
      throws XAMException
   {
      String method = "createProperty(boolean)";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );
      
      createProperty(new Property(fieldName,
         XAMLibrary.STYPE_BOOLEAN_MIME_TYPE, new Boolean(value), binding, false));

      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
}

   public void createProperty(String fieldName, boolean binding, long value)
      throws XAMException
   {
      String method = "createProperty(long)";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      createProperty(new Property(fieldName, XAMLibrary.STYPE_INT_MIME_TYPE,
         new Long(value), binding, false));

      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
}

   public void createProperty(String fieldName, boolean binding, double value)
      throws XAMException
   {
      String method = "createProperty(double)";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      createProperty(new Property(fieldName, XAMLibrary.STYPE_DOUBLE_MIME_TYPE,
         new Double(value), binding, false));

      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
}

   public void createProperty(String fieldName, boolean binding, XUID value)
      throws XAMException
   {
      String method = "createProperty(XUID)";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      createProperty(new Property(fieldName, XAMLibrary.STYPE_XUID_MIME_TYPE,
         value, binding, false));

      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
}

   public void createProperty(String fieldName, boolean binding, String value)
      throws XAMException
   {
      String method = "createProperty(String)";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      createProperty(new Property(fieldName, XAMLibrary.STYPE_STRING_MIME_TYPE,
         value, binding, false));

      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
}

   public void createProperty(String fieldName, boolean binding, Calendar value)
      throws XAMException
   {
      String method = "createProperty(Calendar)";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      createProperty(new Property(fieldName,
         XAMLibrary.STYPE_DATETIME_MIME_TYPE, value, binding, false));

      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
   }

   public void deleteField(String fieldName)
      throws XAMException
   {
      String method = "deleteField";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      synchronized (m_fields)
      {
         checkFieldNameWellFormed(fieldName);
         Field field = getField(fieldName);
         checkProtectedNameSpace(fieldName);
         checkFieldReadOnly(field);
         removeField(field);
      }
      
      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
   }

   public boolean getBoolean(String fieldName)
      throws XAMException
   {
      String method = "getBoolean";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      Property prop = getProperty(fieldName, XAMLibrary.STYPE_BOOLEAN_MIME_TYPE);
      boolean rvalue =  ((Boolean) prop.getValue()).booleanValue();
      
      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
      return rvalue;
   }

   public Calendar getDateTime(String fieldName)
      throws XAMException
   {
      String method = "getDateTime";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );
      
      Property prop = getProperty(fieldName,
         XAMLibrary.STYPE_DATETIME_MIME_TYPE);
      Calendar rvalue = (Calendar) prop.getValue();
      
      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
      return rvalue;
   }

   public boolean getFieldBinding(String fieldName)
      throws XAMException
   {
      String method = "getFieldBinding";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      Field field = getFieldFor(fieldName);
      boolean rvalue = field.getBinding();
      
      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
      return rvalue;
   }

   public long getFieldLength(String fieldName)
      throws XAMException
   {
      String method = "getFieldLength";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      Field field = getFieldFor(fieldName);
      long rvalue = field.getLength();
      
      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
      return rvalue;
   }

   public boolean getFieldReadOnly(String fieldName)
      throws XAMException
   {
      String method = "getFieldReadOnly";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      Field field = getFieldFor(fieldName);
      boolean rvalue = field.getReadOnly();
      
      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
      return rvalue;
   }

   public String getFieldType(String fieldName)
      throws XAMException
   {
      String method = "getFieldType";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      Field field = getFieldFor(fieldName);
      String rvalue = field.getType();
      
      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
      return rvalue;
   }

   public double getDouble(String fieldName)
      throws XAMException
   {
      String method = "getDouble";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      Property prop = getProperty(fieldName, XAMLibrary.STYPE_DOUBLE_MIME_TYPE);
      double rvalue = ((Double) prop.getValue()).doubleValue();
      
      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
      return rvalue;
   }

   public long getLong(String fieldName)
      throws XAMException
   {
      String method = "getLong";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      Property prop = getProperty(fieldName, XAMLibrary.STYPE_INT_MIME_TYPE);
      long rvalue = ((Long) prop.getValue()).longValue();

      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
      return rvalue;
   }

   public String getString(String fieldName)
      throws XAMException
   {
      String method = "getString";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      Property prop = getProperty(fieldName, XAMLibrary.STYPE_STRING_MIME_TYPE);
      String rvalue = (String) prop.getValue();
      
      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
      return rvalue;
   }

   public XUID getXUID(String fieldName)
      throws XAMException
   {
      String method = "getXUID";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      Property prop = getProperty(fieldName, XAMLibrary.STYPE_XUID_MIME_TYPE);
      XUID rvalue = (XUID) prop.getValue();
      
      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
      return rvalue;
   }

   public Object getPropertyObject(String fieldName)
      throws XAMException
   {
      synchronized (m_fields)
      {
         checkFieldNameWellFormed( fieldName );
         Property prop = getPropertyField(fieldName);
         return prop.getValue();
      }
   }

   public XIterator openFieldIterator(String prefix)
      throws XAMException
   {
      String method = "openFieldIterator";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      synchronized (m_fields)
      {
         checkFieldNameWellFormed(prefix);
         XIterator rvalue = fieldIterator(prefix);
         
         if( m_highVerbosity && (m_logger != null ))
            m_logger.exiting( s_logClass, method );
         return rvalue;
      }
   }

   public void setProperty(String fieldName, boolean value)
      throws XAMException
   {
      String method = "setProperty(boolean)";      
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      setProperty(fieldName,
         new Boolean(value),
         XAMLibrary.STYPE_BOOLEAN_MIME_TYPE);

      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
   }

   public void setProperty(String fieldName, Calendar value)
      throws XAMException
   {
      String method = "setProperty(Calendar)";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      setProperty(fieldName, value, XAMLibrary.STYPE_DATETIME_MIME_TYPE);

      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
   }

   public void setFieldAsBinding(String fieldName)
      throws XAMException
   {
      String method = "setFieldAsBinding";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      synchronized (m_fields)
      {
         checkFieldNameWellFormed(fieldName);
         DefaultField field = (DefaultField) getField(fieldName);
         checkFieldReadOnly(field);
         field.setBinding(true);
         field.setModified(true);
      }
      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
   }

   public void setFieldAsNonbinding(String fieldName)
      throws XAMException
   {
      String method = "setFielAsNonbinding";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      synchronized (m_fields)
      {
         checkFieldNameWellFormed(fieldName);
         DefaultField field = (DefaultField) getField(fieldName);
         checkFieldReadOnly(field);
         field.setBinding(false);
         field.setModified(true);
      }

      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
   }

   public void setProperty(String fieldName, double value)
      throws XAMException
   {
      String method = "setProperty(double)";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      setProperty(fieldName,
         new Double(value),
         XAMLibrary.STYPE_DOUBLE_MIME_TYPE);

      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
   }

   public void setProperty(String fieldName, long value)
      throws XAMException
   {
      String method = "setProperty(long)";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      setProperty(fieldName, new Long(value), XAMLibrary.STYPE_INT_MIME_TYPE);
      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
   }

   public void setProperty(String fieldName, String value)
      throws XAMException
   {
      String method = "setProperty(String)";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      setProperty(fieldName, value, XAMLibrary.STYPE_STRING_MIME_TYPE);

      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
   }

   public void setProperty(String fieldName, XUID value)
      throws XAMException
   {
      String method = "setProperty(XUID)";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      setProperty(fieldName, value, XAMLibrary.STYPE_XUID_MIME_TYPE);

      if( m_highVerbosity && (m_logger != null ))
         m_logger.exiting( s_logClass, method );
   }

   /**
    * Clears all the fields from the internal Map. This does NOT delete the
    * fields from the actual system. This method is not thread safe.
    */
   protected void clearFields()
   {
      m_fields.clear();
   }

   /**
    * Returns the number of fields stored in the FieldContainer. This method is
    * not thread safe.
    * 
    * @return The number of fields.
    */
   protected int getFieldCount()
   {
      return m_fields.size();
   }

   /**
    * Checks to see if the specified field already exists. This method is not
    * thread safe.
    * 
    * @param fieldName The field name
    * @return True if the field already exists, false otherwise.
    * @throws XAMException, a general error occurred in communicating to the XAM Storage System.
    */
   public boolean containsField(String fieldName)
       throws InvalidFieldNameException, XAMException
   {
      String method = "containsField";
      if( m_highVerbosity && (m_logger != null ))
         m_logger.entering( s_logClass, method );

      synchronized (m_fields)
      {
         checkFieldNameWellFormed(fieldName);
         boolean rvalue = m_fields.containsKey(fieldName);
         
         if( m_highVerbosity && (m_logger != null ))
            m_logger.exiting( s_logClass, method );
         return rvalue;
      }
   }

   /**
    * Adds a new {@link Field} to the {@link FieldContainer}. If a field
    * already exists with the same name it is overwritten. This method is not
    * thread safe.
    * 
    * @param field The new Field
    */
   protected void addField(Field field)
   {
      m_fields.put(field.getName(), field);
   }

   /**
    * Removes the {@link Field} from the {@link FieldContainer}. Note, this
    * does not delete the field from the system. This method is not thread safe.
    * 
    * @param field
    */
   protected void removeField(Field field)
   {
      m_fields.remove(field.getName());
   }

   /**
    * Updates an existing {@link Field} or adds a new one if it does not exist.
    * This method is not thread safe.
    * 
    * @param field The field to update
    * @throws XAMException If the field value does not match the field type.
    */
   protected void updateField(Field field)
      throws XAMException
   {
      Field old = (Field) m_fields.get(field.getName());
      if (old instanceof DefaultField)
      {
         DefaultField def = (DefaultField) old;
         def.setBinding(field.getBinding());
         def.setLength(field.getLength());
         def.setReadonly(field.getReadOnly());
         if ((def instanceof Property) && (field instanceof Property))
         {
            ((Property) def).setValue(((Property) field).getValue(),
               field.getType());
         }
         else
            def.setType(field.getType());
      }
      else
         addField(field);
   }

   /**
    * Returns the named {@link Field} if it exists or throws a
    * {@link FieldDoesNotExistException} if it does not. This method is not
    * thread safe.
    * 
    * @param fieldName The field name
    * @return The named field
    * @throws XAMException If the field does not exist
    */
   protected Field getField(String fieldName)
      throws XAMException
   {
      Field field = (Field) m_fields.get(fieldName);
      if (field == null) throw new FieldDoesNotExistException(fieldName);
      return field;
   }

   /**
    * Returns the named field as a {@link Property} if it exists or throws a
    * {@link FieldDoesNotExistException} if it does not. If the field is not a
    * property a {@link InvalidFieldTypeException} is thrown. This method is not
    * thread safe.
    * 
    * @param fieldName The property name
    * @return The named Property
    * @throws XAMException If the property does not exist or is a stream.
    */
   protected Property getPropertyField(String fieldName)
      throws XAMException
   {
      checkFieldNameWellFormed( fieldName );
      Field field = getField(fieldName);
      if (!(field instanceof Property))
         throw new InvalidFieldTypeException("Field is not a property "
            + fieldName);
      return (Property) field;
   }

   /**
    * Returns the named field as a {@link Property} of the specifed type if it
    * exists or throws a {@link FieldDoesNotExistException} if it does not. If
    * the field is not a property a {@link InvalidFieldTypeException} is thrown.
    * If the property type does not match the specified type a
    * {@link InvalidFieldTypeException} is thrown This method is not thread
    * safe.
    * 
    * @param fieldName The property name
    * @param type The property mime type
    * @return The named property
    * @throws XAMException If the property does not exist, is a stream or does
    *            not match the requested mime type.
    */
   protected Property getPropertyType(String fieldName, String type)
      throws XAMException
   {
      Property prop = getPropertyField(fieldName);
      if (!prop.getType().equals(type))
         throw new InvalidFieldTypeException("Property '" + fieldName + "' is not type " + type);
      return prop;
   }

   /**
    * Returns the named field as a {@link StreamField} if it exists or throws a
    * {@link FieldDoesNotExistException} if it does not. If the field is not a
    * stream a {@link InvalidFieldTypeException} is thrown. This method is not
    * thread safe.
    * 
    * @param fieldName The name of the stream
    * @return The named stream
    * @throws XAMException If the field does not exist or is not a stream.
    */
   protected StreamField getStreamField(String fieldName)
      throws XAMException
   {
      Field field = getField(fieldName);
      if (field instanceof StreamField) return (StreamField) field;
      throw new InvalidFieldTypeException("Field is not a stream");
   }

   /**
    * Returns an Iterator of all the field names that begin with the specified
    * prefix. To specify all field names pass in a blank string or null. The
    * field names are sorted prior to being iterated over. This method is not
    * thread safe.
    * 
    * @param prefix The field name prefix.
    * @return An Iterator object containing field names
    */
   protected XIterator fieldIterator(String prefix)
   {
      if (prefix == null) prefix = WILDCARD;
      else
         prefix += WILDCARD;
      
      Pattern p = Pattern.compile(prefix);
      List<String> matches = new ArrayList<String>();
      Iterator<String> iter = m_fields.keySet().iterator();
      
      while (iter.hasNext())
      {
         String fieldName = iter.next();
         Matcher m = p.matcher(fieldName);
         if (m.matches()) matches.add(fieldName);
      }
      Collections.sort(matches);
      return new WrappedIterator(matches.iterator());
   }

   /**
    * Adds the specified {@link Property} to the FieldContainer after checking
    * to see a field with the same name exists and that the property name is not
    * in the protected name space.
    * 
    * @param prop The new property
    * @throws XAMException If a field with the same name exists or the property
    *            name is in the protected name space.
    */
   protected void createProperty(Property prop)
      throws XAMException
   {
      synchronized (m_fields)
      {
         checkFieldNameWellFormed(prop.getName());
         checkFieldExists(prop.getName());
         checkProtectedNameSpace(prop.getName());
         prop.setModified(true);
         addField(prop);
      }
   }

   /**
    * Returns the named {@link Field} if it exits and throws a
    * {@link FieldDoesNotExistException} if it does not.
    * 
    * @param fieldName The field name
    * @return The named Field
    * @throws XAMException If the field does not exist
    */
   protected Field getFieldFor(String fieldName)
      throws XAMException
   {
      synchronized (m_fields)
      {
         checkFieldNameWellFormed(fieldName);
         return getField(fieldName);
      }
   }

   /**
    * Returns the named field as a {@link Property} of the specifed type if it
    * exists or throws a {@link FieldDoesNotExistException} if it does not. If
    * the field is not a property a {@link InvalidFieldTypeException} is thrown.
    * If the property type does not match the specified type a
    * {@link InvalidFieldTypeException} is thrown
    * 
    * @param fieldName The property name
    * @param type The property mime type
    * @return The named property
    * @throws XAMException If the property does not exist, is a stream or does
    *            not match the requested mime type.
    */
   protected Property getProperty(String fieldName, String type)
      throws XAMException
   {
      synchronized (m_fields)
      {
         return getPropertyType(fieldName, type);
      }
   }

   /**
    * Modifies the specified property to have the new value and type. Exceptions
    * will be thrown if the field does not exist, is not a property or is
    * readonly.
    * 
    * @param fieldName The property name
    * @param value The new value
    * @param type The new mime type
    * @throws XAMException If the field does not exist, isn't a property or is
    *            read only.
    */
   protected void setProperty(String fieldName, Object value, String type)
      throws XAMException
   {
      synchronized (m_fields)
      {
         checkFieldNameWellFormed(fieldName);
         Property prop = getPropertyField(fieldName);
         checkFieldReadOnly(prop);
         prop.setValue(value, type);
         prop.setModified(true);
      }
   }

   /**
    * Sets the flag determining if creating properties in the protected
    * namespace will be allowed through the public methods.
    * 
    * @param protectNameSpace true if new property names should be checked,
    *           false otherwise
    */
   protected void setProtectNamespace(boolean protectNamespace)
   {
      m_protectNamespace = protectNamespace;
   }

   /**
    * Returns true if the property name space is checked when new properties are
    * created, false if the property name is not checked.
    * 
    * @return true if new property names are checked, false otherwise
    */
   protected boolean getProtectNamespace()
   {
      return m_protectNamespace;
   }

   /**
    * Validates that the value is of the correct class type and returns the
    * associated XAM mime type. If the class type does not match one of the XAM
    * mime types an IllegalArgumentException is thrown.
    * 
    * @param value A property value
    * @return The associated XAM mime type
    * @throws IllegalArgumentException if the class of the value does not match
    *            one of the XAM mime types
    */
   protected String checkPropertyType(Object value)
   {
      String type = null;
      if (value instanceof String) type = XAMLibrary.STYPE_STRING_MIME_TYPE;
      else if (value instanceof Calendar) type = XAMLibrary.STYPE_DATETIME_MIME_TYPE;
      else if (value instanceof XUID) type = XAMLibrary.STYPE_XUID_MIME_TYPE;
      else if (value instanceof Boolean) type = XAMLibrary.STYPE_BOOLEAN_MIME_TYPE;
      else if (value instanceof Long) type = XAMLibrary.STYPE_INT_MIME_TYPE;
      else if (value instanceof Double) type = XAMLibrary.STYPE_DOUBLE_MIME_TYPE;
      else
         throw new IllegalArgumentException(value.getClass().getName());
      return type;
   }

   /**
    * Validates that the field name is not the protected name space.
    * If it is a {@link InvalidFieldNameException} is thrown.
    * 
    * @param fieldName The field name
    * @throws XAMException If the field name is in the protected name space.
    */
   protected void checkProtectedNameSpace(String fieldName)
      throws XAMException
   {
      if (getProtectNamespace()
         && fieldName.startsWith(PROTECTED_NAMESPACE_PREFIX))
         throw new InvalidFieldNameException("Operation not permitted in protected name space " + fieldName);
   }

   /**
    * Checks to see if a field already exists. If it does a
    * {@link FieldExistsException} is thrown.
    * 
    * @param fieldName The field name
    * @throws XAMException If the field already exists
    */
   protected void checkFieldExists(String fieldName)
      throws XAMException
   {
      if (containsField(fieldName)) throw new FieldExistsException(fieldName);
   }
   
   protected void checkFieldNameWellFormed(String fieldName)
      throws InvalidFieldNameException
   {
      if( fieldName == null ) throw new InvalidFieldNameException("Field name is null" );
   }

   /**
    * Checks to see if a field is read only. If it is a
    * {@link FieldReadOnlyException} is thrown.
    * 
    * @param field A field
    * @throws XAMException If the field is read only
    */
   protected void checkFieldReadOnly(Field field)
      throws XAMException
   {
      if (field.getReadOnly())
         throw new FieldReadOnlyException(field.getName());
   }
}
