/* Copyright (c) 2008, The Storage Networking Industry Association.

 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */

/**
 * Based on "Free CRC source code in C"
 * Copyright (c) 2000 by Michael Barr.  This software is placed into
 * the public domain and may be used for any purpose.  However, this
 * notice must not be changed or removed and no warranty is either
 * expressed or implied by its publication or distribution.
 */
package org.snia.xam.base;

/**
 * CRC calculation code ported from Michael Barr's free example C code.
 */
public class Crc
{
   public static final byte[] CHECK_BYTES = "123456789".getBytes();

   private int m_check;

   private int m_finalXor;

   private int m_initRemainder;

   private int m_polynomial;

   private boolean m_reflectData;

   private boolean m_reflectRemainder;

   private int[] m_table;

   private int m_topbit;

   private int m_width;

   /**
    * Creates a a new Crc object with the specified parameters.
    * 
    * @param width The width of the CRC in bytes
    * @param polynomial The CRC polynomial
    * @param initialRemainder The initial remainder
    * @param finalXorValue The final XOR value
    * @param checkValue The CRC for the ASCII bytes '123456789'
    * @param reflectData Flag to reflect the data
    * @param reflectRemainder Flag to reflect the remainder
    */
   public Crc(int width, int polynomial, int initialRemainder,
      int finalXorValue, int checkValue, boolean reflectData,
      boolean reflectRemainder)
   {
      m_width = width;
      m_polynomial = polynomial;
      m_initRemainder = initialRemainder;
      m_finalXor = finalXorValue;
      m_check = checkValue;
      m_reflectData = reflectData;
      m_reflectRemainder = reflectRemainder;
      m_topbit = 1 << (m_width - 1);
      init();
   }

   /**
    * Returns the CRC value that this instance should generate for the ASCII
    * bytes '123456789'.
    * 
    * @return The CRC check value
    */
   public int getCheckValue()
   {
      return m_check;
   }

   /**
    * Returns the CRC for the byte array. This method is thread safe.
    * 
    * @param bytes The data to calculate the CRC over
    * @return The CRC for the data
    */
   public int getCrc(byte[] bytes)
   {
      return getCrc(bytes, 0, bytes.length);
   }

   /**
    * Returns the CRC for some bytes in an array. This method is thread safe.
    * 
    * @param bytes The data to calculate the CRC over
    * @param offset The offset into the data
    * @param length The number of bytes to use
    * @return The CRC for the data
    */
   public int getCrc(byte[] bytes, int offset, int length)
   {
      int remainder = m_initRemainder;
      int end = offset + length;
      // Divide the data by the polynomial one byte at a time.
      for (int i = offset; i < end; i++)
      {
         int data = bytes[i];
         if (m_reflectData) data = reflect(data, 8);
         data = data ^ (remainder >> (m_width - 8));
         remainder = m_table[data] ^ ((remainder << 8) & 0xffff);
      }
      // The final remainder is the CRC
      if (m_reflectRemainder) remainder = reflect(remainder, m_width);
      remainder = remainder ^ m_finalXor;
      return remainder;
   }

   /**
    * Reoders the bits in a binary sequence by reflecting them about the middle
    * position
    * 
    * @param data The data to reflect
    * @param width The size of the data in bytes
    * @return The reflection of the original data
    */
   private int reflect(int data, int width)
   {
      int reflection = 0;
      // Reflect the data about the center bit
      for (int i = 0; i < width; i++)
      {
         // If the LSB bit is set, set the reflection of it
         if ((data & 0x01) != 0) reflection |= (1 << ((width - 1) - i));
         data = data >> 1;
      }
      return reflection;
   }

   /**
    * Populates the CRC lookup table
    */
   private void init()
   {
      m_table = new int[256];
      // Compute the remainder of each possible dividend
      for (int dividend = 0; dividend < 256; dividend++)
      {
         // Start with the dividend followed by zeros
         int remainder = dividend << (m_width - 8);
         // Perform modulo-2 division, a bit at a time
         for (int bit = 8; bit > 0; --bit)
         {
            // Try to divide the current data bit
            if ((remainder & m_topbit) != 0) remainder = (remainder << 1)
               ^ m_polynomial;
            else
               remainder = remainder << 1;
         }
         // Store the result in the table
         m_table[dividend] = remainder & 0xffff;
      }
   }
}
