/* Copyright (c) 2008, The Storage Networking Industry Association.

 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */

package org.snia.xam.base;

import org.snia.xam.InvalidXUIDException;
import org.snia.xam.XUID;

/**
 * Default implementation of the {@link XUID} interface.
 * Static methods are provided to convert to and from both
 * String and byte[] objects.
 */
public class DefaultXUID
   implements XUID, Comparable<XUID>
{
   protected static XAMCrc s_crc = new XAMCrc();

   private String m_base64;

   protected byte[] m_bytes;

   private static final int LENGTH_OFFSET = 5;
   private static final int CRC_OFFSET = 6;

   /**
    * Constructs a new XUID from the specified byte[]. No validation is done on
    * the bytes. It is assumed that the byte[] contains a valid XUID. This
    * constructor should only be called internally after validating the byte[]
    * or by sub-classes that have validated the byte[].
    * 
    * @param bytes The XUID bytes
    * @param offset The offset into the byte[]
    * @param len The number of XUID bytes in the byte[]
    */
   protected DefaultXUID(byte[] bytes, int offset, int len)
   {
      m_bytes = new byte[len];
      System.arraycopy(bytes, offset, m_bytes, 0, len);
   }

   /**
    * Constructs a new XUID from the specified byte[]. No validation is done on
    * the bytes. It is assumed that the byte[] contains a valid XUID. This
    * constructor should only be called internally after validating the byte[]
    * or by sub-classes that have validated the byte[].
    * 
    * @param bytes The XUID bytes
    */
   protected DefaultXUID(byte[] bytes)
   {
      this(bytes, 0, bytes.length);
   }

   /**
    * Build a XUID from a base64 encoded byte string.
    * @param s The byte string.
    */
   public static XUID valueOf(String str)
      throws InvalidXUIDException
   {
      try
      {
         return valueOf(Base64.base64Decode(str));
      }
      catch (Exception x)
      {
         throw new InvalidXUIDException("Base64 decoding failed", x);
      }
   }

   /**
    * Build a XUID from a byte array.
    * @param abyte0 The array of bytes of the XUID
    */
   public static XUID valueOf(byte[] bytes)
      throws InvalidXUIDException
   {
      if (bytes.length < XUID.MIN_LENGTH)
         throw new InvalidXUIDException("byte[] too short " + bytes.length);
      int ptr = 0;
      int version = (int) (bytes[ptr++] & 0xff);
      if (version != 0)
         throw new InvalidXUIDException("Incorrect version " + version);
      // Skip vendor and reserved
      ptr += 4;
      int len = (int) (bytes[ptr++] & 0xff);
      if (len > bytes.length)
         throw new InvalidXUIDException("Encoded length incorrect " + len);
      checkCrc(bytes);
      return new DefaultXUID(bytes, 0, len);
   }

   /**
    * Get the raw array of bytes representing this XUID
    * @return The byte array.
    */
   public byte[] toBytes()
   {
      return m_bytes;
   }

   /**
    * Convert this XUID to a base64 encoded String
    * @return The base 64 version of this XUID
    */
   public String toString()
   {
      if (m_base64 == null)
      {
         m_base64 = new String(Base64.base64Encode(m_bytes));
      }
      return m_base64;
   }

   /**
    * Perform a validity check on this XUID
    * @return True if this XUID value is properly formed.
    */
   public boolean isValid()
   {
      return true;
   }

   /**
    * Compare two XUID values.
    * @return True if the two values are the same
    */
   public boolean equals(Object obj)
   {
      if (obj instanceof XUID)
         return toString().equals(((XUID) obj).toString());
      return false;
   }

   public int hashCode()
   {
      return toString().hashCode();
   }

   public int compareTo(XUID obj)
   {
      return toString().compareTo(obj.toString());
   }

   private static void checkCrc(byte[] bytes)
      throws InvalidXUIDException
   {
      byte msb = bytes[CRC_OFFSET];
      byte lsb = bytes[CRC_OFFSET + 1];
      int crc = ((msb & 0xff) << 8) | (lsb & 0xff);

      byte length = bytes[LENGTH_OFFSET];

      try
      {
         bytes[CRC_OFFSET] = 0;
         bytes[CRC_OFFSET + 1] = 0;
         if (crc != s_crc.getCrc(bytes, 0, length))
            throw new InvalidXUIDException("CRC check failed");
      }
      finally
      {
         bytes[CRC_OFFSET] = msb;
         bytes[CRC_OFFSET + 1] = lsb;
      }
   }
}
