/* Copyright (c) 2008, The Storage Networking Industry Association.

 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */

package org.snia.xam.base;

import java.util.Calendar;

import org.snia.xam.InvalidFieldTypeException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XUID;
import org.snia.xam.util.ISO8601Date;

/**
 * {@link DefaultField} sub-class that represents a XAM property field.
 */
public class Property
   extends DefaultField
{
   public String[] VALID_TYPES = {
      XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
      XAMLibrary.STYPE_DATETIME_MIME_TYPE, 
      XAMLibrary.STYPE_DOUBLE_MIME_TYPE,
      XAMLibrary.STYPE_INT_MIME_TYPE, 
      XAMLibrary.STYPE_STRING_MIME_TYPE,
      XAMLibrary.STYPE_XUID_MIME_TYPE};

   private Object m_value;

   /**
    * Constructs a new XAM property with the specified name, mime-type and
    * value. If the value is not of the appropriate class an
    * {@link InvalidFieldTypeException} is thrown.
    * 
    * @param name The property name
    * @param type The property mime-type
    * @param value The property value
    * @throws XAMException If the mime-type is invalid or the value does not
    *            match the mime-type
    */
   public Property(String name, String type, Object value)
      throws XAMException
   {
      super(name);
      setValue(value, type);
   }

   /**
    * Constructs a new XAM property with the specified name, mime-type, value,
    * binding state and read state. If the value is not of the appropriate class
    * an {@link InvalidFieldTypeException} is thrown.
    * 
    * @param name The property name
    * @param type The property mime-type
    * @param value The property value
    * @param binding The property binding state
    * @param readonly The property read state
    * @throws XAMException If the mime-type is invalid or the value does not
    *            match the mime-type
    */
   public Property(String name, String type, Object value, boolean binding,
      boolean readonly)
      throws XAMException
   {
      super(name);
      setValue(value, type);
      setBinding(binding);
      setReadonly(readonly);
   }

   /**
    * Constructs a new XAM property from an existing property.
    * 
    * @param prop An existing property
    * @throws XAMException If the mime-type is invalid or the value does not
    *            match the mime-type
    */
   public Property(Property prop)
      throws XAMException
   {
      this(prop.getName(), prop.getType(), prop.getValue(), prop.getBinding(),
         prop.getReadOnly());
   }

   public Object getValue()
   {
      return m_value;
   }

   public void setValue(Object value, String type)
      throws XAMException
   {
      validateType(type);
      setType(type);
      m_value = value;
      if (type.equals(XAMLibrary.STYPE_BOOLEAN_MIME_TYPE)) setLength(1L);
      else if (type.equals(XAMLibrary.STYPE_DATETIME_MIME_TYPE))
      {
         Calendar c = (Calendar) value;
         setLength(ISO8601Date.toISO8601String(c).length());
      }
      else if (type.equals(XAMLibrary.STYPE_DOUBLE_MIME_TYPE)) setLength(8L);
      else if (type.equals(XAMLibrary.STYPE_INT_MIME_TYPE)) setLength(8L);
      else if (type.equals(XAMLibrary.STYPE_STRING_MIME_TYPE)) setLength((m_value == null
         ? 0L
         : ((String) m_value).length()));
      else if (type.equals(XAMLibrary.STYPE_XUID_MIME_TYPE))
         setLength((m_value == null ? 0L : ((XUID) m_value).toBytes().length));
   }

   protected void validateType(String type)
      throws XAMException
   {
      for (int i = 0; i < VALID_TYPES.length; i++)
         if (VALID_TYPES[i].equals(type)) return;
      throw new InvalidFieldTypeException(type);
   }
}
