/* Copyright (c) 2008-2009, The Storage Networking Industry Association.

 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */
/* (c) Copyright (c) 2009, Sun Microsystems, Inc.               */

package org.snia.xam.base;

import java.io.FileInputStream;
import java.io.IOException;
import java.text.ParseException;
import java.util.Enumeration;
import java.util.Properties;

import java.util.Collections;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.FieldContainer;
import org.snia.xam.FieldExistsException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidFieldNameException;
import org.snia.xam.InvalidOperationException;
import org.snia.xam.MaximumFieldException;
import org.snia.xam.ObjectInUseException;
import org.snia.xam.XAMException;
import org.snia.xam.XSetAbandonException;
import org.snia.xam.XSetCorruptException;
import org.snia.xam.XSystemAbandonException;
import org.snia.xam.XSystemCorruptException;
import org.snia.xam.util.ISO8601Date;

/**
 * This worker class is used by the XAMLibrary to load the XAM configuration
 * file. It is then later used to determine what the actual class is to be
 * instantiated for a VIM name. Additional parameters to control logging may be
 * inserted into this configuration file. When the XAMLibrary implementation is 
 * instantiated and uses this configuration file, default values from the configuration
 * file (properties) are transfered to the library. To support this, the library must 
 * insert default values of the correct type into the field container prior to 
 * loading the configuration with this class.
 * 
 * The structure of this file is specified by the SNIA XAM SDK TWG paper
 * "Stackable Vendor Interface Module".
 */
public class XAMConfiguration
{
   private static String ALIAS = ".xam.config.vim.alias.";

   private static String ORDER = ".xam.config.vim.lookuporder.";

   private static String PRELOAD = ".xam.config.vim.preload";
   
   // Type tags, used prior to property name
   public static final String TYPE_INT      = "xam_int.";
   public static final String TYPE_DOUBLE   = "xam_double.";
   public static final String TYPE_BOOLEAN  = "xam_boolean.";
   public static final String TYPE_DATETIME = "xam_datetime.";
   public static final String TYPE_XUID     = "xam_xuid.";
   public static final String TYPE_STRING   = "xam_string.";

   private Properties m_vimmap;

   private Logger m_logger;

   /**
    * Creates a new, empty VIM configuration object.
    */
   public XAMConfiguration( Logger externalLogger )
   {
      m_vimmap = new Properties();
      m_logger = externalLogger;
   }

   /**
    * Read the VIM configuration from a Properties object
    * 
    * @param vimconfig A Properties object containing the vim configuration
    */
   public void load(Properties vimconfig, FieldContainer fc)
      throws XAMException
   {
      m_vimmap.putAll(vimconfig);
      copyToFieldContainer(vimconfig, fc);
      logConfig();
   }
   
   public void setLogger( Logger externalLogger )
   {
      m_logger = externalLogger;
   }

   /**
    * Read the VIM configuration from a file.
    * 
    * @param configFilePath Path to the VIM configuration file0.
    * @throws IOException Possible exceptions thrown if the file does not exist,
    *            or is improperly formatted.
    */
   public void load(String configFilePath, FieldContainer fc)
      throws XAMException, IOException
   {
      Properties props = new Properties();
      props.load(new FileInputStream(configFilePath));
      m_vimmap.putAll(props);
      copyToFieldContainer(props, fc);
      logConfig();
   }

   /**
    * Return the Java class name for a specific VIM name.
    * 
    * @param vimName The name of the VIM
    * @return the classname of the VIM.
    */
   public String classForVIMName(String vimName)
   {
      String vimKey = ALIAS + vimName;
      return (String) m_vimmap.getProperty(vimKey);
   }

   /**
    * Return the ordered list of VIM names. Used by the XAMLibrary when it is
    * starting up. The specification allows ordered lists, and this
    * implementation will allow missing index entries. For instance if the vims
    * are listed as VIM1=1, VIM2=2, VIM3=4, the list place VIM3 at position 3.
    * If the configuration file supplies duplicate entries (VIM1=1, VIM2=2,
    * VIM3=2), then the list will contain VIM3 after VIM2. With multiple
    * occurences of the same index value, the vim names are gauranteed to be
    * placed earlier in the list than larger indexes. In otherwords ordering
    * with multiple items having the same index lookup order is not specified.
    * 
    * @return The ordered array of VIM Names.
    */
   public String[] getLookupOrder()
   {
      ArrayList<String> vimNames = new ArrayList<String>();

      Enumeration e = m_vimmap.keys();
      while (e.hasMoreElements())
      {
         String key = (String) e.nextElement();
         if (key.startsWith(ORDER))
         {
            String value = m_vimmap.getProperty(key);
            vimNames.add(key + "=" + value);
         }
      }
      Collections.sort(vimNames);
      int length = vimNames.size();
      String names[] = new String[length];
      for (int x = 0; x < length; x++)
      {
         String raw = (String) vimNames.get(x);
         int orderIdx = raw.indexOf("=");
         String cooked = raw.substring(orderIdx + 1);
         names[x] = cooked;
      }
      return names;
   }

   /**
    * Return the ordered list of VIM names to be preloaded. The XAM Library is
    * expected to preload these VIMs prior to the first connect call.
    * 
    * @return The ordered array of VIM Names.
    */
   public String[] getPreloadOrder()
   {
      ArrayList<String> vimNames = new ArrayList<String>();

      Enumeration e = m_vimmap.keys();
      while (e.hasMoreElements())
      {
         String key = (String) e.nextElement();
         if (key.startsWith(PRELOAD))
         {
            String value = m_vimmap.getProperty(key);
            vimNames.add(key + "=" + value);
         }
      }
      Collections.sort(vimNames);
      int length = vimNames.size();
      String names[] = new String[length];
      for (int x = 0; x < length; x++)
      {
         String raw = (String) vimNames.get(x);
         int orderIdx = raw.indexOf("=");
         String cooked = raw.substring(orderIdx + 1);
         names[x] = cooked;
      }
      return names;
   }

   private void setPropertyFromConfig( Properties     props,
                                       FieldContainer fc,
                                       String         key) 
      throws InvalidFieldNameException, FieldExistsException, MaximumFieldException, XSetAbandonException, 
             XSetCorruptException, XSystemAbandonException, XSystemCorruptException, AuthenticationExpiredException, 
             InvalidArgumentException, InvalidOperationException, ObjectInUseException, XAMException
   {
      String sValue = props.getProperty(key);
      
      // Just in case this is a typed value, strip the type prior to need.
      String strippedKey = key.substring( key.indexOf(".")+1);
      boolean update = fc.containsField(strippedKey);
      
      try
      {
         if( key.startsWith(TYPE_INT))
         {
            if( update )
               fc.setProperty( strippedKey, Long.parseLong(sValue));
            else
               fc.createProperty( strippedKey, false, Long.parseLong(sValue) );
         }
      
         else if( key.startsWith(TYPE_DOUBLE))
         {
            if( update )
               fc.setProperty( strippedKey, Double.parseDouble(sValue));
            else
               fc.createProperty( strippedKey, false, Double.parseDouble(sValue) );
         }
      
         else if( key.startsWith(TYPE_BOOLEAN))
         {
            if( update )
               fc.setProperty( strippedKey, Boolean.parseBoolean(sValue) );
            else
               fc.createProperty( strippedKey, false, Boolean.parseBoolean(sValue) );
         }
      
         else if( key.startsWith(TYPE_DATETIME))
         {
            if( update )
            fc.setProperty( strippedKey, ISO8601Date.toCalendar(sValue));
            else
               fc.createProperty( strippedKey, false, ISO8601Date.toCalendar(sValue));
         }
      
         else if( key.startsWith(TYPE_XUID))
         {
            if( update )
               fc.setProperty( strippedKey, DefaultXUID.valueOf(sValue) );
            else
               fc.createProperty( strippedKey, false, DefaultXUID.valueOf(sValue));
         }
      
         else if( key.startsWith(TYPE_STRING))
         {
            if( update )
               fc.setProperty( strippedKey, sValue );
            else
               fc.createProperty( strippedKey, false, sValue );
         }
         else 
         {
            if( fc.containsField(key) )
               fc.setProperty( key, sValue );
            else
               fc.createProperty( key, false, sValue );
         }
      }
      catch( ParseException pe )
      {
         String msg = "Unable to parse argument '" + sValue + "' for property " + key;
         m_logger.severe( msg );
         throw new InvalidArgumentException( msg );
      }
   }

   /**
    * Copies all the configuration parameters to the supplied FieldContainer, if the value exists 
    * inside the field container. This means that only those values which are defaulted by the 
    * field container are transfered. All others are ignored.
    * (presumably a XAMLibrary).
    */
   private void copyToFieldContainer(Properties p, FieldContainer fc)
      throws XAMException
   {
      // Copy known fields from the config
      Enumeration e = p.keys();
      while( e.hasMoreElements() )
      {
         String key = (String) e.nextElement();
         setPropertyFromConfig( p, fc, key );
      }      
   }

   /**
    * Configure the logging values based on what was specified.
    */
   private void logConfig()
   {
      if (m_logger.isLoggable(Level.INFO))
      {
         Enumeration iter = m_vimmap.propertyNames();
         StringBuffer configLog = new StringBuffer( "Configuration File Entries\n" );
         while (iter.hasMoreElements())
         {
            String name = (String) iter.nextElement();
            String value = m_vimmap.getProperty(name);
            configLog.append(name);
            configLog.append("=");
            configLog.append(value);
            configLog.append("\n");
         }
         m_logger.info( configLog.toString() );
      }
   }
}
