/* Copyright (c) 2008-2009, The Storage Networking Industry Association.

 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */
/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */
package org.snia.xam.base;

import java.io.IOException;
import java.util.Iterator;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.FieldContainer;
import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.FieldExistsException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidFieldNameException;
import org.snia.xam.InvalidOperationException;
import org.snia.xam.MaximumFieldException;
import org.snia.xam.ObjectInUseException;
import org.snia.xam.VIMLoadException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XSetAbandonException;
import org.snia.xam.XSetCorruptException;
import org.snia.xam.XStream;
import org.snia.xam.XSystem;
import org.snia.xam.XSystemAbandonException;
import org.snia.xam.XSystemCorruptException;
import org.snia.xam.util.LogManager;
import org.snia.xam.vim.VIM;

/**
 * This class represents the entry point into the XAM library. The mapping from
 * the VIM name specified in an XRI to the implementing Java class is done using
 * Properties. For each VIM a property must be specified where the key is the
 * VIM name and the value is the fully qualifed VIM class name. For example, if
 * the VIM name is <b>AcmeVim</b> and the class implementing the VIM is
 * <b>com.acme.xam.vim.VIM</b> then the following property must be supplied:
 * 
 * <pre>
 * AcmeVim = com.acme.xam.vim.VIM
 * </pre>
 * 
 * Constructors are provided that take either a Properties object or a
 * properties file.
 * 
 * Applications should only create a single instance of this library. There are
 * logging and clean up issues to deal with, and these are better handled with a single
 * instance. An instance of the library may instantiate additional worker object which 
 * can consume resources and impact the host (threads, etc). 
 */
public class XAMImplementation
   extends AbstractFieldContainer
   implements XAMLibrary
{
   private static final String ENV_CONFIG_PATH = "XAM_CONFIG_PATH";
   private static final String DEFAULT_CONFIG_PATH = "xam.properties";
   
   private static final String IDENTITY = "SNIA XAM Java Library v1.0";

   private static final String API_VERSION = "01.00.00";

   private static final String  DEFAULT_LOG_PATH = "SNIA-XAM.log";
   private static final int     DEFAULT_LOG_SIZE = 10; // kB
   private static final int     DEFAULT_LOG_COUNT = 1;
   private static final boolean DEFAULT_APPEND = false;
   
   private long m_debug;

   private Logger m_logger;

   // State variables describing the logging configuration. If these change, a new 
   // logger is going to need to be created;
   private String  m_logPath;
   private int     m_logSize  = DEFAULT_LOG_SIZE;  // kB
   private int     m_logCount = DEFAULT_LOG_COUNT; // # log files to keep.
   private boolean m_enterExit = false;            // Log enter and exit traces
   
   private static String s_logClass = "org.snia.xam.xam.base.XAMImplementation";
   
   // Used when loading properties from the XAMConfig file, prevents certain updates from occuring.
   private boolean m_initializing = false; 

   private static XAMImplementation s_library;
   
   private XAMConfiguration m_config;

   /**
    * Creates a new XAM Library and initializes default properties.
    * 
    * @throws XAMException
    */
   public XAMImplementation()
      throws XAMException
   {
      createDefaultProperties();
      updateLogger();
      m_config = new XAMConfiguration(m_logger);
  }
   
   /**
    * Methods to support XAMLibraryFactory. This method, and the next will return a singleton library class.
    */
   public static XAMLibrary getLibrary() throws XAMException
   {
      if( s_library == null )
      {
         String configPath = System.getenv(ENV_CONFIG_PATH);
         if( (configPath == null) || configPath.equals("") )
            configPath = DEFAULT_CONFIG_PATH;
         s_library = new XAMImplementation(configPath);
      }
      return s_library;
   }

   /**
    * Methods to support XAMLibraryFactory. This method, and the next will return a singleton library class.
    */
   public static XAMLibrary getLibrary(Properties props)
      throws XAMException
   {
      if (s_library == null)
      {
         s_library = new XAMImplementation();
         s_library.loadVIMConfig(props);
      }
      return s_library;
   }

   /**
    * Creates a new XAM Library, initializes the default properties and load the
    * VIM configuration from a file. This is exactly the same as calling
    * {@link #XAMImplementation()} followed by {@link #loadVIMConfig(String)}.
    * 
    * @param configFile A property file containing the VIM configuration
    * @throws XAMException IF an error occurs reading the property file
    */
   public XAMImplementation(String configFile)
      throws XAMException
   {
      this();
      loadVIMConfig(configFile);
   }

   /**
    * Loads the VIM configuration from a properties file.
    * 
    * @param configFile A property file path
    * @throws IOException If an error occurs reading the property file
    */
   public void loadVIMConfig(String configFile)
      throws XAMException
   {
      try
      {
         m_initializing = true;
         setProtectNamespace(false);
         m_config.load(configFile, this);
         setProtectNamespace(true);
         m_initializing = false;
         
         updateLogger();
         
         if (m_logger != null)
         {
            m_logger.log(Level.INFO,
               "Loaded library configuration from file '" + configFile + "'");
         }
      }
      catch (Exception e)
      {
         throw new XAMException("Error reading the VIM configuration file :"
            + configFile + ": ", e, -1);
      }
   }

   /**
    * Loads the VIM configuration from a Properties object.
    * @param vimconfig Properties object containing the VIM configuration
    */
   public void loadVIMConfig(Properties vimconfig)
      throws XAMException
   {
      m_initializing = true;
      setProtectNamespace(false);
      m_config.load(vimconfig, this);
      setProtectNamespace(true);
      m_initializing = false;
      
      updateLogger();
      
      if (m_logger != null)
      {
         m_logger.log(Level.INFO,
            "Loading library configuration from application supplied properties");
      }
   }

   /**
    * Connect to the specified XAM archive and return am XSystem object to allow
    * interaction. XRI values are in the form
    * "snia-xam://[vimname!]xsystemname[?param=value[{&param=value}]]"
    * 
    * @param xriString The XRI string
    * @return The XSystem object representing the connection
    * @throws XAMException If an error occurs connecting to the XSystem
    */
   public XSystem connect(String xriString)
      throws XAMException
   {
      final String CONNECT_MTHD = "connect";
      
      m_logger.entering( s_logClass, CONNECT_MTHD );
      XRI xri = XRI.valueOf(xriString);
      String vimName = xri.getVimInfo();
      if ((vimName == null) || (vimName.length() == 0))
         throw new VIMLoadException(
            "No VIM specified for storage system named '" + xri.getSystem()
               + "'");
      VIM vim = loadVIM(vimName);
      if (vim == null)
         throw new VIMLoadException("No VIM found for " + vimName);
      XSystem xsys = vim.createXSystem();
      copyFields(xsys);
      xsys.deleteField(XSystem.XAM_XSYSTEM_INITIALIZING);
      xsys.connect(xriString);
      m_logger.exiting( s_logClass, CONNECT_MTHD );
      return xsys;
   }

   /**
    * Updates the XAM Library logger to match the values set in the library
    * properties.
    */
   protected void updateLogger()
      throws XAMException
   {
      if( m_initializing )
         return;
      
      // Read the log configuration from the property files.
      String path   = null;
      long   lsize  = DEFAULT_LOG_SIZE;
      long   lcount = DEFAULT_LOG_COUNT;
      long   level  = XAMLibrary.XAM_LOG_FATAL;
     
      path = getString(XAMLibrary.XAM_LOG_PATH);
      level = getLong(XAMLibrary.XAM_LOG_LEVEL);
      m_enterExit = getLong(XAMLibrary.XAM_LOG_VERBOSITY) > 1;
      
      try { m_debug = (int) getLong(XAMLibrary.XAM_LOG_VERBOSITY); }
      catch (FieldDoesNotExistException x) { }
      
      try 
      { 
         lsize = getLong( XAM_LOG_MAX_SIZE ) * 1000; // The property is in kB 
         if( (lsize > Integer.MAX_VALUE) || (lsize < 0) )
            throw new InvalidArgumentException( "Log file size is too large, or less than zero" );
      }
      catch( FieldDoesNotExistException x ) { }
      
      try 
      { 
         lcount = getLong( XAM_LOG_MAX_ROLLOVERS ); 
         if( (lcount > Integer.MAX_VALUE) || (lcount < 0) )
            throw new InvalidArgumentException( "Log count is too large, or less than zero" );
      }
      catch(FieldDoesNotExistException x) { }
      
      boolean append = getBoolean( XAMImplementation.XAM_LOG_APPEND );
      m_logger = LogManager.GetLogger(path, (int) lsize, (int) lcount, append);
      m_logger.setLevel(LogManager.mapLogLevelToJavaLevel(level));
      m_logPath = path;
      
      if( m_config != null )
         m_config.setLogger(m_logger);
   }

   /**
    * Used to copy the XAMLibrary fields to the XSystem instance for
    * initialization.
    * 
    * @param dest
    */
   private void copyFields(FieldContainer dest)
      throws XAMException
   {
      // Copy the XAMLibrary fields to the XSystem.
      Iterator fi = this.openFieldIterator("");
      while (fi.hasNext())
      {
         String fieldName = (String) fi.next();
         boolean binding = getFieldBinding(fieldName);
         String type = getFieldType(fieldName);
         if (type.equals(XAMLibrary.STYPE_STRING_MIME_TYPE)) 
            try { dest.createProperty(fieldName, binding, getString(fieldName)); }
            catch( FieldExistsException fe ) { dest.setProperty( fieldName, getString(fieldName)); } 
            
         else if (type.equals(XAMLibrary.STYPE_DATETIME_MIME_TYPE)) 
            try { dest.createProperty(fieldName, binding, getDateTime(fieldName));}
            catch( FieldExistsException fe ) { dest.setProperty( fieldName, getDateTime(fieldName )); }
            
         else if (type.equals(XAMLibrary.STYPE_XUID_MIME_TYPE)) 
            try { dest.createProperty(fieldName, binding, getXUID(fieldName)); }
            catch( FieldExistsException fe ) { dest.setProperty( fieldName, getXUID(fieldName)); } 
            
         else if (type.equals(XAMLibrary.STYPE_BOOLEAN_MIME_TYPE)) 
            try { dest.createProperty(fieldName, binding, getBoolean(fieldName)); }
            catch( FieldExistsException fe ) { dest.setProperty( fieldName, getBoolean(fieldName)); }
            
         else if (type.equals(XAMLibrary.STYPE_INT_MIME_TYPE)) 
            try { dest.createProperty(fieldName, binding, getLong(fieldName)); }
            catch( FieldExistsException fe ) { dest.setProperty( fieldName, getLong(fieldName)); } 
            
         else if (type.equals(XAMLibrary.STYPE_DOUBLE_MIME_TYPE)) 
            try { dest.createProperty(fieldName, binding, getDouble(fieldName)); }
            catch( FieldExistsException fe ) { dest.setProperty( fieldName, getDouble(fieldName)); }
            
         else
            throw new XAMException("Unknown type: copyFields() " + type);
      }
   }

   /**
    * Loads the named VIM class, returning a reference to the VIM. The name is
    * translated using the VIMConfig object.
    */
   private VIM loadVIM(String vimName)
      throws VIMLoadException
   {
      String vimClass = m_config.classForVIMName(vimName);
      VIM vim = null;
      if (vimClass == null)
         throw new VIMLoadException("Can't find the class for the VIM '"
            + vimName + "', do you have a line with"
            + " '.xam.config.vim.alias." + vimName + "' in your config?");
      try
      {
         Class cl = Class.forName(vimClass);
         vim = (VIM) cl.newInstance();
         addField(new Property(XAM_VIM_LIST + "." + vimName,
            STYPE_STRING_MIME_TYPE, vimName, false, true));
      }
      catch (Exception x)
      {
         // Log VIM load failure
         if (m_logger != null)
            m_logger.log(Level.SEVERE, "Failed to load VIM " + vimName
               + ", class from config=" + vimClass, x);
      }
      return vim;
   }

   protected synchronized void setProperty(String fieldName,
                                           Object value,
                                           String type)
      throws XAMException
   {
      super.setProperty(fieldName, value, type);
      if (XAM_LOG_LEVEL.equals(fieldName))
      {
         updateLogger();
      }
      else if (XAM_LOG_VERBOSITY.equals(fieldName))
      {
         updateLogger();
      }
      else if (XAM_LOG_PATH.equals(fieldName))
      {
         updateLogger();
      }
   }

   private void createDefaultProperties()
      throws XAMException
   {
      addField(new Property(XAMLibrary.XAM_IDENTITY, STYPE_STRING_MIME_TYPE,
         IDENTITY, false, true));
      addField(new Property(XAMLibrary.XAM_API_LEVEL, STYPE_STRING_MIME_TYPE,
         API_VERSION, false, true));
      addField(new Property(XAMLibrary.XAM_LOG_LEVEL, STYPE_INT_MIME_TYPE,
         new Long(m_debug), false, false));
      addField(new Property(XAMLibrary.XAM_LOG_PATH, STYPE_STRING_MIME_TYPE,
         DEFAULT_LOG_PATH, false, false));
      addField(new Property(XAMLibrary.XAM_LOG_VERBOSITY, STYPE_INT_MIME_TYPE,
         new Long(1), false, false));
      addField(new Property(XAMLibrary.XAM_LOG_LEVEL, STYPE_INT_MIME_TYPE,
         new Long(XAMLibrary.XAM_LOG_FATAL), false, false));
      addField(new Property(XAMImplementation.XAM_LOG_MAX_SIZE, STYPE_INT_MIME_TYPE,
         new Long(XAMImplementation.DEFAULT_LOG_SIZE), false, false ));
      addField(new Property(XAMImplementation.XAM_LOG_MAX_ROLLOVERS, STYPE_INT_MIME_TYPE,
         new Long(XAMImplementation.DEFAULT_LOG_COUNT), false, false ));
      addField(new Property(XAMImplementation.XAM_LOG_APPEND, STYPE_BOOLEAN_MIME_TYPE,
         new Boolean(XAMImplementation.DEFAULT_APPEND), false, false ));
   }

   public XAsync asyncOpenXStream(String name,
                                  String mode,
                                  long xopid,
                                  XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      InvalidFieldNameException, FieldDoesNotExistException,
      ObjectInUseException, XSetAbandonException, XSetCorruptException,
      XSystemAbandonException, XSystemCorruptException, XAMException
   {
      throw new InvalidOperationException(
         "This library does not support async operations... yet");
   }

   public XStream createXStream(String fieldName,
                                boolean binding,
                                String mimeType)
      throws AuthenticationExpiredException, InvalidArgumentException,
      InvalidFieldNameException, InvalidOperationException,
      FieldExistsException, MaximumFieldException, ObjectInUseException,
      XSetAbandonException, XSetCorruptException, XSystemAbandonException,
      XSystemCorruptException, XAMException
   {
      throw new InvalidOperationException(
         "This library does not support creating XStreams");
   }

   public XStream openXStream(String fieldName, String mode)
      throws AuthenticationExpiredException, InvalidFieldNameException,
      FieldDoesNotExistException, ObjectInUseException, XSetAbandonException,
      XSetCorruptException, XSystemAbandonException, XSystemCorruptException,
      XAMException
   {
      throw new InvalidOperationException(
         "This library has no XStreams to open");
   }
   
   /**
    * Close this XAM Library instance.
    * 
    * @throws AuthenticationExpiredException
    * @throws XAMException
    *             <DT><b>Thread Safety:</b>
    *             <DD>This method is thread safe
    *             <DT><b>Blocking:</b>
    *             <DD>This method blocks until completion
    */
   public void close()
      throws AuthenticationExpiredException, ObjectInUseException,
             XAMException
   {
      m_logger.log(Level.INFO, "Closing the Java XAM Library" );
      s_library = null;
      m_debug = XAMLibrary.XAM_LOG_NONE;
      m_logger = null;
   }
   
   /**
    * This finalizer will attempt to close the library instance, causing logging to close properly.
    * Because finalizers are not gauranteed, application implementers should not rely upon this to clean
    * up logging. If the application programmers are willing to accept non-portable code, they can 
    * call XAMImplementation.close() directly. 
    */
   protected void finalize() throws Throwable
   {
      try
      {
         this.close();
      }
      finally
      {
         super.finalize();
      }
   }
}
