/* Copyright (c) 2008, The Storage Networking Industry Association.

 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */
/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */
package org.snia.xam.base;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.snia.xam.InvalidXRIException;

/**
 * A convenience class for parsing XAM XRI strings. A XAM XRI string is of the
 * form:
 * 
 * <pre>
 * snia-xam://[&lt;vim name&gt;]!&lt;system&gt;[?&lt;param&gt;[&amp;param]]
 * vim_name = The name of the VIM to use (optional)
 * system   = The archive system to connect to
 * param    = Optional &amp; separated parameter list where each
 *            parameter is of the form:
 *            name=value
 * 
 */
public class XRI
{
   /** SNIA XAM XRI prefix */
   public static final String PREFIX = "snia-xam://";

   protected static final String PARAM_PATTERN = "[\\?\\&]([^=^&]+)\\=([^=^&]*)";

    // Match any set of characters, EXCEPT for the separator "!"
   protected static final String VIM_PATTERN = "(?:([^!]+)\\!)?";
   
    // Match any set of characters, EXCEPT for the separator "!" and "?"
   protected static final String SYSTEM_PATTERN = "([^?^!]*){1}";
   
   protected static final String XRI_PARAM_PATTERN = "(\\?[^=^&]+=[^=^&]*(?:\\&[^=^&]+\\=[^=^&]*)*)?";
   
   protected static final String XRI_PATTERN = PREFIX + VIM_PATTERN + SYSTEM_PATTERN + XRI_PARAM_PATTERN;

   protected static final Pattern s_xri = Pattern.compile(XRI_PATTERN);

   protected static final Pattern s_params = Pattern.compile(PARAM_PATTERN);

   private String m_vim;

   private String m_system;

   private Map<String,String> m_params;

   /**
    * Creates an XRI instance from an XRI string. If the string does not contain
    * a valid XRI an {@link InvalidXRIException} is thrown.
    * 
    * @param xriString The XRI string
    * @return An XRI instance
    * @throws InvalidXRIException If the string does not contain a valid XRI
    */
   public static XRI valueOf(String xriString)
      throws InvalidXRIException
   {
       try 
	   {
	   xriString = java.net.URLDecoder.decode(xriString, "UTF-8" );
       } 
       catch( Exception e ) 
       { throw new InvalidXRIException( e + xriString ); }

      Matcher m = s_xri.matcher(xriString);

      if (!m.matches()) throw new InvalidXRIException(xriString);

      XRI xri = new XRI();
      xri.setVimInfo(m.group(1));
      xri.setSystem(m.group(2));
      if( (xri.getSystem() == null) || xri.getSystem().equals("") )
	  throw new InvalidXRIException( "Null systemname not allowed in XRI" );

      String parameters = m.group(3);
      if (parameters != null)
      {
         m = s_params.matcher(parameters);
         while (m.find())
            xri.addParameter(m.group(1), m.group(2));
      }
      return xri;
   }

   /**
    * Constructor to create a blank XRI.
    */
   public XRI()
   {
      m_params = new HashMap<String,String>();
   }

   /**
    * Returns the list of parameter names contained in the XRI as a list of
    * String objects.
    * 
    * @return The list of parameter name
    */
   public List<String> getParameterNames()
   {
      return new ArrayList<String>(m_params.keySet());
   }

   /**
    * Returns the parameter value for the specified parameter name as a String.
    * Returns null if the XRI does not contain the parameter.
    * 
    * @param name The parameter name
    * @return The parameter value as a String, null if it doesn't exist.
    */
   public String getParameterValue(String name)
   {
      return (String) m_params.get(name);
   }

   /**
    * Returns the system name specified in the XRI.
    * 
    * @return The system name
    */
   public String getSystem()
   {
      return m_system;
   }

   /**
    * Returns the VIM name specified in the XRI or null if not specified.
    * 
    * @return The VIM name if specified, null otherwise
    */
   public String getVimInfo()
   {
      return m_vim;
   }

   /**
    * Converts the XRI into a properly formatted String.
    * 
    * @return The XRI as a String.
    */
   public String toString()
   {
      StringBuffer buf = new StringBuffer(80);
      buf.append(PREFIX);
      if (m_vim != null)
      {
         buf.append(m_vim);
         buf.append('!');
      }
      buf.append(m_system);
      boolean first = true;
      Iterator<String> iter = m_params.keySet().iterator();
      while (iter.hasNext())
      {
         if (first)
         {
            buf.append('?');
            first = false;
         }
         else
            buf.append('&');
         String name = iter.next();
         buf.append(name);
         buf.append('=');
         buf.append((String) m_params.get(name));
      }
      return buf.toString();
   }

   /**
    * Adds a parameter name value pair to the XRI.
    * 
    * @param name The parameter name
    * @param value The parameter value
    */
   public void addParameter(String name, String value)
   {
      m_params.put(name, value);
   }

   /**
    * Sets the VIM name for the XRI.
    * 
    * @param vimInfo The VIM name.
    */
   public void setVimInfo(String vimInfo)
   {
      m_vim = vimInfo;
   }

   /**
    * Sets the system name for the XRI.
    * 
    * @param system The system name
    */
   public void setSystem(String system)
   {
      m_system = system;
   }
}
