/* Copyright (c) 2008-2009, The Storage Networking Industry Association.

   Redistribution and use in source and binary forms, with or without modification, 
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.
   
   * Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.
   
   * Neither the name of The Storage Networking Industry Association (SNIA) nor 
   the names of its contributors may be used to endorse or promote products 
   derived from this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
   THE POSSIBILITY OF SUCH DAMAGE.
*/

/* (c) Copyright 2008 Sun Microsystems */

package org.snia.xam.util;

import java.io.IOException;
import java.util.Hashtable;
import java.util.logging.FileHandler;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.logging.SimpleFormatter;

import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;

public class LogManager
{
   private static Hashtable<String,Logger> s_cache = new Hashtable<String,Logger>();
   
   // Default log values to be used for logging, if not specified.
   final static boolean APPEND         = false;
   final static int     ROLLOVER_COUNT = 1;
   final static int     LOG_SIZE       = 10000;
   
   public final static int     DEFAULT_LOG_SIZE = 10; // kB
   public final static int     DEFAULT_LOG_COUNT = 1;


   /**
    * Obtain the logger matching the path. Also allows specification of the logger's behavioral
    * parameters. When the logger is obtained, or created, these parameters are used to control 
    * the behavior of the logger. If the logger is to be referenced without modifying the behavior,
    * use the non-parameterized version of the GetLogger(path) call. 
    * @param path The store path associated with this logger instance.
    * @param logSize The log size in bytes
    * @param rolloverCount The number of retained log files when they are rolled.
    * @param append If the current log file should be appended to.
    * @return The logger referred to by the path argument.
    * @throws XAMException
    */
   public static Logger GetLogger( String  path,
                                   int     logSize,
                                   int     rolloverCount,
                                   boolean append ) 
      throws XAMException
   {
      synchronized(s_cache)
      {
         Logger xamLogger = s_cache.get(path);
         if( xamLogger != null )
            UpdateHandlerSettings( xamLogger, path, logSize, rolloverCount, append );
         else
         {
            xamLogger = Logger.getLogger(path);
            s_cache.put( path, xamLogger );
            xamLogger.setUseParentHandlers(false);
                  
            // XAM Spec doesn't provide for multiple files per XAMLibrary Instance. 
            // Remove all existing file handlers and replace with one created exactly for
            // the XSystem instance. This may be because this method has been called
            // in response to changing a logging parameter.
            Handler hList[] = xamLogger.getHandlers();
            for( int h = hList.length-1; h >= 0; h-- )
               if( hList[h] instanceof FileHandler )
               {
                  xamLogger.removeHandler(hList[h]);
               }
            
            try
            {
               FileHandler logHandler = new FileHandler( path, logSize, rolloverCount, append);
               logHandler.setFormatter(new SimpleFormatter());
               xamLogger.addHandler(logHandler);
            }
            catch (IOException x)
            {
               throw new XAMException("Failed to setup XAM Library logging", x, -1);
            }
         }
         return xamLogger;
      }
   }

   /**
    * Obtains the instance of the logger associated with the path. Does not update any logger settings.
    * If the logger is to be created as a result of this call, the specified XAM logging defaults will
    * be used. If later calls to the parameterized version of GetLogger is called, the settings will 
    * be updated by that call.
    * @param path The path of the store location.
    * @return The appropriate logger.
    * @throws XAMException
    */
   public static Logger GetLogger( String  path ) 
      throws XAMException
   {
      synchronized(s_cache)
      {
         Logger xamLogger = s_cache.get(path);
         if( xamLogger == null )
         {
            xamLogger = Logger.getLogger(path);
            s_cache.put( path, xamLogger );
            xamLogger.setUseParentHandlers(false);
                  
            // XAM Spec doesn't provide for multiple files per XAMLibrary Instance. 
            // Remove all existing file handlers and replace with one created exactly for
            // the XSystem instance. This may be because this method has been called
            // in response to changing a logging parameter.
            Handler hList[] = xamLogger.getHandlers();
            for( int h = hList.length-1; h >= 0; h-- )
               if( hList[h] instanceof FileHandler )
               {
                  xamLogger.removeHandler(hList[h]);
               }
            
            try
            {
               FileHandler logHandler = new FileHandler( path, LOG_SIZE, ROLLOVER_COUNT, APPEND);
               logHandler.setFormatter(new SimpleFormatter());
               xamLogger.addHandler(logHandler);
            }
            catch (IOException x)
            {
               throw new XAMException("Failed to setup XAM Library logging", x, -1);
            }
         }
         return xamLogger;
      }
   }

   /**
    * Called when the handler settings are different, and need to be reset.
    * @param logSize in bytes
    * @param rolloverCount
    * @param append
    */
   private static void UpdateHandlerSettings( Logger  logger,
                                              String  path,
                                              int     logSize,
                                              int     rolloverCount,
                                              boolean append)
      throws XAMException
   {
      Handler hList[] = logger.getHandlers();
      for( int h = hList.length-1; h >= 0; h-- )
         if( hList[h] instanceof FileHandler )
         {
            hList[h].close();
            logger.removeHandler(hList[h]);
         }

      try
      {
         FileHandler logHandler = new FileHandler( path, logSize, rolloverCount, append);
         logHandler.setFormatter(new SimpleFormatter());
         logger.addHandler(logHandler);
      }
      catch (IOException x)
      {
         throw new XAMException("Failed to setup XAM Library logging", x, -1);
      }
   }
  
   /**
    * Given a XAM Logging level, turn it into the appropriate Java logging
    * level.
    * 
    * @param xam_level The XAM Logging level
    * @return The appropriate Java logging level.
    */
   public static Level mapLogLevelToJavaLevel(long xam_level)
   {
      if( xam_level > XAMLibrary.XAM_LOG_ALL )
         return Level.ALL;

      switch ((int) xam_level)
      {
         case (int) XAMLibrary.XAM_LOG_ALL :
            return Level.ALL;
         case (int) XAMLibrary.XAM_LOG_INFO :
            return Level.INFO;
         case (int) XAMLibrary.XAM_LOG_WARN :
            return Level.WARNING;
         case (int) XAMLibrary.XAM_LOG_ERROR :
            return Level.SEVERE;
         case (int) XAMLibrary.XAM_LOG_FATAL :
            return Level.SEVERE;
         case (int) XAMLibrary.XAM_LOG_NONE :
            return Level.OFF;
         default :
            return Level.OFF;
      }
   }
}
