/* Copyright (c) 2008, The Storage Networking Industry Association.

 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */
/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */
package org.snia.xam.util;

import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.base.XAMImplementation;
import java.lang.reflect.Method;
import java.util.Properties;
import java.io.FileInputStream;

public class XAMLibraryFactory
{
   private static final String MESSAGE = "Unabled to create instance of XAMLibrary";

   private static final String GET_LIBRARY = "getLibrary";

   public static final String JAVA_XAM_LIBRARY = "org.snia.xam.base.XAMImplementation";

   public static final String C_XAM_LIBRARY = "org.snia.xam.XAMObject";

   /**
    * Return the default library. For this implementation, the default is to
    * return the XAM Library singleton.
    */
   static public XAMLibrary newXAMLibrary()
      throws XAMException
   {
      return XAMImplementation.getLibrary();
   }

   /**
    * Build a library based on the class name. This method will invoke the
    * library's "getLibrary()" method. 
    * Classname is one of the predefined class; 
    * XAMLibraryFactory.JAVA_XAM_LIBRARY or XAMLibraryFactory.C_XAM_LIBARY. 
    * The classname may also point to another class which implements the 
    * org.snia.xam.XAMLibrary interface.
    *
    * @param className The name of the class to instantiate.
    */
   static public XAMLibrary newXAMLibrary(String className)
      throws XAMException
   {
      try
      {
         Class libClass = Class.forName(className);
         Class parameterTypes[] = {};
         Method getLib = libClass.getMethod(GET_LIBRARY, parameterTypes);
         Object parameters[] = {};
         XAMLibrary lib = (XAMLibrary) getLib.invoke(null, parameters);
         return lib;
      }
      catch (Exception e)
      {
         throw new XAMException(MESSAGE, e);
      }
   }

   /**
    * Build a library based on the class name, with the property file. This
    * method will invoke the library's "getLibrary( Properties )" method.
    * Classname is one of the predefined class; 
    * XAMLibraryFactory.JAVA_XAM_LIBRARY or XAMLibraryFactory.C_XAM_LIBARY. 
    * The classname may also point to another class which implements the 
    * org.snia.xam.XAMLibrary interface.
    *
    * @param className The name of the class to instantiate.
    * @param props A Java property file containing the XAM configuration info.
    */
   static public XAMLibrary newXAMLibrary(String className, Properties props)
      throws XAMException
   {
      try
      {
         Class libClass = Class.forName(className);
         Class parameterTypes[] = { Class.forName("java.util.Properties") };
         Method getLib = libClass.getMethod(GET_LIBRARY, parameterTypes);
         Object parameters[] = { props };
         XAMLibrary lib = (XAMLibrary) getLib.invoke(null, parameters);
         return lib;
      }
      catch (Exception e)
      {
         throw new XAMException(MESSAGE, e);
      }
   }

   /**
    * Build a library based on the class name, with the property file. This
    * method will invoke the library's "getLibrary( Properties )" method.
    * Classname is one of the predefined class;
    * XAMLibraryFactory.JAVA_XAM_LIBRARY or XAMLibraryFactory.C_XAM_LIBARY. The
    * classname may also point to another class which implements the
    * org.snia.xam.XAMLibrary interface.
    * 
    * @param className The name of the class to instantiate.
    * @param configFile Path to a Java property file containing the XAM
    *           configuration info.
    */
   static public XAMLibrary newXAMLibrary(String className, String configFile)
      throws XAMException
   {
      try
      {
         Properties props = new Properties();
         props.load(new FileInputStream(configFile));
         return XAMLibraryFactory.newXAMLibrary(className, props);
      }
      catch (java.io.IOException jio)
      {
         throw new XAMException(MESSAGE, jio);
      }
   }
}
