/* Copyright (c) 2008, The Storage Networking Industry Association.

   Redistribution and use in source and binary forms, with or without modification, 
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.
   
   * Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.
   
   * Neither the name of The Storage Networking Industry Association (SNIA) nor 
   the names of its contributors may be used to endorse or promote products 
   derived from this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
   THE POSSIBILITY OF SUCH DAMAGE.
*/

/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */

package org.snia.xam.util;

import java.io.IOException;
import java.io.InputStream;

import org.snia.xam.FieldContainer;
import org.snia.xam.XAMException;
import org.snia.xam.XStream;

/**
 * An InputStream wrapper class for XStream objects.
 */
public class XStreamInputStream extends InputStream
{
	private int BUF_SIZE = 1024; // Internal buffer
	
	private byte[]  m_buf;
	private int     m_ptr;
	private int     m_read;
	private XStream m_stream;

	private XStreamInputStream()
	{
		m_buf = new byte[BUF_SIZE];
	}
	
	/**
	 * Opens an existing XStream in read only mode and
	 * wraps it in a InputStream interface.
	 * 
	 * @param fc The FieldContianer that contains the XStream
	 * @param name The name of the XStream
	 * @throws XAMException If the XStream does not exist, or an error occurs trying to open it.
	 */
	public XStreamInputStream(FieldContainer fc, String name)
		throws XAMException
	{
		this();
		m_stream = fc.openXStream(name, XStream.MODE_READ_ONLY);
	}

	/**
	 * Wraps an open XStream in an InputStream interface.
	 * The XStream must have been opened in read only mode.
	 * 
	 * @param stream An open read only XStream
	 */
	public XStreamInputStream(XStream stream)
	{
		this();
		m_stream = stream;
	}

	/**
	 * Closes the underlying XStream
	 */
	public void close() throws IOException
	{
		try
		{
			m_stream.close();
		}
		catch (XAMException x)
		{
			throw new XAMIOException(x);
		}
		super.close();
	}
	
	/**
	 * Mark not supported
	 * @return false
	 */
	public boolean markSupported()
	{
		return false;
	}
	
	/**
	 * Returns the number of bytes that can be read (or skipped over)
	 * from this input stream without blocking by the next caller of a
	 * method for this input stream. The next caller might be the same
	 * thread or or another thread.
	 * 
	 * @return The number of bytes that can be read from this input stream without blocking.
	 * @throws IOException If an I/O error occurs.
	 */
	public int available() throws IOException
	{
		return m_read - m_ptr;
	}

	/**
	 * @see java.io.InputStream#read()
	 */
	public int read() throws IOException
	{
		if (available() == 0)
		{
			int read = read(m_buf, 0, m_buf.length);
			if (read == -1)
				return -1;
			m_read = read;
			m_ptr = 0;
		}
		
		return m_buf[m_ptr++];
	}
	
	/**
	 * @see java.io.InputStream#read(byte[])
	 */
	public int read(byte[] b) throws IOException
	{
		return read(b, 0, b.length);
	}
	
	/**
	 * @see java.io.InputStream#read(byte[], int, int)
	 */
	public int read(byte[] b, int offset, int length) throws IOException
    {
	   int read = 0;
	   // First use any buffered data that we have
	   int buffered = Math.min(available(), length);
	   if (buffered > 0)
	   {
	      System.arraycopy(m_buf, m_ptr, b, offset, buffered);
	      m_ptr += buffered;
	      read = buffered;
	   }
	   // If we still need more data read from the underlying XStream
	   while ((length - read) > 0)
	   {
	      try
	      {
	         long bytes_read = m_stream.read(b, offset + read, length - read);
	         if (bytes_read == -1)
	         {
	            read = -1;
	            break;
	         }
	         read += bytes_read;
	      }
	      catch (XAMException x)
	      {
	         throw new XAMIOException(x);
	      }
         }
	   return read;
    }
}
