/* Copyright (c) 2008, The Storage Networking Industry Association.

   Redistribution and use in source and binary forms, with or without modification, 
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.
   
   * Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.
   
   * Neither the name of The Storage Networking Industry Association (SNIA) nor 
   the names of its contributors may be used to endorse or promote products 
   derived from this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
   THE POSSIBILITY OF SUCH DAMAGE.
*/

/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */

package org.snia.xam.util;

import java.io.IOException;
import java.io.OutputStream;

import org.snia.xam.FieldContainer;
import org.snia.xam.XAMException;
import org.snia.xam.XStream;

/**
 * An OutputStream wrapper class for XStream objects.
 */
public class XStreamOutputStream extends OutputStream
{
	private int BUF_SIZE = 1024; // Internal buffer
	
	private byte[]  m_buf;
	private int     m_write;
	private XStream m_stream;

	private XStreamOutputStream()
	{
		m_buf = new byte[BUF_SIZE];
	}
	
	/**
	 * Opens an existing XStream and either truncates it or appends to it
	 * based on the value of the append parameter. The opened XStream is
	 * wrapped in an InputStream interface.
	 * 
	 * @param ct The FieldContainer containing the XStream
	 * @param name The name of the XStream
	 * @param append True to append new data, false to overwrite existing data
	 * @throws XAMException If an error occurs trying to open the XStream
	 */
	public XStreamOutputStream(FieldContainer ct, String name, boolean append)
		throws XAMException
	{
		this();
		if (append)
			m_stream = ct.openXStream(name, XStream.MODE_WRITE_APPEND);
		else
			m_stream = ct.openXStream(name, XStream.MODE_WRITE_TRUNCATE);
	}
	
	/**
	 * Creates a new XStream and wraps it in an InputStream interface.
	 * 
	 * @param ct The FieldContainer to create the XStream in
	 * @param name The name of the XStream
	 * @param binding True if the XStream is binding
	 * @param mimeType The mime type of the XStream
	 * @throws XAMException If an error occurs while trying to create the XStream
	 */
	public XStreamOutputStream(FieldContainer ct, String name, boolean binding, String mimeType)
		throws XAMException
	{
		this();
		m_stream = ct.createXStream(name, binding, mimeType);
	}

	/**
	 * Wraps an open XStream in an InputStream interface. The XStream
	 * must have been opened in either write append or write truncate
	 * mode.
	 * 
	 * @param stream An open XStream
	 */
	public XStreamOutputStream(XStream stream)
	{
		this();
		m_stream = stream;
	}

	/**
	 * Flushes all write operations and closes the underlying XStream.
	 */
	public void close() throws IOException
	{
		flush();
		try
		{
			m_stream.close();
		}
		catch (XAMException x)
		{
			throw new XAMIOException(x);
		}
		super.close();
	}
	
	/**
	 * @see java.io.OutputStream#flush()
	 */
	public void flush() throws IOException
	{
		int write = 0;
		while ((m_write - write) > 0)
		{
			try
			{
				write += m_stream.write(m_buf, write, m_write-write);
			}
			catch (XAMException x)
			{
				throw new XAMIOException(x);
			}
		}
		m_write = 0;
	}
	
	/**
	 * @see java.io.OutputStream#write(int)
	 */
	public void write(int b) throws IOException
	{
		if (m_write == m_buf.length)
			flush();
		m_buf[m_write++] = (byte)b;
	}
	
	/**
	 * @see java.io.OutputStream#write(byte[])
	 */
	public void write(byte[] b) throws IOException
	{
		write(b, 0, b.length);
	}
	
	/**
	 * @see java.io.OutputStream#write(byte[], int, int)
	 */
	public void write(byte[] b, int offset, int length) throws IOException
	{
		if (m_write > 0)
			flush();
		int write = 0;
		while (write < length)
		{
			try
			{
				write += m_stream.write(b, offset+write, length-write);
			}
			catch (XAMException x)
			{
				throw new XAMIOException(x);
			}
		}
	}
}
