/* Copyright (c) 2008-2009, The Storage Networking Industry Association.

 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */
/* (c) Copyright 2009 Sun Microsystems, Inc.                    */

package org.snia.xam.util;

import java.util.Iterator;
import java.util.NoSuchElementException;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XStream;
import org.snia.xam.XUID;
import org.snia.xam.base.DefaultXUID;

/**
 * A class to read an XStream containing XAM XUID bytes using a
 * java.util.Iterator interface. It is assumed that the XStream passed to the
 * constructor has been opened in {@link XStream#MODE_READ_ONLY} mode.
 */
public class XUIDIterator
   implements Iterator
{
   // This inner class is used to unblock the read operation in the iterator.
   class xiListener implements XAsyncListener 
   {
      public void XAsyncCallback(XAsync asyncState)
      {
         synchronized(this) 
         {
            notifyAll();
         }
      }
      
   }
   // Iterator will attempt to read 500 XUID values (or less)
   // This turns out to fit nicely into a single TCP packet, so it
   // should be relatively efficient if going over a network.
   public static final int XUIDS_PER_READ = 500;

   private static final int MAX_XUID_LEN = 80;

   protected byte[] m_buf;
   protected byte[] m_async_buf;

   protected int m_ptr; // Cursor into m_buf for next.

   protected int m_read; // The number of bytes from the last read operation

   protected XUID m_next;

   protected XStream m_stream;

   /**
    * Constructs a new XUIDIterator for the specifed XStream. The XStream should
    * have been opened in {@link XStream#MODE_READ_ONLY} mode.
    * 
    * @param stream An open read only XStream
    */
   public XUIDIterator(XStream stream)
   {
      m_stream    = stream;
      m_buf       = new byte[XUIDS_PER_READ * MAX_XUID_LEN];
      m_async_buf = new byte[XUIDS_PER_READ * MAX_XUID_LEN];
   }

   /**
    * Checks to see if the XStream contains another XUID. Returns true if the
    * XStream does, false otherwise. This method may block reading more data
    * from the system.
    * 
    * @return true if the XStream contains another XUID, false otherwise.
    */
   public boolean hasNext()
   {
      if (m_buf == null) return false;

      try
      {
         if (m_next == null) m_next = readNext();
      }
      catch( Exception e )
      {
         throw new RuntimeException(e);
      }

      return m_next != null;
   }

   /**
    * Returns the next XUID in the XStream. If the XStream does not contain any
    * more XUIDs a NoSuchElementException is thrown.
    * 
    * @return The next XUID in the XStream.
    */
   public Object next()
   {
      if (m_buf == null) return null;

      try
      {
         if (m_next == null) m_next = readNext();
      }
      catch( Exception e )
      {
         throw new RuntimeException(e);
      }

      if (m_next == null) throw new NoSuchElementException();

      XUID next = m_next;
      m_next = null;
      return next;
   }

   /**
    * Optional java.util.Iterator method, currently unsupported.
    */
   public void remove()
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Attempts to read the next XUID from the buffered data. If not enough data
    * is available more will be read from the stream. This method may block.
    * 
    * @return The next XUID or null if no more data is available and the query
    *         is complete
    */
   protected XUID readNext()
      throws XAMException
   {
      // Make sure we have enough data for at least one XUID
      if ((m_read - m_ptr) < MAX_XUID_LEN) fillBuffer();

      if (m_buf != null)
      {
         // Create the Xuid object and advance the buffer pointer
         int len = 0xff & m_buf[m_ptr + 5];
         byte[] data = new byte[len];
         System.arraycopy(m_buf, m_ptr, data, 0, len);
         m_ptr += MAX_XUID_LEN;
         return DefaultXUID.valueOf(data);
      }

      // No data left, return null
      return null;
   }

   /**
    * Not enough bytes in the current buffer. Attempt to get more from the Stream. If the stream 
    * returns end of file, and the current buffer is exhausted, the buffer is deleted, indicating end
    * of stream.
    * @throws XAMException
    */
   private void fillBuffer()
      throws XAMException
   {
      long bytes_read = 0;

      // Move remaining data to the beginning of the buffer.
      // Since the remaining bytes must be less than 80 this is
      // guaranteed to be a non-overlapping copy, since the buffer 
      // is at least as large a single XUID.
      System.arraycopy(m_buf, m_ptr, m_buf, 0, m_read - m_ptr);
      m_read -= m_ptr;
      m_ptr = 0;

      try
      {
         while( (m_read - m_ptr) < XUID.MAX_LENGTH )
         {
            long op_id = System.currentTimeMillis();
            xiListener listener = new xiListener();
            long length = m_buf.length - m_read;
            synchronized( listener )
            {
               XAsync async = m_stream.asyncRead( m_async_buf, length, 
                                                  op_id, listener );
               do 
               {
                  try { listener.wait(); }
                  catch( InterruptedException ie ) { /* ignore it */ }
               } while ( ! async.isComplete());
               int actual_length = (int) async.getBytesRead();
               if( actual_length == XStream.EOF )
                  break;
               System.arraycopy(m_async_buf, 0, m_buf, m_read, (int) actual_length );
               m_read += actual_length;
            }
         }
      }
      catch (XAMException x)
      {
         m_buf = null;
         throw x;
      }

      if (m_read == 0)
      {
         // We've reached the end of the stream and are done reading
         m_buf = null;
      }
      else if (m_read < MAX_XUID_LEN)
      {
         // We only have a partial XUID
         m_buf = null;
         throw new XAMException( "XSystem error: XSystem returned a partial XUID" );
      }
   }

}
