/* Copyright (c) 2008, The Storage Networking Industry Association.
 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */
/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */
package org.snia.xam.base;

import java.io.File;
import java.io.RandomAccessFile;
import java.util.Properties;
import java.util.logging.Logger;
import java.util.logging.Level;

import org.snia.xam.XAMLibrary;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestResult;
import junit.framework.TestSuite;

public class LoggingTest
   extends TestCase
{
   public static void main(String[] args)
   {
      TestResult res = junit.textui.TestRunner.run(suite());
      System.exit(Math.max(res.failureCount(), res.errorCount()));
   }

   public static Test suite()
   {
      TestSuite suite = new TestSuite();
      suite.addTest(new LoggingTest("testLogPath"));
      suite.addTest(new LoggingTest("testLogSize"));
      suite.addTest(new LoggingTest("testLogAppend"));
      suite.addTest(new LoggingTest("testLogFinalization"));
      return suite;
   }

   public LoggingTest(String test)
   {
      super(test);
   }

   /**
    * This embedded class is used to filter the directory listing for log files. Used
    * in the next couple of logfile tests.
    */
   class prefixFilter implements java.io.FilenameFilter
   {
      String m_prefix;
      prefixFilter( String prefix )
      {
         m_prefix = prefix;
      }
      
      public boolean accept(File f, String arg1)
      {
         File test = new File(arg1);
         if( arg1.startsWith(m_prefix) && test.isFile() )
            return true;
         else
            return false;
      }
   }
   
   /**
    * This method will count the occurences of a file in the local directory. The files 
    * are those beginning with a pattern. After counting, the files are deleted. Used in
    * the next logging test (or so).
    *
    * This method assumes that the unit tests put the log files in the CWD, if not, this 
    * method is going to fail silently.
    * 
    * @param prefix The first part of the file name to count. 
    * @return The number of files matching the prefix. 
    */
   private int countAndDeleteFiles( String prefix )
   {
      prefixFilter filter = new prefixFilter(prefix);
      File dir = new File(".");
      String list[] = dir.list(filter);
      for( int i = 0; i < list.length; i++ )
      {
         File f = new File(list[i]);
         f.delete();
      }
      return list.length;
   }
   
   /**
    * Test that the barest form of logging is working.
    * This test checks that the named file will exist after instantiating the library.
    */
   public void testLogPath()
      throws Exception
   {
      final String LOG_PREFIX = "testLoggingUnitTest";
      final String PATH       = LOG_PREFIX + ".log.%g";
      final String UNIT_TEST_MESSAGE = "UNIT TEST MESSAGE";
      
      // Do something that gets logged
      Properties vims = new Properties();
      vims.setProperty( XAMLibrary.XAM_LOG_PATH, PATH);
      vims.setProperty( "xam_int." + XAMLibrary.XAM_LOG_LEVEL, String.valueOf(XAMLibrary.XAM_LOG_ALL) );
      
      // Clean old instances of the log file to be sure the test is fair.
      countAndDeleteFiles(LOG_PREFIX);

      XAMImplementation xam = new XAMImplementation();
      xam.loadVIMConfig(vims);
      
      // See if the appropriate file was created
      String path = xam.getString(XAMLibrary.XAM_LOG_PATH);
      String actualPath = path.replaceAll(".%g", ".0");
      File f = new File(actualPath);
      assertTrue("Log file not found " + f.getAbsolutePath(), f.exists());
      
      // Log a message
      String loggerName = path + "." + xam.getLong( XAMLibrary.XAM_LOG_LEVEL );
      Logger logger = Logger.getLogger(loggerName);
      logger.info( UNIT_TEST_MESSAGE );
      
      // Flush all messages on log file
      xam.close();
      
      // Make sure the message made it into the file
      RandomAccessFile logf = new RandomAccessFile( actualPath, "r" );
      String line;
      boolean found = false;
      while( (line = logf.readLine()) != null )
      {
         if( line.contains(UNIT_TEST_MESSAGE) )
            found = true;
      }
      if( ! found )
         fail( "Unit test message not logged to file" );
      
      // Clean up
      countAndDeleteFiles(LOG_PREFIX);
   }

   /**
    * Test to make sure that the log file is sized appropriately.
    * This test will write enough to roll the log over a couple of times, and then check
    * the size of .1 file. The test will pass as long as the file is within about 10% of the
    * intended size. This 10% criteria is highly sensitive to the logging implementation, 
    * but the intent of the test is to make sure that the intended log size is communicated to
    * the logging implementation.
    * @throws Exception
    */
   public void testLogSize()
      throws Exception
   {
      final int    LOG_SIZE   = 33; // in kB
      final int    LOG_COUNT  = 3; 
      final String LOG_PREFIX = "testLoggingSizeTest";
      final String PATH       = LOG_PREFIX + ".log.%g";
      final String UNIT_TEST_MESSAGE = "UNIT TEST MESSAGE";
      
      // Do something that gets logged
      Properties vims = new Properties();
      vims.setProperty( XAMLibrary.XAM_LOG_PATH, PATH );
      vims.setProperty( "xam_int." + XAMImplementation.XAM_LOG_MAX_SIZE, String.valueOf(LOG_SIZE) );
      vims.setProperty( "xam_int." + XAMImplementation.XAM_LOG_MAX_ROLLOVERS, String.valueOf(LOG_COUNT));
      
      // Clean old instances of the log file to be sure the test is fair.
      countAndDeleteFiles(LOG_PREFIX );

      XAMImplementation xam = new XAMImplementation();
      xam.setProperty(XAMLibrary.XAM_LOG_LEVEL, XAMLibrary.XAM_LOG_ALL );
      xam.loadVIMConfig(vims);
      
      String path = xam.getString(XAMLibrary.XAM_LOG_PATH );
      String actualPath = path.replaceAll(".%g", ".1");

      // Log a message
      // The size is not quite correct, since the logging implementation will add overhead,
      // but this is correct enough to allow the test to pass.
      int totalSize = (LOG_SIZE*1000) * LOG_COUNT;
      int msgSize = UNIT_TEST_MESSAGE.length();
      String loggerName = path + "." + xam.getLong( XAMLibrary.XAM_LOG_LEVEL );
      Logger logger = Logger.getLogger(loggerName);
      
      logger.log( Level.ALL, "Start of size test" );
      while( totalSize > 0 )
      {
         logger.log( Level.ALL, UNIT_TEST_MESSAGE );
         totalSize -= msgSize;
      }
      
      // Flush all messages on log file
      xam.close();
      
      // Check that the size of the file is not too far from the desired size.
      File logf = new File( actualPath );
      assertTrue( "Log file did not exist " + actualPath, logf.exists() );
      long delta = Math.abs( (LOG_SIZE*1000) - logf.length() );
      assertTrue( "Log file size varied more than 10% from desired size ", (delta/(LOG_SIZE*1000) < 0.1 ));
      
      // Clean up
      countAndDeleteFiles(LOG_PREFIX);
   }
   
   /**
    * Test the ability to append to the end of a log file. Testing the control of the append
    * function via to properties.
    */
   public void testLogAppend()
      throws Exception
   {
      final String LOG_PREFIX = "testLoggingAppendTest";
      final String PATH       = LOG_PREFIX + ".log.%g";
      final String UNIT_TEST_MESSAGE_1 = "APPEND TEST MESSAGE 1";
      final String UNIT_TEST_MESSAGE_2 = "APPEND TEST MESSAGE 2";
      
      // Do something that gets logged
      Properties vims = new Properties();
      vims.setProperty( XAMLibrary.XAM_LOG_PATH, PATH);
      vims.setProperty( "xam_int." + XAMLibrary.XAM_LOG_LEVEL, String.valueOf(XAMLibrary.XAM_LOG_ALL));
      
      // Clean old instances of the log file to be sure the test is fair.
//      countAndDeleteFiles(LOG_PREFIX);

      // Create first part of the log, with a message, close the library to flush.
      XAMImplementation xam = new XAMImplementation();
      xam.loadVIMConfig(vims);
      String loggerName = xam.getString(XAMLibrary.XAM_LOG_PATH);
      Logger logger = Logger.getLogger(loggerName);
      logger.log( Level.ALL, UNIT_TEST_MESSAGE_1 );
      xam.close();
      
      // Create first part of the log, with a message, close the library to flush.
      vims.setProperty( "xam_boolean." + XAMImplementation.XAM_LOG_APPEND, String.valueOf(true));
      xam = new XAMImplementation();
      xam.loadVIMConfig(vims);
      logger = Logger.getLogger(loggerName);
      logger.log( Level.ALL, UNIT_TEST_MESSAGE_2 );
      xam.close();
      
      // See if the appropriate file was created
      String path = xam.getString(XAMLibrary.XAM_LOG_PATH);
      String actualPath = path.replaceAll(".%g", ".0");
      File f = new File(actualPath);
      assertTrue("Log file not found " + f.getAbsolutePath(), f.exists());
      
      // Make sure the message made it into the file
      RandomAccessFile logf = new RandomAccessFile( actualPath, "r" );
      String line;
      boolean found1 = false;
      boolean found2 = false;
      while( (line = logf.readLine()) != null )
      {
         if( line.contains(UNIT_TEST_MESSAGE_1) )
            found1 = true;
         if( line.contains(UNIT_TEST_MESSAGE_2) )
            found2 = true;
      }
      if( ! found1 || ! found2)
         fail( "Unit test message not appended to file" );
      
      // Clean up
//      countAndDeleteFiles(LOG_PREFIX);
   }

   /** This test will check that the Library will close out log files in response
    * to a finalizer being invoked on the library. This will help prevent a problem
    * when an application improperly creates too many XAMLibrary instances.
    * @throws Exception
    */
   public void testLogFinalization()
      throws Exception
   {
      final int    ROLL_COUNT = 3;
      final String LOG_PREFIX = "testLogFinalization";
      final String PATH = LOG_PREFIX + ".log.%g";
      
      // Remove old occurrences of the log files, if any
      countAndDeleteFiles(LOG_PREFIX);
      
      // Do something that gets logged

      Properties vims = new Properties();
      vims.setProperty(XAMLibrary.XAM_LOG_PATH, PATH);
      vims.setProperty("xam_int." + XAMImplementation.XAM_LOG_MAX_ROLLOVERS, "3" );
      
      XAMImplementation xam = null;
      for( int x = 0; x < 200; x++ )
      {
         xam = new XAMImplementation();
         xam.loadVIMConfig(vims);
         
         // Give finalizers a chance to run.
         xam = null;
      }

      // Attempt to run GC, but of course this is not gauranteed
      Runtime.getRuntime().freeMemory();
      System.runFinalization();
      System.gc();
      
      try{ Thread.sleep( 1000 ); } catch (InterruptedException ie ) {;}
      
      // See if the appropriate file was created
      int count = countAndDeleteFiles( LOG_PREFIX );
   }
}
