/* Copyright (c) 2008, The Storage Networking Industry Association.

 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */

package org.snia.xam.base;

import java.util.Calendar;
import java.util.Random;
import java.util.Vector;

import org.snia.xam.base.AbstractFieldContainer;
import org.snia.xam.base.Property;
import org.snia.xam.testvim.TestXUID;
import org.snia.xam.util.ISO8601Date;

import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.FieldDoesNotExistException;
import org.snia.xam.FieldExistsException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidFieldNameException;
import org.snia.xam.InvalidFieldTypeException;
import org.snia.xam.InvalidOperationException;
import org.snia.xam.MaximumFieldException;
import org.snia.xam.ObjectInUseException;
import org.snia.xam.XAMException;
import org.snia.xam.XAMLibrary;
import org.snia.xam.XIterator;
import org.snia.xam.XSetAbandonException;
import org.snia.xam.XSetCorruptException;
import org.snia.xam.XStream;
import org.snia.xam.XSystemAbandonException;
import org.snia.xam.XSystemCorruptException;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XUID;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestResult;
import junit.framework.TestSuite;

public class TestAbstractFieldContainer
   extends TestCase
{
   private static final String PROP_1 = "com.example.prop1";

   private static final String PROP_2 = "com.example.prop2";

   private static final String PROP_3 = "com.example.prop3";

   private static final String PROP_4 = "com.example.prop4";

   private static final String PROP_5 = "com.example.prop5";

   private static final String PROP_6 = "com.example.prop6";

   private static final String PROP_7 = "com.example.prop7";

   private static final String PROP_8 = "com.example.prop8";

   private static final String PROP_9 = "com.example.prop9";

   private static final String PROP_10 = "com.example.prop10";

   private static final String PROP_11 = "com.example.prop11";

   private static final String PROP_12 = "com.example.prop12";

   class testFieldContainer
      extends AbstractFieldContainer
   {
      public testFieldContainer()
      {
         setProtectNamespace(false);
      }

      public void addTestProperty(Property p)
      {
         addField(p);
      }

      public void protect(boolean p)
      {
         setProtectNamespace(p);
      }

      public XAsync asyncOpenXStream(String name,
                                     String mode,
                                     long xopid,
                                     XAsyncListener listener)
         throws AuthenticationExpiredException, InvalidArgumentException,
         InvalidFieldNameException, FieldDoesNotExistException,
         ObjectInUseException, XSetAbandonException, XSetCorruptException,
         XSystemAbandonException, XSystemCorruptException, XAMException
      {
         throw new InvalidOperationException();
      }

      public XStream createXStream(String fieldName,
                                   boolean binding,
                                   String mimeType)
         throws AuthenticationExpiredException, InvalidArgumentException,
         InvalidFieldNameException, InvalidOperationException,
         FieldExistsException, MaximumFieldException, ObjectInUseException,
         XSetAbandonException, XSetCorruptException, XSystemAbandonException,
         XSystemCorruptException, XAMException
      {
         throw new InvalidOperationException();
      }

      public XStream openXStream(String fieldName, String mode)
         throws AuthenticationExpiredException, InvalidFieldNameException,
         FieldDoesNotExistException, ObjectInUseException,
         XSetAbandonException, XSetCorruptException, XSystemAbandonException,
         XSystemCorruptException, XAMException
      {
         throw new InvalidOperationException();
      }
   }

   public TestAbstractFieldContainer(String name)
   {
      super(name);
   }

   /**
    * Test the core property functions.
    */
   public void testBooleanProperties()
      throws Exception
   {
      testFieldContainer tafc = new testFieldContainer();

      // Boolean
      tafc.createProperty(PROP_1, false, true);
      tafc.createProperty(PROP_2, true, false);
      assertTrue("Bool true property", tafc.getBoolean(PROP_1));
      assertFalse("Bool nonbinding property", tafc.getFieldBinding(PROP_1));
      assertFalse("Bool false property", tafc.getBoolean(PROP_2));
      assertTrue("Bool binding property", tafc.getFieldBinding(PROP_2));
      assertEquals("Bool property length",
         (long) 1,
         tafc.getFieldLength(PROP_1));
      assertEquals("Bool property type",
         XAMLibrary.STYPE_BOOLEAN_MIME_TYPE,
         tafc.getFieldType(PROP_1));

      tafc.setProperty(PROP_1, false);
      assertFalse("Bool set false property", tafc.getBoolean(PROP_1));
   }

   /**
    * Test core long property functions
    */
   public void testLongProperties()
      throws Exception
   {
      testFieldContainer tafc = new testFieldContainer();

      // Long
      tafc.createProperty(PROP_3, false, -12);
      tafc.createProperty(PROP_4, true, 37);
      assertEquals("Long -12 property", -12, tafc.getLong(PROP_3));
      assertFalse("Long nonbinding property", tafc.getFieldBinding(PROP_3));
      assertEquals("Long 37 property", 37, tafc.getLong(PROP_4));
      assertTrue("Long binding property", tafc.getFieldBinding(PROP_4));
      assertEquals("Long property length",
         (long) 8,
         tafc.getFieldLength(PROP_3));
      assertEquals("Long property type",
         XAMLibrary.STYPE_INT_MIME_TYPE,
         tafc.getFieldType(PROP_3));
   }

   /**
    * Test core double property functions
    */
   public void testDoubleProperties()
      throws Exception
   {
      testFieldContainer tafc = new testFieldContainer();

      // Double
      tafc.createProperty(PROP_5, false, -11.3);
      tafc.createProperty(PROP_6, true, 212.1);
      assertEquals("Double -11.3 property", -11.3, tafc.getDouble(PROP_5));
      assertFalse("Double nonbinding property", tafc.getFieldBinding(PROP_5));
      assertEquals("Double 212.1 property", 212.1, tafc.getDouble(PROP_6));
      assertTrue("Double binding property", tafc.getFieldBinding(PROP_6));
      assertEquals("Double property length",
         (long) 8,
         tafc.getFieldLength(PROP_5));
      assertEquals("Double property type",
         XAMLibrary.STYPE_DOUBLE_MIME_TYPE,
         tafc.getFieldType(PROP_5));
   }

   /**
    * Test core String property functions
    */
   public void testStringProperties()
      throws Exception
   {
      final String S1 = "Bandersnatch";
      final String S2 = "Jabberwocky";

      testFieldContainer tafc = new testFieldContainer();

      // String
      tafc.createProperty(PROP_7, false, S1);
      tafc.createProperty(PROP_8, true, S2);
      assertEquals("String " + S1 + " property", S1, tafc.getString(PROP_7));
      assertFalse("String nonbinding property", tafc.getFieldBinding(PROP_7));
      assertEquals("String " + S2 + " property", S2, tafc.getString(PROP_8));
      assertTrue("String binding property", tafc.getFieldBinding(PROP_8));
      assertEquals("String property length",
         (long) S1.length(),
         tafc.getFieldLength(PROP_7));
      assertEquals("String property type",
         XAMLibrary.STYPE_STRING_MIME_TYPE,
         tafc.getFieldType(PROP_7));
   }

   /**
    * Test core datetime property functions
    */
   public void testDateTimeProperties()
      throws Exception
   {
      final Calendar D1 = Calendar.getInstance();
      final Calendar D2 = Calendar.getInstance();

      testFieldContainer tafc = new testFieldContainer();

      // Date
      tafc.createProperty(PROP_9, false, D1);
      tafc.createProperty(PROP_10, true, D2);
      assertEquals("DateTime " + D1 + " property", D1, tafc.getDateTime(PROP_9));
      assertFalse("DateTime nonbinding property", tafc.getFieldBinding(PROP_9));
      assertEquals("DateTime " + D2 + " property",
         D2,
         tafc.getDateTime(PROP_10));
      assertTrue("DateTime binding property", tafc.getFieldBinding(PROP_10));
      String d1_str = ISO8601Date.toISO8601String(D1);
      assertEquals("DateTime property length",
         (long) d1_str.length(),
         tafc.getFieldLength(PROP_9));
      assertEquals("DateTime property type",
         XAMLibrary.STYPE_DATETIME_MIME_TYPE,
         tafc.getFieldType(PROP_9));
   }

   /**
    * Test core XUID property functions
    */
   public void testXUIDProperties()
      throws Exception
   {
      testFieldContainer tafc = new testFieldContainer();

      // XUID
      TestXUID xuid_1 = TestXUID.newXUID();
      TestXUID xuid_2 = TestXUID.newXUID();

      tafc.createProperty(PROP_11, false, xuid_1);
      tafc.createProperty(PROP_12, true, xuid_2);

      assertEquals("XUID 1 property", tafc.getXUID(PROP_11), xuid_1);
      assertFalse("XUID 1 binding", tafc.getFieldBinding(PROP_11));

      assertEquals("XUID 2 property", tafc.getXUID(PROP_12), xuid_2);
      assertTrue("XUID 2 binding", tafc.getFieldBinding(PROP_12));

      int xuid_length = xuid_1.toBytes().length;
      assertEquals("XUID Length", xuid_length, tafc.getFieldLength(PROP_12));

      XUID a_xuid = tafc.getXUID(PROP_11);
      assertEquals("XUIDs compare", xuid_1, a_xuid);

      assertEquals("XUID property type",
         XAMLibrary.STYPE_XUID_MIME_TYPE,
         tafc.getFieldType(PROP_12));
   }

   /**
    * Test that the get/set Fields check property types. This test just creates
    * a single XAM_INT property, and tries to get it via other typed calls.
    * Because of the way that AbstractFieldContainer is written, not all
    * combinations need to be tried. (uses "getPropertyType")
    */
   public void testTypeCheck()
      throws Exception
   {
      testFieldContainer tafc = new testFieldContainer();

      tafc.createProperty(PROP_1, false, 177);
      try
      {
         Boolean test = tafc.getBoolean(PROP_1);
      }
      catch (InvalidFieldTypeException ifte)
      { /* test passes. */
      }
      try
      {
         double test = tafc.getDouble(PROP_1);
      }
      catch (InvalidFieldTypeException ifte)
      { /* test passes. */
      }
      try
      {
         String test = tafc.getString(PROP_1);
      }
      catch (InvalidFieldTypeException ifte)
      { /* test passes. */
      }
      try
      {
         Calendar test = tafc.getDateTime(PROP_1);
      }
      catch (InvalidFieldTypeException ifte)
      { /* test passes. */
      }
      try
      {
         XUID test = tafc.getXUID(PROP_1);
      }
      catch (InvalidFieldTypeException ifte)
      { /* test passes. */
      }
   }

   /**
    * Test the ReadOnly capabilities on fields.
    */
   public void testReadOnly()
      throws Exception
   {
      testFieldContainer tafc = new testFieldContainer();
      Property test = new Property(PROP_1, XAMLibrary.STYPE_STRING_MIME_TYPE,
         "foobar", false, true);
      tafc.addTestProperty(test);

      assertTrue("Property Read Only", tafc.getFieldReadOnly(PROP_1));
   }

   /**
    * Test field iterator.
    */
   public void testFieldIterator()
      throws Exception
   {
      final int TEST_COUNT = 13;
      Random rnd = new Random();

      testFieldContainer tafc = new testFieldContainer();
      Vector<String> names = new Vector<String>();

      for (int i = 0; i < TEST_COUNT; i++)
      {
         long value = rnd.nextInt();
         String name = String.valueOf(value);
         if (i == 0) name = "X" + name;
         names.add(name);

         tafc.createProperty(name, false, value);
      }

      // Make sure the iterator gets everything in the list.
      XIterator xi = tafc.openFieldIterator("");
      while (xi.hasNext())
      {
         String a_name = (String) xi.next();
         names.remove(a_name);
      }
      assertEquals("Iterator covered all names", 0, names.size());

      // Check to see if a simple prefix works.
      xi = tafc.openFieldIterator("X");
      int count = 0;
      while (xi.hasNext())
      {
         count++;
         xi.next();
      }
      assertEquals("Iterator prefix count 1 item", 1, count);
   }

   /**
    * Test the protected name space capabilities of the AFC.
    */
   public void testProtectedNamespace()
      throws Exception
   {
      testFieldContainer tafc = new testFieldContainer();

      tafc.protect(false);
      tafc.createProperty(".protected.foo", false, false);
      tafc.protect(true);
      try
      {
         tafc.createProperty(".protect.bar", false, false);
      }
      catch (InvalidFieldNameException x)
      {
         /* This means the test passed. */
      }
   }

   /**
    * Standalone tests.
    */
   public static void main(String[] args)
   {
      System.exit(runTest(args, suite()));
   }

   public static int runTest(String[] args, Test test)
   {
      TestResult res = junit.textui.TestRunner.run(test);
      return Math.max(res.failureCount(), res.errorCount());
   }

   public static Test suite()
   {
      TestSuite suite = new TestSuite();
      suite.addTest(new TestAbstractFieldContainer("testBooleanProperties"));
      suite.addTest(new TestAbstractFieldContainer("testLongProperties"));
      suite.addTest(new TestAbstractFieldContainer("testDoubleProperties"));
      suite.addTest(new TestAbstractFieldContainer("testStringProperties"));
      suite.addTest(new TestAbstractFieldContainer("testDateTimeProperties"));
      suite.addTest(new TestAbstractFieldContainer("testXUIDProperties"));
      suite.addTest(new TestAbstractFieldContainer("testTypeCheck"));
      suite.addTest(new TestAbstractFieldContainer("testReadOnly"));
      suite.addTest(new TestAbstractFieldContainer("testFieldIterator"));
      suite.addTest(new TestAbstractFieldContainer("testProtectedNamespace"));
      return suite;
   }
}
