/* Copyright (c) 2008, The Storage Networking Industry Association.
 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */
/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */
package org.snia.xam.base;

import org.snia.xam.InvalidXRIException;
import org.snia.xam.base.XRI;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import java.util.List;

/**
 * This test class is to test that AbstractFieldContainer throws the appropriate Exceptions at the 
 * appropriate time. Since there are so many exceptions that may be thrown, not all possibilities are
 * tests. This tests only those that the AbstractFieldContainer is going to throw. This test suite 
 * is likely never "complete" and may be added to as time permits.
 * 
 * @author allisonm
 *
 */
public class XRITest
   extends TestCase
{
   private static String VALID_XRI[] = {
      "snia-xam://vim!localhost",
      "snia-xam://localhost",
      "snia-xam://127.0.0.1",
      "snia-xam://vim!127.0.0.1", // Match ipv4addr from RFC3987
      "snia-xam://vim!my-host-1",
      "snia-xam://vim!host?param1=value",
      "snia-xam://host?param1=value",
      "snia-xam://system?param1=value&param2=foobar",
      "snia-xam://vimmer!hoster?paramx=y&param7=8",
      "snia-xam://v!h?p=c%3a%2fstore", // param value is "c:/store"
      "snia-xam://v!h?p=c%3a%5cstore", // param value is "c:\store"
      "snia-xam://v!h?p=c:/store", "snia-xam://v!h?p=/home/user/store",
      "snia-xam://host?p1=v1&p2=v2&p3=c%3a",
      "snia-xam://v!fe80::20e:7fff:fe23:5212", // IP V6 address
      "snia-xam://v!::20e:7fff:fe23:5212", // IP V6 address
      "snia-xam://v!fe80::20e:7fff:::" // IP V6 address
   };

   private static String INVALID_XRI[] = {"foobar://localhost", "snia-xam://", // System name is required!
      "snia-xam://a!b!c", // Only one ! separator allowed.
      "snia-xam://?params", // System name missing.
      "snia-xam://?&arf=bar" // Bad parameter, first one is empty.
   };

   public static void main(String[] args)
   {
   }

   public static Test suite()
   {
      TestSuite suite = new TestSuite();
      suite.addTest(new XRITest("testValidXRI"));
      suite.addTest(new XRITest("testInvalidXRI"));
      suite.addTest(new XRITest("testXRIComponents"));
      return suite;
   }

   public XRITest(String test)
   {
      super(test);
   }

   /** 
    * Test a series of legal XRI values to make sure they parse correctly.
    */
   public void testValidXRI()
      throws Exception
   {
      for (int x = 0; x < VALID_XRI.length; x++)
      {
         XRI local_xri = XRI.valueOf(VALID_XRI[x]);
         local_xri = null;
      }
   }

   /**
    * Test to make sure that the InvalidXRIException is thrown on bad XRI values
    */
   public void testInvalidXRI()
      throws Exception
   {
      for (int x = 0; x < INVALID_XRI.length; x++)
      {
         try
         {
            XRI local_xri = XRI.valueOf(INVALID_XRI[x]);
            fail("The following XRI should have thrown an InvalidXRIException: "
               + INVALID_XRI[x]);
         }
         catch (InvalidXRIException ixe)
         {
            /* It's good, we expect this. */
         }
      }
   }

   /**
    * Test that the XRI class is actually parsing out the correct components 
    * of an XRI in the appropriate circumstances.
    */
   public void testXRIComponents()
      throws Exception
   {
      final String XRI1 = "snia-xam://vim1!host1";

      XRI xri1 = XRI.valueOf(XRI1);
      assertEquals("XRI1 Vim Name", "vim1", xri1.getVimInfo());
      assertEquals("XRI1 System", "host1", xri1.getSystem());
      List p = xri1.getParameterNames();
      assertEquals("XRI1 Param list", 0, p.size());
      xri1 = null;

      final String XRI2 = "snia-xam://host-only";
      XRI xri2 = XRI.valueOf(XRI2);
      assertNull("XRI2 VIM Name", xri2.getVimInfo());
      assertEquals("XRI2 System", "host-only", xri2.getSystem());
      p = xri2.getParameterNames();
      assertEquals("XRI2 Param list", 0, p.size());
      xri2 = null;

      final String XRI3 = "snia-xam://host?p1=v1&p2=v2&p3=c%3a%5cstore";
      XRI xri3 = XRI.valueOf(XRI3);
      assertNull("XRI2 VIM Name", xri3.getVimInfo());
      assertEquals("XRI3 System", "host", xri3.getSystem());
      p = xri3.getParameterNames();
      assertEquals("XRI3 param count", 3, p.size());

      assertEquals("XRI3 param 1", "v1", xri3.getParameterValue("p1"));
      assertEquals("XRI3 param 2", "v2", xri3.getParameterValue("p2"));
      assertEquals("XRI3 param 3", "c:\\store", xri3.getParameterValue("p3"));
   }
}
