/* Copyright (c) 2008, The Storage Networking Industry Association.

   Redistribution and use in source and binary forms, with or without modification, 
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.
   
   * Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.
   
   * Neither the name of The Storage Networking Industry Association (SNIA) nor 
   the names of its contributors may be used to endorse or promote products 
   derived from this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
   THE POSSIBILITY OF SUCH DAMAGE.
*/

/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */

package org.snia.xam.testvim;

import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidOperationException;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XSetUnderHoldException;
import org.snia.xam.XStream;
import org.snia.xam.XStreamAbandonException;
import org.snia.xam.XStreamCorruptException;
import org.snia.xam.XStreamException;
import org.snia.xam.base.Field;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

/**
 * This class is present to allow testing of the QueryFactory class. It allows creation of
 * a stream, and allowing the writing of stream contents. It is succificient to support the 
 * QueryFactoryTest.
 */
public class TestXStream
   implements XStream, Field
{
   private static final String NI = null;
   File             m_file ;
   FileOutputStream m_fos;
   String           m_name;
   
   // Normalized field attributes
   boolean m_binding;
   boolean m_readonly;
   String  m_type;
   
   TestXStream( String name, boolean binding, boolean ro, String type ) throws XAMException
   {
      m_name     = name;
      m_binding  = binding;
      m_readonly = ro;
      m_type     = type;
      try
      {
         m_file = File.createTempFile( name , ".tmp" );
         m_fos  = new FileOutputStream( m_file );
      }
      catch( IOException ioe )
      {
         throw new XStreamException( "can't create temp xstream" + ioe.getMessage() );
      }
   }
   
   public void abandon()
      throws AuthenticationExpiredException, XAMException
   {
      try
      {
         m_fos.close();
      }
      catch( IOException ioe )
      {
         /* Ignore it, this is for testing, and the XStream shouldn't be used after this point. */
      }
      finally
      {
         m_fos  = null;
         m_file.delete();
         m_file = null;
      }

   }

   public XAsync asyncClose(long xopid, XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamCorruptException, XStreamException, XAMException
   {
      throw new InvalidOperationException( NI );
   }

   public XAsync asyncRead(byte[] buffer,
                           long bufferLength,
                           long xopid,
                           XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XAMException
   {
      throw new InvalidOperationException( NI );
   }

   public XAsync asyncWrite(byte[] buffer,
                            long bufferLength,
                            long xopid,
                            XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XAMException
   {
      throw new InvalidOperationException( NI );
   }

   public void close()
      throws AuthenticationExpiredException, XSetUnderHoldException,
      XStreamCorruptException, XStreamException, XAMException
   {
      if( m_fos != null )
      {
         try 
         { 
            m_fos.close(); 
         } 
         catch( IOException ioe ) 
         {
            throw new XStreamException( "Failed to capture stream data " + ioe.getMessage() );
         }
      }
      m_file = null;
      m_fos  = null;
   }

   public long read(byte[] buffer, long offset, long bytesToRead)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XAMException
   {
      throw new InvalidOperationException( NI );
   }

   public long read(byte[] buffer, long bytesToRead)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XAMException
   {
      throw new InvalidOperationException( NI );
   }

   public long read(byte[] buffer)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XAMException
   {
      throw new InvalidOperationException( NI );
   }

   public void seek(long offset, long whence)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XAMException
   {
      throw new InvalidOperationException( NI );
   }

   public long tell()
      throws AuthenticationExpiredException, XStreamAbandonException,
      XStreamCorruptException, XAMException
   {
      throw new InvalidOperationException( NI );
   }

   public long write(byte[] buffer)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XSetUnderHoldException, XAMException
   {
      this.write( buffer, 0, buffer.length );
      return buffer.length;
   }

   public long write(byte[] buffer, long bytesToWrite)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XSetUnderHoldException, XAMException
   {
      this.write( buffer, 0, bytesToWrite );
      return bytesToWrite;
   }

   public long write(byte[] buffer, long offset, long bytesToWrite)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XSetUnderHoldException, XAMException
   {
      if( m_fos == null )
         throw new XStreamException( "File output stream is invalid in test " );
      
      if( (offset > Integer.MAX_VALUE) || (offset < 0) )
         throw new InvalidArgumentException( "Offset of stream write is bad : " + offset );
      
      if( (bytesToWrite > Integer.MAX_VALUE) || (bytesToWrite < 0 ))
         throw new InvalidArgumentException( "Buffer length in stream write is invalid : " + bytesToWrite );
      try
      {
         m_fos.write( buffer, (int) offset, (int) bytesToWrite );
      }
      catch( IOException ioe )
      {
         throw new XStreamException( "Write failed due to underlying IOExcepton " + ioe.getMessage() );
      }
      return bytesToWrite;
   }

   public boolean getBinding()
   {
      return m_binding;
   }

   public long getLength()
   {
      if( m_file != null )
         return m_file.length();
      else
         return -1;
   }

   public String getName()
   {
      return m_name;
   }

   public boolean getReadOnly()
   {
      return m_readonly;
   }

   public String getType()
   {
      return m_type;
   }

   public boolean isModified()
   {
      return false;
   }

}
