/* Copyright (c) 2008, The Storage Networking Industry Association.

   Redistribution and use in source and binary forms, with or without modification, 
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.
   
   * Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.
   
   * Neither the name of The Storage Networking Industry Association (SNIA) nor 
   the names of its contributors may be used to endorse or promote products 
   derived from this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
   THE POSSIBILITY OF SUCH DAMAGE.
*/

/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */

package org.snia.xam.testvim;

import org.snia.xam.base.DefaultXUID;
import org.snia.xam.util.ISO8601Date;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.text.ParseException;
import java.util.Calendar;
import java.util.Random;
import java.util.StringTokenizer;

/**
 * Generate a system unique XUID value. This class attempts to make values unique across space 
 * and time using MacAddress, time, and sequence numbers. This not yet thread safe.
 */
public class TestXUID
   extends DefaultXUID
{
   // This will be used when finding the MacAddress
   static final String NIX_GET_MAC_ADDR_CMD = "ifconfig eth0";
   static final String WINDOWS_GET_MAC_ADDR_CMD = "ipconfig /all";

   // Useful constants for the SNIA XUID offsets
   final static int RESERVED    = 0;
   final static int OID         = 1;
   final static int RESERVED2   = 4;
   final static int LENGTH      = 5;
   final static int CRC         = 6;
   final static int OPAQUE      = 8;
   final static int XUID_LENGTH = 80;
   
   // SNIA.ORG IANA number
   final static int SNIA_IANA_1 = 0x3a;
   final static int SNIA_IANA_2 = 0x03;
   
   // For cleaner file names, substitute funky characters with this neutral.
   final static char NEUTRAL_CHARACTER = '_';
   
   // Increment this every single time XUID is called. Helps guarantee that the XUID values
   // are unique. Doesn't hurt when the disambiguator is reset next time. 
   static long m_disambiguator = 0;
   
   static byte m_macaddress[];
   
   protected TestXUID( byte bits[] )
   {
      super( bits );
   }
   
   /**
    * Create a XUID from "whole cloth". This is when a NEW XUID value is required.
    * typically for a new XSet, or when an XSet is being changed due to a binding modification.
    * @return a new unique XUID.
    */
   public static final TestXUID newXUID()
   {
      byte raw[] = new byte[80];
      
      raw[RESERVED]  = 0;
      raw[OID + 0]   = 0;
      raw[OID + 1]   = SNIA_IANA_1;
      raw[OID + 2]   = SNIA_IANA_2;
      raw[RESERVED2] = 0;
      raw[CRC + 0]   = 0;
      raw[CRC + 1]   = 0;
      
      if( m_macaddress == null )
         getMacAddress();
      
      // Produce a globally unique value.
      // Step 1 - Add the Mac Address
      int i = OPAQUE;
      for( int x = 0; x < m_macaddress.length; x++ )
      {
         raw[i++] = m_macaddress[x];
      }
      raw[i++] = (byte) '-';
      
      // Step 2 - Add the system time value
      String iso8601Time = ISO8601Date.toISO8601String(Calendar.getInstance());
      for( int c = 0; c < iso8601Time.length(); c++ )
      {
         char the_c = iso8601Time.charAt(c);
         if( (the_c == ':') || (the_c == '.' )) 
            the_c = NEUTRAL_CHARACTER;
         raw[i++] = (byte) the_c;
      }
      
      // Step 3 - Add the disambiguator
      raw[i++] = '-';
      m_disambiguator++;
      String dis = Long.toHexString(m_disambiguator);
      for( int c = 0; c < dis.length(); c++ )
      {
         char the_c = dis.charAt(c);
         raw[i++] = (byte) the_c;
      }
      
      // Step 4 - Set the length
      raw[LENGTH] = (byte) (i);
      
      // Step 5 - Trim the blank stuff.
      byte trimmed[] = new byte[i];
      for( int x = 0; x < i; x++ )
         trimmed[x] = raw[x];

      // Step 6 - Set the CRC
      int crc_value = s_crc.getCrc(trimmed);
      trimmed[CRC + 0] = (byte) ((crc_value >> 8) & 0xff);
      trimmed[CRC + 1] = (byte) (crc_value & 0xff);
        
      return new TestXUID(trimmed);
   }
   
   /**
    * Execute the "ipConfig/ifConfig" command, buffer up the results for later parsing.
    * @param cmd Platform specific config command.
    * @return BufferedReader of all the interface information
    * @throws IOException
    */
   private static BufferedReader getIPInfo( String cmd ) throws IOException 
   {
      Process p = Runtime.getRuntime().exec(cmd);
      BufferedReader output = new BufferedReader(new InputStreamReader(p.getInputStream()));
      return output;
   }

   /** Parse the output from the comamnd and fill in the bytes.
    * @param br BufferedReader with the output of *nix command.
    */
   private static void parseNixMacAddress( BufferedReader br ) throws IOException
   {
      String mac_string = br.readLine();
      StringTokenizer st = new StringTokenizer(mac_string);
      while (st.hasMoreTokens())
      {
         String token = st.nextToken();
         if (token.equals("HWaddr"))
         {
            mac_string = st.nextToken();
            break;
         }
      }
      // Parse the mac_string. The string should
      // look like "00:10:A4:90:CC:94"
      m_macaddress = new byte[6];
      if (mac_string != null)
      {
         StringTokenizer mac_st = new StringTokenizer(mac_string, ":");
         int i = 0;
         
         while (mac_st.hasMoreTokens())
         {
            String token = mac_st.nextToken();
            m_macaddress[i++] = (byte) Integer.parseInt(token, 16);
         }
      }
   }

   /** Parse the output from the comamnd and fill in the bytes.
    * @param br BufferedReader with the output of *nix command.
    */
   private static void windowsParseMacAddress( BufferedReader br ) throws IOException, ParseException
   {
      String line;
      while( (line = br.readLine()) != null )
      {
         if( line.contains( "Physical Address" ))
            break;
      }

      // Parse the mac_string. The string should
      // look like "Physical Address . . . . . : 00-10-A4-90-CC-94"
      m_macaddress = new byte[12];
      int i = 0;
      int start = line.indexOf( ':' );
      StringTokenizer st = new StringTokenizer(line.substring(start+1), " -" );
      
      while( st.hasMoreElements() )
      {
         String token = st.nextToken();
         m_macaddress[i++] = (byte) token.charAt(0);
         m_macaddress[i++] = (byte) token.charAt(1);
      }
   }
   
   /**
    * Used when attempting to get the MacAddress has failed. 
    */
   private static void createRandomMacAddress()
   {
      m_macaddress = new byte[6];
      Random rnd = new Random();
      
      for( int x = 0; x < m_macaddress.length; x++ )
      {
         m_macaddress[x++] = (byte) (byte) rnd.nextInt( 256 );
      }
   }
   
   /**
    * This is basically a hack to get the MacAddress so we can disambiguate the XUID values.
    */
   private static void getMacAddress()
   {
      String os = System.getProperty("os.name");
      
      try 
      {
         if(os.startsWith("Windows")) 
            windowsParseMacAddress(getIPInfo(WINDOWS_GET_MAC_ADDR_CMD));
      
         else if(os.startsWith("Linux")) 
            parseNixMacAddress(getIPInfo(NIX_GET_MAC_ADDR_CMD));
         
         else 
         {
            createRandomMacAddress();
         }
      }
      catch( Exception e )
      {
         createRandomMacAddress();
      }
   }
   
   /**
    * Because this TestXUID is composed of characters, it is supposed to be a filesystem 
    * safe filename. Just convert the bytes to characters and return them.
    * @return The file name version of the XUID.
    */
   public String getFileName()
   {
      StringBuffer sb = new StringBuffer();
      int length = this.m_bytes[LENGTH] - OPAQUE;
      
      for( int i = 0; i < length; i++ )
         sb.append((char) m_bytes[OPAQUE+i] );
      return sb.toString();
   }
   
   static public void main( String args[] )
   {
      for( int i = 0; i < 10; i ++ )
      {
         TestXUID tx = TestXUID.newXUID();
         System.out.println( tx.toString() );
         System.out.println( tx.getFileName() );
      }
   }
}
