/* Copyright (c) 2008-2009, The Storage Networking Industry Association.

 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */
/* (c) Copyright 2009 Sun Microsystems, Inc.                    */

package org.snia.xam.util;

import org.snia.xam.AsyncHaltedException;
import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidOperationException;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XSet;
import org.snia.xam.XSetUnderHoldException;
import org.snia.xam.XStream;
import org.snia.xam.XStreamAbandonException;
import org.snia.xam.XStreamCorruptException;
import org.snia.xam.XStreamException;
import org.snia.xam.XUID;

/**
 * A class to wrap a byte[] in a read only {@link XStream} interface. This is
 * primarily used to wrap the SASL authentication response in an XStream.
 */
public class ByteArrayXStream
   implements XStream
{
   class basXAsync implements XAsync, Runnable
   {
      long    m_bytes_read    = 0;
      long    m_op_id         = 0;
      XStream m_stream        = null;
      
      // Used to do the read
      byte m_buf[]  = null;
      long m_length = 0;
      XAsyncListener m_listener = null;
      Throwable m_exception     = null;
      
      basXAsync( XStream stream, byte buf[], long length, long id, XAsyncListener listener )
      {
         m_op_id    = id;
         m_listener = listener;
         m_buf      = buf;
         m_length   = length;
         m_stream   = stream;
      }
      
      public void close()
         throws AuthenticationExpiredException, XAMException
      { }

      public long getBytesRead()
         throws AuthenticationExpiredException, AsyncHaltedException,
         XAMException
      { return m_bytes_read; }

      public long getBytesWritten()
         throws AuthenticationExpiredException, AsyncHaltedException,
         XAMException
      { return 0; }

      public void getStatus()
         throws AuthenticationExpiredException, AsyncHaltedException,
         XAMException
      { 
         if( m_exception instanceof XAMException )
            throw (XAMException) m_exception;
         else if( m_exception != null )
            throw new XAMException( "Error halted async operation ", m_exception );
      }

      public long getXOPID()
         throws AuthenticationExpiredException, AsyncHaltedException,
         XAMException
      { return m_op_id; }

      public XSet getXSet()
         throws AuthenticationExpiredException, AsyncHaltedException,
         XAMException
      { return null; }

      public XStream getXStream()
         throws AuthenticationExpiredException, AsyncHaltedException,
         XAMException
      { return m_stream; }

      public XUID getXUID()
         throws AuthenticationExpiredException, AsyncHaltedException,
         XAMException
      { return null;  }

      public void halt()
         throws AuthenticationExpiredException, XAMException
      {  }

      public boolean isComplete()
         throws AuthenticationExpiredException, XAMException
      { return true; }
      
      void read() 
      {
         Thread runner = new Thread(this);
         runner.setName( "ByteArrayStream async reader" );
         runner.start();
      }
      
      public void run()
      {
         try
         {
            m_bytes_read = m_stream.read(m_buf, m_length);
            if( m_listener != null )
               m_listener.XAsyncCallback(this);
         }
         catch( Throwable t )
         {
            m_exception = t;
         }
      }
   }
   
   private final int CLOSED    = -1;
   private final int ABANDONED = -2;
   
   private byte[] m_bytes;

   private int m_ptr;

   /**
    * Constructs a new ByteArrayXStream with the specified data. This creates a
    * read only {@link XStream}.
    * 
    * @param bytes The XStream data
    */
   public ByteArrayXStream(byte[] bytes)
   {
      m_bytes = bytes;
      m_ptr = 0;
   }

   public void abandon()
      throws XAMException
   {
      m_ptr = ABANDONED;
   }

   public void close()
      throws XAMException
   {
      checkClosed();
      m_ptr = CLOSED;
   }

   public long read(byte[] buffer, long offset, long length)
      throws XAMException
   {
      checkClosed();
      if( offset > Integer.MAX_VALUE )
         throw new InvalidArgumentException("ByteArrayXStream offset is too large for implementation");
      
      if( length > Integer.MAX_VALUE )
         throw new InvalidArgumentException("ByteArrayXStream offset is too large for implementation");
      
      if (m_ptr == m_bytes.length) return -1;
      int total = Math.min(m_bytes.length - m_ptr, (int) length);
      System.arraycopy(m_bytes, m_ptr, buffer, (int) offset, total);
      m_ptr += total;
      return total;
   }

   public long read(byte[] buffer, long bytesToRead)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XAMException
   {
      return read( buffer, 0, bytesToRead );
   }

   public long read(byte[] buffer)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XAMException
   {
      return read( buffer, 0, buffer.length );
   }

   public void seek(long offset, long whence)
      throws XAMException
   {
      checkClosed();
      long ptr = m_ptr;
      switch ((int) whence)
      {
         case XStream.SEEK_SET :
            ptr = offset;
            break;
         case XStream.SEEK_END :
            ptr = m_bytes.length + offset;
            break;
         case XStream.SEEK_CUR :
            ptr += offset;
            break;
      }
      ptr = Math.max(0, Math.min(ptr, (long) m_bytes.length));
      m_ptr = (int) ptr;
   }

   public long tell()
      throws XAMException
   {
      checkClosed();
      return m_ptr;
   }

   public long write(byte[] buffer, long offset, long length)
      throws XAMException
   {
      checkClosed();
      throw new InvalidOperationException();
   }

   public long write(byte[] buffer)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XSetUnderHoldException, XAMException
   {
      checkClosed();
      throw new InvalidOperationException();
   }

   public long write(byte[] buffer, long bytesToWrite)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XSetUnderHoldException, XAMException
   {
      checkClosed();
      throw new InvalidOperationException();
   }

   /**
    * Checks to see if the XStream has been closed. If it has an
    * {@link InvalidOperationException} is thrown.
    * 
    * @throws XAMException If the XStream is closed
    */
   private void checkClosed()
      throws XAMException
   {
      if (m_ptr == CLOSED) throw new XStreamException("XStream closed");
      else if( m_ptr == ABANDONED) throw new XStreamAbandonException("XStream abandoned");
   }

   public XAsync asyncClose(long xopid, XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamCorruptException, XStreamException, XAMException
   {
      // TODO Auto-generated method stub
      return null;
   }

   public XAsync asyncRead(byte[] buffer,
                           long bufferLength,
                           long xopid,
                           XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XAMException
   {
      basXAsync async = new basXAsync( this, buffer, bufferLength, xopid, listener );
      async.read(); // start the read thread.
     return async;
   }

   public XAsync asyncWrite(byte[] buffer,
                            long bufferLength,
                            long xopid,
                            XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XAMException
   {
      // TODO Auto-generated method stub
      return null;
   }
}
