/* Copyright (c) 2008-2009, The Storage Networking Industry Association.
 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */
/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */
/* (c) Copyright 2009 Sun Microsystems, Inc.                    */

package org.snia.xam.util;

import java.util.Vector;

import org.snia.xam.AsyncHaltedException;
import org.snia.xam.AuthenticationExpiredException;
import org.snia.xam.InvalidArgumentException;
import org.snia.xam.InvalidOperationException;
import org.snia.xam.XAMException;
import org.snia.xam.XAsync;
import org.snia.xam.XAsyncListener;
import org.snia.xam.XSet;
import org.snia.xam.XSetUnderHoldException;
import org.snia.xam.XStream;
import org.snia.xam.XStreamAbandonException;
import org.snia.xam.XStreamCorruptException;
import org.snia.xam.XStreamException;
import org.snia.xam.XUID;

import org.snia.xam.testvim.TestXUID;

/** This class is only for testing. XUID byte buffers are returned when the 
 *  bytes are available. This simulates an XSystem job returning a XUID value every 100 mS. 
 *  It is used to test the ability of the XUIDIterator class to respond to slow XUID streams.
 *  This class provides a stream of padded XUID values. Each 80 bytes.
 */
public class SlowXUIDStream
   implements XStream
{
   class basXAsync implements XAsync, Runnable
   {
      long    m_bytes_read    = 0;
      long    m_op_id         = 0;
      XStream m_stream        = null;
      
      // Used to do the read
      byte m_buf[]  = null;
      long m_length = 0;
      XAsyncListener m_listener = null;
      Throwable m_exception     = null;
      
      basXAsync( XStream stream, byte buf[], long length, long id, XAsyncListener listener )
      {
         m_op_id    = id;
         m_listener = listener;
         m_buf      = buf;
         m_length   = length;
         m_stream   = stream;
      }
      
      public void close()
         throws AuthenticationExpiredException, XAMException
      { }

      public long getBytesRead()
         throws AuthenticationExpiredException, AsyncHaltedException,
         XAMException
      { return m_bytes_read; }

      public long getBytesWritten()
         throws AuthenticationExpiredException, AsyncHaltedException,
         XAMException
      { return 0; }

      public void getStatus()
         throws AuthenticationExpiredException, AsyncHaltedException,
         XAMException
      { 
         if( m_exception instanceof XAMException )
            throw (XAMException) m_exception;
         else if( m_exception != null )
            throw new XAMException( "Error halted async operation ", m_exception );
      }

      public long getXOPID()
         throws AuthenticationExpiredException, AsyncHaltedException,
         XAMException
      { return m_op_id; }

      public XSet getXSet()
         throws AuthenticationExpiredException, AsyncHaltedException,
         XAMException
      { return null; }

      public XStream getXStream()
         throws AuthenticationExpiredException, AsyncHaltedException,
         XAMException
      { return m_stream; }

      public XUID getXUID()
         throws AuthenticationExpiredException, AsyncHaltedException,
         XAMException
      { return null;  }

      public void halt()
         throws AuthenticationExpiredException, XAMException
      {  }

      public boolean isComplete()
         throws AuthenticationExpiredException, XAMException
      { return true; }
      
      void read() 
      {
         Thread runner = new Thread(this);
         runner.setName( "ByteArrayStream async reader" );
         runner.start();
      }
      
      public void run()
      {
         try
         {
            m_bytes_read = m_stream.read(m_buf, m_length);
            if( m_listener != null )
               m_listener.XAsyncCallback(this);
         }
         catch( Throwable t )
         {
            m_exception = t;
         }
      }
   }
   
   private final int CLOSED    = -2;
   private final int ABANDONED = -1;
   private final int OPENED    =  0;
   
   // XUIDs are provided not more than once every 100 ms.
   final long XUID_DELAY = 100;
   
   private      int    m_state     = OPENED;
   private      long   m_last_time = System.currentTimeMillis();
   private      byte   m_buffer[]  = null;
   private      int    m_cursor    = 0; // remember where in the buffer was reading, in case app reads one byte @ a time
   private      int    m_max_xuids = 0;
   private      int    m_count     = 0;
   private      Vector m_xuids     = new Vector();
   
   /**
    * Create a slow XUID stream that produce exactly <i>xuid_count</i> XUID values.
    * @param xuid_count
    */
   public SlowXUIDStream( int xuid_count )
   {
      m_max_xuids = xuid_count;
      m_count     = 0;
      m_state = OPENED;
   }
   
   public void abandon()
      throws AuthenticationExpiredException, XAMException
   {
      if( m_state != OPENED )
         throw new XStreamException( "XStream in wrong state, can not be abandoned" );
      
      m_state = ABANDONED;
   }

   public XAsync asyncClose(long xopid, XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamCorruptException, XStreamException, XAMException
   {
      throw new InvalidOperationException( "Not implemented for the test case" );
   }

   public XAsync asyncRead(byte[] buffer,
                           long bufferLength,
                           long xopid,
                           XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XAMException
   {
      basXAsync async = new basXAsync( this, buffer, bufferLength, xopid, listener );
      async.read(); // start the read thread.
      return async;
   }

   public XAsync asyncWrite(byte[] buffer,
                            long bufferLength,
                            long xopid,
                            XAsyncListener listener)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XAMException
   {
      throw new InvalidOperationException( "Not implemented for the test case" );
   }

   public void close()
      throws AuthenticationExpiredException, XSetUnderHoldException,
      XStreamCorruptException, XStreamException, XAMException
   {
      if( (m_state != OPENED) || (m_state != ABANDONED) )
         throw new XStreamException( "XSteam in wrong state, can't close" );

      m_state = CLOSED;
   }
   
   /**
    * Return the "nth" XUID that was generated by the stream.
    * @param pos
    * @return
    */
   public XUID elementAt( int pos )
   {
      return (XUID) m_xuids.elementAt(pos);
   }

   /** This is the only implemented method of note. XUID byte buffers are returned when the 
    *  bytes are available. This simulates an XSystem job returning a XUID value every 100 mS. 
    *  It is used to test the ability of the XUIDIterator class to respond to slow XUID streams.
    */
   public long read(byte[] buffer, long offset, long bytesToRead)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XAMException
   {
      if( m_count >= m_max_xuids )
         return XStream.EOF;
      
      if( m_buffer == null )
      {
         long current_time = System.currentTimeMillis();
         if( (current_time - m_last_time) < XUID_DELAY )
            return 0;
         else 
         {
            XUID x = TestXUID.newXUID();
            m_xuids.add( x );
            m_buffer = x.toBytes();
            m_cursor = 0;
            m_count++;
         }
      }
      
      // Clamp to no more than the length of the XUID, this stream only 
      // can possibly return 1 XUID at a time, and must not block until more
      // are ready. 
      if( bytesToRead > XUID.MAX_LENGTH )
         bytesToRead = XUID.MAX_LENGTH;
      
      // Copy the actual XUID bytes
      System.arraycopy( m_buffer, m_cursor, buffer, (int) offset, (int) m_buffer.length );
      
      // Pad the rest
      for( int i = m_buffer.length; i < bytesToRead; i++ )
         buffer[i] = 0;
      
      m_cursor += bytesToRead;
      if( m_cursor >= XUID.MAX_LENGTH )
      {
         m_buffer = null;
         m_cursor = 0;
      }
      return bytesToRead;
   }

   public long read(byte[] buffer, long bytesToRead)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XAMException
   {
      return read( buffer, 0, bytesToRead );
   }

   public long read(byte[] buffer)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XAMException
   {
      return read( buffer, 0, buffer.length );
   }

   public void seek(long offset, long whence)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XAMException
   {
      throw new InvalidOperationException( "Not implemented for the test case" );
   }

   public long tell()
      throws AuthenticationExpiredException, XStreamAbandonException,
      XStreamCorruptException, XAMException
   {
      throw new InvalidOperationException( "Not implemented for the test case" );
   }

   public long write(byte[] buffer)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XSetUnderHoldException, XAMException
   {
      throw new InvalidOperationException( "Not implemented for the test case" );
   }

   public long write(byte[] buffer, long bytesToWrite)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XSetUnderHoldException, XAMException
   {
      throw new InvalidOperationException( "Not implemented for the test case" );
   }

   public long write(byte[] buffer, long offset, long bytesToWrite)
      throws AuthenticationExpiredException, InvalidArgumentException,
      XStreamAbandonException, XStreamCorruptException, XStreamException,
      XSetUnderHoldException, XAMException
   {
      throw new InvalidOperationException( "Not implemented for the test case" );
   }

}
