/* Copyright (c) 2008-2009, The Storage Networking Industry Association.
 Redistribution and use in source and binary forms, with or without modification, 
 are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright notice, 
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.
 
 * Neither the name of The Storage Networking Industry Association (SNIA) nor 
 the names of its contributors may be used to endorse or promote products 
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 THE POSSIBILITY OF SUCH DAMAGE.
 */
/* (c) Copyright 2008 Hewlett-Packard Development Company, L.P. */
/* (c) Copyright 2009 Sun Microsystems, Inc.                    */

package org.snia.xam.util;

import org.snia.xam.InvalidXUIDException;
import org.snia.xam.XStream;
import org.snia.xam.XUID;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestResult;
import junit.framework.TestSuite;

import org.snia.xam.util.ByteArrayXStream;

import org.snia.xam.testvim.TestXUID;

import java.util.Vector;

/**
 * Test the XAMLibraryFactory class. This test suite will test the positive
 * create path, with no arguments, and with a property path.
 */
public class XUIDIteratorTest
   extends TestCase
{
   public static Test suite()
   {
      TestSuite suite = new TestSuite();
      suite.addTest(new XUIDIteratorTest("testSimpleIteration"));
      suite.addTest(new XUIDIteratorTest("testMultiBatchIteration"));
      suite.addTest(new XUIDIteratorTest("testCorruptIteration"));
      suite.addTest(new XUIDIteratorTest("testWaitingForStream"));
      return suite;
   }

   public XUIDIteratorTest(String test)
   {
      super(test);
   }

   /**
    * Run suite, so just these unit tests can be run.
    * @param args
    */
   public static void main(String[] args)
   {
      TestResult res = junit.textui.TestRunner.run(suite());
      System.exit(Math.max(res.failureCount(), res.errorCount()));
   }

   void dumpXUID( byte[] buf, int at )
   {
      for( int i = at; (i < (at + XUID.MAX_LENGTH)) && (i < buf.length); i++ )
      {
         System.out.print( buf[i] );
         System.out.print( " " );
      }
      System.out.println(">*<");
   }
   
   /**
    * Simple test, Makes sure that the iterator handles the easy case.
    */
   public void testSimpleIteration()
      throws Exception
   {
      final int XUID_COUNT = 10;
      byte  xuid_bytes[] = new byte[XUID_COUNT*XUID.MAX_LENGTH];
      Vector xuids = new Vector();
      
      for( int i = 0; i < XUID_COUNT; i++ )
      {
         XUID x = TestXUID.newXUID();
         xuids.add(x);
         int xuid_length = x.toBytes().length;
         System.arraycopy( x.toBytes(), 0, xuid_bytes, i*XUID.MAX_LENGTH, xuid_length); 
      }
      
      XStream xuid_stream = new ByteArrayXStream(xuid_bytes);
      
      XUIDIterator xi = new XUIDIterator( xuid_stream );
      int count = 0;
      while( xi.hasNext() )
      {
         XUID test_xuid = (XUID) xi.next();
         XUID src_xuid  = (XUID) xuids.elementAt(count);
         assertEquals( "XUID #" + count + " didn't compare properly", src_xuid, test_xuid );
         count++;
      }
   }

   /**
    * Simple test, Makes sure that the iterator handles the multiple batches
    * of XUIDs.
    */
   public void testMultiBatchIteration()
      throws Exception
   {
      // This will result in 2 complete batches and a partial batch.
      final int XUID_COUNT = (int) (XUIDIterator.XUIDS_PER_READ * 2.5);
      
      byte  xuid_bytes[] = new byte[XUID_COUNT*XUID.MAX_LENGTH];
      Vector xuids = new Vector();
      
      for( int i = 0; i < XUID_COUNT; i++ )
      {
         XUID x = TestXUID.newXUID();
         xuids.add(x);
         int xuid_length = x.toBytes().length;
         System.arraycopy( x.toBytes(), 0, xuid_bytes, i*XUID.MAX_LENGTH, xuid_length); 
      }
      
      XStream xuid_stream = new ByteArrayXStream(xuid_bytes);
      
      XUIDIterator xi = new XUIDIterator( xuid_stream );
      int count = 0;
      while( xi.hasNext() )
      {
         XUID test_xuid = (XUID) xi.next();
         XUID src_xuid  = (XUID) xuids.elementAt(count);
         assertEquals( "XUID #" + count + " didn't compare properly", src_xuid, test_xuid );
         count++;
      }
      assertEquals( "Did not find appropriate number of XUIDs in stream", XUID_COUNT, count );
   }

   /**
    * Insert corrupted (shortened) XUID
    */
   public void testCorruptIteration()
      throws Exception
   {
      final int XUID_COUNT = 10;
      byte  xuid_bytes[] = new byte[XUID_COUNT*XUID.MAX_LENGTH];
      Vector xuids = new Vector();
      
      for( int i = 0; i < XUID_COUNT; i++ )
      {
         XUID x = TestXUID.newXUID();
         xuids.add(x);
         int xuid_length = x.toBytes().length;
         
         // This is the defect insertion and should cause the iterator to fail.
         if( i == 5 )
            xuid_length = xuid_length / 2;
         System.arraycopy( x.toBytes(), 0, xuid_bytes, i*XUID.MAX_LENGTH, xuid_length); 
      }
      
      XStream xuid_stream = new ByteArrayXStream(xuid_bytes);
      
      XUIDIterator xi = new XUIDIterator( xuid_stream );
      int count = 0;
      try
      {
         while( xi.hasNext() )
         {
            XUID test_xuid = (XUID) xi.next();
            XUID src_xuid  = (XUID) xuids.elementAt(count);
            assertEquals( "XUID #" + count + " didn't compare properly", src_xuid, test_xuid );
            count++;
         }
         fail( "Should have thrown a corrupt XUID exception during iteration" );
      }
      catch( Throwable ixe )
      {
         assertTrue( "Exception not caused by InvalidXUIDException", 
            ixe.getCause() instanceof InvalidXUIDException );
      }
   }

   /**
    * Test the case where the stream may not have the XUIDs available as soon as the 
    * iterator needs them. This is simulated using a fake stream that will provide a new 
    * XUID every 100 ms.
    */
   public void testWaitingForStream()
      throws Exception
   {
      final int XUID_COUNT = 29; // Picked this value more or less at random.
      SlowXUIDStream slow_stream = new SlowXUIDStream( XUID_COUNT );
            
      XUIDIterator xi = new XUIDIterator( slow_stream );
      int count = 0;
      while( xi.hasNext() )
      {
         XUID test_xuid = (XUID) xi.next();
         XUID src_xuid  = slow_stream.elementAt(count);
         assertEquals( "XUID #" + count + " didn't compare properly", src_xuid, test_xuid );
         count++;
      }
      assertEquals( "Not enough XUIDs from the SlowXUIDStream", XUID_COUNT, count );
   }
}
