////////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2007, The Storage Networking Industry Association.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - Neither the name of The Storage Networking Industry Association (SNIA)
//      nor the names of its contributors may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
// ============================================================================
//  Copyright (c) 2007, EMC Corporation.
//
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions 
//  are met:
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright 
//      notice, this list of conditions and the following disclaimer in the 
//      documentation and/or other materials provided with the distribution.
//    - The name of EMC Corporation may not be used to endorse or promote 
//      products derived from this software without specific prior written 
//      permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
//  THE POSSIBILITY OF SUCH DAMAGE.
////////////////////////////////////////////////////////////////////////////////
//  XAM SDK Source File Build Version @version.full@
////////////////////////////////////////////////////////////////////////////////

#include "XAM_VIM.h"
#include "XAM_POSIX.h"
#include "XAM_Exception.h"
#include "XAM_Status.h"
#include "XAM_Logger.h"

xam_boolean
XAM_VIM::loadLibrary(const char* inPath)
{
    xam_boolean vSuccess = false;
    char* vAdjustedLibName = (char*)0;

    XAM_LOG_INFO_P2(component_lib, "Attempting to load library: \"%s\"", inPath);

    // Load the library using platform specific dlopen semantics
    mVimLibraryHandle = XAM_POSIX::xam_dlopen(inPath, LAZY);

    if(mVimLibraryHandle == (void*)0)
    {
        // On failure, try again with an adjusted library name (if possible)
        if(mVimLibraryHandle == (void*)0)
        {
            XAM_LOG_ERROR_P2(component_lib, "Failed to load library: \"%s\"", inPath);

            xam_string vLibName;
            xam_string vPath;
            XAM_POSIX::xam_strcpy(vPath, inPath);
            parseLibNameFromPath(&vPath, &vLibName);

            vAdjustedLibName = XAM_POSIX::xam_adjust_library_name(vLibName);
            if(vAdjustedLibName)
            {
                XAM_LOG_INFO_P2(component_lib, "Adjusting library name: \"%s\"", vAdjustedLibName);
                mVimLibraryHandle = XAM_POSIX::xam_dlopen(vAdjustedLibName, LAZY);
                if(mVimLibraryHandle == (void*)0)
                    XAM_LOG_ERROR_P2(component_lib, "Failed to load adjusted library: \"%s\"", vAdjustedLibName);
            }
        }
    }

    if(mVimLibraryHandle != (void*)0)    
    {
        // Check for the connect method to positively identify this 
        // library as a VIM
        if(!XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_CreateXSystem"))
        {
            XAM_LOG_ERROR_P2(component_lib, "Required symbol \"VIM_CreateXSystem\" not found in: \"%s\"", inPath);

            // The symbol was not found, close the open handle
            XAM_POSIX::xam_dlclose(mVimLibraryHandle);
            mVimLibraryHandle = (void*)0;
            vSuccess = false;
        }
        else
        {
            // The symbol lookup was successful
            vSuccess = true;
        }
    }

    if(vSuccess)
    {
        // Store the library path and name
        XAM_POSIX::xam_strcpy(mLibPath, vAdjustedLibName ? vAdjustedLibName : inPath);   
        parseLibNameFromPath(&mLibPath, &mLibName);

        XAM_LOG_INFO_P2(component_lib, "Successfully loaded library: \"%s\"", mLibName);
    }

    if(vAdjustedLibName != (char*)0)
        XAM_POSIX::xam_free(vAdjustedLibName);

    return vSuccess;
}

void
XAM_VIM::closeLibrary()
{
    if(mVimLibraryHandle != (void*)0)
    {
        // Close the open library handle
        XAM_POSIX::xam_dlclose(mVimLibraryHandle);
        mVimLibraryHandle = (void*)0;
    }
}

const char* 
XAM_VIM::getName()
{
    return mLibName;
}   

// ============================================================================
// Methods to handle xam_status 
// ============================================================================

xam_boolean
XAM_VIM::XSystem_GetErrorToken (const xsystem_handle inHandle,
                                const xam_status inStatus,
                                xam_string* const outToken)
{
    xam_boolean vFoundToken = false;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSystem_GetErrorToken (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_param, 
                        "[in] [xam_status] %i", inStatus);
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_param, 
                        "[in] [xam_string*] %p", outToken);
    }

    if(mVimLibraryHandle == (void*)0)
    {
        // The VIM library handle for the object is invalid. 
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_GetErrorToken == 0)
        {
            VIM_XSystem_GetErrorToken = (XSystem_GetErrorTokenFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_GetErrorToken");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_GetErrorToken != 0)
        {
            vFoundToken = VIM_XSystem_GetErrorToken(inHandle, inStatus, outToken);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_GetErrorToken");
    }

    return vFoundToken;
}


// =============================================================================
// Methods for the XIterator 
// =============================================================================

void
XAM_VIM::XSystem_OpenFieldIterator (const xsystem_handle inHandle,  
                                    const xam_string inString,
                                    xiterator_handle* const outIterator)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO,   
                        component_vim_enter, 
                        "VIM_XSystem_OpenFieldIterator (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inString));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xiterator_handle] %"XAM_INT_FORMAT"d", 
                        outIterator);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.         
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_OpenFieldIterator == 0)
        {
            VIM_XSystem_OpenFieldIterator = (XSystem_OpenFieldIteratorFunc)
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_OpenFieldIterator");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_OpenFieldIterator != 0)
        {
            vStatus = VIM_XSystem_OpenFieldIterator(inHandle, 
                                                      inString, 
                                                      outIterator);

        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_OpenFieldIterator");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void
XAM_VIM::XSet_OpenFieldIterator (const xset_handle inHandle,  
                                   const xam_string inString,
                                   xiterator_handle* const outIterator)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO,   
                        component_vim_enter, 
                        "VIM_XSet_OpenFieldIterator (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inString));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xiterator_handle] %"XAM_INT_FORMAT"d", 
                        outIterator);
    }

    if(mVimLibraryHandle == (void*)0)
    {
        // The VIM library handle for the object is invalid.         
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_OpenFieldIterator == 0)
        {
            VIM_XSet_OpenFieldIterator = (XSet_OpenFieldIteratorFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_OpenFieldIterator");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_OpenFieldIterator != 0)
        {
            vStatus = VIM_XSet_OpenFieldIterator(inHandle, 
                                                   inString, 
                                                   outIterator);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_OpenFieldIterator");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void
XAM_VIM::XIterator_HasNext (const xiterator_handle inHandle, 
                            xam_boolean* const outValue)
{
  xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XIterator_HasNext (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xiterator_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean*] %p", 
                        outValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.                 
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XIterator_HasNext == 0)
        {
            VIM_XIterator_HasNext = (XIterator_HasNextFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XIterator_HasNext");
        }

        // If valid, utilize the function pointer
        if(VIM_XIterator_HasNext != 0)
        {
            vStatus = VIM_XIterator_HasNext(inHandle, outValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XIterator_HasNext");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }    
}

void 
XAM_VIM::XIterator_Next (const xiterator_handle inHandle, 
                         xam_string* const outName)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XIterator_Next (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xiterator_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string*] %p", 
                        outName);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.                 
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XIterator_Next == 0)
        {
            VIM_XIterator_Next = (XIterator_NextFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XIterator_Next");
        }

        // If valid, utilize the function pointer
        if(VIM_XIterator_Next != 0)
        {
            vStatus = VIM_XIterator_Next(inHandle, outName);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit XIterator_Next");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 
  
void 
XAM_VIM::XIterator_Close (xiterator_handle inHandle)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO,   
                        component_vim_enter, 
                        "VIM_XIterator_Close (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG,
                        component_vim_param, 
                        "[in] [xiterator_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.                         
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XIterator_Close == 0)
        {
            VIM_XIterator_Close = (XIterator_CloseFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XIterator_Close");
        }

        // If valid, utilize the function pointer
        if(VIM_XIterator_Close != 0)
        {
            vStatus = VIM_XIterator_Close(inHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit XIterator_Close");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

// =============================================================================
// Methods for managing XSystem Fields (properties or XStreams)
// =============================================================================

void 
XAM_VIM::XSystem_ContainsField (const xsystem_handle inHandle,
                              const xam_string inName,
                              xam_boolean* const outExists)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO,   
                        component_vim_enter, 
                        "VIM_XSystem_ContainsField (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean*] %p", 
                        outExists);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.                         
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_ContainsField == 0)
        {
            VIM_XSystem_ContainsField = (XSystem_ContainsFieldFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_ContainsField");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_ContainsField != 0)
        {
            vStatus = VIM_XSystem_ContainsField(inHandle, 
                                              inName, 
                                              outExists);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_ContainsField");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
}

void 
XAM_VIM::XSystem_GetFieldType (const xsystem_handle inHandle,
                               const xam_string inName,
                               xam_string* const outType)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO,   
                        component_vim_enter, 
                        "VIM_XSystem_GetFieldType (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_handle_t] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string*] %p", 
                        outType);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.                         
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_GetFieldType == 0)
        {
            VIM_XSystem_GetFieldType = (XSystem_GetFieldTypeFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_GetFieldType");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_GetFieldType != 0)
        {
            vStatus = VIM_XSystem_GetFieldType(inHandle, 
                                               inName, 
                                               outType);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_GetFieldType");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_GetFieldLength (const xsystem_handle inHandle,
                                 const xam_string inName,
                                 xam_int* const outLength)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO,   
                        component_vim_enter, 
                        "VIM_XSystem_GetFieldLength (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_handle_t] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_int*] %p", 
                        outLength);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.                         
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_GetFieldLength == 0)
        {
            VIM_XSystem_GetFieldLength = (XSystem_GetFieldLengthFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_GetFieldLength");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_GetFieldLength != 0)
        {
            vStatus = VIM_XSystem_GetFieldLength(inHandle, inName, outLength);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_GetFieldLength");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_GetFieldReadOnly (const xsystem_handle inHandle,
                                   const xam_string inName,
                                   xam_boolean* const outReadOnly)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO,   
                        component_vim_enter, 
                        "VIM_XSystem_GetFieldReadOnly (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean*] %p", 
                        outReadOnly);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.                         
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_GetFieldReadOnly == 0)
        {
            VIM_XSystem_GetFieldReadOnly = (XSystem_GetFieldReadOnlyFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_GetFieldReadOnly");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_GetFieldReadOnly != 0)
        {
            vStatus = VIM_XSystem_GetFieldReadOnly(inHandle, 
                                                   inName, 
                                                   outReadOnly);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_GetFieldReadOnly");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_DeleteField (const xsystem_handle inHandle,
                              const xam_string inName)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO,   
                        component_vim_enter, 
                        "VIM_XSystem_DeleteField (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_DeleteField == 0)
        {
            VIM_XSystem_DeleteField = (XSystem_DeleteFieldFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_DeleteField");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_DeleteField != 0)
        {
            vStatus = VIM_XSystem_DeleteField(inHandle, inName);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_DeleteField");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

// =============================================================================
// Methods for managing XSystem property fields
// =============================================================================

void 
XAM_VIM::XSystem_CreateBoolean (const xsystem_handle inHandle,
                                const xam_string inName,
                                const xam_boolean inBinding,
                                const xam_boolean inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO,   
                        component_vim_enter, 
                        "VIM_XSystem_CreateBoolean (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean] %s", 
                        inValue ? "true" : "false");
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_CreateBoolean == 0)
        {
            VIM_XSystem_CreateBoolean = (XSystem_CreateBooleanFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_CreateBoolean");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_CreateBoolean != 0)
        {
            vStatus = VIM_XSystem_CreateBoolean(inHandle, 
                                                inName, 
                                                inBinding, 
                                                inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_CreateBoolean");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_CreateInt (const xsystem_handle inHandle,
                            const xam_string inName,
                            const xam_boolean inBinding,
                            const xam_int inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO,   
                        component_vim_enter, 
                        "VIM_XSystem_CreateInt (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_int] %li", 
                        inValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_CreateInt == 0)
        {
            VIM_XSystem_CreateInt = (XSystem_CreateIntFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_CreateInt");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_CreateInt != 0)
        {
            vStatus = VIM_XSystem_CreateInt(inHandle, 
                                            inName, 
                                            inBinding, 
                                            inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_CreateInt");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_CreateDouble (const xsystem_handle inHandle,
                               const xam_string inName,
                               const xam_boolean inBinding,
                               const xam_double inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO,   
                        component_vim_enter, 
                        "VIM_XSystem_CreateDouble (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_double] %f", 
                        inValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_CreateDouble == 0)
        {
            VIM_XSystem_CreateDouble = (XSystem_CreateDoubleFunc) 
                   XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                        "VIM_XSystem_CreateDouble");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_CreateDouble != 0)
        {
            vStatus = VIM_XSystem_CreateDouble(inHandle, 
                                              inName, 
                                              inBinding, 
                                              inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_CreateDouble");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_CreateXUID (const xsystem_handle inHandle,
                             const xam_string inName,
                             const xam_boolean inBinding,
                             const xam_xuid inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO,   
                        component_vim_enter, 
                        "VIM_XSystem_CreateXUID (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_xuid] \"%s\"", 
                        inValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_CreateXUID == 0)
        {
            VIM_XSystem_CreateXUID = (XSystem_CreateXUIDFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_CreateXUID");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_CreateXUID != 0)
        {
            vStatus = VIM_XSystem_CreateXUID(inHandle, 
                                             inName, 
                                             inBinding, 
                                             inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_CreateXUID");
    }

    if(vStatus != (xam_status)0)
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_CreateString (const xsystem_handle inHandle,
                           const xam_string inName,
                           const xam_boolean inBinding,
                           const xam_string inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_CreateString (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inValue));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_CreateString == 0)
        {
            VIM_XSystem_CreateString = (XSystem_CreateStringFunc) XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XSystem_CreateString");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_CreateString != 0)
        {
            vStatus = VIM_XSystem_CreateString(inHandle, inName, inBinding, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_CreateString");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_CreateDatetime (const xsystem_handle inHandle,
                                 const xam_string inName,
                                 const xam_boolean inBinding,
                                 const xam_datetime inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_CreateDatetime (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_datetime] %s", inValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_CreateDatetime == 0)
        {
            VIM_XSystem_CreateDatetime = (XSystem_CreateDatetimeFunc) XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XSystem_CreateDatetime");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_CreateDatetime != 0)
        {
            vStatus = VIM_XSystem_CreateDatetime(inHandle, inName, inBinding, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_CreateDatetime");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_SetBoolean (const xsystem_handle inHandle,
                             const xam_string inName,
                             const xam_boolean inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_SetBoolean (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", inValue ? "true" : "false");
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_SetBoolean == 0)
        {
            VIM_XSystem_SetBoolean = (XSystem_SetBooleanFunc) XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XSystem_SetBoolean");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_SetBoolean != 0)
        {
            vStatus = VIM_XSystem_SetBoolean(inHandle, inName, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_SetBoolean");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_SetInt (const xsystem_handle inHandle,
                         const xam_string inName,
                         const xam_int inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_SetInt (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int] %li", inValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_SetInt == 0)
        {
            VIM_XSystem_SetInt = (XSystem_SetIntFunc) XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XSystem_SetInt");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_SetInt != 0)
        {
            vStatus = VIM_XSystem_SetInt(inHandle, inName, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_SetInt");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_SetDouble (const xsystem_handle inHandle,
                           const xam_string inName,
                           const xam_double inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSystem_SetDouble (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG,
                        component_vim_param, 
                        "[in] [xam_double] %f", 
                        inValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_SetDouble == 0)
        {
            VIM_XSystem_SetDouble = (XSystem_SetDoubleFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_SetDouble");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_SetDouble != 0)
        {
            vStatus = VIM_XSystem_SetDouble(inHandle, inName, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_SetDouble");
    }

    if(vStatus != (xam_status)0)
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_SetXUID (const xsystem_handle inHandle,
                          const xam_string inName,
                          const xam_xuid inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSystem_SetXUID (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_xuid] \"%s\"", 
                        inValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_SetXUID == 0)
        {
            VIM_XSystem_SetXUID = (XSystem_SetXUIDFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_SetXUID");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_SetXUID != 0)
        {
            vStatus = VIM_XSystem_SetXUID(inHandle, inName, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_SetXUID");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_SetString (const xsystem_handle inHandle,
                            const xam_string inName,
                            const xam_string inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSystem_SetString (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] %s", 
                        _logString(inValue));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_SetString == 0)
        {
            VIM_XSystem_SetString = (XSystem_SetStringFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_SetString");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_SetString != 0)
        {
            vStatus = VIM_XSystem_SetString(inHandle, inName, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_SetString");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_SetDatetime (const xsystem_handle inHandle,
                              const xam_string inName,
                              const xam_datetime inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSystem_SetDatetime (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_datetime] %s", 
                        inValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_SetDatetime == 0)
        {
            VIM_XSystem_SetDatetime = (XSystem_SetDatetimeFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_SetDatetime");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_SetDatetime != 0)
        {
            vStatus = VIM_XSystem_SetDatetime(inHandle, inName, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_SetDatetime");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_GetBoolean (const xsystem_handle inHandle,
                             const xam_string inName,
                             xam_boolean* const outValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSystem_GetBoolean (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean*] %p", 
                        outValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_GetBoolean == 0)
        {
            VIM_XSystem_GetBoolean = (XSystem_GetBooleanFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_GetBoolean");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_GetBoolean != 0)
        {
            vStatus = VIM_XSystem_GetBoolean(inHandle, inName, outValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_GetBoolean");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_GetInt (const xsystem_handle inHandle,
                         const xam_string inName,
                         xam_int* const outValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSystem_GetInt (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_int*] %p", 
                        outValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_GetInt == 0)
        {
            VIM_XSystem_GetInt = (XSystem_GetIntFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_GetInt");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_GetInt != 0)
        {
            vStatus = VIM_XSystem_GetInt(inHandle, inName, outValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_GetInt");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_GetDouble (const xsystem_handle inHandle,
                           const xam_string inName,
                           xam_double* const outValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSystem_GetDouble (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_double*] %p", 
                        outValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_GetDouble == 0)
        {
            VIM_XSystem_GetDouble = (XSystem_GetDoubleFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_GetDouble");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_GetDouble != 0)
        {
            vStatus = VIM_XSystem_GetDouble(inHandle, inName, outValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_GetDouble");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_GetXUID (const xsystem_handle inHandle,
                          const xam_string inName,
                          xam_xuid* const outValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSystem_GetXUID (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_xuid*] %p", 
                        outValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_GetXUID == 0)
        {
            VIM_XSystem_GetXUID = (XSystem_GetXUIDFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_GetXUID");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_GetXUID != 0)
        {
            vStatus = VIM_XSystem_GetXUID(inHandle, inName, outValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_GetXUID");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_GetString (const xsystem_handle inHandle,
                            const xam_string inName,
                            xam_string* const outValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSystem_GetString (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string*] %p", 
                        outValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_GetString == 0)
        {
            VIM_XSystem_GetString = (XSystem_GetStringFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_GetString");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_GetString != 0)
        {
            vStatus = VIM_XSystem_GetString(inHandle, inName, outValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_GetString");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_GetDatetime (const xsystem_handle inHandle,
                              const xam_string inName,
                              xam_datetime* const outValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSystem_GetDatetime (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_datetime*] %p", 
                        outValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_GetDatetime == 0)
        {
            VIM_XSystem_GetDatetime = (XSystem_GetDatetimeFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_GetDatetime");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_GetDatetime != 0)
        {
            vStatus = VIM_XSystem_GetDatetime(inHandle, inName, outValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_GetDatetime");
    }

    if(vStatus != (xam_status)0)
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

// =============================================================================
// Methods for managing XSet Fields (properties or XStreams)
// =============================================================================

void 
XAM_VIM::XSet_ContainsField (const xam_handle_t inHandle,
                           const xam_string inName,
                           xam_boolean* const outExists)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_ContainsField (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean*] %p", 
                        outExists);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_ContainsField == 0)
        {
            VIM_XSet_ContainsField = (XSet_ContainsFieldFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_ContainsField");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_ContainsField != 0)
        {
            vStatus = VIM_XSet_ContainsField(inHandle, inName, outExists);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_ContainsField");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
}

void 
XAM_VIM::XSet_SetFieldAsBinding (const xset_handle inHandle,
                                 const xam_string inName)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_SetFieldAsBinding (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_SetFieldAsBinding == 0)
        {
            VIM_XSet_SetFieldAsBinding = (XSet_SetFieldAsBindingFunc) XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XSet_SetFieldAsBinding");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_SetFieldAsBinding != 0)
        {
            vStatus = VIM_XSet_SetFieldAsBinding(inHandle, inName);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_SetFieldAsBinding");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_SetFieldAsNonbinding (const xset_handle inHandle,
                                    const xam_string inName)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_SetFieldAsNonbinding (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_SetFieldAsNonbinding == 0)
        {
            VIM_XSet_SetFieldAsNonbinding = (XSet_SetFieldAsNonbindingFunc) 
                    XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                    "VIM_XSet_SetFieldAsNonbinding");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_SetFieldAsNonbinding != 0)
        {
            vStatus = VIM_XSet_SetFieldAsNonbinding(inHandle, inName);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_SetFieldAsNonbinding");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_GetFieldType (const xset_handle inHandle,
                            const xam_string inName,
                            xam_string* const outType)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_GetFieldType (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string*] %p", 
                        outType);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_GetFieldType == 0)
        {
            VIM_XSet_GetFieldType = (XSet_GetFieldTypeFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_GetFieldType");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_GetFieldType != 0)
        {
            vStatus = VIM_XSet_GetFieldType(inHandle, inName, outType);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_GetFieldType");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_GetFieldLength (const xset_handle inHandle,
                              const xam_string inName,
                              xam_int* const outLength)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_GetFieldLength (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_int*] %p", 
                        outLength);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_GetFieldLength == 0)
        {
            VIM_XSet_GetFieldLength = (XSet_GetFieldLengthFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_GetFieldLength");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_GetFieldLength != 0)
        {
            vStatus = VIM_XSet_GetFieldLength(inHandle, inName, outLength);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_GetFieldLength");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_GetFieldBinding (const xset_handle inHandle,
                               const xam_string inName,
                               xam_boolean* const outFixed)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_GetFieldBinding (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean*] %p", 
                        outFixed);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_GetFieldBinding == 0)
        {
            VIM_XSet_GetFieldBinding = (XSet_GetFieldBindingFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_GetFieldBinding");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_GetFieldBinding != 0)
        {
            vStatus = VIM_XSet_GetFieldBinding(inHandle, inName, outFixed);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_GetFieldBinding");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_GetFieldReadOnly (const xset_handle inHandle,
                               const xam_string inName,
                               xam_boolean* const outReadOnly)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_GetFieldReadOnly (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean*] %p", 
                        outReadOnly);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_GetFieldReadOnly == 0)
        {
            VIM_XSet_GetFieldReadOnly = (XSet_GetFieldReadOnlyFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_GetFieldReadOnly");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_GetFieldReadOnly != 0)
        {
            vStatus = VIM_XSet_GetFieldReadOnly(inHandle, inName, outReadOnly);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_GetFieldReadOnly");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
} 

void 
XAM_VIM::XSet_DeleteField (const xset_handle inHandle,
                           const xam_string inName)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_DeleteField (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_DeleteField == 0)
        {
            VIM_XSet_DeleteField = (XSet_DeleteFieldFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_DeleteField");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_DeleteField != 0)
        {
            vStatus = VIM_XSet_DeleteField(inHandle, inName);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_DeleteField");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

// =============================================================================
// Methods for managing XSet property fields
// =============================================================================

void 
XAM_VIM::XSet_CreateBoolean (const xset_handle inHandle,
                             const xam_string inName,
                             const xam_boolean inBinding,
                             const xam_boolean inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_CreateBoolean (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean] %s", 
                        inValue ? "true" : "false");
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_CreateBoolean == 0)
        {
            VIM_XSet_CreateBoolean = (XSet_CreateBooleanFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_CreateBoolean");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_CreateBoolean != 0)
        {
            vStatus = VIM_XSet_CreateBoolean(inHandle, inName, inBinding, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_CreateBoolean");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_CreateInt (const xset_handle inHandle,
                         const xam_string inName,
                         const xam_boolean inBinding,
                         const xam_int inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_CreateInt (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_int] %li", 
                        inValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_CreateInt == 0)
        {
            VIM_XSet_CreateInt = (XSet_CreateIntFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_CreateInt");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_CreateInt != 0)
        {
            vStatus = VIM_XSet_CreateInt(inHandle, inName, inBinding, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_CreateInt");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_CreateDouble (const xset_handle inHandle,
                           const xam_string inName,
                           const xam_boolean inBinding,
                           const xam_double inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_CreateDouble (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_double] %f", 
                        inValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_CreateDouble == 0)
        {
            VIM_XSet_CreateDouble = (XSet_CreateDoubleFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_CreateDouble");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_CreateDouble != 0)
        {
            vStatus = VIM_XSet_CreateDouble(inHandle, inName, inBinding, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_CreateDouble");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_CreateXUID (const xset_handle inHandle,
                          const xam_string inName,
                          const xam_boolean inBinding,
                          const xam_xuid inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_CreateXUID (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_xuid] \"%s\"", 
                        inValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_CreateXUID == 0)
        {
            VIM_XSet_CreateXUID = (XSet_CreateXUIDFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_CreateXUID");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_CreateXUID != 0)
        {
            vStatus = VIM_XSet_CreateXUID(inHandle, inName, inBinding, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_CreateXUID");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_CreateString (const xset_handle inHandle,
                            const xam_string inName,
                            const xam_boolean inBinding,
                            const xam_string inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_CreateString (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inValue));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_CreateString == 0)
        {
            VIM_XSet_CreateString = (XSet_CreateStringFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_CreateString");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_CreateString != 0)
        {
            vStatus = VIM_XSet_CreateString(inHandle, inName, inBinding, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_CreateString");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_CreateDatetime (const xset_handle inHandle,
                              const xam_string inName,
                              const xam_boolean inBinding,
                              const xam_datetime inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_CreateDatetime (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_datetime] %s", 
                        inValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_CreateDatetime == 0)
        {
            VIM_XSet_CreateDatetime = (XSet_CreateDatetimeFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_CreateDatetime");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_CreateDatetime != 0)
        {
            vStatus = VIM_XSet_CreateDatetime(inHandle, inName, inBinding, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_CreateDatetime");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_SetBoolean (const xset_handle inHandle,
                          const xam_string inName,
                          const xam_boolean inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_SetBoolean (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean] %s", 
                        inValue ? "true" : "false");
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_SetBoolean == 0)
        {
            VIM_XSet_SetBoolean = (XSet_SetBooleanFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_SetBoolean");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_SetBoolean != 0)
        {
            vStatus = VIM_XSet_SetBoolean(inHandle, inName, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_SetBoolean");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_SetInt (const xset_handle inHandle,
                      const xam_string inName,
                      const xam_int inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_SetInt (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_int] %li", 
                        inValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_SetInt == 0)
        {
            VIM_XSet_SetInt = (XSet_SetIntFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_SetInt");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_SetInt != 0)
        {
            vStatus = VIM_XSet_SetInt(inHandle, inName, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_SetInt");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_SetDouble (const xset_handle inHandle,
                        const xam_string inName,
                        const xam_double inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_SetDouble (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_double] %f", 
                        inValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_SetDouble == 0)
        {
            VIM_XSet_SetDouble = (XSet_SetDoubleFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_SetDouble");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_SetDouble != 0)
        {
            vStatus = VIM_XSet_SetDouble(inHandle, inName, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_SetDouble");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_SetXUID (const xset_handle inHandle,
                       const xam_string inName,
                       const xam_xuid inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_SetXUID (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_xuid] \"%s\"", 
                        inValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_SetXUID == 0)
        {
            VIM_XSet_SetXUID = (XSet_SetXUIDFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_SetXUID");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_SetXUID != 0)
        {
            vStatus = VIM_XSet_SetXUID(inHandle, inName, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_SetXUID");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_SetString (const xset_handle inHandle,
                         const xam_string inName,
                         const xam_string inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_SetString (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] %s", 
                        _logString(inValue));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_SetString == 0)
        {
            VIM_XSet_SetString = (XSet_SetStringFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_SetString");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_SetString != 0)
        {
            vStatus = VIM_XSet_SetString(inHandle, inName, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_SetString");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_SetDatetime (const xset_handle inHandle,
                           const xam_string inName,
                           const xam_datetime inValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_SetDatetime (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_datetime] %s", 
                        inValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_SetDatetime == 0)
        {
            VIM_XSet_SetDatetime = (XSet_SetDatetimeFunc) XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XSet_SetDatetime");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_SetDatetime != 0)
        {
            vStatus = VIM_XSet_SetDatetime(inHandle, inName, inValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_SetDatetime");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_GetBoolean (const xset_handle inHandle,
                          const xam_string inName,
                          xam_boolean* const outValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO,   
                        component_vim_enter, 
                        "VIM_XSet_GetBoolean (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG,
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG,
                        component_vim_param, 
                        "[in] [xam_boolean*] %p", 
                        outValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_GetBoolean == 0)
        {
            VIM_XSet_GetBoolean = (XSet_GetBooleanFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_GetBoolean");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_GetBoolean != 0)
        {
            vStatus = VIM_XSet_GetBoolean(inHandle, inName, outValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO,
                        component_vim_exit, 
                        "Exit VIM_XSet_GetBoolean");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_GetInt (const xset_handle inHandle,
                      const xam_string inName,
                      xam_int* const outValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_GetInt (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_int*] %p", 
                        outValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_GetInt == 0)
        {
            VIM_XSet_GetInt = (XSet_GetIntFunc)                 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_GetInt");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_GetInt != 0)
        {
            vStatus = VIM_XSet_GetInt(inHandle, inName, outValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_GetInt");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_GetDouble (const xset_handle inHandle,
                        const xam_string inName,
                        xam_double* const outValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_GetDouble (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_double*] %p", 
                        outValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_GetDouble == 0)
        {
            VIM_XSet_GetDouble = (XSet_GetDoubleFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_GetDouble");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_GetDouble != 0)
        {
            vStatus = VIM_XSet_GetDouble(inHandle, inName, outValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_GetDouble");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_GetXUID (const xset_handle inHandle,
                       const xam_string inName,
                       xam_xuid* const outValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_GetXUID (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_xuid*] %p", 
                        outValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_GetXUID == 0)
        {
            VIM_XSet_GetXUID = (XSet_GetXUIDFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_GetXUID");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_GetXUID != 0)
        {
            vStatus = VIM_XSet_GetXUID(inHandle, inName, outValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_GetXUID");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
} 

void 
XAM_VIM::XSet_GetString (const xset_handle inHandle,
                         const xam_string inName,
                         xam_string* const outValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_GetString (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string*] %p", 
                        outValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_GetString == 0)
        {
            VIM_XSet_GetString = (XSet_GetStringFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_GetString");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_GetString != 0)
        {
            vStatus = VIM_XSet_GetString(inHandle, inName, outValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_GetString");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_GetDatetime (const xset_handle inHandle,
                           const xam_string inName,
                           xam_datetime* const outValue)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_GetDatetime (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_datetime*] %p", 
                        outValue);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_GetDatetime == 0)
        {
            VIM_XSet_GetDatetime = (XSet_GetDatetimeFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_GetDatetime");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_GetDatetime != 0)
        {
            vStatus = VIM_XSet_GetDatetime(inHandle, inName, outValue);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_GetDatetime");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 


// =============================================================================
// Methods for XStreams and XStream fields
// =============================================================================

void 
XAM_VIM::XSystem_CreateXStream (const xsystem_handle inHandle,
                                const xam_string inName,
                                const xam_boolean inBinding,
                                const xam_string inType,
                                xstream_handle* const outXStream)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSystem_CreateXStream (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inType));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xstream_handle*] %p", 
                        outXStream);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_CreateXStream == 0)
        {
            VIM_XSystem_CreateXStream = (XSystem_CreateXStreamFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_CreateXStream");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_CreateXStream != 0)
        {
            vStatus = VIM_XSystem_CreateXStream(inHandle, 
                                                inName, 
                                                inBinding, 
                                                inType, 
                                                outXStream);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_CreateXStream");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xstream_handle] %"XAM_INT_FORMAT"d", *outXStream);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void
XAM_VIM::XSystem_OpenXStream (const xsystem_handle inHandle,
                              const xam_string inName,
                              const xam_string inMode,
                              xstream_handle* const outXStream)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSystem_OpenXStream (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xsystem_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inMode));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xstream_handle*] %p", 
                        outXStream);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_OpenXStream == 0)
        {
            VIM_XSystem_OpenXStream = (XSystem_OpenXStreamFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_OpenXStream");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_OpenXStream != 0)
        {
            vStatus = VIM_XSystem_OpenXStream(inHandle, 
                                              inName, 
                                              inMode, 
                                              outXStream);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSystem_OpenXStream");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xstream_handle] %"XAM_INT_FORMAT"d", *outXStream);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
}

void 
XAM_VIM::XSet_CreateXStream (const xset_handle inHandle,
                             const xam_string inName,
                             const xam_boolean inBinding,
                             const xam_string inType,
                             xstream_handle* const outXStream)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_CreateXStream (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inType));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xstream_handle*] %p", 
                        outXStream);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_CreateXStream == 0)
        {
            VIM_XSet_CreateXStream = (XSet_CreateXStreamFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_CreateXStream");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_CreateXStream != 0)
        {
            vStatus = VIM_XSet_CreateXStream(inHandle, 
                                             inName, 
                                             inBinding, 
                                             inType, 
                                             outXStream);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XSet_CreateXStream");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xstream_handle] %"XAM_INT_FORMAT"d", *outXStream);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void
XAM_VIM::XSet_OpenXStream (const xset_handle inHandle,
                           const xam_string inName,
                           const xam_string inMode,
                           xstream_handle* const outXStream)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XSet_OpenXStream (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xam_string] \"%s\"", 
                        _logString(inMode));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xstream_handle*] %p", 
                        outXStream);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_OpenXStream == 0)
        {
            VIM_XSet_OpenXStream = (XSet_OpenXStreamFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_OpenXStream");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_OpenXStream != 0)
        {
            vStatus = VIM_XSet_OpenXStream(inHandle, inName, inMode, outXStream);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_OpenXStream");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xstream_handle] %"XAM_INT_FORMAT"d", *outXStream);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
}

void 
XAM_VIM::XStream_Read (const xstream_handle inHandle,
                       char* const ioBuffer,
                       const xam_int inBufferLength,
                       xam_int* const outBytesRead)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_enter, 
                        "VIM_XStream_Read (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
                        component_vim_param, 
                        "[in] [xstream_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [const char*] %p", 
                        ioBuffer);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int] %li", 
                        inBufferLength);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int*] %p", 
                        outBytesRead);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XStream_Read == 0)
        {
            VIM_XStream_Read = (XStream_ReadFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XStream_Read");
        }

        // If valid, utilize the function pointer
        if(VIM_XStream_Read != 0)
        {
            vStatus = VIM_XStream_Read(inHandle, 
                                       ioBuffer, 
                                       inBufferLength, 
                                       outBytesRead);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
                        component_vim_exit, 
                        "Exit VIM_XStream_Read");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XStream_Write (const xstream_handle inHandle,
                        const char* inBuffer,
                        const xam_int inByteCount,
                        xam_int* const outBytesWritten)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XStream_Write (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xstream_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [const char*] %p", 
                        inBuffer);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int] %li", 
                        inByteCount);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int*] %p", 
                        outBytesWritten);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XStream_Write == 0)
        {
            VIM_XStream_Write = (XStream_WriteFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XStream_Write");
        }

        // If valid, utilize the function pointer
        if(VIM_XStream_Write != 0)
        {
            vStatus = VIM_XStream_Write(inHandle, 
                                        inBuffer, 
                                        inByteCount, 
                                        outBytesWritten);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XStream_Write");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XStream_Seek (const xstream_handle inHandle,
                       const xam_int inOffset,
                       const xam_int inWhence)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XStream_Seek (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xstream_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int] %li", 
                        inOffset);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int] %li", 
                        inWhence);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XStream_Seek == 0)
        {
            VIM_XStream_Seek = (XStream_SeekFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XStream_Seek");
        }

        // If valid, utilize the function pointer
        if(VIM_XStream_Seek != 0)
        {
            vStatus = VIM_XStream_Seek(inHandle, inOffset, inWhence);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XStream_Seek");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XStream_Tell (const xstream_handle inHandle,
                       xam_int* const outPosition)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XStream_Tell (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xstream_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int*] %p", outPosition);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XStream_Tell == 0)
        {
            VIM_XStream_Tell = (XStream_TellFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XStream_Tell");
        }

        // If valid, utilize the function pointer
        if(VIM_XStream_Tell != 0)
        {
            vStatus = VIM_XStream_Tell(inHandle, outPosition);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XStream_Tell");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XStream_Abandon (xstream_handle inHandle)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XStream_Abandon (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xstream_handle] %"XAM_INT_FORMAT"d", inHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XStream_Abandon == 0)
        {
            VIM_XStream_Abandon = (XStream_AbandonFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XStream_Abandon");
        }

        // If valid, utilize the function pointer
        if(VIM_XStream_Abandon != 0)
        {
            vStatus = VIM_XStream_Abandon(inHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XStream_Abandon");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XStream_Close (xstream_handle inHandle)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XStream_Close (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xstream_handle] %"XAM_INT_FORMAT"d", inHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XStream_Close == 0)
        {
            VIM_XStream_Close = (XStream_CloseFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XStream_Close");
        }

        // If valid, utilize the function pointer
        if(VIM_XStream_Close != 0)
        {
            vStatus = VIM_XStream_Close(inHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XStream_Close");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

// =============================================================================
// Methods for managing the connection to a XAM storage system
// =============================================================================

void 
XAM_VIM::CreateXSystem (xsystem_handle* const outHandle)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_CreateXSystem (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle*] %p", outHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_CreateXSystem == 0)
        {
            VIM_CreateXSystem = (VIM_CreateXSystemFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_CreateXSystem");
        }

        // If valid, utilize the function pointer
        if(VIM_CreateXSystem != 0)
        {
            vStatus = VIM_CreateXSystem(outHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }
    
    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_CreateXSystem");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xsystem_handle] %"XAM_INT_FORMAT"d", *outHandle);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_Connect (const xsystem_handle inHandle,
                          const xam_string inXRI)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_Connect (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inXRI));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_Connect == 0)
        {
            VIM_XSystem_Connect = (VIM_XSystem_ConnectFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XSystem_Connect");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_Connect != 0)
        {
            vStatus = VIM_XSystem_Connect(inHandle, inXRI);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }
    
    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_Connect");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_Authenticate (const xsystem_handle inHandle,
                               const char* inBuffer,
                               const xam_int inByteCount,
                               xstream_handle* const outXStream)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_Authenticate (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [const char*] %p", inBuffer);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int] %li", inByteCount);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xstream_handle*] %p", outXStream);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_Authenticate == 0)
        {
            VIM_XSystem_Authenticate = (XSystem_AuthenticateFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_Authenticate");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_Authenticate != 0)
        {
            vStatus = VIM_XSystem_Authenticate(inHandle, 
                                               inBuffer, 
                                               inByteCount, 
                                               outXStream);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_Authenticate");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xstream_handle] %"XAM_INT_FORMAT"d", *outXStream);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_Close (const xsystem_handle inHandle)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_Close (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_Close == 0)
        {
            VIM_XSystem_Close = (XSystem_CloseFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_Close");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_Close != 0)
        {
            vStatus = VIM_XSystem_Close(inHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_Close");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_Abandon (const xsystem_handle inHandle)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_Abandon (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_Abandon == 0)
        {
            VIM_XSystem_Abandon = (XSystem_AbandonFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_Abandon");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_Abandon != 0)
        {
            vStatus = VIM_XSystem_Abandon(inHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_Abandon");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

// ============================================================================
// Methods for XSet management
// ============================================================================

void 
XAM_VIM::XSystem_DeleteXSet (const xsystem_handle inHandle,
                                const xam_xuid inXUID)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_DeleteXSet (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_xuid] \"%s\"", inXUID);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_DeleteXSet == 0)
        {
            VIM_XSystem_DeleteXSet = (XSystem_DeleteXSetFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_DeleteXSet");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_DeleteXSet != 0)
        {
            vStatus = VIM_XSystem_DeleteXSet(inHandle, inXUID);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_DeleteXSet");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_HoldXSet (const xsystem_handle inHandle,
                              const xam_xuid inXUID,
                              const xam_string inHoldID)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_HoldXSet (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_xuid] \"%s\"", inXUID);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inHoldID));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_HoldXSet == 0)
        {
            VIM_XSystem_HoldXSet = (XSystem_HoldXSetFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_HoldXSet");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_HoldXSet != 0)
        {
            vStatus = VIM_XSystem_HoldXSet(inHandle, inXUID, inHoldID);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_HoldXSet");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_ReleaseXSet (const xsystem_handle inHandle,
                                 const xam_xuid inXUID,
                                 const xam_string inHoldID)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_ReleaseXSet (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_xuid] \"%s\"", inXUID);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inHoldID));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_ReleaseXSet == 0)
        {
            VIM_XSystem_ReleaseXSet = (XSystem_ReleaseXSetFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_ReleaseXSet");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_ReleaseXSet != 0)
        {
            vStatus = VIM_XSystem_ReleaseXSet(inHandle, inXUID, inHoldID);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_ReleaseXSet");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_AccessXSet (const xsystem_handle inHandle,
                             const xam_xuid inXUID,
                             const xam_int inMode,
                             xam_boolean* const outIsAccessible)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_AccessXSet (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_xuid] \"%s\"", inXUID);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int] %li", inMode);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean*] %p", outIsAccessible);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_AccessXSet == 0)
        {
            VIM_XSystem_AccessXSet = (XSystem_AccessXSetFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_AccessXSet");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_AccessXSet != 0)
        {
            vStatus = VIM_XSystem_AccessXSet(inHandle, 
                                             inXUID, 
                                             inMode, 
                                             outIsAccessible);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_AccessXSet");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
                        "[out] [xam_boolean] %s", *outIsAccessible ? "true" : "false");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_IsXSetRetained (const xsystem_handle inHandle,
                                 const xam_xuid inXUID,
                                 xam_boolean* const outIsRetained)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_IsXSetRetained (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_xuid] \"%s\"", inXUID);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean*] %p", outIsRetained);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_IsXSetRetained == 0)
        {
            VIM_XSystem_IsXSetRetained = (XSystem_IsXSetRetainedFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_IsXSetRetained");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_IsXSetRetained != 0)
        {
            vStatus = VIM_XSystem_IsXSetRetained(inHandle, 
                                                 inXUID, 
                                                 outIsRetained);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_IsXSetRetained");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
                        "[out] [xam_boolean] %s", *outIsRetained ? "true" : "false");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
}

void
XAM_VIM::XSystem_GetXSetAccessTime (const xsystem_handle inHandle,
                                    const xam_xuid inXUID,
                                    xam_datetime* const outAccessTime)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_GetXSetAccessTime (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_xuid] \"%s\"", inXUID);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_datetime*] %p", outAccessTime);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_GetXSetAccessTime == 0)
        {
            VIM_XSystem_GetXSetAccessTime = (XSystem_GetXSetAccessTimeFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_GetXSetAccessTime");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_GetXSetAccessTime != 0)
        {
            vStatus = VIM_XSystem_GetXSetAccessTime(inHandle, 
                                                    inXUID, 
                                                    outAccessTime);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_GetXSetAccessTime");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xam_datetime] %s", *outAccessTime);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
}

// ============================================================================
// Methods for creating an XSet
// ============================================================================

void 
XAM_VIM::XSystem_CreateXSet (const xsystem_handle inHandle,
                             const xam_string inMode,
                             xset_handle* const outXSet)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_CreateXSet (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] %s", _logString(inMode));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle*] %p", outXSet);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_CreateXSet == 0)
        {
            VIM_XSystem_CreateXSet = (XSystem_CreateXSetFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_CreateXSet");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_CreateXSet != 0)
        {
            vStatus = VIM_XSystem_CreateXSet(inHandle, inMode, outXSet);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_CreateXSet");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xset_handle] %"XAM_INT_FORMAT"d", *outXSet);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

// ============================================================================
// Methods for XSet policy management
// ============================================================================

void
XAM_VIM::XSet_ApplyAccessPolicy (const xset_handle inHandle,
                                 const xam_boolean inBinding,
                                 const xam_string inPolicy)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_ApplyAccessPolicy (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inPolicy));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_ApplyAccessPolicy == 0)
        {
            VIM_XSet_ApplyAccessPolicy = (XSet_ApplyAccessPolicyFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_ApplyAccessPolicy");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_ApplyAccessPolicy != 0)
        {
            vStatus = VIM_XSet_ApplyAccessPolicy(inHandle, 
                                                 inBinding, 
                                                 inPolicy);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_ApplyAccessPolicy");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
}

void
XAM_VIM::XSet_ResetAccessFields (const xset_handle inHandle)
{
  xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_ResetAccessFields (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", inHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_ResetAccessFields == 0)
        {
            VIM_XSet_ResetAccessFields = (XSet_ResetAccessFieldsFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_ResetAccessFields");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_ResetAccessFields != 0)
        {
            vStatus = VIM_XSet_ResetAccessFields(inHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_ResetAccessFields");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
}

void 
XAM_VIM::XSet_ApplyManagementPolicy (const xset_handle inHandle,
                                        const xam_boolean inBinding,
                                        const xam_string inPolicy)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_ApplyManagementPolicy (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inPolicy));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_ApplyManagementPolicy == 0)
        {
            VIM_XSet_ApplyManagementPolicy = (XSet_ApplyManagementPolicyFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_ApplyManagementPolicy");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_ApplyManagementPolicy != 0)
        {
            vStatus = VIM_XSet_ApplyManagementPolicy(inHandle, 
                                                     inBinding, 
                                                     inPolicy);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_ApplyManagementPolicy");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void
XAM_VIM::XSet_ResetManagementFields (const xset_handle inHandle)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_ResetManagementFields (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", inHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_ResetManagementFields == 0)
        {
            VIM_XSet_ResetManagementFields = (XSet_ResetManagementFieldsFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_ResetManagementFields");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_ResetManagementFields != 0)
        {
            vStatus = VIM_XSet_ResetManagementFields(inHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_ResetManagementFields");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
}

void
XAM_VIM::XSet_CreateRetention (const xset_handle inHandle,
                               const xam_boolean inBinding,
                               const xam_string inRetentionID)
{
  xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_CreateRetention (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", 
                        _logString(inRetentionID));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_CreateRetention == 0)
        {
            VIM_XSet_CreateRetention = (XSet_CreateRetentionFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_CreateRetention");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_CreateRetention != 0)
        {
            vStatus = VIM_XSet_CreateRetention(inHandle, 
                                               inBinding, 
                                               inRetentionID);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_CreateRetention");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
}

void
XAM_VIM::XSet_SetRetentionEnabledFlag (const xset_handle inHandle,
                                       const xam_string inRetentionID,
                                       const xam_boolean inBinding,
                                       const xam_boolean inEnabled)
{
  xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_SetRetentionEnabledFlag (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", 
                        _logString(inRetentionID));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", 
                        inEnabled ? "true" : "false");
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_SetRetentionEnabledFlag == 0)
        {
            VIM_XSet_SetRetentionEnabledFlag = (XSet_SetRetentionEnabledFlagFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_SetRetentionEnabledFlag");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_SetRetentionEnabledFlag != 0)
        {
            vStatus = VIM_XSet_SetRetentionEnabledFlag(inHandle,                                              
                                                       inRetentionID,
                                                       inBinding, 
                                                       inEnabled);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_SetRetentionEnabledFlag");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
}

void
XAM_VIM::XSet_ApplyRetentionEnabledPolicy (const xset_handle inHandle,
                                           const xam_string inRetentionID,
                                           const xam_boolean inBinding,
                                           const xam_string inPolicy)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_ApplyRetentionEnabledPolicy (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", 
                        _logString(inRetentionID));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", 
                        _logString(inPolicy));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_ApplyRetentionEnabledPolicy == 0)
        {
            VIM_XSet_ApplyRetentionEnabledPolicy = (XSet_ApplyRetentionEnabledPolicyFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_ApplyRetentionEnabledPolicy");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_ApplyRetentionEnabledPolicy != 0)
        {
            vStatus = VIM_XSet_ApplyRetentionEnabledPolicy(inHandle,                                              
                                                           inRetentionID,
                                                           inBinding, 
                                                           inPolicy);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_ApplyRetentionEnabledPolicy");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
}

void
XAM_VIM::XSet_SetRetentionDuration (const xset_handle inHandle,
                                    const xam_string inRetentionID,
                                    const xam_boolean inBinding,
                                    const xam_int inDuration)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_SetRetentionDuration (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", 
                        _logString(inRetentionID));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int] %"XAM_INT_FORMAT"d", 
                        inDuration);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_SetRetentionDuration == 0)
        {
            VIM_XSet_SetRetentionDuration = (XSet_SetRetentionDurationFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_SetRetentionDuration");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_SetRetentionDuration != 0)
        {
            vStatus = VIM_XSet_SetRetentionDuration(inHandle,                                              
                                                    inRetentionID,
                                                    inBinding, 
                                                    inDuration);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_SetRetentionDuration");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
}

void
XAM_VIM::XSet_ApplyRetentionDurationPolicy (const xset_handle inHandle,
                                            const xam_string inRetentionID,
                                            const xam_boolean inBinding,
                                            const xam_string inPolicy)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_ApplyRetentionDurationPolicy (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", 
                        _logString(inRetentionID));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", 
                        _logString(inPolicy));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_ApplyRetentionDurationPolicy == 0)
        {
            VIM_XSet_ApplyRetentionDurationPolicy = (XSet_ApplyRetentionDurationPolicyFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_ApplyRetentionDurationPolicy");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_ApplyRetentionDurationPolicy != 0)
        {
            vStatus = VIM_XSet_ApplyRetentionDurationPolicy(inHandle,                                              
                                                            inRetentionID,
                                                            inBinding, 
                                                            inPolicy);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_ApplyRetentionDurationPolicy");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }    
}

void
XAM_VIM::XSet_SetRetentionStarttime (const xset_handle inHandle,
                                     const xam_string inRetentionID,
                                     const xam_boolean inBinding)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_SetRetentionStarttime (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", 
                        _logString(inRetentionID));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_SetRetentionStarttime == 0)
        {
            VIM_XSet_SetRetentionStarttime = (XSet_SetRetentionStarttimeFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_SetRetentionStarttime");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_SetRetentionStarttime != 0)
        {
            vStatus = VIM_XSet_SetRetentionStarttime(inHandle,                                              
                                                     inRetentionID,
                                                     inBinding);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_SetRetentionStarttime");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }    
}

void
XAM_VIM::XSet_SetBaseRetention (const xset_handle inHandle,
                                const xam_boolean inBinding,
                                const xam_int inDuration)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_SetBaseRetention (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int] %"XAM_INT_FORMAT"d",
                        inDuration);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_SetBaseRetention == 0)
        {
            VIM_XSet_SetBaseRetention = (XSet_SetBaseRetentionFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_SetBaseRetention");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_SetBaseRetention != 0)
        {
            vStatus = VIM_XSet_SetBaseRetention(inHandle,                                              
                                                inBinding, 
                                                inDuration);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_SetBaseRetention");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }   
}

void
XAM_VIM::XSet_ApplyBaseRetentionPolicy (const xset_handle inHandle,
                                        const xam_boolean inBinding,
                                        const xam_string inPolicy)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_ApplyBaseRetentionPolicy (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", 
                        _logString(inPolicy));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_ApplyBaseRetentionPolicy == 0)
        {
            VIM_XSet_ApplyBaseRetentionPolicy = (XSet_ApplyBaseRetentionPolicyFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_ApplyBaseRetentionPolicy");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_ApplyBaseRetentionPolicy != 0)
        {
            vStatus = VIM_XSet_ApplyBaseRetentionPolicy(inHandle,                                              
                                                        inBinding, 
                                                        inPolicy);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_ApplyBaseRetentionPolicy");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }    
}

void 
XAM_VIM::XSet_ApplyAutoDeletePolicy (const xset_handle inHandle,
                                        const xam_boolean inBinding,
                                        const xam_string inPolicy)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_ApplyAutoDeletePolicy (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", 
                        _logString(inPolicy));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_ApplyAutoDeletePolicy == 0)
        {
            VIM_XSet_ApplyAutoDeletePolicy = (XSet_ApplyAutoDeletePolicyFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_ApplyAutoDeletePolicy");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_ApplyAutoDeletePolicy != 0)
        {
            vStatus = VIM_XSet_ApplyAutoDeletePolicy(inHandle, 
                                                     inBinding, 
                                                     inPolicy);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_ApplyAutoDeletePolicy");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_SetAutoDelete (const xset_handle inHandle,
                                const xam_boolean inBinding,
                                const xam_boolean inAutoDelete)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_SetAutoDelete (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", 
                        inAutoDelete ? "true" : "false");
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_SetAutoDelete == 0)
        {
            VIM_XSet_SetAutoDelete = (XSet_SetAutoDeleteFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_SetAutoDelete");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_SetAutoDelete != 0)
        {
            vStatus = VIM_XSet_SetAutoDelete(inHandle, 
                                             inBinding, 
                                             inAutoDelete);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_SetAutoDelete");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void
XAM_VIM::XSet_ApplyShredPolicy (const xset_handle inHandle,
                                const xam_boolean inBinding,
                                const xam_string inPolicy)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_ApplyShredPolicy (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", 
                        _logString(inPolicy));
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_ApplyShredPolicy == 0)
        {
            VIM_XSet_ApplyShredPolicy = (XSet_ApplyShredPolicyFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_ApplyShredPolicy");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_ApplyShredPolicy != 0)
        {
            vStatus = VIM_XSet_ApplyShredPolicy(inHandle, 
                                                inBinding, 
                                                inPolicy);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_ApplyShredPolicy");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
}

void
XAM_VIM::XSet_SetShred (const xset_handle inHandle,
                        const xam_boolean inBinding,
                        const xam_boolean inShred)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_SetShred (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", 
                        inShred ? "true" : "false");
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_SetShred == 0)
        {
            VIM_XSet_SetShred = (XSet_SetShredFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_SetShred");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_SetShred != 0)
        {
            vStatus = VIM_XSet_SetShred(inHandle, 
                                        inBinding, 
                                        inShred);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_SetShred");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
}

void 
XAM_VIM::XSet_ApplyStoragePolicy (const xset_handle inHandle,
                                     const xam_boolean inBinding,
                                     const xam_string inPolicy)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_ApplyStoragePolicy (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean] %s", 
                        inBinding ? "true" : "false");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", 
                        _logString(inPolicy));    
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_ApplyStoragePolicy == 0)
        {
            VIM_XSet_ApplyStoragePolicy = (XSet_ApplyStoragePolicyFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSet_ApplyStoragePolicy");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_ApplyStoragePolicy != 0)
        {
            vStatus = VIM_XSet_ApplyStoragePolicy(inHandle, 
                                                  inBinding, 
                                                  inPolicy);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_ApplyStoragePolicy");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_GetActualRetentionDuration (const xset_handle inHandle,
                                          const xam_string inRetentionID,
                                          xam_int* const outDuration)     
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_GetActualRetentionDuration (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", 
                        _logString(inRetentionID));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int*] %p", 
                        outDuration);  
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_GetActualRetentionDuration == 0)
        {
            VIM_XSet_GetActualRetentionDuration = 
                (XSet_GetActualRetentionDurationFunc) 
                    XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                        "VIM_XSet_GetActualRetentionDuration");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_GetActualRetentionDuration != 0)
        {
            vStatus = VIM_XSet_GetActualRetentionDuration(inHandle, 
                                                          inRetentionID,
                                                          outDuration);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_GetActualRetentionDuration");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void
XAM_VIM::XSet_GetActualRetentionEnabled (const xset_handle inHandle,
                                         const xam_string inRetentionID,
                                         xam_boolean* const outEnabled)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_GetActualRetentionEnabled (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", 
                        _logString(inRetentionID));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean*] %p", 
                        outEnabled); 
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_GetActualRetentionEnabled == 0)
        {
            VIM_XSet_GetActualRetentionEnabled = (XSet_GetActualRetentionEnabledFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                    "VIM_XSet_GetActualRetentionEnabled");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_GetActualRetentionEnabled != 0)
        {
            vStatus = VIM_XSet_GetActualRetentionEnabled(inHandle, 
                                                         inRetentionID,
                                                         outEnabled);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_GetActualRetentionEnabled");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
}

void 
XAM_VIM::XSet_GetActualAutoDelete (const xset_handle inHandle,
                                      xam_boolean* const outEnabled)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_GetActualAutoDelete (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean*] %p", 
                        outEnabled); 
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_GetActualAutoDelete == 0)
        {
            VIM_XSet_GetActualAutoDelete = (XSet_GetActualAutoDeleteFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                    "VIM_XSet_GetActualAutoDelete");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_GetActualAutoDelete != 0)
        {
            vStatus = VIM_XSet_GetActualAutoDelete(inHandle, 
                                                   outEnabled);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_GetActualAutoDelete");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 


void 
XAM_VIM::XSet_GetActualShred (const xset_handle inHandle,
                              xam_boolean* const outEnabled)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_GetActualShred (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean*] %p", 
                        outEnabled); 
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_GetActualShred == 0)
        {
            VIM_XSet_GetActualShred = (XSet_GetActualShredFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                    "VIM_XSet_GetActualShred");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_GetActualShred != 0)
        {
            vStatus = VIM_XSet_GetActualShred(inHandle, 
                                              outEnabled);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_GetActualShred");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

// ============================================================================
// Methods for editing an XSet
// ============================================================================

void 
XAM_VIM::XSystem_OpenXSet (const xsystem_handle inHandle,
                           const xam_xuid inXUID,
                           const xam_string inMode,
                           xset_handle* const outXSet)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_OpenXSet (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_xuid] \"%s\"", inXUID);  
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inMode));  
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle*] %p", outXSet); 
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_OpenXSet == 0)
        {
            VIM_XSystem_OpenXSet = (XSystem_OpenXSetFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_OpenXSet");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_OpenXSet != 0)
        {
            vStatus = VIM_XSystem_OpenXSet(inHandle, 
                                           inXUID, 
                                           inMode, 
                                           outXSet);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_OpenXSet");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xset_handle] %"XAM_INT_FORMAT"d", *outXSet);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSystem_CopyXSet (const xsystem_handle inHandle,
                           const xam_xuid inXUID,
                           const xam_string inMode,
                           xset_handle* const outXSet)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_CopyXSet (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_xuid] \"%s\"", inXUID);  
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inMode));  
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle*] %p", outXSet); 
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_CopyXSet == 0)
        {
            VIM_XSystem_CopyXSet = (XSystem_CopyXSetFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                                     "VIM_XSystem_CopyXSet");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_CopyXSet != 0)
        {
            vStatus = VIM_XSystem_CopyXSet(inHandle, 
                                           inXUID, 
                                           inMode, 
                                           outXSet);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_CopyXSet");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xset_handle] %"XAM_INT_FORMAT"d", *outXSet);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_Commit (const xset_handle inHandle,
                         xam_xuid* const outXUID)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_Commit (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_xuid*] %p", 
                        outXUID); 
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_Commit == 0)
        {
            VIM_XSet_Commit = (XSet_CommitFunc) 
            XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XSet_Commit");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_Commit != 0)
        {
            vStatus = VIM_XSet_Commit(inHandle, outXUID);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_Commit");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_Close (const xset_handle inHandle)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_Close (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", inHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_Close == 0)
        {
            VIM_XSet_Close = (XSet_CloseFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XSet_Close");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_Close != 0)
        {
            vStatus = VIM_XSet_Close(inHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_Close");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_Abandon (const xset_handle inHandle)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_Abandon (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", inHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_Abandon == 0)
        {
            VIM_XSet_Abandon = (XSet_AbandonFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XSet_Abandon");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_Abandon != 0)
        {
            vStatus = VIM_XSet_Abandon(inHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_Abandon");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

// ============================================================================
// Methods for XSet migration
// ============================================================================

void 
XAM_VIM::XSet_OpenExportXStream (const xset_handle inHandle,
                                 xstream_handle* const outXStream)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_OpenExportXStream (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xstream_handle*] %p", outXStream);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_OpenExportXStream == 0)
        {
            VIM_XSet_OpenExportXStream = (XSet_OpenExportXStreamFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                    "VIM_XSet_OpenExportXStream");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_OpenExportXStream != 0)
        {
            vStatus = VIM_XSet_OpenExportXStream(inHandle, outXStream);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_OpenExportXStream");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xstream_handle] %"XAM_INT_FORMAT"d", *outXStream);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_OpenImportXStream (const xset_handle inHandle,
                                 xstream_handle* const outXStream)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_OpenImportXStream (%s)", 
                        this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", 
                        inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xstream_handle*] %p", 
                        outXStream);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_OpenImportXStream == 0)
        {
            VIM_XSet_OpenImportXStream = (XSet_OpenImportXStreamFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                    "VIM_XSet_OpenImportXStream");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_OpenImportXStream != 0)
        {
            vStatus = VIM_XSet_OpenImportXStream(inHandle, outXStream);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_OpenImportXStream");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xstream_handle] %"XAM_INT_FORMAT"d", *outXStream);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
} 

// ============================================================================
// Methods for job control
// ============================================================================

void 
XAM_VIM::XSet_SubmitJob (const xset_handle inHandle)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_SubmitJob (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", inHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_SubmitJob == 0)
        {
            VIM_XSet_SubmitJob = (XSet_SubmitJobFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, 
                    "VIM_XSet_SubmitJob");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_SubmitJob != 0)
        {
            vStatus = VIM_XSet_SubmitJob(inHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_SubmitJob");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    }
} 

void 
XAM_VIM::XSet_HaltJob (const xset_handle inHandle)      
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_HaltJob (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", inHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_HaltJob == 0)
        {
            VIM_XSet_HaltJob = (XSet_HaltJobFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XSet_HaltJob");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_HaltJob != 0)
        {
            vStatus = VIM_XSet_HaltJob(inHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_HaltJob");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
} 


// ====================================================================
// method prototypes asynchronous i/o
// ====================================================================

void 
XAM_VIM::XSystem_AsyncOpenXSet (const xsystem_handle inHandle,
                                const xam_xuid inXUID,
                                const xam_string inMode,
                                const XOPID inXOPID,
                                xasync_callback inCallback,
                                xasync_handle* const outAsyncHandle)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_AsyncOpenXSet (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_xuid] %p", inXUID);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inMode));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [XOPID] %li", inXOPID);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_callback] %p", inCallback);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle*] %p", outAsyncHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_AsyncOpenXSet == 0)
        {
            VIM_XSystem_AsyncOpenXSet = (XSystem_AsyncOpenXSetFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XSystem_AsyncOpenXSet");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_AsyncOpenXSet != 0)
        {
            vStatus = VIM_XSystem_AsyncOpenXSet(inHandle,
                                                inXUID,
                                                inMode,
                                                inXOPID,
                                                inCallback,
                                                outAsyncHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_AsyncOpenXSet");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xasync_handle] %"XAM_INT_FORMAT"d", *outAsyncHandle);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}


void 
XAM_VIM::XSystem_AsyncCopyXSet (const xsystem_handle inHandle,
                                const xam_xuid inXUID,
                                const xam_string inMode,
                                const XOPID inXOPID,
                                xasync_callback inCallback,
                                xasync_handle* const outAsyncHandle)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_AsyncCopyXSet (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_xuid] %p", inXUID);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inMode));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [XOPID] %li", inXOPID);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_callback] %p", inCallback);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle*] %p", outAsyncHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_AsyncCopyXSet == 0)
        {
            VIM_XSystem_AsyncCopyXSet = (XSystem_AsyncCopyXSetFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XSystem_AsyncCopyXSet");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_AsyncCopyXSet != 0)
        {
            vStatus = VIM_XSystem_AsyncCopyXSet(inHandle,
                                                inXUID,
                                                inMode,
                                                inXOPID,
                                                inCallback,
                                                outAsyncHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_AsyncCopyXSet");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xasync_handle] %"XAM_INT_FORMAT"d", *outAsyncHandle);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

void 
XAM_VIM::XSystem_AsyncOpenXStream (const xsystem_handle inHandle,
                                   const xam_string inName,
                                   const xam_string inMode,
                                   const XOPID inXOPID,
                                   xasync_callback inCallback,
                                   xasync_handle* const outAsyncHandle)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSystem_AsyncOpenXStream (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xsystem_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inMode));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [XOPID] %li", inXOPID);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_callback] %p", inCallback);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle*] %p", outAsyncHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSystem_AsyncOpenXStream == 0)
        {
            VIM_XSystem_AsyncOpenXStream = (XSystem_AsyncOpenXStreamFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XSystem_AsyncOpenXStream");
        }

        // If valid, utilize the function pointer
        if(VIM_XSystem_AsyncOpenXStream != 0)
        {
            vStatus = VIM_XSystem_AsyncOpenXStream(inHandle,
                                                   inName,
                                                   inMode,
                                                   inXOPID,
                                                   inCallback,
                                                   outAsyncHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSystem_AsyncOpenXStream");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xasync_handle] %"XAM_INT_FORMAT"d", *outAsyncHandle);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

void 
XAM_VIM::XSet_AsyncOpenXStream (const xset_handle inHandle,
                                const xam_string inName,
                                const xam_string inMode,
                                const XOPID inXOPID,
                                xasync_callback inCallback,
                                xasync_handle* const outAsyncHandle)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_AsyncOpenXStream (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inName));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_string] \"%s\"", _logString(inMode));
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [XOPID] %li", inXOPID);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_callback] %p", inCallback);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle*] %p", outAsyncHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_AsyncOpenXStream == 0)
        {
            VIM_XSet_AsyncOpenXStream = (XSet_AsyncOpenXStreamFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XSet_AsyncOpenXStream");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_AsyncOpenXStream != 0)
        {
            vStatus = VIM_XSet_AsyncOpenXStream(inHandle,
                                                inName,
                                                inMode,
                                                inXOPID,
                                                inCallback,
                                                outAsyncHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_AsyncOpenXStream");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xasync_handle] %"XAM_INT_FORMAT"d", *outAsyncHandle);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

void 
XAM_VIM::XStream_AsyncRead (const xstream_handle inHandle,
                            char* const ioBuffer,
                            const xam_int inBufferLength,
                            const XOPID inXOPID,
                            xasync_callback inCallback,
                            xasync_handle* const outAsyncHandle)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XStream_AsyncRead (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xstream_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [char*] %p", ioBuffer);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int] %li", inBufferLength);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [XOPID] %li", inXOPID);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_callback] %p", inCallback);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle*] %p", outAsyncHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XStream_AsyncRead == 0)
        {
            VIM_XStream_AsyncRead = (XStream_AsyncReadFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XStream_AsyncRead");
        }

        // If valid, utilize the function pointer
        if(VIM_XStream_AsyncRead != 0)
        {
            vStatus = VIM_XStream_AsyncRead(inHandle,
                                            ioBuffer,
                                            inBufferLength,
                                            inXOPID,
                                            inCallback,
                                            outAsyncHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XStream_AsyncRead");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xasync_handle] %"XAM_INT_FORMAT"d", *outAsyncHandle);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

void 
XAM_VIM::XStream_AsyncWrite (const xstream_handle inHandle,
                             const char* inBuffer,
                             const xam_int inByteCount,
                             const XOPID inXOPID,
                             xasync_callback inCallback,
                             xasync_handle* const outAsyncHandle)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XStream_AsyncWrite (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xstream_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [char*] %p", inBuffer);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int] %li", inByteCount);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [XOPID] %li", inXOPID);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_callback] %p", inCallback);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle*] %p", outAsyncHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XStream_AsyncWrite == 0)
        {
            VIM_XStream_AsyncWrite = (XStream_AsyncWriteFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XStream_AsyncWrite");
        }

        // If valid, utilize the function pointer
        if(VIM_XStream_AsyncWrite != 0)
        {
            vStatus = VIM_XStream_AsyncWrite(inHandle,
                                             inBuffer,
                                             inByteCount,
                                             inXOPID,
                                             inCallback,
                                             outAsyncHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XStream_AsyncWrite");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xasync_handle] %"XAM_INT_FORMAT"d", *outAsyncHandle);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

void 
XAM_VIM::XSet_AsyncCommit (const xset_handle inHandle,
                           const XOPID inXOPID,
                           xasync_callback inCallback,
                           xasync_handle* const outAsyncHandle)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XSet_AsyncCommit (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [XOPID] %li", inXOPID);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_callback] %p", inCallback);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle*] %p", outAsyncHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XSet_AsyncCommit == 0)
        {
            VIM_XSet_AsyncCommit = (XSet_AsyncCommitFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XSet_AsyncCommit");
        }

        // If valid, utilize the function pointer
        if(VIM_XSet_AsyncCommit != 0)
        {
            vStatus = VIM_XSet_AsyncCommit(inHandle,
                                           inXOPID,
                                           inCallback,
                                           outAsyncHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XSet_AsyncCommit");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xasync_handle] %"XAM_INT_FORMAT"d", *outAsyncHandle);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

void 
XAM_VIM::XStream_AsyncClose (const xstream_handle inHandle,
                             const XOPID inXOPID,
                             xasync_callback inCallback,
                             xasync_handle* const outAsyncHandle)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XStream_AsyncClose (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xstream_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [XOPID] %li", inXOPID);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_callback] %p", inCallback);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle*] %p", outAsyncHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XStream_AsyncClose == 0)
        {
            VIM_XStream_AsyncClose = (XStream_AsyncCloseFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XStream_AsyncClose");
        }

        // If valid, utilize the function pointer
        if(VIM_XStream_AsyncClose != 0)
        {
            vStatus = VIM_XStream_AsyncClose(inHandle,
                                             inXOPID,
                                             inCallback,
                                             outAsyncHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XStream_AsyncClose");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xasync_handle] %"XAM_INT_FORMAT"d", *outAsyncHandle);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

// ====================================================================
// method prototypes for managing asyncronous operations
// ====================================================================

void 
XAM_VIM::XAsync_Halt (const xasync_handle inHandle)
{
    xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XAsync_Halt (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle] %"XAM_INT_FORMAT"d", inHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XAsync_Halt == 0)
        {
            VIM_XAsync_Halt = (XAsync_HaltFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XAsync_Halt");
        }

        // If valid, utilize the function pointer
        if(VIM_XAsync_Halt != 0)
        {
            vStatus = VIM_XAsync_Halt(inHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XAsync_Halt");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

void 
XAM_VIM::XAsync_IsComplete (const xasync_handle inHandle,
                            xam_boolean* const outIsComplete)
{
   xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XAsync_IsComplete (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_boolean*] %p", outIsComplete);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XAsync_IsComplete == 0)
        {
            VIM_XAsync_IsComplete = (XAsync_IsCompleteFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XAsync_IsComplete");
        }

        // If valid, utilize the function pointer
        if(VIM_XAsync_IsComplete != 0)
        {
            vStatus = VIM_XAsync_IsComplete(inHandle, outIsComplete);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XAsync_IsComplete");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
                        "[out] [xam_boolean] %s", *outIsComplete ? "true" : "false");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

void 
XAM_VIM::XAsync_GetStatus (const xasync_handle inHandle,
                           xam_status* const outStatus)
{
   xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XAsync_GetStatus (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_status*] %p", outStatus);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XAsync_GetStatus == 0)
        {
            VIM_XAsync_GetStatus = (XAsync_GetStatusFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XAsync_GetStatus");
        }

        // If valid, utilize the function pointer
        if(VIM_XAsync_GetStatus != 0)
        {
            vStatus = VIM_XAsync_GetStatus(inHandle, outStatus);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XAsync_GetStatus");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

void 
XAM_VIM::XAsync_GetXOPID (const xasync_handle inHandle,
                          XOPID* const outXOPID)
{
   xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XAsync_GetXOPID (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [XOPID*] %p", outXOPID);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XAsync_GetXOPID == 0)
        {
            VIM_XAsync_GetXOPID = (XAsync_GetXOPIDFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XAsync_GetXOPID");
        }

        // If valid, utilize the function pointer
        if(VIM_XAsync_GetXOPID != 0)
        {
            vStatus = VIM_XAsync_GetXOPID(inHandle, outXOPID);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XAsync_GetXOPID");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

void 
XAM_VIM::XAsync_GetXSet (const xasync_handle inHandle,
                         xset_handle* const outXSet)
{
   xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XAsync_GetXSet (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xset_handle*] %p", outXSet);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XAsync_GetXSet == 0)
        {
            VIM_XAsync_GetXSet = (XAsync_GetXSetFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XAsync_GetXSet");
        }

        // If valid, utilize the function pointer
        if(VIM_XAsync_GetXSet != 0)
        {
            vStatus = VIM_XAsync_GetXSet(inHandle, outXSet);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XAsync_GetXSet");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xset_handle] %"XAM_INT_FORMAT"d", *outXSet);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

void 
XAM_VIM::XAsync_GetXStream (const xasync_handle inHandle,
                            xstream_handle* const outXStream)
{
   xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XAsync_GetXStream (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xstream_handle*] %p", outXStream);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XAsync_GetXStream == 0)
        {
            VIM_XAsync_GetXStream = (XAsync_GetXStreamFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XAsync_GetXStream");
        }

        // If valid, utilize the function pointer
        if(VIM_XAsync_GetXStream != 0)
        {
            vStatus = VIM_XAsync_GetXStream(inHandle, outXStream);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XAsync_GetXStream");
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[out] [xstream_handle] %"XAM_INT_FORMAT"d", *outXStream);
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

void 
XAM_VIM::XAsync_GetXUID (const xasync_handle inHandle,
                         xam_xuid* const outXUID)
{
   xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XAsync_GetXUID (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_xuid*] %p", outXUID);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XAsync_GetXUID == 0)
        {
            VIM_XAsync_GetXUID = (XAsync_GetXUIDFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XAsync_GetXUID");
        }

        // If valid, utilize the function pointer
        if(VIM_XAsync_GetXUID != 0)
        {
            vStatus = VIM_XAsync_GetXUID(inHandle, outXUID);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XAsync_GetXUID");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

void 
XAM_VIM::XAsync_GetBytesRead (const xasync_handle inHandle,
                              xam_int* const outBytesRead)
{
   xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XAsync_GetBytesRead (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int*] %p", outBytesRead);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XAsync_GetBytesRead == 0)
        {
            VIM_XAsync_GetBytesRead = (XAsync_GetBytesReadFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XAsync_GetBytesRead");
        }

        // If valid, utilize the function pointer
        if(VIM_XAsync_GetBytesRead != 0)
        {
            vStatus = VIM_XAsync_GetBytesRead(inHandle, outBytesRead);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XAsync_GetBytesRead");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

void 
XAM_VIM::XAsync_GetBytesWritten (const xasync_handle inHandle,
                                 xam_int* const outBytesWritten)
{
   xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XAsync_GetBytesWritten (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle] %"XAM_INT_FORMAT"d", inHandle);
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xam_int*] %p", outBytesWritten);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XAsync_GetBytesWritten == 0)
        {
            VIM_XAsync_GetBytesWritten = (XAsync_GetBytesWrittenFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XAsync_GetBytesWritten");
        }

        // If valid, utilize the function pointer
        if(VIM_XAsync_GetBytesWritten != 0)
        {
            vStatus = VIM_XAsync_GetBytesWritten(inHandle, outBytesWritten);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XAsync_GetBytesWritten");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

void 
XAM_VIM::XAsync_Close (const xasync_handle inHandle)
{
   xam_status vStatus = (xam_status) XAM_UNKNOWN_ERROR;

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_enter,
						"VIM_XAsync_Close (%s)", this->getName());
        XAM_Logger::Log(LOGLEVEL_DEBUG, 
						component_vim_param,
						"[in] [xasync_handle] %"XAM_INT_FORMAT"d", inHandle);
    }

    if(mVimLibraryHandle == (void*)0)
    {        
        // The VIM library handle for the object is invalid.
        XAMVimNotFoundException e;
        throw e;
    }
    else
    {
        // Grab the function pointer if not already valid
        if(VIM_XAsync_Close == 0)
        {
            VIM_XAsync_Close = (XAsync_CloseFunc) 
                XAM_POSIX::xam_dlsym(mVimLibraryHandle, "VIM_XAsync_Close");
        }

        // If valid, utilize the function pointer
        if(VIM_XAsync_Close != 0)
        {
            vStatus = VIM_XAsync_Close(inHandle);
        }
        else
        {
            XAMSymbolNotFoundException e;
            throw e;
        }
    }

    if(XAM_Logger::isEnabled())
    {
        XAM_Logger::Log(LOGLEVEL_INFO, 
						component_vim_exit,
						"Exit VIM_XAsync_Close");
    }

    if(vStatus != (xam_status)0) 
    {
        XAM_Exception e(vStatus);
        throw e;
    } 
}

void
XAM_VIM::parseLibNameFromPath(xam_string* inPath, xam_string* const outName)
{
    xam_int vPathLen = XAM_POSIX::xam_strlen((char*)inPath);
    char* vStartLibNamePtr = (char*)inPath + vPathLen;

    // Strip the path off of the library name (if any)
    while(vStartLibNamePtr != (char*)inPath)
    {      
        vStartLibNamePtr--;
        if(*vStartLibNamePtr == XAM_FILE_SEP)
        {
            vStartLibNamePtr++;
            break;
        }
    }

    xam_int vSz = XAM_POSIX::xam_strlen(vStartLibNamePtr);
    XAM_POSIX::xam_strncpy((char*)outName, vStartLibNamePtr, vSz);
    (*outName)[vSz] = '\0';
}

XAM_VIM::XAM_VIM(const char* inLibName) :
    VIM_XSystem_GetErrorToken(0),
    VIM_XSystem_OpenFieldIterator(0),
    VIM_XSet_OpenFieldIterator(0),
    VIM_XIterator_HasNext(0),
    VIM_XIterator_Next(0),
    VIM_XIterator_Close(0), 
    VIM_XSystem_ContainsField(0),
    VIM_XSystem_GetFieldType(0),   
    VIM_XSystem_GetFieldLength(0),   
    VIM_XSystem_GetFieldReadOnly(0),   
    VIM_XSystem_DeleteField(0),  
    VIM_XSystem_CreateBoolean(0),   
    VIM_XSystem_CreateInt(0),    
    VIM_XSystem_CreateDouble(0),    
    VIM_XSystem_CreateXUID(0),    
    VIM_XSystem_CreateString(0),    
    VIM_XSystem_CreateDatetime(0),   
    VIM_XSystem_SetBoolean(0),    
    VIM_XSystem_SetInt(0),      
    VIM_XSystem_SetDouble(0),    
    VIM_XSystem_SetXUID(0),   
    VIM_XSystem_SetString(0),    
    VIM_XSystem_SetDatetime(0),    
    VIM_XSystem_GetBoolean(0),    
    VIM_XSystem_GetInt(0),     
    VIM_XSystem_GetDouble(0),
    VIM_XSystem_GetXUID(0), 
    VIM_XSystem_GetString(0),  
    VIM_XSystem_GetDatetime(0),   
    VIM_XSystem_CreateXStream(0), 
    VIM_XSystem_OpenXStream(0),  
    VIM_XSet_ContainsField(0),
    VIM_XSet_GetFieldLength(0),   
    VIM_XSet_GetFieldBinding(0), 
    VIM_XSet_GetFieldType(0), 
    VIM_XSet_GetFieldReadOnly(0),   
    VIM_XSet_SetFieldAsBinding(0),    
    VIM_XSet_SetFieldAsNonbinding(0),  
    VIM_XSet_DeleteField(0),  
    VIM_XSet_CreateBoolean(0),   
    VIM_XSet_CreateInt(0),    
    VIM_XSet_CreateDouble(0),    
    VIM_XSet_CreateXUID(0),    
    VIM_XSet_CreateString(0),    
    VIM_XSet_CreateDatetime(0),   
    VIM_XSet_SetBoolean(0),    
    VIM_XSet_SetInt(0),      
    VIM_XSet_SetDouble(0),    
    VIM_XSet_SetXUID(0),   
    VIM_XSet_SetString(0),    
    VIM_XSet_SetDatetime(0),    
    VIM_XSet_GetBoolean(0),    
    VIM_XSet_GetInt(0),     
    VIM_XSet_GetDouble(0),
    VIM_XSet_GetXUID(0), 
    VIM_XSet_GetString(0),  
    VIM_XSet_GetDatetime(0),   
    VIM_XSet_CreateXStream(0), 
    VIM_XSet_OpenXStream(0),  
    VIM_XStream_Read(0),   
    VIM_XStream_Write(0),  
    VIM_XStream_Seek(0),   
    VIM_XStream_Tell(0),   
    VIM_XStream_Close(0),
    VIM_XStream_Abandon(0),
    VIM_CreateXSystem(0),
    VIM_XSystem_Connect(0),
    VIM_XSystem_Authenticate(0),   
    VIM_XSystem_Close(0),    
    VIM_XSystem_Abandon(0),     
    VIM_XSystem_DeleteXSet(0),   
    VIM_XSystem_HoldXSet(0),    
    VIM_XSystem_ReleaseXSet(0),   
    VIM_XSystem_AccessXSet(0),  
    VIM_XSystem_IsXSetRetained(0),
    VIM_XSystem_GetXSetAccessTime(0),
    VIM_XSystem_CreateXSet(0),    
    VIM_XSet_ApplyAccessPolicy(0), 
    VIM_XSet_ResetAccessFields(0), 
    VIM_XSet_ApplyManagementPolicy(0),  
    VIM_XSet_ResetManagementFields(0),
    VIM_XSet_CreateRetention(0),
    VIM_XSet_SetRetentionEnabledFlag(0),
    VIM_XSet_ApplyRetentionEnabledPolicy(0),
    VIM_XSet_SetRetentionDuration(0),
    VIM_XSet_ApplyRetentionDurationPolicy(0),
    VIM_XSet_SetRetentionStarttime(0),
    VIM_XSet_SetBaseRetention(0),
    VIM_XSet_ApplyBaseRetentionPolicy(0),
    VIM_XSet_ApplyAutoDeletePolicy(0),  
    VIM_XSet_SetAutoDelete(0), 
    VIM_XSet_ApplyShredPolicy(0),  
    VIM_XSet_SetShred(0),   
    VIM_XSet_ApplyStoragePolicy(0),   
    VIM_XSet_GetActualRetentionDuration(0),
    VIM_XSet_GetActualRetentionEnabled(0),
    VIM_XSet_GetActualAutoDelete(0),  
    VIM_XSet_GetActualShred(0),
    VIM_XSystem_OpenXSet(0),   
    VIM_XSystem_CopyXSet(0),
    VIM_XSet_Commit(0),    
    VIM_XSet_Close(0),    
    VIM_XSet_Abandon(0),   
    VIM_XSet_OpenExportXStream(0),    
    VIM_XSet_OpenImportXStream(0),   
    VIM_XSet_SubmitJob(0),    
    VIM_XSet_HaltJob(0),
    VIM_XSystem_AsyncOpenXSet(0),
    VIM_XSystem_AsyncCopyXSet(0),
    VIM_XSystem_AsyncOpenXStream(0),
    VIM_XSet_AsyncOpenXStream(0),
    VIM_XStream_AsyncRead(0),
    VIM_XStream_AsyncWrite(0),
    VIM_XSet_AsyncCommit(0),
    VIM_XStream_AsyncClose(0),
    VIM_XAsync_Halt(0),
    VIM_XAsync_IsComplete(0),
    VIM_XAsync_GetStatus(0),
    VIM_XAsync_GetXOPID(0),
    VIM_XAsync_GetXSet(0),
    VIM_XAsync_GetXStream(0),
    VIM_XAsync_GetXUID(0),
    VIM_XAsync_GetBytesRead(0),
    VIM_XAsync_GetBytesWritten(0),
    VIM_XAsync_Close(0)
{
    mVimLibraryHandle = (void*)0;

    // Load the library
    if(!loadLibrary(inLibName))
    {
        // An error occurred while loading the library
        XAMVimNotFoundException e;
        throw e;
    }   
}

XAM_VIM::~XAM_VIM()
{
    // Unload the library
    closeLibrary();
}



